//=========================================================
//
// AI
//
//=========================================================

#include "pch.h"
#include <stdio.h>
#include <fstream>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"
#include "../Utils/Utils_Convert.h"
//#include "../Utils/NotXML.h"
#include "../Utils/Utils_Others.h"

#include <chrono>
#include <pthread.h>
#include <thread>

//// WIN API
#include <time.h>
#include <windows.h>
#include <process.h>

using namespace AI_Modules;

//=========================================================
// static
//=========================================================
//AI* AI::_self = NULL;

//=========================================================
// ルールはアプリケーション内で共通ですよね？ というつもり。
//=========================================================
Rule AI::_rule;
map<int, Card> AI::_cardList;
map<int, string> AI::_cardNameList;

// スレッドに号令をかけるための変数：static でいいの？ 
bool AI::_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure = false;   // これがtrueになったら全スレッド中止（どこかで詰みが発見されたときに使う）

int AI::_numberOfThread; // うーん、これもstatic？


//=========================================================
// コンストラクタ
//=========================================================
AI::AI() {

	//AI::_self = this;

	//_threadHandle     = NULL;

#if defined(_WIN32)

	// クリティカルセクション用意 WinAPI版
	InitializeCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );

#elif defined(__linux__)

	///// Linux版
	//_criticalSection_For_DigInto_Mate_Thread_Result = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
	_criticalSection_For_DigInto_Mate_Thread_Result = PTHREAD_RECURSIVE_MUTEX_INITIALIZER;

#endif
	

	///// 地形評価テーブル：無印
	int terrainEvalTable_Throuable[5][5] = {
		{50, 60, 70, 60, 50},
		{40, 80, 90, 80, 40},
		{30, 60, 100, 60, 30},
		{7, 10, 15, 10, 7},
		{5, 1, 2, 1, 5}
	};
	memcpy(_terrainEvalTable_Throuable, terrainEvalTable_Throuable, sizeof(terrainEvalTable_Throuable));

	///// 地形評価テーブル：前進
	int terrainEvalTable_GoUP[5][5] = {
		{30, 40, 60, 40, 30},
		{40, 60, 70, 60, 40},
		{10, 30, 80, 30, 10},
		{4, 7, 10, 7, 4},
		{3, 1, 2, 1, 3}
	};
	memcpy(_terrainEvalTable_GoUP, terrainEvalTable_GoUP, sizeof(terrainEvalTable_GoUP));

	///// 地形評価テーブル：後ろ向き
	int terrainEvalTable_GoDOWN[5][5] = {
		{-200, -300, -600, -300, -200},
		{-200, -400, -700, -400, -200},
		{10, 20, -800, 20, 10},
		{10, 30, 40, 30, 10},
		{8, 10, 20, 10, 8}
	};
	memcpy(_terrainEvalTable_GoDOWN, terrainEvalTable_GoDOWN, sizeof(terrainEvalTable_GoDOWN));

	///// 地形評価テーブル：右向き
	int terrainEvalTable_GoRIGHT[5][5] = {
		{-6, -8, -10, -8, -6},
		{-5, -10, -20, -10, -5},
		{-17, -25, -30, -25, -17},
		{-5, -10, -27, -10, -5},
		{-8, -17, -20, -17, -8}
	};
	memcpy(_terrainEvalTable_GoRIGHT, terrainEvalTable_GoRIGHT, sizeof(terrainEvalTable_GoRIGHT));

	///// 地形評価テーブル：左向き
	int terrainEvalTable_GoLEFT[5][5] = {
		{-6, -8, -10, -8, -6},
		{-5, -10, -20, -10, -5},
		{-17, -25, -30, -25, -17},
		{-5, -10, -27, -10, -5},
		{-8, -17, -20, -17, -8}
	};
	memcpy(_terrainEvalTable_GoLEFT, terrainEvalTable_GoLEFT, sizeof(terrainEvalTable_GoLEFT));

	///// 地形評価テーブル：パネルなし
	int terrainEvalTable_NotThrouable[5][5] = {
		{0, 0, 0, 0, 0},
		{0, 70, 80, 70, 0},
		{0, 0, -100, 0, 0},
		{0, 0, 0, 0, 0},
		{0, 0, 0, 0, 0}
	};
	memcpy(_terrainEvalTable_NotThrouable, terrainEvalTable_NotThrouable, sizeof(terrainEvalTable_NotThrouable));

	///// 地形評価：高さ重み値
	//_terrainEval_hightBias = 1.0f;
	_terrainEval_hightBias = 0.1f;

	// 地形探索用のメモリプール
	MemoryPool_ForTreeSearch_Setup( _memoryPool_ForTerrainSearch, 100 );

	_recursionBlock_Evaluate_Terrain = false;
	_isResign = false;
	
	_PC_InvokeCount = 0;
	_pieceCount_black = 0;
	_pieceCount_white = 0;

	// PCメソッド用の
	pillarList_Black_to_White_S = NULL;
	pillarList_White_to_Black_S = NULL;

	piecePositionList_Black = NULL;
	piecePositionList_White = NULL;

	_pillarMold_Black_S = NULL;
	_pillarMold_White_S = NULL;

	_pillarMold_Black_SizeList = NULL;
	_pillarMold_White_SizeList = NULL;

	// 柱の本数、柱内の要素数：黒用
	pillarListSizeMax_Black = 0;     // 柱型、の件数の最大値
	pillarListSize_Black = 0;        // 柱型、の件数
	pillarGridListSizeMax_Black = 0; // 柱内の要素数の最大値（盤サイズの縦幅に等しい）

	// 柱の本数、柱内の要素数：白用
	pillarListSizeMax_White = 0;     // 柱型、の件数の最大値
	pillarListSize_White = 0;        // 柱型、の件数
	pillarGridListSizeMax_White = 0; // 柱内の要素数の最大値（盤サイズの縦幅に等しい）

	PEE_Init();
	
	
	///// 論理距離測定調査用の通過済み経路記録リスト
	for ( unsigned int y = 0; y < 9; ++y ) {
		for ( unsigned int x = 0; x < 9; ++x ) {
			_passReminderYX_GetDistanceToGoal_Init[y][x] = 0;
		}
	}
	
	
	
}

//==========================================================
// 駒評価素子関連の値を初期設定
//==========================================================
void AI::PEE_Init() {

	///// これは固定値
	_heightValue_Goal = 9000;

	///// 初期設定値を読み込む
	//_parameters_PieceEvaluation = PEE_Read_File( "savePEE_Plain.txt" );
	//_parameters_PieceEvaluation = PEE_Read_File( "savePEE_Plain_安定90以上を目指す.txt" );
	_parameters_PieceEvaluation = PEE_Read_File( "savePEE_Plain_001.txt" );
	
}

//==========================================================
// 駒評価素子関連の値を初期設定
//==========================================================
void AI::PEE_Init( Parameters_PieceEvaluation parameters_PieceEvaluation ) {

	///// これは固定値
	_heightValue_Goal = 9000;

	///// 受け取る
	_parameters_PieceEvaluation = parameters_PieceEvaluation;
}

/*
//=========================================================
// GetInstance
//=========================================================
AI* AI::GetInstance() {

	AI *result = NULL;
	
	if (AI::_self == NULL) {
		result = new AI();
	} else {
		result = AI::_self;
	}
	
	return result;
}
*/

//=========================================================
// デストラクタ
//=========================================================
AI::~AI() {

	// 地形探索用のメモリプール後始末
	MemoryPool_ForTreeSearch_End ( _memoryPool_ForTerrainSearch );


#if defined(_WIN32)

	// クリティカルセクション、後始末
	DeleteCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );
	
#elif defined(__linux__)
	
	// なし。で、いいんでしたっけ？

#endif


	// PCメソッド用のメモリ、後始末
	pillarList_Black_to_White_S = NULL;
	pillarList_White_to_Black_S = NULL;

	piecePositionList_Black = NULL;
	piecePositionList_White = NULL;

	Evaluate_PC_S_Prepare_PillarMold_Clear();
}
//=========================================================
// ルールをセット
//=========================================================
void AI::SetRule( Rule rule, map<int, Card> cardList ) {

	_rule     = rule;
	_cardList = cardList;
}
void AI::SetCardNameList( map<int, string> cardNameList) {

	_cardNameList = cardNameList;
}

//=========================================================
// 次の一手を教えて
//=========================================================
Move AI::GetNextMove( InnerExpression *innerExpression ) {

	Move result;
	
	///// 現局面で可能な手を全て取得
	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList( allMoveList );

	//printf("開始\n");
	//long t =  timeGetTime();
	auto t = std::chrono::system_clock::now();

	/*
	vector<Move> getAllMoveList;
	innerExpression->GetLegalNextMoveList(getAllMoveList);
	printf("可能な手数：%d\n", getAllMoveList.size());

	///// 詰み判定そのもののベンチマーク
	for (int i = 0; i < 500; ++i) {

		vector<Move> mateProcedure;
		mateProcedure.clear();
		GetMateProcedure_Thread( mateProcedure, innerExpression, 1, true);

		if ( mateProcedure.size() > 0 && i%100 == 0) {
			printf("詰みあり：手数[%d]\n", mateProcedure.size());
			result = mateProcedure[0];
		}
	}
	*/
	/*
		GoNextを使って、インスタンスを新たに生成することを繰り返すケースと
		GoBackを使って、同一インスタンスを使い回すケースとで、
		実行速度を比較する
	*/
	/*
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(10000);
	
	vector<Move> mateProcedure;
	vector<Move> moveCandidateList;
	for ( unsigned int i = 0; i < allMoveList.size(); ++i) {
		
		innerExpression->GoNext( allMoveList[i] );

		mateProcedure.clear();
		GetMateProcedure_Thread( mateProcedure, innerExpression, 7, true);

		if (mateProcedure.size() <= 0) {
			moveCandidateList.push_back( allMoveList[i] );
		}
		// 巻き戻し
		if (!innerExpression->GoBack()) {
			printf("巻き戻し失敗！\n");
		}
		
		//InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], allocator );

		//mateProcedure.clear();
		//GetMateProcedure_Thread( mateProcedure, nextState, 7, true);

		//if (mateProcedure.size() <= 0) {
		//	moveCandidateList.push_back( allMoveList[i] );
		//}
		
	}
	printf("絞込み前の手数：%d  絞り込み後の手数：%d\n", allMoveList.size(), moveCandidateList.size());
	delete allocator;
	*/

	// ランダム版
	//result = GetNextMove_Random( innerExpression ); 

	// モンテカルロ版
	//result = GetNextMove_Monte( innerExpression );

	// 木探索版
	//result = GetNextMove_Tree( innerExpression );

	// 混合その１
	//result = GetNextMove_Composite( innerExpression );

	// 混合その２
	result = GetNextMove_Composite2( innerExpression );

	//long passTime = timeGetTime() - t;
	auto passTime = std::chrono::system_clock::now() - t;
	//printf( "%d ms\n", passTime );


	return result;
}

//==========================================================
// 次の一手：AI自己対戦用
//==========================================================
Move AI::GetNextMove_For_AI_SelfMatchLearning( InnerExpression *innerExpression ) {

	Move result;

	//printf("次の一手：AI自己対戦用\n");

	///// 現局面で可能な手を全て取得
	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList( allMoveList );

	///// 一手で勝つのがあればそれを選ぶ
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		if ( IsWinMove( allMoveList[i]) ) {
			
			result = allMoveList[i];
			return result;
		}
		// 先頭に「移動」がきてるはずなので、「移動」じゃなくなったら、見るのをやめる。
		if ( allMoveList[i].action != MOVE::ACTION::MOVE_PIECE ) {
			break;
		}
	}

	
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(1500);

	///// 一番評価の高い手を覚えておくのです。
	int maxEval = 0;

	vector<Move> bestMoveList;

	///// 次の一手でループ
	for (unsigned int i = 0; i < allMoveList.size(); ++i) 
	{

		///// 次の局面
		InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], allocator );

		///// 駒の価値だけで局面を評価：AIの学習対象、という意味
		int eval = Evaluate_Piece( nextState, innerExpression->_currentTurn );

		///// 地形評価もしてみる。
		//eval += Evaluate_Terrain( nextState, innerExpression->_currentTurn, 3);
		
		///// 王手評価も
		//eval += Evaluate_Check( nextState, innerExpression->_currentTurn );

		///// 一番いい手？
		if ( i == 0 || eval > maxEval) {
			maxEval = eval;
			bestMoveList.clear();
		}
		if (eval == maxEval) {
			bestMoveList.push_back( allMoveList[i] );
		}
	}

	delete allocator;

	///// 結果
	if ( bestMoveList.size() > 0 ) {
		int lotNum = Utils_Cal::GerRandom( bestMoveList.size());
		result = bestMoveList[lotNum];
	} else {
		printf("あれ？ なんかヘンですよ？\n");
	}

	return result;
}

//==========================================================
// 次の一手：AI自己対戦用＿学習者用
//==========================================================
Move AI::GetNextMove_For_AI_SelfMatchLearning_ForCandidate( InnerExpression *innerExpression )  {
	return GetNextMove_For_AI_SelfMatchLearning( innerExpression );
}
//==========================================================
// 次の一手：AI自己対戦用＿対戦相手用
//==========================================================
Move AI::GetNextMove_For_AI_SelfMatchLearning_ForOpponent( InnerExpression *innerExpression )  {

	Move result;

	//printf("次の一手：AI自己対戦用\n");

	///// 現局面で可能な手を全て取得
	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList( allMoveList );

	///// 一手で勝つのがあればそれを選ぶ
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		if ( IsWinMove( allMoveList[i]) ) {
			
			result = allMoveList[i];
			return result;
		}
		// 先頭に「移動」がきてるはずなので、「移動」じゃなくなったら、見るのをやめる。
		if ( allMoveList[i].action != MOVE::ACTION::MOVE_PIECE ) {
			break;
		}
	}

	
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(1500);

	///// 一番評価の高い手を覚えておくのです。
	int maxEval = 0;

	vector<Move> bestMoveList;

	///// 次の一手でループ
	for (unsigned int i = 0; i < allMoveList.size(); ++i) 
	{

		///// 次の局面
		InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], allocator );

		///// 駒の価値だけで局面を評価：AIの学習対象、という意味
		int eval = Evaluate_Piece( nextState, innerExpression->_currentTurn );

		///// 地形評価もしてみる。
		//eval += Evaluate_Terrain( nextState, innerExpression->_currentTurn, 1);

		///// 王手評価もしてみる
		//eval += Evaluate_Check( nextState, innerExpression->_currentTurn);

		///// 一番いい手？
		if ( i == 0 || eval > maxEval) {
			maxEval = eval;
			bestMoveList.clear();
		}
		if (eval == maxEval) {
			bestMoveList.push_back( allMoveList[i] );
		}
	}

	delete allocator;

	///// 結果
	if ( bestMoveList.size() > 0 ) {
		int lotNum = Utils_Cal::GerRandom( bestMoveList.size());
		result = bestMoveList[lotNum];
	} else {
		printf("あれ？ なんかヘンですよ？\n");
	}

	return result;
}

//==========================================================
// PEEファイルの実験用：手作業でいじりながら効果を確かめるための。
//==========================================================
void AI::GetNextMove_PEE_Test ( InnerExpression *innerExpression, ProgressData &progressData ) {
	
	///// 開始
progressData.progressRate = 0.0f;
	progressData.messageLogs.push_back ( "思考開始" );
	_isResign = false;
	
	// 前回の「直近候補手」をクリアしておく
	_evaluatedMoveList_theLatest.clear();
	
	// 所要ms計測
	//auto t = std::chrono::system_clock::now();
	clock_t t = clock();
	
	
	///// モンテカルロ試し
	/*
	Move moveFromMonte = GetNextMove_Monte ( innerExpression );
	progressData.answerMove = moveFromMonte;
	progressData.progressRate = 100.0f;
	progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::EVEN;
	
	//auto passTime = std::chrono::system_clock::now() - t;
	clock_t passTime0 =  clock() - t;
	printf ("----- AI Think Process Done.\n");
	printf( "----- %d ms\n", passTime0);
	
	return;
	*/

	Move result;

	///// PCメソッドを使う設定なら、柱の型を更新。
	if ( _parameters_PieceEvaluation.pcEvalInfo.isEnabled ) {
		//Evaluate_PC_Prepare_PillarMold();
		Evaluate_PC_S_Prepare_PillarMold();
		_PC_InvokeCount = 0;
	}
	
	///// 駒の個数（手駒＋盤上の個数）を確認：駒の評価判定の中で、個数分を見終えたらループから打ち切るためのもの。

	_pieceCount_black = innerExpression->_handPiece_BLACK;
	_pieceCount_white = innerExpression->_handPiece_WHITE;

	for (int y = 0; y < _rule.boardHeight; ++y) {
		for (int x = 0; x < _rule.boardWidth; ++x) {
		
			if ( innerExpression->_gridList_Piece[y][x] ==  GRID::BLACK) {
				++_pieceCount_black;
			}
			else if ( innerExpression->_gridList_Piece[y][x] ==  GRID::WHITE) {
				++_pieceCount_white;
			}
		}
	}
	
	printf ( "NumberOfPiece BLACK:[%d], WHITE:[%d]\n", _pieceCount_black, _pieceCount_white );

progressData.progressRate = 5.0f;
	if ( progressData.isStop ) { _recordList.clear(); return; } // 停止可能ポイント

	///// 現局面で可能な手を全て取得
	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList( allMoveList );
	
	
	// 千日手になる手を省く
	/*
		AIはルールに関係なく、自主的に千日手を避けるってことでいいんじゃないですか？ 
		うん。そんな気がしてきました。
		というわけでここではルールを見ずに千日手を避けています。
	*/
	printf ( "千日手になる手を省く これまでの手数：%d\n", _recordList.size() );
	vector<Move> allMoveList_examined;
	vector<Move> repetitionMoveList;
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {
		
		InnerExpression *nextState = innerExpression->GetNext (allMoveList[i]);
		
		
		bool isExist = false;
		for ( unsigned int n = 0; n < _recordList.size(); ++n) {
			
			// 形だけ比較するので、対象外の値をそろえてしまう
			nextState->_currentTurn           = _recordList[n]->_currentTurn;
			nextState->_totalMoveCount        = _recordList[n]->_totalMoveCount;
			nextState->_isThrouableInfoLatest = _recordList[n]->_isThrouableInfoLatest;
			
			if ( nextState->IsEquals( _recordList[n] ) ) {
				isExist = true;
				break;
			}
		}
		if ( isExist ) {
			repetitionMoveList.push_back ( allMoveList[i] ); 
			printf ( "千日手になる手 %d\n", allMoveList[i].action );
		} else {
			
			allMoveList_examined.push_back ( allMoveList[i] );
		}

		delete nextState;
	}
	allMoveList = allMoveList_examined;
	
	

progressData.progressRate = 10.0f;
	if ( progressData.isStop ) { _recordList.clear(); return; } // 停止可能ポイント
	
	// この時点で合法的な手がない場合、ここで思考停止とする
	if ( allMoveList.size() == 0 ) {
		
		/*
			手詰まりのルール上の扱いについては、要検討事項とします。
			
			後日追記：手詰まりは負けということになりました。試合進行管理側で面倒を見ているので、ここではこれでよし。
		*/
		
		printf ( "手詰まりなので思考終了\n" );
		
		/*
			本当に手詰まりならば、そもそもこの思考処理が呼び出されることはないはず。
			ここで手詰まりとなる場合があるとすれば、千日手になる手を自主的に省いているから。
			その意味で「投了」とする
		*/
		_isResign = true;
	
progressData.progressRate = 100.0f;
		progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::LOSE_MAX;
		_recordList.clear(); return;
	}

	// この時点で一手しかない場合、ここで思考停止とする
	if ( allMoveList.size() == 1 ) {
			
		printf ( "一手しかなかったので思考終了\n" );
	
progressData.progressRate = 100.0f;
		progressData.answerMove = allMoveList[0];
		progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::UNKNOWN;
		
		// 選択肢バトル向けのメモ作成：vector< pair<Move, int> > _evaluatedMoveList_theLatest;
		pair<Move, int> onlyCandidate;
		onlyCandidate.first  = progressData.answerMove;
		onlyCandidate.second = 0;
		_evaluatedMoveList_theLatest.push_back ( onlyCandidate );
		
		_recordList.clear(); return;
	}

	// 手の数
	int allMoveListSize_Original = allMoveList.size();
	

	///// 絞り込み：戻り値が空じゃない＝詰みがあった。
	vector<Move> mateProcedure = GetNextMove_PEE_Test_PreSearch( allMoveList, innerExpression, progressData );

progressData.progressRate = 20.0f;
	if ( progressData.isStop ) { _recordList.clear(); return; } // 停止可能ポイント

	// 手の数＿絞込み後
	int allMoveListSize_AfterPicked = allMoveList.size();

	///// 一手だけになった
	if ( allMoveListSize_AfterPicked == 1 ) {

		/*
			ここで、外部向け総括評価値は、WIN_MAX か WIN_4 となっている、というわけですね。
			絞込み処理の中で、progressData を入れてくれているはず。
		*/
progressData.progressRate = 100.0f;
		progressData.answerMove = allMoveList[0];
		
		// 選択肢バトル向けのメモ作成：vector< pair<Move, int> > _evaluatedMoveList_theLatest;
		pair<Move, int> onlyCandidate;
		onlyCandidate.first  = progressData.answerMove;
		onlyCandidate.second = 0;
		_evaluatedMoveList_theLatest.push_back ( onlyCandidate );

		_recordList.clear(); return;
	}

	///// 相手からの詰め手順があれば妨害する手を探す
	vector<Move> moveList_MateGuard = GetNextMove_PEE_Test_DisturbMated ( innerExpression );
	/*
		ここで、妨害する手が返ってきた＝相手からの詰みがある、
	*/
	
	
	///// 明らかに不利になる手を省く。つまり、矢印を自陣側に向けて設置する手。
	/*
		確かに手数は減らせますが、微差である上に、終盤にはこんな大雑把に省くわけにいかず、また、序盤であればそもそも処理は軽め。
		まともに実装するなら、「序盤何手まではこの方法で手数を減らす」のようなことを設定ファイルで設定できるようにしたいですが、
		あんまり意味のない軽量化だと思うと気力がわかないので、この処理自体をボツとします。
	*/
	/*
	vector<Move> allMoveList_withoutSuicideMoves;
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {
		
		Move m = allMoveList[i];
		
		bool isNGMove = false;
		
		if ( m.action == MOVE::ACTION::PUT_CARD ) {
			
			if      ( innerExpression->_currentTurn == TURN::BLACK && _cardNameList [ m.putCardId ] == "iinogare" && m.putCardDirection == DIRECTION::DOWN ) { isNGMove = true; }
			else if ( innerExpression->_currentTurn == TURN::BLACK && _cardNameList [ m.putCardId ] == "keno"     && m.putCardDirection == DIRECTION::UP   ) { isNGMove = true; }
			else if ( innerExpression->_currentTurn == TURN::WHITE && _cardNameList [ m.putCardId ] == "iinogare" && m.putCardDirection == DIRECTION::UP   ) { isNGMove = true; }
			else if ( innerExpression->_currentTurn == TURN::WHITE && _cardNameList [ m.putCardId ] == "keno"     && m.putCardDirection == DIRECTION::DOWN ) { isNGMove = true; }
			
		}
		
		if ( !isNGMove ) {
			allMoveList_withoutSuicideMoves.push_back ( m );
		}
	}
	printf ( "Reduce Deprecated Moves : allMoveList[%d] -> allMoveList[%d]\n", allMoveList.size(), allMoveList_withoutSuicideMoves.size() );
	allMoveList = allMoveList_withoutSuicideMoves;
	*/

progressData.progressRate = 25.0f;
	if ( progressData.isStop ) { _recordList.clear(); return; } // 停止可能ポイント

	///// 以降、評価関数を使う。
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(1500);

	///// ここらでシャッフルしておく。同率のときにマンネリにならないように。
	UTILS_OTHERS::RandomShuffle<Move>( allMoveList );

	///// 手に評価値をつける。後でソートして上位のを使う。
	vector< pair<Move, int> > evaluatedMoveList;

	///// 手札の枚数を数えておく
	int myHandCardsNumber = (innerExpression->_currentTurn == TURN::BLACK) ? innerExpression->_handCards_BLACK_Index : innerExpression->_handCards_WHITE_Index;
	
	///// 圧力判定用の値を準備しておく
	if ( _parameters_PieceEvaluation.parameters_PressureEvaluation.isEnabled ) {
		Evaluate_Move_By_PressureBalance_Init ( innerExpression );
	}
	
	///// 圧力判定を使うかどうか
	bool isPressureEvaluationEnabled = false;
	if ( _parameters_PieceEvaluation.parameters_PressureEvaluation.isEnabled ) {
		
		printf (" _parameters_PieceEvaluation.parameters_PressureEvaluation.isEnabled : true\n ");
		
		if ( Evaluate_Move_By_PressureBalance_IsWithinTheGameProgress () ) {
			
			isPressureEvaluationEnabled = true;
			
			printf ( "Pressure Balance to use.\n" );
		} else {
			printf ( "Pressure Balance not to use.\n" );
		}
	}
	
	
	///// 距離比べ判定での絞り込み
	int distanceDiffNarrowDownSize = allMoveList.size();
	map<string, InnerExpression*> nextStateList;
	
	if ( Evaluate_Move_by_Distance_isScreeningEnabled ( innerExpression ) ) {
		
		DistanceDiffScreening dsParams = _parameters_PieceEvaluation.parameters_DistanceToGoal.distanceDiffScreening;
		
		distanceDiffNarrowDownSize = dsParams.screeningLength;
		Evaluate_Move_by_Distance_Init ( innerExpression );
		
		///// この時点では、evaluatedMoveList がまだ空ってことでいいですよね？ いいハズ。
		
		///// 次の一手でループ
		for (unsigned int i = 0; i < allMoveList.size(); ++i) 
		{
			progressData.progressRate = 25.0f + ( 25.0f * i / allMoveList.size());
			if ( progressData.isStop ) { delete allocator; _recordList.clear(); return; } // 停止可能ポイント

			///// 次の局面
			InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], allocator );
			
			string moveUString = moveToUSTRING ( allMoveList[i] );
			
			// この後、もう一回ループをまわして毎回次の局面を求めるのは避けたいので、ここで覚えておくこととする。
			nextStateList [ moveUString ] = nextState;
			
			int eval = 0;
			
			///// 距離比べ判定
			//int distDifEval = Evaluate_Move_by_Distance (innerExpression, allMoveList[i], nextState, innerExpression->_currentTurn ); 
			int distDifEval = Evaluate_DistanceDiff ( nextState, innerExpression->_currentTurn ); 
			eval += distDifEval;
			
			
			///// 評価値をつけて、リストに入れておく。あとでソート
			pair<Move, int> evaluatedMove;
			evaluatedMove.first  = allMoveList[i];
			evaluatedMove.second = eval;
			
			evaluatedMoveList.push_back( evaluatedMove );
		}
		
		///// ここで一度ソート
		GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate ( evaluatedMoveList );
		
		//printf ( "okay? allMoveList.size() = %d, evaluatedMoveList.size() = %d\n", allMoveList.size(), evaluatedMoveList.size() );
		
		///// allMoveList も同じ並び順にしておく
		allMoveList.clear();
		
		///// 果たして、この時点で点数がついている（絞り込めている）か？
		bool isScreeningValid = false;
		
		for ( unsigned int i = 0; i < evaluatedMoveList.size(); ++i ) {
			
			if ( evaluatedMoveList[i].second != 0 ) {
				isScreeningValid = true;
			}
			
			allMoveList.push_back ( evaluatedMoveList[i].first );
		}
		
		
		//// 絞込み：上位n件以降のポイントを潰しておく
		if ( isScreeningValid ) {
			
			printf ( "DISTANCE DIFF SCREENING succeeded.\n" );
			
			for ( unsigned int i = distanceDiffNarrowDownSize; i < evaluatedMoveList.size(); ++i ) {
				
				evaluatedMoveList[i].second = -99999;
			}
			
		// 絞り込めませんでした
		} else {
			printf ( "DISTANCE DIFF SCREENING failed.\n" );
			distanceDiffNarrowDownSize = allMoveList.size();
		}
		
		
	}
	
progressData.progressRate = 50.0f;

	
	///// 次の一手でループ
	for (unsigned int i = 0; i < allMoveList.size(); ++i) 
	{
		if ( i >= distanceDiffNarrowDownSize ) {
			printf ( "Break for distanceDiffNarrowDownSize[%d]\n", distanceDiffNarrowDownSize );
			break;
		}
		
//progressData.progressRate = 25.0f + ( 45.0f * i / allMoveList.size());
progressData.progressRate = 50.0f + ( 20.0f * i / distanceDiffNarrowDownSize );
		if ( progressData.isStop ) { delete allocator; _recordList.clear(); return; } // 停止可能ポイント

		///// 次の局面
		InnerExpression *nextState = NULL;
		
		if ( nextStateList.size() == 0 ) {
			
			// 絞り込んでいなかった場合
			nextState = innerExpression->GetNext( allMoveList[i], allocator );
			
		} else {
			
			// 絞り込んでいた場合
			string moveUString = moveToUSTRING ( allMoveList[i] );
			nextState = nextStateList [ moveUString ];
		}
		
		
		int eval = 0;
		
		/*
		// あんまり実用性を引き出せない感じなので、さしあたりモンテカルロはお蔵入りとします。
		// 実用性がないというのは、つまり、処理が重く、そのわりに強くもならない、という意味。
		// モンテカルロのトライアウトをするかどうか
		if ( true ) {
			
			float winRateByMonte = GetWinRateByMonte ( innerExpression, allMoveList[i] );
			
			int biasForMonte = 100;
			
			eval += winRateByMonte * biasForMonte;
		}
		*/
		
		// 半モンテ
		// 重すぎるので断念
		/*
		//if ( true ) {
		if ( distanceDiffNarrowDownSize < 300 ) {	// 絞り込んでいた場合のみ
		
			int semiMonteEval = Evaluate_SemiMonte ( nextState, 3, innerExpression->_currentTurn );
			
			eval += semiMonteEval;
		}
		*/
		
		
		// 序盤専用フラグ
		bool isOpeningProccessOnly = isPressureEvaluationEnabled;
		
		// 圧力判定を使いますか？
		if ( isOpeningProccessOnly ) {
			int pressureBalanceEval = Evaluate_Move_By_PressureBalance ( innerExpression, allMoveList[i], nextState, innerExpression->_currentTurn ); 
			
			eval += pressureBalanceEval;
			
			// 判定結果がゼロの場合、他の処理への抑止フラグは解除しておく
			if ( pressureBalanceEval == 0 ) {
				isOpeningProccessOnly = 0;
			}
		}
		
		// 駒の距離判定を使いますか？
		if ( _parameters_PieceEvaluation.parameters_DistanceToGoal.isEnabled ) {
			
			int distanceEval = Evaluate_DistanceToGoal ( nextState, innerExpression->_currentTurn );
			eval += distanceEval;
			
		}
		
		
		// 手筋評価を使いますか？
		if ( _parameters_PieceEvaluation.handyProcList.size() > 0 ) {
			
			//printf ( "HandyProc(s) are set : num[ %d ]\n", _parameters_PieceEvaluation.handyProcList.size() );
			
			// 現在の局面と、候補手と、次の局面を渡して、加算される評価得点が返ってくる。
			int handyProcEval = Evaluate_HandyProc ( innerExpression, allMoveList[i], nextState );
			
			eval += handyProcEval;
			
			//if ( handyProcEval > 0 ) {
			//	printf ( "Get handyProcEval [%d] -> eval[%d]\n", handyProcEval, eval );
			//}
		}
		
		
		// 傘と柱だけ
		//eval = Evaluate_PC_S ( nextState, innerExpression->_currentTurn );

		///// 地形評価を使う？
		int terrainDepth = _parameters_PieceEvaluation.terrainSearchUsage.depth;
		if ( terrainDepth >= 0 && !isOpeningProccessOnly ) {
		
		
			// 地形評価のバイアスが 0 以上ならば、使う。
			if ( _parameters_PieceEvaluation.terrainSearchUsage.bias_TerrainEvaluate > 0 ) {
			
				//printf("地形評価開始 バイアス[%d]\n", _parameters_PieceEvaluation.terrainSearchUsage.bias_TerrainEvaluate);

				// 地形評価の末端で使っている評価関数が、まさにその「駒の価値評価」
				// したがって、地形評価を使う場合には、駒評価を重ねて使う必要はないのではないか？ そう思う場合は設定ファイルでバイアス０にしてもらうこととする。
				
				int evalTerrain = Evaluate_Terrain( nextState, innerExpression->_currentTurn, terrainDepth);

				// バイアス
				evalTerrain *= (float)_parameters_PieceEvaluation.terrainSearchUsage.bias_TerrainEvaluate / 100.0f;
				
				eval += evalTerrain;
			}
			
			// 駒評価のバイアスが 0 以上ならば、使う。
			if ( _parameters_PieceEvaluation.terrainSearchUsage.bias_PeeEvaluate > 0 ) {
			
				//printf("駒評価開始 バイアス[%d]\n",_parameters_PieceEvaluation.terrainSearchUsage.bias_PeeEvaluate);
				
				int evalPEE = Evaluate_Piece( nextState, innerExpression->_currentTurn );
				
				// バイアス
				evalPEE *= (float)_parameters_PieceEvaluation.terrainSearchUsage.bias_PeeEvaluate / 100.0f;

				eval += evalPEE;
			}

		///// 地形評価を使わない(深度[-1]の場合)（isOpeningProccessOnlyでも、PEEの平適用はする、ということなのですね？）
		} else {

			///// 駒の価値だけで局面を評価
			eval += Evaluate_Piece( nextState, innerExpression->_currentTurn );
		}

		///// 柱と傘での評価をする？
		if ( !isOpeningProccessOnly && _parameters_PieceEvaluation.pcEvalInfo.isEnabled && _parameters_PieceEvaluation.pcEvalInfo.applyBias_forPlain != 0 ) {

			//int pcEval = Evaluate_PC ( nextState, innerExpression->_currentTurn );
			int pcEval = Evaluate_PC_S ( nextState, innerExpression->_currentTurn );
			pcEval = (int)((float)pcEval * (float)_parameters_PieceEvaluation.pcEvalInfo.applyBias_forPlain / 100.0f);
			eval += pcEval;
		}
		
		///// 位置での評価をする？
		if ( !isOpeningProccessOnly && _parameters_PieceEvaluation.parameters_PositionEvaluation.isEnabled && _parameters_PieceEvaluation.parameters_PositionEvaluation.applyBias_forPlain != 0 ) {
			
			int positionEval = Evaluate_PositionForWholeBoard ( nextState, innerExpression->_currentTurn );
			
			positionEval = (int)((float)positionEval * (float)_parameters_PieceEvaluation.parameters_PositionEvaluation.applyBias_forPlain / 100.0f);
			
			eval += positionEval;
			
		}
		
		///// 特定位置での評価をする？
		if ( !isOpeningProccessOnly && _parameters_PieceEvaluation.positionCareSettings.isEnabled && _parameters_PieceEvaluation.positionCareSettings.applyBias_forPlain != 0 ) {
			
			int specificPositionsCareEval = Evaluate_PositionForSpecificPositions ( nextState, innerExpression->_currentTurn );
			
			specificPositionsCareEval = (int)((float)specificPositionsCareEval * (float)_parameters_PieceEvaluation.positionCareSettings.applyBias_forPlain / 100.0f);
			
			eval += specificPositionsCareEval;
			
		}

		///// 手札評価をする？
		if ( _parameters_PieceEvaluation.handCardEvaluate.isEnabled ) {

			eval += Evaluate_HandCards ( nextState, innerExpression->_currentTurn );
		}

		///// 手札補充手を評価する？
		if ( _parameters_PieceEvaluation.takeCardMoveEvaluate.isEnabled ) {

			// 発動条件：手札残り枚数
			if ( myHandCardsNumber <= _parameters_PieceEvaluation.takeCardMoveEvaluate.handCardsLessThanNum ) {

				// 今の指し手、カードを取る手？
				if ( allMoveList[i].action == MOVE::ACTION::TAKE_CARD ) {

					// 評価値を求めて
					int takeCardMoveEval = Evaluate_TakeCardMove ( innerExpression, allMoveList[i], innerExpression->_currentTurn );
					
					// 単に上乗せすると、現在も指し手実行後もマイナス値の場合、たとえ値が向上するのだとしても、マイナスしてしまうこととなる。
					// したがいまして、現状と、指し手後の値の差、を上乗せすることとする。
					int dif = takeCardMoveEval - eval;
					
					// レートに沿って上乗せする
					eval += (float)dif * (float)_parameters_PieceEvaluation.takeCardMoveEvaluate.addRate / 100.0f;
				}
			}
		}

		///// 王手評価もしてみる
		//eval += Evaluate_Check( nextState, innerExpression->_currentTurn);

		
		///// 評価値をつけて、リストに入れておく。あとでソート
		
		// 絞り込んでいなかった場合
		if ( nextStateList.size() == 0 ) {
			
			// 新規にレコードを作成して追加
			pair<Move, int> evaluatedMove;
			evaluatedMove.first  = allMoveList[i];
			evaluatedMove.second = eval;

			evaluatedMoveList.push_back( evaluatedMove );
			
		// 絞り込んでいた場合
		} else {
			
			// すでに存在しているはずのレコードに加算
			
			for ( unsigned int n = 0; n < evaluatedMoveList.size(); ++n ) {
				
				if ( isEqual ( allMoveList[i], evaluatedMoveList[n].first ) ) {
					
					evaluatedMoveList[n].second += eval;
					break;
				}
			}
		}

	}

	delete allocator;

progressData.progressRate = 70.0f;
	if ( progressData.isStop ) { _recordList.clear(); return; } // 停止可能ポイント

	///// 結果をソート
	GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate ( evaluatedMoveList );


progressData.progressRate = 72.0f;
	
	///// 詰めろを評価
	if ( _parameters_PieceEvaluation.preMate.isEnabled ) {

		GetNextMove_PEE_Test_PreMate ( evaluatedMoveList, innerExpression, progressData );

		///// それを踏まえてもう一度ソート
		GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate ( evaluatedMoveList );
	}

progressData.progressRate = 85.0f;

	if ( progressData.isStop ) { _recordList.clear(); return; } // 停止可能ポイント

	///// ログ：上位ｎ件を出力
	int outputBestNumber = 30;
	printf("候補手ベスト %d\n", outputBestNumber);
	for ( int i = 0; i < outputBestNumber && i < evaluatedMoveList.size(); ++i ) {

		printf( "%d 評価値 %d  %s\n", 
			i + 1, 
			evaluatedMoveList[i].second, 
			ShowMoveToConsole(evaluatedMoveList[i].first, true).c_str() );
	}

progressData.progressRate = 90.0f;

	///// 詰まされる手を除外する
	bool isMatedAvoided = true; // （何をしても詰まされる場合は false）
	if ( _parameters_PieceEvaluation.matedSearchDepth > 0 && _parameters_PieceEvaluation.matedSearchSubjectNumber > 0 ) {
		
		isMatedAvoided = GetNextMove_PEE_Test_AvoidMated ( evaluatedMoveList, innerExpression, progressData );

		///// それを踏まえてもう一度ソート
		GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate ( evaluatedMoveList );
		
		//printf ( "何をしても詰まされる判定：%s\n", isMatedAvoided ? "true":"false" );
	}

progressData.progressRate = 99.0f;

	///// 結果
	if ( evaluatedMoveList.size() > 0 ) {
		
		PickResultFrom_ShowSettingsForDebug();
		int pickedIndex = PickResultFrom_GetPickedIndex ( evaluatedMoveList, innerExpression );
		
		result = evaluatedMoveList[pickedIndex].first;
		
		/*
			外部向け評価値総括を適当に入れる。
		*/
		int e = evaluatedMoveList[pickedIndex].second;
		
		// デフォルト値として「互角」を入れておく。
		progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::EVEN;
		
		if ( !isMatedAvoided ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::LOSE_4; }
		
		else if ( e > 10000 ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::WIN_3; }
		else if ( e > 1000  ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::WIN_2; }
		else if ( e > 200   ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::WIN_1; }
		
		else if ( e < -10000 ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::LOSE_3; }
		else if ( e < -1000  ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::LOSE_2; }
		else if ( e < -200   ) { progressData.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::LOSE_1; }
		
		
		// 今回選んだものをリストの先頭にしておく。選択肢バトルの際、どれを選ぶのかの便宜として。
		// vector< pair<Move, int> > evaluatedMoveList;
		pair<Move, int> top    = evaluatedMoveList[0];
		pair<Move, int> picked = evaluatedMoveList[pickedIndex];
		
		evaluatedMoveList[0]           = picked;
		evaluatedMoveList[pickedIndex] = top;
		
		// 直近の候補手リストとして覚えておく。
		_evaluatedMoveList_theLatest = evaluatedMoveList;

	} else {
		printf("何かがおかしいですね？\n");
	}

	printf ("_PC_InvokeCount : %d\n", _PC_InvokeCount);
	
	///// 終了
	progressData.messageLogs.push_back ( "思考完了" );
	progressData.answerMove = result;
	
	_recordList.clear();
	
	//auto passTime = std::chrono::system_clock::now() - t;
	clock_t passTime =  clock() - t;
	printf ("----- AI Think Process Done.\n");
	printf( "----- %d ms\n", passTime );
	
	HandyProc_KakuninForDebug();
	Evaluate_Move_By_PressureBalance_ShowSettingsForDebug();
	ShowPositionCareSettingsForDebug();
	Evaluate_DistanceToGoal_ShowSettingsForDebug();
	
	/*
		思考の進捗が100%になっているのを見た時点で「完了してるのね」と判定されてしまう。
		answerMove に result を入れた後に、100 にしないと、運が悪いと指し手が取得できないことがありえる。
		
		というわけで、最後の最後に、100 を入れる。
	*/
	progressData.progressRate = 100.0f;
	
	return;
}


//==========================================================
// PEEファイルの実験用：手作業でいじりながら効果を確かめるための。
// 事前の絞込み処理
// 戻り値：詰みがあった場合、詰み手順。
//==========================================================
vector<Move> AI::GetNextMove_PEE_Test_PreSearch ( vector<Move> &allMoveList, InnerExpression *innerExpression ) {
	ProgressData dammy;
	return GetNextMove_PEE_Test_PreSearch ( allMoveList, innerExpression, dammy );
}
vector<Move> AI::GetNextMove_PEE_Test_PreSearch ( vector<Move> &allMoveList, InnerExpression *innerExpression, ProgressData &progressData ) {

	vector<Move> mateProcedure_asResult;

	///// 一手で勝つのがあればそれを選ぶ
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		if ( IsWinMove( allMoveList[i]) ) {
			
			Move winMove = allMoveList[i];
			allMoveList.clear();
			allMoveList.push_back( winMove );

			printf("一手で勝つ手があった\n");

			// 総括評価：勝ち
			progressData.evaluationSumUp = EVALUATION::SUM_UP::WIN_MAX;

			// 終了。
			mateProcedure_asResult.clear();
			return mateProcedure_asResult;
		}
		// 先頭に「移動」がきてるはずなので、「移動」じゃなくなったら、見るのをやめる。
		if ( allMoveList[i].action != MOVE::ACTION::MOVE_PIECE ) {
			break;
		}
	}


	///// 詰む手があればそれを選ぶ：一手勝ちと兼用にできますよね、多分。
	int mateSearchDepth = _parameters_PieceEvaluation.mateSearchDepth;

	if (mateSearchDepth  > 0) { 

		printf("詰み探索開始 (深度 %d) \n", mateSearchDepth);

		vector<Move> mateProcedure;
		GetMateProcedure_Thread( mateProcedure, allMoveList, innerExpression, mateSearchDepth);
		if ( mateProcedure.size() > 0 ) {

			printf("詰む手があった %d手\n", mateProcedure.size());

			// 総括評価：勝ち寸前
			progressData.evaluationSumUp = EVALUATION::SUM_UP::WIN_4;
			mateProcedure_asResult = mateProcedure;

			// シャッフル後、先頭の詰み手順一つだけを入れて終了
			Move mateMove = *(mateProcedure.begin());
			allMoveList.clear();
			allMoveList.push_back( mateMove );

			return mateProcedure_asResult;
		}

		printf("詰む手なし\n");
	}

	///// ここに来たということは、戻り値は空のリストであるはず。
	return mateProcedure_asResult;
}

//==========================================================
// PEEファイルの実験用：手作業でいじりながら効果を確かめるための。
// 詰まされる手があれば、妨害する手を捜す
//==========================================================
vector<Move> AI::GetNextMove_PEE_Test_DisturbMated ( InnerExpression *innerExpression ) {

	vector<Move> result;
	/*
	///// もし、今相手の手番だったら。
	InnerExpression *virtualOpponentTurnsState = innerExpression->GetCopy();
	virtualOpponentTurnsState->_currentTurn = SwitchTurn( virtualOpponentTurnsState->_currentTurn );

	
	
	
	
	///// 魅惑の未実装ワールド /////
	
	
	
	
	
	
	delete virtualOpponentTurnsState;
	*/
	return result;
}

//==========================================================
// PEEファイルの実験用：手作業でいじりながら効果を確かめるための。
// 詰まされる手を除外する
// 戻り値：詰まされる手を避けることができたかどうか。
//         何をしても詰まされる場合は false とする。それ以外は処理成功という意味でtrueとする
//==========================================================
bool AI::GetNextMove_PEE_Test_AvoidMated ( vector< pair<Move, int>> &allMoveList, InnerExpression *innerExpression, ProgressData &progressData ) {
	
	bool result = true;
	//auto t = std::chrono::system_clock::now();

	///// 上位何件に適用する？
	int subjectNumber = _parameters_PieceEvaluation.matedSearchSubjectNumber;

	// 候補手の数を超える必要はないですよね。
	if ( subjectNumber > allMoveList.size() ) {
		subjectNumber = allMoveList.size();
	}

	// やらないなら、終了
	if ( subjectNumber <= 0 ) {
		return result;
	}

	///// 読み深度
	int depth = _parameters_PieceEvaluation.matedSearchDepth;

	printf("詰まされる手を除外：上位 %d 件をチェック 深度 %d\n", subjectNumber, depth);


	///// 上位から順に、何手で詰まされるかを入れていく
	vector< pair < Move, int> > moveAndMatedLengthList;

	vector<Move> mateProcedure;
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(1500);
	
	//MemoryPool memoryPool;
	//MemoryPool_ForTreeSearch_Setup( memoryPool, 10 );

	// なんか、最初の一個を捨てておくと、落ちずに済むみたい？
	//allocator->GenerateObj();

	vector<Move> nextOpponentMove;
	
	///// 進捗バーが固まらないように
	float currentProgressRate = progressData.progressRate;
	float goalProgressRate    = 99.0f;
	
	
	//// 時間がかかる場合に、探索深度をゆるめる措置
	clock_t t = clock();
	
	///// 評価値上位から順に見ていく
	for ( int i = 0; i < subjectNumber; ++i ) {
		
		//printf("%d / %d 件目を調査中。\n", i, subjectNumber);
		
		if ( (i % 100) == 0) { 
			printf("%d / %d 件目を調査中。\n", i, subjectNumber);
			
		}
		
		// この時点で、前回の計即時から一定時間以上以上経過していた場合、探索深度をゆるめる
		clock_t currentT = clock();
		
		if ( currentT - t > 3000 ) {
			
			depth -= 2;
			
			printf ( "depth decreased for timeover. depth:[%d]\n", depth );
			
			if ( depth <= 0 ) {
				printf ( "break for timeover. depth:[%d] \n", depth );
				break;
			}
			
			t = clock();
		}
		
		
		///// 進捗バーじりじり：subjectNumberがゼロの場合はこのループに入ってこないハズなので、ゼロ除算の心配はないハズ。
		progressData.progressRate = currentProgressRate + (goalProgressRate - currentProgressRate)*i/subjectNumber;
		
		Move &move = allMoveList[i].first;

		// この手を指した次の局面に、
		InnerExpression *nextState = innerExpression->GetNext( move, allocator );
		/*
		// なんか、アロケータを使うと、落ちるので、代わりにメモリプールを使う。
		//InnerExpression *nextState = NULL;
		//MemoryPool_ForTreeSearch_Allocate ( memoryPool, nextState );
		//innerExpression -> GetCopy( nextState );
		//nextState       -> GoNext( move );

		// それも怪しいので、直に生成
		InnerExpression *nextState = innerExpression->GetNext( move );
		*/

		//printf("関数内＿前\n");
		//InnerExpression::GetCheckMoveList( nextState, nextOpponentMove );
		//printf("関数内＿後\n");

		// 相手に１手で勝つ手があったら、詰め探索を呼ぶまでもなく、ダメ。
		// １手では負けないときのみ、詰み探索を呼ぶ
		nextOpponentMove.clear();
		nextState->GetLegalWinMoveList ( nextOpponentMove );
		if ( nextOpponentMove.size() ==  0 ) {

			// 詰み手順を求める
			mateProcedure.clear();
			//mateProcedure = GetMateProcedure( nextState, depth, true);
			GetMateProcedure_Thread( mateProcedure, nextState, depth, true);

		} else {

			// 相手からのトドメ。適当に入れておく。
			mateProcedure.push_back ( nextOpponentMove[0] );
			allMoveList[i].second -= 99999; 
		}

		// この時点で、nextState は解放してよい
		//MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );
		//delete nextState;

		// 詰みがある？
		if (mateProcedure.size() > 0) {

			// 詰み手順の長さとともに、リストに保存していく
			pair < Move, int> moveAndMatedLength;
			moveAndMatedLength.first  = move;
			moveAndMatedLength.second = mateProcedure.size();

			moveAndMatedLengthList.push_back(moveAndMatedLength);

			// 詰め手順が短いほど、得点が低くなるように？
			//allMoveList[i].second -= 10000 - mateProcedure.size() * 100;

		// 詰みがない
		} else {

			// 終了
			break;
		}
	}

	delete allocator;
	//MemoryPool_ForTreeSearch_End ( memoryPool );

	///// 詰まされる心配なし？
	if ( moveAndMatedLengthList.size() <= 0 ) {

		printf ( "詰まされる心配なし\n" );

	///// 詰まされずに済む手が残っている？
	} else if ( moveAndMatedLengthList.size() < subjectNumber ) {

		printf("詰まされずに済む手が残っていた：上位%d件 が NG\n", moveAndMatedLengthList.size());
		for (unsigned int i = 0; i < moveAndMatedLengthList.size(); ++i) {
			string moveString = ShowMoveToConsole( moveAndMatedLengthList[i].first, true);
			printf ("  削った候補手  %s ：評価 %d \n",moveString.c_str(), moveAndMatedLengthList[i].second);
		}
		/*
			詰まされる件数分、リストを削る
		*/
		// リストが何件になればいい？ 
		int goalLength = allMoveList.size() - moveAndMatedLengthList.size();


		// 上位から順に削っていく
		while ( allMoveList.size() > goalLength ) {
			
			// 高評価値順に並んでいる前提で、先頭から削っていく。
			allMoveList.erase( allMoveList.begin() );
		}
		/*
			これで、トップに出た候補手は詰まされない手、ということ。
		*/

	///// 何をしても詰まされる
	} else { // moveAndMatedLengthList.size() == subjectNumber ということのはず

		printf("何をしても詰まされる\n");
		
		result = false;
		// ここでは入れない。「避けようがない」という情報を戻り値 false として返すのみとする。
		//progressData.evaluationSumUp = EVALUATION::SUM_UP::LOSE_4;

		/*
			もっとも長手数の詰まされになる候補手を選ぶ
			いや、これ、手数の長さじゃなくて、手の多さ、か。
			それはそれでいいのかも。もっとも紛れが多い手順にする、ということ。
		*/
		/*
		vector<Move> longestMatedMoves;
		int longestMatedLength = 0;
		for ( unsigned int i = 0; i < moveAndMatedLengthList.size(); ++i ) {

			// 最長手を更新？
			if ( i == 0 || longestMatedLength < moveAndMatedLengthList[i].second ) {

				longestMatedLength = moveAndMatedLengthList[i].second;
				longestMatedMoves.clear();
			}
			if ( longestMatedLength == moveAndMatedLengthList[i].second ) {

				longestMatedMoves.push_back( moveAndMatedLengthList[i].first );
			}
		}

		// 一つにする
		int lotNum = Utils_Cal::GetRandom( moveAndMatedLengthList.size() );
		allMoveList.clear();
		pair<Move, int> evaluatedMove;
		evaluatedMove.first  = moveAndMatedLengthList[lotNum].first;
		evaluatedMove.second = 99999 - moveAndMatedLengthList[lotNum].second; // あまり深い意味はない数字。負けを表してくれればよい。
		allMoveList.push_back( evaluatedMove );

		printf("最長手数？ %d の手を選択\n", moveAndMatedLengthList[lotNum].second);
		
		*/
		// 最長手数といっても、なんだか怪しいので、素直に、もとの評価順位のままにしておく
	}
	
	
	//auto passTime = std::chrono::system_clock::now() - t;
	//printf("詰まされる手を除外：所要時間：%d\n", passTime);

	return result;
}

//==========================================================
// 詰めろを評価
//==========================================================
void AI::GetNextMove_PEE_Test_PreMate ( vector< pair<Move, int>> &allMoveList, InnerExpression *innerExpression, ProgressData &progressData ) {

	// Are you sure?
	if (_parameters_PieceEvaluation.preMate.depth <= 0) {
		return;
	}

	// How many evaluate?
	int subjectNum = ( _parameters_PieceEvaluation.preMate.subjectNum < allMoveList.size() ) ? _parameters_PieceEvaluation.preMate.subjectNum : allMoveList.size();

	printf ("詰めろ探索 : 上位 %d 件に適用  深度:%d  評価値 : %d\n", subjectNum, _parameters_PieceEvaluation.preMate.depth, _parameters_PieceEvaluation.preMate.value);

	// 王手になる手を取得：除外対象として。
	vector<Move> checkMoveList;
	if ( !_parameters_PieceEvaluation.preMate.isIncludeDirectCheck ) { // 王手になる手も含めていいなら、ここで王手リストを取得する必要はない。
		InnerExpression::GetCheckMoveList ( innerExpression, checkMoveList );
	}

	vector<Move> mateProcedure;
	vector<Move> nextMoveList;
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(subjectNum);

	// count, for logging.
	int preMateCount = 0;
	
	int depth = _parameters_PieceEvaluation.preMate.depth;
	
	//// 時間がかかる場合に、探索深度をゆるめる措置
	clock_t t = clock();
	
	float currentProgressRate = progressData.progressRate;
	float goalProgressRate    = 85.0f;
	
	for ( int i = 0; i < subjectNum; ++i ) {
		
		clock_t currentT = clock();
		
		if ( currentT - t > 3000 ) {
			
			depth -= 2;
			
			printf ( "AI::GetNextMove_PEE_Test_PreMate    depth decreased for timeover : depth[%d]\n", depth );
			
			
			if ( depth <= 0 ) {
				printf ("break \n");
				break;
			}
			t = clock();
		}
		
		///// 進捗バーじりじり：subjectNum がゼロの場合はこのループに入ってこないハズなので、ゼロ除算の心配はないハズ。
		progressData.progressRate = currentProgressRate + (goalProgressRate - currentProgressRate)*i/subjectNum;
		
		Move &move = allMoveList[i].first;

		// この手自体が王手である手は「詰めろ」の対象とはしない
		if ( !_parameters_PieceEvaluation.preMate.isIncludeDirectCheck) {

			for ( unsigned int n = 0; n < checkMoveList.size(); ++n ) {
				if ( AI::MoveEquals ( move, checkMoveList[n] ) ) {
					continue;
				}
			}
		}

		// Apply the move to next
		InnerExpression *nextState = innerExpression->GetNext( move, allocator );
		
		// If it were still my turn continuously...
		nextState->_currentTurn = SwitchTurn ( nextState->_currentTurn  );

		// Get the next moves to win.
		nextMoveList.clear();
		nextState->GetLegalWinMoveList ( nextMoveList ); 
		if (nextMoveList.size() > 0) {

			printf("一手勝ち：\n");
			for ( unsigned int n = 0; n < nextMoveList.size(); ++n ) {
				printf ( "%s\n", ShowMoveToConsole(nextMoveList[n], true).c_str() );
			}
			allMoveList[i].second += _parameters_PieceEvaluation.preMate.value;
			++preMateCount;
			continue;
		}

		// Get the next possible moves.
		//nextState->GetLegalNextMoveList ( nextMoveList, MOVE::ACTION::PUT_CARD | MOVE::ACTION::PUT_PIECE | MOVE::ACTION::MOVE_PIECE );
		InnerExpression::GetCheckMoveList ( nextState, nextMoveList);

		// Search a Mate Procedure.
		GetMateProcedure_Thread ( mateProcedure, nextMoveList, nextState, depth, true );

		// If any, add the point
		if ( mateProcedure.size() > 0 ) {

			printf("詰めろ：\n");
			printf ( "%s\n", ShowMoveToConsole(move, true).c_str() );
			for ( unsigned int n = 0; n < mateProcedure.size(); ++n ) {
				printf ( "%s\n", ShowMoveToConsole(mateProcedure[n], true).c_str() );
			}

			allMoveList[i].second += _parameters_PieceEvaluation.preMate.value;

			++preMateCount;
		}
	}

	printf ("詰めろ件数 : %d\n", preMateCount);

	delete allocator;
}

//=========================================================
// 次の一手を教えて：子処理：ランダム
//=========================================================
Move AI::GetNextMove_Random( InnerExpression *innerExpression ) {

	Move result;

	//vector<Move> allMoveList = innerExpression->GetLegalNextMoveList();
	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList(allMoveList);

	if (allMoveList.size() == 0) {

		result.action = MOVE::ACTION::PASS;
		result.turn   = innerExpression->_currentTurn;

		return result;
	}

	/*
		現在の手番のプレイヤーが最善を尽くそうとする。
	*/
	vector<Move> moveCandidateList;
	for ( unsigned int i = 0; i < allMoveList.size(); ++i) {

		Move move = allMoveList[i];
		InnerExpression *nextState = innerExpression->GetNext( move );

		moveCandidateList.push_back( move );

		// 勝つ手があった。
		if (  IsWin( nextState, innerExpression->_currentTurn ) ) {
			moveCandidateList.clear();
			moveCandidateList.push_back( move );
			delete nextState;
			break;
		}
		delete nextState;
	}
	int lotIndex = Utils_Cal::GerRandom( moveCandidateList.size());

	result = moveCandidateList[lotIndex];


	return result;
}


//=========================================================
// 次の一手を教えて：木探索版
//=========================================================
Move AI::GetNextMove_Tree( InnerExpression *innerExpression ) {

	//_allReadState = 0;
	DigInto_Window_OptionArgs optionArg;


	//pair<Move, int> result = DigInto( innerExpression, 2, innerExpression->_currentTurn );
	

	//_allocator = new Allocator<InnerExpression>();
	//_allocator->Alloc(10000);

	MemoryPool memoryPool;
	MemoryPool_ForTreeSearch_Setup ( memoryPool, 100 );
	/*
	_memoryPool_ForTreeSearch = new char[ 100 * sizeof(InnerExpression) ];
	_memoryPool_ForTreeSearch_currentIt = _memoryPool_ForTreeSearch;
	ZeroMemory(_memoryPool_ForTreeSearch, 100 * sizeof(InnerExpression));
	*/

	pair<Move, int> result = DigInto_Window ( memoryPool, innerExpression, 3, innerExpression->_currentTurn, 99999, -99999, optionArg);
	printf("評価値：%d 全 %d 手\n", result.second, optionArg.allReadState);

	//delete[] _memoryPool_ForTreeSearch;
	MemoryPool_ForTreeSearch_End ( memoryPool );
	//delete _allocator;

	return result.first;
}




//=========================================================
// 掘り進む
//=========================================================
pair<Move, int> AI::DigInto( InnerExpression *currentState, int level, int fromThisTurnsViewPoint ) {

	pair<Move, int> result;

	///// 可能な指し手リストを出して
	vector<Move> allMoveList;
	currentState->GetLegalNextMoveList( allMoveList );

	int bestEval = 0;

	///// 可能な指し手でループ
	for (unsigned int i = 0; i < allMoveList.size(); ++i) {

		++_allReadState;

		// こう指すと、こうなる。
		Move move = allMoveList[i];
		InnerExpression *nextState = currentState->GetNext( move );

		// 評価値は…
		int eval = 0;

		// 勝負あり？
		bool isEnd = IsWinMove( move );

		// まだ掘る？
		pair<Move, int> digResult;
		if ( level > 0 && !isEnd) {

			digResult =	DigInto( nextState, level - 1, fromThisTurnsViewPoint);
			eval = digResult.second;

		// もう掘らない
		} else {

			// 評価：ルートプレイヤーの立場で
			eval = Evaluate( nextState, fromThisTurnsViewPoint );

			if ( _allReadState % 1000 == 0 ) {	printf ( "%d 手目 評価値:%d\n", _allReadState, eval );}
		}

		// ベスト？
		bool isBest = false;

		// 自分の手番
		if ( currentState->_currentTurn == fromThisTurnsViewPoint ) {
		
			// MAXを求める
			isBest = ( eval > bestEval || i == 0 );

		// 相手の手番
		} else {

			// MINを求める
			isBest = ( eval < bestEval || i == 0 );
		}

		// ベスト値を更新？
		if ( isBest ) {

			bestEval = eval;
			result.first  = move;
			result.second = eval;
		}

		delete nextState;
	}

	return result;
}

//=========================================================
// 次の手番を取得
//=========================================================
int AI::SwitchTurn( int turn ) {

	int result = turn;

	if ( turn == TURN::BLACK ) {
		result = TURN::WHITE;
	}
	else if ( turn == TURN::WHITE ) {
		result = TURN::BLACK;
	}

	return result;
}

//=========================================================
// 勝敗判定：純粋に、現時点で、ルール上
//【引数】この手番として、勝ってますか？
//【戻値】true 勝ってる false まだ勝ってない
//=========================================================
bool AI::IsWin( InnerExpression *innerExpression, int turn ) {

	bool result = false;

	int opponentAreaY = 0;
	int myPiece       = 0;

	///// どっちの手番として見ますか？
	if ( turn == TURN::BLACK ) {
		opponentAreaY = 0;
		myPiece       = GRID::BLACK;
	}
	else if ( turn == TURN::WHITE ) {
		opponentAreaY = _rule.boardHeight - 1;
		myPiece       = GRID::WHITE;

	// え？
	} else {
		return false;
	}


	/////  一つでも相手陣に自分の駒があれば、勝ち。
	for (int x = 0; x < _rule.boardWidth; ++x) {

		if (innerExpression->_gridList_Piece[opponentAreaY][x] == myPiece) {
			result = true;
			break;
		}
	}


	return result;
}

//=========================================================
// 勝つ指し手
//【引数】この指し手は、この手番として、勝ってますか？
//【戻値】true 勝ってる false まだ勝ってない
//=========================================================
//bool AI::IsWinMove( Move &move, int turn ) {
bool AI::IsWinMove( Move &move) {

	int turn = move.turn;

	bool result = false;

	///// 相手陣に移動する指し手なら、勝ち。（「移動」以外は勝ちでは有り得ない）
	if (move.action == MOVE::ACTION::MOVE_PIECE) {

		int opponentAreaY = 0;

		///// どっちの手番として見ますか？
		if ( turn == TURN::BLACK ) {
			opponentAreaY = 0;
		}
		else if ( turn == TURN::WHITE ) {
			opponentAreaY = _rule.boardHeight - 1;

		// え？
		} else {
			return false;
		}

		if ( move.movePieceToY == opponentAreaY ) {
			result = true;
		}
	}

	return result;
}


//=========================================================
// カード情報をください
//=========================================================
Card AI::GetCardInfo( int cardId ) {

	Card result;
	
	map<int, Card>::iterator it = _cardList.find( cardId );
	if ( it != _cardList.end() ) {

		result = it->second;

	} else {
		printf("Something wrong is going on. Yeah! : (cardId = %d)\n", cardId);
	}
	
	return result;
}

//=========================================================
// 指し手の同一判定
//=========================================================
bool AI::isEqual ( const Move &m1, const Move &m2 ) {
	
	bool result = false;
	
	if ( m1.action == m2.action && m1.turn == m2.turn ) {
		
		if ( m1.action == MOVE::ACTION::PUT_PIECE ) {
			
			result = 
			(
				m1.putPieceX == m2.putPieceX &&
				m1.putPieceY == m2.putPieceY
			);
			
		}
		else if ( m1.action == MOVE::ACTION::MOVE_PIECE ) {
		
			result = 
			(
				m1.movePieceFromX       == m2.movePieceFromX &&
				m1.movePieceFromY       == m2.movePieceFromY &&
				m1.movePieceToX         == m2.movePieceToX &&
				m1.movePieceToY         == m2.movePieceToY &&
				m1.movePieceAndIsCllide == m2.movePieceAndIsCllide
			);
		
		}
		else if ( m1.action == MOVE::ACTION::PUT_CARD ) {
			
			result = 
			(
				m1.putCardX         == m2.putCardX &&
				m1.putCardY         == m2.putCardY &&
				m1.putCardId        == m2.putCardId &&
				m1.putCardDirection == m2.putCardDirection 
			);
		
		}
		else if ( m1.action == MOVE::ACTION::TAKE_CARD ) {
			
			result = 
			(
				m1.takeCardX == m2.takeCardX &&
				m1.takeCardY == m2.takeCardY 
			);
		}
		else if ( m1.action == MOVE::ACTION::PASS ) {
			result = true;
		}
		
		
	}
	
	
	
	return result;

}

//=========================================================
// 指し手をユニーク文字列に変換。mapのキーにしたいのです
//=========================================================
string AI::moveToUSTRING ( const Move &m ) {
	
	string result = "";
	
	if ( m.turn == TURN::BLACK ) {
		result += "b";
	}
	else if ( m.turn == TURN::WHITE ) {
		result += "w";
	}
	
	if ( m.action == MOVE::ACTION::PUT_PIECE ) {
		
		result += "pp";
		result += "x" + Utils_Convert::Int_to_String ( m.putPieceX );
		result += "y" + Utils_Convert::Int_to_String ( m.putPieceY );
	}
	else if ( m.action == MOVE::ACTION::MOVE_PIECE ) {
		
		result += "mp";
		result += "fx" + Utils_Convert::Int_to_String ( m.movePieceFromX );
		result += "fy" + Utils_Convert::Int_to_String ( m.movePieceFromY );
		result += "tx" + Utils_Convert::Int_to_String ( m.movePieceToX );
		result += "ty" + Utils_Convert::Int_to_String ( m.movePieceToY );
		result += m.movePieceAndIsCllide ? "c" : "nc";
	}
	else if ( m.action == MOVE::ACTION::PUT_CARD ) {
		
		result += "pc";
		result += "x" + Utils_Convert::Int_to_String ( m.putCardX );
		result += "y" + Utils_Convert::Int_to_String ( m.putCardY );
		result += "d" + Utils_Convert::Int_to_String ( m.putCardDirection );
		result += "i" + Utils_Convert::Int_to_String ( m.putCardId );
	}
	else if ( m.action == MOVE::ACTION::TAKE_CARD ) {
		
		result += "tc";
		result += "x" + Utils_Convert::Int_to_String ( m.takeCardX );
		result += "y" + Utils_Convert::Int_to_String ( m.takeCardY );
	}
	else if ( m.action == MOVE::ACTION::PASS ) {
		result += "p";
	}
	
	
	return result;
}

//=========================================================
// AIから得た指し手をコンソールに表示
//=========================================================
string AI::ShowMoveToConsole( const AI_Modules::Move &move, bool isOnlyReturnValue ) {

	string result = "";
	char buf[512];

	///// 手番
	if ( move.turn == AI_Modules::TURN::BLACK ) {
		if (!isOnlyReturnValue) {printf( "BlackTurn\n" );}
		sprintf( buf, "BlackTurn " );
		result += buf;
	}
	else if ( move.turn == AI_Modules::TURN::WHITE ) {
		if (!isOnlyReturnValue) {printf( "WhiteTurn\n" );}

		sprintf( buf, "WhiteTurn " );
		result += buf;
	}

	///// 駒設置の場合
	if ( move.action == AI_Modules::MOVE::ACTION::PUT_PIECE) {

		if (!isOnlyReturnValue) {printf( "PUT_PIECE：(%d,%d)\n", move.putPieceX, move.putPieceY );}
		sprintf( buf, "PUT_PIECE：(%d,%d)", move.putPieceX, move.putPieceY );
	}
	///// 駒移動の場合
	else if ( move.action == AI_Modules::MOVE::ACTION::MOVE_PIECE) {

		string isCollide = ( move.movePieceAndIsCllide ) ? "Collide" : "not Collide";

		if (!isOnlyReturnValue) {printf( "MOVE_PIECE：(%d, %d)→(%d, %d) %s\n", move.movePieceFromX, move.movePieceFromY, move.movePieceToX, move.movePieceToY, isCollide.c_str() );}
		sprintf( buf, "MOVE_PIECE：(%d, %d)→(%d, %d) %s", move.movePieceFromX, move.movePieceFromY, move.movePieceToX, move.movePieceToY, isCollide.c_str() );
	}
	///// カード設置の場合
	else if ( move.action == AI_Modules::MOVE::ACTION::PUT_CARD) {

		string directionString = "";
		if      ( move.putCardDirection == AI_Modules::DIRECTION::UP    ) { directionString = "UP"; }
		else if ( move.putCardDirection == AI_Modules::DIRECTION::LEFT  ) { directionString = "LEFT"; }
		else if ( move.putCardDirection == AI_Modules::DIRECTION::RIGHT ) { directionString = "RIGHT"; }
		else if ( move.putCardDirection == AI_Modules::DIRECTION::DOWN  ) { directionString = "DOWN"; }

		// あるはず。
		string cardName = _cardNameList[move.putCardId];

		if (!isOnlyReturnValue) {printf( "PUT_CARD：(%d, %d) に %s を %s で\n", move.putCardX, move.putCardY, cardName.c_str(), directionString.c_str() );}
		sprintf( buf, "PUT_CARD：(%d, %d) に %s を %s で", move.putCardX, move.putCardY, cardName.c_str(), directionString.c_str() );
	}
	///// カードを取る場合
	else if ( move.action == AI_Modules::MOVE::ACTION::TAKE_CARD) {

		if (!isOnlyReturnValue) {printf( "TAKE_CARD：(%d, %d)\n", move.takeCardX, move.takeCardY);}
		sprintf( buf, "TAKE_CARD：(%d, %d)", move.takeCardX, move.takeCardY);
	}
	///// パス
	if ( move.action == AI_Modules::MOVE::ACTION::PASS) {

		if (!isOnlyReturnValue) {printf( "PASS\n" );}
		sprintf( buf, "PASS" );
	}
	result += buf;

	return result;
}




//=========================================================
// 結果取得の際にランダム要素を入れる処理関連：デバッグ用。設定内容を確認
//=========================================================
void AI::PickResultFrom_ShowSettingsForDebug() {
	
	PickResultFrom params = _parameters_PieceEvaluation.pickResultFrom;
	
	printf ("<PICK_RESULT_FROM>\n");
	
	printf ("\t<ENABLED>%s</ENABLED>\n", params.isEnabled ? "on" : "off");
	printf ("\t<GAME_PROGRESS from=\"%d\" to=\"%d\" />\n", params.enabledTurn_from, params.enabledTurn_to);
	
	printf ("\t<TOP>%d</TOP>\n", params.topLength);
	
	printf ("</PICK_RESULT_FROM>\n");

	
	
	
}
//=========================================================
// 結果取得の際にランダム要素を入れる処理：戻り値は「この要素が今回の当たり」というインデックス
//=========================================================
int AI::PickResultFrom_GetPickedIndex ( const vector< pair<Move, int> > &evaluatedMoveList, InnerExpression *innerExpression ) {
	
	int result = 0;
	
	PickResultFrom params = _parameters_PieceEvaluation.pickResultFrom;
	
	// 無効ならば何もしない
	if ( !params.isEnabled ) {
		return 0;
	}
	
	// 進行度合いでの判定
	int progress = GetGameProgressDegree ( innerExpression );
	
	if ( params.enabledTurn_from > progress ) { 
		return 0;
	}
	
	if ( params.enabledTurn_to != -1 && params.enabledTurn_to < progress ) {
		return 0;
	}
	
	// 件数が超過している場合の対応
	int topLength = params.topLength;
	
	if ( evaluatedMoveList.size() < topLength ) {
		topLength = evaluatedMoveList.size();
	}

printf ( "topLength : %d \n", topLength );	
	
	result = Utils_Cal::GetRandom ( topLength );
	
printf ( "result : %d \n", result );
	
	return result;
}









