//=========================================================
//
// AI_実装ファイル_評価関数関係
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"


using namespace AI_Modules;


//=========================================================
// 評価関数
//【引数】局面、どっちの立場で評価しますか？
//=========================================================
int AI::Evaluate( InnerExpression* innerExpression, int forThisTurn, int rutinFlgs) {

	int result = 0;

	///// 手番を考慮：自分の手番なら、少し評価高い
	if ( rutinFlgs == -1 || ( (rutinFlgs & EVALUATION::RUTIN::CONSIDER_ONES_TURN) == EVALUATION::RUTIN::CONSIDER_ONES_TURN ) ) {

		if ( innerExpression->_currentTurn == forThisTurn) {
			result += 10;
		} else {
			result -= 10;
		}
	}
	///// 勝ってたら評価最高
	if ( rutinFlgs == -1 || ( (rutinFlgs & EVALUATION::RUTIN::GAME_OVER) == EVALUATION::RUTIN::GAME_OVER ) ) {
		if ( IsWin( innerExpression, forThisTurn ) ) {

			result = 99999 - innerExpression->GetTotalMoveCount();

			// もう考えない
			return result;
		}

		///// 負けてたら評価最低
		if ( IsWin( innerExpression, SwitchTurn( forThisTurn ) ) ) {

			result = -99999 + innerExpression->GetTotalMoveCount();

			// もう考えない
			return result;
		}
	}
	///// 駒の価値を計算
	if ( rutinFlgs == -1 || ( (rutinFlgs & EVALUATION::RUTIN::VALUE_OF_PIECES) == EVALUATION::RUTIN::VALUE_OF_PIECES ) ) {
		result += Evaluate_Piece ( innerExpression, forThisTurn );
	}
	///// 地形の価値を計算
	if ( rutinFlgs == -1 || ( (rutinFlgs & EVALUATION::RUTIN::VALUE_OF_TERRAIN) == EVALUATION::RUTIN::VALUE_OF_TERRAIN ) ) {

		/*
			地形を価値を判定するための読み深度はどう指定しますか？
		*/
		result += Evaluate_Terrain ( innerExpression, forThisTurn, 0);
	}
	///// カードの位置を計算：カードというか、移動制限。自陣付近に通れるマス目が多いと不利、みたいな。
	if ( rutinFlgs == -1 || ( (rutinFlgs & EVALUATION::RUTIN::VALUE_OF_CARDS) == EVALUATION::RUTIN::VALUE_OF_CARDS ) ) {
		result += Evaluate_Card ( innerExpression, forThisTurn );
	}
	///// 王手がかかっていたら高得点ですよね？
	if ( rutinFlgs == -1 || ( (rutinFlgs & EVALUATION::RUTIN::VALUE_OF_CHECK) == EVALUATION::RUTIN::VALUE_OF_CHECK ) ) {
		result += Evaluate_Check ( innerExpression, forThisTurn );
	}

	return result;
}

//=========================================================
// 評価関数子処理：勝ち負け判定
//=========================================================
bool AI::Evaluate_IsWin( int &result, InnerExpression* innerExpression, int &forThisTurn) {

	///// 勝負がついていたら、もう考えない、ということにしたいですよね？
	bool isWinLose = false;

	if ( IsWin( innerExpression, forThisTurn ) ) {

		result = 99999 - innerExpression->GetTotalMoveCount();

		// もう考えない
		isWinLose = true;


	///// 負けてたら評価最低
	} else if ( IsWin( innerExpression, SwitchTurn( forThisTurn ) ) ) {

		result = -99999 + innerExpression->GetTotalMoveCount();

		// もう考えない
		isWinLose = true;
	}

	return isWinLose;
}


//=========================================================
// 評価関数子処理：駒の価値
//=========================================================
int AI::Evaluate_Piece( const InnerExpression* innerExpression, const int forThisTurn) {
	++_count_Evaluate_Piece;
	float result = 0;

	///// 自分の駒、相手の駒
	int myPiece = ( forThisTurn == TURN::BLACK ) ? GRID::BLACK : GRID::WHITE;
	int opPiece = ( forThisTurn == TURN::BLACK ) ? GRID::WHITE : GRID::BLACK;

	///// 自分にとっての「前」「後」
	int throuableDirectionForward  = ( forThisTurn == TURN::BLACK ) ? THROUGHABLE_INFO::MUST_TO_UP   : THROUGHABLE_INFO::MUST_TO_DOWN;
	int throuableDirectionBackward = ( forThisTurn == TURN::BLACK ) ? THROUGHABLE_INFO::MUST_TO_DOWN : THROUGHABLE_INFO::MUST_TO_UP;

	///// 相手陣・自陣の高さ
	int opponentAreaY = ( forThisTurn == TURN::BLACK ) ? 0 : _rule.boardHeight - 1;
	int myAreaY       = ( forThisTurn == TURN::BLACK ) ? _rule.boardHeight - 1 : 0;

	///// 自分の手駒の数
	int my_numberOfHandPieces = ( forThisTurn == TURN::BLACK ) ? innerExpression->_handPiece_BLACK : innerExpression->_handPiece_WHITE;
	int op_numberOfHandPieces = ( forThisTurn == TURN::BLACK ) ? innerExpression->_handPiece_WHITE : innerExpression->_handPiece_BLACK;
	
	int myPieceMax = ( forThisTurn == TURN::BLACK ) ? _pieceCount_black : _pieceCount_white;
	int opPieceMax = ( forThisTurn == TURN::BLACK ) ? _pieceCount_white : _pieceCount_black;
	
	///// （盤面上の）駒についてだけの判定なので、存在するはずの駒の数だけ判定したら、そこで処理を打ち切っていい。
	int my_numberOfPiece = myPieceMax - my_numberOfHandPieces;
	int op_numberOfPiece = opPieceMax - op_numberOfHandPieces;
	
	int doneCount_my = 0;
	int doneCount_op = 0;
	
	
	///// 盤上の駒について計算
	/*
		相手の駒はどう扱う？ 無視 or 符号マイナス
	*/
	bool loopEnd = false;
	for (int y = 0; y < _rule.boardHeight; ++y) { if ( loopEnd ) {break;}
		for (int x = 0; x < _rule.boardWidth; ++x) {
			
			
			if ( doneCount_my >= my_numberOfPiece && doneCount_op >= op_numberOfPiece ) {
				//printf ("全駒見終えたのでループ終了\n");
				loopEnd = true;
				break;
			}
			
			
			// 自分の駒が、ありました。
			if ( innerExpression->_gridList_Piece[y][x] == myPiece ) {
				
				++doneCount_my;

				// パネルの密集度
				//result += Evaluate_DensityOfPanel( innerExpression, x, y, myPiece );

				// ブラックホール判定
				//int blackHoleDegree = Evaluate_BlackholeExaminaion ( innerExpression->_gridList_ThrouableInfo, x, y );

				// ブラックホール度をマイナス：相手陣地に入っている場合は高さ判定の箇所で打ち消すので問題なし
				//result -= blackHoleDegree;

				// すでに相手陣地に入っていたら高得点
				//if ( y == opponentAreaY ) {
				/*
				if ( false ) { // PEEファイル で OPPONENT_AREA (0, 0) で指定することにする。
	
					result += _heightValue_Goal;

				} else {
				*/
				// まだ相手陣地には達していない：通常の評価を行う

					// 評価素子での判定
					result += Evaluate_PEE ( innerExpression, myPiece, x, y, _parameters_PieceEvaluation.pieceEvaluateElementList );
					
					// 「高さ」での加点
					int height = 0;
					if      (myPiece == GRID::BLACK) { height = _rule.boardHeight - y;}
					else if (myPiece == GRID::WHITE) { height = y + 1;}

					//float heightUnit = 1.0f / (float)_rule.borardHeight;
					//result += ((float)height * heightUnit) * (float)_parameters_PieceEvaluation.heightValue_Bias;
					
					// 普通に、一段ごとに加算。
					result += _parameters_PieceEvaluation.heightValue_Bias * height;
					
				//}

			}

			// 相手の駒が、ありました。
			else if ( innerExpression->_gridList_Piece[y][x] == opPiece ) {

				++doneCount_op;

				// パネルの密集度
				//result -= Evaluate_DensityOfPanel( innerExpression, x, y, opPiece );

				// ブラックホール判定
				//int blackHoleDegree = Evaluate_BlackholeExaminaion ( innerExpression->_gridList_ThrouableInfo, x, y );

				// ブラックホール度をマイナス：相手陣地に入っている場合は高さ判定の箇所で打ち消すので問題なし
				//result += blackHoleDegree;

				// すでに相手陣地に入っていたら高得点
				//if ( y == myAreaY ) {
				/*
				if ( false ) { // PEEファイル で OPPONENT_AREA (0, 0) で指定することにする。

					result -= _heightValue_Goal;

				} else {
				*/
				// まだゴールには達していない場合、通常の評価を行う

					// 評価素子での判定
					result -= Evaluate_PEE ( innerExpression, opPiece, x, y, _parameters_PieceEvaluation.pieceEvaluateElementList );
					
					// 「高さ」での加点
					int height = 0;
					if      (opPiece == GRID::BLACK) { height = _rule.boardHeight - y;}
					else if (opPiece == GRID::WHITE) { height = y + 1;}

					//float heightUnit = 1.0f / (float)_rule.borardHeight;
					//result -= ((float)height * heightUnit) * (float)_parameters_PieceEvaluation.heightValue_Bias;

					// 普通に、一段ごとに加算。
					result -= _parameters_PieceEvaluation.heightValue_Bias * height;
				//}
			}
		}
	}
	

	
	///// 手駒について計算：自駒

	// 手駒は損、とみなしてみる。
	//result -= my_numberOfHandPieces * 100;
	//result += op_numberOfHandPieces * 100;

	//return 0;
	return result;
}

//=========================================================
// 評価関数子処理：カードの価値
//=========================================================
int AI::Evaluate_Card( const InnerExpression* innerExpression, const int forThisTurn) {

	int result = 0;

	///// 手札の枚数
	int myHandCards = ( forThisTurn == TURN::BLACK ) ? innerExpression->_handCards_BLACK_Index : innerExpression->_handCards_WHITE_Index;
	int opHandCards = ( forThisTurn == TURN::BLACK ) ? innerExpression->_handCards_WHITE_Index : innerExpression->_handCards_BLACK_Index;

	///// 手札が多い方が有利とする
	result += myHandCards * 1;
	result -= opHandCards * 1;

	///// 陣地の高さ
	int myAreaY = ( forThisTurn == TURN::BLACK ) ? _rule.boardHeight - 1 : 0;
	int opAreaY = ( forThisTurn == TURN::BLACK ) ? 0 :  _rule.boardHeight - 1;

	///// 前・後ろ
	int throuableForward  = ( forThisTurn == TURN::BLACK ) ? THROUGHABLE_INFO::MUST_TO_UP   : THROUGHABLE_INFO::MUST_TO_DOWN;
	int throuableBackward = ( forThisTurn == TURN::BLACK ) ? THROUGHABLE_INFO::MUST_TO_DOWN : THROUGHABLE_INFO::MUST_TO_UP;

	///// 相手陣に近い位置にカードというか、通れるマス目があるほど有利とする
	for ( int y = 0; y < _rule.boardHeight; ++y) {
		for ( int x = 0; x < _rule.boardWidth; ++x) {

			const int &throuableInfo = innerExpression->_gridList_ThrouableInfo[y][x];

			if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE) {

				result += _rule.boardHeight - Utils_Cal::Abs( opAreaY - y );
				result -= _rule.boardHeight - Utils_Cal::Abs( myAreaY - y );

			} else if ( throuableInfo == throuableForward) {

				// 前進パネルは、自陣に近くてもマイナスにしない。
				result += _rule.boardHeight - Utils_Cal::Abs( opAreaY - y );

			} else if ( throuableInfo == throuableBackward) {

				// 後退パネルは、自陣に近いほどマイナス。
				result -= _rule.boardHeight - Utils_Cal::Abs( myAreaY - y );

			} else if ( throuableInfo == THROUGHABLE_INFO::MUST_TO_LEFT) {

				result += _rule.boardHeight - Utils_Cal::Abs( opAreaY - y );
				result -= _rule.boardHeight - Utils_Cal::Abs( myAreaY - y );

			} else if ( throuableInfo == THROUGHABLE_INFO::MUST_TO_RIGHT) {

				result += _rule.boardHeight - Utils_Cal::Abs( opAreaY - y );
				result -= _rule.boardHeight - Utils_Cal::Abs( myAreaY - y );
			}

		}
	}

	return result;
}

//=========================================================
// 評価関数子処理：王手の価値
//=========================================================
int AI::Evaluate_Check( InnerExpression *innerExpression, const int forThisTurn) {

	int result = 0;

	/*
		これを評価値に組み込めば、
		事前チェックで重い受け処理をしなくても、
		「王手をかけられる手＝評価の低い手」として回避してくれるのではないか？
	*/
	/*
		「王手」と「勝つ手」の違いは？

		「王手」は、手番じゃないけど、もし手番だったら、勝つ手があるよ、という「状態」
		それが、手番が逆だと「勝つ手がある」＝「すでに勝っている」
	*/
	/*
		IsWin では、現に勝負がついているかどうかを見ている。
		ここでは、そういう意味ではない。
		もしかしたら、勝てるのに、見逃すかもしれない。
	*/

	///// 現在の手番のままで、勝つ手を取得
	vector<Move> winMoveList;
	innerExpression->GetLegalWinMoveList( winMoveList );

	///// 現在の手番で、勝つ手がある
	if (winMoveList.size() > 0) {

		// 自分の手番？
		if ( innerExpression->_currentTurn == forThisTurn ) {

			// この手に気付いたなら勝てる状態、ということ。
			result += winMoveList.size() * 5000 * winMoveList.size();

		// 相手の手番なのですね
		} else {

			// 相手がこの手に気付いたなら負ける状態、ということ。
			result -= winMoveList.size() * 5000 * winMoveList.size();
		}
	}

	///// 手番を入れ替えて、勝つ手を取得
	innerExpression->_currentTurn = SwitchTurn( innerExpression->_currentTurn );
	innerExpression->GetLegalWinMoveList( winMoveList );
	innerExpression->_currentTurn = SwitchTurn( innerExpression->_currentTurn );

	///// つまり王手になっている、ということ
	if (winMoveList.size() > 0) {

		// 自分の手番？
		if ( innerExpression->_currentTurn == forThisTurn ) {

			// 受けなければ負け、ということ
			result -= winMoveList.size() * 3000 * winMoveList.size();

		// 相手の手番なのですね
		} else {

			// 自分が王手をかけている。相手が受け間違えればこちらの勝ち。
			result += winMoveList.size() * 3000 * winMoveList.size();
		}
	}


	return result;
}

//=========================================================
// グリッドを移動可不可情報にそって探索、価値を評定
//=========================================================
int AI::Evaluate_Grid ( const int gridList_ThrouableInfo[15][15], int x, int y, int currentMax, int restStep, const int forThisTurn) {

	int result = 0;

	int throuableInfo = gridList_ThrouableInfo[y][x];

	///// 終了判定
	if ( restStep <= 0 ) {
		return currentMax;
	}

	///// 移動不可？
	if ( throuableInfo == THROUGHABLE_INFO::NOT_THROUGHABLE ) {

		return currentMax;
	}

	///// 上にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE || throuableInfo == THROUGHABLE_INFO::MUST_TO_UP ) {

		if ( y > 0 && gridList_ThrouableInfo[y - 1][x] != THROUGHABLE_INFO::NOT_THROUGHABLE) {

			int nextMax = currentMax;
			
			// 黒は盤面の上ほど高得点
			if (forThisTurn == TURN::BLACK) {
				nextMax = ( _rule.boardHeight -  y - 1 > currentMax ) ? y + 1 : currentMax;
			}

			int resultCandidate = Evaluate_Grid ( gridList_ThrouableInfo, x, y - 1, nextMax, restStep - 1, forThisTurn );

			result = resultCandidate;
		}
	}
	///// 下にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE || throuableInfo == THROUGHABLE_INFO::MUST_TO_DOWN ) {

		if ( y < _rule.boardHeight && gridList_ThrouableInfo[y + 1][x] != THROUGHABLE_INFO::NOT_THROUGHABLE ) {
			
			int nextMax = currentMax;

			// 白は盤面の下ほど高得点
			if (forThisTurn == TURN::WHITE) {
				nextMax = ( y + 1 > currentMax ) ? y + 1 : currentMax;
			}

			int resultCandidate =  Evaluate_Grid ( gridList_ThrouableInfo, x, y + 1, nextMax, restStep - 1, forThisTurn );

			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}
	///// 右にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE || throuableInfo == THROUGHABLE_INFO::MUST_TO_RIGHT ) {

		if ( x < _rule.boardWidth  && gridList_ThrouableInfo[y][x + 1] != THROUGHABLE_INFO::NOT_THROUGHABLE) {

			int resultCandidate =  Evaluate_Grid ( gridList_ThrouableInfo, x + 1, y, currentMax, restStep - 1, forThisTurn );
			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}
	///// 左にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE || throuableInfo == THROUGHABLE_INFO::MUST_TO_LEFT ) {

		if ( x > 0 && gridList_ThrouableInfo[y][x - 1] != THROUGHABLE_INFO::NOT_THROUGHABLE ) {

			int resultCandidate =  Evaluate_Grid ( gridList_ThrouableInfo, x - 1, y, currentMax, restStep - 1, forThisTurn );
			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}

	///// 右上にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE) {

		if ( x < _rule.boardWidth && y > 0 && gridList_ThrouableInfo[y - 1][x + 1] != THROUGHABLE_INFO::NOT_THROUGHABLE ) {

			int nextMax = currentMax;
			
			// 黒は盤面の上ほど高得点
			if (forThisTurn == TURN::BLACK) {
				nextMax = ( _rule.boardHeight -  y - 1 > currentMax ) ? y + 1 : currentMax;
			}

			int resultCandidate = Evaluate_Grid ( gridList_ThrouableInfo, x + 1, y - 1, nextMax, restStep - 1, forThisTurn );
			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}
	///// 左上にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE) {

		if ( x > 0 && y > 0 && gridList_ThrouableInfo[y - 1][x -1] != THROUGHABLE_INFO::NOT_THROUGHABLE ) {

			int nextMax = currentMax;
			
			// 黒は盤面の上ほど高得点
			if (forThisTurn == TURN::BLACK) {
				nextMax = ( _rule.boardHeight -  y - 1 > currentMax ) ? y + 1 : currentMax;
			}

			int resultCandidate = Evaluate_Grid ( gridList_ThrouableInfo, x - 1, y - 1, nextMax, restStep - 1, forThisTurn );
			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}
	///// 右下にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE) {

		if ( x < _rule.boardWidth && y < _rule.boardHeight && gridList_ThrouableInfo[y + 1][x + 1] != THROUGHABLE_INFO::NOT_THROUGHABLE ) {

			int nextMax = currentMax;

			// 白は盤面の下ほど高得点
			if (forThisTurn == TURN::WHITE) {
				nextMax = ( y + 1 > currentMax ) ? y + 1 : currentMax;
			}

			int resultCandidate = Evaluate_Grid ( gridList_ThrouableInfo, x + 1, y + 1, nextMax, restStep - 1, forThisTurn );
			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}
	///// 左下にいける？
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE) {

		if ( x > 0 && y < _rule.boardHeight && gridList_ThrouableInfo[y + 1][x -1] != THROUGHABLE_INFO::NOT_THROUGHABLE ) {

			int nextMax = currentMax;

			// 白は盤面の下ほど高得点
			if (forThisTurn == TURN::WHITE) {
				nextMax = ( y + 1 > currentMax ) ? y + 1 : currentMax;
			}

			int resultCandidate = Evaluate_Grid ( gridList_ThrouableInfo, x - 1, y + 1, nextMax, restStep - 1, forThisTurn );
			result = ( resultCandidate > result ) ? resultCandidate : result;
		}
	}


	return result;
}

//==========================================================
// グリッドの価値判定
//==========================================================
int AI::Evaluate_Terrain 
(
	const InnerExpression *currentState  // この局面における評価
	,int x, int y                        // 探索の現在地
	,int currentMoveDirection            // 現在、この方向へ移動中（直線移動なら低コスト）
	,int energy                          // 残り移動エネルギー
	,int myColor                         // 自分の色
	,vector<vector<bool>> walkedGrid     // すでに通った位置へ印をつける（サイズは大元の呼び出し元で定義されているとする）
	,int level
) 
{
	//printf("(%d, %d) 深さ：%d  energy：%d\n", x, y, level, energy);

	///// ゴールした時点での残り energy を「結果」とする。
	int result = energy;

	///// あまりにも道のりが長くなっていたら、袋小路扱いにする
	int limitLebel = _rule.boardWidth + _rule.boardHeight;
	if (level >= limitLebel) {
		result = EVALUATION::TERRAIN::BLIND_ALLEY;
		return result;
	}

	//==========================================================
	// 仮に、パラメータをここで定義
	//==========================================================
	int cost_changeDirection         = 100; // 方向転換コスト
	int cost_resolveNonThrouable     = 100; // 入場不可を解消するコスト
	int cost_getMyPieceAway          = 100; // 自分の駒をどかすコスト
	int cost_getOpponentPieceAway    = 1000; // 相手の駒をどかすコスト
	int cost_resolvleNotMove         = 100; // 移動不可を解消するコスト
	int cost_resolvleNotMoveOpposite = 1000; // 移動不可（正反対）を解消するコスト


	//==========================================================
	// 入場コストを計算
	//==========================================================

	///// 現在地が入場不可である？
	if ( currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE ) {
		energy -= cost_resolveNonThrouable;
	}

	///// 現在地に先客がいる？
	if ( currentState->_gridList_Piece[y][x] != GRID::ENPTY ) {

		// 味方の駒
		if ( currentState->_gridList_Piece[y][x] == myColor ) { 
			energy -= cost_getMyPieceAway;
		// 敵の駒
		} else { 
			energy -= cost_getOpponentPieceAway;
		}
	}

	///// では入場しましょう：印をつける
	walkedGrid[y][x] = true;

	//==========================================================
	// ゴール判定
	//==========================================================

	///// ゴールの位置を確認
	int goalY = 0;
	if      ( myColor == GRID::BLACK ) { goalY = 0; }
	else if ( myColor == GRID::WHITE ) { goalY = _rule.boardHeight - 1;}

	///// ゴールしていたら終了
	if ( y == goalY ) {

		///// 残り energy を元に、点数・グリッドの価値を決める
		/*
			energy が多く残っているほど、価値が高い。 
		*/

		result = energy;
		return result;
	}

	//==========================================================
	// 移動コストを計算
	//==========================================================

	///// 行き止まりだった？
	bool isBlindAlley = true;

	///// ハイスコア入れ
	int highScoreResult = 0;

	///// 下へ移動しようとする
	do {
		if (myColor == GRID::BLACK) { break; }

		int nextX = x;
		int nextY = y + 1;
		int nextMoveDirection = DIRECTION::DOWN;

		// 移動できないなら、しない
		if ( nextY >= _rule.boardHeight ) { break;}

		// すでに通った場所へは行かない（逆方向へは行かない、もこれでついでに判定できる）
		if ( walkedGrid[nextY][nextX] ) { 
			break;
		} else { 
			isBlindAlley = false; 
		}

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_DOWN) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_UP) {nextEnergy -= cost_resolvleNotMoveOpposite;}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？ 
		highScoreResult = resultCandidate;
		//if ( resultCandidate > result ) { result = resultCandidate; }

	} while(false);

	///// 上へ移動しようとする
	do {
		if (myColor == GRID::WHITE) { break; }

		int nextX = x;
		int nextY = y - 1;
		int nextMoveDirection = DIRECTION::UP;

		// 移動できないなら、しない
		if ( nextY < 0 ) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_UP) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_DOWN) {
			nextEnergy -= cost_resolvleNotMoveOpposite;
		}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);
	
	/*
	///// 右へ移動しようとする
	do {
		int nextX = x + 1;
		int nextY = y;
		int nextMoveDirection = DIRECTION::RIGHT;

		// 移動できないなら、しない
		if ( nextX >= _rule.borardWidth ) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_RIGHT) {}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);
	
	///// 左へ移動しようとする
	do {
		int nextX = x - 1;
		int nextY = y;
		int nextMoveDirection = DIRECTION::LEFT;

		// 移動できないなら、しない
		if ( nextX < 0 ) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_LEFT) {}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);
	*/
	///// 左上へ移動しようとする
	do {
		if (myColor == GRID::WHITE) { break; }

		int nextX = x - 1;
		int nextY = y - 1;
		int nextMoveDirection = DIRECTION::LEFT_UP;

		// 移動できないなら、しない
		if ( nextX < 0 ) { break;}
		if ( nextY < 0 ) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_DOWN) {nextEnergy -= cost_getOpponentPieceAway;}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);
	
	///// 左下へ移動しようとする
	do {
		if (myColor == GRID::BLACK) { break; }

		int nextX = x - 1;
		int nextY = y + 1;
		int nextMoveDirection = DIRECTION::LEFT_DOWN;

		// 移動できないなら、しない
		if ( nextX < 0 ) { break;}
		if ( nextY >= _rule.boardHeight ) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_UP) {nextEnergy -= cost_getOpponentPieceAway;}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);

	///// 右上へ移動しようとする
	do {
		if (myColor == GRID::WHITE) { break; }

		int nextX = x + 1;
		int nextY = y - 1;
		int nextMoveDirection = DIRECTION::RIGHT_UP;

		// 移動できないなら、しない
		if ( nextX >= _rule.boardWidth ) { break;}
		if ( nextY < 0 ) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_DOWN) {nextEnergy -= cost_getOpponentPieceAway;}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);

	///// 右下へ移動しようとする
	do {
		if (myColor == GRID::BLACK) { break; }

		int nextX = x + 1;
		int nextY = y + 1;
		int nextMoveDirection = DIRECTION::RIGHT_DOWN;

		// 移動できないなら、しない
		if ( nextX >= _rule.boardWidth ) { break;}
		if ( nextY >= _rule.boardHeight) { break;}

		// すでに通った場所へは行かない
		if ( walkedGrid[nextY][nextX] ) { break;} else { isBlindAlley = false; }

		// 方向転換コストを計算
		int nextEnergy = energy;
		if ( nextMoveDirection != currentMoveDirection ) { nextEnergy -= cost_changeDirection;}

		// 移動コストを計算
		if      (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::NOT_THROUGHABLE) { nextEnergy -= cost_resolvleNotMove;}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::THROUGHABLE) {}
		else if (currentState->_gridList_ThrouableInfo[y][x] == THROUGHABLE_INFO::MUST_TO_UP) {nextEnergy -= cost_getOpponentPieceAway;}
		else    {nextEnergy -= cost_resolvleNotMove;}

		// 移動
		int resultCandidate = Evaluate_Terrain ( currentState ,nextX, nextY ,nextMoveDirection ,nextEnergy ,myColor, walkedGrid, level + 1);

		// ハイスコアを更新？
		if ( resultCandidate > highScoreResult ) { highScoreResult = resultCandidate; }

	} while(false);
	
	//==========================================================
	// 移動判定 後 の まとめ
	//==========================================================

	///// どこへも行けなかった
	if ( isBlindAlley ) {

		result = EVALUATION::TERRAIN::BLIND_ALLEY;

	///// どこかには行けた
	} else {

		result = highScoreResult;
	}


	return result;
}


//==========================================================
// グリッドの価値を、周辺にあるパネルの密集度で判断
//==========================================================
int AI::Evaluate_DensityOfPanel(
	const InnerExpression *currentState  // この局面における評価
	,int x, int y                        // 探索の現在地
	,int myPiece                         // 自分の色
) {

	/*
		ブラックホール判定は、さし当たって別処理としておく。
	*/
	int result = 0;


	int radius = 2; // 半径

	if ( myPiece == GRID::BLACK ) {

		// 自分にとっての前後
		int throuable_UP   = THROUGHABLE_INFO::MUST_TO_UP;
		int throuable_DOWN = THROUGHABLE_INFO::MUST_TO_DOWN;

		// 自分にとっての「高さ」
		int hight = _rule.boardHeight - y;

		// ゴール位置
		int goalY = 0;

		// ゴールに達していたら、高得点。
		if ( y == goalY) {
			result = 9000;

		// ゴールに達していないなら…
		} else {

			for ( int defY = -radius; defY <= radius; ++defY ) {
				for ( int defX = -radius; defX <= radius; ++defX ) {

					// 評価値テーブル上の位置
					int tableX = radius + defX;
					int tableY = radius + defY;

					// 盤面上の、この位置を見る
					int lookX = x + defX;
					int lookY = y + defY;

					// 盤をはみ出していたら見ない
					if ( lookY < 0) { continue; }
					if ( lookX < 0) { continue; }
					if ( lookY >= _rule.boardHeight ) { continue; }
					if ( lookX >= _rule.boardWidth  ) { continue; }

					int lookPanel = currentState->_gridList_ThrouableInfo[lookY][lookX];

					int tableValue = 0;

					if      (lookPanel == THROUGHABLE_INFO::THROUGHABLE    ) { tableValue = _terrainEvalTable_Throuable[ tableY ][ tableX ];}
					else if (lookPanel == throuable_UP                 ) { tableValue = _terrainEvalTable_GoUP     [ tableY ][ tableX ];}
					else if (lookPanel == throuable_DOWN               ) { tableValue = _terrainEvalTable_GoDOWN   [ tableY ][ tableX ];}
					else if (lookPanel == THROUGHABLE_INFO::MUST_TO_RIGHT) { tableValue = _terrainEvalTable_GoRIGHT  [ tableY ][ tableX ];}
					else if (lookPanel == THROUGHABLE_INFO::MUST_TO_LEFT ) { tableValue = _terrainEvalTable_GoLEFT   [ tableY ][ tableX ];}
					else if (lookPanel == THROUGHABLE_INFO::NOT_THROUGHABLE) { tableValue = _terrainEvalTable_NotThrouable[ tableY ][ tableX ];}

					if (tableValue > 0) {
						result += tableValue * (hight * _terrainEval_hightBias);
					} else {
						result += tableValue / (hight * _terrainEval_hightBias);
					}
				}
			}
		}
	}
	if ( myPiece == GRID::WHITE ) {

		// 自分にとっての前後
		int throuable_UP   = THROUGHABLE_INFO::MUST_TO_DOWN;
		int throuable_DOWN = THROUGHABLE_INFO::MUST_TO_UP;

		// 自分にとっての「高さ」
		int hight = y + 1;

		// ゴール位置
		int goalY = _rule.boardHeight - 1;

		// ゴールに達していたら、高得点。
		if ( y == goalY) {
			result = 9000;

		// ゴールに達していないなら…
		} else {

			for ( int defY = -radius; defY <= radius; ++defY ) {
				for ( int defX = -radius; defX <= radius; ++defX ) {

					// 評価値テーブル上の位置
					int tableX = radius + defX;
					int tableY = radius + defY;

					// 盤面上の、この位置を見る
					int lookX = x + defX;
					int lookY = y - defY;

					// 盤をはみ出していたら見ない
					if ( lookY < 0) { continue; }
					if ( lookX < 0) { continue; }
					if ( lookY >= _rule.boardHeight ) { continue; }
					if ( lookX >= _rule.boardWidth  ) { continue; }


					int lookPanel = currentState->_gridList_ThrouableInfo[lookY][lookX];

					int tableValue = 0;
					if      (lookPanel == THROUGHABLE_INFO::THROUGHABLE    ) { tableValue = _terrainEvalTable_Throuable[ tableY ][ tableX ];}
					else if (lookPanel == throuable_UP                 ) { tableValue = _terrainEvalTable_GoUP     [ tableY ][ tableX ];}
					else if (lookPanel == throuable_DOWN               ) { tableValue = _terrainEvalTable_GoDOWN   [ tableY ][ tableX ];}
					else if (lookPanel == THROUGHABLE_INFO::MUST_TO_RIGHT) { tableValue = _terrainEvalTable_GoRIGHT  [ tableY ][ tableX ];}
					else if (lookPanel == THROUGHABLE_INFO::MUST_TO_LEFT ) { tableValue = _terrainEvalTable_GoLEFT   [ tableY ][ tableX ];}
					else if (lookPanel == THROUGHABLE_INFO::NOT_THROUGHABLE) { tableValue = _terrainEvalTable_NotThrouable[ tableY ][ tableX ];}

					if (tableValue > 0) {
						result += tableValue * (hight * _terrainEval_hightBias);
					} else {
						result += tableValue / (hight * _terrainEval_hightBias);
					}
				}
			}
		}
	}
	
	return result;

}


//=========================================================
// ブラックホール判定
//=========================================================
int AI::Evaluate_BlackholeExaminaion ( const int gridList_ThrouableInfo[15][15], int x, int y, int stepCount) {
	
	int result = 0;

	///// いつまでたっても終わらない
	if ( stepCount >= 10 ) {

		// ブラックホールなんじゃないの？
		return EVALUATION::BLACKHOLE::MIGHT_BE_BLACKHOLE;
	}

	int throuableInfo = gridList_ThrouableInfo[y][x];

	///// 無印パネルにたどり着きました
	if ( throuableInfo == THROUGHABLE_INFO::THROUGHABLE) {

		// ブラックホールではない。
		return EVALUATION::BLACKHOLE::FAR_FROM_BLACKHOLE;
	}
	///// パネルがありません。
	if ( throuableInfo == THROUGHABLE_INFO::NOT_THROUGHABLE) {

		// 当面はブラックホールにならない
		return EVALUATION::BLACKHOLE::FREE_FROM_BLACK_HOLE;
	}



	///// 左向き矢印ですよ
	if ( throuableInfo == THROUGHABLE_INFO::MUST_TO_LEFT ) {

		// 端？
		if ( x == 0 ) {

			// ブラックホールです
			return EVALUATION::BLACKHOLE::BLACKHOLE;

		// 端ではないようです
		} else {

			int nextX = x - 1;
			int nextY = y;
			int nextGrid = gridList_ThrouableInfo[nextY][nextX];
			int adverseThrouableRestriction = THROUGHABLE_INFO::MUST_TO_RIGHT;

			// 次のグリッドにパネルがない
			if ( nextGrid == THROUGHABLE_INFO::NOT_THROUGHABLE ) {

				// いつブラックホール化してもおかしくない
				return EVALUATION::BLACKHOLE::CAN_BE_BLACK_HOLE;

			// 次のグリッドが逆向き
			} else if ( nextGrid == adverseThrouableRestriction ) {

				// ブラックホールです
				return EVALUATION::BLACKHOLE::BLACKHOLE;

			// 次のグリッドが無印矢印です
			} else if ( nextGrid == THROUGHABLE_INFO::THROUGHABLE ) {

				// ブラックホールではない
				return EVALUATION::BLACKHOLE::FAR_FROM_BLACKHOLE;

			// 次のグリッドにも何か矢印があるみたいですよ
			} else {

				// 探索を続ける
				return Evaluate_BlackholeExaminaion( gridList_ThrouableInfo, nextX, nextY, stepCount + 1 );
			}
		}
	}
	///// 右向き矢印ですよ
	else if ( throuableInfo == THROUGHABLE_INFO::MUST_TO_RIGHT ) {

		// 端？
		if ( x == _rule.boardWidth - 1 ) {

			// ブラックホールです
			return EVALUATION::BLACKHOLE::BLACKHOLE;

		// 端ではないようです
		} else {

			int nextX = x + 1;
			int nextY = y;
			int nextGrid = gridList_ThrouableInfo[nextY][nextX];
			int adverseThrouableRestriction = THROUGHABLE_INFO::MUST_TO_LEFT;

			// 次のグリッドにパネルがない
			if ( nextGrid == THROUGHABLE_INFO::NOT_THROUGHABLE ) {

				// いつブラックホール化してもおかしくない
				return EVALUATION::BLACKHOLE::CAN_BE_BLACK_HOLE;

			// 次のグリッドが逆向き
			} else if ( nextGrid == adverseThrouableRestriction ) {

				// ブラックホールです
				return EVALUATION::BLACKHOLE::BLACKHOLE;

			// 次のグリッドが無印矢印です
			} else if ( nextGrid == THROUGHABLE_INFO::THROUGHABLE ) {

				// ブラックホールではない
				return EVALUATION::BLACKHOLE::FAR_FROM_BLACKHOLE;

			// 次のグリッドにも何か矢印があるみたいですよ
			} else {

				// 探索を続ける
				return Evaluate_BlackholeExaminaion( gridList_ThrouableInfo, nextX, nextY, stepCount + 1 );
			}
		}
	}
	///// 上向き矢印ですよ
	else if ( throuableInfo == THROUGHABLE_INFO::MUST_TO_UP ) {

		// 端？
		if ( y == 0 ) {

			// ブラックホールです
			return EVALUATION::BLACKHOLE::BLACKHOLE;

		// 端ではないようです
		} else {

			int nextX = x;
			int nextY = y - 1;
			int nextGrid = gridList_ThrouableInfo[nextY][nextX];
			int adverseThrouableRestriction = THROUGHABLE_INFO::MUST_TO_DOWN;

			// 次のグリッドにパネルがない
			if ( nextGrid == THROUGHABLE_INFO::NOT_THROUGHABLE ) {

				// いつブラックホール化してもおかしくない
				return EVALUATION::BLACKHOLE::CAN_BE_BLACK_HOLE;

			// 次のグリッドが逆向き
			} else if ( nextGrid == adverseThrouableRestriction ) {

				// ブラックホールです
				return EVALUATION::BLACKHOLE::BLACKHOLE;

			// 次のグリッドが無印矢印です
			} else if ( nextGrid == THROUGHABLE_INFO::THROUGHABLE ) {

				// ブラックホールではない
				return EVALUATION::BLACKHOLE::FAR_FROM_BLACKHOLE;

			// 次のグリッドにも何か矢印があるみたいですよ
			} else {

				// 探索を続ける
				return Evaluate_BlackholeExaminaion( gridList_ThrouableInfo, nextX, nextY, stepCount + 1 );
			}
		}
	}
	///// 下向き矢印ですよ
	else if ( throuableInfo == THROUGHABLE_INFO::MUST_TO_DOWN ) {

		// 端？
		if ( y == _rule.boardHeight - 1 ) {

			// ブラックホールです
			return EVALUATION::BLACKHOLE::BLACKHOLE;

		// 端ではないようです
		} else {

			int nextX = x;
			int nextY = y + 1;
			int nextGrid = gridList_ThrouableInfo[nextY][nextX];
			int adverseThrouableRestriction = THROUGHABLE_INFO::MUST_TO_UP;

			// 次のグリッドにパネルがない
			if ( nextGrid == THROUGHABLE_INFO::NOT_THROUGHABLE ) {

				// いつブラックホール化してもおかしくない
				return EVALUATION::BLACKHOLE::CAN_BE_BLACK_HOLE;

			// 次のグリッドが逆向き
			} else if ( nextGrid == adverseThrouableRestriction ) {

				// ブラックホールです
				return EVALUATION::BLACKHOLE::BLACKHOLE;

			// 次のグリッドが無印矢印です
			} else if ( nextGrid == THROUGHABLE_INFO::THROUGHABLE ) {

				// ブラックホールではない
				return EVALUATION::BLACKHOLE::FAR_FROM_BLACKHOLE;

			// 次のグリッドにも何か矢印があるみたいですよ
			} else {

				// 探索を続ける
				return Evaluate_BlackholeExaminaion( gridList_ThrouableInfo, nextX, nextY, stepCount + 1 );
			}
		}
	}

	return result;
}

//==========================================================
// 地形の価値を見定める：地形を変えない指し手だけで探索してみる
//==========================================================
int AI::Evaluate_Terrain ( InnerExpression *currentState, int forThisTurnsView, int depth ) {

	/*
		depth は 奇数を指定しておくのが、
		両者ともに同じ回数ずつ指す、という意味で、公平な判定になるのかも。
	*/
	/*
	// 再帰禁止フラグ → 評価値「0」で素通りする
	if (_recursionBlock_Evaluate_Terrain) {
		return 0;
	}
	_recursionBlock_Evaluate_Terrain = true;
	*/

	int result = 0;

	// オプション引数で、地形を変えない手だけで探索することを指定する
	DigInto_Window_OptionArgs optionArgs;
	optionArgs.isCurrentTerrainOnly   = true;
	optionArgs.isDoShallowSearchPrune = false;
	optionArgs.evaluateFunctionFlgs   = EVALUATION::RUTIN_SET::FOR_TERRAIN_SEARCH;

	//MemoryPool memoryPool;
	//MemoryPool_ForTreeSearch_Setup( memoryPool, 100 );

	// _memoryPool_ForTerrainSearch が用意されている前提。

	pair<Move, int> searchResult = 
		DigInto_Window ( _memoryPool_ForTerrainSearch, currentState, depth, forThisTurnsView, 99999, -99999, optionArgs);

	//MemoryPool_ForTreeSearch_End ( memoryPool );

	//printf("評価値：%d 全 %d 手\n", searchResult.second, optionArgs.allReadState);

	// 結果のうち、「指し手」は無意味と考えてよい？
	searchResult.first;

	// 得られた評価値のみを結果とする
	result = searchResult.second;


	//_recursionBlock_Evaluate_Terrain = false;
	return result;
}


//==========================================================
// 駒評価素子を使った評価
// 大元では、駒があった位置を指定してこれを呼ぶ。
// 以降の再帰時には、子要素すべてを呼ぶ。
//【引数】
// int x, y 座標。この位置を中心とする
//==========================================================
int AI::Evaluate_PEE ( const InnerExpression* innerExpression, const int &myPieceColor, int x, int y, const vector<PieceEvaluateElement> &PEE_List ) {

	int result = 0;

	// 手番による加点値反転フラグ
	//int pointReverseSign = ( innerExpression->_currentTurn == forThisTurn ) ? 1 : -1;

	// 手番による上下反転：黒の上は 上、白の上は下。
	//int throuable_UP   = ( myPieceColor == GRID::BLACK ) ? THROUGHABLE_INFO::MUST_TO_UP   : THROUGHABLE_INFO::MUST_TO_DOWN;
	//int throuable_DOWN = ( myPieceColor == GRID::BLACK ) ? THROUGHABLE_INFO::MUST_TO_DOWN : THROUGHABLE_INFO::MUST_TO_UP;
	/*
		↑「上下読み替え」をするので、これは不要っぽい。
	*/

	///// 評価素子リストでループ：大元では、６枚あるはず。
	for (unsigned int i = 0; i < PEE_List.size(); ++i) {

		// 素子
		PieceEvaluateElement pee = PEE_List[i];

		// 対象となる座標
		int targetGridX = x + pee.x;
		int targetGridY = ( myPieceColor == GRID::BLACK ) ? y + pee.y : y - pee.y;

		// 条件が合った？
		bool isMatch = false;

		// この座標、盤の範囲外だったりする？
		if ( targetGridX < 0 || targetGridY < 0 || targetGridX >= _rule.boardWidth || targetGridY >= _rule.boardHeight ) {

			// 範囲外を評価することにしてる？
			if ( pee.throuableInfo == THROUGHABLE_INFO::EX::OUT_OF_BOARD) {

				result += pee.point;
				isMatch = true;
			}

		// 盤の範囲内です。
		} else {

			// 対象グリッドの駒を見ます
			if ( pee.throuableInfo >= THROUGHABLE_INFO::EX::PIECE_EMPTY && pee.throuableInfo <= THROUGHABLE_INFO::EX::PIECE_OPPONENT ) {
				
				int targetPiece = innerExpression->_gridList_Piece[ targetGridY ][ targetGridX ];

				// 駒なし、を評価します
				if (pee.throuableInfo == THROUGHABLE_INFO::EX::PIECE_EMPTY ) { 

					// 対象位置、駒なし、ですか？
					if ( targetPiece == GRID::ENPTY  ) {
						result += pee.point;
						isMatch = true;
					}
				}
				// 味方の駒、を評価します
				else if (pee.throuableInfo == THROUGHABLE_INFO::EX::PIECE_MY ) { 

					// 対象位置に味方の駒あり、ですか？
					if ( myPieceColor == GRID::BLACK && targetPiece == GRID::BLACK ) {
						result += pee.point;
						isMatch = true;
					}
					else if ( myPieceColor == GRID::WHITE && targetPiece == GRID::WHITE ) {
						result += pee.point; 
						isMatch = true;
					}
				}
				// 敵の駒、を評価します。
				else if (pee.throuableInfo == THROUGHABLE_INFO::EX::PIECE_OPPONENT ) { 
					
					// 対象位置に敵の駒あり、ですか？
					if ( myPieceColor == GRID::BLACK && targetPiece == GRID::WHITE ) {
						result += pee.point; 
						isMatch = true;
					}
					if ( myPieceColor == GRID::WHITE && targetPiece == GRID::BLACK ) {
						result += pee.point; 
						isMatch = true;
					}
				}
			}
			// 対象グリッドのパネルを見ます
			else if ( pee.throuableInfo >= THROUGHABLE_INFO::EX::NOT_THROUGHABLE && pee.throuableInfo <= THROUGHABLE_INFO::EX::MUST_TO_DOWN ) {
				
				
				//printf ("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa(%d, %d)[%d]\n", targetGridX, targetGridY, pee.throuableInfo);
				
				int targetThrouableInfo = innerExpression->_gridList_ThrouableInfo[ targetGridY ][ targetGridX ];

				/*
					「上下」パネルを駒の白黒で読み替える
					素子内では、自分視点で「上下」となっている。
					つまり、実際の上下を、白の場合は逆に読み替える
				*/
				if ( myPieceColor == GRID::WHITE ) {
					
					// 盤面上 の 上矢印は
					if (targetThrouableInfo == THROUGHABLE_INFO::MUST_TO_UP) {
						// 白にとっては「下向き」ですよ。
						targetThrouableInfo = THROUGHABLE_INFO::MUST_TO_DOWN;
					}
					// 盤面上 の 下矢印は
					else if (targetThrouableInfo == THROUGHABLE_INFO::MUST_TO_DOWN) {
						// 白にとっては「上向き」ですよ。
						targetThrouableInfo = THROUGHABLE_INFO::MUST_TO_UP;
					}
				}
				
				// THROUGHABLE_INFO と THROUGHABLE_INFO::EX ならば互換性があるつもり
				if ( targetThrouableInfo == pee.throuableInfo) { 
					result += pee.point; 
					isMatch = true;
					
					/*
					if ( targetThrouableInfo == THROUGHABLE_INFO::NOT_THROUGHABLE ) {
						printf ( "NOT_THROUGHABLE(%d, %d):pee.point[%d]\n",targetGridX,targetGridY, pee.point );
					}
					*/
				}

			// 相手陣地？
			} else if ( pee.throuableInfo == THROUGHABLE_INFO::EX::OPPONENT_AREA ) {

				// 黒にとっては、盤の上の端が相手陣地ですよ。
				if ( myPieceColor == GRID::BLACK ) {
					
					if (targetGridY == 0) {
						result += pee.point;
						isMatch = true;
					}

				// 白にとっては、盤の下の端が相手陣地ですよ。
				} else if ( myPieceColor == GRID::WHITE ) {

					if (targetGridY == _rule.boardHeight - 1) {
						result += pee.point;
						isMatch = true;
					}
				}
				
			// カードを取れるかどうか
			} else if ( pee.throuableInfo == THROUGHABLE_INFO::EX::CARD_TAKEABLE ) {
				
				//printf ("カードを取れるかどうかの判定（THROUGHABLE_INFO::EX::CARD_TAKEABLE）\n");
				
				int                          targetPiece = innerExpression->_gridList_Piece[ targetGridY ][ targetGridX ];
				InnerExpression::CardOnBoard targetCard  = innerExpression->_gridList_Card [ targetGridY ][ targetGridX ];
				
				// カード自体がある＝その位置に駒があれば取れる
				if ( targetCard.cardId != -1 ) 
				{
					// 自分の駒が乗っている＝取れる
					if ( (targetPiece == GRID::BLACK || targetPiece == GRID::WHITE ) && myPieceColor == targetPiece ) {
						
						result += pee.point;
						isMatch = true;
						
						//printf ("自分の駒(%s)が乗っている＝取れる(%d, %d) -> +%d\n",( (myPieceColor==GRID::BLACK?"BLACK":"WHITE") ) , targetGridX, targetGridY, pee.point);
						
					}
					// 相手の駒が乗っている＝取られる
					else if ( (targetPiece == GRID::BLACK || targetPiece == GRID::WHITE ) && myPieceColor != targetPiece ) {
						
						result -= pee.point;
						isMatch = true;
						
						//printf ("相手の駒が乗っている＝取れる(%d, %d) -> -%d\n", targetGridX, targetGridY, pee.point);
					}
				}
				
			}
		}

		// 条件が合っていた
		if ( isMatch ) {

			// 周辺素子が定義されていれば再帰
			if (pee.relativeList.size() > 0) { // 0件判定はしなくて、内部でループに入らないはずですが、再帰呼び出しのオーバーヘッド？ を抑える意図。一応。

				// myPieceColor はずっと受け継ぐ。
				result += Evaluate_PEE( innerExpression, myPieceColor, targetGridX, targetGridY, pee.relativeList);
			}
		}
	}

	return result;
}

//==========================================================
// 手札を評価
//==========================================================
int AI::Evaluate_HandCards ( const InnerExpression* innerExpression, int forThisTurn ) {

	int result = 0;

	map<int, Card>::iterator findIt;

	///// 黒の手札ポイント
	int pointBlack = 0;
	for (int i = 0; i < innerExpression->_handCards_BLACK_Index; ++i) {

		int cardId = innerExpression->_handCards_BLACK[i];

		 findIt = _cardList.find( cardId );
		 if ( findIt != _cardList.end() ) {

			 Card card = findIt->second;

			 // カード自体のポイント
			 pointBlack += _parameters_PieceEvaluation.handCardEvaluate.point_Card;

			 for (unsigned int n = 0; n < card.gridAffectionList.size(); ++n) {

				 GridAffection gridAffection = card.gridAffectionList[n];

				// 無印パネル
				if ( gridAffection.movableDirection == DIRECTION::NONE ) {
					pointBlack += _parameters_PieceEvaluation.handCardEvaluate.point_Throughable;
				// 矢印つきパネル
				} else {
					pointBlack += _parameters_PieceEvaluation.handCardEvaluate.point_MustGoSomeway;
				}
			 }
		 
		 } else {
			 printf("指定のカードがありません。ID:%d 黒\n", cardId);
		 }
	}

	///// 白の手札ポイント
	int pointWhite = 0;
	for (int i = 0; i < innerExpression->_handCards_WHITE_Index; ++i) {

		int cardId = innerExpression->_handCards_WHITE[i];

		 findIt = _cardList.find( cardId );
		 if ( findIt != _cardList.end() ) {

			 Card card = findIt->second;

			 // カード自体のポイント
			 pointWhite += _parameters_PieceEvaluation.handCardEvaluate.point_Card;

			 for (unsigned int n = 0; n < card.gridAffectionList.size(); ++n) {

				 GridAffection gridAffection = card.gridAffectionList[n];

				// 無印パネル
				if ( gridAffection.movableDirection == DIRECTION::NONE ) {
					pointWhite += _parameters_PieceEvaluation.handCardEvaluate.point_Throughable;
				// 矢印つきパネル
				} else {
					pointWhite += _parameters_PieceEvaluation.handCardEvaluate.point_MustGoSomeway;
				}
			 }
		 
		 } else {
			 printf("指定のカードがありません。ID:%d 黒\n", cardId);
		 }
	}


	///// 相手の得点をマイナスとする。
	if ( forThisTurn == TURN::BLACK ) {
		pointWhite *= -1;
	} else if ( forThisTurn == TURN::WHITE ) {
		pointBlack *= -1;
	}

	///// 加算合成
	result = pointBlack + pointWhite;

	return result;
}
