//=========================================================
//
// AI_実装ファイル_評価関数関係
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"


using namespace AI_Modules;


//==========================================================
// 評価関数：セット：地形判定の探索ツリーの末端で使うもの
//==========================================================
int AI::Evaluate_Set_For_TerrainSearch ( InnerExpression* innerExpression, int forThisTurn) {

	int result = 0;

	///// 勝ってたら評価最高
	/*
	int winResult = 0;
	if ( Evaluate_IsWin ( winResult, innerExpression, forThisTurn) ) {

		// 勝負がついているなら、もう考えない。
		result = winResult;
		return result;
	}
	*/
	///// 駒の価値
	/*
		駒の価値評価の中でも、相手陣地に入っていたら高さ評価で高得点にしているのですよね。
		それって、上の、勝ってたら評価最高、というのと かぶってる。それは大丈夫？

		手番の影響を受けて大幅に変わってしまうようなので、
		とりあえす、上の、勝ってたら評価最高、というのは、封印してみる。
	*/
	result += Evaluate_Piece ( innerExpression, forThisTurn );

	///// 柱と傘での評価をする？
	if ( _parameters_PieceEvaluation.pcEvalInfo.isEnabled && _parameters_PieceEvaluation.pcEvalInfo.applyBias_forTerrain != 0 ) {

		//int pcEval = Evaluate_PC ( innerExpression, forThisTurn );
		int pcEval = Evaluate_PC_S ( innerExpression, forThisTurn );
		pcEval = (int)((float)pcEval * (float)_parameters_PieceEvaluation.pcEvalInfo.applyBias_forTerrain / 100.0f);
		result += pcEval;
	}
	
	///// 位置での評価をする？
	if ( _parameters_PieceEvaluation.parameters_PositionEvaluation.isEnabled && _parameters_PieceEvaluation.parameters_PositionEvaluation.applyBias_forTerrain != 0 ) {
		
		int positionEval = Evaluate_PositionForWholeBoard (innerExpression, forThisTurn );
		positionEval= (int)((float)positionEval * (float)_parameters_PieceEvaluation.parameters_PositionEvaluation.applyBias_forTerrain / 100.0f);
		result += positionEval;
		
	}
	///// 特定位置での評価をする？
	if ( _parameters_PieceEvaluation.positionCareSettings.isEnabled && _parameters_PieceEvaluation.positionCareSettings.applyBias_forTerrain != 0 ) {
		
		int specificPositionsCareEval = Evaluate_PositionForSpecificPositions ( innerExpression, forThisTurn );
		specificPositionsCareEval = (int)((float)specificPositionsCareEval * (float)_parameters_PieceEvaluation.positionCareSettings.applyBias_forTerrain / 100.0f);
		result += specificPositionsCareEval;
		
	}
	///// 論理距離比較での評価をする？
	if ( _parameters_PieceEvaluation.parameters_DistanceToGoal.distanceDiffScreening.isEnabled && _parameters_PieceEvaluation.parameters_DistanceToGoal.distanceDiffScreening.bias_TerrainEvaluate != 0 ) {
		
		int dsDiffEval = Evaluate_DistanceDiff ( innerExpression, forThisTurn );
		dsDiffEval = (int)((float)dsDiffEval * (float)_parameters_PieceEvaluation.parameters_DistanceToGoal.distanceDiffScreening.bias_TerrainEvaluate / 100.0f);
		result += dsDiffEval;
		
	}

	///// 王手判定
	//result += Evaluate_Check ( innerExpression, forThisTurn );

	return result;
}

//==========================================================
// 評価関数：セット：中盤用：探索深度は浅めにして、評価判定には地形判定を使う
//==========================================================
int AI::Evaluate_Set_For_MiddleStage ( InnerExpression* innerExpression, int forThisTurn ) {

	int result = 0;

	///// 勝ってたら最高評価
	int winResult = 0;
	if ( Evaluate_IsWin ( winResult, innerExpression, forThisTurn) ) {

		// 勝負がついているなら、もう考えない。
		result = winResult;
		return result;
	}

	///// 地形判定
	result += Evaluate_Terrain ( innerExpression, forThisTurn, 3);

	return result;
}


//==========================================================
// 評価関数：セット：ツリー探索内で、選択肢が多いときの絞込みに使う評価
//==========================================================
int AI::Evaluate_Set_For_PickUp ( InnerExpression* innerExpression, int forThisTurn ) {

	int result = 0;

	///// 勝ってたら最高評価
	int winResult = 0;
	if ( Evaluate_IsWin ( winResult, innerExpression, forThisTurn) ) {

		// 勝負がついているなら、もう考えない。
		result = winResult;
		return result;
	}

	///// 地形判定： 0 → 自分が一回指すだけ
	result += Evaluate_Terrain ( innerExpression, forThisTurn, 0);

	return result;

}

