//=========================================================
//
// AI_実装ファイル_評価関数関係
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"


using namespace AI_Modules;


//==========================================================
// 手札を補充する手を評価
//==========================================================
int AI::Evaluate_TakeCardMove ( InnerExpression* innerExpression, Move &takeCardMove, int forThisTurn) {

	int result = 0;

	// カードを取る手ですよね？
	if ( takeCardMove.action != MOVE::ACTION::TAKE_CARD ) {
		return result;
	}
	
	string moveToString = ShowMoveToConsole( takeCardMove, true );
	printf ("手札を補充する手を評価：%s \n", moveToString.c_str());

	///// どのカードを取るのですか？
	InnerExpression::CardOnBoard takenCardOnBoard = innerExpression->_gridList_Card [ takeCardMove.takeCardY ][ takeCardMove.takeCardX ];

	///// カードを取った次の局面を求めて、二度指しの準備をする
	InnerExpression *virtualNextState = innerExpression->GetNext( takeCardMove );
	virtualNextState->_currentTurn = SwitchTurn( virtualNextState->_currentTurn );

	///// カードを設置する手だけを取得
	vector<Move> allMoveList;
	virtualNextState->GetLegalNextMoveList( allMoveList, MOVE::ACTION::PUT_CARD );

	///// 絞込み：詰み探索：一手で勝つ手は原理的にありえないので、処理としては無駄といえる。
	GetNextMove_PEE_Test_PreSearch( allMoveList, virtualNextState );


	///// ここで選択肢が１つだけになっている＝詰みがあった
	/*
		ここで指定された「カードを補充」しなくても詰みがあったのかもしれない？
		その場合は、ここを呼び出す前に結論が出ているので、ここにくることはない。
	*/
	if ( allMoveList.size() == 1 ) {

		printf ("詰めろを発見：%s\n", ShowMoveToConsole ( allMoveList[0], true).c_str() );

		delete virtualNextState;

		// 詰めろの評価値を返す
		return _parameters_PieceEvaluation.takeCardMoveEvaluate.nextCheckMateValue;
	}






	///// 以降、評価関数を使う。
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(1500);

	///// 手に評価値をつける。後でソートして上位のを使う。
	vector< pair<Move, int> > evaluatedMoveList;

	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		// 取ったカードを置くのじゃないなら、素通りしますよ
		if ( allMoveList[i].putCardId != takenCardOnBoard.cardId) {
			continue;
		}

		// 同じ位置に同じ角度で置き戻すなら、素通りしますよ
		if 
		( 
			allMoveList[i].putCardX == takeCardMove.takeCardX && 
			allMoveList[i].putCardY == takeCardMove.takeCardY &&
			allMoveList[i].putCardDirection == takenCardOnBoard.direction
		) 
		{
			continue;
		}

		///// 次の局面
		InnerExpression *nextState = virtualNextState->GetNext( allMoveList[i], allocator );

		int eval = 0;


		///// 地形評価を使う？
		int terrainDepth = _parameters_PieceEvaluation.terrainSearchUsage.depth;
		if ( terrainDepth >= 0) {

			eval = Evaluate_Terrain( nextState, innerExpression->_currentTurn, terrainDepth);

			// バイアス
			eval *= (float)_parameters_PieceEvaluation.terrainSearchUsage.bias_TerrainEvaluate / 100.0f;

			// PEEのバイアスが 0 以上ならば、使う。
			if ( _parameters_PieceEvaluation.terrainSearchUsage.bias_PeeEvaluate > 0 ) {

				int evalPEE = Evaluate_Piece( nextState, innerExpression->_currentTurn );
				evalPEE *= (float)_parameters_PieceEvaluation.terrainSearchUsage.bias_PeeEvaluate / 100.0f;

				eval += evalPEE;
			}

		///// 地形評価を使わない
		} else {

			///// 駒の価値だけで局面を評価
			eval = Evaluate_Piece( nextState, innerExpression->_currentTurn );
		}

		///// 柱と傘での評価をする？
		if ( _parameters_PieceEvaluation.pcEvalInfo.isEnabled ) {

			eval += Evaluate_PC ( nextState, innerExpression->_currentTurn );
		}

		///// 手札評価をする？
		if ( _parameters_PieceEvaluation.handCardEvaluate.isEnabled ) {

			eval += Evaluate_HandCards ( nextState, innerExpression->_currentTurn );
		}

		///// 評価値をつけて、リストに入れておく。あとでソート
		pair<Move, int> evaluatedMove;
		evaluatedMove.first  = allMoveList[i];
		evaluatedMove.second = eval;

		evaluatedMoveList.push_back( evaluatedMove );
	}

	delete allocator;

	///// 結果をソート
	GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate ( evaluatedMoveList );


	///// 最上位の評価値を戻り値とする。
	if ( evaluatedMoveList.size() > 0 ) {
	
		result = evaluatedMoveList[0].second;
		printf ("最善手 %s  評価値 %d\n", ShowMoveToConsole(evaluatedMoveList[0].first, true).c_str(), result);
	
	} else {
		result = 0;
		printf ("評価値 %d\n", result);
	}


	delete virtualNextState;

	return result;
}
