//=========================================================
//
// AI_実装ファイル_詰み探索関連
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"


using namespace AI_Modules;


//=========================================================
// 詰み手順を教えて
//=========================================================
vector<Move> AI::GetMateProcedure( InnerExpression *currentState, int level, bool isCheckAlreadyMate ) {

	vector<Move> result;


	///// すでに勝ってる？：アルゴリズムによっては、処理速度の無駄になる？
	if (isCheckAlreadyMate) {

		vector<Move> winMoveList;
		currentState->GetLegalWinMoveList(winMoveList);
		if (winMoveList.size() > 0 ) {
			result.push_back( *(winMoveList.begin()) );
			printf("すでに勝ってる");
			return result;
		}
	}

	printf("◆ 開始 -----------------------------\n");
	//DWORD t =  timeGetTime();
	auto t = std::chrono::system_clock::now();

	MemoryPool memoryPool;
	MemoryPool_ForTreeSearch_Setup ( memoryPool, 100 );
	/*
	_memoryPool_ForTreeSearch = new char[ 100 * sizeof(InnerExpression) ];
	_memoryPool_ForTreeSearch_currentIt = _memoryPool_ForTreeSearch;
	ZeroMemory(_memoryPool_ForTreeSearch, 100 * sizeof(InnerExpression));
	*/

	vector<Move> route;
	_allReadState = 0;

	pair<int, vector<Move>> returnResult = DigInto_Mate( memoryPool, currentState, level, currentState->_currentTurn, route, 9000, -9000 );
	int eval = returnResult.first;

	//DWORD passTime = timeGetTime() - t;
	auto passTime = std::chrono::system_clock::now() - t;
	
	printf( "%d ms 読んだ手数：%d\n", passTime, _allReadState );

	if ( eval > 9000 ) {	
		printf( "詰み評価：%d(%s)\n", eval, "詰む");
		result = returnResult.second;
	} else {
		printf( "詰み評価：%d(%s)\n", eval, "詰まない" );
	}

	MemoryPool_ForTreeSearch_End ( memoryPool );
	//delete[] _memoryPool_ForTreeSearch;

	return result;
}

//=========================================================
// 掘り進む：詰み探索用
//=========================================================
pair<int, vector<Move>> AI::DigInto_Mate( MemoryPool &memoryPool, InnerExpression *currentState, int level, int fromThisTurnsViewPoint, vector<Move> &route, int windowUpperLimit, int windowLowerLimit) {

	///// 戻り値：評価値（詰む詰まない基準）
	pair<int, vector<Move>> result;
	result.first = 0;
	result.second.clear();

	// これでいいの？ 戻り値とか。そもそもここにくることとか。
	if (level <= 0) {
		//printf(" level = %d\n ", level);
		return result;
	}

	///// 自分の手番：一階層目は自分の番である前提
	if ( currentState->_currentTurn == fromThisTurnsViewPoint ) {

		///// 全指し手を取得
		vector<Move> allMoveList;
		//currentState->GetLegalNextMoveList( allMoveList );
		InnerExpression::GetCheckMoveList ( currentState, allMoveList );


		pair<int, vector<Move>> bestReturnResult;
		bestReturnResult.first = windowLowerLimit;

		///// 王手になる手が一つでもあった？
		bool isCheckMoveExist = false;
		//int checkMoveNumber = 0;

		vector<Move> moveList;

		bool isUpdate = false;

		///// 可能な指し手でループ
		for (unsigned int i = 0; i < allMoveList.size(); ++i) {

			///// スレッド経由で来ている場合、他のスレッドで詰みが見つかったなら強制終了。
			if ( _numberOfThread > 0 ) {
				if ( _DigInto_Mate_Thread_Signal_isDiscoverMateProcedure ) {
					//printf(" 詰みがみつかったの？ じゃあ、撤収しますね\n ");
					return result;
				}
			}

			// 指した結果
			Move &move = allMoveList[i];

			//InnerExpression *nextState = currentState->GetNext( move );
			
			InnerExpression *nextState;
			//MemoryPool_ForTreeSearch_Allocate ( nextState );
			MemoryPool_ForTreeSearch_Allocate ( memoryPool, nextState );
			
			currentState->GetCopy( nextState );
			nextState->GoNext( move );

			// 王手になるか？
			bool isCheck = false;
			{
				// 相手に上がる手がないなら、この手で王手していい。
				moveList.clear();
				nextState->GetLegalWinMoveList( moveList );
				
				if ( moveList.size() == 0 ) {

					isCheck = true;

					/*
					// もしこのまま二手指しできるとして、こちらに上がる手があるか？
					nextState->_currentTurn = SwitchTurn(nextState->_currentTurn);

					moveList.clear();
					nextState->GetLegalWinMoveList( moveList );
					isCheck = (moveList.size() > 0);

					nextState->_currentTurn = SwitchTurn(currentState->_currentTurn);
					*/
				}
			}

			// 王手になるなら続ける
			if ( isCheck ) {
				
				//++checkMoveNumber;

				isCheckMoveExist = true;

				route.push_back( move );
				pair<int, vector<Move>> returnResult = DigInto_Mate( memoryPool, nextState, level - 1, fromThisTurnsViewPoint, route, windowUpperLimit, bestReturnResult.first);
				route.pop_back();

				// 更新判定
				//if ( returnResult.first > bestReturnResult.first || i == 0) {
				if ( returnResult.first > bestReturnResult.first || !isUpdate) {
					
					bestReturnResult = returnResult;

					isUpdate = true;
				}
			}

			//delete nextState;
			//MemoryPool_ForTreeSearch_Free ( nextState );
			MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );

			// 枝狩り？
			if ( bestReturnResult.first > windowUpperLimit ) {
				break;
			}

		} // 可能な指し手でループ

		//printf("王手になる手数：%d\n", checkMoveNumber);

		// 王手になる手があった
		if (isCheckMoveExist) {

			result = bestReturnResult;

		// 王手になる手が一つもなかった
		} else {
			result.first = 0;
			result.second = route;
			//++_allReadState;
		}

	///// 相手の手番：２階層目以降という前提。つまり、今王手がかかっているはず。
	} else {

		///// 全指し手を取得
		vector<Move> allMoveList;
		currentState->GetLegalNextMoveList( allMoveList, MOVE::ACTION::GUARDABLE );

		///// 王手を一度でも解除できたか？
		bool isAvoidCheck = false;
		//int avoidCheckNumber = 0;

		pair<int, vector<Move>> bestReturnResult;
		//bestReturnResult.first = 99999;
		bestReturnResult.first = windowUpperLimit;

		vector<Move> moveList;

		bool isUpdate = false;

		///// 可能な指し手でループ
		for (unsigned int i = 0; i < allMoveList.size(); ++i) {

			///// スレッド経由で来ている場合、他のスレッドで詰みが見つかったなら強制終了。
			if ( _numberOfThread > 0 ) {
				if ( _DigInto_Mate_Thread_Signal_isDiscoverMateProcedure ) {
					//printf(" 詰みがみつかったの？ じゃあ、撤収しますねー\n ");
					return result;
				}
			}

			// 指した結果
			Move &move = allMoveList[i];
			
			//InnerExpression *nextState = currentState->GetNext( move );
			InnerExpression *nextState;
			//MemoryPool_ForTreeSearch_Allocate ( nextState );
			MemoryPool_ForTreeSearch_Allocate ( memoryPool, nextState );

			currentState->GetCopy( nextState );
			nextState->GoNext( move );

			// 王手を解除することができたか？ つまり、この時点で王手が継続しているかどうか、ということ。
			bool isCheck = false;
			{
				// すでに自分が勝っているなら、それもよし。
				if (IsWinMove( move ) ) {

					// 王手はされているけど、こっちが先に勝ってしまった、ということですよね。
					// じゃあチェックフラグは解除して、深さも、底についたことにしてしまってはどうか？
					// つまり「受け切った」といいたい。相手の攻め失敗。
					//isCheck = true;
					isCheck = false; // 先に勝ったのだから相手の王手も無効
					level   = 0;     // 底についたことにする。

				} else {

					// 相手に上がる手がないこと
					moveList.clear();
					nextState->GetLegalWinMoveList( moveList );

					isCheck = ( moveList.size() > 0);
				}
			}

			// 王手を解除できなかった
			if ( isCheck ) {
				
				//delete nextState;
				//MemoryPool_ForTreeSearch_Free ( nextState );
				MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );
				continue;
			}

			// 一度でも王手を解除できた。
			isAvoidCheck = true;
			//++avoidCheckNumber;

			// まだ掘る？：王手を解除できた

			pair<int, vector<Move>> returnResult;


			// すでに底に達している → 受けきった、ということ
			if ( level <= 0 ) {

				//++_allReadState;

				// 一つでも受けきる手順があったならば、詰みなしということ
				result.first  = 0;
				result.second = route;
				result.second.push_back(move);

				//delete nextState;
				//MemoryPool_ForTreeSearch_Free ( nextState );
				MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );

				return result;

			// まだ底に達していない → 続ける
			} else {

				route.push_back( move );
				returnResult = DigInto_Mate( memoryPool, nextState, level - 1, fromThisTurnsViewPoint, route, bestReturnResult.first, windowLowerLimit);
				route.pop_back();
				
				// 更新判定
				//if ( returnResult.first < bestReturnResult.first || i == 0) {
				if ( returnResult.first < bestReturnResult.first || !isUpdate ) {

					bestReturnResult = returnResult;

					isUpdate = true;
				}
			}

			//delete nextState;
			//MemoryPool_ForTreeSearch_Free ( nextState );
			MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );

			// 枝狩り？
			if ( bestReturnResult.first < windowLowerLimit ) {
				break;
			}

		} // 可能な指し手でループ

		// 一度も王手を解除できなかった＝詰み成立
		if ( !isAvoidCheck  ) {
			//++_allReadState;
			
			result.first = 9999 - currentState->_totalMoveCount;
			result.second = route;

		// 一度でも王手を解除した
		} else {
			result = bestReturnResult;
		}

		if (result.first == 9000) {
			printf( " result.first == %d : %s, %s level == %d\n ", result.first, (isUpdate) ? "更新された" : "更新されなかった", (isAvoidCheck) ? "王手を解除した":"王手を解除できなかった", level );
		}

		//printf("王手を解除できる指し手の数：%d\n", avoidCheckNumber);
	}

	return result;
}









//=========================================================
// 掘り進む：詰み探索用
//=========================================================
void AI::DigInto_Mate( MemoryPool &memoryPool, pair<int, vector<Move>> &result, InnerExpression *currentState, int level, int fromThisTurnsViewPoint, vector<Move> &route, int windowUpperLimit, int windowLowerLimit) {

	if (level < 0) {
		//printf(" *level = %d\n ", level);
		return;
	}

	///// 戻り値：評価値（詰む詰まない基準）
	//pair<int, vector<Move>> result;
	result.first = 0;
	result.second.clear();

	///// 自分の手番：一階層目は自分の番である前提
	if ( currentState->_currentTurn == fromThisTurnsViewPoint ) {

		///// 全指し手を取得：攻め用の手のみ
		vector<Move> allMoveList;
		//currentState->GetLegalNextMoveList( allMoveList, MOVE::ACTION::CHECKABLE );
		InnerExpression::GetCheckMoveList( currentState, allMoveList );


		pair<int, vector<Move>> bestReturnResult;
		bestReturnResult.first = windowLowerLimit;

		///// 王手になる手が一つでもあった？
		bool isCheckMoveExist = false;
		//int checkMoveNumber = 0;

		vector<Move> moveList;

		bool isUpdate = false;

		///// 可能な指し手でループ
		for (unsigned int i = 0; i < allMoveList.size(); ++i) {

			///// スレッド経由で来ている場合、他のスレッドで詰みが見つかったなら強制終了。
			if ( _numberOfThread > 0 ) {
				if ( _DigInto_Mate_Thread_Signal_isDiscoverMateProcedure ) {
					//printf(" 詰みがみつかったの？ じゃあ、撤収しますね\n ");
					return;
				}
			}

			// 指した結果
			Move &move = allMoveList[i];

			//InnerExpression *nextState = currentState->GetNext( move );
			
			InnerExpression *nextState;
			//MemoryPool_ForTreeSearch_Allocate ( nextState );
			MemoryPool_ForTreeSearch_Allocate ( memoryPool, nextState );
			
			currentState->GetCopy( nextState );
			nextState->GoNext( move );

			// 王手になるか？
			bool isCheck = false;
			{
				// 相手に上がる手がないなら、よし。
				moveList.clear();
				nextState->GetLegalWinMoveList( moveList );
				
				if ( moveList.size() == 0 ) {
					isCheck = true;
				}
			}

			// 王手になるなら続ける
			if ( isCheck ) {
				
				//++checkMoveNumber;

				isCheckMoveExist = true;

				route.push_back( move );
				
				pair<int, vector<Move>> returnResult;
				DigInto_Mate( memoryPool, returnResult, nextState, level - 1, fromThisTurnsViewPoint, route, windowUpperLimit, bestReturnResult.first);
				
				route.pop_back();

				// 更新判定
				//if ( returnResult.first > bestReturnResult.first || i == 0) {
				if ( returnResult.first > bestReturnResult.first || !isUpdate) {
					
					bestReturnResult = returnResult;

					isUpdate = true;
				}
			}

			//delete nextState;
			//MemoryPool_ForTreeSearch_Free ( nextState );
			MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );

			// 枝狩り？
			if ( bestReturnResult.first > windowUpperLimit ) {
				break;
			}

		} // 可能な指し手でループ

		//printf("王手になる手数：%d\n", checkMoveNumber);

		// 王手になる手があった
		if (isCheckMoveExist) {

			result = bestReturnResult;

		// 王手になる手が一つもなかった
		} else {
			result.first = 0;
			result.second = route;
			//++_allReadState;
		}

	///// 相手の手番：２階層目以降という前提。つまり、今王手がかかっているはず。
	} else {

		///// 全指し手を取得：受け用の手のみ
		vector<Move> allMoveList;
		currentState->GetLegalNextMoveList( allMoveList, MOVE::ACTION::GUARDABLE );

		///// 王手を一度でも解除できたか？
		bool isAvoidCheck = false;
		//int avoidCheckNumber = 0;

		pair<int, vector<Move>> bestReturnResult;
		//bestReturnResult.first = 99999;
		bestReturnResult.first = windowUpperLimit;

		vector<Move> moveList;

		bool isUpdate = false;

		///// 可能な指し手でループ
		for (unsigned int i = 0; i < allMoveList.size(); ++i) {

			///// スレッド経由で来ている場合、他のスレッドで詰みが見つかったなら強制終了。
			if ( _numberOfThread > 0 ) {
				if ( _DigInto_Mate_Thread_Signal_isDiscoverMateProcedure ) {
					//printf(" 詰みがみつかったの？ じゃあ、撤収しますねー\n ");
					return;
				}
			}

			// 指した結果
			Move &move = allMoveList[i];
			
			//InnerExpression *nextState = currentState->GetNext( move );
			InnerExpression *nextState;
			//MemoryPool_ForTreeSearch_Allocate ( nextState );
			MemoryPool_ForTreeSearch_Allocate ( memoryPool, nextState );

			currentState->GetCopy( nextState );
			nextState->GoNext( move );

			// 王手を解除することができたか？
			bool isCheck = false;
			{
				// すでに自分が勝っているなら、それもよし。
				if (IsWinMove( move ) ) {
					isCheck = true;

				} else {

					// 相手に上がる手がないこと
					moveList.clear();
					nextState->GetLegalWinMoveList( moveList );

					isCheck = ( moveList.size() > 0);
				}
			}

			// 王手を解除できなかった
			if ( isCheck ) {
				
				//delete nextState;
				//MemoryPool_ForTreeSearch_Free ( nextState );
				MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );
				continue;
			}

			// 一度でも王手を解除できた。
			isAvoidCheck = true;
			//++avoidCheckNumber;

			// まだ掘る？：王手を解除できた

			pair<int, vector<Move>> returnResult;


			// すでに底に達している → 受けきった、ということ
			if ( level <= 0 ) {

				//++_allReadState;

				// 一つでも受けきる手順があったならば、詰みなしということ
				result.first  = 0;
				result.second = route;
				result.second.push_back(move);

				//delete nextState;
				//MemoryPool_ForTreeSearch_Free ( nextState );
				MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );

				return;

			// まだ底に達していない → 続ける
			} else {

				route.push_back( move );
				DigInto_Mate( memoryPool, returnResult, nextState, level - 1, fromThisTurnsViewPoint, route, bestReturnResult.first, windowLowerLimit);
				route.pop_back();
				
				// 更新判定
				//if ( returnResult.first < bestReturnResult.first || i == 0) {
				if ( returnResult.first < bestReturnResult.first || !isUpdate ) {

					bestReturnResult = returnResult;

					isUpdate = true;
				}
			}

			//delete nextState;
			//MemoryPool_ForTreeSearch_Free ( nextState );
			MemoryPool_ForTreeSearch_Free ( memoryPool, nextState );

			// 枝狩り？
			if ( bestReturnResult.first < windowLowerLimit ) {
				break;
			}

		} // 可能な指し手でループ

		// 一度も王手を解除できなかった＝詰み成立
		if ( !isAvoidCheck  ) {
			//++_allReadState;
			
			result.first = 9999 - currentState->_totalMoveCount;
			result.second = route;

		// 一度でも王手を解除した
		} else {
			result = bestReturnResult;
		}

		if (result.first == 9000) {
			printf( " result.first == %d : %s, %s level == %d\n ", result.first, (isUpdate) ? "更新された" : "更新されなかった", (isAvoidCheck) ? "王手を解除した":"王手を解除できなかった", level );
		}

		//printf("王手を解除できる指し手の数：%d\n", avoidCheckNumber);
	}

	return;
}


//==========================================================
// 詰み判定：見直し編：使用者用のインターフェイス関数
//==========================================================
bool AI::GetMateProcedure_Revise( Move &mateMove, int &mateDepth, InnerExpression *currentState, int level) {

	bool isMate = false;

	MemoryPool memoryPool;
	MemoryPool_ForTreeSearch_Setup (memoryPool, 1000);

	mateDepth = level;

	isMate = DigInto_Mate_Revise ( memoryPool, mateMove, mateDepth, currentState, level );

	MemoryPool_ForTreeSearch_End (memoryPool);

	// 詰みが見つかった深さの意味を、指定された深度を元に、整える
	mateDepth = level - mateDepth;

	return isMate;
}

//==========================================================
// 詰み判定：詰みがなければ 戻り値の Moveにはパスが入こととする。
//==========================================================
bool AI::DigInto_Mate_Revise( MemoryPool &memoryPool, Move &mateMove, int &mateDepth, InnerExpression *currentState, int level) {

	string indent = "";
	for (int i = 0; i < 8 - level; ++i) {
		indent += " ";
	}

	///// 次の可能な手を取得：攻め手なので、カード取得は不要
	vector<Move> allMoveList;
	currentState->GetLegalNextMoveList(allMoveList, MOVE::ACTION::CHECKABLE);

	///// 一手で勝てるか？（必ず攻め手は一回指す）
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		// 移動手じゃなくなったら、一手では勝てないと判断
		if ( allMoveList[i].action != MOVE::ACTION::MOVE_PIECE ) {
			break;
		}

		if ( IsWinMove( allMoveList[i]) ) {
			return true;
		}
	}

	///// 必ず攻め手は一回指して、深度チェック。
	/*
		勝つ手がなかったから、ここに来た。
		深度が尽きた＝受け手から見ると、受け切った。
	*/
	if (level <= 0) {
		return false;
	}

	// 次の手リスト
	vector<Move> nextAllMoveList;

	///// 手を進める
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		//printf("%s 攻め手[%d] %s\n",indent.c_str(), i, ShowMoveToConsole( allMoveList[i], true ).c_str());

		// 手を進める
		currentState->GoNext( allMoveList[i] );

		// 王手判定 ＝ もしこのまま二手指しできるとして、こちらに上がる手があるか？
		nextAllMoveList.clear();
		currentState->_currentTurn = SwitchTurn(currentState->_currentTurn);
		currentState->GetLegalWinMoveList( nextAllMoveList );
		currentState->_currentTurn = SwitchTurn(currentState->_currentTurn);

		bool isCheck = (nextAllMoveList.size() > 0);

		// 王手じゃないなら、流す
		if ( !isCheck ) { 
			currentState->GoBack();
			continue; 
		}

		// 受け側の手番。可能な手を求める：「王手を受ける手」なので、カード設置は不要。
		nextAllMoveList.clear();
		currentState->GetLegalNextMoveList( nextAllMoveList, MOVE::ACTION::GUARDABLE );

		// 勝つ手＝攻め手が負けになる手 があるか？
		bool isDefenderWin = false;
		for (unsigned int n = 0; n < nextAllMoveList.size(); ++n) {

			if ( nextAllMoveList[n].action != MOVE::ACTION::MOVE_PIECE) {
				break;
			}
			if ( IsWinMove( nextAllMoveList[n]) ) {
				isDefenderWin = true;
				break;
			}
		}
		// 攻め手さん、別の手を指してください。
		if (isDefenderWin) {
			currentState->GoBack();
			continue;
		}


		/*
			ここに来たということは、
			攻め手の自殺手ではなかったようだ。
		*/

		////// 受け手の指し手を進める
		bool isAllLose = true;
		for (unsigned int n = 0; n < nextAllMoveList.size(); ++n) {

			//printf(" %s 受け手[%d]\n",indent.c_str(), n);

			// 局面をコピーして、
			//InnerExpression *nextState;
			//MemoryPool_ForTreeSearch_Allocate ( memoryPool, nextState );
			//currentState->GetCopy( nextState );
			InnerExpression *nextState = currentState->GetCopy();

			// 手を進める
			nextState->GoNext( nextAllMoveList[n] );
			//result.push_back( nextAllMoveList[n] );

			// 再帰：ここで何をしても負けになるなら、直前の攻め手の手が詰みになっていたということ。
			bool isLose = DigInto_Mate_Revise( memoryPool, mateMove, mateDepth, nextState, level - 2);

			// コピーした局面を解放
			//MemoryPool_ForTreeSearch_Free( memoryPool, nextState );
			delete nextState;

			if (isLose) {

				// 処理なし

			} else {

				isAllLose = false;

				// 受ける手があったということだから、このときの攻め手は流せばいいはず。
				break;
			}
		}

		// 何をしても受け手の負け？
		if (isAllLose) {
			
			// この時点で初めて、結果に攻め手を付け足す：階層を引き返す過程で上書きされていく。最終的に「最初の一手」になる。
			mateMove = allMoveList[i];

			if (mateDepth > level) {
				mateDepth = level;
			}
			return true;

		// 受け手に受ける手があった
		} else {

			// 受ける手があったということだから、今回の攻め手は流す。
			currentState->GoBack();
			continue;
		}

	} // for 攻め手の合法手ループ


	// ここまで来た＝false どの攻め手でも詰められなかった
	return false;
}


