//=========================================================
//
// AI_実装ファイル_詰み探索関連（スレッド）
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"

#include <chrono>
#include <thread>

//// WIN API
#include <time.h>
#include <windows.h>
#include <process.h>


using namespace AI_Modules;


//=========================================================
// 詰み手順を教えて：スレッドで分散する版
//=========================================================
void AI::GetMateProcedure_Thread( vector<Move> &result, InnerExpression *currentState, int level, bool isCheckAlreadyMate ) {

	///// 王手になる手しか取得しない。
	/*
		一手で勝てる局面の場合、結果は保証しない。
	*/
	vector<Move> allMoveList;
	//currentState->GetLegalNextMoveList( allMoveList );
	InnerExpression::GetCheckMoveList( currentState, allMoveList );

	return GetMateProcedure_Thread( result, allMoveList, currentState, level, isCheckAlreadyMate );

}
///// 指定局面からの初手リストを引数でもらう版
void AI::GetMateProcedure_Thread( vector<Move> &result, vector<Move> &allMoveList, InnerExpression *currentState, int level, bool isCheckAlreadyMate ) {

	/*
		初手を外からもらうようになっているので、
		デフォルト＝王手のみだけれども、
		そうじゃない指し手候補リストが来るかもしれない。
		なので、この階層では、王手チェックをする。
		デフォルト＝王手のみ、でも、手数が少ない、というメリットは残るはず。
	*/

	//vector<Move> result;

	//printf("◆ 開始 -----------------------------\n");
	//long t =  timeGetTime();
	//auto t = std::chrono::system_clock::now();

	///// メモリプールを用意
	/*
		ここで用意するのは最上位の分だけ。
		ツリー用のは使えない。
	*/
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc ( allMoveList.size() ); // この一階層で可能な手数がこれを超えないように。いや、超えてもいいんですけど。

	_allocator_DigInto_Mate_Thread_Arg = new Allocator<DigInto_Mate_Thread_Arg>();
	_allocator_DigInto_Mate_Thread_Arg->Alloc(allMoveList.size()); // これは、この一階層で王手になる手数、つまりスレッドの本数分必要。

	///// スレッドに号令を出す変数を準備。
	_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure   = false;

	///// スレッドハンドルリストをクリア：既存のスレッドが終了しているかをチェックするべきかも。
	//_threadHandleList.clear();
	_numberOfThread = 0;
	_DigInto_Mate_Thread_Result.clear();
	
	//vector<DigInto_Mate_Thread_Arg*> threadArgList; // あとで終了を判定するために確保しておく
	


	///// クリティカルセクション、使いますよね？
	//InitializeCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );
	// コンストラクタ・デストラクタで行うこととした。

	///// 一手目の選択肢リストを取得
	/*
		GoToEnd から呼ぶ場合を想定して、このリストを引数で受け取る版も用意する、
		あるいは、そういうインターフェイスにしてしまう。
	*/
	//vector<Move> allMoveList;
	//currentState->GetLegalNextMoveList( allMoveList );


	vector<Move> moveList;
	moveList.reserve(1300);

	///// 可能な指し手でループ
	for (unsigned int i = 0; i < allMoveList.size(); ++i) {
		
		// 終了フラグが立っている：どこかのスレッドで詰みが見つかったということ。
		if (_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure) {

			//printf(" 詰みがみつかったの？ じゃあ、撤収しますね〜\n ");
			// これが、いわば、枝狩り In a mannaer of speaking, this is so called a pruning.
			break;
		}

		// 指した結果
		Move move = allMoveList[i];

		// すでに勝ってるかどうかを見る：第一階層目でだけ、チェックすればいい。
		if (isCheckAlreadyMate) {

			if ( IsWinMove( allMoveList[i]) ) {
			
				_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure = true;
				_DigInto_Mate_Thread_Result.clear();
				_DigInto_Mate_Thread_Result.push_back( allMoveList[i] );

				// 終わっていいですよね？
				break;
			}
		}
		// アルゴリズムの加減で、このチェックをここでしなくて済むなら、その方がいい。
		// 詰み手順探索処理は、あくまでも詰みになるまでを取得、と割り切る。
		
		//InnerExpression *nextState = currentState->GetNext( move );
		InnerExpression *nextState = currentState->GetNext( move, allocator );
		
		// 王手になるか？
		bool isCheck = false;
		{
			// 相手に上がる手があるなら、ダメ
			moveList.clear();
			nextState->GetLegalWinMoveList( moveList );
			
			// 相手に上がる手がないなら、
			if ( moveList.size() == 0 ) {

				// もしこのまま二度指しできるとして、勝つ手があるか？
				nextState->_currentTurn = SwitchTurn(nextState->_currentTurn);

				moveList.clear();
				nextState->GetLegalWinMoveList( moveList );
				isCheck = (moveList.size() > 0);

				nextState->_currentTurn = SwitchTurn(currentState->_currentTurn);
			}
		}

		// 王手になるなら採用。スレッドを立てる。
		if ( isCheck ) {
			
			//DigInto_Mate_Thread_Arg *arg = new DigInto_Mate_Thread_Arg();
			DigInto_Mate_Thread_Arg *arg = _allocator_DigInto_Mate_Thread_Arg->GenerateObj();

			arg->ai = this;
			arg->innerExpression = nextState;
			arg->previousMove    = move;
			arg->level = level - 1; // 一段降りる。
			
			/*
			EnterCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );
				threadArgList.push_back (arg);
			LeaveCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );
			*/
			
			
#if defined(_WIN32)
			// WIN API 版
			//_threadHandleList.push_back(
			//);
			HANDLE hThread = (HANDLE)_beginthreadex ( NULL, 0, AI::DigInto_Mate_Thread, arg, 0, NULL );
			if ( hThread ) {
				
				// 立ち上げに成功したら、スレッドカウントを増やす
				EnterCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );
				{
					++_numberOfThread;
				}
				LeaveCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );
			} else {
				printf ( " AI::DigInto_Mate_Thread launch failed somehow.\n " );
			}
			
#elif defined(__linux__)
			
			// C++11 版
			std::thread t1 ( AI::DigInto_Mate_Thread, arg );
			t1.detach();
			
			// C++11版では、スレッド立ち上げに成功したかどうかの判定はどうするんでしたっけ？？？ しなくていいの？ え？ まじで？
			pthread_mutex_lock( &_criticalSection_For_DigInto_Mate_Thread_Result );
			{
				++_numberOfThread;
			}
			pthread_mutex_unlock( &_criticalSection_For_DigInto_Mate_Thread_Result );

			
			
#endif
		}

	} // 可能な指し手でループ


	// 全スレッドが終了するまで、待機してないと、
	// 今のだと、スレッドを立てるだけ立てて、素通りしている。
	
	//printf ( "_numberOfThread : %d \n", _numberOfThread );
	
	clock_t t = clock();
	int emergencyEscapeCount = 3;
	do 
	{	
		// 多少、何らかの処理がある方が、空ループよりもスレッド終了の待機時間が短くなる模様。
		// マシンによって違うんじゃねーの？ そのときの気分とか天気とか。
		
		clock_t lapse = clock() - t;
		
		if ( lapse >= 3000) {
			--emergencyEscapeCount;
			t = clock();
			
			// このメッセージが出るということは処理が止まっている可能性大。
			printf("Wait until all threads end 全スレッド終了まで待機中（残りスレッド数 [%d]）\n", _numberOfThread);
			
			// 何らかの理由でスレッドか止まっている、もしくはカウント数がズレている可能性があるため、強制ループ解除
			if ( emergencyEscapeCount <= 0 ) {
				
				_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure = true;
				
				printf ( "forced break from the thread wait loop as it gets stucked somehow. Look forward what is gonna happen afterwards!\n " );
				break;
			}
		}
		
		std::this_thread::sleep_for(std::chrono::milliseconds(1));
		
	} while ( _numberOfThread > 0 );
	
	//clock_t t = clock();
	/*
	do 
	{
		//clock_t lapse = clock() - t;
		
		//if ( lapse % 100 == 0 ) {
			
			int liveThreadNum = 0;
			for ( unsigned int i = 0; i < threadArgList.size(); ++i ) {
				
				if ( !threadArgList[i]->isEnd ) {
					++liveThreadNum;
					break;
				}
			}
			if ( liveThreadNum == 0 ) {
				break;
			}
			if ( lapse >= 10000) {
				
				t = clock();
				
				// このメッセージが出るということは処理が止まっている可能性大。
				printf("Wait until all threads end 全スレッド終了まで待機中（残りスレッド数 [%d / %d]）……\n", liveThreadNum, threadArgList.size());
			}
			//std::this_thread::sleep_for(std::chrono::milliseconds(1));
			
			
			if ( liveThreadNum == 0 ) {
				break;
			}
		//}
		
	} while (true);
	*/
	
	/*
	int waitCount = 0;
	while (_numberOfThread > 0) {
		++waitCount;
		if (waitCount >= 3000) {
			waitCount = 0;
			
			// このメッセージが出るということは処理が止まっている可能性大。
			printf("全スレッド終了まで待機中（残りスレッド数 [%d]）……\n", _numberOfThread);
		}
		std::this_thread::sleep_for(std::chrono::milliseconds(1));
	}
	*/

	// ここで、最上位階層用のアロケータを始末
	delete allocator;
	delete _allocator_DigInto_Mate_Thread_Arg;

	// 結局、詰みは見つかった？ In the end, did you find any mate procedures?
	if (_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure) {
		//printf("詰みあり\n");
	} else {
		//printf("詰みなし\n");
	}
	result = _DigInto_Mate_Thread_Result;

	//long passTime = timeGetTime() - t;
	//auto passTime = std::chrono::system_clock::now() - t;
	//printf( "%d ms \n", passTime );


	/*
		このへんに、全スレッドをキチンと終わらせる処理が必要なのでは？
		いや、他殺はしてないから、そういうのは、いらないはず。理屈の上では。
	*/

	// クリティカルセクション、後始末
	//DeleteCriticalSection( &_criticalSection_For_DigInto_Mate_Thread_Result );


	//return result;
}

//=========================================================
// 詰み手順掘り進みスレッド
//=========================================================
unsigned int AI::DigInto_Mate_Thread(void *obj) {

	DigInto_Mate_Thread_Arg *arg = (DigInto_Mate_Thread_Arg*)(obj);
	AI              *ai              = arg->ai;
	InnerExpression *innerExpression = arg->innerExpression;
	Move             previousMove    = arg->previousMove;    // この指し手で、ここに来た。
	int              level           = arg->level;

	// 引数用に確保したメモリは、ここで解放してしまってよい。
	//delete arg;
	// アロケータを使っているなら、個別には解放しない。


	///// 一階層進んできたわけですから、route にはその分、ゲタを履かせておきますよ。
	vector<Move> route;
	//route.reserve(20); // これをするとなぜか遅くなる。実際には件数を入れないことの方が多いのに確保しているから割が悪い？
	route.push_back( previousMove );

	///// 掘り進め処理開始
	pair<int, vector<Move>> returnResult;

	// newを使っているから、別インスタンス。
	// メモリプールを引数で渡して初期化するようにしてあるので、別インスタンスにしなくてもいい。
	// DigInto_Mateの中にも、スレッドで問題になるような箇所はない。
	// ここでAIインスタンスを別個に生成しない方が、処理速度が大幅に早い。
	//AI *ownAI = new AI();
	AI *ownAI = ai;  
	MemoryPool memoryPool;
	ownAI->MemoryPool_ForTreeSearch_Setup( memoryPool, level + 1 );

	// 掘る
	//returnResult = ownAI->DigInto_Mate( innerExpression, level, SwitchTurn(innerExpression->_currentTurn), route, 9000, -9000 );
	ownAI->DigInto_Mate( memoryPool, returnResult, innerExpression, level, SwitchTurn(innerExpression->_currentTurn), route, 9000, -9000 );

	// 詰みが見つかった？：強制終了でここに来たばあい、9000を超えてることはないですよね？
	if ( returnResult.first > 9000 ) {

		//printf("pthread_mutex_lock\n");

#if defined(_WIN32)
		
		// WinAPI
		EnterCriticalSection( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );

#elif defined(__linux__)

		// Linux
		pthread_mutex_lock( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );

#endif
		// 排他処理
		{
			if ( !AI::_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure ) {
			
				//printf("in mutex\n");
				// その戻り値を結果に入れる。
				ai->_DigInto_Mate_Thread_Result = returnResult.second;

				// 終了（他のスレッドにも終わってもらう）
				AI::_DigInto_Mate_Thread_Signal_isDiscoverMateProcedure = true;
			}
		}

#if defined(_WIN32)

		// WinAPI
		LeaveCriticalSection( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );

#elif defined(__linux__)

		// Linux
		pthread_mutex_unlock( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );

#endif
	}
	//printf("mutex unlocked.\n");
	
	//printf("delete memory pool\n");
	ownAI->MemoryPool_ForTreeSearch_End(memoryPool);
	
	// 別個に生成しないので、delete もしない。したらダメですよ。
	//delete ownAI;
	
	
	//printf("memory pool deleted\n");
	
	// スレッド数カウントを減らす
	//ai->_numberOfThread -= 1;
#if defined(_WIN32)

	// クリティカルセクション WinAPI
	EnterCriticalSection( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );
	{
		ai->_numberOfThread -= 1;
	}
	LeaveCriticalSection( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );

#elif defined(__linux__)

	// クリティカルセクション Linux
	pthread_mutex_lock( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );
	{
		ai->_numberOfThread -= 1;
	}
	pthread_mutex_unlock( &(ai->_criticalSection_For_DigInto_Mate_Thread_Result) );

#endif
	
	//arg->isEnd = true;
	
	//printf( "スレッド数  : %d\n", ai->_numberOfThread );
	//printf("is going to return\n");
	
	return 0;
}


