//=========================================================
//
// AI
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include <fstream>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"
#include "../Utils/Utils_Convert.h"
#include "../Utils/Utils_Others.h"


using namespace AI_Modules;


//=========================================================
// 木探索用メモリプールを操作する関数群：プールを作る
//=========================================================
void AI::MemoryPool_ForTreeSearch_Setup    ( MemoryPool &pool, int numberOfInnerExpressions ) {
	
	pool.memoryPool = new char[ numberOfInnerExpressions * sizeof(InnerExpression) ];
	pool.currentIt  = pool.memoryPool;
	
	//[[deprecated("怪しい処理：これで本当にメモリを初期化できているの？")]]
	
	//ZeroMemory( pool.memoryPool, numberOfInnerExpressions * sizeof(InnerExpression));
	// というか、むしろループ版の方が処理が速い気がする。
	// 初期化自体不要なような気もしなくもない。処理速度に影響はなさそうなので、やるだけはやっておく。
	for (int i = 0; i < numberOfInnerExpressions * sizeof(InnerExpression); ++ i) {
        pool.memoryPool[i] = NULL;
	}
	
}
//=========================================================
// 木探索用メモリプールを操作する関数群：プールから割り当てる
//=========================================================
void AI::MemoryPool_ForTreeSearch_Allocate ( MemoryPool &pool, InnerExpression* &memory ) {

	// ここを使ってね
	memory = (InnerExpression*)(pool.currentIt);

	// インデックス？ を進める
	pool.currentIt += sizeof( InnerExpression );
}

//=========================================================
// 木探索用メモリプールを操作する関数群：プールへ返す
//=========================================================
void AI::MemoryPool_ForTreeSearch_Free ( MemoryPool &pool, InnerExpression* &memory ) {
	
	/*
		これは、あくまでも、木探索用のメモリプール、というわけで、
		生成と解放のタイミングは常にそろっている前提。 memoryの位置を探る必要はない。
		スタック的な感じ。
		本当に大丈夫なんですか？ の、はず。くわばらくわばら。
	*/
	
	memory->~InnerExpression();
	InnerExpression::operator delete( memory, pool.currentIt);

	// インデックス？ を一つ戻す
	pool.currentIt -= sizeof( InnerExpression );
}

//=========================================================
// 木探索用メモリプールを操作する関数群：プールを片付ける
//=========================================================
void AI::MemoryPool_ForTreeSearch_End (MemoryPool &pool) {

	delete[] pool.memoryPool;

	pool.memoryPool = NULL;
	pool.currentIt  = NULL;
}
/*
//=========================================================
// 木探索用メモリプールを操作する関数群：プールを作る
//=========================================================
void AI::MemoryPool_ForTreeSearch_Setup    ( int numberOfInnerExpressions ) {
	
	_memoryPool_ForTreeSearch = new char[ numberOfInnerExpressions * sizeof(InnerExpression) ];
	_memoryPool_ForTreeSearch_currentIt = _memoryPool_ForTreeSearch;
	ZeroMemory(_memoryPool_ForTreeSearch, numberOfInnerExpressions * sizeof(InnerExpression));

}
//=========================================================
// 木探索用メモリプールを操作する関数群：プールから割り当てる
//=========================================================
void AI::MemoryPool_ForTreeSearch_Allocate ( InnerExpression* &memory ) {

	// ここを使ってね
	memory = (InnerExpression*)_memoryPool_ForTreeSearch_currentIt;

	// インデックス？ を進める
	_memoryPool_ForTreeSearch_currentIt += sizeof( InnerExpression );
}

//=========================================================
// 木探索用メモリプールを操作する関数群：プールへ返す
//=========================================================
void AI::MemoryPool_ForTreeSearch_Free ( InnerExpression* &memory ) {

	memory->~InnerExpression();
	InnerExpression::operator delete( memory, _memoryPool_ForTreeSearch_currentIt);

	// インデックス？ を一つ戻す
	_memoryPool_ForTreeSearch_currentIt -= sizeof( InnerExpression );
}

//=========================================================
// 木探索用メモリプールを操作する関数群：プールを片付ける
//=========================================================
void AI::MemoryPool_ForTreeSearch_End () {

	delete[] _memoryPool_ForTreeSearch;

	_memoryPool_ForTreeSearch = NULL;
	_memoryPool_ForTreeSearch_currentIt = NULL;
}
*/
