//=========================================================
//
// AI_実装ファイルその２
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"


using namespace AI_Modules;


//=========================================================
// 次の一手を教えて：子処理：モンテカルロ
//=========================================================
Move AI::GetNextMove_Composite( InnerExpression *innerExpression ) {

	Move result;

	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList( allMoveList );

	int tryCount = 10;
	if      ( allMoveList.size() > 1200 ) { tryCount = 1;}
	else if ( allMoveList.size() > 1000 ) { tryCount = 5;}
	else if ( allMoveList.size() > 500 ) { tryCount = 10;}
	else if ( allMoveList.size() > 300 ) { tryCount = 20;}
	else if ( allMoveList.size() > 150 ) { tryCount = 30;}
	else if ( allMoveList.size() > 100 ) { tryCount = 40;}
	else if ( allMoveList.size() > 50 ) { tryCount = 100;}
	else if ( allMoveList.size() > 0 ) { tryCount  = 200;}

	int maxWinCountRecord = -1;
	float minTurnCountAvarage = 1000000.0f;

	///// 一手で勝つのがあればそれを選ぶ
	for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

		if ( IsWinMove( allMoveList[i]) ) {
			result = allMoveList[i];
			printf("[%d / %d手目] 勝つ手を選択\n", i, allMoveList.size());
			break;
		}
		// 先頭に「移動」がきてるはずなので、「移動」じゃなくなったら、見るのをやめる。
		if ( allMoveList[i].action != MOVE::ACTION::MOVE_PIECE ) {
			break;
		}
	}
	// この段階で PASS じゃない ＝ 勝つ手があったということ
	if ( result.action != MOVE::ACTION::PASS ) {
		return result;
	}

	///// 詰む手があればそれを選ぶ：一手勝ちと兼用にできますよね、多分。
	vector<Move> mateProcedure;
	GetMateProcedure_Thread( mateProcedure, innerExpression, 7);
	if ( mateProcedure.size() > 0 ) {

		printf("詰む手があった\n");
		return *(mateProcedure.begin());
	}

	/*
		相手に詰まされる手を避ける
	*/
	_allocator = new Allocator<InnerExpression>();
	_allocator->Alloc(10000);

	vector<Move> moveCandidateList;
	for ( unsigned int i = 0; i < allMoveList.size(); ++i) {

		InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], _allocator );

		mateProcedure.clear();
		GetMateProcedure_Thread( mateProcedure, nextState, 7, true);

		if (mateProcedure.size() <= 0) {
			moveCandidateList.push_back( allMoveList[i] );
		}
	}
	// 救済措置
	if (moveCandidateList.size() == 0) {
		printf("何をしても詰まされる場合の救済措置\n");
		moveCandidateList = allMoveList;
	}
	delete _allocator;


	_allocator = new Allocator<InnerExpression>();
	_allocator->Alloc(10000);

	///// 探索
	for ( unsigned int i = 0; i < moveCandidateList.size(); ++i ) {

		InnerExpression *nextState = innerExpression->GetNext( moveCandidateList[i], _allocator );

		int winCount  = 0;
		int drawCount = 0;

		// 試行
		float turnCountAvarage = 0.0f;
		for (int n = 0; n < tryCount; ++n) {

			// 残り全部勝利したとしても記録を更新できない見込みなら、打ち切る。
			if (maxWinCountRecord >= winCount + tryCount - n) {
				break;
			}

			vector<int> tryResult = GoToEnd(nextState);
			int winner    = tryResult[0];
			int turnCount = tryResult[1];
			turnCountAvarage += (float)turnCount;

			if ( winner == innerExpression->_currentTurn) {

				++winCount;
			} else if ( winner == TURN::NONE) {
				++drawCount;
			}

		}
		turnCountAvarage /= (float)tryCount;

		// 勝率記録更新？
		bool isUpdate = false;
		if (winCount > maxWinCountRecord) {

			isUpdate = true;
			
		} else if (winCount == maxWinCountRecord) {

			if (minTurnCountAvarage > turnCountAvarage) {
				isUpdate = true;
			}
			
		}

		if ( isUpdate ) {

			maxWinCountRecord = winCount;
			result = moveCandidateList[i];

			if (minTurnCountAvarage > turnCountAvarage) {
				minTurnCountAvarage = turnCountAvarage;
			}
		}

		//delete nextState;
		if ( i % 10 == 0 ) {
			printf("[%d / %d手目] 勝率%d , 引き分け：%d\n", i, moveCandidateList.size(), winCount, drawCount);
		}
	}

	delete _allocator;
	_allocator = NULL;

	return result;
}
