//=========================================================
//
// AI_実装ファイルその３
//
//=========================================================
#include "pch.h"

#include <stdio.h>
#include "AI.h"
#include "InnerExpression.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Allocator.h"
#include "../Utils/Utils_Others.h"


using namespace AI_Modules;


//=========================================================
// 次の一手を教えて：木探索版変形
//=========================================================
Move AI::GetNextMove_Composite2( InnerExpression *innerExpression ) {
	_count_Evaluate_Piece = 0;
	Move result;

	///// 現局面で可能な手を全て取得
	vector<Move> allMoveList;
	innerExpression->GetLegalNextMoveList( allMoveList );

	// 絞り込む前に、件数を確保しておく。
	int allMoveListSize = allMoveList.size();

	///// 探索前の絞込み
	GetNextMove_SubProcedure_Pre( allMoveList, innerExpression );

	///// 読むまでもない？
	if ( allMoveList.size() == 1 ) {

		result = allMoveList[0];

	///// 探索しましょう。
	} else {

		///// シャッフルしておく：事前絞込み内で「勝つ手があればそれを選択」しているため、それ以前にシャッフルされていてはいけない。
		UTILS_OTHERS::RandomShuffle<Move>( allMoveList );

		MemoryPool memoryPool;
		MemoryPool_ForTreeSearch_Setup( memoryPool, 100 );
		//_allReadState = 0;

		DigInto_Window_OptionArgs optionArgs;

		int depth = 5;
		if      ( allMoveListSize > 400 ) { depth = 0; }
		else if ( allMoveListSize > 300 ) { depth = 0; }
		else if ( allMoveListSize > 200 ) { depth = 0; }
		else if ( allMoveListSize > 100 ) { depth = 0; }
		else if ( allMoveListSize > 0   ) { depth = 1; }

		// 中盤用評価関数を使います。
		/*
			allMoveListSize で、読み深度とともに設定を分けたり、
			絞込みの結果、詰みを避けたような場合には「終盤」と判定するとか。
		*/
		optionArgs.evaluateFunctionFlgs = EVALUATION::RUTIN_SET::FOR_MIDDLE_STAGE;

		pair<Move, int> searchResult = DigInto_Window ( memoryPool, allMoveList, innerExpression, depth, innerExpression->_currentTurn, 99999, -99999, optionArgs);
		//printf("評価値：%d（深度：%d） 全 %d 手\n", searchResult.second, depth, optionArgs.allReadState);

		result = searchResult.first;

		MemoryPool_ForTreeSearch_End ( memoryPool );
	}

	//printf("駒の価値評価を実行した回数：%d\n", _count_Evaluate_Piece);
	return result;
}


//=========================================================
// 次の一手を教えて：基本子処理：事前下見
//=========================================================
void AI::GetNextMove_SubProcedure_Pre( vector<Move> &allMoveList, InnerExpression *innerExpression ) {

	///// 学習中は詰め探索をしない。
	bool isLearning = false;

	//printf("本探索前の絞込み開始  絞込み前の手数 %d\n", allMoveList.size());
	//long t =  timeGetTime();

	/*
		探索前の下準備。
		一手で勝つ手、詰む手、つまされる手を避ける


		vector<Move> &allMoveList を絞り込む。
	*/

	///// 可能な手のもともとの数を確保しておく。絞り込む前に。
	int allMoveListSize = allMoveList.size();
	
	if ( !isLearning ) {

		///// 一手で勝つのがあればそれを選ぶ
		for ( unsigned int i = 0; i < allMoveList.size(); ++i ) {

			if ( IsWinMove( allMoveList[i]) ) {
			
				Move winMove = allMoveList[i];
				allMoveList.clear();
				allMoveList.push_back( winMove );

				//printf("勝つ手を選択 %d ms\n", timeGetTime() - t );
				return;
			}
			// 先頭に「移動」がきてるはずなので、「移動」じゃなくなったら、見るのをやめる。
			if ( allMoveList[i].action != MOVE::ACTION::MOVE_PIECE ) {
				break;
			}
		}
		//printf("一手で勝つ手はナシ %d ms\n", timeGetTime() - t );

		///// 詰み探索は、何手読みますか？
		int mateSearchLevel = 7;
		{
			if      (allMoveList.size() > 600) { mateSearchLevel = 1; }
			else if (allMoveList.size() > 500) { mateSearchLevel = 3; }
			else if (allMoveList.size() > 400) { mateSearchLevel = 5; }
			else if (allMoveList.size() > 300) { mateSearchLevel = 7; }
			else if (allMoveList.size() > 200) { mateSearchLevel = 9; }
			else if (allMoveList.size() > 100) { mateSearchLevel = 11; }
			else if (allMoveList.size() >   0) { mateSearchLevel = 13; }
		}
		///// 詰む手があればそれを選ぶ：一手勝ちと兼用にできますよね、多分。

		vector<Move> mateProcedure;
		GetMateProcedure_Thread( mateProcedure, innerExpression, mateSearchLevel);
		if ( mateProcedure.size() > 0 ) {

			//printf("詰む手があった %d ms\n", timeGetTime() - t );

			Move mateMove = *(mateProcedure.begin());
			allMoveList.clear();
			allMoveList.push_back( mateMove );

			return;
		}
		//printf("詰む手ナシ %d ms\n", timeGetTime() - t );
	}

	///// 浅く読んで、見込みのある手を残し、見込みのない手を省く
	GetNextMove_SubProcedure_Pre_shallowSearch ( allMoveList, innerExpression );

	if ( !isLearning ) {
		///// 相手に詰まされる手を避ける
		GetNextMove_SubProcedure_Pre_avoid_Mated( allMoveList, innerExpression, allMoveListSize);
	}

	//printf( "本探索前の絞込み終了 %d ms  絞込み後の手数 %d\n", timeGetTime() - t, allMoveList.size() );


}

//=========================================================
// 本探索前の絞込み＿子処理：浅く読んで絞り込む
//=========================================================
void AI::GetNextMove_SubProcedure_Pre_shallowSearch( vector<Move> &allMoveList, InnerExpression *innerExpression ) {

	///// 浅く読んで、見込みのある手を残し、見込みのない手を省く
	//printf("浅く読んで、見込みのある手を残し、見込みのない手を省く\n");
	//printf("探索開始\n");

	///// ここでシャッフルしてもいいはず。
	// 上位を残す処理のとき、同率順位で並び順が固定しないように。
	UTILS_OTHERS::RandomShuffle<Move>( allMoveList );

	///// 全ての手は評価しない。現状よりいい手がいくつ見つかったら満足する？
	int howManyBetterMoveDoIWant = allMoveList.size();

	///// 満足基準
	int satisfyBase = Evaluate( innerExpression, innerExpression->_currentTurn );
	int satisfyBaseOriginally = satisfyBase;

	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(1500);

	// 上位いくつを残す？
	/*
		ここで絞り込むのって、結局、詰みを避ける処理の負担を減らすのが主効果かも。
		その意味では、
		パッと見、詰みがなさそう＝王手がない、ならば、広めに手を残して、
		そうでないなら、狭めに、手を絞り込んでおく、
		というのが、よさそうな戦略？
	*/
	int topN = 20;

	// 可能な手・評価値(浅)
	vector< pair<Move, int> > point_CandidateMove;

	// 読み深度
	int depth = 1;
	if      ( allMoveList.size() > 250 ) { depth = -1; topN = 50; howManyBetterMoveDoIWant = 100;}
	else if ( allMoveList.size() >  50 ) { depth = 0; topN = 25; howManyBetterMoveDoIWant = 100;}
	else if ( allMoveList.size() >   0 ) { depth = 0; topN = allMoveList.size(); howManyBetterMoveDoIWant = 50;} // 悪い手があった場合に省くだけ。
	
	MemoryPool memoryPool;
	MemoryPool_ForTreeSearch_Setup(  memoryPool, 100 );

	vector<Move> nextMoveList;
	DigInto_Window_OptionArgs digInto_Window_OptionArgs;
	digInto_Window_OptionArgs.isDoShallowSearchPrune = false;

	for (unsigned int i = 0; i < allMoveList.size(); ++i) 
	{

		InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], allocator );

		pair<Move, int> searchResult;

		// 探索しない場合：depth 0 のときは、depth 0 で探索する
		//if ( depth < 0 ) {

			//int eval = Evaluate ( nextState, innerExpression->_currentTurn);
			int eval = Evaluate_Terrain ( nextState, innerExpression->_currentTurn, 1);

			searchResult.first  = allMoveList[i];
			searchResult.second = eval;
		/*
		// 探索する場合
		} else {

			digInto_Window_OptionArgs.allReadState = 0;

			nextMoveList.clear();
			nextState->GetLegalNextMoveList(nextMoveList);

			// 次の番の人が、その人の立場で、全力で指す、とする。
			searchResult = DigInto_Window ( memoryPool, nextMoveList, nextState, depth, nextState->_currentTurn, 99999, -99999, digInto_Window_OptionArgs);

			// この手を指した場合、ということでしたよね、もともと。
			searchResult.first = allMoveList[i];

			// 符号を入れ替える：「次の番の人が」その人の立場で、全力で指す、と、したから。
			searchResult.second *= -1;
		}
		*/

		//printf("%d / %d  評価値 : %d(深度:%d)\n", i, allMoveList.size(), searchResult.second, depth);

		// 一時リストに入れる：ソートは後回しで。
		point_CandidateMove.push_back( searchResult );

		
		// 現状よりいい手の数
		if (  searchResult.second >= satisfyBase ) {
			--howManyBetterMoveDoIWant;
		}

		// 満足基準が浮き沈みする
		satisfyBase -= (satisfyBase  - searchResult.second) / 10;

		// 満足したら出る
		if ( howManyBetterMoveDoIWant <= 0 ) {
			//printf("現状（%d）よりいい手がたくさん見つかって満足したので探索終了\n", satisfyBase);
			break;
		}

	}
	//printf("満足基準、元々の値：%d\n", satisfyBaseOriginally);
	//printf("満足基準、最終値  ：%d\n", satisfyBase);
	//printf("howManyBetterMoveDoIWant 残り値：%d\n", howManyBetterMoveDoIWant);
	
	// ソート
	//printf("ソート\n");
	GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate(point_CandidateMove);
	//printf("ソート終了\n");

	MemoryPool_ForTreeSearch_End ( memoryPool );
	delete allocator;

	//printf("探索終了\n");
	//printf("上位を残す\n");

	///// 上位を残す。
	vector< pair<Move, int> > point_CandidateMove_topN;
	for (unsigned int i = 0; i < point_CandidateMove.size() && i < topN; ++i) {

		point_CandidateMove_topN.push_back ( point_CandidateMove[i] );
	}
	//printf("残り %d 件\n", point_CandidateMove_topN.size());


	///// あまりにも悪い手は省く（たとえ上位に残っていたとしても）
	//printf("極端に悪い手を省く\n");
	int uncertificatablePoint = -5000;


	vector< pair<Move, int> > point_CandidateMove_topN_noBad;

	for (unsigned int i = 0; i < point_CandidateMove_topN.size(); ++i) {

		// 耐える点数を下回ったら、ループ終了
		if ( point_CandidateMove_topN[i].second < uncertificatablePoint ) {
			break;
		}

		// 候補に残す
		point_CandidateMove_topN_noBad.push_back ( point_CandidateMove_topN[i] );
	}
		
	// 候補リストを置き換える：悪い手しか残ってない場合は、置き換えない。
	if ( point_CandidateMove_topN_noBad.size() > 0 ) {
		point_CandidateMove_topN = point_CandidateMove_topN_noBad;
	} else {
		//printf("悪い手しか残ってなかったので置き換えない\n");
	}

	//printf("残り %d 件\n", point_CandidateMove_topN.size());
	

	///// 絞り込み結果を反映させる
	allMoveList.clear();
	for ( unsigned int i = 0; i < point_CandidateMove_topN.size(); ++i ) {

		allMoveList.push_back( point_CandidateMove_topN[i].first );
	}

	//printf("Shallow探索終了\n");
}
///// ソート
void AI::GetNextMove_SubProcedure_Pre_shallowSearch_SortCandidate( vector< pair<Move, int> > &point_CandidateMove) {

	bool isSwap = true;
	while (isSwap) { 

		isSwap = false;

		for ( unsigned int i = 0; i < point_CandidateMove.size() - 1; ++i ) {

			// 入れ替える？
			if ( point_CandidateMove[i].second < point_CandidateMove[i + 1].second ) {

				pair<Move, int> temp = point_CandidateMove[i];
				point_CandidateMove[i    ] = point_CandidateMove[i + 1];
				point_CandidateMove[i + 1] = temp;

				isSwap = true;
			}
		}
	}
}


//=========================================================
// 本探索前の絞込み＿子処理：詰まされる手を避ける
//=========================================================
void AI::GetNextMove_SubProcedure_Pre_avoid_Mated( vector<Move> &allMoveList, InnerExpression *innerExpression, int allMoveListSize ) {

	/*
		相手に詰まされる手を避ける
	*/
	//printf("相手に詰まされる手を避ける ： 現時点 %d 件\n", allMoveList.size());
	vector<Move> mateProcedure;
	Allocator<InnerExpression> *allocator = new Allocator<InnerExpression>();
	allocator->Alloc(10000);

	///// 詰みを避ける手をいくつ求める？
	int howManysaftyMovesDoIWant = allMoveList.size();

	///// 相手に詰みが生じる手の数
	int matedMoveLength = 0;

	int depth = 1;
	if      (allMoveListSize > 600) { depth = 3; howManysaftyMovesDoIWant = 10;}
	else if (allMoveListSize > 500) { depth = 3; howManysaftyMovesDoIWant = 10;}
	else if (allMoveListSize > 400) { depth = 3; howManysaftyMovesDoIWant = 10;}
	else if (allMoveListSize > 300) { depth = 5; howManysaftyMovesDoIWant = 10;}
	else if (allMoveListSize > 200) { depth = 5; howManysaftyMovesDoIWant = 10;}
	else if (allMoveListSize > 100) { depth = 5; howManysaftyMovesDoIWant = 10;}
	else if (allMoveListSize >   0) { depth = 7; howManysaftyMovesDoIWant = 10;}

	vector<Move> moveCandidateList;      // 詰まない手を入れる
	vector<Move> moveCandidateList_Mate; // 詰む手を入れる
	int mateProcedureLength = 0;         // 詰む場合の、手数最大値（できるだけ長い手数で逃げるため）
	for ( unsigned int i = 0; i < allMoveList.size(); ++i) {

		InnerExpression *nextState = innerExpression->GetNext( allMoveList[i], allocator );

		mateProcedure.clear();
		GetMateProcedure_Thread( mateProcedure, nextState, depth, true);

		// 詰み手順が存在しない場合のみ、選択肢に入れる。
		if (mateProcedure.size() <= 0) {
			moveCandidateList.push_back( allMoveList[i] );

			// 満足したら出る
			if ( matedMoveLength > 0 &&  // そもそも詰まされる心配がないなら、そうそう満足しない。
				moveCandidateList.size() >= howManysaftyMovesDoIWant ) {
				break;
			}

		// 詰む手も一応とっておく（何をしても詰まされる場合に、マシな手を使う）
		} else {

			++matedMoveLength;

			if ( mateProcedure.size() > mateProcedureLength ) {
				
				moveCandidateList_Mate.clear();
				mateProcedureLength = mateProcedure.size();
			}

			if ( mateProcedure.size() == mateProcedureLength ) {
				
				moveCandidateList_Mate.push_back( allMoveList[i] );

			}
		}
	}
	delete allocator;

	// 救済措置
	if (moveCandidateList.size() == 0) {
		//printf("何をしても詰まされる場合の救済措置\n");
		//moveCandidateList = allMoveList;
		moveCandidateList = moveCandidateList_Mate;
	}

	// 絞り込んだ結果に置き換える
	allMoveList = moveCandidateList;

	//printf("相手に詰まされる手を避ける＿終了：残り %d 件\n", allMoveList.size());
}
