//=========================================================
//
// 局面、内部表現
//
//=========================================================

#ifndef _INNER_EXPRESSION_
#define _INNER_EXPRESSION_

#include <stdlib.h>
#include <vector>
//#include <set>
#include <new>

using namespace std;

#include "Move.h"
#include "Position.h"
#include "../Utils/Allocator.h"

namespace AI_Modules {

	class InnerExpression {
	
	public:
		
		InnerExpression();
		~InnerExpression();

		//static char *_memoryPool; 
		//static void Initialize();
		//static InnerExpression* GenerateInstance();
		//static void Release();

		// placement newとplacement deleteを定義
		
		static void* operator new(size_t size, void* buf){ return buf; }
		static void* operator new(size_t size){ 
			return malloc(size); 
		}
		static void operator delete(void* p, void* buf){}
		static void operator delete(void* p){ free(p); }  // VC6.0では必要
		
		//=========================================================
		// static
		//=========================================================
		//static int _numberOfInstances;
		//static set<InnerExpression*> _instanceList;
		//static void LeakRelease();


		//=========================================================
		// 盤上のカード
		//=========================================================
		struct CardOnBoard {

			int cardId;       // どのカードが
			int direction;    // どの向きか

			CardOnBoard() {
				cardId    = -1; // マイナスで、「なし」ということにする。
				direction = -1;
			}

			bool isEmpty() {
				return (cardId < 0);
			}
		};


		//=========================================================
		// 初期設定
		//=========================================================
		void Init(); // 盤のサイズをそろえて、後は真っ白にする。
		void Init_SetCurrentTurn( int turn );
		void Init_PutCard( int x, int y, int cardId, int direction );
		void Init_PutPiece( int x, int y, int piece );
		void Init_SetThrouableInfo( int x, int y, int throuableInfo );
		void Init_SetHandPiece_BLACK( int num );
		void Init_SetHandPiece_WHITE( int num );
		void Init_AddHandCard_BLACK( int cardId );
		void Init_AddHandCard_WHITE( int cardId );
		void Init_ReflectCardAffection();

		//=========================================================
		// 次の局面を取得
		//【引数】こう指すと
		//【戻値】こうなる
		//=========================================================
		InnerExpression* GetNext( Move move );
		InnerExpression* GetNext( Move move, Allocator<InnerExpression> *allocator );

		//=========================================================
		// 次の局面になっておくれ
		//【引数】こう指してね
		//=========================================================
		void GoNext( const Move &move );

		//=========================================================
		// 手を一つ巻き戻す
		//【戻り値】Undoできたら true 
		//=========================================================
		bool GoBack();

		//=========================================================
		// 可能な次の手のうち、勝つ手を取得
		//（現局面に王手が生じているかどうかを判定する主旨）
		//=========================================================
		void GetLegalWinMoveList(vector<Move> &result);
		bool IsCheck();


		//=========================================================
		// 可能な次の手を列挙
		//=========================================================
		//vector<Move> GetLegalNextMoveList();
		void GetLegalNextMoveList(vector<Move> &result);
		void GetLegalNextMoveList(vector<Move> &result, int moveChoice);

		//==========================================================
		// 王手になる手だけを取得（現状、王手になってない前提）
		// 手番は、現在の手番を使う。
		//==========================================================
		static void GetCheckMoveList ( InnerExpression *currentState, vector<Move> &result );
		static void GetCheckMoveList_PutCard           ( InnerExpression *currentState, vector<Move> &result, vector < pair<int, int> > &piecePositionListMy, vector < pair<int, int> > &piecePositionListOpponent);
		static void GetCheckMoveList_PutPiece          ( InnerExpression *currentState, vector<Move> &result);
		static void GetCheckMoveList_MovePiece         ( InnerExpression *currentState, vector<Move> &result, vector < pair<int, int> > &piecePositionListMy, vector < pair<int, int> > &piecePositionListOpponent);
		static void GetCheckMoveList_MovePiece_Collide ( InnerExpression *currentState, vector<Move> &result, vector < pair<int, int> > &piecePositionListMy, vector < pair<int, int> > &piecePositionListOpponent);

		static void GetPiecePositionList ( InnerExpression *currentState, vector<pair<int, int>> &resultBlack, vector<pair<int, int>> &resultWhite );
		static void GetToGoalCourseList ( InnerExpression *currentState, vector<vector<pair<int, int>>> &result);
		static bool IsMovable ( InnerExpression *currentState, int pieceX, int pieceY, int toX, int toY);
		static bool IsMovable ( InnerExpression *currentState, int pieceX, int pieceY, int toX, int toY, pair<int, int> &ignoredPiecePosition );
		static bool IsCollidable ( InnerExpression *currentState, int pieceX, int pieceY, int toX, int toY, int &moveToCollideX, int &moveToCollideY );
		static bool ToGoalThroughThere ( InnerExpression *currentState, int pieceX, int pieceY, int checkPointX, int checkPointY, pair<int, int> &ignoredPiecePosition);
		static void GetGoalCourseListWithPanel( InnerExpression *currentState, vector< pair<vector< pair<int, int>>, int> > &result, vector < pair<int, int> > &piecePositionListMy, vector < pair<int, int> > &piecePositionListOpponent );
		static void GetCardIdList( InnerExpression *currentState, int turn, set<int> &result );
		static void GetCheckMoveList_PutCard_ToThisCourse ( InnerExpression *currentState, pair<vector<pair<int, int>>, int> &course, int cardId, vector<Move> &result );

		//=========================================================
		// コピーを取得
		//=========================================================
		InnerExpression* GetCopy();
		InnerExpression* GetCopy(Allocator<InnerExpression> *allocator);
		void GetCopy( InnerExpression *useThisMemory );

		//=========================================================
		// 局面が同一かどうか
		//【戻り値】 true で「同一」
		//=========================================================
		bool IsEquals( InnerExpression *anotherOne );
		static int compare_int( const void *a, const void *b );

		//=========================================================
		// コンソールに様子を表示：デバッグ用
		//=========================================================
		void ShowToConsole();

		///// 手番情報
		int _currentTurn; // 現在どっちの手番か？ TURN
		
		///// 盤面情報
		int         _gridList_ThrouableInfo[15][15]; // 通行情報レイヤー：通行情報　 THROUABLE_INFO
		int         _gridList_Piece[15][15];         // 駒レイヤー      ：黒/白/なし GRID
		CardOnBoard _gridList_Card[15][15];          // カードレイヤー  ：カードと向き
		static int           _gridList_Init[15][15];     // 初期化用
		static CardOnBoard   _gridList_Init_Card[15][15]; // 初期化用
		int _boardSizeX;
		int _boardSizeY;

		///// 駒台情報：何個あるか、だけ。
		int _handPiece_BLACK;
		int _handPiece_WHITE;

		///// 手札情報：カード番号のリスト
		int _handCards_BLACK[100];
		int _handCards_WHITE[100];
		static int _handCards_Init[100];
		int _handCards_BLACK_Index;
		int _handCards_WHITE_Index;

		///// 今、何手目？
		int _totalMoveCount;
		void AddTotalMoveCount()       { ++_totalMoveCount; }
		void DecrementTotalMoveCount() { --_totalMoveCount; }
		int GetTotalMoveCount()        { return _totalMoveCount;}

		///// パス情報
		bool _isPassPreviousMove;
		bool _isPassPass;


		///// 盤上のカードの状態に対して通行可能情報が最新か？
		bool _isThrouableInfoLatest;

	private:


		///// Undo用
		MoveForBack _moveForBack;


		//=========================================================
		// 可能な次の手を列挙：子処理群
		//=========================================================
		void GetLegalNextMoveList_PutPiece  (vector<Move> &result );
		void GetLegalNextMoveList_MovePiece (vector<Move> &result );
		void GetLegalNextMoveList_TakeCard  (vector<Move> &result );
		void GetLegalNextMoveList_PutCard   (vector<Move> &result );

		//void GetLegalNextMoveList_PutCard_TurnDetermined (vector<Move> &result, vector<int> &handCards );
		void GetLegalNextMoveList_PutCard_TurnDetermined (vector<Move> &result, int handCards[], int length );

		void GetLegalNextMoveList_MovePiece_toUP    (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toDOWN  (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toLEFT  (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toRIGHT (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toRIGHT_DOWN (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toRIGHT_UP   (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toLEFT_UP    (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toLEFT_DOWN  (vector<Move> &result, int x, int y );

		void GetLegalNextMoveList_MovePiece_toUP_toWin    (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toDOWN_toWin  (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toRIGHT_DOWN_toWin (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toRIGHT_UP_toWin   (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toLEFT_UP_toWin    (vector<Move> &result, int x, int y );
		void GetLegalNextMoveList_MovePiece_toLEFT_DOWN_toWin  (vector<Move> &result, int x, int y );


		//=========================================================
		// 各方向への移動可？
		//=========================================================
		bool IsThrouableTo( int throuableInfo, int direction );

		//=========================================================
		// 次の局面になっておくれ：子処理群
		//=========================================================
		void GoNext_PutPiece  ( const Move &move );
		void GoNext_MovePiece ( const Move &move );
		void GoNext_PutCard   ( const Move &move );
		void GoNext_TakeCard  ( const Move &move );
		void GoNext_Pass      ( const Move &move );

		//=========================================================
		// 手を巻き戻して：子処理群
		//=========================================================
		bool GoBack_PutPiece  ();
		bool GoBack_MovePiece ();
		bool GoBack_PutCard   ();
		bool GoBack_TakeCard  ();
		bool GoBack_Pass      ();

		//=========================================================
		// 盤の範囲内？
		//=========================================================
		bool IsWithinBoardLimit( int x, int y );

		//=========================================================
		// このカードがありますか？
		//=========================================================
		int GetHandCardIndexOf_Black( int cardId);
		int GetHandCardIndexOf_White( int cardId);

		//=========================================================
		// n番目のカードを手放す
		//=========================================================
		void ReleaseCard_ByIndex_Black( int index);
		void ReleaseCard_ByIndex_White( int index);
	};
};
#endif
