//=========================================================
//
// 簡易XMLパーサークラス（宣言）
//
//=========================================================

#ifndef _NOTXML_
#define _NOTXML_

#include <vector>
#include <map>
#include <string>

using namespace std;


class NotXML {

public:

	//=========================================================
	// タグ構造体
	//=========================================================
	struct Tag {

		string name;                   // タグ名
		string value;                  // 値
		map<string, string> attribute; // 属性
		vector<Tag> childList;         // 子要素

		Tag() {
			name  = "";
			value = "";
		}
	};
	
	//=========================================================
	// 読み込む（ファイル名を指定）
	//=========================================================
	vector<Tag> Read     (string fileName);
	Tag         ReadRoot (string fileName);

	//=========================================================
	// 読み込む（タグが書かれているであろう文字列をタグ化する）
	//=========================================================
	vector<Tag> Read_fromString     (string str);
	Tag         ReadRoot_fromString (string str);


	//=========================================================
	// Utilities
	//=========================================================
	static Tag GetTagFromName( const Tag& parent, string name ) {

		Tag result;

		for (unsigned int i = 0 ; i < parent.childList.size(); ++i) {

			if ( parent.childList[i].name == name) {
				result = parent.childList[i];
				break;
			}
		}

		return result;
	}
	static vector<Tag> GetTagListFromName( const Tag& parent, string name ) {

		vector<Tag> result;

		for (unsigned int i = 0 ; i < parent.childList.size(); ++i) {

			if ( parent.childList[i].name == name) {
				result.push_back( parent.childList[i] );
			}
		}

		return result;
	}
	static string GetAttribute( const Tag& tag, string name, string defaultValue = "" ) {

		string result = defaultValue;

		if (tag.attribute.find(name) != tag.attribute.end()) {
			result = tag.attribute.find(name)->second;
		}

		return result;
	}
	static bool IsExist ( const Tag& tag, string name) {
		bool result = false;
		
		for (unsigned int i = 0 ; i < tag.childList.size(); ++i) {

			if ( tag.childList[i].name == name) {
				result = true;
				break;
			}
		}
		return result;
	}

private:

	///// タグの種類（読み取り中専用）
	enum PURE_TAG_TYPE {

		NOT_PURE_TAG_TYPE,     // 無効値
		OPEN,    // 開始タグ    <***>
		CLOSE,   // 閉じタグ    </***>
		EMPTY,   // 空要素タグ  <***/>
		COMMENT, // コメント    <!-- ***  -->

		PROCESS, // 処理命令    <?***?> Processing Instruction
		DOCTYPE, // DOCTYPE宣言 <!DOCTYPE ****/>

		VALUE    // 値          上記以外（タグではない）
	};


	//=========================================================
	// ファイルを読み込んで文字列にする
	//=========================================================
	string ReadFile_toString(string fileName);

	//=========================================================
	// 文字列からタグっぽい箇所を抽出
	//=========================================================
	vector<string> Parse_PureTagList(string str);

	//=========================================================
	// タグの種類を判別する
	//=========================================================
	PURE_TAG_TYPE JadgePureTag( string pureTag );

	//=========================================================
	// 属性を解釈
	//=========================================================
	map<string, string> GetAttribute(string str);

	//=========================================================
	// タグとしての情報を取得する
	//=========================================================
	Tag GetTagInfo(string str, PURE_TAG_TYPE type);

	//=========================================================
	// タグ間の構造を解釈
	//=========================================================
	vector<Tag> Parse(vector<string> pureTagList);

	//=========================================================
	// 前後Trim
	//【引数】
	// string
	//【戻り値】
	// 前後の半角スペース、\n、\r、\t を取り除いた文字列
	//=========================================================
	string Trim(const string &str);
};
#endif




