//=========================================================
//
// アニメーション
//
//=========================================================

#ifndef _ANIMATION_
#define _ANIMATION_

#include <string>
#include <vector>
#include <map>
#include <SFML/Graphics.hpp>
#include <SFML/Audio.hpp>

using namespace std;


namespace GUI_SFML {

	class Animation {

	public:

		//=========================================================
		// アニメーションの１コマ
		//=========================================================
		struct AnimationFrame {

			int           frameCount;  // 表示時間
			string        imageName;   // 画像名
			sf::Rect<int> rect;        // 表示部分
			int           offsetX;     // 画像表示オフセットX
			int           offsetY;     // 画像表示オフセットY
			string        soundFile;   // 効果音ファイル名
			int           addRotate;   // 回転角度追加
			string        setOriginX;  
			string        setOriginY;
			bool          isReverseX;
			bool          isReverseY;
			string        scaleX;
			string        scaleY;
			string        color;
			string label;     // このコマのラベル
			string nextLabel; // 次のラベル（空文字で、リストの次。デフォルト）

			AnimationFrame () {
				frameCount = 1;
				imageName  = "";
				label      = "";
				nextLabel  = "";
				offsetX    = 0;
				offsetY    = 0;
		        soundFile  = "";
				addRotate  = 0;
				setOriginX = "";
				setOriginY = "";
				isReverseX  = false;
				isReverseY  = false;
				color       = "";
				scaleX      = "";
				scaleY      = "";

			}
		};

		Animation();
		~Animation();

		void Init ( string fileName );

		void Update ( sf::RenderWindow &window );
		void Draw   ( sf::RenderWindow &window );

		void         SetPosition ( sf::Vector2f pos ) { _position = pos; }
		sf::Vector2f GetPosition () { return _position; }

		bool IsActive () { return _isActive; }

	private:

		///// 稼動中？
		bool _isActive;

		///// 表示位置
		sf::Vector2f _position;
		sf::Vector2f _offset;

		///// 元となる画像
		map < string, sf::Image>   _imageList;
		map < string, sf::Texture> _textureList;

		///// コマリスト
		vector < AnimationFrame > _frameList;

		///// ループフラグ
		bool _isLoop;

		///// フレームカウント：今のコマになってから何フレーム経過したか？
		int _frameCount;

		///// 今表示中のコマ番号
		int _currentFrameIndex;

		///// 表示するもの
		sf::Sprite _sprite;

		//=========================================================
		// 新たなフレームを開始する
		//=========================================================
		void StartNewFrame();

	};


};
#endif
