//=========================================================
//
// 読み返しログビューワー
//
//=========================================================

#include "ScenarioLogViewer.h"
#include "ScrollBar.h"

#include <fstream>
#include <iostream>

#include "SFML_Utils.h"
#include "TextImageFactory.h"

#include "../ScriptEngine/Application.h"
#include "../ScriptEngine/interfaces/IScenarioLogger.h"

#include "../Utils/Utils_String.h"


using namespace GUI_SFML;

//=========================================================
// コンストラクタ
//=========================================================
ScenarioLogViewer::ScenarioLogViewer () {
	
	_scenarioLogger = NULL;
	
	_position = sf::Vector2f ( 50.0f, 50.0f );
	
	_base = sf::RectangleShape( sf::Vector2f(710.0f, 500.0f) );
	_base.setOutlineThickness(1.0f);
	_base.setOutlineColor ( sf::Color ( 0,0,0,255 ) );
	//_base.setFillColor    ( sf::Color ( 120,120,120, 200 ) );
	_base.setFillColor    ( sf::Color ( 50,50,50, 200 ) );
	_base.setPosition( _position - sf::Vector2f(5.0f, 0.0f) );
	
	//_font.loadFromFile ( "/home/kifc000/Documents/フォント/umefont_530/ume-tgc4.ttf" );
	_font.loadFromFile ( TextImageFactory::_fontFileName );
	
	_displayRowNum = 16;
	
	_scrollBar = new ScrollBar();
	_scrollBar->SetActualLength(0);
	_scrollBar->SetDisplayLength(_displayRowNum);
	_scrollBar->SetFrameSize(sf::Vector2f(20.0f, 500.0f));
	_scrollBar->SetPosition( _position + sf::Vector2f(710.0f, 0.0f)  );
	_scrollBar->IsUpDownButton(true);
	
	_isGoOverTheBottom = false;
}

//=========================================================
// デストラクタ
//=========================================================
ScenarioLogViewer::~ScenarioLogViewer() {
	
	delete _scrollBar;
}

//=========================================================
// ファイルから読み込み
//=========================================================
//void ScenarioLogViewer::ReadFile ( string fileName ) {
void ScenarioLogViewer::ReadLog () {
	
	_isGoOverTheBottom = false;
	_rowList.clear();
	
	//int lineCount = GetLineCount ( fileName );
	int lineCount = 0;
	//int maxLineCount = 500; // 表示する最大行数
	
	
	//std::ifstream ifs;
	//ifs.open ( fileName.c_str(), std::ios::in );
	string buffer = "";
	
	float width = 700.0f;
	
	string allLog = "";
	if ( _scenarioLogger != NULL ) {
		allLog = _scenarioLogger->ReadScenarioLog();
	} else {
		return;
	}
	
	
	//printf ( "%s\n", allLog.c_str() );
	
	vector<string> allRowList = Utils_String::Split ( allLog, "\n" );
	lineCount = allRowList.size();
	
	printf ("Log Line Count : %d\n", lineCount);
	
	//int currentLineCount = 0; // ファイル上の行数、ということで充分。
	//while ( !ifs.eof() ) {
	for ( unsigned int i = 0; i < allRowList.size(); ++i ) {
	
		//getline ( ifs, buffer );
		buffer = allRowList[i];
		
		//++currentLineCount;
		
		// 末尾のn行のみを表示対象とする
		//if (lineCount - currentLineCount > maxLineCount ) {
		//	continue;
		//}
		
		//result += buffer + "\n";
		
		sf::String sfString = SFML_Utils::GetSfString(buffer);
		
		int charNum = sfString.getSize(); // この文字数分登録できたら次の行にいくよ！
		int currentRegisteredCharNum = 0;
		
		///// 改行処理を交えつつ追加していく
		while (true) {
		
			sf::Text text ( sfString, _font, 18);
			
			
			sf::String leftHalf  = "";
			sf::String rightHalf = "";
			
			bool alreadyOver    = false;
			bool alreadyNewLine = false;
			
			for (unsigned int i = 0; i < sfString.getSize(); ++i) 
			{
				float x = text.findCharacterPos(i).x;
				
				
				alreadyOver = ( x + text.getCharacterSize() < width );
				
				if ( alreadyOver ) {
						
					leftHalf += sfString[i];
					
				} else {
					
					///// TODO 禁則処理：足りなければ増やす。余分ならば削る
					if ( !alreadyNewLine ) {
						
						if (
							   sfString[i] == SFML_Utils::GetSfString("。") 
							|| sfString[i] == SFML_Utils::GetSfString("、") 
							|| sfString[i] == SFML_Utils::GetSfString("？") 
							|| sfString[i] == SFML_Utils::GetSfString("！") 
							|| sfString[i] == SFML_Utils::GetSfString("」") 
							|| sfString[i] == SFML_Utils::GetSfString("）") 
						) 
						{
								
							leftHalf += sfString[i];
							
						} else {
							rightHalf += sfString[i];
							alreadyNewLine = true;
						}
						
					} else {
					
						rightHalf += sfString[i];
						alreadyNewLine = true;
					}
				}
			}
			
			sf::Text textLeft ( leftHalf, _font, 18);
			textLeft.setColor ( sf::Color::White );
		
			_rowList.push_back(textLeft);
			
			// ハミ出した分がなくなったら次の行へ
			if (rightHalf == "") {
				break;
			}
			
			sfString = rightHalf;
		}
	}
	
	//ifs.close();

		
	_scrollBar->SetActualLength(_rowList.size());
	
	_scrollBar->MoveToBottom();
	
}


//=========================================================
// 事前に行数を数える目的のもの
//=========================================================
int ScenarioLogViewer::GetLineCount ( string fileName ) {
	
	int result = 0;

	std::ifstream ifs;
	
	ifs.open ( fileName.c_str(), std::ios::in );
	
	string buffer;
	
	
	while ( !ifs.eof() ) {
		
		getline ( ifs, buffer );
		
		++result;
	}
	ifs.close();
	
	return result;
}

//=========================================================
// 更新
//=========================================================
void ScenarioLogViewer::Update(sf::RenderWindow &window) {
	
	
	_scrollBar->Update(window);
	
	int mouseWheel = 0;
	
	if ( Application::GetInstance()->HasFocus() ) {
		
		mouseWheel = Application::GetInstance()->_mouseWheel; // TODO 一応分離させてるのに、こんなことしていいのかな？
//printf ("[ScenarioLogViewer] Application::GetInstance()->_mouseWheel : %d\n", Application::GetInstance()->_mouseWheel);
		
		if ( sf::Keyboard::isKeyPressed ( sf::Keyboard::Up ) ) {
//printf ("[ScenarioLogViewer] sf::Keyboard::Up\n");
			mouseWheel = 1;
		}
		if ( sf::Keyboard::isKeyPressed ( sf::Keyboard::Down ) ) {
//printf ("[ScenarioLogViewer] sf::Keyboard::Down\n");
			mouseWheel = -1;
		}
//printf ("[ScenarioLogViewer] mouseWheel : %d\n", mouseWheel);
		_scrollBar->MoveUp(mouseWheel);
		
	}
	
	
	
	// 最下段でさらに下へ行こうとした？
	_isGoOverTheBottom = (_scrollBar->IsBottom() && mouseWheel < 0);
	
	int currentScrollIndex = _scrollBar->GetCurrentIndex();
	
	sf::Vector2f offset(10.0f, 10.0f);
	int rowHeight = 30;
	
	for (int i = 0; i < _displayRowNum; ++i) {
		
		if ( i + currentScrollIndex >= _rowList.size() ) {
			break;
		}
		
		_rowList[i + currentScrollIndex].setPosition ( offset + _position + sf::Vector2f(0.0f, rowHeight * i )  );
		
	}
	
}

//=========================================================
// 描画
//=========================================================
void ScenarioLogViewer::Draw(sf::RenderWindow &window) {
	
	window.draw(_base);
	
	int currentScrollIndex = _scrollBar->GetCurrentIndex();
	
	for (int i = 0; i < _displayRowNum; ++i) {
		
		if ( i + currentScrollIndex >= _rowList.size() ) {
			break;
		}
		
		sf::Text text = _rowList[i + currentScrollIndex];
		window.draw(text);
	}
	
	
	_scrollBar->Draw(window);
	
}



