//=========================================================
//
// GUI SelectBox(declare)
//
//=========================================================


#ifndef _SELECT_BOX_
#define _SELECT_BOX_

#include <map>
#include <vector>
#include <string>

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>

using namespace std;

namespace GUI_SFML {

	class ISelectBoxItem;
	class ScrollBar;

	class SelectBox {

	public:

		SelectBox();
		~SelectBox();

		void Update(sf::RenderWindow &window);
		void Draw(sf::RenderWindow &window);

		void AddItem( ISelectBoxItem *item );
		void RemoveItem( ISelectBoxItem *item );
		ISelectBoxItem* GetPointingItem();
		ISelectBoxItem* GetClickItem();
		int GetPointingItemIndex() { return _currentPointingItenIndex; }

		void SetPosition( sf::Vector2f pos );
		void SetPosition_Offset( sf::Vector2f pos ) { _position_Offset = pos ;}
		sf::Vector2f GetPosition() { return _position; }

		void SetDisplayCount( int arg);
		void SetFrameSize( sf::Vector2f arg);

		void SetLengthOfOneItem( float arg );
		
		void MoveToBottom(); // 最下段まで移動
		void MoveToTop();    // 最上段まで移動
		
		void Makeup_OutofUpperBounds   (int index); // 指定インデックスの項目が上にハミ出ていたら、スクロールを追いつかせる
		void Makeup_OutofDownnerBounds (int index); // 指定インデックスの項目が下にハミ出ていたら、スクロールを追いつかせる
		
		void SetFrameColor             ( sf::Color color );
		void SetFrameOutlineColor      ( sf::Color color );
		void SetFrameOutlineThickness  ( float value     );
		
		void SetDefaultCursorOn ( bool arg  ) { _isDefaultCorsorOn = arg; }
		
		bool IsMouseOver() { return _isMouseOver; }
		void ScrollUp();
		void ScrollDown();
		
		ScrollBar* GetScrollBar() { return _scrollBar; }
		
		
	private:

		vector<ISelectBoxItem*> _itemList;
	
		sf::Vector2f _frameSize;
		int _displayCount;

		sf::Vector2f _position;
		sf::Vector2f _position_Offset;
		int _currentDisplayPosition;
		
		sf::RectangleShape _frame;
		sf::RectangleShape _cursor;

		float _length_of_oneItem;

		ScrollBar *_scrollBar;
		
		bool _isMouseRelease;
		bool _isMousePress;

		ISelectBoxItem *_currentPointingItem;
		int _currentPointingItenIndex;
		
		bool _isDefaultCorsorOn;
		
		bool _isMouseOver;
	};

};
#endif
