//=========================================================
//
// Slider (implement)
//
//=========================================================

#include "Slider.h"
#include "Button.h"
#include "SFML_Utils.h"

using namespace GUI_SFML;

//=========================================================
// Slider
//=========================================================
Slider::Slider(InitProp initProp) {

	_position = initProp.position;
	_length   = initProp.length;

	_range_min = initProp.range_min;
	_range_max = initProp.range_max;

	_trigerSize = sf::Vector2f( 15.0f, 25.0f);

	_currentTriggerPosition = 0.0f;


	_rail = sf::RectangleShape ( sf::Vector2f(_length, 6.0f) );
	_rail.setOutlineThickness(1.0f);
	_rail.setOutlineColor (sf::Color(0,0,0,0));
	_rail.setFillColor    (initProp.railColor);

	_trigger = sf::RectangleShape ( _trigerSize );
	_trigger.setOutlineThickness(1.0f);
	_trigger.setOutlineColor (sf::Color(0,0,0,255));
	_trigger.setFillColor    (initProp.trigerColor);

	_trigger.setOrigin( _trigerSize / 2.0f );


	_isGrabbing = false;
	_mouseX_previous = 0;
	
	_isMousePressed = false;
	_isMousePushedOnThis = false;

	SetCurrentValue( initProp.value );

	_isButton = initProp.isButton;
	_button_toMin = new GUI_SFML::Button( "<<", sf::Vector2f(20.0f, 25.0f) );
	_button_toMax = new GUI_SFML::Button( ">>", sf::Vector2f(20.0f, 25.0f) );


	SetPosition(initProp.position);

}
//=========================================================
// ~Slider
//=========================================================
Slider::~Slider() {

	delete _button_toMin;
	delete _button_toMax;

}

//=========================================================
// Update
//=========================================================
void Slider::Update(sf::RenderWindow &window) {

	if ( ! window.hasFocus() ) {
		return;
	}
	
	_rail.setPosition( _position );
	_trigger.setPosition( 
		_position + 
		sf::Vector2f(_currentTriggerPosition, 3.0f ) 
	);


	sf::Vector2i mousePos = sf::Mouse::getPosition(window);
	int mouseX = mousePos.x;
	int mouseY = mousePos.y;
	
	SFML_Utils::PosToScreenPos(mouseX, mouseY, window);
	
	bool isMousePush = sf::Mouse::isButtonPressed( sf::Mouse::Left );
	
	// マウスを押した時点で、マウスが乗っていた？
	if 
	(
		!_isMousePressed &&
		isMousePush &&
		(
			isMouseOverTheTriger_LeftPart(mouseX, mouseY)  ||
			isMouseOverTheTriger_RightPart(mouseX, mouseY) ||
			isMouseOverTheTriger(mouseX, mouseY) 
		)
	) 
	{
		_isMousePushedOnThis = true;
	}
	
	if ( !isMousePush ) 
	{
		_isMousePushedOnThis = false;
	}
	
	// ツマミ を つかんでいるかどうか
	if (_isGrabbing) {

		if (!isMousePush) {
			_isGrabbing = false;
		}

	} else {
		
		if (
			_isMousePushedOnThis && // 押した時点でマウスが乗っていて
			isMouseOverTheTriger( mouseX, mouseY ) // マウスが乗っていれば
		) 
		{ 
			// つかんだことにする
			_isGrabbing = true;
		}
	}

	if (_isGrabbing) {

		_currentTriggerPosition += mouseX - _mouseX_previous;

	} else {

		if (isMousePush && _isMousePushedOnThis) {

			if ( isMouseOverTheTriger_LeftPart( mouseX, mouseY )) { 
				_currentTriggerPosition -= GetOnePointLength();
			}

			if ( isMouseOverTheTriger_RightPart(mouseX, mouseY )) { 
				_currentTriggerPosition += GetOnePointLength();
			}
		}
	}

	if (_isButton) {
		_button_toMin->Update(window);
		_button_toMax->Update(window);

		if (_button_toMin->IsRelease()) {
			_currentTriggerPosition -= GetOnePointLength();
		}
		if (_button_toMax->IsRelease()) {
			_currentTriggerPosition += GetOnePointLength();
		}
	}



	if (_currentTriggerPosition < 0.0f) {
		_currentTriggerPosition = 0.0f;
	}
	if (_currentTriggerPosition > _length) {
		_currentTriggerPosition = _length;
	}

	_mouseX_previous = mouseX;
	
	_isMousePressed = isMousePush;

}
///// isMouseOverTheTriger
bool Slider::isMouseOverTheTriger( int mouseX, int mouseY ) {

	bool result = false;

	if
	(
		_trigger.getPosition().x - _trigerSize.x / 2.0f < mouseX && mouseX < _trigger.getPosition().x + _trigerSize.x / 2.0f
		&&
		_trigger.getPosition().y - _trigerSize.y / 2.0f < mouseY && mouseY < _trigger.getPosition().y + _trigerSize.y / 2.0f
	)
	{
		result = true;
	}

	return result;
}
///// isMouseOverTheTriger_LeftPart
bool Slider::isMouseOverTheTriger_LeftPart( int mouseX, int mouseY ) {

	bool result = false;

	if
	(
		_position.x < mouseX && mouseX < _trigger.getPosition().x - _trigerSize.x / 2.0f
		&&
		_trigger.getPosition().y - _trigerSize.y / 2.0f < mouseY && mouseY < _trigger.getPosition().y + _trigerSize.y / 2.0f
	)
	{
		result = true;
	}

	return result;
}
///// isMouseOverTheTriger_RightPart
bool Slider::isMouseOverTheTriger_RightPart( int mouseX, int mouseY ) {

	bool result = false;

	if
	(
		_trigger.getPosition().x + _trigerSize.x / 2.0f < mouseX && mouseX < _position.x + _length
		&&
		_trigger.getPosition().y - _trigerSize.y / 2.0f < mouseY && mouseY < _trigger.getPosition().y + _trigerSize.y / 2.0f
	)
	{
		result = true;
	}

	return result;
}


//=========================================================
// Draw
//=========================================================
void Slider::Draw(sf::RenderWindow &window) {

	window.draw( _rail );
	window.draw( _trigger );

	if (_isButton) {
		_button_toMin->Draw(window);
		_button_toMax->Draw(window);
	}


}
//=========================================================
// SetPosition
//=========================================================
void Slider::SetPosition( sf::Vector2f pos ) {

	_position = pos;
	_button_toMin->SetPosition( sf::Vector2f( _position.x -20.0f - _trigerSize.x/2.0f - 1, _position.y - _trigerSize.y/2.0f + 3.0f ) );
	_button_toMax->SetPosition( sf::Vector2f( _position.x + _length + _trigerSize.x/2.0f + 1, _position.y  - _trigerSize.y/2.0f + 3.0f) );
}

//=========================================================
// GetCurrentValue
//=========================================================
int Slider::GetCurrentValue() {

	int result = 0;

	/*
		--- factors ---

		_range_min;
		_range_max;
		_length;
		_currentTriggerPosition;
	*/

	float range = (float)(_range_max - _range_min);

	result = _range_min + range * (_currentTriggerPosition / _length);

	return result;
}

//=========================================================
// SetCurrentValue
//=========================================================
void Slider::SetCurrentValue( int value ) {

	if ( value < _range_min ) { value = _range_min; }
	if ( value > _range_max ) { value = _range_max; }

	float maxRange   = (float)(_range_max - _range_min);
	float currentPos = (float)(value - _range_min);


	_currentTriggerPosition =  _length * (currentPos / maxRange);

}
//=========================================================
// GetOnePointLength
//=========================================================
float Slider::GetOnePointLength() {

	float result = 0.0f;

	/*
		--- factors ---

		_range_min;
		_range_max;
		_length;
		_currentTriggerPosition;
	*/
	float range = (float)(_range_max - _range_min);

	result = _length / range;

	return result;
}

