//=========================================================
//
// TextImageFactory Class for SFML (implement)
//
//=========================================================

#include "TextImageFactory.h"
#include <stdio.h>
#include "../Utils/Utils_String.h"

sf::Color TextImageFactory::_defaultColor    = sf::Color(255, 255, 255, 255);
int       TextImageFactory::_defaultFontSize = 16;
sf::Font  TextImageFactory::_font = sf::Font();
string    TextImageFactory::_fontFileName = "";


//int TextImageFactory::_width = 0;

TextImageFactory::TextImageFactory() {}
TextImageFactory::~TextImageFactory() {}

//=========================================================
// フォントを登録
//=========================================================
bool TextImageFactory::SetFont ( std::string fontFileName ) {

	_fontFileName = fontFileName;

	return _font.loadFromFile ( fontFileName );
	
}


//=========================================================
// 文字画像を作成
//=========================================================
void TextImageFactory::Create(sf::Image &resultImage, string text, int fontSize, sf::Color color) {

	Create(
		-1,
		0,
		resultImage, text, fontSize, color
	);
}
//=========================================================
// 文字画像を作成・改行指定と行間指定つき
//=========================================================
void TextImageFactory::Create(
	int newLineWidth,
	int spaceBetweenRows,
	sf::Image &resultImage, string text, int fontSize, sf::Color color)
{
	
	//=========================================================
	// 一文字ずつ、フォント情報を取得していく
	//=========================================================
	unsigned int index = 0;
	unsigned char *input = (unsigned char*)text.c_str();
	
	vector<FontDotData> fontDataList;
	
	while ( index < text.size() ) 
	{
		FontDotData fontDotData;
		
		int bytesInSequence = 0;
		fontDotData.codePoint = utf8ToUtf32 (input + index, bytesInSequence);
		index += bytesInSequence;
		
		const sf::Texture& texture = _font.getTexture(fontSize);
		fontDotData.glyph = _font.getGlyph( fontDotData.codePoint, fontSize, false );
		
		//printf ( "%04X size(%d, %d) %d \n", (int)fontDotData.codePoint, fontDotData.glyph.bounds.width, fontDotData.glyph.bounds.height, fontDotData.glyph.advance );
		
		sf::Image image = texture.copyToImage();
		
		for ( int y = 0; y < fontDotData.glyph.textureRect.height; ++y) {
			
			vector<sf::Color> newRow;
			
			for ( int x = 0; x < fontDotData.glyph.textureRect.width; ++x) {
				
				sf::Color pixelColor 
					= image.getPixel(x + fontDotData.glyph.textureRect.left, y + fontDotData.glyph.textureRect.top);
				
				// 透明度以外の色は指定の色に置き換える
				pixelColor.r = color.r;
				pixelColor.g = color.g;
				pixelColor.b = color.b;
				
				newRow.push_back(pixelColor);
			}
			fontDotData.dotTable.push_back( newRow );
		}
		
		fontDataList.push_back(fontDotData);
		
	} // while
	

	//=========================================================
	// 生成する画像のサイズを求める
	//=========================================================
	int imageWidth = 0;
	int imageHeight = 0;
	
	int currentWidth = 0;
	bool isOnceNewLined = false;
	for ( unsigned int i = 0; i < fontDataList.size(); ++i ) {
		
		sf::Glyph glyph = fontDataList[i].glyph;

		if ( imageHeight < glyph.bounds.height ) {
			imageHeight = fontSize;// glyph.bounds.height;
		}
		
		// 改行判定
		if ( newLineWidth != -1 && currentWidth + glyph.bounds.width > newLineWidth ) {
			currentWidth = 0;
			imageHeight +=   fontSize + spaceBetweenRows; // glyph.bounds.height + spaceBetweenRows;
			isOnceNewLined = true;
		}
		
		currentWidth += glyph.advance;

		fontDataList[i].offsetX = currentWidth - glyph.advance;
		//fontDataList[i].offsetY = (imageHeight - glyph.bounds.height) / 2;
		fontDataList[i].offsetY = imageHeight - fontSize +  ( (fontSize - glyph.bounds.height)/2  );
	
	}
	imageWidth = isOnceNewLined ? newLineWidth : currentWidth;
	

	//=========================================================
	// Image を初期化
	//=========================================================
	resultImage.create(imageWidth, imageHeight, sf::Color::Transparent);
	
	//=========================================================
	// 画像にドットを置いていく
	//=========================================================
	for ( unsigned int i = 0; i < fontDataList.size(); ++i ) {
		
		FontDotData fontDotData = fontDataList[i];
		
		for ( unsigned int y = 0; y < fontDotData.dotTable.size(); ++y) {
			for ( unsigned int x = 0; x < fontDotData.dotTable[y].size(); ++x) {
				
				sf::Color color = fontDotData.dotTable[y][x];
				
				resultImage.setPixel ( fontDataList[i].offsetX + x, fontDataList[i].offsetY + y, color);
			}
		}
	}
	
}


//=========================================================
// sf::Text で。
//=========================================================
sf::Text TextImageFactory::Create ( std::string text, int fontSize ) {
	
	sf::Text result;
	
	unsigned int index = 0;
	unsigned char *input = (unsigned char*)text.c_str();
	
	vector<unsigned int> codePointList;
	
	//_width = 0;
	while ( index < text.size() ) 
	{
		int bytesInSequence = 0;
		unsigned int codePoint = utf8ToUtf32 (input + index, bytesInSequence);
		index += bytesInSequence;
		codePointList.push_back(codePoint);
		
		/*
		sf::Glyph glyph = _font.getGlyph(codePoint, fontSize, false);
		
		bool isHalf = Utils_String::isHalfWidth ( text.substr( (index - bytesInSequence), bytesInSequence ) );
		if (isHalf) {
			_width += glyph.advance / 2.0f;
		} else {
			_width += glyph.advance;
		}
		*/
		
	} // while	
	
	//printf("文字数 : %d\n", codePointList.size());
	
	// 怪しいコード。確保する要素数は単に codePointList.size() じゃなくて？ それにすると末尾にゴミがつく。なぜなのか？
	// 文字列として末尾にNULLを追加するべき？
	// 末尾に NULL をつけるとゴミが消える。そのために一つ余分に確保している。これはいいのか？
	unsigned int *codePointPoiner = new unsigned int[ codePointList.size() + 1  ];
	
	for ( unsigned int i = 0; i < codePointList.size(); ++i ) {
		
		codePointPoiner[i] = codePointList[i];
	}
	codePointPoiner[codePointList.size()] = NULL;
	
	result.setFont(_font);
	result.setCharacterSize(fontSize);
	result.setString( codePointPoiner );
	
	delete[] codePointPoiner;
	
	return result;
}


//=========================================================
// 参考：http://www.ryusuke.net/archives/601
//=========================================================
//unsigned long TextImageFactory::utf8ToUtf32(unsigned char *input, int &bytesInSequence)
unsigned int TextImageFactory::utf8ToUtf32(unsigned char *input, int &bytesInSequence)
{

	unsigned char c1, c2, c3, c4, c5, c6;

	bytesInSequence = 1;
	if (!input)
	{
		return 0;
	}

	//0xxxxxxx (ASCII)      7bit
	c1 = input[0];
	if ((c1 & 0x80) == 0x00)
	{
		return c1;
	}

	//10xxxxxx              high-order byte
	if ((c1 & 0xc0) == 0x80)
	{
		return 0;
	}

	//0xFE or 0xFF          BOM (not utf-8)
	if (c1 == 0xfe || c1 == 0xFF )
	{
		return 0;
	}

	//110AAAAA 10BBBBBB     5+6bit=11bit
	c2 = input[1];
	if (((c1 & 0xe0) == 0xc0) &&
		((c2 & 0xc0) == 0x80))
	{
		bytesInSequence = 2;


		return ((c1 & 0x1f) << 6) | (c2 & 0x3f);
	}

	//1110AAAA 10BBBBBB 10CCCCCC        4+6*2bit=16bit
	c3 = input[2];
	if (((c1 & 0xf0) == 0xe0) &&
		((c2 & 0xc0) == 0x80) &&
		((c3 & 0xc0) == 0x80))
	{
		bytesInSequence = 3;
		return ((c1 & 0x0f) << 12) | ((c2 & 0x3f) << 6) | (c3 & 0x3f);
	}

	//1111 0AAA 10BBBBBB 10CCCCCC 10DDDDDD      3+6*3bit=21bit
	c4 = input[3];
	if (((c1 & 0xf8) == 0xf0) &&
		((c2 & 0xc0) == 0x80) &&
		((c3 & 0xc0) == 0x80) &&
		((c4 & 0xc0) == 0x80))
	{
		bytesInSequence = 4;
		return ((c1 & 0x07) << 18) | ((c2 & 0x3f) << 12) | ((c3 & 0x3f) << 6) | (c4 & 0x3f);
	}

	//1111 00AA 10BBBBBB 10CCCCCC 10DDDDDD 10EEEEEE     2+6*4bit=26bit
	c5 = input[4];
	if (((c1 & 0xfc) == 0xf0) &&
		((c2 & 0xc0) == 0x80) &&
		((c3 & 0xc0) == 0x80) &&
		((c4 & 0xc0) == 0x80) &&
		((c5 & 0xc0) == 0x80))
	{
		//bytesInSequence = 4; // 参考元では 4 となっていた。5ではなく？
		bytesInSequence = 5;
		return ((c1 & 0x03) << 24) | ((c2 & 0x3f) << 18) | ((c3 & 0x3f) << 12) | ((c4 & 0x3f) << 6) | (c5 & 0x3f);
	}

	//1111 000A 10BBBBBB 10CCCCCC 10DDDDDD 10EEEEEE 10FFFFFF        1+6*5bit=31bit
	c6 = input[5];
	if (((c1 & 0xfe) == 0xf0) &&
		((c2 & 0xc0) == 0x80) &&
		((c3 & 0xc0) == 0x80) &&
		((c4 & 0xc0) == 0x80) &&
		((c5 & 0xc0) == 0x80) &&
		((c6 & 0xc0) == 0x80))
	{
		//bytesInSequence = 4; // 参考元では 4 となっていた。6ではなく？
		bytesInSequence = 6;
		return ((c1 & 0x01) << 30) | ((c2 & 0x3f) << 24) | ((c3 & 0x3f) << 18) | ((c4 & 0x3f) << 12) | ((c5 & 0x3f) << 6) | (c6 & 0x3f);
	}

	return 0;
}







