//=========================================================
//
// テキストメッセージ欄
//
//=========================================================

#include "../ScriptEngine/Application.h"
#include "TextMessageWindow.h"
#include "TextChar.h"
#include "Animation.h"
#include "SFML_Utils.h"
#include "TextImageFactory.h"
#include "../Utils/Utils_String.h"
#include "../ScriptEngine/interfaces/IScenarioLogger.h"
#include "ScenarioLogViewer.h"
#include "../ScriptEngine/ReadTextManager.h"
#include "../ScriptEngine/effect/Effect_Shake.h"


using namespace GUI_SFML;


//=========================================================
// 内部定数
//=========================================================
const int TextMessageWindow::MOUSE_STATE_NONE    = 0;  // 押していない
const int TextMessageWindow::MOUSE_STATE_PRESS   = 1;  // 押している
const int TextMessageWindow::MOUSE_STATE_RELEASE = 2;  // 放したところ


//=========================================================
// 特殊制御用文字
//=========================================================
string TextMessageWindow::CONTROL_CHAR_NEW_LINE   = ";"; // 改行
string TextMessageWindow::CONTROL_CHAR_NEW_PAGE   = "/"; // 改ページ
string TextMessageWindow::CONTROL_CHAR_WAIT_CLICK = ":"; // クリック待ち


//=========================================================
// 予約文字
//=========================================================
string TextMessageWindow::ALIAS_START = "[--";   // エイリアス開始タグ
string TextMessageWindow::ALIAS_END   = "--]";   // エイリアス終了タグ
		
const string TextMessageWindow::SPEAKER_NAME_DISPLAY_TYPE_OUT = "OUT"; // 外側（名前欄に表示する）
const string TextMessageWindow::SPEAKER_NAME_DISPLAY_TYPE_IN  = "IN";  // 内側（行の冒頭に挿入する）

	
//=========================================================
// static
//=========================================================
std::map<string, string> TextMessageWindow::_textAliasList;
std::vector<string>      TextMessageWindow::_protagonists;
bool                     TextMessageWindow::_isProtagonistShown = false;

//=========================================================
// コンストラクタ
//=========================================================
TextMessageWindow::TextMessageWindow ( IReadManagerReceptionist *readManagerReceptionist ) {

	// 受け取るけど、解放はしません。
	_readManagerReceptionist = readManagerReceptionist;
	if ( _readManagerReceptionist == NULL ) {
		_readManagerReceptionist = this;
	}

	_readFinishedLength = 0;

	_size.x = 600;
	_size.y = 160;
	_rowHeight = 35.0f;
	_fontSize  = 25;
	_fontColor = sf::Color(255,255,255,255);
	_fontColor_reRead = sf::Color(255,255,0,255);

	_offset.x = 40.0f;
	_offset.y = 20.0f;

	_isFrameOn = true;
	_frame = sf::RectangleShape ( _size );
	_frame.setOutlineThickness ( 1.0f );
	_frame.setOutlineColor ( sf::Color(0,0,0,255) );
	_frame.setFillColor    ( sf::Color(0,0,0,50));

	_isBackImageOn = false;
	_backImageFileName = "";


	_cursorPosition = sf::Vector2f( 0.0f, 0.0f );

	_charShowCountMax = 3;
	_charShowCount = _charShowCountMax;

	_isWaitingClick = false;
	_mouseStateL = MOUSE_STATE_NONE;
	_mouseStateR = MOUSE_STATE_NONE;

	_waitingClickCursor = NULL;
	//_waitingClickCursor = new Animation();
	//_waitingClickCursor->Init( "./data/textMessageWindowWaitingClick.txt" );

	_isTemporaryQuick = false;

	///// 話者名
	_speakerName = "";
	_isSpeakerNameBackImageOn = false;
	_speakerOffset_centering  = 0.0f;
	_speakerNameDisplayType   = SPEAKER_NAME_DISPLAY_TYPE_IN;
	_speakFrom                = "「";
	_speakEnd                  = "」";

	///// 読み返しモード関連
	_isReReadMode      = false; // 今、読み返しモード？
	_reReadRangeFrom   = 0;     // 読み返し行を表示する範囲
	_reReadRangeTo     = 0;     // 読み返し行を表示する範囲
	_reReadLimitRowNum = 100;   // 読み返せる行の上限 
	
	_scenarioLogViewer = new ScenarioLogViewer ();
	
	_scenarioLogBuffer = "";
	
	_isSpeakingStarted = false;
	_isSpeakingEndedRightBefore = false;


	_addText_latest = ""; 
	_charCount_sinceLastAddText = 0;
	
	_isShown = true;
	
	_scenarioLogger = NULL;
	
	_currentamePlateNum = -1;
	
	
	
	_notCareOfCameraPos = false;
	//_notCareOfCameraPos = true;
}

void TextMessageWindow::SetScenarioLogger (IScenarioLogger *obj) { 
	_scenarioLogger = obj;
	_scenarioLogViewer->SetScenarioLogger ( obj );
}


//=========================================================
// デザイン情報を取得
//=========================================================
TextMessageWindow::Design TextMessageWindow::GetDesign () {
	
	TextMessageWindow::Design result;
	
	///// 表示 / 非表示
	result.isShown = _isShown;

	result.position          = _position;
	result.size              = _size;           // 横幅も、縦幅も、サイズ？ 行数ではなく？
	result.offset            = _offset;         // 文字の表示を開始する位置のオフセット
	result.fontSize          = _fontSize;              // 文字のサイズ
	result.rowHeight         = _rowHeight;             // 行間（間というか、１行の高さ）
	result.fontColor         = _fontColor;         // 文字色
	result.fontColor_reRead  = _fontColor_reRead;  // 文字色：読み返し行
	result.backImageFileName = _backImageFileName;

	result.speakerOffset           = _speakerOffset;
	result.speakerOffset_centering = _speakerOffset_centering;
	result.speakerNameDisplayType  = _speakerNameDisplayType;
	result.speakFrom               = _speakFrom;  // 話者名挿入時の目印
	result.speakEnd                = _speakEnd;   // セリフ終了の目印

	result.speakerNameBackImage   = _speakerNameBackImageFileName;
	result.speakerBackImageOffset = _speakerBackImageOffset;
	
	for ( unsigned int i = 0; i < _speakerNameBackImageFileName_wide.size(); ++i ) {
		
		result.speakerNameBackImage_wide.push_back (_speakerNameBackImageFileName_wide[i]);
	}
	//result.speakerNameBackImage_wide = _speakerNameBackImageFileName_wide;

			
	// カーソル速度
	result.charShowCountMax = _charShowCountMax;
			
	// 画像を使わずに矩形を背景にする場合の情報
	result.frameFillColor        = _frame.getFillColor();
	result.frameOutlineColor     = _frame.getOutlineColor();
	result.frameOutlineThickness = _frame.getOutlineThickness();
	
	return result;
}


//=========================================================
// デザイン情報をセット
//=========================================================
void TextMessageWindow::SetDesign ( TextMessageWindow::Design design) {

	SetShown            ( design.isShown   );
	SetPosition         ( design.position  );
	SetWidth            ( design.size.x    );
	SetHeight           ( design.size.y    ); 
	SetOffset           ( design.offset    );
	SetFontSize         ( design.fontSize  );
	SetRowHeight        ( design.rowHeight );
	SetFontColor        ( design.fontColor );
	SetFontColor_reRead ( design.fontColor_reRead  );
	SetBackImage        ( design.backImageFileName );
	
	SetSpeakerNameBackImage_wide ( design.speakerNameBackImage_wide);
	
	//_speakerNameBackImageFileName_wide = design.speakerNameBackImage_wide;
	
	SetSpeakerNameOffset           ( design.speakerOffset           );
	SetSpeakerNameOffsetCentering  ( design.speakerOffset_centering );
	SetSpeakerNameDisplayType      ( design.speakerNameDisplayType  );
	SetSpeakFrom                   ( design.speakFrom               );
	SetSpeakEnd                    ( design.speakEnd                );

	SetSpeakerNameBackImage        ( design.speakerNameBackImage     );
	SetSpeakerNameBackImageOffsetX ( design.speakerBackImageOffset.x );
	SetSpeakerNameBackImageOffsetY ( design.speakerBackImageOffset.y );
	
	// カーソル速度
	SetCursorSpeed ( design.charShowCountMax );
	
	// 画像を使わずに矩形を背景にする場合の情報
	SetFrameFillColor        ( design.frameFillColor        );
	SetFrameOutlineColor     ( design.frameOutlineColor     );
	SetFrameOutlineThickness ( design.frameOutlineThickness );

}
		
//=========================================================
// 話者名表示方式をセット
//=========================================================
void TextMessageWindow::SetSpeakerNameDisplayType ( string type ) {
	
	_speakerNameDisplayType = type;
	
	if      ( type == SPEAKER_NAME_DISPLAY_TYPE_IN  ) { _speakerNameDisplayType = type;}
	else if ( type == SPEAKER_NAME_DISPLAY_TYPE_OUT ) { _speakerNameDisplayType = type;}
	else {
			
		printf("Unknown speakerNameDisplayType : %s\n", type.c_str());
		
		_speakerNameDisplayType = SPEAKER_NAME_DISPLAY_TYPE_IN;
	}
	
}

//=========================================================
// 下敷き画像をセット
//=========================================================
void TextMessageWindow::SetBackImage ( string fileName ) {

	_backImageFileName = fileName;

	if ( fileName != "" ) {

		_image.loadFromFile( fileName );
		_texture.loadFromImage ( _image );
		_sprite.setTexture ( _texture);

		// 画像を使うなら、フレームは使わない。
		_isBackImageOn = true;
		_isFrameOn     = false;

	///// ファイル名が空文字＝画像オフ、とみなす。
	} else {
		_isBackImageOn = false;
		_isFrameOn     = true;
	}
}

//=========================================================
// 名札の下敷き画像をセット
//=========================================================
void TextMessageWindow::SetSpeakerNameBackImage( string fileName ) {
	
	_speakerNameBackImageFileName = fileName;
	
	if ( fileName != "" ) {
		
		_isSpeakerNameBackImageOn = true;
		
		_speakerNameBackImage_image.loadFromFile( fileName );
		_speakerNameBackImage_texture.loadFromImage ( _speakerNameBackImage_image );
		_speakerNameBackImage_sprite.setTexture ( _speakerNameBackImage_texture);
		
		sf::IntRect rect;
		rect.left = 0;
		rect.top = 0;
		rect.width  = _speakerNameBackImage_texture.getSize().x;
		rect.height = _speakerNameBackImage_texture.getSize().y;
		
		_speakerNameBackImage_sprite.setTextureRect( rect );
		
		
	} else {
		
		_isSpeakerNameBackImageOn = false;
	}
	
}
///// ワイド版
void TextMessageWindow::SetSpeakerNameBackImage_wide( string fileName ) {
	
	_speakerNameBackImageFileName_wide.push_back ( fileName );
	
	int index = _speakerNameBackImageFileName_wide.size() - 1;
	
	_speakerNameBackImage_wide_texture.push_back ( sf::Texture() );
	_speakerNameBackImage_wide_sprite.push_back ( sf::Sprite() );
	
	_speakerNameBackImage_wide_texture[index].loadFromFile ( _speakerNameBackImageFileName_wide[index] );
	_speakerNameBackImage_wide_sprite[index].setTexture ( _speakerNameBackImage_wide_texture[index]);
		
	sf::IntRect rect;
	rect.left = 0;
	rect.top = 0;
	rect.width  = _speakerNameBackImage_wide_texture[index].getSize().x;
	rect.height = _speakerNameBackImage_wide_texture[index].getSize().y;
	
	_speakerNameBackImage_wide_sprite[index].setTextureRect( rect );

}
///// ワイド版：リストで総入れ替え
void TextMessageWindow::SetSpeakerNameBackImage_wide( vector<string> fileNameList ) {
	
	for ( unsigned int i = 0; i < _speakerNameBackImageFileName_wide.size(); ++i ) {
		
		// TODO 必要なら解法処理を
	}
	
	//printf ("総入れ替え発生\n");
	
	_speakerNameBackImageFileName_wide.clear();
	_speakerNameBackImage_wide_texture.clear();
	_speakerNameBackImage_wide_sprite.clear();
	
	
	for ( unsigned int i = 0; i < fileNameList.size(); ++i ) {
			
		//printf ( "ワイド版名札画像：%s\n", fileNameList[i].c_str() );
		
		SetSpeakerNameBackImage_wide (fileNameList[i]);
		
	}
}
///// クリアしたいときだってあるんですよ。
void TextMessageWindow::ClearSpeakerNameBackImage_wide () {
	
	
	for ( unsigned int i = 0; i < _speakerNameBackImageFileName_wide.size(); ++i ) {
		
		// TODO 必要なら解法処理を
	}
	
	//printf ("総入れ替え発生\n");
	
	_speakerNameBackImageFileName_wide.clear();
	_speakerNameBackImage_wide_texture.clear();
	_speakerNameBackImage_wide_sprite.clear();	
	
}

//=========================================================
// クリック待ちのときのカーソルを設定。アニメーション設定ファイル
//=========================================================
void TextMessageWindow::SetClickWaitCorsor ( string fileName ) {

	// すでに設定されていたら、前のは消しておく。
	if (_waitingClickCursor != NULL) {
		delete _waitingClickCursor;
	}
	if (fileName != "" ) {
		_waitingClickCursor = new Animation();
		_waitingClickCursor->Init( fileName );
	} else {
		_waitingClickCursor = NULL;
	}
}


//=========================================================
// 主人公名を登録
//=========================================================
void TextMessageWindow::RegistProtagonist ( string name ) {
	_protagonists.push_back(name);
}
void TextMessageWindow::ShowProtagonistName () { _isProtagonistShown = true; }
void TextMessageWindow::HideProtagonistName () { _isProtagonistShown = false;}


//=========================================================
// 文字列のエイリアスを登録
//=========================================================
void TextMessageWindow::RegistTextAlias ( string alias, string value ) {
	
	///// 同じ名前のものがあったら警告でも出しておいてあげる
	if ( _textAliasList.find ( alias ) != _textAliasList.end() ) {

		printf ( " 文字列のエイリアス %s(%s) が上書きされて %s(%s) になりました。\n", 
			alias.c_str(), _textAliasList[alias].c_str(),
			alias.c_str(), value.c_str()
		);
	}

	// ラベルと、ファイル名の対応を覚えておく。
	// TODO セーブ/ロード対象、にするべきであるはず。いえ、これも「設定」ですから、セーブしてしまうと、変更が反映されなくなってしまう。
	_textAliasList [ alias ] = value;
	
}

void TextMessageWindow::DeleteTextAlias ( string alias ) {
	
	if ( _textAliasList.find ( alias ) != _textAliasList.end() ) {
		
		_textAliasList.erase ( alias );
	}	
}


//=========================================================
// テキスト内のエイリアスを置換
//=========================================================
void TextMessageWindow::ReplaceWithAlias ( string &text ) {
	
	string aliasStart = ALIAS_START;
	string aliasEnd   = ALIAS_END;
	
	
	for (auto const &entity : _textAliasList ) {
		
		string alias = entity.first;
		string value = entity.second;
		
		string replaceFrom = aliasStart + alias + aliasEnd;
		
		text = Utils_String::Replace ( text, replaceFrom, value );
		
		
	}
	
}
	
	

//=========================================================
// デストラクタ
//=========================================================
TextMessageWindow::~TextMessageWindow() {

	///// 残っている文字は解放、でいいですよね？
	Clear();

	delete _waitingClickCursor;

	if ( _readManagerReceptionist == this ) {
		// あ、別に、何もしなくていいのか。一心同体なのだから。
	}


	Release_ReReadCharList();
}

//=========================================================
// 読み返し行をクリアする
//=========================================================
void TextMessageWindow::Release_ReReadCharList() {

	for ( unsigned int i = 0; i < _reReadRowList.size(); ++i ) {
		for ( unsigned int n = 0; n < _reReadRowList[i].size(); ++n ) {
			delete _reReadRowList[i][n];
		}
	}
	_reReadRowList.clear();
}


//=========================================================
// クリア
//=========================================================
void TextMessageWindow::Clear() {

	// 表示されているものを、という意味。
	for ( unsigned int i = 0; i < _textCharList.size(); ++i ) {

		delete _textCharList[i];
	}
	_textCharList.clear();
	for ( unsigned int i = 0; i < _textCharList_Reserved.size(); ++i ) {

		delete _textCharList_Reserved[i];
	}
	_textCharList_Reserved.clear();
}

//=========================================================
// 更新
//=========================================================
void TextMessageWindow::Update( sf::RenderWindow &window ) {

	///// 非表示モード？
	if ( !_isShown ) {
		_isReReadMode = false; // ついでに読み返しモードも解除。
		return;
	}
	
	///// カメラに合わせて位置を調整する
	sf::Vector2f innerPos;
	if ( !_notCareOfCameraPos ) {
	
		//sf::Vector2f screenPos = Application::GetInstance()->_window.mapPixelToCoords( sf::Vector2i( _position.x, _position.y ) );
		sf::Vector2i screenPos = Application::GetInstance()->_window.mapCoordsToPixel( 
			sf::Vector2f( _position.x, _position.y ),
			Application::GetInstance()->_window.getDefaultView()
		);
		innerPos = _position;
		//_position = screenPos;
		_position.x = screenPos.x;
		_position.y = screenPos.y;
	}
	
	_frame.setPosition  ( _position );
	_sprite.setPosition ( _position );
	
	_speakerNameBackImage_sprite.setPosition ( _position + _speakerBackImageOffset );
	
	// TODO とりあえず、ワイド版は一つのみ
	//_speakerNameBackImage_wide_sprite.setPosition ( _position + _speakerBackImageOffset );
	for ( unsigned int i = 0; i < _speakerNameBackImage_wide_sprite.size(); ++i ) {
		_speakerNameBackImage_wide_sprite[i].setPosition ( _position + _speakerBackImageOffset );
	}


	///// 読み返しモード？
	if ( _isReReadMode ) {

		Update_ReReadMode (window);

	///// 通常モード？
	} else {
		
		ShowCharOneByOne(window);

		///// クリック待ち？
		if ( _isWaitingClick ) {

			// クリック待ちカーソルがあれば、使う
			if ( _waitingClickCursor != NULL ) {
				_waitingClickCursor->SetPosition ( _cursorPosition + _position + _offset );
				_waitingClickCursor->Update ( window );
			}

			if ( _mouseStateL == TextMessageWindow::MOUSE_STATE_RELEASE ) {
				_isWaitingClick = false;
			}
		}

		///// 更新
		for ( unsigned int i = 0; i < _textCharList.size(); ++i ) {

			_textCharList[i]->Update( window );
		}

		///// 読み返しモードに移行する？
		ToReReadModeIfNeccesary( window );
	}


	///// マウスボタン監視
	ObserveMouseState(window);
	
	
	// カメラに合わせて調整した内部座標を元に戻す
	if ( !_notCareOfCameraPos ) {
		_position = innerPos;
	}
}

//=========================================================
// 読み返しモードにする？ ：マウスの操作を監視して、モードを切り替える（読み返しモードじゃないときに呼ぶ）
//=========================================================
void TextMessageWindow::ToReReadModeIfNeccesary ( sf::RenderWindow &window )  {

	if 
	(
		Application::GetInstance()->HasFocus()
		&&
		(
			Application::GetInstance()->_mouseWheel > 0 
			||
			sf::Keyboard::isKeyPressed ( sf::Keyboard::Up )
		)
	) 
	{
		
		//_scenarioLogViewer->ReadFile ( "textLog.txt" );
		_scenarioLogViewer->ReadLog();
		
		_isReReadMode = true;
		
		/*
		///// つまり、読み返すテキストがあるときのみ、読み返しモードにする
		if ( _reReadRowList.size() > 0) {

			_isReReadMode = true;

			///// 表示範囲の初期状態を設定
			_reReadRangeTo   = _reReadRowList.size() - 1;

			int rowLimit  = (_rowHeight > 0) ? _size.y / _rowHeight : 0;

			_reReadRangeFrom = _reReadRangeTo -  rowLimit + 1;

			if ( _reReadRangeFrom < 0 ) {
				_reReadRangeFrom = 0;
			}

			SetUpCurrentReReadCharList();

			// みなさん、位置についてね。
			Update_ReReadCharList ( window );
		}
		*/
	}

}
//=========================================================
// 通常モードにする？：マウスの操作を監視して、モードを切り替える（読み返しモードのときに呼ぶ）
//=========================================================
void TextMessageWindow::ToNomalModeIfNeccesary () {

	/*
		もしかすると、この処理を分ける必要はないのかも？
	*/
}

//=========================================================
// 読み返しモードのときの更新
//=========================================================
void TextMessageWindow::Update_ReReadMode ( sf::RenderWindow &window ) {

	/*
		操作を監視して、
		表示範囲を変更。

		最下段の状態でさらに下段を表示しようとしたならば、読み返しモード解除
		マウス左クリックで、読み返しモード解除

		現在の表示範囲の読み返し文字に Update をかける。
	*/

	///// 操作
	ReReadMode_Operation ();

	///// 操作した後で、今でもまだ、読み返しモード？
	if ( !_isReReadMode ) {
		return;
	}
	
	_scenarioLogViewer->Update(window);
	
	///// 読み返し文字さん、こんにちは。ごきげんいかがですか？
	//Update_ReReadCharList ( window );

}
//=========================================================
// 読み返し文字さんたちに位置についてもらう
//=========================================================
void TextMessageWindow::Update_ReReadCharList( sf::RenderWindow &window ) {

	///// 読み返し文字さん、こんにちは。ごきげんいかがですか？
	sf::Vector2f cursorPos = _offset;
	for ( unsigned int i = 0; i < _reReadRowListCurrent.size(); ++i) {

		for ( unsigned int n = 0; n < _reReadRowListCurrent[i].size(); ++n) {

			_reReadRowListCurrent[i][n]->SetPosition ( _position + cursorPos );
			_reReadRowListCurrent[i][n]->Update( window );

			// 参考：通常モードのときにセットしている位置
			//tc->SetPosition ( _cursorPosition + _position + _offset );

			cursorPos.x += _reReadRowListCurrent[i][n]->GetWidth();
		}

		cursorPos.y += _rowHeight;
		cursorPos.x = _offset.x;
	}
}


//=========================================================
// 読み返しモードのときの操作受付
//=========================================================
void TextMessageWindow::ReReadMode_Operation () {

	///// マウス右クリックで解除
	if ( _mouseStateR == TextMessageWindow::MOUSE_STATE_RELEASE ) {
		
		_isReReadMode = false;
		return;
	}
	
	///// 最下段でホイールグリグリで解除
	if ( _scenarioLogViewer->IsGoOverTheBottom() ) {
		
		_isReReadMode = false;
		return;
	}
	/*
	///// 上へ
	if ( Application::GetInstance()->_mouseWheel > 0 ) {

		// 上限判定：上に行く余地があれば、上に行く
		if (_reReadRangeFrom > 0) { 

			_reReadRangeFrom += -1;
			_reReadRangeTo   += -1;

			///// From から To の分の行を確保
			SetUpCurrentReReadCharList();
		}

	///// 下へ
	} else if ( Application::GetInstance()->_mouseWheel < 0 ) {


		// 下限判定というか、下限にひっかかったら、読み返しモードを解除する。
		if ( _reReadRangeTo >= _reReadRowList.size() - 1 ) {

			_isReReadMode = false;
			_reReadRowListCurrent.clear();

		// 下に行く
		} else {

			_reReadRangeFrom += 1;
			_reReadRangeTo   += 1;

			///// From から To の分の行を確保
			SetUpCurrentReReadCharList();
		}
	}
	*/
}

//=========================================================
// 現在の分の読み返し行を求める
//=========================================================
void TextMessageWindow::SetUpCurrentReReadCharList() {

	///// From から To の分の行を確保
	_reReadRowListCurrent.clear();
	for ( int i = _reReadRangeFrom; i <= _reReadRangeTo; ++i ) {

		if ( i >= 0 && i < _reReadRowList.size() ) {
			_reReadRowListCurrent.push_back( _reReadRowList[i] );
		}
	}
}

//=========================================================
// 読み返し用：１行の幅を求める
//=========================================================
/*
int TextMessageWindow::GetRowHeight ( vector< vector< TextChar* > > &reReadRowList ) {

	int result = 0;

	vector<int> rowHeightList;
	for ( unsigned int i = 0; i < reReadRowList.size(); ++i ) {

		vector<int> heightList;
		for ( unsigned int n = 0; n < reReadRowList[i].size(); ++n ) {

			heightList.push_back ( reReadRowList[i][n]->GetHeight() );
		}

		// この行内の文字の高さの平均値
		int avarage = 0;
		for ( unsigned int n = 0; n < heightList.size(); ++n ) {
			avarage += heightList[n];
		}
		avarage = ( heightList.size() > 0 ) ? avarage / heightList.size() : 0;


		// この行の高さ、とする
		rowHeightList.push_back ( avarage );
	}

	///// 各行の高さの平均値
	result = 0;
	for ( unsigned int n = 0; n < rowHeightList.size(); ++n ) {
		result += rowHeightList[n];
	}
	result = ( rowHeightList.size() > 0 ) ? result / rowHeightList.size() : 0;


	return result;
}
*/

//=========================================================
// 全ての文字を表示し終えているか？
//=========================================================
bool TextMessageWindow::IsAllCharShown() {

	return ((_textCharList_Reserved.size() == 0) && !_isWaitingClick);
}

//=========================================================
// １文字ずつ表示するトリック
//=========================================================
void TextMessageWindow::ShowCharOneByOne(sf::RenderWindow &window) {

	--_charShowCount;
	
	bool isSkipForced = false;
	/*
		既読スキップするかどうかのスイッチ。

		ON だとして、
		既読だったら、_charShowCountを待たずに進める？
		というか、既読じゃなくなるまで、ProgressOneChar() を回すことしましょう。
		あと何回 ProgressOneChar() を実行すれば既読じゃなくなるのか、最後まで回すことになるのか、
		を、既読管理クラスに問い合わせる。スキップが ON だったら。

	*/
	//if (_readManagerReceptionist->IsSkip() ) {
	if (_readManagerReceptionist->IsWannaSkip() ) {

		// どこまで既読？
		//int readFinishedLength = _readManagerReceptionist->GetReadFinishedLength();
		
		
		/*
			AddText()の段階でユニットごと既読である場合は、丸ごと既読でテキスト追加している。
			ここに来るのは、
			
			・ユニットの途中まで既読の場合
			・既読ユニットだけれど、途中でSkip開始した場合
			
			後者の場合に、既読ユニットならば残りすべて一気表示してしまいたい。
			
		*/
		
		// 丸ごと既読ユニットである場合
		if ( _readFinishedLength == ReadTextManager::ALL_FINISHED ) {

			_isWaitingClick = false;
			isSkipForced = true;
			
			// 現在、スキップ中
			_readManagerReceptionist->SkippingNow(true);
		
		// 丸ごと既読ユニットというわけではない場合
		} else {
			
			// 今、既読の範囲内？
			if ( _textCharList.size() < _readFinishedLength) {

				// さっさと進めてしまいましょう。
				_charShowCount  = 0;
				_isWaitingClick = false;
				
				// 現在、スキップ中
				_readManagerReceptionist->SkippingNow(true);

			} else {

				// スキップ終了：メニューのスキップ項目をOFFにする
				//_readManagerReceptionist->SkipEnd();
				
				// 現在、スキップ中ではない
				_readManagerReceptionist->SkippingNow(false);
			}
		}
	}
	
	///// 既読じゃないけど強制スキップ
	//bool isSkipForced = false; // 定義を関数冒頭に移動
	if ( Application::GetInstance()->HasFocus() && sf::Keyboard::isKeyPressed ( sf::Keyboard::Return ) ) {
		
		//printf("既読じゃないけど強制スキップ\n");
		/*
		// さっさと進めてしまいましょう。
		_charShowCount  = 0;
		_isWaitingClick = false;
		*/
		
		_isWaitingClick = false;
		isSkipForced = true;
	}

	///// 文字が流れている最中に左クリックして早送りになるアレ
	if ( !_isWaitingClick && _mouseStateL == MOUSE_STATE_RELEASE ) {
		_isTemporaryQuick = true;
	}
	// いつ解除されますか？：全部表示しきったら解除
	if ( _isTemporaryQuick ) {
		if (_textCharList_Reserved.size() == 0) {
			_isTemporaryQuick = false;
		}
	}
	
	// 強制スキップです
	if (isSkipForced) {
		
		while ( _textCharList_Reserved.size() ) {
			ProgressOneChar();
		}
			
			
	// 強制スキップではありません
	} else if ( _charShowCount <= 0 || _isTemporaryQuick) {

		// クリック待ち？
		if ( _isWaitingClick ) {

			_charShowCount = 0;

			_isTemporaryQuick = false;

		// クリック待ちじゃない
		} else {

			// 進める：なぜこの処理を分けているか？ ロード時にクリック待ちを無視して進めるため。
			ProgressOneChar();

		} // クリック待ち？
	}
}

//=========================================================
// １文字進める
//=========================================================
void TextMessageWindow::ProgressOneChar() {
	
	_charShowCount = _charShowCountMax;
	
	//printf ("ProgressOneChar()：|");
	//for ( unsigned int i = 0; i <_textCharList_Reserved.size(); ++i ) {
	//	printf( "%s|", _textCharList_Reserved[i]->GetChar().c_str());
	//}
	//printf("\n");

	// 予約リストから１文字追加
	if ( _textCharList_Reserved.size() ) {

		TextChar *tc = *_textCharList_Reserved.begin();
		_textCharList_Reserved.erase ( _textCharList_Reserved.begin() );
		_textCharList.push_back( tc );
		
		// 改行判定用に「次の文字」も取得しておく
		TextChar *nextChar = NULL;
		if ( _textCharList_Reserved.size() > 0 ) {
			nextChar = *_textCharList_Reserved.begin();
		}

		_charCount_sinceLastAddText++;
		
		_isWaitingClick = false; // ロード時、まとめて進める際、途中にクリック待ちがあると、それ以降クリック待ちのままになってしまう。ゆえに、クリック待ちじゃないならfalseにしておく

		// 既読管理さんに、既読情報を通知する。「１文字進んだ」
		// これだと上書きになってしまいません？
		_readManagerReceptionist->Finished ( _textCharList.size() );

		StoreCharToScenarioLog ( tc->GetChar() );

		// 改ページ？
		if ( tc->GetType() == TextChar::TEXT_CHAR_TYPE_NEW_PAGE ) {

			NewPage();
			
			FlushToScenarioLog();

			// すぐに次の文字を表示してね。
			_charShowCount = 0;
			

		// クリック待ち？
		} else if ( tc->GetType() == TextChar::TEXT_CHAR_TYPE_WAIT_CLICK ) {

			_isWaitingClick = true;
			
			// すぐに次の文字を表示してね。いや、表示はされないんですけども。
			_charShowCount = 0;
			
			FlushToScenarioLog();
			
		} else {

			// 改行する？
			bool isNewLine = false;
			
			// 改行するかどうか：改行文字である。
			if ( tc->GetType() == TextChar::TEXT_CHAR_TYPE_NEW_LINE ) {

				isNewLine = true;

				// すぐに次の文字を表示してね。
				_charShowCount = 0; 

			
			// 明示的な改行文字はなくても、改行しなければならないときがある
			} else {
				
				// 文末に来てはいけない文字かもしれない
				if ( IsHyphenatedChar_notSuitableForTheTailOfLine (tc->GetChar()) ) {
				
					//printf ("文末に来てはいけない文字だぞ [   %s   ]\n", tc->GetChar().c_str());
					
					// 次の文字を表示するとハミ出るなら、先に改行する
					int nextCharWidth = (nextChar != NULL) ? nextChar->GetWidth() : 0;
					
					//if ( nextChar != NULL ) {
					//	printf ( "次の文字：%s\n", nextChar->GetChar().c_str()  );
					//}
					
					if ( _cursorPosition.x + tc->GetWidth() + nextCharWidth > _size.x) {
						
						//printf ("文末判定！ 改行するぞ！\n");
						isNewLine = true;
					} else {
						//printf("まだ文末じゃないぜ\n");
					}
				}
				

				// 改行するかどうか：カーソルの位置に表示すると、行をハミ出す？ 改行が必要？
				if ( _cursorPosition.x + tc->GetWidth() > _size.x) {
				
					//printf ("改行しない方の禁則処理にひっかかってる？\n");

					// 禁則処理
					if ( !IsHyphenatedChar_notSuitableForTheTopOfLine ( tc->GetChar() ) )
					{
						isNewLine = true;
					}
				}
			}

			// 改行する？
			if ( isNewLine) {

				_cursorPosition.x = 0;
				_cursorPosition.y += _rowHeight;

				// 縦幅をハミ出しているなら、スクロール
				if ( _cursorPosition.y + _rowHeight > _size.y ) {
					Scroll();
				}
				
				
			}
			
			// カーソルの位置に表示するのですよ。
			tc->SetPosition ( _cursorPosition + _position + _offset );

			// カーソルを動かすのですよ。
			_cursorPosition.x += tc->GetWidth();
		}

	} // 予約リストから１文字追加
	
	// 全部表示し終えたらログへフラッシュ
	else {
		FlushToScenarioLog();
	}

}

//=========================================================
// 改ページ
//=========================================================
void TextMessageWindow::NewPage() {

	/*
		改ページとは？

		・今表示されているものを全てクリア
		・カーソルを原点に戻す
	*/

	///// クリア
	PutScrolledRowAway ( _textCharList );
	_textCharList.clear();

	///// カーソルを戻す
	_cursorPosition.x = 0.0f;
	_cursorPosition.y = 0.0f;


	// 改ページのタイミングで、セーブ/ロード用の履歴をクリアする。
	_addTextHistory.clear();
}


//=========================================================
// スクロール
//=========================================================
void TextMessageWindow::Scroll() {

	/*
		スクロールとは？

		・最初の行を削除
		・以降の文字の表示位置を、１行の高さ分、上に移動
		・カーソルも、１行の高さ分、上に移動
	*/

	///// １行目とそれ以降を分離して、それ以降を残す
	/*
		表示幅が規定をハミ出る、または改行文字が現れるまで。
	*/
	vector<TextChar*> firstRow;
	vector<TextChar*> restCharList;

	int rowWidth = 0;
	bool isAlreadyPassFirstRow = false;
	
	
	// 一番左上にくる文字の位置、って、こういうことですよね？
	int firstCharPosY = _position.y + _offset.y;

	for ( unsigned int i = 0; i < _textCharList.size(); ++i) {
		
		
		// １文字目と表示位置のY値が同じなら「一行目」と判定する
		if ( _textCharList[i]->GetPosition().y == firstCharPosY ) {
				
			firstRow.push_back ( _textCharList[i] );
			
		} else {
			
			restCharList.push_back ( _textCharList[i] );
		}
		
		/*

		// 横幅加算。どこまでも。
		rowWidth += _textCharList[i]->GetWidth();
		
		// 禁則判定
		bool ishyphenation = IsHyphenatedChar_notSuitableForTheTopOfLine ( _textCharList[i]->GetChar() );

		// １行目判定
		if ( rowWidth > _size.x && !ishyphenation && _textCharList[i]->GetType() == TextChar::TEXT_CHAR_TYPE_CHARACTER ) {
			isAlreadyPassFirstRow = true;
		}

		if ( !isAlreadyPassFirstRow ) {

			firstRow.push_back ( _textCharList[i] );
		} else {
			restCharList.push_back ( _textCharList[i] );
		}

		// 改行判定
		if ( _textCharList[i]->GetType() == TextChar::TEXT_CHAR_TYPE_NEW_LINE ) {
			isAlreadyPassFirstRow = true;
		}
		*/

	}
	
	/*
	printf("１行目：");
	for ( unsigned int i = 0; i < firstRow.size(); ++i ) {
		printf("%s", firstRow[i]->GetChar().c_str());
	}
	printf("\n");
	*/

	///// 以降の行、表示位置を１行分上にずらす
	for ( unsigned int i = 0; i < restCharList.size(); ++i ) {

		sf::Vector2f pos = restCharList[i]->GetPosition();
		pos.y -= _rowHeight;
		restCharList[i]->SetPosition ( pos );
	}

	///// カーソルも、１行分上にずらす
	_cursorPosition.y -= _rowHeight;

	///// １番上の行を後始末
	_textCharList = restCharList;
	PutScrolledRowAway ( firstRow );

}

//=========================================================
// 行に分割する
//=========================================================
/*
vector < vector<TextChar*> > TextMessageWindow::SplitIntoRows( vector<TextChar*> &charList ) {
	
	// TODO SplitIntoRows これも、読み返しログを専用のログクラスに委託するようにしてから、使ってませんよね？
	//	横幅が、１文字の幅より狭いケースは想定してません。そんなことしないでね。
	

	vector < vector<TextChar*> > rowList;

	int rowWidth = 0;
	vector<TextChar*> row;
	for ( unsigned int i = 0; i < charList.size(); ++i ) {

		// 横幅加算
		rowWidth += charList[i]->GetWidth();

		// 禁則判定
		bool ishyphenation = IsHyphenatedChar_notSuitableForTheTopOfLine ( charList[i]->GetChar() );

		// 横幅ハミ出た？
		if ( rowWidth > _size.x && !ishyphenation) {

			// この時点での行を、行リストに移す。
			rowList.push_back ( row );

			// 行をクリアして、今の文字を追加（改行）
			row.clear();
			row.push_back( charList[i] );
			rowWidth = charList[i]->GetWidth();

		// ハミ出てない
		} else {
			
			// 現在の行に追加
			row.push_back ( charList[i] );
			
			// TODO ここを使うなら、ここにも禁則処理（行頭・行末）を入れる必要があるわけですが……
		
		}
	}
	if ( row.size() > 0 ) {
		rowList.push_back(row);
	}
	

	return rowList;
}
*/

//=========================================================
// スクロール時、削った最上行の文字を始末
//=========================================================
void TextMessageWindow::PutScrolledRowAway ( vector<TextChar*> &firstRow ) {

	/*
		改ページのときにもここを呼んでいるので、最上行とは限らない。

		行で分割して、読み返し行に移す。
	*/
	
	// 解放はしておきたいですよね？
	for ( unsigned int i = 0; i < firstRow.size(); ++i ) {

		delete firstRow[i];
	}
	firstRow.clear();

	
	// TODO ここ結局使ってませんよね？ 読み返し機能を専用のログクラスに委託することになってから
	/*
	vector< vector<TextChar*> > rowList =  SplitIntoRows ( firstRow );

	for ( unsigned int i = 0; i < rowList.size(); ++i ) {

		//printf("削られる行:\n------------\n");
		
		///// 読み返し用に色を変えてみる
		for ( unsigned int n = 0; n < rowList[i].size(); ++n) {
			rowList[i][n]->SetColor ( _fontColor_reRead );
			
			//printf("%s", rowList[i][n]->GetChar().c_str());
		}
		//printf("\n-------------\n");

		_reReadRowList.push_back ( rowList[i] );

		// 読み返し行の上限
		if ( _reReadRowList.size() > _reReadLimitRowNum ) {

			for ( unsigned int n = 0; n < (*_reReadRowList.begin()).size(); ++n) {
				delete (*_reReadRowList.begin())[n];
			}

			_reReadRowList.erase ( _reReadRowList.begin() );
		}
	}

	// 引数で渡された入れ物はクリアしておく。
	firstRow.clear();
	
	*/


	/*
		どう始末する？
		既読リストに移したい。

		とりあえず、解放しておく。
	*/
	/*
	for ( unsigned int i = 0; i < firstRow.size(); ++i ) {

		delete firstRow[i];
	}
	firstRow.clear();
	*/
}

//=========================================================
// 描画
//=========================================================
void TextMessageWindow::Draw  ( sf::RenderWindow &window ) {

	///// 非表示モード？
	if ( !_isShown ) {
		return;
	}
	
	sf::View defaultView = window.getDefaultView();
	sf::View currentView = window.getView();
	
	sf::Transform defaultTransForm = defaultView.getTransform();
	sf::Transform currentTransForm = currentView.getTransform();
	
	float defaultRotation = defaultView.getRotation();
	float currentRotation = currentView.getRotation();
	
	// 位置はデフォルト、回転は反映させる
	defaultView.setRotation ( currentRotation );
	
	// 画面ゆれエフェクトを反映させる
	defaultView.move ( Effect_Shake::_shakeXforTheWorld, Effect_Shake::_shakeYforTheWorld );

	// ではよろしくー
	window.setView ( defaultView );

	///// フレームで描く？
	if ( _isFrameOn ) {
		window.draw ( _frame );
	}
	///// 画像で下敷きを描く？
	if ( _isBackImageOn ) {
		window.draw ( _sprite );
	}

	///// 話者名表示する？
	if ( _speakerName != "" && _speakerNameDisplayType == SPEAKER_NAME_DISPLAY_TYPE_OUT) {
		
		// 話者名の枠画像を表示する？
		if ( _isSpeakerNameBackImageOn ) {
			
			// デフォルト版
			if ( _currentamePlateNum < 0 ) {
				//printf("デフォルト版を使用 : %d\n", _currentamePlateNum);
				window.draw ( _speakerNameBackImage_sprite );
				
			// ワイド版
			} else {
				//printf("ワイド版を使用 : %d\n", _currentamePlateNum);
				// TODO とりあえず、ワイド版は1つのみ
				window.draw ( _speakerNameBackImage_wide_sprite[_currentamePlateNum] );
			}
		}
		
		window.draw ( _speakerName_sprite );
	}


	///// 通常モード？
	if ( !_isReReadMode ) {

		for ( unsigned int i = 0; i < _textCharList.size(); ++i ) {

			_textCharList[i]->Draw ( window );

		}

		///// クリック待ち？
		if ( _isWaitingClick ) {

			// クリック待ちカーソルがあれば、使う
			if ( _waitingClickCursor != NULL ) {
				_waitingClickCursor->Draw ( window );
			}
		}

	///// 読み返しモード
	} else {
		
		_scenarioLogViewer->Draw(window);
		
		/*
		for ( unsigned int i = 0; i < _reReadRowListCurrent.size(); ++i) {
			for ( unsigned int n = 0; n < _reReadRowListCurrent[i].size(); ++n) {

				_reReadRowListCurrent[i][n]->Draw ( window );
			}
		}
		*/
	}
	
	window.setView ( currentView );

}


//=========================================================
// 読み返しログ用の文字列を溜め込んでいく
//=========================================================
void TextMessageWindow::StoreCharToScenarioLog ( string log ) {
	
	
	//log = Utils_String::Trim ( log );
	
	/*
	log = Utils_String::Replace ( log, "\t", "" );
	//log = Utils_String::Replace ( log, CONTROL_CHAR_NEW_LINE, "\n" );
	log = Utils_String::Replace ( log, CONTROL_CHAR_NEW_LINE, "" );
	log = Utils_String::Replace ( log, CONTROL_CHAR_NEW_PAGE, "" );
	log = Utils_String::Replace ( log, CONTROL_CHAR_WAIT_CLICK, "" );
	*/
			
	_scenarioLogBuffer = _scenarioLogBuffer + log;
	
}

//=========================================================
// 溜まっている読み返しログを書き込む
//=========================================================
void TextMessageWindow::FlushToScenarioLog () {
	
	if ( _scenarioLogBuffer == "" ) {
		//printf ("_scenarioLogBuffer が 空なので読み返しログを書き込まない\n");
	}
	else if (_scenarioLogger != NULL) {

			// TODO 「」で囲っているかどうかを見るためだけに全部を配列化するのは効率が悪いですか？
			vector<string> l = SplitOneByOne(_scenarioLogBuffer);
			
			// セリフですか？（一文字目が "「" で、行末以外のところで "」" で閉じてなければセリフとみなす。
			// たとえば、行の冒頭でカギカッコの引用が入っているものをセリフとみなさないように。
			// セリフが終わる前に改行をしている可能性もあるので、行末で閉じていなくてもセリフとみなす。
			bool isNarative = false;
			if ( l[0] == _speakFrom ) {
				
				if (l[l.size()-1] == _speakEnd) {
					
					isNarative = true;
					
				} else {
					
					isNarative = true;
					
					for ( int i = 1; i < l.size()-1; ++i ) {
						
						if ( i >= l.size() ) {
							break;
						}
						
						if ( l[i] == _speakEnd) {
							isNarative = false;
						}
					}
					
				}
			}
			
			//if ( l[0] == _speakFrom && l[l.size()-1] == _speakEnd ) {
			if ( isNarative ) {
				
				if ( _speakerName == "" ) {
					_scenarioLogBuffer = "\n" + _scenarioLogBuffer;
				} else {
					_scenarioLogBuffer = "\n" + _speakerName + "\n" + _scenarioLogBuffer;
				}
				
				// 次の書き込みのために、「直前のログはセリフだった」と言う。
				//_isSpeakingRightBefore = true;
				
				// セリフの最中ですよ
				_isSpeakingStarted = true;
				
				// 最中なのであって、まだ終わっていない。終わってるかもしれないけれど、その判定をこの後でするのですよ。
				_isSpeakingEndedRightBefore = false;
				
			} else {
				

				//_isSpeakingRightBefore = false;
				
				// 直前のログがセリフであったならば、行を空ける。見た目の問題。
				if ( _isSpeakingEndedRightBefore ) {
					_scenarioLogBuffer = "\n" + _scenarioLogBuffer;
				}
			
				_isSpeakingEndedRightBefore = false;
			}
			
			
			// セリフ終了？ ならばこの次、一行空けて欲しいですよね。
			// ただし、セリフじゃないなら。
			// セリフが連続するなら、セリフの側で「セリフの冒頭」に改行を追加している。
			// 「セリフの末尾」としての改行を重ねる必要はない。見た目の問題。
			if ( _isSpeakingStarted ) {
				
				if ( l[l.size()-1] == _speakEnd ) {
					_isSpeakingEndedRightBefore = true;
				}
			}
		
		
		_scenarioLogger->WriteScenarioLog ( _scenarioLogBuffer );
		
		_scenarioLogBuffer = "";
		
	} else {
		printf ( "IScenarioLogger が NULL なので読み返しログを書き込めない\n" );
	}
}

//=========================================================
///// 全部表示し終わっているか？ クリック待ちでもなく（試合中のアクターのセリフとして、「完了」しているかどうかの判定が目的）
//=========================================================
bool TextMessageWindow::IsAllFinished() {
	
	bool result = false;
	
	if ( _textCharList_Reserved.size() == 0 ) {
		
		if ( !_isWaitingClick ) {
			result = true;
		}
	}
	
	return result;
}

//=========================================================
// テキストを追加：既読として
//=========================================================
void TextMessageWindow::AddText_asFinishedText ( string text ) {

	// 色指定なし：デフォルト色を使う
	AddText_asFinishedText ( text, _fontColor );

}
void TextMessageWindow::AddText_asFinishedText ( string text, sf::Color fontColor ) {

	// まずは普通に追加して、
	AddText (text, fontColor);

	// 全部既読にしてしまう。
	while (_textCharList_Reserved.size() > 0) {

		ProgressOneChar();
	}

	//printf("既読扱いで AddText \n");
}

//=========================================================
// テキストを追加
//=========================================================
void TextMessageWindow::AddText ( string text ) {

	// 色指定なし：デフォルト色を使う
	AddText ( text, _fontColor );
}

void TextMessageWindow::AddText ( string text, sf::Color fontColor ) {
	
	/*
	if (_scenarioLogger != NULL) {
		
		string scenarioLog = text;
		
		scenarioLog = Utils_String::Trim ( scenarioLog );
		
		scenarioLog = Utils_String::Replace ( scenarioLog, "\t", "" );
		scenarioLog = Utils_String::Replace ( scenarioLog, CONTROL_CHAR_NEW_LINE, "\n" );
		scenarioLog = Utils_String::Replace ( scenarioLog, CONTROL_CHAR_NEW_PAGE, "" );
		scenarioLog = Utils_String::Replace ( scenarioLog, CONTROL_CHAR_WAIT_CLICK, "" );
		
		
		if (_speakerName != "") {
			scenarioLog = Utils_String::Replace ( scenarioLog, _speakFrom, _speakerName + _speakFrom );
		}
		
		_scenarioLogger->WriteScenarioLog ( scenarioLog );
		
	}
	*/
	/*
		結局、必ずこの関数を通る。
	*/
	
	_addText_latest = text; 
	_charCount_sinceLastAddText = 0;


	// セーブデータ用に保存しておく
	_addTextHistory.push_back ( pair<string, string>( text, SFML_Utils::Convert_color_to_string(fontColor) ) );

	// どこまで既読？
	_readFinishedLength = _readManagerReceptionist->GetReadFinishedLength();

	// テキスト追加処理本体
	AddText_Body(text, fontColor);
}
/*
//=========================================================
// テキストを追加：履歴には残さない
//=========================================================
void TextMessageWindow::AddText_outOfHistory ( string text ) {

	// 色指定なし：デフォルト色を使う
	AddText_outOfHistory ( text, _fontColor );
}
void TextMessageWindow::AddText_outOfHistory ( string text, sf::Color fontColor ) {

	// まずは普通に追加して、（処理本体のみ）
	AddText_Body (text, fontColor);

	// 全部既読にしてしまう。
	while (_textCharList_Reserved.size() > 0) {

		ProgressOneChar();
	}
}
*/

//=========================================================
// テキスト追加：処理本体
//=========================================================
void TextMessageWindow::AddText_Body ( string text, sf::Color fontColor ) {
	
	// えっと、引数で渡された全体色と、途中で変わるかもしれない現在色
	sf::Color baseColor    = fontColor;
	sf::Color currentColor = baseColor;
	
	// ここで文字列のエイリアスに置換
	ReplaceWithAlias ( text );
	
	// TODO この時点でログに書き込んでしまうの？ 
	//StoreCharToScenarioLog(text);
	//FlushToScenarioLog();

	// １文字ずつに分割
	vector<string> charList =  SplitOneByOne ( text );

	// 制御タグ
	string tag_Color = "[color:";
	string tag_Ruby  = "[ruby:";
	
	// 次の文字につけるルビ：これが空文字でないならば、次に文字を生成するときにルビをつけて、これを空文字に戻す
	string ruby = "";
	
	for ( unsigned int i = 0; i < charList.size(); ++i ) {

		TextChar *textChar = NULL;

		// 色変え制御タグ？
		// この場合、textChar が NULL の状態で _textCharList_Reserved.push_back ( textChar ); されるけど、いいの？
		// いや、そうはならない。「通常の文字」としてカウントされる。
		if ( IsColorTag ( charList, i, tag_Color, currentColor ) ) {
			
			// 末尾にタグが書かれていると。i がハミ出る
			if ( i >= charList.size() ) {
				break;
			}
		}

		// ルビ指定タグ？
		if ( IsRubyTag ( charList, i, tag_Ruby, ruby ) ) {
			
			if ( i >= charList.size() ) {
				break;
			}
		}

		// 改行文字？
		// if ( charList[i] == CONTROL_CHAR_NEW_LINE ) {
		if ( IsControlTag ( charList, i, CONTROL_CHAR_NEW_LINE ) ) {

			//textChar = new TextChar ( charList[i], 0, TextChar::TEXT_CHAR_TYPE_NEW_LINE );
			textChar = new TextChar ( "", 0, TextChar::TEXT_CHAR_TYPE_NEW_LINE );


		// 改ページ文字？
		//} else if ( charList[i] == CONTROL_CHAR_NEW_PAGE ) {
		} else if ( IsControlTag ( charList, i, CONTROL_CHAR_NEW_PAGE ) ) {

			//textChar = new TextChar ( charList[i], 0, TextChar::TEXT_CHAR_TYPE_NEW_PAGE );
			textChar = new TextChar ( "", 0, TextChar::TEXT_CHAR_TYPE_NEW_PAGE );

		// クリック待ち文字？
		//} else if ( charList[i] == CONTROL_CHAR_WAIT_CLICK ) {
		} else if ( IsControlTag ( charList, i, CONTROL_CHAR_WAIT_CLICK ) ) {

			//textChar = new TextChar ( charList[i], 0, TextChar::TEXT_CHAR_TYPE_WAIT_CLICK );
			textChar = new TextChar ( "", 0, TextChar::TEXT_CHAR_TYPE_WAIT_CLICK );

		// 通常の文字
		} else {
			textChar = new TextChar ( charList[i], _fontSize, TextChar::TEXT_CHAR_TYPE_CHARACTER, currentColor );
			
			if ( ruby != "" ) {
				textChar->SetRuby(ruby);
				ruby = "";
			}
			
		}

		// すぐに表示するのではなく、予約リストに追加する。
		_textCharList_Reserved.push_back ( textChar );

	}


}


//=========================================================
// 話者名をセット
//=========================================================
void TextMessageWindow::SetSpeakerName ( string speakerName ) {

	// この時点で話者名もエイリアスを通す
	ReplaceWithAlias (speakerName); 
	
	// エイリアスを通した上で、主人公名に該当するかどうかを判定する
	if ( !_isProtagonistShown ) {

		bool isProtagonist = false;
		
		for ( string name : _protagonists ) {
			
			if ( speakerName == name ) {
				isProtagonist = true;
				break;
			}
		}
		
		// 主人公名なので表示しない
		if ( isProtagonist ) {
			_speakerName = "";
			speakerName = "";
		}
	}
	
	
	///// 今設定されている名前と違うものであれば、変更処理を行う
	if ( _speakerName != speakerName) {

		_speakerName = speakerName;
	
		
		if ( _speakerName != "" && _speakerNameDisplayType == SPEAKER_NAME_DISPLAY_TYPE_OUT) {
			
			/*
			TextImageFactory::Create ( _speakerName_image, _speakerName, _fontSize, _fontColor );

			_speakerName_texture.loadFromImage ( _speakerName_image );
			_speakerName_texture.setSmooth ( true );
			
			// TODO こういうのってメモリリークしないんですか？
			// ちなみにこれをしないと、名札を変更しても、以前のサイズのままになってしまう。
			_speakerName_sprite = sf::Sprite();

			_speakerName_sprite.setTexture ( _speakerName_texture );
			*/
			// sf::Textを使う版
			_speakerName_sprite.setFont ( TextImageFactory::_font );
			_speakerName_sprite.setCharacterSize(_fontSize);
			_speakerName_sprite.setColor(_fontColor);
			_speakerName_sprite.setString ( SFML_Utils::GetSfString(_speakerName) );

			_speakerName_sprite.setPosition ( _position + _speakerOffset );
			
			// あとで中央寄せのために参照したいのです。
			sf::Sprite *currentNamePlate = &_speakerNameBackImage_sprite;

			///// 名札をハミ出る場合は名札を拡大する？
			_speakerNameBackImage_sprite.setScale(1.0f, 1.0f);
			float plateWidth = _speakerNameBackImage_sprite.getGlobalBounds().width;
			float nameWidth  = _speakerName_sprite.getGlobalBounds().width;
			
			// あんまりギリギリに収まるっていうのも困るので、文字サイズ分を余白とする
			float margin = _fontSize;
			
			// TODO とりあえずワイド版は一つのみ：リストにして複数持てるようにできたらいいな
			if ( plateWidth > 0.0f && (nameWidth + margin) > plateWidth ) {
				
				//printf ("ハミ出る\n");
				
				// ワイド版があればそれを使う
				if ( _speakerNameBackImageFileName_wide.size() > 0) {
					
					for ( unsigned int i = 0; i < _speakerNameBackImageFileName_wide.size(); ++i ) {
						
						float wideWidth = _speakerNameBackImage_wide_sprite[i].getGlobalBounds().width;
						
						// 収まるかな？
						if ( (nameWidth + margin) < wideWidth ) {
							_currentamePlateNum = i;
							currentNamePlate = &_speakerNameBackImage_wide_sprite[_currentamePlateNum];
							
							//printf ("ワイド版名札[%d]、収まる。これを使用\n", i);
							
							break;
							
						} else {
							//printf ("ワイド版名札[%d]、収まらず\n", i);
							
						}
					}
				}
				
				
				// ワイド版がないまたは収まるのがなかった場合はデフォルトのを拡大して使う
				if ( _currentamePlateNum < 0 ) {
					
					//printf ("デフォルトのを拡大して使うこととする\n");
					//printf ("%f / %f \n", nameWidth + margin, plateWidth);
					
					float scale = (nameWidth + margin) / plateWidth;
					//float margin = 1.1f;
					//printf ("%f\n", scale);
					
					_speakerNameBackImage_sprite.setScale ( scale, 1.0f );
					
				}
				
			// ハミ出ないならデフォルトのを等倍で使う
			} else {
				//printf ("ハミ出ない\n");
				_currentamePlateNum = -1;
				_speakerNameBackImage_sprite.setScale ( 1.0f, 1.0f );
			}
			

			// 中央寄せする？
			// TODO なんか中央寄せと言いつつ、当初はオフセットになっていたのですが、結局、「中央寄せか否か」になりました。じゃあ「数字での指定」はよろしくないのではありませんか？ ごもっとも。
			if ( _speakerOffset_centering != 0.0f ) {
				
				float moveXToCenter = 0.0f;
				moveXToCenter -= _speakerName_sprite.getGlobalBounds().width / 2.0f;
				//moveXToCenter += _speakerOffset_centering / 2.0f;
				moveXToCenter += currentNamePlate->getGlobalBounds().width / 2.0f;
				
				// 小数点以下が存在すると画像がブレるので整数に丸めるのです。
				int moveToXCenter = (int)(moveXToCenter + 0.5f);
				
				_speakerName_sprite.move ( moveToXCenter, 0.0f );
			}
			

			
			
			
		}
	}

}

		
//=========================================================
// 制御文字？
//【引数】
// int &currentIndex      ：制御文字だったら、いい位置まで進めてしまう。
// sf::Color &currentColor：制御文字だったら、色を変えてしまう。 
//=========================================================
bool TextMessageWindow::IsColorTag ( vector<string> &charList, unsigned int &currentIndex, const string &colorTag, sf::Color &currentColor ) {

	/*
		タグは半角だけの前提。
	*/
	
	bool result = false;

	///// タグの長さよりも、残り文字列の方が短いので、もうタグである可能性はない
	if (  (currentIndex + colorTag.size()) > charList.size()) {

		return result;
	}

	///// タグの文字分、文字列リストから取り出す
	string counterPart = "";
	for ( unsigned int i = 0; i < colorTag.size(); ++i ) {

		counterPart += charList[ currentIndex + i ];
	}

	///// タグかな？
	if ( colorTag == counterPart ) {

		///// じゃあ、タグがちゃんと書いてあるよね！
		// タグが閉じられるまでの文字列を取得
		string close = "]";
		string tagValue = "";
		for ( unsigned int i = currentIndex + colorTag.size(); i < charList.size(); ++i ) {

			if ( charList[i] == close) {
				break;
			}

			tagValue += charList[i];
		}

		///// 色買えタグだから、色情報だよね！
		currentColor = SFML_Utils::Convert_string_to_color ( tagValue );

		///// インデックスはいくつにすればいいのかな？
		currentIndex = currentIndex + colorTag.size() + tagValue.size() + close.size();
		
		result = true;
	}
	
	return result;
}

//=========================================================
// ルビ指定タグ？
//【引数】
// int &currentIndex ：タグだったら、いい位置まで進めてしまう。
// string &ruby      ：ルビ
//=========================================================
bool TextMessageWindow::IsRubyTag ( vector<string> &charList, unsigned int &currentIndex, const string &rubyTag, string &ruby ) {
	
	bool result = false;
	
	///// タグの長さよりも、残り文字列の方が短いので、もうタグである可能性はない
	if (  (currentIndex + rubyTag.size()) > charList.size()) {

		return result;
	}

	///// タグの文字分、文字列リストから取り出す
	string counterPart = "";
	for ( unsigned int i = 0; i < rubyTag.size(); ++i ) {

		counterPart += charList[ currentIndex + i ];
	}

	///// タグかな？
	if ( rubyTag == counterPart ) {

		///// じゃあ、タグがちゃんと書いてあるよね！
		// タグが閉じられるまでの文字列を取得
		string closeTag = "]";
		string tagValue = "";
		for ( unsigned int i = currentIndex + rubyTag.size(); i < charList.size(); ++i ) {

			if ( charList[i] == closeTag) {
				break;
			}

			tagValue += charList[i];
		}

		///// ルビタグだから、フリガナだよね！
		ruby = tagValue;
		
		int rubyLength = GUI_SFML::SFML_Utils::GetSfString(tagValue).getSize();

		///// インデックスはいくつにすればいいのかな？
		currentIndex = currentIndex + rubyTag.size() + rubyLength + closeTag.size();
		
		///// タグでした。
		result = true;
	}
	
	
	return result;
}



//=========================================================
// 制御文字？
//=========================================================
bool TextMessageWindow::IsControlTag ( vector<string> &charList, unsigned int &currentIndex, const string &controlTag ) {
	
	/*
		タグは半角だけの前提。
	*/
	
	bool result = false;

	///// タグの長さよりも、残り文字列の方が短いので、もうタグである可能性はない
	if (  (currentIndex + controlTag.size()) > charList.size()) {

		return result;
	}

	///// タグの文字分、文字列リストから取り出す
	string counterPart = "";
	for ( unsigned int i = 0; i < controlTag.size(); ++i ) {

		counterPart += charList[ currentIndex + i ];
	}

	///// タグかな？
	if ( controlTag == counterPart ) {
		
		result = true;
		
		///// インデックスはいくつにすればいいのかな？
		currentIndex = currentIndex + controlTag.size()
		-1 // 色変更タグと違って、処理上は「１文字」としてカウントされる。
		;
	}	
	
	
	return result;
}



//=========================================================
// マウスの状態を監視
//=========================================================
void TextMessageWindow::ObserveMouseState ( sf::RenderWindow &window ) {

	///// ウィンドウが最前面じゃないときには、マウス操作を受け付けない。
	/*
	HWND handle     =  window.getSystemHandle();
	HWND foreWindow = GetForegroundWindow();

	bool isTopWindow = ( handle == foreWindow );
	*/
	
	bool isTopWindow = Application::GetInstance()->HasFocus();
	
	//if ( !isTopWindow ) {
	if ( !Application::GetInstance()->HasFocus() ) {
		_mouseStateL = TextMessageWindow::MOUSE_STATE_NONE;
		_mouseStateR = TextMessageWindow::MOUSE_STATE_NONE;
		return;
	}
	
	/* フォーカスの有無だけでいいんじゃないですか？
	///// マウスがウィンドウの上に乗ってないときには、マウス操作を受け付けない
	if ( !Application::GetInstance()->_isMouseOn ) {
		_mouseStateL = TextMessageWindow::MOUSE_STATE_NONE;
		_mouseStateR = TextMessageWindow::MOUSE_STATE_NONE;
		return;
	}
	*/

	///// 左ボタン
	if ( _mouseStateL == TextMessageWindow::MOUSE_STATE_NONE ) {

		if ( sf::Mouse::isButtonPressed ( sf::Mouse::Left ) ) {
			_mouseStateL = TextMessageWindow::MOUSE_STATE_PRESS;
		}
	}
	else if ( _mouseStateL == TextMessageWindow::MOUSE_STATE_PRESS ) {

		if ( !sf::Mouse::isButtonPressed ( sf::Mouse::Left ) ) {
			_mouseStateL = TextMessageWindow::MOUSE_STATE_RELEASE;
		}
	}
	else if ( _mouseStateL == TextMessageWindow::MOUSE_STATE_RELEASE ) {

		_mouseStateL = TextMessageWindow::MOUSE_STATE_NONE;
	}
	
	///// 左クリック・キーボード編
	if ( Application::GetInstance()->_releasedKeyCode == sf::Keyboard::Space ) {
		_mouseStateL = TextMessageWindow::MOUSE_STATE_RELEASE;
	}
	
	///// 右ボタン
	if ( _mouseStateR == TextMessageWindow::MOUSE_STATE_NONE ) {

		if ( sf::Mouse::isButtonPressed ( sf::Mouse::Right ) ) {
			_mouseStateR = TextMessageWindow::MOUSE_STATE_PRESS;
		}
	}
	else if ( _mouseStateR == TextMessageWindow::MOUSE_STATE_PRESS ) {

		if ( !sf::Mouse::isButtonPressed ( sf::Mouse::Right ) ) {
			_mouseStateR = TextMessageWindow::MOUSE_STATE_RELEASE;
		}
	}
	else if ( _mouseStateR == TextMessageWindow::MOUSE_STATE_RELEASE ) {

		_mouseStateR = TextMessageWindow::MOUSE_STATE_NONE;
	}
}


//=========================================================
// string を １文字ごとに区切る
//=========================================================
vector<string> TextMessageWindow::SplitOneByOne ( string text ) {

	vector<string> result;
	
	int index = 0;
	int charStride = 0;
	
	while ( index < text.size() )
	{
		unsigned char c = text[index];
		if      ( (c >= 0b00000000) && (c <= 0b01111111) ) { charStride = 1;} 
		else if ( (c >= 0b11000010) && (c <= 0b11011111) ) { charStride = 2;} 
		else if ( (c >= 0b11100000) && (c <= 0b11101111) ) { charStride = 3;} 
		else if ( (c >= 0b11110000) && (c <= 0b11110111) ) { charStride = 4;} 
		else if ( (c >= 0b11111000) && (c <= 0b11111011) ) { charStride = 5;} 
		else if ( (c >= 0b11111100) && (c <= 0b11111101) ) { charStride = 6;}
		else                                               { printf("else\n");}
		
		string one_char = text.substr ( index, charStride );
		
		// 改行とタブ文字は省く
		if ( one_char != "\n" && one_char != "\r" && one_char != "\r\n" && one_char != "\n\r" && one_char != "\t") {

			result.push_back ( one_char );
		}
		
		index += charStride;
	}
	
	return result;
}



//=========================================================
// セーブ用のデータを作成
//=========================================================
TextMessageWindow::SaveData TextMessageWindow::GetSaveData() {

	TextMessageWindow::SaveData result;

	//result.addTextHistory   = _addTextHistory;
	result.addText_latest     = _addText_latest;
	result.stepped            = _charCount_sinceLastAddText;
	
	/*
	result.isWaitingClick   = _isWaitingClick;
	result.textCharListLength = _textCharList.size();

	result.reReadRowList.clear();
	for ( unsigned int i = 0; i < _reReadRowList.size(); ++i ) {

		string row = "";
		for ( unsigned int n = 0; n < _reReadRowList[i].size(); ++n ) {
			row += _reReadRowList[i][n]->GetChar();
		}

		result.reReadRowList.push_back ( row );
	}
	*/
	
	
	result.isShown = _isShown;
	result.speakerName = _speakerName;

	return result;
}

//=========================================================
// セーブ用のデータを元に再現
//=========================================================
void TextMessageWindow::Load ( SaveData &saveData ) {
	
	_addText_latest = saveData.addText_latest;
	Clear();
	
	// 読み返し履歴はクリア
	Release_ReReadCharList();
	
	_cursorPosition.x = 0.0f;
	_cursorPosition.y = 0.0f;
	
	_isWaitingClick = false;
	
	AddText ( _addText_latest );
	
	///// 表示されていたところまで進める
	for ( int i = 0; i < saveData.stepped; ++i ) {
		ProgressOneChar();
	}
	
	/*
	_addTextHistory.clear();
	Clear();

	///// カーソルを戻す
	_cursorPosition.x = 0.0f;
	_cursorPosition.y = 0.0f;

	///// AddText の履歴通りに再現
	for ( unsigned int i = 0; i < saveData.addTextHistory.size(); ++i ) {

		string    text      = saveData.addTextHistory[i].first;
		sf::Color fontColor = SFML_Utils::Convert_string_to_color ( saveData.addTextHistory[i].second );

		AddText ( text, fontColor );
	}

	///// 表示されていたところまで進める つまり、textCharListLength -1 ってことで、いいの？ -1でいいんでしたっけ。いいようなよくないような。
	// どうもよくないらしい。クリック待ち記号が二重に処理されてしまう。
	// クリック待ち状態もセーブされているのだから、変にきを利かせなくていい、ということなのだと思う。
	//for ( int i = 0; i < saveData.textCharListLength - 1; ++i ) {
	for ( int i = 0; i < saveData.textCharListLength; ++i ) {
		ProgressOneChar();
	}

	///// 読み返し行を再現
	Release_ReReadCharList();
	for ( unsigned int i = 0; i < saveData.reReadRowList.size(); ++i ) {

		string             row      = saveData.reReadRowList[i];
		vector <string>    charList = SplitOneByOne ( row );
		vector <TextChar*> rowList;

		for ( unsigned int n = 0; n < charList.size(); ++n ) {

			TextChar *textChar = NULL;
			int type = TextChar::TEXT_CHAR_TYPE_CHARACTER;
			if      ( charList[n] == CONTROL_CHAR_NEW_LINE   ) { type = TextChar::TEXT_CHAR_TYPE_NEW_LINE;}
			else if ( charList[n] == CONTROL_CHAR_NEW_PAGE   ) { type = TextChar::TEXT_CHAR_TYPE_NEW_PAGE;}
			else if ( charList[n] == CONTROL_CHAR_WAIT_CLICK ) { type = TextChar::TEXT_CHAR_TYPE_WAIT_CLICK;}

			textChar = new TextChar ( charList[n], _fontSize, type, _fontColor_reRead );

			rowList.push_back( textChar );
		}

		_reReadRowList.push_back ( rowList );
	}
	
	*/

	///// ロード後は読み返しモード解除でいいですよね。
	_isReReadMode = false;
	
	
	///// クリック待ちだった？
	//_isWaitingClick = saveData.isWaitingClick;
	

	///// クリアしておくといいかな、と思うのでそのようにしておくのですよー。
	_charShowCount = _charShowCountMax;
	
	
	
	_isShown = saveData.isShown;
	SetSpeakerName ( saveData.speakerName );
}


//=========================================================
// コピーというか、同じ設定のインスタンスを取得
//=========================================================
TextMessageWindow* TextMessageWindow::GetSameSetuppedInstance() {

	TextMessageWindow *result = new TextMessageWindow (NULL); // 既読管理者さんは 不在でいいですよね。

	/*
		設定値をコピー
	*/
	result->_position         = this->_position;
	result->_size             = this->_size;           // 横幅も、縦幅も、サイズ？ 行数ではなく？
	result->_offset           = this->_offset;         // 文字の表示を開始する位置のオフセット
	result->_fontSize         = this->_fontSize;       // 文字のサイズ
	result->_rowHeight        = this->_rowHeight;      // 行間（間というか、１行の高さ）
	//result->_cursorPosition   = this->_cursorPosition; // カーソル位置
	result->_fontColor        = _fontColor;            // 文字色
	result->_fontColor_reRead = _fontColor_reRead;     // 文字色：読み返し行

	//// フレームを背景に使う？
	result->_isFrameOn = _isFrameOn;
	result->_frame = _frame;

	//// 画像を背景に使う？
	result->SetBackImage (_backImageFileName);
	/*
		// これらは、SetBackImageの中でセットされる。
		result->_backImageFileName = _backImageFileName;
		result->_isBackImageOn = _isBackImageOn;
		result->_image   = _image;
		result->_texture = _texture;
		result->_sprite  = _sprite;
	*/

	///// 中に表示されているテキスト は コピーしない。
	/*
		vector<TextChar*> _textCharList;          // 実際に表示されている分
		vector<TextChar*> _textCharList_Reserved; // まだ表示されてない分
	*/

	/////// 話者名
	result->_speakerOffset = _speakerOffset; // 先に位置をセットしてから
	result->SetSpeakerName ( _speakerName );
	result->SetSpeakerNameDisplayType  ( _speakerNameDisplayType );
	result->SetSpeakFrom               ( _speakFrom );
	/*
		// これらは SetSpeakerName の 中 で セットされる。
		string _speakerName;
		sf::Image   _speakerName_image;
		sf::Texture _speakerName_texture;
		sf::Sprite  _speakerName_sprite;
	*/


	///// コピーしない
	// セーブ/ロード用
	//vector < pair <string, string > > _addTextHistory;

	///// コピーしない：設定値ではないから
	//int _charShowCount; // １文字ずつ表示するためのトリック
	//int _charShowCountMax;

	///// コピーしない：設定値ではないから
	//bool _isWaitingClick;           // クリック待ち？
	//Animation *_waitingClickCursor; // クリック待ちのときに表示するアイコン

	///// コピーしない：設定値ではないから
	//bool _isTemporaryQuick;         // 文字が流れている最中に左クリックしたときのアレ。

	///// コピーしない：コンストラクタでNULLを渡している
	//// 既読管理インターフェイス
	//IReadManagerReceptionist *_readManagerReceptionist;

	///// コピーしない：設定値ではない
	// 何文字目まで既読か？ AddTextの時点で、受け取っておく。
	//int _readFinishedLength; 

	///// コピーしない：設定値ではない
	//int _mouseState; // マウスの状態

	///// コピーしない：設定値ではない
	///// 読み返しモード関連
	//vector< vector< TextChar* > > _reReadRowList;        // 読み返し行リスト
	//vector< vector< TextChar* > > _reReadRowListCurrent; // 読み返し行リスト。現時点で表示している分
	//bool _isReReadMode;      // 今、読み返しモード？
	//int  _reReadRangeFrom;   // 読み返し行を表示する範囲
	//int  _reReadRangeTo;     // 読み返し行を表示する範囲
	//int  _reReadLimitRowNum; // 読み返せる行数の上限


	return result;
}


//=========================================================
// 禁則処理の文字かどうか？
//=========================================================
bool TextMessageWindow::IsHyphenatedChar_notSuitableForTheTopOfLine ( const string &c ) {
	
	bool ishyphenation = 
	(
		   c == "。"
		|| c == "、"
		|| c == "."
		|| c == ","
		|| c == "」"
		|| c == ")"
		|| c == "）"
		|| c == "!"
		|| c == "?"
		|| c == "！"
		|| c == "？"
		|| c == "\""
		|| c == "\'"
		|| c == "”"
		|| c == "’"
		|| c == " "
		|| c == "　"
		|| c == "ー"
		|| c == "っ"
		|| c == "ゃ"
		|| c == "ゅ"
		|| c == "ょ"
		|| c == "ぁ"
		|| c == "ぃ"
		|| c == "ぅ"
		|| c == "ぇ"
		|| c == "ぉ"
		|| c == "ッ"
		|| c == "ャ"
		|| c == "ュ"
		|| c == "ョ"
		|| c == "ァ"
		|| c == "ィ"
		|| c == "ゥ"
		|| c == "ェ"
		|| c == "ォ"
	);
	
	return ishyphenation;
}
bool TextMessageWindow::IsHyphenatedChar_notSuitableForTheTailOfLine ( const string &c ) {
	
	bool ishyphenation = 
	(
		   c == "「"
		|| c == "("
		|| c == "（"
		|| c == "\""
		|| c == "”"
		|| c == "\'"
		|| c == "’"

	);
	
	return ishyphenation;
}	


