//=========================================================
//
// AIさんのお世話係
// 
// 使う側は、スレッドのことを気にしなくていい、というつもり。
// 
// AIインスタンスそのものは「お客さん」として、その都度受け取る。
// つまり、AIそのものの設定は、済ませておいてね、ということです。
//
//=========================================================

#ifndef _AI_MANAGER_
#define _AI_MANAGER_

#include "../AI/AI.h"
#include "../AI/Move.h"
#include "../AI/InnerExpression.h"

#include <thread>

class AI_Manager {

public:

	//=========================================================
	// 思考の進行状況。
	//=========================================================
	struct ThinkProgress {

		///// 進捗率：負の値のとき、「開始前」とする。100.0f のとき、「思考終了後」とする。
		AI_Modules::ProgressData progressData;

		///// この局面について考え中です。：コピーを持つ。解放はこちらで行う。
		AI_Modules::InnerExpression *problemState;

		///// 進捗率 100% のとき、ここにある値が「回答」ということになる。
		//AI_Modules::Move answerMove;


		ThinkProgress() {
			progressData.progressRate = -1.0f;
			problemState = NULL; 
		}
	};

	AI_Manager();
	~AI_Manager();

	//=========================================================
	// 思考開始：開始できないときは false  
	//【引数】
	// こっちでコピーをとって持ち回ります。
	// なので、そっちはそっちで、そっちで生成したインスタンスの面倒を見てね。
	//=========================================================
	bool ThinkStart ( AI_Modules::AI *AI, AI_Modules::InnerExpression *currentState );

	//=========================================================
	// 進行状況を教えて。
	//【戻り値】
	// AIさんに参照を渡している変数のコピー。なのでスレッド的にも変なことにはならない？ と、いいな。
	//=========================================================
	AI_Modules::ProgressData GetThinkProgress();

	//=========================================================
	// クリア
	// 進行状況を -1.0f に戻すとか、そのへんの処理
	// また次回もよろしくおねがいします。
	//=========================================================
	void Clear();

	//=========================================================
	// 中止。
	//=========================================================
	void ThinkStop();

private:

	AI_Modules::AI *_guestAI; // お客さん
	ThinkProgress _currentProgresData;

#if defined(_WIN32)

	HANDLE _AI_ThreadHandle;
	CRITICAL_SECTION _criticalSection_For_AI;

#elif defined(__linux__)
	
	std::thread _AI_ThreadHandle;
	unsigned long _AI_Thread_Id;
	pthread_mutex_t _criticalSection_For_AI;

#endif

	

	//=========================================================
	// AI思考スレッド：入り口
	// 引数に GameManagerインスタンスを渡して、インスタンスメソッドの思考スレッドを呼ぶ
	//=========================================================
	static unsigned int AI_Thread_Entrance(void *obj);

	//=========================================================
	// AI思考スレッド：インスタンスごとの本体
	//=========================================================
	void AI_Thread();

	//=========================================================
	// AI思考スレッド（スレッドそのもの）が生きているかどうか？
	//=========================================================
	bool IsStillActive_AI_Thread ();

	//=========================================================
	// スレッドが残っていれば終了させる
	//=========================================================
	void EndThread();
};
#endif






