//=========================================================
//
// Application (Implement)
//
//=========================================================


#include "Application.h"
#include "Define.h"

#include "SceneScript.h"
#include "Piece.h"
#include "Card.h"

#include "Effect_SceneTransition.h"
#include "IDrawable.h"
#include "IMouseClingableObject.h"
#include "EffectManager.h"

#include "Delegate.h"
#include "../GUI_SFML/ImageManager.h"
#include "../GUI_SFML/SoundPlayer.h"
#include "../GUI_SFML/SFML_Utils.h"
#include "../GUI_SFML/TextImageFactory.h"
#include "../Utils/Utils_File.h"
#include "../Utils/Utils_String.h"

#include <SFML/Window.hpp>
#include <SFML/Audio.hpp>

#include <thread>
#include <chrono>


#include "Effect_Shake.h"

//=========================================================
// static
//=========================================================
Application* Application::_obj;


//=========================================================
// コンストラクタ
//=========================================================
Application::Application() {
	
	TextImageFactory::SetFont ( "./font/umefont_530/ume-tgc4.ttf" );
	
	_obj = this;
	_windowTitle = "ScriptEngine";
	
	sf::VideoMode videoMode = sf::VideoMode::getDesktopMode();
	unsigned int bitsPerPixel = videoMode.bitsPerPixel;
	printf ( " bitsPerPixel = %d\n", bitsPerPixel );
	
	//unsigned int bitsPerPixel = 32;
	
	_window.create(sf::VideoMode(APPLICATION::WINDOW_SIZE::WIDTH, APPLICATION::WINDOW_SIZE::HEIGHT, bitsPerPixel), GUI_SFML::SFML_Utils::GetSfString(_windowTitle), sf::Style::Close);
	_window.setFramerateLimit(60);
	_window.setVerticalSyncEnabled(true);
	_isFullScreen = false;
	
#if defined(_WIN32)
	// 多重起動防止用にタイトルをファイルに保存しておく（以後、SetWindowTitle ごとに）
	Utils_File::CreateAndOverwrite ( "title", _windowTitle);
#endif

	_window.setPosition ( sf::Vector2i(
		(sf::VideoMode::getDesktopMode().width  - _window.getSize().x) / 2,
		(sf::VideoMode::getDesktopMode().height - _window.getSize().y) / 2
	));

	//sf::View view = _window.getView();
	//sf::Vector2f viewCenter = view.getCenter();


	_isUpdateHalt = false;
	
	_effectManager    = new EffectManager();

	_mouseClingObj = NULL;

	
	
	_sceneList["SCENE_SCRIPT"] = new SceneScript();
	//_sceneList["SCENE_A"] = new SceneA();
	//_sceneList["SCENE_B"] = new SceneB();
	//_sceneList["GAME"]    = new SceneGame();
	// ↑こうやってシーンを追加する、というサンプル

	_currentScene = _sceneList["SCENE_SCRIPT"];
	_currentScene->Ready(_window);

	_mouseX = 0;
	_mouseY = 0;
	_mouseWheel = 0;

	_mouseState_Right  = MOUSE::BUTTON::STATE::NONE;
	_mouseState_Left   = MOUSE::BUTTON::STATE::NONE;
	_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
	
}

//=========================================================
// デストラクタ
//=========================================================
Application::~Application() {

printf ("0 Application::~Application()\n");

	while (_sceneList.size() > 0) {
		delete _sceneList.begin()->second;
		_sceneList.erase(_sceneList.begin());
	}

printf ("1 Application::~Application()\n");

	ImageManager::Release();

printf ("2 Application::~Application()\n");

	//CardFactory::Release();

printf ("3 Application::~Application()\n");

	GUI_SFML::SoundPlayer::Release();

printf ("4 Application::~Application()\n");
	
	delete _effectManager;

printf ("Application::~Application() done\n");
}

//=========================================================
// エフェクトを全部解放
//=========================================================
void Application::CleanUpAllEffect() {
	
	// エフェクトが出ているタイミングでロードする場合に備えて
	
	printf ( "エフェクトを全部解放 開始\n" );
	
	delete _effectManager;
	_effectManager = new EffectManager();

	printf ( "エフェクトを全部解放 完了\n" );
	
}

//=========================================================
// 実行
//=========================================================
void Application::Run() {

	bool running = true;
	while (running)
	{

		///// Mouse Management
		ManageMouseState();
		
		_releasedKeyCode = sf::Keyboard::Unknown;
		
		///// Watch Close Event
		sf::Event windowEvent;
		while ( _window.pollEvent(windowEvent))
		{

			if (windowEvent.type == sf::Event::Closed) {
				_window.close();
				running = false;
			/*
			} else  if (windowEvent.type == sf::Event::MouseWheelMoved) {

				_mouseWheel = windowEvent.mouseWheel.delta;
			*/
			} else  if (windowEvent.type == sf::Event::MouseWheelScrolled) {
				
				if (windowEvent.mouseWheelScroll.wheel == sf::Mouse::VerticalWheel) {
					_mouseWheel = windowEvent.mouseWheelScroll.delta;
				}

			} else  if (windowEvent.type == sf::Event::MouseEntered) {
				
				_isMouseOn = true;

			} else  if (windowEvent.type == sf::Event::MouseLeft) {
				
				_isMouseOn = false;
				
			} else  if (windowEvent.type == sf::Event::MouseMoved) {
				
				_isMouseOn = true;
				
			} else  if (windowEvent.type == sf::Event::KeyReleased) {
				
				_releasedKeyCode = windowEvent.key.code;
				
				if ( windowEvent.key.code == sf::Keyboard::Return && windowEvent.key.alt) {
					
					/*
					///// TODO Windows に 依存しているコード
					HWND hWindow = _window.getSystemHandle();
					char buf[1024];
					GetWindowText ( hWindow, buf, 1024);
					string title = buf;
					*/
					
					string title = GetWindowTitle();

					if ( _isFullScreen ) {
						
						
						sf::VideoMode videoMode = sf::VideoMode::getDesktopMode();
						unsigned int bitsPerPixel = videoMode.bitsPerPixel;
						printf ( " bitsPerPixel = %d\n", bitsPerPixel );
						
						int width  = APPLICATION::WINDOW_SIZE::WIDTH;
						int height = APPLICATION::WINDOW_SIZE::HEIGHT;

						_window.create(sf::VideoMode ( width, height, bitsPerPixel), title, sf::Style::Close);
						_window.setFramerateLimit(60);
						_window.setVerticalSyncEnabled(true);
						_isFullScreen = false;
						
						SetWindowTitle ( title );
						
						_window.setPosition ( sf::Vector2i(
							(sf::VideoMode::getDesktopMode().width  - _window.getSize().x) / 2,
							(sf::VideoMode::getDesktopMode().height - _window.getSize().y) / 2
						));

					} else {
						
						std::vector<sf::VideoMode> videoModes = sf::VideoMode::getFullscreenModes();
						if ( videoModes.size() > 0 ) {
							
							//unsigned int bitsPerPixel = 32; //videoModes[0].bitsPerPixel;
							unsigned int bitsPerPixel = videoModes[0].bitsPerPixel;
							
							printf ( "bitsPerPixel = %d\n", bitsPerPixel );
								
							_window.create(sf::VideoMode(APPLICATION::WINDOW_SIZE::WIDTH, APPLICATION::WINDOW_SIZE::HEIGHT, bitsPerPixel), title, sf::Style::Fullscreen);
							_window.setFramerateLimit(60);
							_window.setVerticalSyncEnabled(true);
							_isFullScreen = true;
							
						} else {
							printf ("Fullscreen failed!\n");
						}
						
						///// モニターの実サイズに合わせる版：ドットが粗くなるし、表示位置も崩れるし、メリットがあんまりないので封印。
						/*
						if ( sf::VideoMode::getFullscreenModes().size() > 0 ) {
							
							sf::VideoMode fullMode =sf::VideoMode::getFullscreenModes()[0];
							
							int displaySizeW = fullMode.width;
							int displaySizeH = fullMode.height;
							
							printf("(%d, %d)\n", displaySizeW, displaySizeH);
							
							_window.create(fullMode, title, sf::Style::Fullscreen);
							
							sf::View defaultView = _window.getView();
							defaultView.zoom ( (float)( (float)APPLICATION::WINDOW_SIZE::HEIGHT) / (float)displaySizeH);
							defaultView.move ( (displaySizeW - APPLICATION::WINDOW_SIZE::WIDTH)/-2.0f, (displaySizeH - APPLICATION::WINDOW_SIZE::HEIGHT)/-2.0f );
							_window.setView(defaultView);
							
							
							_window.setFramerateLimit(60);
							_window.setVerticalSyncEnabled(true);
							_isFullScreen = true;
						
						} else {
							printf("フルスクリーンにできないらしいですよ。\n");
						}
						*/
					}
					
					//hWindow = _window.getSystemHandle();
					//SetWindowText ( hWindow, title.c_str());
					SetWindowTitle ( title );

				}
			}
		}
		
		///// Update
		int updateResult = SCENE::UPDATE_RESULT::UPDATE_RESULT_NONE;
		if (!_isUpdateHalt) {
				
			// TODO _currentScene->IsRendered() が false のときにはマウスくっつきオブジェクトとエフェクトの更新も休むべきではないか？

			MouseClingObject_Update();
			updateResult = _currentScene->Update(_window);
			_effectManager->Update(_window);
			
			
			/*
			//カメラ移動実験
			
			
			sf::View view = _window.getView();
			
			if (sf::Keyboard::isKeyPressed(sf::Keyboard::Right)) {
				view.move ( 1.0f, 0.0f );
			}
			if (sf::Keyboard::isKeyPressed(sf::Keyboard::Left)) {
				view.move ( -1.0f, 0.0f );
			}
			if (sf::Keyboard::isKeyPressed(sf::Keyboard::A)) {
				
				sf::Vector2f center = view.getCenter();
				center += sf::Vector2f(10.0f, 0.0f);
				view.setCenter ( center );
				
			}
			if (sf::Keyboard::isKeyPressed(sf::Keyboard::S)) {
				
				sf::Vector2f center = view.getCenter();
				center += sf::Vector2f(-10.0f, 0.0f);
				view.setCenter ( center );
				
			}

			if (sf::Keyboard::isKeyPressed(sf::Keyboard::Return)) {
				
				sf::Vector2f center = view.getCenter();
				
				//center += sf::Vector2f(20.0f, 10.0f);
				center = sf::Vector2f(20.0f, 10.0f);
				
				view.setCenter ( center );
				
			}
			if (sf::Keyboard::isKeyPressed(sf::Keyboard::Insert)) {
				
				sf::Vector2f center = view.getCenter();
				
				//center += sf::Vector2f(20.0f, 10.0f);
				center = sf::Vector2f(200.0f, 10.0f);
				
				view.setCenter ( center );
				
			}
			
			_window.setView(view);


			if (sf::Keyboard::isKeyPressed(sf::Keyboard::Space)) {
				_window.setView(_window.getDefaultView());
			}
			*/

		}

		///// Draw
		if ( _currentScene->IsRendered() )  {
			
			_window.clear(sf::Color(0,0,0));
			//_window.clear(sf::Color(255,255,255));
			_currentScene->Draw(_window);
			DrawLaterObjList_Card();
			DrawLaterObjList_Piece();
			DrawLaterObjList();
			_effectManager->Draw(_window);
			MouseClingObject_Draw();

			_window.display();
			
		}

		///// シーン切り替え
		if ( updateResult == SCENE::UPDATE_RESULT::UPDATE_RESULT_CHANGE_SCENE) {

			string  nextSceneName = _currentScene->GetNextScene();

			DelegateBase *del =  Delegate<Application>::createDelegator_string( this, &Application::Transition );
			Effect_SceneTransition *effect = new Effect_SceneTransition(del, nextSceneName);
			_effectManager->Add(effect);

		}
		///// 終了
		else if ( updateResult == SCENE::UPDATE_RESULT::UPDATE_RESULT_END_APPLICATION) {

			_window.close();
			running = false;
		}
		
		//Sleep(1);
		std::this_thread::sleep_for(std::chrono::microseconds(1));
	}


}



//=========================================================
// Transite Scene
//=========================================================
void Application::Transition(string nextSceneName) {

	if (_sceneList.find( nextSceneName ) == _sceneList.end()) {
		printf("指定シーンが登録されていません：%s\n", nextSceneName.c_str());
		return;
	}

	IScene* nextScene     = _sceneList.find( nextSceneName )->second;


	_currentScene->Clear(_window);
	nextScene->Ready(_window);

	_currentScene = nextScene;

}

//=========================================================
// HasFouus with _isFullScreen in mind
//=========================================================
bool Application::HasFocus() {
	
	return _window.hasFocus() || _isFullScreen;
	
}

//=========================================================
// Manage MouseState
//=========================================================
void Application::ManageMouseState() {

	///// ウィンドウが最前面じゃないときには、マウス操作を受け付けない。
	/*
	HWND handle     =  _window.getSystemHandle();
	HWND foreWindow = GetForegroundWindow();

	bool isTopWindow = ( handle == foreWindow );
	*/
	bool isTopWindow = HasFocus();
	

	if ( !isTopWindow ) {
		_mouseState_Left   = MOUSE::BUTTON::STATE::NONE;
		_mouseState_Right  = MOUSE::BUTTON::STATE::NONE;
		_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
		
		return;
	}

	_mouseWheel = 0;

	sf::Vector2i mousePos = sf::Mouse::getPosition ( _window );
	
	_mouseX = mousePos.x;
	_mouseY = mousePos.y;

	// カメラの状態を考慮して「スクリーン上のマウスの位置」に変換する
	sf::Vector2f mouseInnerPos =  _window.mapPixelToCoords( mousePos );
	_mouseX = mouseInnerPos.x;
	_mouseY = mouseInnerPos.y;	
	
	
	///// MouseState_Left
	if ( _mouseState_Left == MOUSE::BUTTON::STATE::PRESS ) {
		
		if ( sf::Mouse::isButtonPressed( sf::Mouse::Left )) {
			_mouseState_Left = MOUSE::BUTTON::STATE::PRESS;
		} else {
			_mouseState_Left = MOUSE::BUTTON::STATE::RELEASE;
		}

	} else if (_mouseState_Left == MOUSE::BUTTON::STATE::NONE) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Left )) {
			_mouseState_Left = MOUSE::BUTTON::STATE::HAVE_PRESSED;
		} else {
			_mouseState_Left = MOUSE::BUTTON::STATE::NONE;
		}

	} else if (_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Left )) {
			_mouseState_Left = MOUSE::BUTTON::STATE::HAVE_PRESSED;
		} else {
			_mouseState_Left = MOUSE::BUTTON::STATE::NONE;
		}
	} else if (_mouseState_Left == MOUSE::BUTTON::STATE::HAVE_PRESSED) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Left )) {
			_mouseState_Left = MOUSE::BUTTON::STATE::PRESS;
		} else {
			_mouseState_Left = MOUSE::BUTTON::STATE::NONE;
		}
	} else {
		_mouseState_Left = MOUSE::BUTTON::STATE::NONE;
	}

	///// MouseState_Right
	if ( _mouseState_Right == MOUSE::BUTTON::STATE::PRESS ) {
		
		if ( sf::Mouse::isButtonPressed( sf::Mouse::Right )) {
			_mouseState_Right = MOUSE::BUTTON::STATE::PRESS;
		} else {
			_mouseState_Right = MOUSE::BUTTON::STATE::RELEASE;
		}

	} else if (_mouseState_Right == MOUSE::BUTTON::STATE::NONE) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Right )) {
			_mouseState_Right = MOUSE::BUTTON::STATE::HAVE_PRESSED;
		} else {
			_mouseState_Right = MOUSE::BUTTON::STATE::NONE;
		}

	} else if (_mouseState_Right == MOUSE::BUTTON::STATE::RELEASE) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Right )) {
			_mouseState_Right = MOUSE::BUTTON::STATE::HAVE_PRESSED;
		} else {
			_mouseState_Right = MOUSE::BUTTON::STATE::NONE;
		}
	} else if (_mouseState_Right == MOUSE::BUTTON::STATE::HAVE_PRESSED) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Right )) {
			_mouseState_Right = MOUSE::BUTTON::STATE::PRESS;
		} else {
			_mouseState_Right = MOUSE::BUTTON::STATE::NONE;
		}
	} else {
		_mouseState_Right = MOUSE::BUTTON::STATE::NONE;
	}

	///// MouseState_Middle
	if ( _mouseState_Middle == MOUSE::BUTTON::STATE::PRESS ) {
		
		if ( sf::Mouse::isButtonPressed( sf::Mouse::Middle )) {
			_mouseState_Middle = MOUSE::BUTTON::STATE::PRESS;
		} else {
			_mouseState_Middle = MOUSE::BUTTON::STATE::RELEASE;
		}

	} else if (_mouseState_Middle == MOUSE::BUTTON::STATE::NONE) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Middle )) {
			_mouseState_Middle = MOUSE::BUTTON::STATE::HAVE_PRESSED;
		} else {
			_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
		}

	} else if (_mouseState_Middle == MOUSE::BUTTON::STATE::RELEASE) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Middle )) {
			_mouseState_Middle = MOUSE::BUTTON::STATE::HAVE_PRESSED;
		} else {
			_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
		}
	} else if (_mouseState_Middle == MOUSE::BUTTON::STATE::HAVE_PRESSED) {

		if ( sf::Mouse::isButtonPressed( sf::Mouse::Middle )) {
			_mouseState_Middle = MOUSE::BUTTON::STATE::PRESS;
		} else {
			_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
		}
	} else {
		_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
	}
}

//=========================================================
// マウスにくっつけるオブジェクトを登録
//=========================================================
//int Application::AddMouseClingObj(IDrawable *obj) {
IMouseClingableObject* Application::AddMouseClingObj(IMouseClingableObject *obj) {

	IMouseClingableObject *result = _mouseClingObj;

	_mouseClingObj = obj;

	return result;

	/*
	int registNumber = _mouseClingObjTotalCount;
	++_mouseClingObjTotalCount;

	_mouseClingObjList[registNumber] = obj;


	return registNumber;
	*/
}

//=========================================================
// マウスにくっついているオブジェクトを取り外す
//=========================================================
//IDrawable* Application::DrawMouseClingObj(int registNum) {
IMouseClingableObject* Application::DrawMouseClingObj() {

	IMouseClingableObject *result = NULL;

	result = _mouseClingObj;
	_mouseClingObj = NULL;

	return result;
	/*
	IDrawable *result = NULL;


	if (_mouseClingObjList.find(registNum) != _mouseClingObjList.end()) {

		result =_mouseClingObjList.find(registNum)->second;
		_mouseClingObjList.erase(_mouseClingObjList.find(registNum));
	}


	return result;
	*/
}


//=========================================================
// マウスにくっついているモノを参照
//=========================================================
IMouseClingableObject* Application::RefferMouseClingObj() {

	IMouseClingableObject *result = _mouseClingObj;

	return result;
}


//=========================================================
// マウスにくっつているモノが何か？
//=========================================================
bool Application::IsMouseClingObj_Piece() {

	bool result = false;

	if (_mouseClingObj != NULL) {

		Piece *piece = dynamic_cast<Piece*>(_mouseClingObj);
		if ( piece != NULL ) {

			result = true;
		}
	}

	return result;
}
bool Application::IsMouseClingObj_Card() {

	bool result = false;

	if (_mouseClingObj != NULL) {
		
		
		Card *card = dynamic_cast<Card*>(_mouseClingObj);
		if ( card != NULL ) {
			result = true;
		}
		
		
		//result = (_mouseClingObj->GetType() == Card::
	}
	return result;
}




//=========================================================
// Update / Draw for MouseClingObect
//=========================================================
void Application::MouseClingObject_Update() {

	//sf::Vector2i mousePos = sf::Mouse::getPosition( _window );
	sf::Vector2i mousePos; // カメラの状態を考慮した位置にするため
	mousePos.x = _mouseX;
	mousePos.y = _mouseY;


	if (_mouseClingObj != NULL) {
	
		sf::Vector2f pos ( (float)mousePos.x + _mouseClingObj->GetMouseDrawOffsetX(), (float)mousePos.y + _mouseClingObj->GetMouseDrawOffsetY() );

		_mouseClingObj->Update( _window,  pos);
	}
	/*
	map<int, IDrawable*>::iterator it = _mouseClingObjList.begin();
	while (it != _mouseClingObjList.end()) {

		it->second->Update( _window,  mousePos);

		++it;
	};
	*/
}

void Application::MouseClingObject_Draw() {

	if (_mouseClingObj != NULL) {
		_mouseClingObj->Draw( _window);
	}
	/*
	map<int, IDrawable*>::iterator it = _mouseClingObjList.begin();
	while (it != _mouseClingObjList.end()) {

		if (it->second != NULL) {
			it->second->Draw( _window );

		// suspicious
		} else {
			_mouseClingObjList.erase(it);
			break;
		}

		++it;
	};
	*/
}

//=========================================================
// 後でより前面に描画してね。
//=========================================================
void Application::AddLaterDraw(IDrawable *obj) {

	_laterDrawObjList.push_back( obj );

}
void Application::AddLaterDraw_Piece(IDrawable *obj) {

	_laterDrawObjList_Piece.push_back( obj );

}
void Application::AddLaterDraw_Card(IDrawable *obj) {

	_laterDrawObjList_Card.push_back( obj );

}
//=========================================================
// LaterRendering
//=========================================================
void Application::DrawLaterObjList() {

	for (unsigned int  i = 0; i < _laterDrawObjList.size(); ++i) {

		_laterDrawObjList[i]->DrawActual(_window);

	}
	_laterDrawObjList.clear();

}
void Application::DrawLaterObjList_Piece() {

	for (unsigned int  i = 0; i < _laterDrawObjList_Piece.size(); ++i) {

		_laterDrawObjList_Piece[i]->DrawActual(_window);

	}
	_laterDrawObjList_Piece.clear();

}
void Application::DrawLaterObjList_Card() {

	for (unsigned int  i = 0; i < _laterDrawObjList_Card.size(); ++i) {

		_laterDrawObjList_Card[i]->DrawActual(_window);

	}
	_laterDrawObjList_Card.clear();

}


void Application::SetWindowTitle ( string title ) {
	
	sf::String sfTitle = GUI_SFML::SFML_Utils::GetSfString (title);
	_window.setTitle ( sfTitle );
	
	_windowTitle = title;
	
#if defined(_WIN32)
	HANDLE handle = (HANDLE)_window.getSystemHandle();
	
	char buf[1000];
	GetWindowText(handle, buf, 1000);
	
	string actualTitle = buf;
	
	Utils_File::CreateAndOverwrite ( "title", Utils_String::Trim (actualTitle) );
	
	printf ( " actualTitle : %s\n ", actualTitle.c_str() );
#endif
	
}





