//=========================================================
//
// Application (Declare)
//
//=========================================================

#ifndef _APPLICAITON_
#define _APPLICAITON_

#include <map>
#include <vector>
#include <string>

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>

//#include "../GUI_SFML/SoundManager.h"

class IScene;
class EffectManager;
class IDrawable;
class IMouseClingableObject;

using namespace std;


class Application {

public:

	Application();
	~Application();

	void Run();

	static Application* _obj;
	static Application* GetInstance() {return _obj;}


	sf::RenderWindow _window;

	EffectManager     *_effectManager;
	//GUI_SFML::SoundManager      *_soundManager;

	int _mouseState_Left;
	int _mouseState_Right;
	int _mouseState_Middle;
	int _mouseX;
	int _mouseY;
	int _mouseWheel;
	bool _isMouseOn;
	
	sf::Keyboard::Key _releasedKeyCode;

	///// 後でより前面に描画して欲しいものリスト
	vector<IDrawable*> _laterDrawObjList;
	vector<IDrawable*> _laterDrawObjList_Piece;
	vector<IDrawable*> _laterDrawObjList_Card;

	//=========================================================
	// 後でより前面に描画してね。
	//=========================================================
	void AddLaterDraw(IDrawable *obj);
	void AddLaterDraw_Piece(IDrawable *obj);
	void AddLaterDraw_Card(IDrawable *obj);

	//=========================================================
	// マウスにくっつけるオブジェクトを登録
	//=========================================================
	//int Application::AddMouseClingObj(IDrawable *obj);
	IMouseClingableObject* AddMouseClingObj(IMouseClingableObject *obj); // 戻り値は先客

	//=========================================================
	// マウスにくっついているオブジェクトを取り外す
	//=========================================================
	//IDrawable* Application::DrawMouseClingObj(int registNum);
	IMouseClingableObject* DrawMouseClingObj();

	//=========================================================
	// マウスにくっついているものがある？
	//=========================================================
	//int GetMouseClingObjNum() { return _mouseClingObjList.size(); }
	bool IsMouseClingObjExist() { return (_mouseClingObj != NULL); }

	//=========================================================
	// マウスにくっついているモノを参照
	//=========================================================
	IMouseClingableObject* RefferMouseClingObj();

	
	//=========================================================
	// マウスにくっつているモノ（先頭）が何か？
	//=========================================================
	bool IsMouseClingObj_Piece();
	bool IsMouseClingObj_Card();
	

	int GetScreenWidth() { return _window.getSize().x; }
	int GetScreenHeight(){ return _window.getSize().y; }
	
	void   SetWindowTitle ( string title );
	string GetWindowTitle () { return _windowTitle; };
	
	bool HasFocus();
	
	void CleanUpAllEffect();

private:

	map<string, IScene*> _sceneList;

	IScene* _currentScene;
	

	bool _isUpdateHalt;

	//int _mouseClingObjTotalCount; // registNumを出すために数えているだけ。
	//map<int, IDrawable*> _mouseClingObjList;
	IMouseClingableObject* _mouseClingObj;

	bool _isFullScreen;
	
	string _windowTitle;

	//=========================================================
	// Transite Scene
	//=========================================================
	void Transition(string nextSceneName);

	//=========================================================
	// Manage MouseState
	//=========================================================
	void ManageMouseState();

	//=========================================================
	// Update / Draw for MouseClingObect
	//=========================================================
	void MouseClingObject_Update();
	void MouseClingObject_Draw();

	//=========================================================
	// LaterRendering
	//=========================================================
	void DrawLaterObjList();
	void DrawLaterObjList_Piece();
	void DrawLaterObjList_Card();


};
#endif
