//=========================================================
//
// BGMクラス。SFML的な。
//
//=========================================================

#ifndef _BGM_
#define _BGM_

#include <SFML/Audio.hpp>
#include <string>
#include <map>
using namespace std;


class BGM {

public:

	//=========================================================
	// 定数：トランジションエフェクト
	//=========================================================
	static const string CHANGE_BGM_EFFECT_NONE;              // エフェクトなし
	static const string CHANGE_BGM_EFFECT_FADE_OUT_IN;       // 元のをフェードアウト してから 次のをフェードイン
	static const string CHANGE_BGM_EFFECT_FADE_OUT_IN_SIMUL; // 元のをフェードアウト しながら 次のをフェードイン
	

	BGM();
	~BGM();

	///// ファイル名とラベルを指定して登録。エイリアス。
	void Regist ( string fileName, string label );

	///// BGM変更。
	void Change ( string label, string transitionEffect = "" );

	///// 毎フレーム呼んでもらう。
	void Update();
	
	void  SetVolume ( float value );
	float GetVolume ();
	
	void  SetPitch ( float value );
	float GetPitch ();
	
	void SetLoop (bool loop);

	//=========================================================
	// セーブ/ロード用のインターフェイス
	//=========================================================
	struct SaveData {

		//map < string, string> registHistory; // バッファ登録履歴 _registHistory 
		string                currentPlayingLabel;      // 現在演奏中の ラベル。トランジション中であれば、_sound_Next のラベル
		string                pitch;
		bool                  loop;
		
		SaveData() {
			loop = true;
		}
	};
	SaveData GetSaveData ();
	void Load ( SaveData &saveData );

private:

	map < string, string> _registHistory;              // map < ラベル, ファイル名 > セーブ/ロード対象。
	/*
		
		このお話は、sf::Sound を使っていたバージョンのときのこと。
		今となっては、もうなんだか、意味がわかりません。


		セーブ/ロードのことを考える。
		バッファの内容をそのまま再現すると、読み込みに時間がかかってしまう。

		ロード時に再現するのは、そのときに鳴っている音だけにしたい。
		しかし使う側にとっては、それまでに Regist したものは、バッファにあるもの、として扱いたい。

		Change でラベルを指定するわけですが、
		その時点で、ラベルに対応するバッファがない場合、「読み込んだつもりリスト」から読み込みなおす。
		なんのこっちゃ。

		つまり、Regist で「ファイル名」「ラベル」を指定する。_registHistory
		その情報を、覚えておく、ということ。

		で、Changeで指定したラベルが、ないとき、まず、そのメモを見に行く。
		対応するファイル名が登録されていれば、そのファイルを読み込んで、バッファを作成する。

	*/


	sf::Music *_sound_Current;
	sf::Music *_sound_Next;

	string _currentPlayingLabel; // 現在演奏中のラベル。セーブ/ロード対象 空＝無演奏。
	
	// トランジションを実行中？
	string _transitionEffect;
	
	// ボリューム
	float _volume;
	
	// ピッチ
	float _pitch;
	
	// ループ
	bool _loop;
	
	//=========================================================
	// トランジションエフェクト準備
	//=========================================================
	void Transition_Preparation_NONE              ( string label );
	void Transition_Preparation_FADE_OUT_IN       ( string label );
	void Transition_Preparation_FADE_OUT_IN_SIMUL ( string label );

	//=========================================================
	// トランジションエフェクト実行中
	//=========================================================
	void Transition_Processing_NONE              ();
	void Transition_Processing_FADE_OUT_IN       ();
	void Transition_Processing_FADE_OUT_IN_SIMUL ();
	
	//=========================================================
	// トランジションのためのボリューム増分値を取得
	// 「音量設定」が演出としてのトランジションの時間に影響させないため
	//=========================================================
	float GetVolumeIncrementalValue_forTransition ();
	
	
	
	//=========================================================
	// 存分に空文字や存在しないファイルを指定できるようにするために
	//=========================================================
	bool _sound_Current_OpenResult;
	bool _sound_Next_OpenResult;
	
	void OPEN_FROM_FILE ( sf::Music *music, string fileName, bool &openResult );
	void PLAY_MUSIC ( sf::Music *music, bool openResult );

};
#endif
