//=========================================================
//
// クリックできる画像
//
//=========================================================

#ifndef _CLICKABLE_IMAGE_
#define _CLICKABLE_IMAGE_

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>

#include <string>

using namespace std;

class ClickableImage {

public:

	struct Property {
		
		string       imageFile;
		sf::Vector2f size;
		sf::IntRect  rect_Normal;
		sf::IntRect  rect_Over;
		sf::IntRect  rect_Push;
		
		string name;
	};

	ClickableImage (Property prop);
	~ClickableImage ();
	
	void Update (sf::RenderWindow &window);
	void Draw   (sf::RenderWindow &window);


	bool IsClick () { return _isClick; }

	void         SetPosition ( sf::Vector2f pos ) { _position = pos; }
	sf::Vector2f GetPosition () { return _position; }
	void SetImageFile  ( string fileName );
	void SetSize       ( sf::Vector2f size );
	void SetRectNormal ( sf::IntRect rect );
	void SetRectOrver  ( sf::IntRect rect );
	void SetRectPush   ( sf::IntRect rect );
	
	//=========================================================
	// セーブ/ロード用のインターフェイス
	//=========================================================
	struct SaveData {
		
		Property prop;
		sf::Vector2f position;
		
	};
	SaveData GetSaveData ();
	void Load ( ClickableImage::SaveData &saveData );
	
	
private:

	static const int STATE_NORMAL = 0;
	static const int STATE_OVER   = 1;
	static const int STATE_PUSH   = 2;
	
	Property _prop; // セーブ/ロード用にとっておく
	
	int _currentState;
	
	

	sf::Vector2f _position;

	sf::Texture        _texture;
	sf::RectangleShape _sprite;
	sf::IntRect        _rect_Normal;
	sf::IntRect        _rect_Over;
	sf::IntRect        _rect_Push;

	bool _isClick;

	//=========================================================
	// マウスが乗っている？
	//=========================================================
	bool IsMouseOver();
	
	//=========================================================
	// 初期設定
	//=========================================================
	void Init();
	
	
};
#endif

