//=========================================================
//
// EffectManager (implement)
//
//=========================================================

#include "EffectManager.h"

#include "../Utils/Utils_Others.h"


//=========================================================
// Constructor
//=========================================================
EffectManager::EffectManager() {

	// nothing
}

//=========================================================
// Destructor
//=========================================================
EffectManager::~EffectManager() {

	while (_effectList.size() > 0) {
		
		printf ("EffectManager::~EffectManager() エフェクトが残っていた。解放します\n");
		
		delete *(_effectList.begin());
		_effectList.erase(_effectList.begin());
	}
}

//=========================================================
// Update
//=========================================================
void EffectManager::Update(sf::RenderWindow &window) {

	for (unsigned int i = 0; i < _effectList.size(); ++i) {

		IEffect *effect = _effectList[i];

		bool isEnd = effect->Update(window);

		if (isEnd) {
			delete effect;
			UTILS_OTHERS::deleteFromVector<IEffect*>( _effectList, i );
			--i;
		}
	}

}

//=========================================================
// Draw
//=========================================================
void EffectManager::Draw(sf::RenderWindow &window) {

	for ( vector<IEffect*>::iterator it = _effectList.begin(); it != _effectList.end(); ++it) {

		IEffect *effect = (*it);

		effect->Draw(window);

		/*
		int state = effect->GetState();
		if (state == EFFECT::STATE::END) {

			delete effect;

			it = _effectList.erase( it );
			
			if (_effectList.size() == 0) {
				break;
			}

			++it;
			if (it == _effectList.end()) {
				break;
			}
		}
		*/
	}
}

//=========================================================
// Add
//=========================================================
void EffectManager::Add (IEffect *effect) {
	
	if ( effect == NULL ) {
		printf( "EffectManager::Add() を呼んでおいて引数が NULL ですって？ 意味がわかりませんわね！ 何もせずにメソッドをリターンしますわ！\n" );
		return;
	}
	
	for (unsigned int i = 0; i < _effectList.size(); ++i) {
		
		if ( _effectList[i] == effect ) {
			
			printf ( "EffectManager::Add() リスト内に同一のオブジェクトが既に存在しています。何もせず（重複登録はせずに）にメソッドを終了します\n" );
			return;
		}
		
	}
	_effectList.push_back(effect);

}

//=========================================================
// 強制終了
//=========================================================
void EffectManager::Terminate ( IEffect *effect ) {
	
	printf ( "エフェクトを強制終了\n" );
	
	bool isExist = false;
	
	for (unsigned int i = 0; i < _effectList.size(); ++i) {

		IEffect *effectInList = _effectList[i];
		
		if ( effect == effectInList ) {
			
			isExist = true;
			
			printf ( "対象のエフェクトを管理リスト内に発見。強制終了します\n" );
			effectInList->Terminate();
			
			delete effect;
			UTILS_OTHERS::deleteFromVector<IEffect*>( _effectList, i );
			
			// 同じエフェクトがリストの中に複数存在するなんてことはないはず。
			break;
		}
	}	
	
	if (!isExist) {
		printf("指定のエフェクトオブジェクトはリストに存在しませんでした。（意訳：無事に終了してリストから外されてたんじゃないでしょうか？）\n");
	}
	
	printf ( "エフェクトを強制終了・通過\n" );
}

//=========================================================
// 存在確認
//=========================================================
bool EffectManager::IsExist (  IEffect *effect ) {
	
	bool result = false;
	
	for (unsigned int i = 0; i < _effectList.size(); ++i) {

		IEffect *effectInList = _effectList[i];
		
		if ( effect == effectInList ) {
			result = true;
			break;
		}
	}
	
	return result;
}









