//=========================================================
//
// フェード（アウトもインも）
//
//=========================================================

#include "Fade.h"
#include "Application.h"
#include "Define.h"

//=========================================================
// コンストラクタ
//=========================================================
Fade::Fade() {

	_currentState = STATE_FADE_IN_ed;

	_color_FadeOutStart = sf::Color ( 255,255,255,0 );
	_color_FadeInStart  = sf::Color ( 255,255,255,255 );

	_rectangleShape = sf::RectangleShape ( sf::Vector2f ( APPLICATION::WINDOW_SIZE::WIDTH, APPLICATION::WINDOW_SIZE::HEIGHT ) );
	_rectangleShape.setOutlineThickness ( 0 );
	_rectangleShape.setFillColor ( _color_FadeOutStart );
	_rectangleShape.setPosition ( sf::Vector2f ( 0.0f, 0.0f ) );
	
	_rectangleShape.setOrigin ( _rectangleShape.getGlobalBounds().width/2, _rectangleShape.getGlobalBounds().height/2 );
	
	_textureFileName = "";

}

//=========================================================
// デストラクタ
//=========================================================
Fade::~Fade() {

}

//=========================================================
// テクスチャをセット
//=========================================================
void Fade::SetFadeTexture ( string fileName ) {

	_texture.loadFromFile ( fileName ) ;//"./image/menu/fadeTexture.png" );
	_rectangleShape.setTexture ( &_texture );
	
	// セーブデータ用にファイル名を確保しておくのですよー。
	_textureFileName = fileName;
}

//=========================================================
// 更新
//=========================================================
void Fade::Update ( sf::RenderWindow &window ) {
	
	// カメラの位置に関わりなく、常に画面全体を覆うように
	sf::Vector2i pos = Application::GetInstance()->_window.mapCoordsToPixel (
		sf::Vector2f(
			APPLICATION::WINDOW_SIZE::WIDTH/2,
			APPLICATION::WINDOW_SIZE::HEIGHT/2
		) 
	);
	_rectangleShape.setPosition(pos.x, pos.y);


	if      ( _currentState == STATE_FADE_IN_ing  ) { Process_FadeIn();}
	else if ( _currentState == STATE_FADE_OUT_ing ) { Process_FadeOut();}
}


//=========================================================
// 動きが終了している？
//=========================================================
bool Fade::IsStabilized() {

	bool result = false;

	if      ( _currentState == STATE_FADE_IN_ed  ) { result = true; }
	else if ( _currentState == STATE_FADE_OUT_ed ) { result = true; }

	return result;
}

//=========================================================
// 動きを完遂させる
//=========================================================
void Fade::Stabilize() {
	
	if ( _currentState == STATE_FADE_IN_ing  ) { 

		sf::Color color = _rectangleShape.getFillColor();
		color.a = 0;
		_rectangleShape.setFillColor ( color );		

		_currentState = STATE_FADE_IN_ed;

	}
	else if ( _currentState == STATE_FADE_OUT_ing ) { 
		
		sf::Color color = _rectangleShape.getFillColor();
		color.a = 255;
		_rectangleShape.setFillColor ( color );		

		_currentState = STATE_FADE_OUT_ed;
		
	}
}


//=========================================================
// 描画
//=========================================================
void Fade::Draw   ( sf::RenderWindow &window ) {

	///// フェードインしきっているなら、描画しなくていい
	if ( _currentState != STATE_FADE_IN_ed ) {
		window.draw ( _rectangleShape );
	}
}

//=========================================================
// フェードアウトしてー。
//=========================================================
void Fade::FadeOut( int speed ) {

	_prop_FadeOut.speed = speed;
	ToStart_FadeOut();
}
//=========================================================
// フェードインしてー。
//=========================================================
void Fade::FadeIn ( int speed ) {

	_prop_FadeIn.speed = speed;
	ToStart_FadeIn();
}



//=========================================================
// フェードアウトを開始する準備
//=========================================================
void Fade::ToStart_FadeOut() {

	_rectangleShape.setFillColor ( _color_FadeOutStart );
	_currentState = STATE_FADE_OUT_ing;

}
//=========================================================
// フェードインを開始する準備
//=========================================================
void Fade::ToStart_FadeIn() {

	_rectangleShape.setFillColor ( _color_FadeInStart );
	_currentState = STATE_FADE_IN_ing;
}

//=========================================================
// 各動作の処理：フェードアウト
//=========================================================
void Fade::Process_FadeOut() {

	sf::Color color = _rectangleShape.getFillColor();
	int a = color.a;

	a += _prop_FadeOut.speed;
	color.a = a;

	if ( a >= 255 ) {
		color.a = 255;
		_currentState = STATE_FADE_OUT_ed;
	}

	_rectangleShape.setFillColor ( color );
}
//=========================================================
// 各動作の処理：フェードイン
//=========================================================
void Fade::Process_FadeIn() {

	sf::Color color = _rectangleShape.getFillColor();
	int a = color.a;

	a -= _prop_FadeIn.speed;
	color.a = a;

	if ( a <= 0 ) {
		color.a = 0;
		_currentState = STATE_FADE_IN_ed;
	}

	_rectangleShape.setFillColor ( color );
}

