//=========================================================
//
// フリーバトルセレクター
//
//=========================================================

#include "Define.h"
#include "Application.h"
#include "FreeBattleSelector.h"
#include "../GUI_SFML/TextImageFactory.h"
#include "../GUI_SFML/ScrollBar.h"
#include "ScriptEngine.h"
#include "ClickableImage.h"
#include "Image.h"

using namespace GUI_SFML;

//=========================================================
// 個別のアイコン：コンストラクタ
//=========================================================
FreeBattleIcon::FreeBattleIcon ( FreeBattleIconProp prop,  ScriptEngine *scriptEngine, sf::Texture &lockedIconTexture ) {
	
	_prop = prop;
	
	_scriptEngine      = scriptEngine;
	
	if ( _prop.lockedIconImageFile == "" ) {
	
		_lockedIconTexture = lockedIconTexture;
		
	} else {
		
		_lockedIconTexture.loadFromFile ( _prop.lockedIconImageFile );
	}
	
	_lockedIconSprite.setTexture ( _lockedIconTexture );
	
	int sizeWidth  = 100;
	int sizeHeight = 100;
	
	if ( _prop.isSecret ) {
		sizeWidth  = _prop.secretWidth;
		sizeHeight = _prop.secretHeight;
	}
	
	
	_back_Frame = sf::RectangleShape ( sf::Vector2f ( sizeWidth, sizeHeight ) );
	_back_Frame.setFillColor ( sf::Color ( 50,50,150,150 ) );
	_back_Frame.setOutlineThickness (1.0f);
	_back_Frame.setOutlineColor ( sf::Color ( 0,0,0,255 ) );
	_back_Frame.setOrigin ( 0, 0 );
	
	
	_cursor_Frame = sf::RectangleShape ( sf::Vector2f ( sizeWidth, sizeHeight ) );
	_cursor_Frame.setFillColor ( sf::Color ( 150,150,150,100 ) );
	_cursor_Frame.setOutlineThickness (1.0f);
	_cursor_Frame.setOutlineColor ( sf::Color ( 50,50,255,255 ) );
	_cursor_Frame.setOrigin ( 0, 0 );
	
	if ( _prop.isSecret ) {
		_back_Frame.setPosition ( sf::Vector2f ( _prop.secretX, _prop.secretY ) );
		_cursor_Frame.setPosition ( sf::Vector2f ( _prop.secretX, _prop.secretY ) );
	}
	
	
	//sf::Texture _iconTexture;
	//sf::Sprite _iconSprite;
	//sf::Text   _battleTitleLabel;
	
	_iconTexture.loadFromFile ( _prop.imageFile );
	_iconSprite.setTexture( _iconTexture );

	_battleTitleLabel = TextImageFactory::Create ( _prop.battleTitle, 30 );
	
	_battleTitleLabel.setOutlineThickness(1);
	_battleTitleLabel.setOutlineColor ( sf::Color::Black );
	_battleTitleLabel.setFillColor    ( sf::Color::White );
	
	
	_isMouseover = false;
	_clickedHere = false;
	_isSelected = false;
	_isPressed = false;
}
//=========================================================
// 個別のアイコン：デストラクタ
//=========================================================
FreeBattleIcon::~FreeBattleIcon() {
}

//=========================================================
// 個別のアイコン：サイズを取得
//=========================================================
sf::Vector2f FreeBattleIcon::GetSize() {
	
	sf::Vector2f result;
	
	if ( _prop.isSecret ) {
		
		result.x = _prop.secretWidth;
		result.y = _prop.secretHeight;
		
	} else {
	
		result.x = _back_Frame.getGlobalBounds().width;
		result.y = _back_Frame.getGlobalBounds().height;
	}
	
	return result;
}

//=========================================================
// 個別のアイコン：位置をセット
//=========================================================
void FreeBattleIcon::SetPosition ( int x, int y ) {
	
	if ( _prop.isSecret ) {
		
		// シークレット枠の場合はコンストラクタで枠の位置を固定している。
		
	} else {
	
		_back_Frame.setPosition ( sf::Vector2f ( x, y ) );
		_cursor_Frame.setPosition ( sf::Vector2f ( x, y ) );
		_iconSprite.setPosition ( sf::Vector2f ( x, y ) );
		_lockedIconSprite.setPosition ( sf::Vector2f ( x, y ) );
		
	}
	
	
	// これはマウスオーバーのときに出すものだから、固定位置
	_battleTitleLabel.setPosition ( sf::Vector2f ( 400.0f,  500.0f) );
	
	_battleTitleLabel.setOrigin (
		(int)(_battleTitleLabel.getGlobalBounds().width / 2),
		(int)(_battleTitleLabel.getGlobalBounds().height / 2)
	);
	
}

//=========================================================
// 個別のアイコン：更新
//=========================================================
void FreeBattleIcon::Update ( sf::RenderWindow &window ) {
	
	///// 空欄の場合はマウス判定等しない
	if ( _prop.id == "" ) {
		return;
	}
	
	_isUnlocked = IsUnlocked();
	
	if ( _isUnlocked ) {
		
		///// マウス、乗ってる？
		int mouseX = Application::GetInstance()->_mouseX;
		int mouseY = Application::GetInstance()->_mouseY;
		
		sf::Vector2f currentPos(
			_back_Frame.getGlobalBounds().left,
			_back_Frame.getGlobalBounds().top
		);
		int width  = _back_Frame.getGlobalBounds().width;
		int height = _back_Frame.getGlobalBounds().height;
		
		if
		(
			currentPos.x <= (float)mouseX && (float)mouseX <= currentPos.x + width  &&
			currentPos.y <= (float)mouseY && (float)mouseY <= currentPos.y + height
		)
		{
			_isMouseover = true;
		} else {
			_isMouseover = false;
		}
		
		
		if ( _isMouseover ) {
			
			if ( Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::HAVE_PRESSED ) {
				_clickedHere =true;
			}
			
			
			if ( _clickedHere && Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::PRESS ) {
				_isPressed =true;
			}
			
			if ( _clickedHere ) {

				if ( Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE ) {
					_isSelected =true;
				}
			}

		} else {
			
			_clickedHere = false;
			_isSelected = false;
			_isPressed = false;
		}
		
	// 解放されていない
	} else {
		
		// 処理なしでいいかな？
		
	}

}

//=========================================================
// 個別のアイコン：選択状況をクリア
//=========================================================
void FreeBattleIcon::ClearSelectStatus() {

	_isMouseover = false;
	_clickedHere = false;
	_isSelected = false;
	_isPressed = false;
}

//=========================================================
// 個別のアイコン：描画
//=========================================================
void FreeBattleIcon::Draw   ( sf::RenderWindow &window ) {
	
	///// 空欄の場合
	if ( _prop.id == "" ) {
	
		//window.draw ( _back_Frame );
	
	
	///// 空欄ではない
	} else {

		///// 解放されているかどうか
		if ( _isUnlocked ) {

			if ( _isPressed ) {
				_iconSprite.move(0,1);
			}
			
			
			if ( _prop.isSecret ) {
				
				// シークレット枠の場合はアイコンなし
				
			} else {
				window.draw ( _back_Frame );
				window.draw ( _iconSprite );
			}
			
			
			if ( _isMouseover ) {
				window.draw(_cursor_Frame);
				window.draw ( _battleTitleLabel );
			}

			if ( _isPressed ) {
				_iconSprite.move(0,-1);
			}
		
		
		} else {
		
			if ( _prop.isSecret ) {
				
				// シークレット枠の場合はアイコンなし
				
			} else {
				window.draw ( _lockedIconSprite  );
			}
		}
	}


}

//=========================================================
// 個別のアイコン：アンロックされているかどうか
//=========================================================
bool FreeBattleIcon::IsUnlocked() {

	///// 動作確認用
	//return true;
	
	bool result = false;
	
	// 変数未指定＝無条件で表示
	if ( _prop.variableForUnlock == "") {
		return true;
	}
	
	
	result = _scriptEngine->IsUnlockedFreeBattle ( _prop.variableForUnlock );
	
	
	//printf ( "%s : %s \n", _prop.variableForUnlock.c_str(), (result)?"解放":"ロック" );
	
	return result;
}




//=========================================================
// コンストラクタ
//=========================================================
FreeBattleSelector::FreeBattleSelector ( ScriptEngine *scriptEngine ) {
	
	_scriptEngine = scriptEngine;

	// 表示中？
	_isActive = false;
	
	_position = sf::Vector2f ( 50.0f, 100.0f );
	
	_back_Frame = sf::RectangleShape ( sf::Vector2f ( 680, 480 ) );
	_back_Frame.setFillColor ( sf::Color ( 50,50,50,150 ) );
	_back_Frame.setOutlineThickness (1.0f);
	_back_Frame.setOutlineColor ( sf::Color ( 0,0,0,255 ) );
	_back_Frame.setPosition ( sf::Vector2f ( APPLICATION::WINDOW_SIZE::WIDTH/2, APPLICATION::WINDOW_SIZE::HEIGHT/2 ) );
	_back_Frame.setOrigin ( _back_Frame.getGlobalBounds().width/2, _back_Frame.getGlobalBounds().height/2 );
	
	
	_lockedIconTexture.loadFromFile ( "./image/freeBattle/lockedIcon.png" );
	
	///// 上部のタイトル画像「Free Battle」
	_caption = new Image();
	_caption->ChangeImage("./image/freeBattle/freeBattleCaption.png");
	_caption->SetPosition ( sf::Vector2f(400, 90) );
	
	
	///// 閉じる（戻る）ボタン
	{
		ClickableImage::Property closeButtonProp;
		
		closeButtonProp.imageFile = "./image/menu/backFromSaveloadWindow.png";
		closeButtonProp.size = sf::Vector2f(100.0f, 30.0f);
		closeButtonProp.rect_Normal = sf::IntRect ( 0, 0,  100, 30 );
		closeButtonProp.rect_Over   = sf::IntRect ( 0, 30, 100, 30 );
		closeButtonProp.rect_Push   = sf::IntRect ( 0, 60, 100, 30 );
		
		_closeButton = new ClickableImage(closeButtonProp);
		
		_closeButton->SetPosition ( sf::Vector2f(695, 550) );
	}		
	
	_scrollBar = NULL;
	_displayVCount = 0;

	int _gridXCount = 0;
	int _gridYCount = 0;
	
	//vector<vector<FreeBattleIcon*>> _itemList; // フリーバトルのリスト
	
	_isSelected = false;
	_selectedId = "";

}
//=========================================================
// デストラクタ
//=========================================================
FreeBattleSelector::~FreeBattleSelector() {

	AllClear();
	delete _caption;
	
}

///// 中身を全部クリア
void FreeBattleSelector::AllClear() {

	//vector<vector<FreeBattleIcon*>> _itemList; // フリーバトルのリスト
	
	for ( unsigned int y = 0; y < _itemList.size(); ++y ) {
		
		
		for ( unsigned int x = 0; x < _itemList[y].size(); ++x ) {
			
			FreeBattleIcon *icon = _itemList[y][x];
			
			delete icon;
			
		}
		_itemList[y].clear();
		
	}
	
	_itemList.clear();

	_gridXCount = 0;
	_gridYCount = 0;

	for ( unsigned int i = 0; i < _itemList_secret.size(); ++i ) {
		
		FreeBattleIcon *icon = _itemList_secret[i];
		delete icon;
	}
	_itemList_secret.clear();
		
	
		
	delete _scrollBar;
	_scrollBar = NULL;
}

//=========================================================
// 表の枠を設定：じゃあ、これをすると中身全部消えるってことで。
//=========================================================
void FreeBattleSelector::SetMatrix ( int hCount, int vCount, int displayVCount ) {

	AllClear();
	
	
	for ( unsigned int y = 0; y < vCount; ++y ) {
		
		vector<FreeBattleIcon*> row;
			
		
		for ( unsigned int x = 0; x < hCount; ++x ) {
			
			FreeBattleIcon::FreeBattleIconProp prop;
			
			FreeBattleIcon *icon = new FreeBattleIcon ( prop, _scriptEngine, _lockedIconTexture );
			
			
			row.push_back ( icon );
		}

		_itemList.push_back ( row );
	}
	
	// ここではシークレット枠は考慮しなくていいはず。
	
	
	///// この段階でスクロールバーのサイズが決まる
	
	_displayVCount = displayVCount;
	
	_scrollBar = new ScrollBar();
	//_scrollBar->SetMaxLength(_saveDataList.size());
	_scrollBar->SetActualLength(_itemList.size());
	//_scrollBar->SetLengthOfOneItem(100.0f);
	_scrollBar->SetDisplayLength(_displayVCount);// LengthOfOneItem(100.0f);
	_scrollBar->SetFrameSize(sf::Vector2f(20.0f, 500.0f));
	//_scrollBar->SetPosition( sf::Vector2f(1000.0f, 0.0f) );
	//_scrollBar->MoveUp(1000);
	_scrollBar->IsUpDownButton(false);

	_gridXCount = hCount;
	_gridYCount = vCount;

}


//=========================================================
// 一件追加
//=========================================================
void FreeBattleSelector::AddIcon ( FreeBattleIcon::FreeBattleIconProp prop, int gridX, int gridY ) {
	
	///// シークレットアイコン
	if ( prop.isSecret ) {
		
		FreeBattleIcon *icon = new FreeBattleIcon ( prop, _scriptEngine, _lockedIconTexture );
		
		_itemList_secret.push_back ( icon );
		
		printf ("Add free battle secret icon. %s(%d, %d)-(%d, %d)\n", prop.id.c_str(), prop.secretX, prop.secretY, prop.secretWidth, prop.secretHeight);
		
	///// 通常アイコン
	} else {
	
		if ( gridY < _itemList.size() ) {
			
			if ( gridX < _itemList[gridY].size() ) {
				

				FreeBattleIcon *old = GetIcon ( gridX, gridY );
				
				if ( old != NULL ) {
					delete old;
				}

				FreeBattleIcon *icon = new FreeBattleIcon ( prop, _scriptEngine, _lockedIconTexture );
				
				_itemList[gridY][gridX] = icon;
				
				printf ("Add free battle icon. %s(%d, %d)\n", prop.id.c_str(), gridX, gridY);
			}
		}
	}
}

//=========================================================
// 一件取得
//=========================================================
FreeBattleIcon* FreeBattleSelector::GetIcon ( int x, int y ) {
	
	FreeBattleIcon *icon = NULL;
	
	if ( y < _itemList.size() ) {
		
		if (  x < _itemList[y].size()) {
			
			icon = _itemList[y][x];
		}
	}
	
	
	return icon;
}

//=========================================================
// 選択状況をクリア
//=========================================================
void FreeBattleSelector::ClearSelectStatus() {
	
	///// 個別のアイコンもクリア
	for ( unsigned int y = 0; y < _itemList.size(); ++y ) {
		for ( unsigned int x = 0; x < _itemList[y].size(); ++x ) {

			FreeBattleIcon *icon = GetIcon ( x, y );
			
			if ( icon != NULL ) {
				
				icon->ClearSelectStatus();
				
			}
		}
	}
	// シークレットアイコンもクリア
	for ( unsigned int i = 0; i < _itemList_secret.size(); ++i ) {

		_itemList_secret[i]->ClearSelectStatus();
	}
	
	///// 根元をクリア
	_isSelected = false;
	_selectedId = "";
}

//=========================================================
// 更新
//=========================================================
void FreeBattleSelector::Update ( sf::RenderWindow &window ) {
	
	if ( !_isActive ) { return; }

	_isSelected = false;
	_selectedId = "";
	
	///// 閉じる（戻る）ボタン
	_closeButton->Update(window);
	
	// 何も選ばなかったことにして終了
	if ( _closeButton->IsClick() ) {
		_isSelected = true;
		_selectedId = "";
		return;
	}
	
	_caption->Update(window);
	
	
	///// 全体の表示サイズと位置を事前に計算
	//
	
	int paddingX = 20;
	int paddingY = 20;
	
	int displayTotalWidth = 0;
	int displayTotalHeight = 0;
	
	int iconWidth = 0;
	int iconHeight = 0;

	for ( unsigned int y = 0; y < 1; ++y ) {
		for ( unsigned int x = 0; x < _itemList[y].size(); ++x ){

			FreeBattleIcon *icon = _itemList[y][x];

			if ( x > 0 ) {
			
				displayTotalWidth += paddingX;
				
			} else {
			
				iconWidth  = icon->GetSize().x;
				iconHeight = icon->GetSize().y;
				
				displayTotalHeight = icon->GetSize().y * _displayVCount + paddingY*(_displayVCount-1);
			}
			
			displayTotalWidth += icon->GetSize().x;
			
		}
	}
	
	
	_position.x = (APPLICATION::WINDOW_SIZE::WIDTH  - displayTotalWidth) /2;
	_position.y = (APPLICATION::WINDOW_SIZE::HEIGHT - displayTotalHeight) /2;
	
	
	///// スクロールバーのサイズと位置
	_scrollBar->SetPosition ( sf::Vector2f (
		_position.x + displayTotalWidth + 5,
		_position.y
	) );
	
	_scrollBar->SetFrameSize(sf::Vector2f(20.0f, displayTotalHeight));
	
	
	///// スクロールバー
	if (_scrollBar != NULL) { // ここだけ突発的にNULLチェックしても意味ないんじゃないの？ そうかもしれない。
		
		_scrollBar->Update(window);
		
		
		// マウスがウィンドウに乗っているかどうか
		int mouseX = Application::GetInstance()->_mouseX;
		int mouseY = Application::GetInstance()->_mouseY;
		
		bool isMouseOver = 
		(
			mouseX > 0 && mouseX < APPLICATION::WINDOW_SIZE::WIDTH
			&&
			mouseY > 0 && mouseY < APPLICATION::WINDOW_SIZE::HEIGHT
		);
		
		if ( isMouseOver ) {
			// ホイールでスクロールしたい
			int mouseWheel = Application::GetInstance()->_mouseWheel;
			_scrollBar->MoveUp(mouseWheel);
		}
		
	}
	
	
	///// 表の中身を更新
	//
	int currentCursorPosX = _position.x;
	int currentCursorPosY = _position.y;
	
	for ( unsigned int i = 0; i < _displayVCount; ++i ) {
		

		int y = i + _scrollBar->GetCurrentIndex();
		if ( y >= _itemList.size() ) {
			break;
		}
		
		currentCursorPosX = _position.x;
		
		for ( unsigned int x = 0; x < _itemList[y].size(); ++x ) {
			
			FreeBattleIcon *icon = _itemList[y][x];
			
			icon->SetPosition ( currentCursorPosX, currentCursorPosY );
			
			icon->Update(window);

			currentCursorPosX += iconWidth + paddingX;
			
			if ( icon->IsSelected() ) {
				
				_isSelected = true;
				_selectedId = icon->GetId();
			}
			
		}
		
		currentCursorPosY += iconHeight + paddingY;
		
	}
	
	///// シークレット枠を見る
	for ( unsigned int i = 0; i < _itemList_secret.size(); ++i ) {
		
		FreeBattleIcon *icon = _itemList_secret[i];
		
		icon->Update(window);
		icon->SetPosition ( 0, 0 ); // スジ悪ですが、これをしないとタイトル文字の位置が設定されない。数字はダミー。
		
		if ( icon->IsSelected() ) {
			
			_isSelected = true;
			_selectedId = icon->GetId();
		}
	}
}

//=========================================================
// 描画
//=========================================================
void FreeBattleSelector::Draw   ( sf::RenderWindow &window ) {

	if ( !_isActive ) { return; }

	
	window.draw ( _back_Frame );

	for ( unsigned int i = 0; i < _displayVCount; ++i ) {
		
		int y = i + _scrollBar->GetCurrentIndex();
		if ( y >= _itemList.size() ) {
			break;
		}
		
		for ( unsigned int x = 0; x < _itemList[y].size(); ++x ) {
			
			FreeBattleIcon *icon = _itemList[y][x];
			
			icon->Draw(window);
			
		}
		
	}
	
	// シークレット枠
	for ( unsigned int i = 0; i < _itemList_secret.size(); ++i ) {
		
		FreeBattleIcon *icon = _itemList_secret[i];
		
		icon->Draw(window);
		
	}
	

		
	_caption->Draw(window);
	
	if (_scrollBar != NULL) { 
		_scrollBar->Draw(window);
	}
	
	_closeButton->Draw(window);
}

	
	
//=========================================================
// 選択されているかどうか
//=========================================================
bool FreeBattleSelector::IsSelected() {
	
	bool result = false;
	
	result = _isSelected;
	
	return result;
}
	
//=========================================================
// 選択されているもののIDを取得
//=========================================================
string FreeBattleSelector::GetSelectedId() {
	
	string result = "";
	
	result = _selectedId;
	
	return result;
}	



