//=========================================================
//
// セーブデータ間で共有のデータ、のセーブデータ
//
//=========================================================

#include "GlobalSaveData.h"
#include "ReadTextManager.h"

#include "../Utils/Utils_String.h"
#include "../Utils/Utils_Convert.h"


//=========================================================
// コンストラクタ
// ファイル名を必ず受け取るのです。
//=========================================================
GlobalSaveData::GlobalSaveData( string fileName ) {

	_fileName = fileName; 
}

//=========================================================
// デストラクタ
//=========================================================
GlobalSaveData::~GlobalSaveData() {

}

//=========================================================
// 読み込み
// ファイルを読み込んで内容を受け取る。
// もともと持っていた変数の値とかは、消される。
//=========================================================
void GlobalSaveData::Load () {

	///// 読み込みまして、
	map< string, SaveData::SaveDataUnit > loadDataList;
	ReadSaveData ( loadDataList );

	// ファイルがあれば、中身もそれなりであるはず、という思い込み、または思い切り。
	// 中身がない＝ファイル読み込みエラー、とみなす
	if ( loadDataList.size() <= 0 ) {

		return;
	}



	///// セーブ用のシリアライズデータ再現：変数 数値 _variable_Global_Int
	{
		_saveData_variable_Int.clear();

		for ( map<string, SaveData::SaveDataUnit>::iterator it = loadDataList.begin(); it != loadDataList.end(); ++it) {

			
			// キーに "GVint_"がついてる？
			string suffix = "GVint_";
			string key    = it->first;

			string key_suffix     = ( key.size() >= suffix.size() ) ? key.substr(0, suffix.size()) : "";
			string key_secondHalf = ( key.substr ( key_suffix.size()) );

			if ( key_suffix == suffix ) {

				// ここでは、キーのままに受け取る
				_saveData_variable_Int [ key ] = it->second.GetData_as_Int();
			}
		}
	}

	///// セーブ用のシリアライズデータ再現：変数 文字列 _variable_Global_String
	{
		_saveData_variable_String.clear();

		for ( map<string, SaveData::SaveDataUnit>::iterator it = loadDataList.begin(); it != loadDataList.end(); ++it) {

			
			// キーに "GVstring_"がついてる？
			string suffix = "GVstring_";
			string key    = it->first;

			string key_suffix     = ( key.size() >= suffix.size() ) ? key.substr(0, suffix.size()) : "";
			string key_secondHalf = ( key.substr ( key_suffix.size()) );

			if ( key_suffix == suffix ) {

				// ここでは、キーのままに受け取る
				_saveData_variable_String [ key ] = it->second.GetData_as_String();
			}
		}
	}

	///// セーブ用のシリアライズデータ再現：既読データ
	{
		// 一応、手元のを消しておくよ。
		_saveData_readData.clear();

		// 件数を取得
		int length = loadDataList [ "ReadDataLength" ].GetData_as_Int();

		// １件ずつ取得
		for ( int i = 0; i < length; ++i ) {

			string index = Utils_Convert::Int_to_String ( i );

			ReadTextManager::SaveData readData;

			readData.processLabel = loadDataList [ "ReadData_processLabel-" + index ].GetData_as_String();
			readData.unitIndex    = loadDataList [ "ReadData_unitIndex-"    + index ].GetData_as_Int();
			readData.byCharIndex  = loadDataList [ "ReadData_byCharIndex-"  + index ].GetData_as_Int();

			_saveData_readData.push_back ( readData );
		}
	}
}

//=========================================================
// 書き込み
// ファイルに書き込む。
// ファイルのもともとの内容は、消される。
//=========================================================
void GlobalSaveData::Save () {

	vector< SaveData::SaveDataUnit > saveDataList;


	///// セーブ用のシリアライズデータ作成：変数 数値 _variable_Local_Int
	{
		for ( map<string, int>::iterator it = _saveData_variable_Int.begin(); it != _saveData_variable_Int.end(); ++it) {

			// 元の変数リストのキーに "GVint_" がついている。
			saveDataList.push_back ( SaveData::SaveDataUnit ( it->first,  it->second ));
		}
	}

	///// セーブ用のシリアライズデータ作成：変数 文字列 _variable_Local_String
	{
		for ( map<string, string>::iterator it = _saveData_variable_String.begin(); it != _saveData_variable_String.end(); ++it) {

			// 元の変数リストのキーに "GVstring_" がついている。
			saveDataList.push_back ( SaveData::SaveDataUnit ( it->first,  it->second ));
		}
	}

	///// セーブ用のシリアライズデータ作成：既読データ
	{
		// 件数を記録
		int length = _saveData_readData.size();
		saveDataList.push_back ( SaveData::SaveDataUnit ( "ReadDataLength",  length ));

		// １件ずつ記録
		for ( int i = 0; i < length; ++i ) {

			string index = Utils_Convert::Int_to_String ( i );


			saveDataList.push_back ( SaveData::SaveDataUnit ( "ReadData_processLabel-" + index, _saveData_readData[i].processLabel ));
			saveDataList.push_back ( SaveData::SaveDataUnit ( "ReadData_unitIndex-"    + index, _saveData_readData[i].unitIndex    ));
			saveDataList.push_back ( SaveData::SaveDataUnit ( "ReadData_byCharIndex-"  + index, _saveData_readData[i].byCharIndex  ));
		}
	}

	WriteSaveData ( saveDataList );

}




//=========================================================
// 状態をセーブ：変数＿Int
//=========================================================
void GlobalSaveData::Save_GlobalVariable_Int ( map <string, int>    &variable_Global_Int ) {

	_saveData_variable_Int.clear();

	for ( map<string, int>::iterator it = variable_Global_Int.begin(); it != variable_Global_Int.end(); ++it ) {

		///// キーに目印をつけて、受け取る。ほかのキーとかぶらないようにするため。
		string key = "GVint_" + it->first;
		_saveData_variable_Int [ key ] = it->second;
	}
}

//=========================================================
// 状態をロード：変数＿Int
//=========================================================
void GlobalSaveData::Load_GlobalVariable_Int ( map <string, int>    &variable_Global_Int ) {

	variable_Global_Int.clear();

	for ( map<string, int>::iterator it = _saveData_variable_Int.begin(); it != _saveData_variable_Int.end(); ++it ) {

		///// キーに目印をつけた目印をはずして、渡す。
		string key    = it->first;

		// キーに "LVstring_"がついてる？
		string suffix = "GVint_";
		
		string key_suffix     = ( key.size() >= suffix.size() ) ? key.substr(0, suffix.size()) : "";
		string key_secondHalf = ( key.substr ( key_suffix.size()) );

		variable_Global_Int [ key_secondHalf ] = it->second;
	}
}

//=========================================================
// 状態をセーブ：変数＿String
//=========================================================
void GlobalSaveData::Save_GlobalVariable_String ( map <string, string> &variable_Global_String ) {

	_saveData_variable_String.clear();

	for ( map<string, string>::iterator it = variable_Global_String.begin(); it != variable_Global_String.end(); ++it ) {

		///// キーに目印をつけて、受け取る。ほかのキーとかぶらないようにするため。
		string key = "GVstring_" + it->first;
		_saveData_variable_String [ key ] = it->second;
	}
}

//=========================================================
// 状態をロード：変数＿String
//=========================================================
void GlobalSaveData::Load_GlobalVariable_String ( map <string, string> &variable_Global_String ) {

	variable_Global_String.clear();

	for ( map<string, string>::iterator it = _saveData_variable_String.begin(); it != _saveData_variable_String.end(); ++it ) {

		///// キーに目印をつけた目印をはずして、渡す。
		string key    = it->first;

		// キーに "LVstring_"がついてる？
		string suffix = "GVstring_";
		
		string key_suffix     = ( key.size() >= suffix.size() ) ? key.substr(0, suffix.size()) : "";
		string key_secondHalf = ( key.substr ( key_suffix.size()) );

		variable_Global_String [ key_secondHalf ] = it->second;
	}
}
//=========================================================
// 状態をセーブ：既読データ
//=========================================================
void GlobalSaveData::Save_ReadData ( ReadTextManager *readTextManager  ) {

	_saveData_readData = readTextManager->GetSaveData();

}
//=========================================================
// 状態をロード：既読データ
//=========================================================
void GlobalSaveData::Load_ReadData ( ReadTextManager *readTextManager  ) {

	readTextManager->Load( _saveData_readData );
}

//=========================================================
// 書き込める形式で書き込む
//=========================================================
void GlobalSaveData::WriteSaveData ( vector<SaveData::SaveDataUnit> &saveDataUnitList ) {

	/*
		この段階で、キーの前後に<>をつけて書き込む。
	*/

	///// ファイルを開く。
	FILE *pFile;
	pFile = fopen( _fileName.c_str(), "wb");

	for ( unsigned int i = 0; i < saveDataUnitList.size(); ++i ) {

		// キーを書き込む。末尾に : をつけて。
		string key = saveDataUnitList[i].key + ":";
		for ( int n = 0; n < key.size(); ++n ) {
			fwrite ( &key[n], sizeof(char), 1, pFile);
		}

		// 型を書き込む
		fwrite ( &saveDataUnitList[i].type, sizeof(saveDataUnitList[i].type), 1, pFile);

		// 長さを書き込む
		fwrite ( &saveDataUnitList[i].dataLength, sizeof(saveDataUnitList[i].dataLength), 1, pFile);

		// データを書き込む
		fwrite ( saveDataUnitList[i].data, saveDataUnitList[i].dataLength, 1, pFile);

		// この段階で、データは消してしまっていい？
		saveDataUnitList[i].Clear();
	}

	///// 書き込み終了。ファイルを閉じる。
	fclose(pFile);
}

//=========================================================
// ともかく読み込みます。
//=========================================================
void GlobalSaveData::ReadSaveData ( map<string, SaveData::SaveDataUnit> &result ) {

	///// ファイルを開く
	FILE *pFile;
	pFile = fopen (_fileName.c_str(), "rb");

	///// ファイル、開けた？
	if (pFile != NULL) {

		bool isContinule = true;
		while ( isContinule ) {

			///// 一件分
			SaveData::SaveDataUnit unit;

			///// キーを読み込む。
			string key = "";
			while (true) {
				char c;
				
				isContinule = fread (&c, sizeof(char), 1, pFile);

				if ( c == ':' ) {
					break;
				} else {
					key.push_back ( c );
				}
				if ( !isContinule ) { break; }
			}
			if ( !isContinule ) { break; }

			///// 同じキーがあった？ それは変ですね。
			if ( result.find ( key ) != result.end() ) {
				printf ( "Error : 同じキーがあった。(%s)\n", key.c_str());
				
			}

			///// 大丈夫のようなので、キーを受け取る。
			unit.key = key;

			///// 型を読み込む
			isContinule = fread ( &unit.type, sizeof(int), 1, pFile);
			if ( !isContinule ) { break; }

			///// サイズを読み込む
			isContinule = fread ( &unit.dataLength, sizeof(int), 1, pFile);
			if ( !isContinule ) { break; }

			///// データ本体を読み込む
			if ( unit.dataLength > 0 ) {
				unit.data = new unsigned char[ unit.dataLength ];
				isContinule = fread ( unit.data, unit.dataLength, 1, pFile);
			} else {
				unit.data = NULL;
			}
			if ( !isContinule ) { break; }
			
			///// 確保しておく。
			result [ unit.key ] = unit;

		}

		///// ファイルを閉じる
		fclose(pFile);

	} else {
		printf ("ファイルを開けませんでした（%s）\n",_fileName.c_str());
	}

}

