//=========================================================
//
// 画像。画面に表示するモノとしての。
//
//=========================================================

#include "Image.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Utils_Convert.h"
#include "../GUI_SFML/SFML_Utils.h"
#include "Application.h"

//=========================================================
// 定数：画像差し替えエフェクト
//=========================================================
const string Image::CHANGE_IMAGE_EFFECT_NONE              = "";                  // エフェクトなし。即座に差し替え
const string Image::CHANGE_IMAGE_EFFECT_FADE_OUT_IN       = "FADE_OUT_IN";       // フェードアウト → フェードイン
const string Image::CHANGE_IMAGE_EFFECT_FADE_IN_OUT       = "FADE_IN_OUT";       // フェードイン   → フェードアウト
const string Image::CHANGE_IMAGE_EFFECT_FADE_OUT_IN_SIMUL = "FADE_OUT_IN_SIMUL"; // フェードアウト と フェードインを同時進行
const string Image::CHANGE_IMAGE_EFFECT_FADE_OUT          = "FADE_OUT";          // フェードアウト → 新画像の表示エフェクトはナシ。
const string Image::CHANGE_IMAGE_EFFECT_FADE_IN           = "FADE_IN";           // 現画像の上に新画像をフェードイン
const string Image::CHANGE_IMAGE_EFFECT_WIPE_OUT_IN       = "WIPE_OUT_IN";       // ワイプアウト → イン
const string Image::CHANGE_IMAGE_EFFECT_WIPE_OUT          = "WIPE_OUT";          // 現画像をワイプアウト：新画像は裏側で待機
const string Image::CHANGE_IMAGE_EFFECT_WIPE_IN           = "WIPE_IN";           // 新画像を現画像の上にワイプイン

//=========================================================
// 定数：現画像・新画像、表示順（トランジション中の都合）
//=========================================================
const int Image::DRAW_ORDER_CURRENT_NEW = 1;
const int Image::DRAW_ORDER_NEW_CURRENT = 2;

//=========================================================
// 定数：移動方式
//=========================================================
const string Image::MOVE_FORM_SPRING = "SPRING";        // 移動方式。バネみたいにびよ〜ん
const string Image::MOVE_FORM_LINEAR = "LINEAR";        // 移動方式。線形。スィー。
	

//=========================================================
// コンストラクタ
//=========================================================
Image::Image() {

	_position_Goal.x = 0.0f;
	_position_Goal.y = 0.0f;
	_position_Current.x = 0.0f;
	_position_Current.y = 0.0f;
	_z = 0;
	_label = "";

	_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;

	_drawOrder = DRAW_ORDER_NEW_CURRENT;

	_wipeOutSpeed = 5;
	_wipeInSpeed  = 5;

	_fadeOutSpeed = 5;
	_fadeInSpeed  = 5;


	// 移動方式関連
	_moveForm                  = Image::MOVE_FORM_SPRING; // 移動方式
	_moveForm_spring_softness  = 3.0f;  // バネの柔らかさ
	_moveForm_linear_increment = 10.0f; // 線形移動の増分 
	
	_isSkipping = false;

	_isActive = true;
	
	_isMouseOverSeeThrough = false;
	_isMouseOverToSeeThrough = false;
}
//=========================================================
// デストラクタ
//=========================================================
Image::~Image() {
	printf ("Image::~Image()[%s]\n", _label.c_str());
}

//=========================================================
// 表示する画像を変更する
//=========================================================
void Image::ChangeImage ( string fileName, string transitionEffect ) {

	if      ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_NONE              ) { Transition_Preparation_NONE        ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_OUT_IN       ) { Transition_Preparation_FADE_OUT_IN ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_IN_OUT       ) { Transition_Preparation_FADE_IN_OUT ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_OUT_IN_SIMUL ) { Transition_Preparation_FADE_OUT_IN_SIMUL ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_OUT          ) { Transition_Preparation_FADE_OUT    ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_IN           ) { Transition_Preparation_FADE_IN     ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_WIPE_OUT_IN       ) { Transition_Preparation_WIPE_OUT_IN ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_WIPE_OUT          ) { Transition_Preparation_WIPE_OUT    ( fileName ); }
	else if ( transitionEffect == Image::CHANGE_IMAGE_EFFECT_WIPE_IN           ) { Transition_Preparation_WIPE_IN     ( fileName ); }

	///// そうそう、変化中ですよ。
	_transitionEffect = transitionEffect;
}

//=========================================================
// トランジションエフェクト＿準備処理：エフェクトなし
//=========================================================
void Image::Transition_Preparation_NONE ( string newImageFileName ) {

	///// 新イメージ：なし
	_newImage.fileName = "";
	
	// 反転の状態などを保存しておいて……
	float rotation = _image.sprite.getRotation();
	sf::Vector2f scale = _image.sprite.getScale();
	sf::Vector2f position = _image.sprite.getPosition();

	///// 現イメージを直に差し替え
	_image.LoadFromFile ( newImageFileName );
	
	// 反転の状態などを戻す
	_image.sprite.setRotation ( rotation );
	_image.sprite.setScale    ( scale );
	_image.sprite.setPosition ( position );
	
	///// 新画像が裏で待機
	_drawOrder = DRAW_ORDER_NEW_CURRENT;
}
//=========================================================
// トランジションエフェクト＿準備処理：フェードアウトして、イン
//=========================================================
void Image::Transition_Preparation_FADE_OUT_IN ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// 新イメージのアルファ値：ゼロスタート。
	_newImage.SetAlpha (0);

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );
	
	///// 新画像が裏で待機
	_drawOrder = DRAW_ORDER_NEW_CURRENT;
}
//=========================================================
// トランジションエフェクト＿準備処理：フェードインしてアウト
//=========================================================
void Image::Transition_Preparation_FADE_IN_OUT ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// 新イメージのアルファ値：ゼロスタート。
	_newImage.SetAlpha (0);

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );
	
	///// 新画像が表に出る
	//_drawOrder = DRAW_ORDER_NEW_CURRENT;
	_drawOrder = DRAW_ORDER_CURRENT_NEW;
}
//=========================================================
// トランジションエフェクト＿準備処理：フェードアウトとインを同時進行
//=========================================================
void Image::Transition_Preparation_FADE_OUT_IN_SIMUL ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// 新イメージのアルファ値：ゼロスタート。
	_newImage.SetAlpha (0);

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );
	
	///// 新画像が裏で待機？ 裏でいい？
	_drawOrder = DRAW_ORDER_NEW_CURRENT;
}
//=========================================================
// トランジションエフェクト＿準備処理：フェードアウト。新画像は、エフェクトなし
//=========================================================
void Image::Transition_Preparation_FADE_OUT    ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );
	
	///// 新画像が裏で待機
	_drawOrder = DRAW_ORDER_NEW_CURRENT;
}
//=========================================================
// トランジションエフェクト＿準備処理：現画像の上に新画像をフェードイン
//=========================================================
void Image::Transition_Preparation_FADE_IN ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// 新イメージのアルファ値：ゼロスタート。
	_newImage.SetAlpha (0);

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );

	///// 新画像を上に表示
	_drawOrder = DRAW_ORDER_CURRENT_NEW;
}

//=========================================================
// トランジションエフェクト＿準備処理：ワイプアウト → イン
//=========================================================
void Image::Transition_Preparation_WIPE_OUT_IN ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// まずは、全ピクセルを、ワイプアウト状態にしておく。
	for ( int y = 0; y < _newImage.image.getSize().y; ++y ) {
		for ( int x = 0; x < _newImage.image.getSize().x; ++x ) {
			sf::Color color = _newImage.image.getPixel( x, y );
			color.a = 0;
			_newImage.image.setPixel( x, y, color );
		}
	}
	_newImage.texture.loadFromImage ( _newImage.image );
	_newImage.sprite.setTexture (_newImage.texture );

	///// ワイプイン後の完成形は、これですよ。
	_newImage_forWipeInned.LoadFromFile ( newImageFileName );

	_wipeInAlphaLimit  = 0;
	_wipeOutAlphaLimit = 0;

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );
	
	///// 新画像が裏で待機：たぶん、どっちでもいい。
	_drawOrder = DRAW_ORDER_NEW_CURRENT;
}

//=========================================================
// トランジションエフェクト＿準備処理：現画像をワイプアウト。新画像は裏で待機。
//=========================================================
void Image::Transition_Preparation_WIPE_OUT    ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );


	_wipeOutAlphaLimit = 0;
	

	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );

	///// 新画像が裏で待機
	_drawOrder = DRAW_ORDER_NEW_CURRENT;
}

//=========================================================
// トランジションエフェクト＿準備処理：新画像を現画像の上にワイプイン
//=========================================================
void Image::Transition_Preparation_WIPE_IN ( string newImageFileName ) {

	///// 新イメージとして、読み込む
	_newImage.LoadFromFile ( newImageFileName );

	///// まずは、全ピクセルを、ワイプアウト状態にしておく。
	for ( int y = 0; y < _newImage.image.getSize().y; ++y ) {
		for ( int x = 0; x < _newImage.image.getSize().x; ++x ) {
			sf::Color color = _newImage.image.getPixel( x, y );
			color.a = 0;
			_newImage.image.setPixel( x, y, color );
		}
	}
	_newImage.texture.loadFromImage ( _newImage.image );
	_newImage.sprite.setTexture (_newImage.texture );

	///// ワイプイン後の完成形は、これですよ。
	_newImage_forWipeInned.LoadFromFile ( newImageFileName );

	_wipeInAlphaLimit = 0;


	///// 回転値とスケール（反転）状態を引き継ぐ
	_image.TransferTransform ( _newImage );
	
	///// 新画像が表に出る
	_drawOrder = DRAW_ORDER_CURRENT_NEW;
}

//=========================================================
// 透過色を指定
//=========================================================
void Image::SetMask ( sf::Color color ) {

	_image.image.createMaskFromColor ( color );
	_image.texture.loadFromImage ( _image.image );
	_image.sprite.setTexture ( _image.texture );

	_maskColorString = GUI_SFML::SFML_Utils::Convert_color_to_string ( color );
}

//=========================================================
// トランジションエフェクト＿処理中：エフェクトなし
//=========================================================
void Image::Transition_Update_NONE () {

	// じゃあ、ここにはこないのではないの？ 来たかったらきてもいいですよ。ふわふわ。
}
//=========================================================
// トランジションエフェクト＿処理中：フェードアウトして、フェードイン
//=========================================================
void Image::Transition_Update_FADE_OUT_IN () {

	///// 現イメージがフェードアウトし終わってない？
	if (_image.GetAlpha() > 0 && _image.fileName != "") {

		// 現イメージをフェードアウト
		_image.AddAlpha ( -_fadeOutSpeed );

	///// 現イメージはフェードアウトし終わっている
	} else {

		///// 新イメージがフェードインし終わってない？
		if ( _newImage.GetAlpha() < 255 && _newImage.fileName != "") {

			// 新イメージをフェードイン
			_newImage.AddAlpha ( _fadeInSpeed );

		///// 新イメージがフェードインし終わっている
		} else {

			// イメージを入れ替えて、エフェクト終了。
			_image.LoadFromFile ( _newImage.fileName );
			_newImage.TransferTransform ( _image );
			_newImage.fileName  = "";

			_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
		}
	}

}

//=========================================================
// トランジションエフェクト＿処理中：フェードインして、フェードアウト
//=========================================================
void Image::Transition_Update_FADE_IN_OUT () {

	///// 新イメージがフェードインし終わってない？
	if ( _newImage.GetAlpha() < 255 && _newImage.fileName != "") {

		// 新イメージをフェードイン
		_newImage.AddAlpha ( _fadeInSpeed );

	///// 現イメージはフェードアウトし終わっている
	} else {

		///// 現イメージがフェードアウトし終わってない？
		if (_image.GetAlpha() > 0 && _image.fileName != "") {

			// 現イメージをフェードアウト
			_image.AddAlpha ( -_fadeOutSpeed );

		///// 現イメージがフェードアウトし終わっている
		} else {

			// イメージを入れ替えて、エフェクト終了。
			_image.LoadFromFile ( _newImage.fileName );
			_newImage.TransferTransform ( _image );
			_newImage.fileName  = "";

			_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
		}
	}

}


//=========================================================
// トランジションエフェクト＿処理中：フェードアウトとインを同時進行
//=========================================================
void Image::Transition_Update_FADE_OUT_IN_SIMUL () {

	///// 現イメージがフェードアウトし終わってない？
	bool isEnd_FadeOut = true;
	if (_image.GetAlpha() > 0 && _image.fileName != "") {

		// 現イメージをフェードアウト
		_image.AddAlpha ( -_fadeOutSpeed );

		isEnd_FadeOut = false;
	}
	///// 新イメージがフェードインし終わってない？
	bool isEnd_FadeIn = true;
	if ( _newImage.GetAlpha() < 255 && _newImage.fileName != "") {

		// 新イメージをフェードイン
		_newImage.AddAlpha ( _fadeInSpeed );

		isEnd_FadeIn = false;
	}


	///// フェードインもアウトも終了している
	if ( isEnd_FadeOut && isEnd_FadeIn ) {

		// イメージを入れ替えて、エフェクト終了。
		_image.LoadFromFile ( _newImage.fileName );
		_newImage.TransferTransform ( _image );
		_newImage.fileName  = "";

		_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
	}
}

//=========================================================
// トランジションエフェクト＿処理中：フェードアウト。新イメージはエフェクトなし
//=========================================================
void Image::Transition_Update_FADE_OUT    () {

	///// 現イメージがフェードアウトし終わってない？
	if (_image.GetAlpha() > 0 && _image.fileName != "") {

		// 現イメージをフェードアウト
		_image.AddAlpha ( -_fadeOutSpeed );

	///// 現イメージはフェードアウトし終わっている
	} else {

		// イメージを入れ替えて、エフェクト終了。
		_image.LoadFromFile ( _newImage.fileName );
		_newImage.TransferTransform ( _image );
		_newImage.fileName  = "";

		_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
	}
}
//=========================================================
// トランジションエフェクト＿処理中：現画像の上に新画像をフェードイン
//=========================================================
void Image::Transition_Update_FADE_IN () {

	///// 新イメージがフェードインし終わってない？
	if ( _newImage.GetAlpha() < 255 && _newImage.fileName != "") {

		// 新イメージをフェードイン
		_newImage.AddAlpha ( _fadeInSpeed );

	///// 新イメージがフェードインし終わっている
	} else {

		// イメージを入れ替えて、エフェクト終了。
		_image.LoadFromFile ( _newImage.fileName );
		_newImage.TransferTransform ( _image );
		_newImage.fileName  = "";

		_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
	}
}
//=========================================================
// トランジションエフェクト＿処理中：ワイプアウト → イン
//=========================================================
void Image::Transition_Update_WIPE_OUT_IN () {

	///// 現画像をワイプアウトし終わった？
	if ( _wipeOutAlphaLimit > 255 ) {

		///// 新画像をワイプインし終わった？
		if ( _wipeInAlphaLimit > 255 ) {

			// イメージを入れ替えて、エフェクト終了。
			_image.LoadFromFile ( _newImage.fileName );
			_newImage.TransferTransform ( _image );
			_newImage.fileName  = "";

			_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;

		///// まだ。
		} else {

			_wipeInAlphaLimit += _wipeInSpeed;

			for ( int y = 0; y < _wipeInImage.image.getSize().y; ++y ) {
				for ( int x = 0; x < _wipeInImage.image.getSize().x; ++x ) {

					sf::Color pixColor = _wipeInImage.image.getPixel( x, y );
					if ( pixColor.r < _wipeInAlphaLimit ) {
				
						_newImage.image.setPixel( x, y, 
							_newImage_forWipeInned.image.getPixel( x, y )
						);
					}
				}
			}
			_newImage.texture.loadFromImage(_newImage.image);
			_newImage.sprite.setTexture (_newImage.texture);

		}

	///// まだ。
	} else {

		_wipeOutAlphaLimit += _wipeOutSpeed;

		for ( int y = 0; y < _wipeOutImage.image.getSize().y; ++y ) {
			for ( int x = 0; x < _wipeOutImage.image.getSize().x; ++x ) {

				sf::Color pixColor = _wipeOutImage.image.getPixel( x, y );
				if ( pixColor.r < _wipeOutAlphaLimit ) {
				
					pixColor.a = 0;
					_image.image.setPixel( x, y, pixColor );
				}
			}
		}
		_image.texture.loadFromImage(_image.image);
		_image.sprite.setTexture (_image.texture);
	}




}

//=========================================================
// トランジションエフェクト＿処理中：現画像をワイプアウト。新画像は裏で待機。
//=========================================================
void Image::Transition_Update_WIPE_OUT () {

	/*
		ワイプ用の画像は、サイズが合っているものとする。
	*/
	_wipeOutAlphaLimit += _wipeOutSpeed;

	for ( int y = 0; y < _wipeOutImage.image.getSize().y; ++y ) {
		for ( int x = 0; x < _wipeOutImage.image.getSize().x; ++x ) {

			sf::Color pixColor = _wipeOutImage.image.getPixel( x, y );
			if ( pixColor.r < _wipeOutAlphaLimit ) {
				
				pixColor.a = 0;
				_image.image.setPixel( x, y, pixColor );
			}
		}
	}
	_image.texture.loadFromImage(_image.image);
	_image.sprite.setTexture (_image.texture);

	// 現画像をワイプアウトしきった？
	if ( _wipeOutAlphaLimit > 255 ) {

		// イメージを入れ替えて、エフェクト終了。
		_image.LoadFromFile ( _newImage.fileName );
		_newImage.TransferTransform ( _image );
		_newImage.fileName  = "";

		_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
	}
}

//=========================================================
// トランジションエフェクト＿処理中：新画像を現画像の上にワイプイン
//=========================================================
void Image::Transition_Update_WIPE_IN     () {

	_wipeInAlphaLimit += _wipeInSpeed;

	for ( int y = 0; y < _wipeInImage.image.getSize().y; ++y ) {
		for ( int x = 0; x < _wipeInImage.image.getSize().x; ++x ) {

			sf::Color pixColor = _wipeInImage.image.getPixel( x, y );
			if ( pixColor.r < _wipeInAlphaLimit ) {
				
				_newImage.image.setPixel( x, y, 
					_newImage_forWipeInned.image.getPixel( x, y )
				);
			}
		}
	}
	_newImage.texture.loadFromImage(_newImage.image);
	_newImage.sprite.setTexture (_newImage.texture);

	// 新画像をワイプインしきった？
	if ( _wipeInAlphaLimit > 255 ) {

		// イメージを入れ替えて、エフェクト終了。
		_image.LoadFromFile ( _newImage.fileName );
		_newImage.TransferTransform ( _image );
		_newImage.fileName  = "";

		_transitionEffect = Image::CHANGE_IMAGE_EFFECT_NONE;
	}
}


//=========================================================
// 移動方式関連の設定処理
//=========================================================
void Image::SetMoveForm ( string form ) {
	
	if ( form == Image::MOVE_FORM_SPRING ) {
		_moveForm = Image::MOVE_FORM_SPRING;
	}
	else if ( form == Image::MOVE_FORM_LINEAR ) {
		_moveForm = Image::MOVE_FORM_LINEAR;
	}
	else {
		
		printf ("未定義の移動方式が指定されました : %s\n", form.c_str());
		printf ("現在の移動方式(%s)のままにしておきます\n", _moveForm.c_str());
	}
}
void Image::SetMoveForm_SPRING_softness  ( float value ) {
	
	if ( value < 0.0f ) {
		printf ( "移動方式 SPRING の柔らかさとして 0未満が指定されました。0除算になってしまうので、ちょっとゲタを履かせておきます\n" );
		_moveForm_spring_softness = 0.1f;
		
		printf (" _moveForm_spring_softness : %f\n", _moveForm_spring_softness);

	} else {
		_moveForm_spring_softness = value;
	}

}
void Image::SetMoveForm_LINEAR_increment ( float value ) {
	
	if ( value <= 0.0f ) {
		printf ( "移動方式 LINEAR の増分 0以下が指定されました。意味わかんないので、ちょっとゲタを履かせておきます\n" );

		_moveForm_linear_increment = 1.0f;
		
		printf (" _moveForm_linear_increment : %f\n", _moveForm_linear_increment);

	} else {
		_moveForm_linear_increment = value;
	}

	
}
	

//=========================================================
// 更新子処理：目的地へ移動
//=========================================================
void Image::Update_MoveToGoal() {
	
	if      ( _moveForm == Image::MOVE_FORM_SPRING ) { Update_MoveToGoal_SPRING();}
	else if ( _moveForm == Image::MOVE_FORM_LINEAR ) { Update_MoveToGoal_LINEAR();}

}

//=========================================================
// 更新子処理：目的地へ移動：SPRING
//=========================================================
void Image::Update_MoveToGoal_SPRING() {

	// 現在地と目標地の距離の、
	float distance = Utils_Cal::GetDistance( _position_Current.x, _position_Current.y, _position_Goal.x, _position_Goal.y );

	// 半分？の距離を、
	float moveDistance = ( distance < _moveForm_spring_softness ) ? distance : distance / _moveForm_spring_softness;
	
	// 移動するのですけれども、あまりに増分が小さくなると気が遠くなるので、適当に加減する
	if ( moveDistance < 1.0f) {
		moveDistance = 1.0f;
	}
	
	// 移動する
	pair<float, float> moveResult = 
		Utils_Cal::MoveTo ( _position_Current.x, _position_Current.y, _position_Goal.x, _position_Goal.y, moveDistance );

	_position_Current.x = moveResult.first;
	_position_Current.y = moveResult.second;


	// 反映
	_image.sprite.setPosition( _position_Current );
	_newImage.sprite.setPosition( _position_Current );
	_newImage_forWipeInned.sprite.setPosition( _position_Current );
}

//=========================================================
// 更新子処理：目的地へ移動：LINEAR
//=========================================================
void Image::Update_MoveToGoal_LINEAR() {
	
	// 増分
	float moveForm_linear_increment = _moveForm_linear_increment;
	if ( _isSkipping ) {
		moveForm_linear_increment *= 5.0f;
		//printf ("Image skipping : %f -> %f \n", _moveForm_linear_increment, moveForm_linear_increment);
	}
	
	// 増分の分、移動する
	pair<float, float> moveResult = 
		Utils_Cal::MoveTo ( _position_Current.x, _position_Current.y, _position_Goal.x, _position_Goal.y, moveForm_linear_increment );

	_position_Current.x = moveResult.first;
	_position_Current.y = moveResult.second;
	
	// 反映
	_image.sprite.setPosition( _position_Current );
	_newImage.sprite.setPosition( _position_Current );
	_newImage_forWipeInned.sprite.setPosition( _position_Current );

}
	

//=========================================================
// 更新
//=========================================================
void Image::Update ( sf::RenderWindow &window ) {

	if (!_isActive) { return; }

	///// トランジションエフェクト処理中？
	if      ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_OUT_IN       ) { Transition_Update_FADE_OUT_IN (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_IN_OUT       ) { Transition_Update_FADE_IN_OUT (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_OUT_IN_SIMUL ) { Transition_Update_FADE_OUT_IN_SIMUL (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_OUT          ) { Transition_Update_FADE_OUT    (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_FADE_IN           ) { Transition_Update_FADE_IN     (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_WIPE_OUT_IN       ) { Transition_Update_WIPE_OUT_IN (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_WIPE_OUT          ) { Transition_Update_WIPE_OUT    (); }
	else if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_WIPE_IN           ) { Transition_Update_WIPE_IN     (); }


	Update_MoveToGoal();

	/*
	_image.sprite.setOrigin(
		_image.image.getSize().x / 2,
		_image.image.getSize().y / 2
	);
	_newImage.sprite.setOrigin(
		_newImage.image.getSize().x / 2,
		_newImage.image.getSize().y / 2
	);
	_newImage_forWipeInned.sprite.setOrigin(
		_newImage_forWipeInned.image.getSize().x / 2,
		_newImage_forWipeInned.image.getSize().y / 2
	);
	*/
	
	//_image.image.getPixel();
	
	// もしかして、これが重い？
	if ( _isMouseOverSeeThrough ) {
		
		_isMouseOverToSeeThrough = false;
			
		int mouseX = Application::GetInstance()->_mouseX;
		int mouseY = Application::GetInstance()->_mouseY;
		//printf ( "マウス(%d, %d)\n", mouseX, mouseY );
		
		// 重かった原因：毎フレーム、テクスチャから sf::Image を コピーする必要はなかった。
		/*
		sf::Texture *sfTexture = _image.sprite.getTexture();
		//printf ( "%d\n", sfTexture );
		if ( sfTexture ) {
			
		}
		*/
		
		sf::Image sfImage = _image.image;
		
		//int posX = mouseX - _position_Current.x + _image.sprite.getOrigin().x;
		//int posY = mouseY - _position_Current.y + _image.sprite.getOrigin().y;
		
		int posX = mouseX - _image.sprite.getPosition().x + _image.sprite.getOrigin().x;
		int posY = mouseY - _image.sprite.getPosition().y + _image.sprite.getOrigin().y;
		
		if ( posX > 0 && posY > 0 && posX < sfImage.getSize().x && posY < sfImage.getSize().y ) {
				
			sf::Color pixColor =  sfImage.getPixel ( posX, posY );
			
			if ( pixColor.a > 0 ) {
				
				_isMouseOverToSeeThrough = true;
				//printf ( "マウスオーバー\n" );
				//printf ( "  色(%d,%d,%d,%d)(RGBA)\n", pixColor.r, pixColor.b, pixColor.g, pixColor.a );
				
			}
		}
	}
	
	// スキップフラグは毎フレームリセット
	_isSkipping = false;
}

//=========================================================
// 描画
//=========================================================
void Image::Draw   ( sf::RenderWindow &window ) {

	if (!_isActive) { return; }

	if ( _drawOrder == DRAW_ORDER_CURRENT_NEW ) {

		if ( _image.fileName != "" ) {
			
			sf::Color color;

			if ( _isMouseOverToSeeThrough ) {
				color = _image.sprite.getColor();
				sf::Color colorAlpha = color;
				colorAlpha.a = colorAlpha.a / 2;
				_image.sprite.setColor(colorAlpha);
			}
			
			window.draw ( _image.sprite );

			if ( _isMouseOverToSeeThrough ) {
				_image.sprite.setColor(color);
			}
		}
		if ( _newImage.fileName != "" ) {
			window.draw ( _newImage.sprite );
		}
	}
	else if ( _drawOrder == DRAW_ORDER_NEW_CURRENT ) {
		if ( _newImage.fileName != "" ) {
			window.draw ( _newImage.sprite );
		}
		if ( _image.fileName != "" ) {
				
			sf::Color color;

			if ( _isMouseOverToSeeThrough ) {
				color = _image.sprite.getColor();
				sf::Color colorAlpha = color;
				colorAlpha.a = colorAlpha.a / 2;
				_image.sprite.setColor(colorAlpha);
			}
			
			window.draw ( _image.sprite );

			if ( _isMouseOverToSeeThrough ) {
				_image.sprite.setColor(color);
			}
		}
	}
}

//=========================================================
// 現在の画像ファイル名が欲しい
//=========================================================
string Image::GetCurrentImageFileName() {
	
	string result = "";
	
	///// 画像遷移中であれば、新画像の情報を「現在」ということにする
	if ( _transitionEffect == Image::CHANGE_IMAGE_EFFECT_NONE ) {
		
		result = _image.fileName;
		
	} else {
		
		result = _newImage.fileName;
	}
	
	
	return result;
}

//=========================================================
// セーブ用のデータを作成
//=========================================================
Image::SaveData Image::GetSaveData () {

	Image::SaveData result;

	result.position_Goal    = _position_Goal;     // 目的地
	result.position_Current = _position_Current;  // 現在ただいまの表示位置
		
	result.z     = _z;
	result.label = _label;
	
	result.rotation = _image.sprite.getRotation();
	result.scale    = _image.sprite.getScale();

	/*
		画像トランジション中には、セーブは行われないものとする。

		ところが行われるので、
		トランジション中であるならば、トランジション後の情報をセーブデータとする。してね。
		
	*/
	if ( _transitionEffect == CHANGE_IMAGE_EFFECT_NONE) {
		result.image_fileName  = _image.fileName;
	} else {
		result.image_fileName  = _newImage.fileName;
	}


	result.maskColorString = _maskColorString;

	result.wipeOutImage_fileName = _wipeOutImage.fileName;
	result.wipeInImage_fileName  = _wipeInImage.fileName;

	result.wipeOutSpeed = _wipeOutSpeed;
	result.wipeInSpeed  = _wipeInSpeed;

	result.fadeOutSpeed = _fadeOutSpeed;
	result.fadeInSpeed  = _fadeInSpeed;
	
	result.moveForm = _moveForm;
	result.moveForm_spring_softness  = Utils_Convert::Float_to_String (_moveForm_spring_softness);
	result.moveForm_linear_increment = Utils_Convert::Float_to_String (_moveForm_linear_increment);

	return result;
}
//=========================================================
// セーブデータから再現
//=========================================================
void Image::Load ( SaveData &saveData ) {

	_position_Goal    = saveData.position_Goal;     // 目的地
	_position_Current = saveData.position_Current;  // 現在ただいまの表示位置
		
	_z     = saveData.z;
	_label = saveData.label;
	_image.LoadFromFile ( saveData.image_fileName );

	if ( saveData.maskColorString != "" ) {
		sf::Color maskColor = GUI_SFML::SFML_Utils::Convert_string_to_color ( saveData.maskColorString );
		SetMask ( maskColor );
	}

	_image.sprite.setRotation ( saveData.rotation );
	_image.sprite.setScale    ( saveData.scale    );
	_image.sprite.setPosition ( _position_Current );

	SetWipeOutImage ( saveData.wipeOutImage_fileName );
	SetWipeInImage  ( saveData.wipeInImage_fileName );

	_wipeOutSpeed = saveData.wipeOutSpeed;
	_wipeInSpeed  = saveData.wipeInSpeed;

	_fadeOutSpeed = saveData.fadeOutSpeed;
	_fadeInSpeed  = saveData.fadeInSpeed;
	
	_moveForm = saveData.moveForm;
	_moveForm_spring_softness  = Utils_Convert::String_to_Float( saveData.moveForm_spring_softness );
	_moveForm_linear_increment = Utils_Convert::String_to_Float( saveData.moveForm_linear_increment );


}

