//=========================================================
//
// 画像。画面に表示するモノとしての。
//
//=========================================================

#ifndef _IMAGE_AS_AN_OBJECT_THAT_IS_PUT_ON_THE_SCREEN_
#define _IMAGE_AS_AN_OBJECT_THAT_IS_PUT_ON_THE_SCREEN_

#include <SFML/Graphics.hpp>
#include <string>

#include "../Utils/Utils_Cal.h"

using namespace std;

class Image {

public:

	struct ImageSet_as_SFML {
		
		string      fileName;
		sf::Image   image;
		sf::Texture texture;
		sf::Sprite  sprite;


		//=========================================================
		// コンストラクタ
		//=========================================================
		ImageSet_as_SFML () {
			fileName  = "";
		}

		//=========================================================
		// ファイルを読み込む
		//=========================================================
		void LoadFromFile ( string file ) {

			fileName = file;
			
			if (fileName == "") {
				
				// TODO つまりテクスチャを解除したいのですけれど……
				//sprite.setTexture(0);
				//sprite.setTexture(sf::Texture());
				sprite.setScale(0.0f, 0.0f);
					
			} else {
				//image = sf::Image();
				//texture = sf::Texture();
				sprite = sf::Sprite();

				image.loadFromFile ( fileName );
				texture.loadFromImage ( image );
				sprite.setTexture ( texture );

				// デフォルトで、中央指定。
				
				sprite.setOrigin(
					image.getSize().x / 2,
					image.getSize().y / 2
				);
				

				sf::Color color = sprite.getColor();
				color.a = 255;
				sprite.setColor(color);
			}
		}
		//=========================================================
		// アルファ値だけ変更
		//=========================================================
		void SetAlpha( int a ) {
			sf::Color color = sprite.getColor();
			
			if      ( a > 255 ) { color.a = 255; }
			else if ( a < 0   ) { color.a = 0; }
			else                { color.a = a; }

			sprite.setColor(color);
		}
		void AddAlpha( int a ) {
			SetAlpha( GetAlpha() + a );
		}
		int GetAlpha() {
			sf::Color color = sprite.getColor();
			return color.a;
		}

		//=========================================================
		// 反転
		//=========================================================
		void FlipX() {
			sprite.scale ( -1.0f, 1.0f );
		}
		void FlipY() {
			sprite.scale ( 1.0f, -1.0f );
		}
		
		//=========================================================
		// 回転値とスケール値を引き継ぐ
		//=========================================================
		void TransferTransform ( ImageSet_as_SFML &next ) {
			
			next.sprite.setRotation ( sprite.getRotation() );
			next.sprite.setScale    ( sprite.getScale()    );
			next.sprite.setPosition ( sprite.getPosition() );
		}
		

	};

	//=========================================================
	// 定数：画像差し替えエフェクト
	//=========================================================
	static const string CHANGE_IMAGE_EFFECT_NONE;              // エフェクトなし。即座に差し替え
	static const string CHANGE_IMAGE_EFFECT_FADE_OUT_IN;       // フェードアウト → フェードイン
	static const string CHANGE_IMAGE_EFFECT_FADE_IN_OUT;       // フェードイン   → フェードアウト
	static const string CHANGE_IMAGE_EFFECT_FADE_OUT_IN_SIMUL; // フェードアウト と フェードインを同時進行
	static const string CHANGE_IMAGE_EFFECT_FADE_OUT;          // フェードアウト → 新画像の表示エフェクトはナシ。
	static const string CHANGE_IMAGE_EFFECT_FADE_IN;           // 現画像の上に新画像をフェードイン。
	static const string CHANGE_IMAGE_EFFECT_WIPE_OUT_IN;       // ワイプアウト → イン
	static const string CHANGE_IMAGE_EFFECT_WIPE_OUT;          // 現画像をワイプアウト：新画像は裏側で待機
	static const string CHANGE_IMAGE_EFFECT_WIPE_IN;           // 新画像を現画像の上にワイプイン

	//=========================================================
	// 定数：移動方式
	//=========================================================
	static const string MOVE_FORM_SPRING;              // 移動方式。バネみたいにびよ〜ん
	static const string MOVE_FORM_LINEAR;              // 移動方式。線形。スィー。


	Image();
	~Image();

	void Update ( sf::RenderWindow &window );
	void Draw   ( sf::RenderWindow &window );

	void ChangeImage ( string fileName, string transitionEffect = "" );
	void SetWipeOutImage ( string fileName ) { _wipeOutImage.LoadFromFile( fileName ); }
	void SetWipeInImage  ( string fileName ) { _wipeInImage. LoadFromFile( fileName ); }

	void SetMask     ( sf::Color color );

	void SetPosition  ( sf::Vector2f value ) { _position_Goal = value; MoveEnd(); }
	void SetPositionX ( int value ) { SetPosition ( sf::Vector2f( (float)value, _position_Goal.y ) ); }
	void SetPositionY ( int value ) { SetPosition ( sf::Vector2f( _position_Goal.x, (float)value ) ); }

	void MoveTo  ( sf::Vector2f value) { _position_Goal = value; }
	void MoveToX ( int value )         { MoveTo ( sf::Vector2f( (float)value, _position_Goal.y ) ); }
	void MoveToY ( int value )         { MoveTo ( sf::Vector2f( _position_Goal.x, (float)value ) ); }

	void MoveTo_Relative  ( sf::Vector2f value) { _position_Goal += value; }
	void MoveToX_Relative ( int value )         { _position_Goal.x += (float)value; }
	void MoveToY_Relative ( int value )         { _position_Goal.y += (float)value; }

	void MoveEnd () { _position_Current = _position_Goal; }


	bool IsMoveFinished () { return ( _position_Goal == _position_Current ); }
	bool IsTransitionFinished () { return ( _transitionEffect == CHANGE_IMAGE_EFFECT_NONE ); }

	void SetZ ( int z ) { _z = z; }
	int  GetZ () { return _z; }

	void FlipX () { _image.FlipX(); _newImage.FlipX();}
	void FlipY () { _image.FlipY(); _newImage.FlipY();}

	void   SetLabel ( string value ) { _label = value; }
	string GetLabel ()               { return _label; }

	void SetWipeOutSpeed ( int value ) { _wipeOutSpeed = value; }
	void SetWipeInSpeed  ( int value ) { _wipeInSpeed  = value; }
	void SetFadeOutSpeed ( int value ) { _fadeOutSpeed = value; }
	void SetFadeInSpeed  ( int value ) { _fadeInSpeed  = value; }
	
	void SetMoveForm ( string form );
	void SetMoveForm_SPRING_softness  ( float value );
	void SetMoveForm_LINEAR_increment ( float value );
	
	void SetMouseOverSeeThrough ( bool arg ) { _isMouseOverSeeThrough = arg;  }

	//=========================================================
	// セーブ/ロード用のインターフェイス
	//=========================================================
	struct SaveData {

		sf::Vector2f position_Goal;     // 目的地
		sf::Vector2f position_Current;  // 現在ただいまの表示位置
		
		int z;
		string label;

		string image_fileName; // 画像
		string maskColorString; // マスク未設定 を空文字で表したい。だからあえて、この段階で文字列化しておくのです。
		
		int rotation;     // 回転値
		sf::Vector2f scale; // スケール値

		string wipeOutImage_fileName;
		string wipeInImage_fileName;

		int wipeOutSpeed;
		int wipeInSpeed;

		int fadeOutSpeed;
		int fadeInSpeed;
		
		// 移動方式関連
		string moveForm; // 移動方式
		string moveForm_spring_softness;  // バネのゆるさ
		string moveForm_linear_increment; // 線形移動の増分 
	
	};
	SaveData GetSaveData ();
	void Load ( SaveData &saveData );

	void SetActive ( bool arg ) { _isActive = arg; }
	bool GetActive ()           { return _isActive; }
	
	string GetCurrentImageFileName();
	
	sf::Vector2f GetCurrentPosition() { return _position_Current; }
	
	void Skipping() { _isSkipping = true; }
	
private:
	
	bool _isMouseOverSeeThrough;    // マウスオーバーでの半透明機能オン/オフ
	bool _isMouseOverToSeeThrough;  // 半透明のためにマウスが乗っているところ 

	bool _isActive; // 活性状態？

	sf::Vector2f _position_Goal;     // 目的地
	sf::Vector2f _position_Current;  // 現在ただいまの表示位置

	int _z;
	string _label;

	ImageSet_as_SFML _image; // 画像
	string           _maskColorString;   // マスクカラー（セーブ/ロード時の再現用）

	///// 画像差し替え時用のあれこれ
	ImageSet_as_SFML _newImage;
	ImageSet_as_SFML _newImage_forWipeInned; // ワイプイン後の完成形。

	// ワイプ用のマスク画像
	ImageSet_as_SFML _wipeOutImage;
	ImageSet_as_SFML _wipeInImage;

	int _wipeOutAlphaLimit;
	int _wipeInAlphaLimit;

	int _wipeOutSpeed;
	int _wipeInSpeed;

	int _fadeOutSpeed;
	int _fadeInSpeed;

	// イメージ差し替え中？
	string _transitionEffect;

	// 現画像と新画像の表示順。トランジション中の都合。
	int _drawOrder;
	
	// 移動方式関連
	string _moveForm; // 移動方式
	float _moveForm_spring_softness;  // バネの柔らかさ
	float _moveForm_linear_increment; // 線形移動の増分 
	
	bool _isSkipping; // スキップ中は移動を早めるとかなんとか。

	//=========================================================
	// 定数：現画像・新画像、表示順（トランジュション中の都合）
	//=========================================================
	static const int DRAW_ORDER_CURRENT_NEW;
	static const int DRAW_ORDER_NEW_CURRENT;

	//=========================================================
	// 更新子処理：目的地へ移動
	//=========================================================
	void Update_MoveToGoal();
	void Update_MoveToGoal_SPRING();
	void Update_MoveToGoal_LINEAR();


	//=========================================================
	// トランジションエフェクト＿準備処理
	//=========================================================
	void Transition_Preparation_NONE        ( string newImageFileName );
	void Transition_Preparation_FADE_OUT_IN ( string newImageFileName );
	void Transition_Preparation_FADE_IN_OUT ( string newImageFileName );
	void Transition_Preparation_FADE_OUT_IN_SIMUL ( string newImageFileName );
	void Transition_Preparation_FADE_OUT    ( string newImageFileName );
	void Transition_Preparation_FADE_IN     ( string newImageFileName );
	void Transition_Preparation_WIPE_OUT_IN ( string newImageFileName );
	void Transition_Preparation_WIPE_OUT    ( string newImageFileName );
	void Transition_Preparation_WIPE_IN     ( string newImageFileName );

	//=========================================================
	// トランジションエフェクト＿処理中
	//=========================================================
	void Transition_Update_NONE        ();
	void Transition_Update_FADE_OUT_IN ();
	void Transition_Update_FADE_IN_OUT ();
	void Transition_Update_FADE_OUT_IN_SIMUL ();
	void Transition_Update_FADE_OUT    ();
	void Transition_Update_FADE_IN     ();
	void Transition_Update_WIPE_OUT_IN ();
	void Transition_Update_WIPE_OUT    ();
	void Transition_Update_WIPE_IN     ();

	//=========================================================
	// トランジションエフェクト＿描画、ってのも、もしかして、あるとよい？
	//=========================================================

};
#endif




