//=========================================================
//
// Main
//
//=========================================================

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <time.h>
#include "Application.h"
#include "../Utils/mt19937ar.h"
#include "../Utils/Utils_Convert.h"
#include "../Utils/Utils_File.h"
#include "../Utils/Utils_String.h"
#include <lua.hpp>
//#include <lua.h>

#include <fstream>
#include <iostream>

#if defined(_WIN32)

	#include <Windows.h>

#endif

using namespace std;




char* FormatFunc(char* buffer, const char* format, ...)
{
	va_list ap;
	va_start(ap, format);
	vsprintf(buffer, format, ap);
	va_end(ap);
	return buffer;
}

void OreOrePrintf ( const char* format, ...)
{
	string additional = "[info]";
	additional += format;

	va_list ap;
	va_start(ap, format);
	vprintf(additional.c_str(), ap);
	va_end(ap);
}

// ファイルの存在確認
bool isExistFile ( string fileName ) {
	
	bool result = false;
	
	FILE *f = fopen ( fileName.c_str(), "r" );
	if ( f == NULL ) {
		return false;
	} else {
		result = true;
	}
	
	fclose(f);
	
	return result;
}



//=========================================================
// メインここから
//=========================================================
int main(int argc,char *argv[]) {
	
	bool isConsoleMode = isExistFile ( "console" );
	
#if defined(_WIN32)

	if ( isConsoleMode ) 
	{
		// コンソールを作成する
		AllocConsole();
		
		// 標準入出力に割り当てる
		FILE* fp = NULL;
		freopen("CONOUT$", "w", stdout);
		freopen("CONIN$", "r", stdin);
	}
	
	
	printf ("defined(_WIN32)\n");

#elif defined(__linux__)

	printf ("defined(__linux__)\n");

#endif





#if defined(_WIN32)
	
	printf ("check dounble launch\n");
	
	// 多重起動防止  参考 https://www.petitmonte.com/c_cplus/capi_sample19.html
	/*
	// ミューテックスオブジェクト作成
	HANDLE hMutex = CreateMutex(NULL, TRUE, "mindll_stop_double_launch");
	
	// 既に作成されていたら終了
	if ( GetLastError() == ERROR_ALREADY_EXISTS ) {
		return FALSE;
	} else {
		printf ("OK.\n");
	}
	*/
	
	bool isFirstLaunch = true;
	
	
	// 多重起動防止： 参考 https://cjohn.hatenadiary.org/entry/20081204/1228397004
	
	// フルアクセス
	SECURITY_DESCRIPTOR sd;
	InitializeSecurityDescriptor(&sd,SECURITY_DESCRIPTOR_REVISION);
	SetSecurityDescriptorDacl(&sd, TRUE, 0, FALSE);	    
	SECURITY_ATTRIBUTES secAttribute;
	secAttribute.nLength = sizeof (secAttribute);
	secAttribute.lpSecurityDescriptor = &sd;
	secAttribute.bInheritHandle = TRUE; 

	HANDLE hMutex = ::CreateMutex(&secAttribute, FALSE, "mindll_stop_double_launch");
	
	if(hMutex == NULL)
	{
		printf ("could not get mutex.\n");
	}
	else if(::GetLastError() == ERROR_ALREADY_EXISTS)
	{
		printf ("already launched.\n");
		isFirstLaunch = false;
	}
	else
	{
		printf ("this is the first launch.\n");
		isFirstLaunch = true;
	}
	
	if ( !isFirstLaunch ) {
		
		string title = Utils_String::Trim ( Utils_File::ReadTextFile ( "title" ) );
		
		printf ( "first window title : [%s]\n", title.c_str() );
		
		HWND handle = FindWindow(NULL, title.c_str() );
		if ( handle != NULL ) {
			SetForegroundWindow ( handle );
		} else {
			printf ( "could not find the window.\n " );
		}
		
		if ( isConsoleMode ) {
		
	#if defined(_WIN32)
		
			FreeConsole();

	#elif defined(__linux__)
	#endif

		}
	
		return 0;
	}
#elif defined(__linux__)


#endif





	///// 乱数を初期化しておいてあげる。
	srand( static_cast<unsigned int>( time(NULL) ) );
	init_genrand( static_cast<unsigned int>( time(NULL) ) );
	
	
	

	Application *obj = new Application();
	obj->Run();
	delete obj;
	
	
	if ( isConsoleMode ) {
	
#if defined(_WIN32)
	
		FreeConsole();

#elif defined(__linux__)
#endif

	}

	if ( hMutex != NULL ) {
		CloseHandle(hMutex);
	}

	return 0;
}


