//=========================================================
//
// いわゆる右クリックメニューですよー。
//
//=========================================================

#include "Menu.h"
#include "Define.h"
#include "SaveDataManager.h"
#include "MenuItem.h"
#include "../GUI_SFML/TextImageFactory.h"
#include "../GUI_SFML/SFML_Utils.h"
#include "ScriptEngine.h"
#include "Application.h"
#include "SaveData.h"
#include "SceneScript.h"

using namespace GUI_SFML;

//=========================================================
// 内部定数
//=========================================================
const int Menu::MOUSE_STATE_NONE    = 0; // 押していない
const int Menu::MOUSE_STATE_PRESS   = 1; // 押している
const int Menu::MOUSE_STATE_RELEASE = 2; // 放したところ


//=========================================================
// コンストラクタ
//=========================================================
Menu::Menu( ScriptEngine *scriptEngine ) {
	
	_scriptEngine = scriptEngine;
	
	///// デフォルトでは非表示
	_isActive = false;
	
	_closeSoonAfter = false;
	
	_isConfirmBacktoTitleDisplay = false;

	///// デフォルトの見た目を用意
	_back_Frame = sf::RectangleShape (
		sf::Vector2f ( 
			APPLICATION::WINDOW_SIZE::WIDTH,
			APPLICATION::WINDOW_SIZE::HEIGHT
		)
	);
	_back_Frame.setFillColor ( sf::Color ( 50,50,50,150 ) );
	_back_Frame.setOutlineThickness (0.0f);
	_back_Frame.setOutlineColor ( sf::Color ( 0,0,0,0 ) );
	_back_Frame.setOrigin(0, 0);

	// 一応。背景画像は無効ですよと念を押しておく。
	_back_Image.fileName = "";

	_position.x = 0.0f;
	_position.y = 0.0f;


	_mouseState = MOUSE_STATE_NONE;
	
	// タイトル画面に戻るかどうかの確認ウィンドウ用
	_back_confirmWindow = sf::RectangleShape ( sf::Vector2f ( 400, 200 ) );
	_back_confirmWindow.setFillColor ( sf::Color ( 50,50,150,150 ) );
	_back_confirmWindow.setOutlineThickness (0.0f);
	_back_confirmWindow.setOutlineColor ( sf::Color ( 0,0,0,0 ) );
	_back_confirmWindow.setPosition ( sf::Vector2f ( APPLICATION::WINDOW_SIZE::WIDTH/2, APPLICATION::WINDOW_SIZE::HEIGHT/2 ) );
	_back_confirmWindow.setOrigin ( _back_confirmWindow.getGlobalBounds().width/2, _back_confirmWindow.getGlobalBounds().height/2 );
	
	_text_backToTitleConfirm = TextImageFactory::Create ( "タイトル画面に戻りますか？", 24);
	_text_backToTitleConfirm.setFillColor   ( sf::Color::White );
	_text_backToTitleConfirm.setOutlineColor( sf::Color::Black );
	_text_backToTitleConfirm.setOutlineThickness ( 1.0f );
	_text_backToTitleConfirm.setPosition ( 250, 250 );
	
	_button_yes = new Button ( "戻る",     sf::Vector2f ( 150, 40 ), 24 );
	_button_no  = new Button ( "戻らない", sf::Vector2f ( 150, 40 ), 24 );
	_button_yes->SetPosition ( sf::Vector2f ( 235, 320 ) );
	_button_no ->SetPosition ( sf::Vector2f ( 405, 320 ) );
	
	
	_button_Save        = new Button ( "SAVE", sf::Vector2f ( 200, 40 ), 24 );
	_button_Load        = new Button ( "LOAD", sf::Vector2f ( 200, 40 ), 24 );
	_button_backToTitle = new Button ( "タイトル画面へ戻る", sf::Vector2f ( 200, 40 ), 20 );

	_button_Save->SetPosition        ( sf::Vector2f ( 75, 70 ) );
	_button_Load->SetPosition        ( sf::Vector2f ( 75, 120 ) );
	_button_backToTitle->SetPosition ( sf::Vector2f ( 75, 170 ) );
	
	{
		_text_Volume_BGM = TextImageFactory::Create ( "[BGM音量]", 18 );
		_text_Volume_BGM.setPosition ( 75, 240 - _text_Volume_BGM.getLocalBounds().height / 2 );

		Slider::InitProp initProp;
		initProp.length = 200.0f;
		initProp.range_min = 0;
		initProp.range_max = 100;
		initProp.position = sf::Vector2f ( 75, 265 );
		_slider_Volume_BGM = new Slider (initProp);
		
		
	}
	{
		_text_Volume_SE = TextImageFactory::Create ( "[効果音音量]", 18 );
		_text_Volume_SE.setPosition ( 75, 310 - _text_Volume_BGM.getLocalBounds().height / 2 );

		Slider::InitProp initProp;
		initProp.length = 200.0f;
		initProp.range_min = 0;
		initProp.range_max = 100;
		initProp.position = sf::Vector2f ( 75, 335 );
		_slider_Volume_SE = new Slider (initProp);

	}
	{
		_shortcutKeys_Image.fileName = "./image/menu/shortcutKeys.png";
		_shortcutKeys_Image.texture.loadFromFile(_shortcutKeys_Image.fileName);
		_shortcutKeys_Image.sprite.setTexture(_shortcutKeys_Image.texture);
		_shortcutKeys_Image.sprite.setOrigin(0,0);
		_shortcutKeys_Image.sprite.setPosition(0,0);

	}

	// SaveDataManager から ScriptEngine のSave/Load を呼ぶために、ポインタをはるばると受け渡す。
	_saveDataManager = new SaveDataManager ( _scriptEngine );



	{
		vector< pair<string, string>> optionList;
		{
			optionList.push_back ( pair<string, string>( "ON", "on" ) );
			optionList.push_back ( pair<string, string>( "OFF", "off" ) );
		}

		MenuItem *item = new MenuItem( "IsSkip", optionList, "[既読早送り]");

		item->SetValue("off");
		item->SetPosition ( sf::Vector2f( 75, 370 ) );

		_itemList.push_back ( item );
	}
}

//=========================================================
// デストラクタ
//=========================================================
Menu::~Menu() {

	delete _button_Save;
	delete _button_Load;
	delete _button_backToTitle;
	delete _button_yes;
	delete _button_no;
	
	
	delete _saveDataManager;
	
	delete _slider_Volume_BGM;
	delete _slider_Volume_SE;

	for ( unsigned int i = 0; i < _itemList.size(); ++i) {
		delete _itemList[i];
	}
}


//=========================================================
// 更新
//=========================================================
void Menu::Update ( sf::RenderWindow &window ) {

	// 非表示なら、何もしませんですよ。
	if ( !_isActive ) {
		return;
	}
	
	// TODO カメラに合わせて位置を調整する処理。さしあたり、メニューの背景にだけ適用している。
	//sf::Vector2f screenPos = Application::GetInstance()->_window.mapPixelToCoords( sf::Vector2i(_position.x, _position.y) );

	//int x = _position.x;
	//int y = _position.y;
	//SFML_Utils::PosToScreenPos ( x, y, window );
	//sf::Vector2f screenPos ( x, y );
	
	// セーブデータ一覧が非表示になったらメニューも閉じるの？
	if ( _saveDataManager->GetMode() == SaveDataManager::MODE_NONE && _closeSoonAfter ) {
		
		_isActive = false;
	
	// セーブデータ一覧が非表示であります！
	} else if ( _saveDataManager->GetMode() == SaveDataManager::MODE_NONE ) {
		
		
		// TODO カメラ移動に合わせて位置を変えているところ。ここを参考に、その他の部品にも情報を伝えることとしましょう。必要になったら。
		_back_Frame.setPosition ( _position );
		_back_Image.sprite.setPosition ( _position );
		//_back_Frame.setPosition ( screenPos );
		//_back_Image.sprite.setPosition ( screenPos );


		///// タイトルに戻るの確認ウィンドウが出ている場合
		if ( _isConfirmBacktoTitleDisplay ) {
			
			_button_yes -> Update(window);
			_button_no  -> Update(window);

			///// NOボタン押下？
			if ( _button_no->IsRelease() ) {
				printf("タイトルに戻らない押下\n");
				_isConfirmBacktoTitleDisplay = false;
			}

			///// YESボタン押下？
			if ( _button_yes->IsRelease() ) {
				printf("タイトルに戻る押下\n");
				_isConfirmBacktoTitleDisplay = false; // 次回メニューを開いたときに確認ウィンドウが閉じているように
				
				// 旧方式
				//SceneScript::BACK_TO_TITLE = 1;

				string fileName   = "./data/save/qsave/backToTitle.sav";
				//string fileName   = "./data/save/qsave/qsave.sav";
				
				SaveData *saveDataToBackTotitle = new SaveData ( fileName, 0 ); // 第２引数は「スロット番号」

				_scriptEngine->LoadFrom ( saveDataToBackTotitle, "タイトルに戻りました" );
			}

		
		///// タイトルに戻るの確認ウィンドウが出ていない場合
		} else {

			_button_Save->Update ( window );
			_button_Load->Update ( window );
			_button_backToTitle->Update ( window );

			///// セーブボタン押下？
			if ( _button_Save->IsRelease() ) {
				printf("セーブボタン押下\n");
				_saveDataManager->SetMode ( SaveDataManager::MODE_SAVE );
				_closeSoonAfter = false;
			}

			///// ロードボタン押下？
			if ( _button_Load->IsRelease() ) {
				printf("ロードボタン押下\n");
				_saveDataManager->SetMode ( SaveDataManager::MODE_LOAD );
				_closeSoonAfter = false;
			}
			
			///// タイトルに戻るボタン押下？
			if ( _button_backToTitle->IsRelease() ) {
				
				printf ( "タイトルに戻るボタンが押された。\n" );
				
				_isConfirmBacktoTitleDisplay = true;
				
			}
			
			///// 音量：BGM
			int volumeBefore = 0; 

			volumeBefore = _slider_Volume_BGM->GetCurrentValue();
			_slider_Volume_BGM->Update ( window );
			
			if ( volumeBefore != _slider_Volume_BGM->GetCurrentValue() ) {
				_scriptEngine->SetVolumeBGM ( _slider_Volume_BGM->GetCurrentValue() );
			}
			
			///// 音量：SE
			volumeBefore = _slider_Volume_SE->GetCurrentValue();
			
			_slider_Volume_SE->Update ( window );
			
			if ( volumeBefore != _slider_Volume_SE->GetCurrentValue() ) {
				_scriptEngine->SetVolumeSE ( _slider_Volume_SE->GetCurrentValue() );
			}
			
			///// ショートカットキー一覧
			
			
			// マウスの状態を監視：右クリックの状態を
			ObserveMouseState ( window );

			// 右クリックで、非表示になる。
			if ( _mouseState == MOUSE_STATE_RELEASE ) {
				SetActive ( false );
			}

			///// 設定項目リスト
			for ( unsigned int i = 0; i < _itemList.size(); ++i) {
				_itemList[i]->Update ( window );
			}
		} 


	// セーブデータ一覧が表示されている
	} else {

		_saveDataManager->Update ( window );
	}

	
}

//=========================================================
// 描画
//=========================================================
void Menu::Draw   ( sf::RenderWindow &window ) {

	// 非表示なら、何もしませんですよ。
	if ( !_isActive ) {
		return;
	}
	
	sf::View defaultView = window.getDefaultView();
	sf::View currentView = window.getView();
	
	sf::Transform defaultTransForm = defaultView.getTransform();
	sf::Transform currentTransForm = currentView.getTransform();
	
	float defaultRotation = defaultView.getRotation();
	float currentRotarion = currentView.getRotation();
	
	// 回転は反映させて、位置はデフォルトのまま
	defaultView.setRotation ( currentRotarion );
	window.setView ( defaultView );
	
	// 背景画像が設定されていれば、そっちを優先。
	if ( _back_Image.fileName != "" ) {

		window.draw ( _back_Image.sprite );

	// 背景画像が設定されていないときは、フレームを使う。
	} else {

		window.draw ( _back_Frame );
	}


	///// セーブデータ一覧が非表示
	if ( _saveDataManager->GetMode() == SaveDataManager::MODE_NONE ) {
		
		///// ショートカットキー一覧 兼 下地
		window.draw(_shortcutKeys_Image.sprite);

		///// ボタン。
		_button_Save->Draw ( window );
		_button_Load->Draw ( window );
		_button_backToTitle->Draw(window);

		///// 設定項目リスト
		for ( unsigned int i = 0; i < _itemList.size(); ++i) {
			_itemList[i]->Draw ( window );
		}
		
		///// 音量調整
		_slider_Volume_BGM->Draw ( window );
		_slider_Volume_SE ->Draw ( window );
		
		window.draw(_text_Volume_BGM);
		window.draw(_text_Volume_SE);
		
		
		if ( _isConfirmBacktoTitleDisplay ) {
			
			window.draw ( _back_confirmWindow );
			window.draw ( _text_backToTitleConfirm );
			_button_yes->Draw ( window );
			_button_no->Draw ( window );
		}
		

	///// セーブデータリスト表示中
	} else {
		_saveDataManager->Draw ( window );
	}
	
	window.setView ( currentView );
}



//=========================================================
// マウスの状態を監視
//=========================================================
void Menu::ObserveMouseState ( sf::RenderWindow &window ) {

	///// ウィンドウが最前面じゃないときには、マウス操作を受け付けない。
	/*
	HWND handle     =  window.getSystemHandle();
	HWND foreWindow = GetForegroundWindow();

	bool isTopWindow = ( handle == foreWindow );
	*/
	bool isTopWindow = Application::GetInstance()->HasFocus();

	if ( !isTopWindow ) {
		_mouseState = MOUSE_STATE_NONE;
		return;
	}


	if ( _mouseState == MOUSE_STATE_NONE ) {

		if ( sf::Mouse::isButtonPressed ( sf::Mouse::Right ) ) {
			_mouseState = MOUSE_STATE_PRESS;
		}
	}
	else if ( _mouseState == MOUSE_STATE_PRESS ) {

		if ( !sf::Mouse::isButtonPressed ( sf::Mouse::Right ) ) {
			_mouseState = MOUSE_STATE_RELEASE;
		}
	}
	else if ( _mouseState == MOUSE_STATE_RELEASE ) {

		_mouseState = MOUSE_STATE_NONE;
	}
}
//=========================================================
// 表示/非表示
//=========================================================
void Menu::SetActive ( bool value ) { 

	// セーブデータ一覧画面も閉じておく。
	_isActive = value; 
	_saveDataManager->SetMode ( SaveDataManager::MODE_NONE );
	_closeSoonAfter = false;
}

//=========================================================
// セーブ/ロードウィンドウを開く
//=========================================================
void Menu::OpenSaveWindow() {
	
	_isActive = true;
	_closeSoonAfter = true;
	_saveDataManager->SetMode ( SaveDataManager::MODE_SAVE );
	
}
void Menu::OpenLoadWindow() {
	
	_isActive = true;
	_closeSoonAfter = true;
	_saveDataManager->SetMode ( SaveDataManager::MODE_LOAD );
}
	
//=========================================================
// 設定項目の値を取得
//=========================================================
string Menu::GetItemValue ( string itemName ) {

	string result = "";

	for ( unsigned int i = 0; i < _itemList.size(); ++i ) {

		if ( _itemList[i]->GetName() == itemName ) {
			result = _itemList[i]->GetValue();
			break;
		}
	}

	return result;
}

//=========================================================
// 設定項目に値を設定
//=========================================================
void Menu::SetItemValue ( string itemName, string value ) {

	for ( unsigned int i = 0; i < _itemList.size(); ++i ) {

		if ( _itemList[i]->GetName() == itemName ) {
			_itemList[i]->SetValue(value);
			break;
		}
	}
}

int Menu::GetVolumeBGM () {

	return _slider_Volume_BGM->GetCurrentValue();
}

void  Menu::SetVolumeBGM ( int value ) {
	_slider_Volume_BGM->SetCurrentValue ( value );
}

int Menu::GetVolumeSE () {

	return _slider_Volume_SE->GetCurrentValue();
}

void  Menu::SetVolumeSE ( int value ) {
	_slider_Volume_SE->SetCurrentValue ( value );
}







