//=========================================================
//
// メニューでの設定項目
//
//=========================================================

#include "Application.h"
#include "MenuItem.h"
#include "MenuItemOption.h"
#include "Define.h"
#include "../GUI_SFML/TextImageFactory.h"


//=========================================================
// ラベルと値のセットを受け取る
//【引数】
// この項目の名前と、選択肢のリスト
//=========================================================
MenuItem::MenuItem( string name, vector < pair <string, string> > initData, string label ) {

	_name = name;

	for ( unsigned int i = 0; i < initData.size(); ++i ) {

		string label = initData[i].first;
		string value = initData[i].second;

		MenuItemOption *option = new MenuItemOption ( label, value );

		_optionList.push_back ( option );
	}


	//TextImageFactory::Create ( _imageSet.image, label, 20, sf::Color(255,255,255,255) );
	//_imageSet.Update();
	
	_label = TextImageFactory::Create ( label, 18 );
	
	//_label.setOutlineThickness( 1.0f );
	//_label.setOutlineColor (sf::Color::Black );
	_label.setFillColor    ( sf::Color::White );

}

//=========================================================
// デストラクタ
//=========================================================
MenuItem::~MenuItem() {

	for ( unsigned int i = 0; i < _optionList.size(); ++i ) {

		delete _optionList[i];
	}
}

//=========================================================
// 更新
//=========================================================
void MenuItem::Update ( sf::RenderWindow &window ) {

	ObserveOptionsSelectState();

	sf::Vector2f cursorPos = _position;

	//_imageSet.sprite.setPosition ( cursorPos );
	
	_label.setPosition( cursorPos );

	//cursorPos.x += _imageSet.image.getSize().x + 10.0f;
	cursorPos.x += _label.getGlobalBounds().width + 10.0f;
	cursorPos.y += 2.0f; // 適当

	for ( unsigned int i = 0; i < _optionList.size(); ++i ) {

		_optionList[i]->SetPosition ( cursorPos );

		_optionList[i]->Update(window);

		cursorPos.x += _optionList[i]->GetSize().x + 5.0f;


	}
}

//=========================================================
// 描画
//=========================================================
void MenuItem::Draw   ( sf::RenderWindow &window ) {

	//window.draw ( _imageSet.sprite );
	window.draw(_label);

	for ( unsigned int i = 0; i < _optionList.size(); ++i ) {
		_optionList[i]->Draw(window);
	}
}


//=========================================================
// 値をセット。
//=========================================================
void MenuItem::SetValue ( string arg ) {

	for ( unsigned int i = 0; i < _optionList.size(); ++i ) {
		
		_optionList[i]->SetSelected ( ( _optionList[i]->GetValue() == arg ) );
	}
}

//=========================================================
// 選択されている項目の値を取得
//=========================================================
string MenuItem::GetValue() {

	string result = "";

	for ( unsigned int i = 0; i < _optionList.size(); ++i ) {

		if (_optionList[i]->IsSelected() ) {
			result = _optionList[i]->GetValue();
			break;
		}

	}

	return result;
}

//=========================================================
// 選択肢の選択状態を監視
//=========================================================
void MenuItem::ObserveOptionsSelectState () {

	///// マウスがどうにもなってなければ、何もしない。
	if (Application::GetInstance()->_mouseState_Left != MOUSE::BUTTON::STATE::RELEASE) {
		return;
	}

	///// この時点でマウスオーバーしてる選択肢が選択されるのです。
	int selectIndex = -1;
	for ( unsigned int i = 0; i < _optionList.size(); ++i ) {
		if ( _optionList[i]->IsMouseOver() ) {
			selectIndex = i;
			break;
		}
	}

	///// なんか選択したなら、それを選択状態にする。それ以外を選択じゃない状態にする。
	if ( selectIndex != -1 ) {
		for ( unsigned int i = 0; i < _optionList.size(); ++i ) {
			_optionList[i]->SetSelected (( selectIndex == i ));
		}
	}
}

