//=========================================================
//
// スクリプトエンジン（フリーバトル選択タグ）
//
//=========================================================

#include "Application.h"
#include "ScriptEngine.h"
#include "ScriptProcess.h"
#include "ScriptUnit.h"
#include "Image.h"
#include "GlobalSaveData.h"

#include "../Utils/NotXML.h"
#include "../Utils/Utils_Convert.h"
#include "../Utils/Utils_String.h"
#include "../Utils/Utils_Others.h"
#include "../GUI_SFML/SFML_Utils.h"
#include "../GUI_SFML/TextMessageWindow.h"

#include "FreeBattleSelector.h"


using namespace std;
using namespace GUI_SFML;


//=========================================================
// ユニット開始処理：フリーバトル選択タグ開始
//=========================================================
void ScriptEngine::UnitStart_FreeBattleSelector ( ScriptUnit &unit ) {

	///// 名前をもとに既存のボタンを取得。なければ新規に生成。あればそれに対して操作を行う。
	string name = NotXML::GetAttribute ( unit.unitTag, "name" );
	
	
	///// この後、待ち受けに入るのか、それとも各種定義だけで終了するのか
	bool isGoingToStartWaiting = false;
	
	////  なければ作成
	if ( _freeBattleSelector == NULL ) {
		_freeBattleSelector = new FreeBattleSelector ( this );
	}
	
	
	///// 内部のタグでループ
	for ( unsigned int i = 0; i < unit.unitTag.childList.size(); ++i) {

		NotXML::Tag tag = unit.unitTag.childList[i];

		///// 待ち受けに入るのかどうか：このタグがあるかないか、だけで設定
		if ( tag.name == "WaitStart" ) {
			
			isGoingToStartWaiting = true;
			
			_freeBattleSelector->SetActive(true);
			_freeBattleSelector->ClearSelectStatus();
		
		}
		///// サイズ
		else if ( tag.name == "Matrix" ) {

			int hCount        = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( tag, "h", "1") );
			int vCount        = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( tag, "v", "1") );
			int displayVCount = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( tag, "displayVCount", "1") );
			
			_freeBattleSelector->SetMatrix ( hCount, vCount, displayVCount );
		}
		///// 結果を入れる変数名前
		else if ( tag.name == "Variable" ) {
			
			string variableName = tag.value;
			
			_freeBattleSelector->SetVariableForResult(variableName);
		}
		///// 個別のフリーバトル
		else if ( tag.name == "FreeBattle" ) {
			
			FreeBattleIcon::FreeBattleIconProp prop;
			
			int gridX = 0;
			int gridY = 0;
			
			for ( unsigned int n = 0; n < tag.childList.size(); ++n) {

				NotXML::Tag ctag = tag.childList[n];
				
				// ID
				if ( ctag.name == "Id" ) {
					prop.id = ctag.value;
				}
				// セレクター内での表示位置
				else if ( ctag.name == "Grid" ) {

					gridX = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( ctag, "x", "0") );
					gridY = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( ctag, "y", "0") );
				}
				// シークレット枠で指定
				else if ( ctag.name == "SecretFrame" ) {
					
					prop.isSecret     = true;
					prop.secretX      = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( ctag, "x", "0") );
					prop.secretY      = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( ctag, "y", "0") );
					prop.secretWidth  = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( ctag, "width", "0") );
					prop.secretHeight = Utils_Convert::String_to_Int ( NotXML::GetAttribute ( ctag, "height", "0") );
				}
				// アンロックのためのフラグ名
				else if ( ctag.name == "UnlockFlagName" ) {
					prop.variableForUnlock = ctag.value;
				}
				// 試合名
				else if ( ctag.name == "BattleTitle" ) {
					
					prop.battleTitle = ctag.value;

					TextMessageWindow::ReplaceWithAlias(prop.battleTitle);
					
				}
				// アイコン画像
				else if ( ctag.name == "Image" ) {
					prop.imageFile = ctag.value;
				}
				// ロックアイコン
				else if ( ctag.name == "LockedIconImage" ) {
					prop.lockedIconImageFile = ctag.value;
				}
				
			}
			
			_freeBattleSelector->AddIcon(prop, gridX, gridY);
		
		}
	}

	
	///// この後、待ち受けモードに入るのかどうか
	if ( isGoingToStartWaiting ) {
	
		_state = ScriptEngine::STATE_WAIT_UNTIL_UNIT_END;
	
	} else {
		// 処理終了
		_state = ScriptEngine::STATE_CURRENT_UNIT_FINISHED;
	}
	
	
	// 今回は描画しなくていいですよね？
	//_isRestDraw = true;
	
}




//=========================================================
// ユニット実行中：クリック画像・クリック待ち受け
//=========================================================
void ScriptEngine::UnitProcessing_FreeBattleSelector ( ScriptUnit &unit ) {
	
	// 選択されていたら、選択されているものの名前を取得して、指定の変数に入れて終了
	if ( _freeBattleSelector->IsSelected() ) {
		
		///// 選択されているもののIDを取得
		string selectedId = _freeBattleSelector->GetSelectedId();
		
		printf ( "フリーバトル選択されました。：%s\n", selectedId.c_str() );
		
		
		///// 結果を入れる変数名を取得
		string variableForResult = _freeBattleSelector->GetVariableForResult();
		
		///// ローカル変数に結果を入れる
		SetLocalString ( variableForResult, selectedId );
		
		///// セレクターはこれにて非表示。 
		_freeBattleSelector->SetActive(false);
		_freeBattleSelector->ClearSelectStatus();
		
		///// 待ち受け完了
		_state = ScriptEngine::STATE_CURRENT_UNIT_FINISHED;
		
	} else {
		
		if ( !_freeBattleSelector->IsActive() ) {
			
			// おそらく、フリーバトル選択画面でセーブ/ロードしたものと思われる。
			printf ( "フリーバトル待ち受け中なのに非表示なので、SetActive(true)しますよ。\n" );
			_freeBattleSelector->SetActive(true);
			
		}
		
	}

}



//=========================================================
// フリーバトルが解放されているかどうか。フリーバトル専用のドロナワメソッド。
//=========================================================
bool ScriptEngine::IsUnlockedFreeBattle ( string variableName ) {
	
	bool result = false;
	
	///// グローバル変数を見ます。
	if ( _variable_Global_Int.find( variableName ) != _variable_Global_Int.end() ) {
	
		int value = _variable_Global_Int.find( variableName )->second;
		
		if ( value == 1 ) {
			result = true;
		}
		
	}
	
	return result;
}





