//=========================================================
//
// スクリプトエンジン
//
//=========================================================

#include "Application.h"
#include "ScriptEngine.h"
#include "ScriptProcess.h"
#include "ScriptUnit.h"

#include "ScriptButton.h"
#include "Game_Utils.h"
#include "GameManager.h"
#include "Image.h"
#include "ClickableImage.h"

#include "../Utils/NotXML.h"
#include "../Utils/Utils_String.h"

using namespace std;
using namespace GUI_SFML;


//=========================================================
// ユニット開始処理：検討モード
//=========================================================
void ScriptEngine::UnitStart_AstralProjection( ScriptUnit &unit ) {
	
	printf("検討モード開始\n");

	///// 最低限あるはずのもの：どのゲーム盤？
	string gameManagerName = NotXML::GetTagFromName ( unit.unitTag, "Game" ).value;

	///// ゲーム盤、ないとは言わせませんよ？
	GameManager *gameManager = GameManager_Get ( gameManagerName );

	// え、ないの？ マジで？ おはなしになりませんね。
	if ( gameManager == NULL ) {
		_state = ScriptEngine::STATE_CURRENT_UNIT_FINISHED;
		printf ( "Warning：ゲーム盤がないので検討モードにできません：%s\n", gameManagerName.c_str() );
		return;
	}
	
	// 検討モードの間、非表示にしておく画像（検討モード終了時に元に戻すためのメモ）
	_hideImageLabelList_ForAstralProjection.clear();
	
	// 終了ボタンはこちら側（ScriptEngine側）で担当するのです。
	if ( _endAstralProjectionButton != NULL ) {
		delete _endAstralProjectionButton;
	}
	
	ClickableImage::Property endButtonProp;
		endButtonProp.imageFile   = "./image/gameBoard/endAstralProjection.png";
		endButtonProp.size        = sf::Vector2f(120.0f, 32.0f);
		endButtonProp.rect_Normal = sf::IntRect(0, 0, 120, 32);
		endButtonProp.rect_Over   = sf::IntRect(0, 32, 120, 32);
		endButtonProp.rect_Push   = sf::IntRect(0, 64, 120, 32);
	_endAstralProjectionButton = new ClickableImage(endButtonProp);
	
	// 終了ボタンの位置：真ん中版
	//_endAstralProjectionButton->SetPosition ( sf::Vector2f(220.0f, 520.0f) );
	
	// 終了ボタンの位置：右下版
	_endAstralProjectionButton->SetPosition ( sf::Vector2f(400.0f, 540.0f) );

	// ゲーム盤の方での準備処理
	gameManager->AstralProjectionStart();

	///// 子要素さんこんにちわ。
	for ( unsigned int i = 0; i < unit.unitTag.childList.size(); ++i ) {

		NotXML::Tag tag = unit.unitTag.childList[i];

		// 検討モードの間、非表示にしておく画像
		if ( tag.name == "HideImage" ) {
			
			Image *image = Image_Get ( tag.value );
			if ( image != NULL ) {
				
				if ( image->GetActive() ) {
					
					// 非表示にして
					image->SetActive(false);
					
					// 覚えておく。検討モード終了時に元に戻す。
					_hideImageLabelList_ForAstralProjection.push_back ( tag.value );
				}
			}

		}

	}




	///// 処理開始するよ！ 
	_state = ScriptEngine::STATE_WAIT_UNTIL_UNIT_END;
}


//=========================================================
// ユニット実行中（終了を待機）：検討モード
//=========================================================
void ScriptEngine::UnitProcessing_AstralProjection( ScriptUnit &unit ) {

	///// えっと、どのゲーム盤、でしたっけ？
	string gameManagerName = NotXML::GetTagFromName ( unit.unitTag, "Game" ).value;

	GameManager *gameManager = GameManager_Get ( gameManagerName );
	if ( gameManager == NULL ) {
		printf ( "Error：ゲーム盤がないので検討できません。 %s\n", _battleData.gameManagerName.c_str());
		_state = ScriptEngine::STATE_CURRENT_UNIT_FINISHED;	
	}

	///// ゲーム盤の方での更新を進めてね。
	gameManager->Update_AstralProjection();
	
	///// つまり、盤面を作り直したときに一瞬表示がブレるのを防ぎたい。通常の更新のところでもこのフラグを見ているのは形作り。
	///// GameManager_Update ( window );を参照のこと
	if ( !gameManager->IsItOkayToBeRendered() ) {
		_isRendered = false;
	}

	///// 終わった？
	//if ( sf::Keyboard::isKeyPressed(sf::Keyboard::Space) ) {
	if ( _endAstralProjectionButton->IsClick() ) {		
		
		printf ( "検討モード終了シークエンス1\n" );
		
		// ゲーム盤の側でも、終了処理をする
		gameManager->AstralProjectionEnd();

		printf ( "検討モード終了シークエンス2\n" );
		
		// 非表示にしていた画像（だけ）を元に戻す。この方式なら、もともと非表示だったものまで表示されてしまうことはない。
		// TODO 試合タグでもこうすることにしましょう。
		for ( unsigned int i = 0; i < _hideImageLabelList_ForAstralProjection.size(); ++i ) {

			Image *image = Image_Get ( _hideImageLabelList_ForAstralProjection[i] );
			if ( image != NULL ) {
				image->SetActive(true);
			}
		}

		printf ( "検討モード終了シークエンス3\n" );
		
		// 終了ボタンを消しておくよ
		delete _endAstralProjectionButton;
		_endAstralProjectionButton = NULL;
		
		printf ( "検討モード終了シークエンス4\n" );

		///// 検討タグ終了
		_state = ScriptEngine::STATE_CURRENT_UNIT_FINISHED;	
	}
}







