//=========================================================
//
// 衝撃波
//
//=========================================================

#include "Effect_ImpactWave.h"
#include "../GUI_SFML/SoundPlayer.h"


//=========================================================
// コンストラクタ
//=========================================================
Effect_ImpactWave::Effect_ImpactWave ( Property prop ) {

	_prop = prop;
	_isEnd = false;
	_frameCount = 0;

	_circleShape = sf::CircleShape (1, 50);
	_circleShape.setPosition ( _prop.position );
	_circleShape.setOutlineColor ( _prop.color );
	_circleShape.setFillColor    ( _prop.color );

	_phase = PHASE_CIRCLE_SPREADS;
}

//=========================================================
// デストラクタ
//=========================================================
Effect_ImpactWave::~Effect_ImpactWave() {

}

//=========================================================
// 更新
//=========================================================
bool Effect_ImpactWave::Update(sf::RenderWindow &window) {

	if ( _isEnd ) {
		return _isEnd;
	}

	if ( _frameCount == 0 ) {

		if ( _prop.soundName != "" ) {
			GUI_SFML::SoundPlayer::Play ( _prop.soundName );
		}

	}

	++_frameCount;

	///// フェイズ：円を広げる
	if ( _phase == PHASE_CIRCLE_SPREADS ) {

		int thickness = _circleShape.getOutlineThickness ();
		if ( thickness > _prop.maxRadius ) {

			_phase = PHASE_CIRCLE_DISAPPEAR; 

			_circleShape.setFillColor ( sf::Color(0,0,0,0) );
			_circleShape.setOutlineColor ( _prop.color );

		} else {

			_circleShape.setOutlineThickness ( thickness + _prop.speed );

		}
	}
	///// フェイズ：円を消す
	else if ( _phase == PHASE_CIRCLE_DISAPPEAR ) {

		int currentRadius = _circleShape.getRadius();
		int thickness = _circleShape.getOutlineThickness ();

		currentRadius += _prop.speed;

		if ( currentRadius > _prop.maxRadius ) {

			_isEnd = true;

		} else {

			_circleShape.setRadius ( currentRadius );

			_circleShape.setOrigin ( 
				 (_circleShape.getGlobalBounds().width / 2) - _circleShape.getOutlineThickness(), 
				 (_circleShape.getGlobalBounds().height / 2) - _circleShape.getOutlineThickness()
			);
			

		}
	}

	return _isEnd;
}

//=========================================================
// 描画
//=========================================================
void Effect_ImpactWave::Draw(sf::RenderWindow &window) {


	window.draw(_circleShape);
}

