//=========================================================
//
// 着手エフェクト＿駒移動（ぶつける）
//
//=========================================================

#ifndef _EFFECT_MOVE_PIECE_COLLIDE_
#define _EFFECT_MOVE_PIECE_COLLIDE_

#include "IEffect.h"
#include "Delegate.h"
#include "../GUI_SFML/Animation.h"
#include "../AI/Move.h"

#include <vector>

class Piece;
class PieceBox;
class Grid;

class Effect_MovePiece_Collide : public IEffect {

public:
	
	static Piece* LAST_COLLIDED_PIECE;
	
	///// オフセットアニメーション用
	struct OffsetAnimationFrame {

		int x;
		int y;
		int frameCount_Max;
		int frameCount_Current;

		OffsetAnimationFrame () {
			x = 0;
			y = 0;
			frameCount_Max     = 0;
			frameCount_Current = 0;
		}
		OffsetAnimationFrame ( int arg_x, int arg_y, int arg_frame_Max) {
			x = arg_x;
			y = arg_y;
			frameCount_Max     = arg_frame_Max;
			frameCount_Current = 0;
		}
	};

	///// オフセットアニメーション用
	struct OffsetAnimation {
		vector < OffsetAnimationFrame > offsetList;
		int index;

		OffsetAnimation() {
			index = -1;
		}
	};

	Effect_MovePiece_Collide (
		
		DelegateBase     *endMethod,
		AI_Modules::Move  move,
		int               effectLevel,
		Grid             *grid_From,      // このグリッドから
		Grid             *grid_To,        // このグリッドへ移動
		Piece            *piece,          // この駒を移動
		Piece            *piece_Collided, // この駒にぶつける
		PieceBox         *pieceBox,       // ぶつけられた駒が戻る駒台
		Grid             *grid_Collided   // ぶつけられる駒がいるグリッド
	);


	~Effect_MovePiece_Collide();

	bool Update(sf::RenderWindow &window);
	void Draw(sf::RenderWindow &window);

private:

	bool _isEnd;

	///// 着手終了時に呼ぶメソッド。GameManager側で用意する。
	DelegateBase *_endMethod;

	///// 監視対象のオブジェクト
	AI_Modules::Move _move;       // つまりこの着手ですよ。
	Grid        *_grid_From;      // このグリッドから
	Grid        *_grid_To;        // このグリッドへ移動
	Piece       *_piece;          // この駒を移動
	Piece       *_piece_Collided; // この駒にぶつける
	PieceBox    *_pieceBox;       // ぶつけられた駒が戻る駒台
	Grid        *_grid_Collided;  // ぶつけられる駒がいるグリッド

	///// アニメーションエフェクト
	//GUI_SFML::Animation *_animation_TakeCardWind;   // カードを取得時の竜巻
	//GUI_SFML::Animation *_animation_CardArrivesBox; // カードがカード置き場に到着
	GUI_SFML::Animation *_animation_Collide; // ぶつかりエフェクト
	GUI_SFML::Animation *_animation_Sight;   // 照準

	///// オフセットアニメーション
	//vector < OffsetAnimationFrame > _offsetList_For2_SoarCard; // カード取得時、上空に飛ばされる
	//int _offsetList_PutCard_index;
	//OffsetAnimation _offsetAnimation_For2_SoarCard; // カード取得時、上空に飛ばされる

	///// エフェクトレベルはいくつか？
	int _effectLevel;

	///// 今の状態
	int _currentState;


	///// 使い捨てフレームカウント
	int _frameCount_Disposable;

	//=========================================================
	// 動作のステージ
	//=========================================================
	static const int STATE_0_BEGGINING_TO_GO; // 駒が動き始める
	static const int STATE_0_GOING;           // 駒が動いている
	static const int STATE_0_ARRIVE_AT;       // 駒が移動先に到着した

	static const int STATE_1_BEGGINING_TO_GO; // 駒が動き始める
	static const int STATE_1_GOING;           // 駒が動いている
	static const int STATE_1_ARRIVE_AT;       // 駒が移動先に到着した

	static const int STATE_2_BEGGINING_TO_AIM; // 照準エフェクトを出す
	static const int STATE_2_AIMING;           // 照準エフェクト進行中
	static const int STATE_2_SHOOT;            // 発射
	static const int STATE_2_GOING;            // 駒が動いている
	static const int STATE_2_ARRIVE_AT;        // 駒が移動先に到着した

	static const int STATE_3_CHARGE;          // 駒がタメを作る
	static const int STATE_3_WAIT_TO_GO;      // たまった状態で待つ
	static const int STATE_3_GOING;           // 駒が動いている
	static const int STATE_3_ARRIVE_AT;       // 駒が移動先に到着した

	static const int STATE_FINISHED; // レベル共通：終了している。


	//=========================================================
	// オフセットリスト生成
	//=========================================================
	//void MakeOffsetList_For_0();
	//void MakeOffsetList_For_3();
	//void MakeOffsetList_For_2_SoarCard();


	//=========================================================
	// エフェクトレベルごとに更新処理を分ける
	//=========================================================
	void Update_EffectLevel_0 (sf::RenderWindow &window);
	void Update_EffectLevel_1 (sf::RenderWindow &window);
	void Update_EffectLevel_2 (sf::RenderWindow &window);
	void Update_EffectLevel_3 (sf::RenderWindow &window);

	//=========================================================
	// エフェクトレベルごとに描画処理を分ける
	//=========================================================
	void Draw_EffectLevel_0 (sf::RenderWindow &window);
	void Draw_EffectLevel_1 (sf::RenderWindow &window);
	void Draw_EffectLevel_2 (sf::RenderWindow &window);
	void Draw_EffectLevel_3 (sf::RenderWindow &window);


	//=========================================================
	// 駒の現在地に旋風エフェクトを出す
	//=========================================================
	void WhirlWind();
};

#endif
