//=========================================================
//
// カード設置エフェクト。AI用
//
//=========================================================

#include "Effect_PutCard_AI.h"
#include "Card.h"
#include "CardBox.h"
#include "Board.h"
#include "Grid.h"
#include "../GUI_SFML/SoundPlayer.h"
#include "../Utils/Utils_Cal.h"
#include "EffectList.h"
#include "Effect_Animation.h"
#include "../GUI_SFML/Animation.h"
#include "Application.h"
#include "EffectManager.h"


//=========================================================
// カード設置動作のステージ
//=========================================================
const int Effect_PutCard_AI::STATE_PICK_UP           = 0; // カードをカード台から持ち上げた
const int Effect_PutCard_AI::STATE_GOING_TO_THE_GRID = 1; // カードが設置先グリッドに向かっている
const int Effect_PutCard_AI::STATE_AT_THE_GRID       = 2; // カードが設置先グリッドに到着した
const int Effect_PutCard_AI::STATE_FINISHED          = 3; // カードをグリッドに設置し終えた


//=========================================================
// コンストラクタ：必要なオブジェクトを受け取る
//=========================================================
Effect_PutCard_AI::Effect_PutCard_AI ( 
		DelegateBase     *endMethod,   // 終了時の通知先
		AI_Modules::Move  move,        // この着手ですよ。
		int               effectLevel, // エフェクトレベル
		Card             *card,       // 設置するカード
		Grid             *grid         // 設置先グリッド
)
{

	_endMethod   = endMethod;
	_move        = move;
	_effectLevel = effectLevel;
	_card       = card;
	_goalGrid    = grid; 

	_isEnd = false;

	_currentState = STATE_PICK_UP;


	_animation_PutCard    = NULL;
	_animation_HoldCard   = NULL;
	_animation_RisingCard = NULL;
	_offsetList_PutCard_index = -1;

	printf ( "カード設置エフェクト（AI用）開始\n" );

}

//=========================================================
// カード叩きつけオフセットリスト生成＿For0
//=========================================================
void Effect_PutCard_AI::MakeOffsetList_For_0() {

	_offsetList_PutCard.clear();

}
//=========================================================
// カード叩きつけオフセットリスト生成＿For3
//=========================================================
void Effect_PutCard_AI::MakeOffsetList_For_3() {

	_offsetList_PutCard.clear();

	int y = 0; int jump = -30; int gravity = 2;
	while ( y >= 0 ) { // 着地したら終了

		int frameCount = 1;

		// 上昇中はゆっくり
		//if ( jump < 0 ) {
		//	frameCount = 2;
		//}
		_offsetList_PutCard.push_back ( OffsetAnimationFrame ( 0, jump, frameCount ) );

		y    -= jump;
		jump += gravity;

		// 下降中は速度統一
		if ( jump > 0 ) {
			jump = 30;
		}
	}

}

//=========================================================
// デストラクタ
//=========================================================
Effect_PutCard_AI::~Effect_PutCard_AI() {

	printf ( "カード設置エフェクト（AI用）終了\n" );


	delete _animation_PutCard;
	delete _animation_HoldCard;
	delete _animation_RisingCard;
}


//=========================================================
// 更新
//=========================================================
bool Effect_PutCard_AI::Update(sf::RenderWindow &window) {

	if (_isEnd) { return _isEnd; }


	///// 現在の状態：カードを持ったところ
	if ( _currentState == STATE_PICK_UP ) {

		///// ここでもう、早めに、カードの向きを変えてしまいましょう。それともくるくる回したりしたい？
		_card->SetCurrentDirection ( _move.putCardDirection );

		///// エフェクトレベル０
		if ( _effectLevel == 0 ) {

			// カードつかみ効果音：AIには不要かもしれない。
			//GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open26.wav" );

			// カードを設置先グリッドに設置する。
			_goalGrid->SetCard ( _card );

		///// エフェクトレベル１
		} else if ( _effectLevel == 1 ) {

			// カードつかみ効果音：AIには不要かもしれない。
			//GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open26.wav" );

			// カードを設置先グリッドに設置する。
			_goalGrid->SetCard ( _card );

		///// エフェクトレベル２
		} else if ( _effectLevel == 2 ) {

			// カードつかみ効果音：AIには不要かもしれない。
			//GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open26.wav" );

			// カードを設置先グリッドに設置する。
			_goalGrid->SetCard ( _card );

		///// エフェクトレベル３
		} else if ( _effectLevel == 3 ) {

			// カード保持エフェクトを作成
			_animation_HoldCard = new GUI_SFML::Animation();
			_animation_HoldCard->Init("./effect/holdCardTest/define.dat");
			_animation_HoldCard->SetPosition ( _card->GetGoalPosition() );

			// カードつかみ音
			GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/mecha05.wav" );
		}

		// 状態遷移 → 設置先グリッドに向かっているところ
		_currentState = STATE_GOING_TO_THE_GRID;


	///// 現在の状態：カードが設置先グリッドに向かっているところ
	} else if ( _currentState == STATE_GOING_TO_THE_GRID ) {

		///// 到着した？
		if ( _card->IsMovingStop() ) {

			///// エフェクトレベル０
			if ( _effectLevel == 0 ) {

				// ここで終わってもいいけど、１フレーム伸ばしてみる。音のタイミング的に。
				
				// 状態遷移 → 設置先グリッドにいる
				_currentState = STATE_AT_THE_GRID;

			///// エフェクトレベル１
			} else if ( _effectLevel == 1 ) {

				// ここで終わってもいいけど、１フレーム伸ばしてみる。音のタイミング的に。
				
				// 状態遷移 → 設置先グリッドにいる
				_currentState = STATE_AT_THE_GRID;

			///// エフェクトレベル２
			} else if ( _effectLevel == 2 ) {

				// ここで終わってもいいけど、１フレーム伸ばしてみる。音のタイミング的に。
				
				// 状態遷移 → 設置先グリッドにいる
				_currentState = STATE_AT_THE_GRID;

			///// エフェクトレベル３
			} else if ( _effectLevel == 3 ) {

				///// カードを目的のグリッドの位置に移動させる
				sf::Vector2f cardPos = _card->GetGoalPosition();
				sf::Vector2f goalPos  = _goalGrid->GetCenterPosition();

				pair<float, float> nextPos = Utils_Cal::MoveTo ( cardPos.x, cardPos.y, goalPos.x,  goalPos.y, 10.0f );

				_card->SetPosition ( sf::Vector2f( nextPos.first, nextPos.second ) );


				///// 到着した？
				if ( nextPos.first == goalPos.x && nextPos.second == goalPos.y) {

					// カード保持エフェクトを停止
					delete _animation_HoldCard;
					_animation_HoldCard = NULL;

					// カード上昇エフェクトを生成
					_animation_RisingCard = new GUI_SFML::Animation();
					_animation_RisingCard->Init("./effect/risingCardTest/define.dat");

					// カード上昇音
					GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/noise05.wav" );

					// カード叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
					_offsetList_PutCard_index = 0;
					MakeOffsetList_For_3();

					// グリッドにカードを入れる
					_goalGrid->SetCard ( _card );

					// 状態遷移 → 設置先グリッドにいる
					_currentState = STATE_AT_THE_GRID;

				}
			}
		}

	///// 現在の状態：カードが設置先グリッドにいる
	} else if ( _currentState == STATE_AT_THE_GRID ) {

		///// エフェクトレベル０
		if ( _effectLevel == 0 ) {

			// カード設置
			GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/hit15.wav" );

			printf ("カード移動完了（AI用）\n");

			(*_endMethod)(_move);
			_isEnd = true;

			// 終了
			_currentState = STATE_FINISHED;

		///// エフェクトレベル１
		} else if ( _effectLevel == 1 ) {

			// カード設置
			ImpactWave ( _goalGrid->GetCenterPosition() );
			GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/swing16.wav" );

			printf ("カード移動完了（AI用）\n");

			(*_endMethod)(_move);
			_isEnd = true;

			// 終了
			_currentState = STATE_FINISHED;

		///// エフェクトレベル２
		} else if ( _effectLevel == 2 ) {

			// カード設置
			Energy ( _goalGrid->GetCenterPosition() );
			GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/mecha23_b.wav" );

			printf ("カード移動完了（AI用）\n");

			(*_endMethod)(_move);
			_isEnd = true;

			// 終了
			_currentState = STATE_FINISHED;

		///// エフェクトレベル３
		} else if ( _effectLevel == 3 ) {

			///// カードたたきつけオフセットアニメーション
			if (_offsetList_PutCard_index != -1) {

				// 現在のコマのカウントが0（初回）のときのみ、カードにオフセットを追加する。
				if ( _offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current == 0 ) {

					_card->SetOffsetX ( _card->GetOffset().x + _offsetList_PutCard[_offsetList_PutCard_index].x );
					_card->SetOffsetY ( _card->GetOffset().y + _offsetList_PutCard[_offsetList_PutCard_index].y );


					///// 叩きつけオフセットアニメが下降を始めたら、カード設置エフェクトを作成
					if ( _animation_PutCard == NULL && _offsetList_PutCard[_offsetList_PutCard_index].y > 0 ) {

						_animation_PutCard = new GUI_SFML::Animation();
						_animation_PutCard->Init("./effect/putCardTest/define.dat");
						_animation_PutCard->SetPosition ( _goalGrid->GetCenterPosition() );

						// 上昇エフェクトは終了
						delete _animation_RisingCard;
						_animation_RisingCard = NULL;
					}

				}

				// Maxになったら
				if (_offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current >= _offsetList_PutCard[_offsetList_PutCard_index].frameCount_Max) {

					// カウントを０に戻して
					_offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current = 0;

					// 次へ
					++_offsetList_PutCard_index;

					// 終了？
					if (_offsetList_PutCard_index >= _offsetList_PutCard.size()) {
						_offsetList_PutCard_index = -1; // 終了。

						// カードのオフセットをクリア
						_card->SetOffset ( 0, 0 );
					}

				// まだMaxじゃないなら、カウントを進める
				} else {
					_offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current += 1;
				}
			}

			///// 移動完了？
			if ( _card->IsMovingStop() && _offsetList_PutCard_index == -1 ) {
			
				if ( !_animation_PutCard->IsActive() )  {

					printf ("カード移動完了（AI用）\n");

					(*_endMethod)(_move);
					_isEnd = true;
				}
			}

		} ///// エフェクトレベルで分岐 if else

	///// 現在の状態：終了
	} else if ( _currentState == STATE_FINISHED) {

		/*
			ここにはこないはず
		*/
		_isEnd = false;
	}




	////// アニメーションエフェクト；カード設置
	if ( _animation_PutCard != NULL ) {
		_animation_PutCard->Update(window);
	}
	////// アニメーションエフェクト；カード保持
	if ( _animation_HoldCard != NULL ) {
		_animation_HoldCard->SetPosition ( _card->GetCurrentPosition() );
		_animation_HoldCard->Update(window);
	}
	////// アニメーションエフェクト；カード上昇
	if ( _animation_RisingCard != NULL ) {
		_animation_RisingCard->SetPosition ( _card->GetCurrentPosition() + _card->GetOffset() );
		_animation_RisingCard->Update(window);
	}

	///// カードの面倒を見てあげないといけないかもしれない。どこにも所属していない状態だから。
	if ( _goalGrid->GetCard() != _card ) {
		_card->Update ( window, _card->GetCurrentPosition() );
	}

	return _isEnd;
}

//=========================================================
// 描画
//=========================================================
void Effect_PutCard_AI::Draw(sf::RenderWindow &window) {

	if (_isEnd) {return;}

	///// カードの面倒を見てあげないといけないかもしれない。どこにも所属していない状態だから。
	if ( _goalGrid->GetCard() != _card ) {
		_card->Draw ( window );
	}

	///// アニメーションエフェクト：カード設置
	if ( _animation_PutCard != NULL ) {
		_animation_PutCard->Draw ( window );
	}
	////// アニメーションエフェクト；カード保持
	if ( _animation_HoldCard != NULL ) {
		_animation_HoldCard->Draw ( window );
	}
	////// アニメーションエフェクト；カード上昇
	if ( _animation_RisingCard != NULL ) {
		_animation_RisingCard->Draw ( window );
	}
	
}
//=========================================================
// 設置１用衝撃波発生
//=========================================================
void Effect_PutCard_AI::ImpactWave ( sf::Vector2f positon ) {

	map<unsigned int, IEffect*> effectList;
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_wind/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 0 ] = effect;
	}
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_wind/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 15 ] = effect;
	}
	IEffect *effect = new EffectList ( effectList );
	Application::GetInstance()->_effectManager->Add(effect);
}

//=========================================================
// 設置２用効果
//=========================================================
void Effect_PutCard_AI::Energy ( sf::Vector2f positon ) {

	map<unsigned int, IEffect*> effectList;
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_energy/define_bottom.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 0 ] = effect;
	}
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_energy/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 15 ] = effect;
	}
	IEffect *effect = new EffectList ( effectList );
	Application::GetInstance()->_effectManager->Add(effect);
}