//=========================================================
//
// 着手エフェクト＿カード設置＿人間の操作
//
//=========================================================

#include "Effect_PutCard_Human.h"
#include "Card.h"
#include "CardBox.h"
#include "Board.h"
#include "Grid.h"
#include "Application.h"
#include "../AI/Move.h"
#include "../AI/Define.h"
#include "../GUI_SFML/SoundPlayer.h"
#include "Effect_Animation.h"
#include "EffectList.h"
#include "Application.h"
#include "EffectManager.h"
#include "PsychoGage.h"

//=========================================================
// カード設置動作のステージ
//=========================================================
const int Effect_PutCard_Human::STATE_BEFORE            = 0; // まだ何もしていない
const int Effect_PutCard_Human::STATE_HOLDING_NOW       = 1; // カードをマウスで持っている
const int Effect_PutCard_Human::STATE_GOING_TO_THE_GRID = 2; // 設置先グリッドに向かっている
const int Effect_PutCard_Human::STATE_FINISHED          = 3; // カードをグリッドに設置し終えた



//=========================================================
///// コンストラクタ：必要なオブジェクトを受け取る
//=========================================================
Effect_PutCard_Human::Effect_PutCard_Human ( 

				DelegateBase *endMethod,
				Card         *card,
				CardBox      *cardBox,
				Board        *board,
				int           turn,
				PsychoGage   *psychoGage

) 
{
	_endMethod = endMethod;
	_card     = card;
	_cardBox  = cardBox;
	_board    = board;
	_turn     = turn;
	_goalGrid = NULL;
	_psychoGage = psychoGage;

	_allGridList = _board->GetAllGrid();

	_isEnd = false;

	_currentState = STATE_BEFORE;


	_animation_PutCard    = NULL;
	_animation_HoldCard   = NULL;
	_animation_RisingCard = NULL;
	_offsetList_PutCard_index = -1;

	_effectLevel_toHoldCard = 0;
	_effectLevel_toPutCard  = 0;

	printf ( "カード設置エフェクト開始\n" );
}

//=========================================================
// カード叩きつけオフセットリスト生成＿For0
//=========================================================
void Effect_PutCard_Human::MakeOffsetList_For_0() {

	_offsetList_PutCard.clear();

}
//=========================================================
// カード叩きつけオフセットリスト生成＿For3
//=========================================================
void Effect_PutCard_Human::MakeOffsetList_For_3() {

	_offsetList_PutCard.clear();

	int y = 0; int jump = -30; int gravity = 2;
	while ( y >= 0 ) { // 着地したら終了

		int frameCount = 1;

		// 上昇中はゆっくり
		//if ( jump < 0 ) {
		//	frameCount = 2;
		//}
		_offsetList_PutCard.push_back ( OffsetAnimationFrame ( 0, jump, frameCount ) );

		y    -= jump;
		jump += gravity;

		// 下降中は速度統一
		if ( jump > 0 ) {
			jump = 30;
		}
	}

}

//=========================================================
// デストラクタ
//=========================================================
Effect_PutCard_Human::~Effect_PutCard_Human() {

	printf ( "カード設置エフェクト終了\n" );

	delete _animation_PutCard;
	delete _animation_HoldCard;
	delete _animation_RisingCard;
}


//=========================================================
// エフェクトレベル用の操作判定
//=========================================================
bool Effect_PutCard_Human::IsEffectInput1() {

	bool result = false;
	//result = sf::Keyboard::isKeyPressed ( sf::Keyboard::Num1 )
	result = ( _psychoGage->GetValueAsEffectLevel() == 1 );
	return result;
}
bool Effect_PutCard_Human::IsEffectInput2() {
	bool result = false;
	//result = sf::Keyboard::isKeyPressed ( sf::Keyboard::Num2 )
	result = ( _psychoGage->GetValueAsEffectLevel() == 2 );
	return result;
}
bool Effect_PutCard_Human::IsEffectInput3() {
	bool result = false;
	//result = sf::Keyboard::isKeyPressed ( sf::Keyboard::Num3 )
	result = ( _psychoGage->GetValueAsEffectLevel() == 3 );
	return result;
}

//=========================================================
// 更新
//=========================================================
bool Effect_PutCard_Human::Update(sf::RenderWindow &window) {

	if (_isEnd) {return _isEnd;}


	///// 現在の状態：まだ何もしていない
	if ( _currentState == STATE_BEFORE ) {

		///// マウスに持たれた？
		if ( IsCardHeldByMouse () ) { 
			printf ( "マウスに持たれた。\n" );

			//=========================================================
			// キー「１」押下時
			//=========================================================
			if ( IsEffectInput1() ) {

				_effectLevel_toHoldCard = 1;

				// カードつかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open25.wav" );

			}
			//=========================================================
			// キー「２」押下時
			//=========================================================
			else if ( IsEffectInput2() ) {

				_effectLevel_toHoldCard = 2;

				// カードつかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/power07.wav" );

			}
			//=========================================================
			// キー「３」押下時
			//=========================================================
			else if ( IsEffectInput3() ) {

				_effectLevel_toHoldCard = 3;

				// カード保持エフェクトを作成
				_animation_HoldCard = new GUI_SFML::Animation();
				_animation_HoldCard->Init("./effect/holdCardTest/define.dat");
				_animation_HoldCard->SetPosition ( _card->GetGoalPosition() );

				// カードつかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/mecha05.wav" );

			//=========================================================
			// ノーマル
			//=========================================================
			} else {

				_effectLevel_toHoldCard = 0;

				// カードつかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open26.wav" );
			}

			_currentState = STATE_HOLDING_NOW;
		}

		
	///// 現在の状態：カードを持っている
	} else if ( _currentState == STATE_HOLDING_NOW ) {

		///// カード台に戻した？
		if ( IsCardInCardBox  () ) { 
			printf ( "カード台に戻した。\n" ); 

			// カード保持エフェクトを消す
			delete _animation_HoldCard;
			_animation_HoldCard = NULL;

			//=========================================================
			// エフェクトレベル：通常
			//=========================================================
			if (_effectLevel_toHoldCard == 0) {

				// カード戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi38.wav" );


			//=========================================================
			// エフェクトレベル：１
			//=========================================================
			} else if (_effectLevel_toHoldCard == 1) {

				// カード戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi38.wav" );

			//=========================================================
			// エフェクトレベル：２
			//=========================================================
			} else if (_effectLevel_toHoldCard == 2) {

				// カード戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi38.wav" );

			//=========================================================
			// エフェクトレベル：３
			//=========================================================
			} else if (_effectLevel_toHoldCard == 3) {

				// カード戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi34.wav" );
			}


			// 事務手続き
			_currentState = STATE_BEFORE;
			_isEnd = true;

		///// 設置先グリッドを決めた？
		} else if ( IsCardSetInGrid () ){

			// どこに設置することにしたのかな？
			pair<int, int> pos = GetCardSettingGrid();
			_goalGrid = _board->GetGrid ( pos.first, pos.second );

			// 着手構造体を作るよ！
			_move.action    = AI_Modules::MOVE::ACTION::PUT_CARD;
			_move.turn      = _turn;
			_move.putCardId = _card->GetProperty().processId;
			_move.putCardX  = pos.first;
			_move.putCardY  = pos.second;
			_move.putCardDirection = _card->GetCurrentDirection();

			printf ("設置先を決めた (%d, %d)\n", pos.first, pos.second);

			_currentState = STATE_GOING_TO_THE_GRID;

			// カード保持エフェクトを解除
			delete _animation_HoldCard;
			_animation_HoldCard = NULL;

			//=========================================================
			// キー「１」押下時
			//=========================================================
			if ( IsEffectInput1() ) {

				_effectLevel_toPutCard = 1;

				// カード叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutCard_index = 0;
				MakeOffsetList_For_0();
			}
			//=========================================================
			// キー「２」押下時
			//=========================================================
			else if ( IsEffectInput2() ) {

				_effectLevel_toPutCard = 2;

				// カード叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutCard_index = 0;
				MakeOffsetList_For_0();
			}
			//=========================================================
			// キー「３」押下時
			//=========================================================
			else if ( IsEffectInput3() ) {

				_effectLevel_toPutCard = 3;

				// カード上昇エフェクトを生成
				_animation_RisingCard = new GUI_SFML::Animation();
				_animation_RisingCard->Init("./effect/risingCardTest/define.dat");

				// カード上昇音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/noise05.wav" );

				// カード叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutCard_index = 0;
				MakeOffsetList_For_3();

			//=========================================================
			// ノーマル
			//=========================================================
			} else {

				_effectLevel_toPutCard = 0;

				// カード叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutCard_index = 0;
				MakeOffsetList_For_0();
			}

		}

	///// 現在の状態：設置先グリッドに向かっている
	} else if ( _currentState == STATE_GOING_TO_THE_GRID ) {

		//=========================================================
		// エフェクトレベル０
		//=========================================================
		if ( _effectLevel_toPutCard == 0 ) {

			///// 移動完了？
			if ( _card->IsMovingStop() ) {
			
				// カード設置
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/hit15.wav" );

				printf ("カード移動完了\n");

				(*_endMethod)(_move);
				_isEnd = true;
			}

		}
		//=========================================================
		// エフェクトレベル１
		//=========================================================
		else if ( _effectLevel_toPutCard == 1 ) {

			///// 移動完了？
			if ( _card->IsMovingStop() ) {
			
				// カード設置
				ImpactWave ( _goalGrid->GetCenterPosition() );
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/swing16.wav" );

				printf ("カード移動完了\n");

				(*_endMethod)(_move);
				_isEnd = true;
			}
		}
		//=========================================================
		// エフェクトレベル２
		//=========================================================
		else if ( _effectLevel_toPutCard == 2 ) {

			///// 移動完了？
			if ( _card->IsMovingStop() ) {
			
				// カード設置
				Energy ( _goalGrid->GetCenterPosition() );
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/mecha23_b.wav" );

				printf ("カード移動完了\n");

				(*_endMethod)(_move);
				_isEnd = true;
			}
		}
		//=========================================================
		// エフェクトレベル３
		//=========================================================
		else if ( _effectLevel_toPutCard == 3 ) {

			///// カードたたきつけオフセットアニメーション
			if (_offsetList_PutCard_index != -1) {

				// 現在のコマのカウントが0（初回）のときのみ、カードにオフセットを追加する。
				if ( _offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current == 0 ) {

					_card->SetOffsetX ( _card->GetOffset().x + _offsetList_PutCard[_offsetList_PutCard_index].x );
					_card->SetOffsetY ( _card->GetOffset().y + _offsetList_PutCard[_offsetList_PutCard_index].y );


					///// 叩きつけオフセットアニメが下降を始めたら、カード設置エフェクトを作成
					if ( _animation_PutCard == NULL && _offsetList_PutCard[_offsetList_PutCard_index].y > 0 ) {
						_animation_PutCard = new GUI_SFML::Animation();
						_animation_PutCard->Init("./effect/putCardTest/define.dat");
						_animation_PutCard->SetPosition ( _goalGrid->GetCenterPosition() );

						// 上昇エフェクトは終了
						delete _animation_RisingCard;
						_animation_RisingCard = NULL;
					}

				}

				// Maxになったら
				if (_offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current >= _offsetList_PutCard[_offsetList_PutCard_index].frameCount_Max) {

					// カウントを０に戻して
					_offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current = 0;

					// 次へ
					++_offsetList_PutCard_index;

					// 終了？
					if (_offsetList_PutCard_index >= _offsetList_PutCard.size()) {
						_offsetList_PutCard_index = -1; // 終了。

						// カードのオフセットをクリア
						_card->SetOffset ( 0, 0 );
					}

				// まだMaxじゃないなら、カウントを進める
				} else {
					_offsetList_PutCard[_offsetList_PutCard_index].frameCount_Current += 1;
				}
			}



			///// 移動完了？
			if ( _card->IsMovingStop() && _offsetList_PutCard_index == -1 ) {
			
				if ( !_animation_PutCard->IsActive() )  {

					printf ("カード移動完了\n");

					(*_endMethod)(_move);
					_isEnd = true;
				}
			}
		} ///// エフェクトレベルで分岐 if else

	///// 現在の状態：設置し終えた
	} else if ( _currentState == STATE_FINISHED ) {

		/*
			ここには来ないはず
		*/

		_isEnd = true;
	}


	////// アニメーションエフェクト；カード設置
	if ( _animation_PutCard != NULL ) {
		_animation_PutCard->Update(window);
	}
	////// アニメーションエフェクト；カード保持
	if ( _animation_HoldCard != NULL ) {
		_animation_HoldCard->SetPosition ( _card->GetCurrentPosition() );
		_animation_HoldCard->Update(window);
	}
	////// アニメーションエフェクト；カード上昇
	if ( _animation_RisingCard != NULL ) {
		_animation_RisingCard->SetPosition ( _card->GetCurrentPosition() + _card->GetOffset() );
		_animation_RisingCard->Update(window);
	}


	return _isEnd;
}

//=========================================================
// カードの現在地を探る：カード台？（つまり、戻された＝キャンセル）
//=========================================================
bool Effect_PutCard_Human::IsCardInCardBox() {

	bool result = false;

	vector<Card*> cardList = _cardBox->GetAllCards();
	for ( unsigned int i = 0; i < cardList.size(); ++i ) {

		if ( cardList[i] == _card ) {
			result = true;
			break;
		}
	}

	return result;
}
//=========================================================
// カードの現在地を探る：マウスに持たれてる？
//=========================================================
bool Effect_PutCard_Human::IsCardHeldByMouse() {

	bool result = false;

	if ( _card ==  Application::GetInstance()->RefferMouseClingObj() ) {
		result = true;
	}

	return result;
}
//=========================================================
// カードの現在地を探る：目的のグリッドにセットされた？
//=========================================================
bool Effect_PutCard_Human::IsCardSetInGrid() {

	bool result = false;

	for ( unsigned int y = 0; y < _allGridList.size(); ++y ) {
		for ( unsigned int x = 0; x < _allGridList[y].size(); ++x ) {

			if ( _card ==  _allGridList[y][x]->GetCard() ) {

				result = true;
				break;
			}
		}
	}

	return result;
}


//=========================================================
// カード、設置済みだとして、どの座標に設置した？
//=========================================================
pair<int, int> Effect_PutCard_Human::GetCardSettingGrid() {

	pair<int, int> result( -1, -1  );

	for ( unsigned int y = 0; y < _allGridList.size(); ++y ) {
		for ( unsigned int x = 0; x < _allGridList[y].size(); ++x ) {

			if ( _card ==  _allGridList[y][x]->GetCard() ) {

				result.first  = x;
				result.second = y;
				break;
			}
		}
	}

	return result;
}



//=========================================================
// 描画
//=========================================================
void Effect_PutCard_Human::Draw(sf::RenderWindow &window) {

	if (_isEnd) {return;}

	///// アニメーションエフェクト：カード設置
	if ( _animation_PutCard != NULL ) {
		_animation_PutCard->Draw ( window );
	}
	////// アニメーションエフェクト；カード保持
	if ( _animation_HoldCard != NULL ) {
		_animation_HoldCard->Draw ( window );
	}
	////// アニメーションエフェクト；カード上昇
	if ( _animation_RisingCard != NULL ) {
		_animation_RisingCard->Draw ( window );
	}

}

//=========================================================
// 設置１用衝撃波発生
//=========================================================
void Effect_PutCard_Human::ImpactWave ( sf::Vector2f positon ) {

	map<unsigned int, IEffect*> effectList;
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_wind/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 0 ] = effect;
	}
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_wind/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 15 ] = effect;
	}
	IEffect *effect = new EffectList ( effectList );
	Application::GetInstance()->_effectManager->Add(effect);
}

//=========================================================
// 設置２用効果
//=========================================================
void Effect_PutCard_Human::Energy ( sf::Vector2f positon ) {

	map<unsigned int, IEffect*> effectList;
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_energy/define_bottom.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 0 ] = effect;
	}
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putCard_energy/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		effectList [ 15 ] = effect;
	}
	IEffect *effect = new EffectList ( effectList );
	Application::GetInstance()->_effectManager->Add(effect);
}


