//=========================================================
//
// 着手エフェクト＿駒設置＿人間の操作
//
//=========================================================

#include "Effect_PutPiece_Human.h"
#include "Piece.h"
#include "PieceBox.h"
#include "PieceBox.h"
#include "Board.h"
#include "Grid.h"
#include "Application.h"
#include "../AI/Move.h"
#include "../AI/Define.h"
#include "../GUI_SFML/SoundPlayer.h"
#include "Effect_ImpactWave.h"
#include "Effect_ImpactWave_Rash.h"
#include "Effect_Animation.h"
#include "EffectManager.h"
#include "Define.h"
#include "PsychoGage.h"

//=========================================================
// 駒設置動作のステージ
//=========================================================
const int Effect_PutPiece_Human::STATE_BEFORE            = 0; // まだ何もしていない
const int Effect_PutPiece_Human::STATE_HOLDING_NOW       = 1; // 駒をマウスで持っている
const int Effect_PutPiece_Human::STATE_GOING_TO_THE_GRID = 2; // 設置先グリッドに向かっている
const int Effect_PutPiece_Human::STATE_FINISHED          = 3; // 駒をグリッドに設置し終えた



//=========================================================
///// コンストラクタ：必要なオブジェクトを受け取る
//=========================================================
Effect_PutPiece_Human::Effect_PutPiece_Human ( 

			DelegateBase *endMethod,       // 終了時の通知先メソッド
			Piece        *piece,           // 駒
			PieceBox     *pieceBox,        // 駒台
			Board        *board,           // 盤
			int           turn,            // 手番：Moveを作るのに必要
			PsychoGage   *psychoGage       // 意志ゲージ

) 
{
	_endMethod = endMethod;
	_piece     = piece;
	_pieceBox  = pieceBox;
	_board     = board;
	_turn      = turn;
	_goalGrid  = NULL;
	_psychoGage = psychoGage;

	_allGridList = _board->GetAllGrid();

	_isEnd = false;

	_currentState = STATE_BEFORE;


	_animation_PutPiece    = NULL;
	_animation_HoldPiece   = NULL;
	_animation_RisingPiece = NULL;
	_offsetList_PutPiece_index = -1;

	_effectLevel_toHoldPiece = 0;
	_effectLevel_toPutPiece  = 0;

	printf ( "駒設置エフェクト開始\n" );
}

//=========================================================
// 駒叩きつけオフセットリスト生成＿For0
//=========================================================
void Effect_PutPiece_Human::MakeOffsetList_For_0() {

	_offsetList_PutPiece.clear();

}
//=========================================================
// 駒叩きつけオフセットリスト生成＿For3
//=========================================================
void Effect_PutPiece_Human::MakeOffsetList_For_3() {

	_offsetList_PutPiece.clear();

	int y = 0; int jump = -30; int gravity = 2;
	while ( y >= 0 ) { // 着地したら終了

		int frameCount = 1;

		// 上昇中はゆっくり
		//if ( jump < 0 ) {
		//	frameCount = 2;
		//}
		_offsetList_PutPiece.push_back ( OffsetAnimationFrame ( 0, jump, frameCount ) );

		y    -= jump;
		jump += gravity;

		// 下降中は速度統一
		if ( jump > 0 ) {
			jump = 30;
		}
	}

}

//=========================================================
// デストラクタ
//=========================================================
Effect_PutPiece_Human::~Effect_PutPiece_Human() {

	printf ( "駒設置エフェクト終了\n" );
	
	delete _animation_PutPiece;
	delete _animation_HoldPiece;
	delete _animation_RisingPiece;
}

//=========================================================
// 強制終了
//=========================================================
void Effect_PutPiece_Human::Terminate() {
	
	printf ("あっ、強制終了させられるのですね Effect_PutPiece_Human::Terminate()\n");
}

//=========================================================
// エフェクトレベル用の操作判定
//=========================================================
bool Effect_PutPiece_Human::IsEffectInput1() {

	bool result = false;
	//result = sf::Keyboard::isKeyPressed ( sf::Keyboard::Num1 )
	result = ( _psychoGage->GetValueAsEffectLevel() == 1 );
	return result;
}
bool Effect_PutPiece_Human::IsEffectInput2() {
	bool result = false;
	//result = sf::Keyboard::isKeyPressed ( sf::Keyboard::Num2 )
	result = ( _psychoGage->GetValueAsEffectLevel() == 2 );
	return result;
}
bool Effect_PutPiece_Human::IsEffectInput3() {
	bool result = false;
	//result = sf::Keyboard::isKeyPressed ( sf::Keyboard::Num3 )
	result = ( _psychoGage->GetValueAsEffectLevel() == 3 );
	return result;
}

//=========================================================
// 更新
//=========================================================
bool Effect_PutPiece_Human::Update(sf::RenderWindow &window) {

	if (_isEnd) {return _isEnd;}


	///// 現在の状態：まだ何もしていない
	if ( _currentState == STATE_BEFORE ) {

		///// マウスに持たれた？
		if ( IsPieceHeldByMouse () ) { 
			printf ( "マウスに持たれた。\n" );

			//=========================================================
			// キー「１」押下時
			//=========================================================
			if ( IsEffectInput1() ) {

				_effectLevel_toHoldPiece = 1;

				// 駒つかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open25.wav" );

			}
			//=========================================================
			// キー「２」押下時
			//=========================================================
			else if ( IsEffectInput2() ) {

				_effectLevel_toHoldPiece = 2;

				// 駒つかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/power25.wav" );

			}
			//=========================================================
			// キー「３」押下時
			//=========================================================
			else if ( IsEffectInput3() ) {

				_effectLevel_toHoldPiece = 3;

				// 駒保持エフェクトを作成
				_animation_HoldPiece = new GUI_SFML::Animation();
				_animation_HoldPiece->Init("./effect/holdPieceTest/define.dat");
				_animation_HoldPiece->SetPosition ( _piece->GetGoalPosition() );

				// 駒つかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/mecha05.wav" );

			//=========================================================
			// ノーマル
			//=========================================================
			} else {

				_effectLevel_toHoldPiece = 0;

				// 駒つかみ音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/open26.wav" );
			}

			_currentState = STATE_HOLDING_NOW;
		}

		
	///// 現在の状態：駒を持っている
	} else if ( _currentState == STATE_HOLDING_NOW ) {

		///// 駒台に戻した？
		if ( IsPieceInPieceBox  () ) { 
			printf ( "駒台に戻した。\n" ); 

			// 駒保持エフェクトを消す
			delete _animation_HoldPiece;
			_animation_HoldPiece = NULL;

			//=========================================================
			// エフェクトレベル：通常
			//=========================================================
			if (_effectLevel_toHoldPiece == 0) {

				// 駒戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi38.wav" );

			}
			//=========================================================
			// エフェクトレベル：１
			//=========================================================
			else if (_effectLevel_toHoldPiece == 1) {

				// 駒戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi38.wav" );

			}
			//=========================================================
			// エフェクトレベル：２
			//=========================================================
			else if (_effectLevel_toHoldPiece == 2) {

				// 駒戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi38.wav" );

			}
			//=========================================================
			// エフェクトレベル：３
			//=========================================================
			else if (_effectLevel_toHoldPiece == 3) {

				// 駒戻し音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/kachi34.wav" );
			}


			// 事務手続き
			_currentState = STATE_BEFORE;
			_isEnd = true;

		///// 設置先グリッドを決めた？
		} else if ( IsPieceSetInGrid () ){

			// どこに設置することにしたのかな？
			pair<int, int> pos = GetPieceSettingGrid();
			_goalGrid = _board->GetGrid ( pos.first, pos.second );

			// 着手構造体を作るよ！
			_move.action    = AI_Modules::MOVE::ACTION::PUT_PIECE;
			_move.turn      = _turn;
			_move.putPieceX = pos.first;
			_move.putPieceY = pos.second;

			printf ("設置先を決めた (%d, %d)\n", pos.first, pos.second);

			_currentState = STATE_GOING_TO_THE_GRID;

			// 駒保持エフェクトを解除
			delete _animation_HoldPiece;
			_animation_HoldPiece = NULL;

			//=========================================================
			// キー「１」押下時
			//=========================================================
			if ( IsEffectInput1() ) {

				_effectLevel_toPutPiece = 1;

				// 駒叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutPiece_index = 0;
				MakeOffsetList_For_0();

			}
			//=========================================================
			// キー「２」押下時
			//=========================================================
			else if ( IsEffectInput2() ) {

				_effectLevel_toPutPiece = 2;

				// 駒叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutPiece_index = 0;
				MakeOffsetList_For_0();

			}
			//=========================================================
			// キー「３」押下時
			//=========================================================
			else if ( IsEffectInput3()) {

				_effectLevel_toPutPiece = 3;

				// 駒上昇エフェクトを生成
				_animation_RisingPiece = new GUI_SFML::Animation();
				_animation_RisingPiece->Init("./effect/risingPieceTest/define.dat");

				// 駒上昇音
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/noise05.wav" );

				// 駒叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutPiece_index = 0;
				MakeOffsetList_For_3();

			//=========================================================
			// ノーマル
			//=========================================================
			} else {

				_effectLevel_toPutPiece = 0;

				// 駒叩きつけオフセットアニメーション開始：0にすると開始になる。普段は -1
				_offsetList_PutPiece_index = 0;
				MakeOffsetList_For_0();
			}

		}

	///// 現在の状態：設置先グリッドに向かっている
	} else if ( _currentState == STATE_GOING_TO_THE_GRID ) {

		//=========================================================
		// エフェクトレベル０
		//=========================================================
		if ( _effectLevel_toPutPiece == 0 ) {

			///// 移動完了？
			if ( _piece->IsMovingStop() ) {
			
				// 駒設置
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/hit15.wav" );

				printf ("駒移動完了\n");

				(*_endMethod)(_move);
				_isEnd = true;
			}

		}
		//=========================================================
		// エフェクトレベル１
		//=========================================================
		else if ( _effectLevel_toPutPiece == 1 ) {

			///// 移動完了？
			if ( _piece->IsMovingStop() ) {
			
				// 駒設置
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/hit27.wav" );

				// 衝撃波
				ImpactWave ( _goalGrid->GetCenterPosition() );

				printf ("駒移動完了\n");

				(*_endMethod)(_move);
				_isEnd = true;
			}

		}
		//=========================================================
		// エフェクトレベル２
		//=========================================================
		else if ( _effectLevel_toPutPiece == 2 ) {

			///// 移動完了？
			if ( _piece->IsMovingStop() ) {
			
				// 駒設置
				GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/mecha04.wav" );

				// エフェクト
				Energy ( _goalGrid->GetCenterPosition() );

				printf ("駒移動完了\n");

				(*_endMethod)(_move);
				_isEnd = true;
			}

		}
		//=========================================================
		// エフェクトレベル３
		//=========================================================
		else if ( _effectLevel_toPutPiece == 3 ) {

			///// 駒たたきつけオフセットアニメーション
			if (_offsetList_PutPiece_index != -1) {

				// 現在のコマのカウントが0（初回）のときのみ、駒にオフセットを追加する。
				if ( _offsetList_PutPiece[_offsetList_PutPiece_index].frameCount_Current == 0 ) {

					_piece->SetOffsetX ( _piece->GetOffset().x + _offsetList_PutPiece[_offsetList_PutPiece_index].x );
					_piece->SetOffsetY ( _piece->GetOffset().y + _offsetList_PutPiece[_offsetList_PutPiece_index].y );


					///// 叩きつけオフセットアニメが下降を始めたら、駒設置エフェクトを作成
					if ( _animation_PutPiece == NULL && _offsetList_PutPiece[_offsetList_PutPiece_index].y > 0 ) {
						_animation_PutPiece = new GUI_SFML::Animation();
						_animation_PutPiece->Init("./effect/putPieceTest/define.dat");
						_animation_PutPiece->SetPosition ( _goalGrid->GetCenterPosition() );

						// 上昇エフェクトは終了
						delete _animation_RisingPiece;
						_animation_RisingPiece = NULL;
					}

				}

				// Maxになったら
				if (_offsetList_PutPiece[_offsetList_PutPiece_index].frameCount_Current >= _offsetList_PutPiece[_offsetList_PutPiece_index].frameCount_Max) {

					// カウントを０に戻して
					_offsetList_PutPiece[_offsetList_PutPiece_index].frameCount_Current = 0;

					// 次へ
					++_offsetList_PutPiece_index;

					// 終了？
					if (_offsetList_PutPiece_index >= _offsetList_PutPiece.size()) {
						_offsetList_PutPiece_index = -1; // 終了。

						// 駒のオフセットをクリア
						_piece->SetOffset ( 0, 0 );
					}

				// まだMaxじゃないなら、カウントを進める
				} else {
					_offsetList_PutPiece[_offsetList_PutPiece_index].frameCount_Current += 1;
				}
			}



			///// 移動完了？
			if ( _piece->IsMovingStop() && _offsetList_PutPiece_index == -1 ) {
			
				if ( !_animation_PutPiece->IsActive() )  {

					printf ("駒移動完了\n");

					(*_endMethod)(_move);
					_isEnd = true;
				}
			}
		} ///// エフェクトレベルで分岐 if else

	///// 現在の状態：設置し終えた
	} else if ( _currentState == STATE_FINISHED ) {

		/*
			ここには来ないはず
		*/

		_isEnd = true;
	}


	////// アニメーションエフェクト；駒設置
	if ( _animation_PutPiece != NULL ) {
		_animation_PutPiece->Update(window);
	}
	////// アニメーションエフェクト；駒保持
	if ( _animation_HoldPiece != NULL ) {
		_animation_HoldPiece->SetPosition ( _piece->GetCurrentPosition() );
		_animation_HoldPiece->Update(window);
	}
	////// アニメーションエフェクト；駒上昇
	if ( _animation_RisingPiece != NULL ) {
		_animation_RisingPiece->SetPosition ( _piece->GetCurrentPosition() + _piece->GetOffset() );
		_animation_RisingPiece->Update(window);
	}


	return _isEnd;
}

//=========================================================
// 駒の現在地を探る：駒台？（つまり、戻された＝キャンセル）
//=========================================================
bool Effect_PutPiece_Human::IsPieceInPieceBox() {

	bool result = false;

	vector<Piece*> pieceList = _pieceBox->GetPieceList();
	for ( unsigned int i = 0; i < pieceList.size(); ++i ) {

		if ( pieceList[i] == _piece ) {
			result = true;
			break;
		}
	}

	return result;
}
//=========================================================
// 駒の現在地を探る：マウスに持たれてる？
//=========================================================
bool Effect_PutPiece_Human::IsPieceHeldByMouse() {

	bool result = false;

	if ( _piece ==  Application::GetInstance()->RefferMouseClingObj() ) {
		result = true;
	}

	return result;
}
//=========================================================
// 駒の現在地を探る：目的のグリッドにセットされた？
//=========================================================
bool Effect_PutPiece_Human::IsPieceSetInGrid() {

	bool result = false;

	for ( unsigned int y = 0; y < _allGridList.size(); ++y ) {
		for ( unsigned int x = 0; x < _allGridList[y].size(); ++x ) {

			if ( _piece ==  _allGridList[y][x]->GetPiece() ) {

				result = true;
				break;
			}
		}
	}

	return result;
}


//=========================================================
// 駒、設置済みだとして、どの座標に設置した？
//=========================================================
pair<int, int> Effect_PutPiece_Human::GetPieceSettingGrid() {

	pair<int, int> result( -1, -1  );

	for ( unsigned int y = 0; y < _allGridList.size(); ++y ) {
		for ( unsigned int x = 0; x < _allGridList[y].size(); ++x ) {

			if ( _piece ==  _allGridList[y][x]->GetPiece() ) {

				result.first  = x;
				result.second = y;
				break;
			}
		}
	}

	return result;
}



//=========================================================
// 描画
//=========================================================
void Effect_PutPiece_Human::Draw(sf::RenderWindow &window) {

	if (_isEnd) {return;}

	///// アニメーションエフェクト：駒設置
	if ( _animation_PutPiece != NULL ) {
		_animation_PutPiece->Draw ( window );
	}
	////// アニメーションエフェクト；駒保持
	if ( _animation_HoldPiece != NULL ) {
		_animation_HoldPiece->Draw ( window );
	}
	////// アニメーションエフェクト；駒上昇
	if ( _animation_RisingPiece != NULL ) {
		_animation_RisingPiece->Draw ( window );
	}

}


//=========================================================
// 衝撃波エフェクト作成：設置レベル１用のアレ
//=========================================================
void Effect_PutPiece_Human::ImpactWave ( sf::Vector2f positon ) {

	// 音はここで出す
	//GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/bom11.wav" );

	Effect_ImpactWave::Property prop;
	prop.color = sf::Color ( 20,20,100, 100 );
	prop.maxRadius = 100; // APPLICATION::WINDOW_SIZE::WIDTH / 1.5;
	prop.speed = 15;
	prop.position = positon;
	prop.soundName = "";

		
	map<unsigned int, Effect_ImpactWave::Property> impactList;
	for ( int i = 0; i < 5; ++i ) {
		impactList [ i * 2 ] = prop;
	}

	Effect_Impact_Rash *effect = new Effect_Impact_Rash ( impactList );

	Application::GetInstance()->_effectManager-> Add ( effect );
}

//=========================================================
// 駒設置２用効果
//=========================================================
void Effect_PutPiece_Human::Energy ( sf::Vector2f positon ) {

	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putPiece_energy/define_bottom.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		Application::GetInstance()->_effectManager->Add(effect);
	}
	{
		GUI_SFML::Animation *animation = new GUI_SFML::Animation();
		animation->Init("./effect/putPiece_energy/define.dat");
		animation->SetPosition( positon );

		IEffect *effect = new Effect_Animation ( animation );
		Application::GetInstance()->_effectManager->Add(effect);
	}

}

