//=========================================================
//
// Luaのスクリプトで制御するエフェクト
//
// 参考：http://gamedevgeek.com/tutorials/calling-c-functions-from-lua/
//=========================================================

#include "Effect_ScriptedOperation.h"
#include "Define.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/Utils_Convert.h"
#include "../GUI_SFML/SoundPlayer.h"
#include "../Utils/Utils_Others.h"
#include "../Utils/Utils_String.h"
#include "../Utils/Utils_Lua.h"
#include "../GUI_SFML/SoundPlayer.h"

using namespace std;

//=========================================================
// コンストラクタ
//=========================================================
Effect_ScriptedOperation::Effect_ScriptedOperation ( DelegateBase *endMethod, int effectLevel, GUI_SFML::TextMessageWindow *textMessageWindow, map<string, string> parameters ) {

	// 引数を受け取る
	_textMessageWindow = textMessageWindow;
	_endMethod         = endMethod;
	_effectLevel       = effectLevel;
	
	
	_otherParams = "";
	
	//_faceImageFile1 = UTILS_OTHERS::GetValueFromMap ( parameters, "faceImage1", "" );
	/*
		cutInParameters["backImage2" ] = "IMAGE<>./image/cg/ぬいぐるみ.png";
		
	*/
	for ( map<string, string>::iterator it = parameters.begin(); it != parameters.end(); ++it ) {
		
		string paramKey   = it->first; 
		string paramValue = it->second;
		
		printf ( "key : %s = value : %s\n", paramKey.c_str(), paramValue.c_str() );
		
		vector<string> valueParsing = Utils_String::Split ( paramValue, "<>" );
		string objectType    = ( valueParsing.size() >= 1 ) ? valueParsing[0] : "";
		string objectContent = ( valueParsing.size() >= 2 ) ? valueParsing[1] : "";
		
		/*
			///// 内部的に持つ部品
			map<string, ImageSet*> _imageList; // 画像
			map<string, string>    _soundList; // 音（ファイル名）
			map<string, string>    _text;      // セリフ
			map<string, string>    _nameLabel; // 名札
		*/
		
		// その他のパラメータが指定されている場合
		if ( paramKey == "OTHER_PARAMS" ) {
			
			_otherParams = paramValue;
		
		// 画像を登録
		} else if ( objectType == "IMAGE" ) {

			if ( _imageList.find(paramKey) != _imageList.end() ) {
				printf ( "同じキー(%s)を重複して画像を登録(%s)したようですよ。\n", paramKey.c_str(), objectContent.c_str() );
				map<string, ImageSet*>::iterator it = _imageList.find(paramKey);
				delete it->second;
			}
			
			ImageSet *imageSet = new ImageSet(objectContent);
			_imageList[paramKey] = imageSet;
			
			
		// サウンドを登録
		} else if ( objectType == "SOUND" ) {
			
			_soundList[paramKey] = objectContent;
			
		// セリフを登録
		} else if ( objectType == "REMARK" ) {
			_textList[paramKey] = objectContent;

		// 名札を登録
		} else if ( objectType == "NAMELABEL" ) {
			_nameLabel[paramKey] = objectContent;
		}
	}
	
	_lua = NULL;
	_scriptFileName = UTILS_OTHERS::GetValueFromMap ( parameters, "scriptFileName", "" );
	Script_Init();

	_frameCount = 0;
	_isEnd = false;
}

//=========================================================
// デストラクタ
//=========================================================
Effect_ScriptedOperation::~Effect_ScriptedOperation () {

	///// Lua 終了
	if ( _lua != NULL ) {
		lua_close ( _lua );
	}

	for ( map<string, ImageSet*>::iterator it = _imageList.begin(); it != _imageList.end(); ++it ) {
		
		delete it->second;
	}
}


//=========================================================
// Lua使用中のインスタンス
//=========================================================
Effect_ScriptedOperation* Effect_ScriptedOperation::LuaedObject = NULL;


//=========================================================
// Luaスクリプト使用の初期設定
//=========================================================
void Effect_ScriptedOperation::Script_Init() {

	// 既存のLuaオブジェクトがあったらクリアしておく
	if ( _lua != NULL ) {
		lua_close(_lua);
	}
	// Lua 初期設定
	_lua = luaL_newstate();
	luaL_openlibs ( _lua );

	// Luaファイル読み込み
	if ( luaL_dofile (_lua, _scriptFileName.c_str()) ) {

		// エラーメッセージ
		string error_message = lua_tostring(_lua, -1);
		printf ( "Error：luaL_dofile: %s\n", error_message.c_str());

		///// Lua 終了
		lua_close ( _lua );
		_lua = NULL;

	// エラーなし
	} else {
		// よかったよかった。

		// Lua側から呼び出すC++側のメソッドを登録
		lua_register ( _lua, "Remark",         Effect_ScriptedOperation::FromLua_Remark                );
		lua_register ( _lua, "PlaySound",      Effect_ScriptedOperation::FromLua_PlaySound             );
		lua_register ( _lua, "PutImage",       Effect_ScriptedOperation::FromLua_PutImage              );
		lua_register ( _lua, "SetAlpha",       Effect_ScriptedOperation::FromLua_SetAlphaToImage       );
		lua_register ( _lua, "SetTextureRect", Effect_ScriptedOperation::FromLua_SetTextureRectToImage );
		lua_register ( _lua, "SetOrigin",      Effect_ScriptedOperation::FromLua_SetOriginToImage      );
		
	
		// Lua使用中のインスタンスは私ですよ。
		Effect_ScriptedOperation::LuaedObject = this;
		{
				
			// Lua側の初期設定関数を呼ぶ
			lua_getglobal ( _lua, "Init" );
			lua_pushstring ( _lua, _otherParams.c_str() ); // パラメータとか

			// 関数呼び出し
			if ( lua_pcall ( _lua, 1, 0, 0 ) != 0 ) {
				printf("error : %s\n", lua_tostring ( _lua, -1) );
				lua_pop ( _lua, 1);
			}	
		}
		// Lua使用中のインスタンスはもう私じゃないですよ。
		Effect_ScriptedOperation::LuaedObject = NULL;
	}
}


	
//=========================================================
// Lua側から呼び出される処理群
//=========================================================
// セリフを言う
static int Effect_ScriptedOperation::FromLua_Remark( lua_State *lua ) {
	return Effect_ScriptedOperation::LuaedObject->FromLua_Remark_Object(lua);
}
int Effect_ScriptedOperation::FromLua_Remark_Object( lua_State *lua ) {

	printf ( "【Effect_ScriptedOperation::FromLua_Remark_Object】Lua経由でセリフを言う。\n" );

	// 引数を受け取る
	string nameLabelID = lua_tostring( lua, 1 );
	string remarkID    = lua_tostring( lua, 2 );

	printf ("arg1 : %s\n", nameLabelID.c_str());
	printf ("arg2 : %s\n", remarkID.c_str());
	
	string nameLabel = _nameLabel[nameLabelID];
	string remark    = _textList[remarkID];
	
	_textMessageWindow->SetSpeakerName         ( nameLabel );
	_textMessageWindow->AddText_asFinishedText ( remark    );


	// スタックから降ろさなくていいのかな？
	//lua_pop ( _lua, -1 );
		
	// 戻り値をセット
	//lua_pushinteger(lua, 1000 + arg1);
	//lua_pushstring(lua, a.c_str());

	//return 2; // 戻り値の数

	return 0;
	
	
}

// 音を鳴らす
static int Effect_ScriptedOperation::FromLua_PlaySound( lua_State *lua ) {
	return Effect_ScriptedOperation::LuaedObject->FromLua_PlaySound_Object(lua);
}
int Effect_ScriptedOperation::FromLua_PlaySound_Object( lua_State *lua ) {
	
	printf ( "【Effect_ScriptedOperation::FromLua_PlaySound_Object】Lua経由で音を鳴らす\n" );

	// 引数を受け取る
	string soundID = lua_tostring( lua, 1 );

	printf ("arg1 : %s\n", soundID.c_str());
	
	string soundFileName = _soundList [soundID];
	
	GUI_SFML::SoundPlayer::Play ( soundFileName );
	
	
	return 0;
}
	
// 画像を表示する
int Effect_ScriptedOperation::FromLua_PutImage( lua_State *lua ) {
	return Effect_ScriptedOperation::LuaedObject->FromLua_PutImage_Object(lua);
}
int Effect_ScriptedOperation::FromLua_PutImage_Object( lua_State *lua ) {
	
	//printf ( "【Effect_ScriptedOperation::FromLua_PutImage_Object】Lua経由で画像を表示する\n" );

	// 引数を受け取る
	string imageID = lua_tostring  ( lua, 1 );
	int    posX    = lua_tointeger ( lua, 2 );
	int    posY    = lua_tointeger ( lua, 3 );
	
	

	//printf ("imageID : %s(%d, %d)\n", imageID.c_str(), posX, posY);
	
	//ImageSet *imageSet = NULL;
	//imageSet = ( _imageList.find(imageID) != _imageList.end() ) ? _imageList [imageID] : NULL;

	ImageSet *imageSet = ( _imageList.find(imageID) != _imageList.end() ) ? _imageList [imageID] : NULL;

	
	if ( imageSet != NULL ) {
		
		printf( " さぁ、どうにかして %s を表示するぞ！ \n", imageID.c_str() );
	
		//sf::RenderWindow &window;
		//ImageSet->sprite
		//window.draw ( _imageFace.sprite  );
		
		// sf::RenderWindowが必要になるので、
		// 実際の描画はDrawメソッドの中で描画することととして、
		// ここでは、その予約だけをしておく。
		
		ImagePutReserveInfo reserve;
		reserve.imageSet  = imageSet;
		reserve.positionX = posX;
		reserve.positionY = posY;
		
		_imagePutReserveList.push_back ( reserve );
		
	} else {
		printf ( "登録画像が存在しない\n" );
	}
	
	return 0;
	
}
// 画像のアルファ値を指定する
int Effect_ScriptedOperation::FromLua_SetAlphaToImage( lua_State *lua ) {
	return Effect_ScriptedOperation::LuaedObject->FromLua_SetAlphaToImage_Object(lua);
}
int Effect_ScriptedOperation::FromLua_SetAlphaToImage_Object( lua_State *lua ) {
	
	printf ( "【Effect_ScriptedOperation::FromLua_SetAlphaToImage_Object】Lua経由で画像のアルファ値を設定する\n" );

	// 引数を受け取る
	string imageID = lua_tostring  ( lua, 1 );
	int    alpha   = lua_tointeger ( lua, 2 );
	
	printf ("imageID : %s(%d)\n", imageID.c_str(), alpha );
	
	ImageSet *imageSet = ( _imageList.find(imageID) != _imageList.end() ) ? _imageList [imageID] : NULL;
	
	if ( imageSet != NULL ) {
			
		sf::Color color = imageSet->sprite.getColor();
		color.a = alpha;
		imageSet->sprite.setColor(color);
		
	} else {
		printf ( "登録画像が存在しない\n" );
	}
	
	return 0;
}

// 画像の描画領域を設定する
int Effect_ScriptedOperation::FromLua_SetTextureRectToImage( lua_State *lua ) {
	return Effect_ScriptedOperation::LuaedObject->FromLua_SetTextureRectToImage_Object(lua);
}
int Effect_ScriptedOperation::FromLua_SetTextureRectToImage_Object( lua_State *lua ) {
	
	printf ( "【Effect_ScriptedOperation::FromLua_SetTextureRectToImage_Object】Lua経由で画像の描画テクスチャ領域を設定する\n" );

	// 引数を受け取る
	string imageID = lua_tostring  ( lua, 1 );
	int    left    = lua_tointeger ( lua, 2 );
	int    top     = lua_tointeger ( lua, 3 );
	int    width   = lua_tointeger ( lua, 4 );
	int    height  = lua_tointeger ( lua, 5 );
	
	printf ("imageID : %s(%d,%d,%d,%d)\n", imageID.c_str(), left,top,width,height );
	
	ImageSet *imageSet = ( _imageList.find(imageID) != _imageList.end() ) ? _imageList [imageID] : NULL;
	
	if ( imageSet != NULL ) {
		
		sf::IntRect rect;
		rect.left = left;
		rect.top  = top;
		rect.width  = width;  //_imageBack.image.getSize().x;
		rect.height = height; //_imageFace.image.getSize().y;

		imageSet->sprite.setTextureRect ( rect );
		
	} else {
		printf ( "登録画像が存在しない\n" );
	}
	
	return 0;
}


// 画像の中心点を設定する
int Effect_ScriptedOperation::FromLua_SetOriginToImage( lua_State *lua ) {
	
	printf ( "画像の中心点を設定する\n" );
	printf ( "%d\n", Effect_ScriptedOperation::LuaedObject );
	
	return Effect_ScriptedOperation::LuaedObject->FromLua_SetOriginToImageToImage_Object(lua);
}
int Effect_ScriptedOperation::FromLua_SetOriginToImageToImage_Object( lua_State *lua ) {

	printf ( "【Effect_ScriptedOperation::FromLua_SetOriginToImageToImage_Object】画像の中心点を設定する\n" );

	// これで引数の数が取得できるらしいですよ。
	int numberOfArgs = lua_gettop(lua);
	printf ("引数の数 : %d\n", numberOfArgs );
	
	
	// 引数を受け取る：まずは登録画像のIDを先に
	string imageID = lua_tostring  ( lua, 1 );
	printf ("imageID [%s]\n", imageID.c_str());
	
	printf ( "_imageListを確認するよ %d \n", _imageList.size() );

	ImageSet *imageSet = ( _imageList.find(imageID) != _imageList.end() ) ? _imageList [imageID] : NULL;
	
	printf ( "画像ID : %s\n", imageID.c_str() );
	
	// 以降、存在するIDを指定された前提で
	if ( imageSet != NULL ) {

		printf ( "登録画像あり\n");

		int x = 0;
		int y = 0;
		
		// 第２引数あり？
		if ( numberOfArgs >= 2  ) { 
			x = lua_tointeger ( lua, 2 );
		
		// なし＝画像の中心点
		} else {
			printf ( "第２引数なしにつき、画像サイズの中心点とする\n" );
			x = imageSet->sprite.getGlobalBounds().width / 2.0f;
		}

		// 第３引数あり？
		if ( numberOfArgs >= 3  ) { 
			y = lua_tointeger ( lua, 3 );
		
		// なし＝画像の中心点
		} else {
			printf ( "第３引数なしにつき、画像サイズの中心点とする\n" );
			y = imageSet->sprite.getGlobalBounds().height / 2.0f;
		}
		printf ( "中心点を設定（%d, %d）\n", x, y );
		imageSet->sprite.setOrigin( x, y );
		
	} else {
		printf ( "登録画像が存在しない\n" );
	}
	
	return 0;
}
	
	

	
//=========================================================
// 更新
//=========================================================
bool Effect_ScriptedOperation::Update(sf::RenderWindow &window){
	
	//printf ("スクリプトエフェクト：更新\n");
	
	// C++側でフレームをカウントする必要はないのかもしれない
	_frameCount++;
	
	// とりあえず１フレームだけ。
	//_isEnd = ( _frameCount > 1 );
	
	string updateResult = Update_with_Lua();
	
	if ( updateResult == "end" ) {
		_isEnd = true;
	}

	if (_isEnd) {
		// 着手メソッドを実行する
		if ( _endMethod != NULL ) {
			(*_endMethod)( _effectLevel );
		}
	}
	
	
	return _isEnd;
	
}
string Effect_ScriptedOperation::Update_with_Lua() {
	
	// Lua使用中のインスタンスは私ですよ。
	Effect_ScriptedOperation::LuaedObject = this;
	
	// Lua関数をスタックに積む
	lua_getglobal ( _lua, "Update" );
	
	// 関数呼び出し
	if ( lua_pcall ( _lua, 0, 1, 0 ) != 0 ) {

		printf("error : %s\n", lua_tostring ( _lua, -1) );
		lua_pop ( _lua, 1);
	}

	// 戻り値：更新結果を文字列で
	string updateResult = Utils_Lua::GetReturnValue_asString ( _lua, 1 );

	// 戻り値をスタックから降ろす
	lua_pop ( _lua, -1 );


	// 私はもうLua使用中のインスタンスではありませんよ。
	Effect_ScriptedOperation::LuaedObject = NULL;
	
	return updateResult;
}


//=========================================================
// 描画
//=========================================================
void Effect_ScriptedOperation::Draw(sf::RenderWindow &window) {

	//printf ("スクリプトエフェクト：描画\n");
	
	Draw_with_Lua(window);
	
	// 予約されている画像を表示する
	// つまり、Lua側でどのタイミングで描画を指定しようとも、実際に描画されるのはこのタイミングとなる。
	// それまでの間に同じ画像を重ねて指定すると、重ねて描画が実行される
	for ( unsigned int i = 0; i < _imagePutReserveList.size(); ++i ) {
		
		ImagePutReserveInfo reserveInfo = _imagePutReserveList[i];
		
		// 位置を指定して
		reserveInfo.imageSet->sprite.setPosition ( reserveInfo.positionX, reserveInfo.positionY );
	
		// 描画
		window.draw ( reserveInfo.imageSet->sprite );
		
	}
	_imagePutReserveList.clear();
	
}
void Effect_ScriptedOperation::Draw_with_Lua(sf::RenderWindow &window) {

	// Lua使用中のインスタンスは私ですよ。
	Effect_ScriptedOperation::LuaedObject = this;
	
	// Lua関数をスタックに積む
	lua_getglobal ( _lua, "Draw" );
	
	// 関数呼び出し
	if ( lua_pcall ( _lua, 0, 0, 0 ) != 0 ) {

		printf("error : %s\n", lua_tostring ( _lua, -1) );
		lua_pop ( _lua, 1);
	}

	// 私はもうLua使用中のインスタンスではありませんよ。
	Effect_ScriptedOperation::LuaedObject = NULL;
}	
	
	
	
	
	
	

/*
static int Effect_ScriptedOperation::LuaTest( lua_State *lua ) {
	
	return Effect_ScriptedOperation::LuaedObject->LuaTest_Object(lua);
	
}
int Effect_ScriptedOperation::LuaTest_Object( lua_State *lua ) {

	string a = "文字列";
	printf ("C++側のコード：クラスメソッドの実験ですよ。\n");
	
	
	// これで引数の数が取得できるらしいですよ。
	int n = lua_gettop(lua);
	printf ("引数の数 : %d\n", n );
	
	// 引数を受け取る
	string arg1 = lua_tostring( lua, 1 );

	printf ("arg1 : %s\n", arg1.c_str());


	// スタックから降ろさなくていいのかな？ どうも、いいっぽい。
	//lua_pop ( _lua, -1 );
		
	// 戻り値をセット
	//lua_pushinteger(lua, 1000 + arg1);
	//lua_pushstring(lua, a.c_str());

	//return 2; // 戻り値の数

	return 0;
}
*/
	
	
	

