//=========================================================
//
// 画面ゆれ
//
//=========================================================

#include "Effect_Shake.h"
#include "../Utils/Utils_Cal.h"
#include "../Utils/mt19937ar.h"
#include "../Camera.h"

int Effect_Shake::_shakeXforTheWorld = 0;
int Effect_Shake::_shakeYforTheWorld = 0;

//=========================================================
// コンストラクタ
//=========================================================
Effect_Shake::Effect_Shake ( Property prop ) {
	
	_prop = prop;
	
	//_originalView = _prop.window->getDefaultView(); 
	
	_viewOfBeginning = _prop.window->getView(); 
	
	// デフォルトビューではなく現時点のビューを取得したいところですが、がめんゆれエフェクトが重ねがけされると、終了時にズレる
	// ゆえに、終了時にはデフォルトビューに戻すことにする
	// 別経路でビューを変更していた場合、画面ゆれエフェクトを実行することでデフォルトビューに戻ってしまう。大変だ！
	//
	// エフェクト終了時に、Cameraの状態を適用しなおすこととした。
	// Cameraオブジェクトはエフェクト開始時にScriptEngine側から _prop内に受け取っている。
	
	
	_isEnd = false;
	_frameCount = 0;
	_interval = 0;
	
	if ( _prop.isShakeEnd != NULL ) {
		*(_prop.isShakeEnd) = "false";
	}
	
	init_genrand( static_cast<unsigned int>( time(NULL) ) );

	//printf("画面ゆれエフェクト開始\n");
	
}
//=========================================================
// デストラクタ
//=========================================================
Effect_Shake::~Effect_Shake() {
	//printf("画面ゆれ終了\n");
	
	if ( _prop.isShakeEnd != NULL ) {
		*(_prop.isShakeEnd) = "true";
	}
	
	_shakeXforTheWorld = 0;
	_shakeYforTheWorld = 0;
	
	if ( _prop.camera != NULL ) {
		_prop.camera->ApplyThisCameraToTheActualScreen();
	}
	
}
//=========================================================
// 更新
//=========================================================
bool Effect_Shake::Update(sf::RenderWindow &window) {
	
	if ( _isEnd ) {
		return _isEnd;
	}
	
	if ( _prop.textWindow != NULL ) {
		_prop.textWindow->_notCareOfCameraPos = true;
	}
	
	--_interval;
	
	if ( _interval < 0 ) {
			
		_interval = _prop.interval;
	
		//sf::View currentView = _prop.window->getDefaultView();
		sf::View currentView = _viewOfBeginning;
		
		currentView.setSize ( currentView.getSize().x - _prop.magnitude, currentView.getSize().y - _prop.magnitude );
		
	
		int randomShakeX = genrand_int31() % _prop.magnitude;
		int randomShakeY = genrand_int31() % _prop.magnitude;
		
		int shakeX = randomShakeX - _prop.magnitude/2;
		int shakeY = randomShakeY - _prop.magnitude/2;
		
		//int shakeX = Utils_Cal::GetRandom(_prop.magnitude) - _prop.magnitude/2;
		//int shakeY = Utils_Cal::GetRandom(_prop.magnitude) - _prop.magnitude/2;
		
		// 外部からの取得用
		_shakeXforTheWorld = shakeX;
		_shakeYforTheWorld = shakeY;
		
		currentView.move ( shakeX, shakeY );
		
		_prop.window->setView(currentView);
	}
	

	++_frameCount;
	if ( _frameCount > _prop.frameCount ) {
		
		// デフォルトビューに戻すのではなく、カメラの状態を適用しなおす
		//_prop.window->setView(_originalView);	
		_prop.camera->ApplyThisCameraToTheActualScreen();
		
		if ( _prop.textWindow != NULL ) {
			_prop.textWindow->_notCareOfCameraPos = false;
		}
	
		_isEnd = true;
		
		if ( _prop.isShakeEnd != NULL ) {
			*(_prop.isShakeEnd) = "true";
		}

		_shakeXforTheWorld = 0;
		_shakeYforTheWorld = 0;

	}
	
	return _isEnd;
}


//=========================================================
// 描画
//=========================================================
void Effect_Shake::Draw(sf::RenderWindow &window) {
	// 処理なし
}

