//=========================================================
//
// 盤
//
//=========================================================

#include "Board.h"
#include "Grid.h"
#include "Piece.h"
#include "Card.h"
#include "CardFactory.h"
#include "../Utils/Utils_Convert.h"
#include "../Utils/Utils_String.h"

#include <map>
#include <SFML/Graphics.hpp>

using namespace std;


//=========================================================
// コンストラクタ
//=========================================================
Board::Board(int width, int height) {

	sf::Vector2f baseSize(0.0f, 0.0f);
	_baseMargin = sf::Vector2f(5.0f, 5.0f);

	baseSize += _baseMargin;

	sf::Vector2f baseSizeMeasurer = baseSize;

	///// Generate Grids
	for (int y = 0; y < height; ++y) {

		baseSizeMeasurer.x = _baseMargin.x;

		vector<Grid*> row;

		sf::Vector2f gridSize;

		for (int x = 0; x < width; ++x) {

			Grid *grid = new Grid();
			row.push_back( grid );

			gridSize = grid->GetSize();

			baseSizeMeasurer.x += gridSize.x;

			if (baseSizeMeasurer.x > baseSize.x) {
				baseSize.x = baseSizeMeasurer.x;
			}

		}

		baseSizeMeasurer.y = baseSize.y + gridSize.y;
		if (baseSizeMeasurer.y > baseSize.y) {
			baseSize.y = baseSizeMeasurer.y;
		}


		_gridList.push_back(row);

	}

	baseSize += _baseMargin;


	_base = sf::RectangleShape ( baseSize );
	_base.setFillColor         ( sf::Color(150,150,150,255) );
	_base.setOutlineColor      ( sf::Color(0,0,0,255) );
	_base.setOutlineThickness  ( 1.0f );


	//=========================================================
	// 相対位置のお仲間たちを各自におしらせ。
	//=========================================================
	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];

			for (unsigned int inner_y = 0; inner_y < _gridList.size(); ++inner_y) {

				vector<Grid*> inner_row = _gridList[inner_y];

				for (unsigned int inner_x = 0; inner_x < inner_row.size(); ++inner_x) {

					Grid *inner_grid = inner_row[inner_x];

					int relativeX = inner_x - x;
					int relativeY = inner_y - y;
					grid->_relativeGridList[ relativeY ][ relativeX ] = inner_grid;

				} // inner_x
			} // inner_y

		} // x
	} // y
}

//=========================================================
// コンストラクタ：ファイルから再現するためのサイズ文字列で
//=========================================================
Board* Board::GetInstance ( string outPutSizeData ) {

	Board *result = NULL;

	int width  = 0;
	int height = 0;

	vector<string> list = Utils_String::Split( outPutSizeData, ":" );
	if ( list.size() >= 2 ) {
		width  = Utils_Convert::String_to_Int(list[0]);
		height = Utils_Convert::String_to_Int(list[1]);
	}

	result = new Board ( width, height );

	return result;
}

//=========================================================
// デストラクタ
//=========================================================
Board::~Board() {

	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];

			delete grid;
		}

	}

}

//=========================================================
// 更新
//=========================================================
void Board::Update(sf::RenderWindow &window, sf::Vector2f pos) {

	_base.setPosition( pos );

	sf::Vector2f drawOffSet_FromBoardLeftUp = _baseMargin;
	sf::Vector2f currentDrawPos = pos + drawOffSet_FromBoardLeftUp;
	sf::Vector2f drawOffSet_GridtoGrid( 0.0f, 0.0f);

	// 後で触りたいもの
	vector<Grid*> laterGridList;

	// 普通にループ
	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		sf::Vector2f gridSize;

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			laterGridList.push_back( grid );

			grid->SetPosition ( currentDrawPos );
			grid->Update( window );

			grid->ClearCardEffect();
			//grid->ReflectCardEffect();

			gridSize = grid->GetSize();
			currentDrawPos.x += gridSize.x + drawOffSet_GridtoGrid.x;

		}

		currentDrawPos.x = pos.x + drawOffSet_FromBoardLeftUp.x;
		currentDrawPos.y += gridSize.y + drawOffSet_GridtoGrid.y;
	}

	///// カードの影響範囲を反映させる
	for (unsigned int i = 0; i < laterGridList.size(); ++i) {

		laterGridList[i]->ReflectCardEffect();
	}

}

//=========================================================
// 描画
//=========================================================
void Board::Draw(sf::RenderWindow &window) {

	window.draw( _base );

	for (unsigned int y = 0; y < _gridList.size(); ++y) {
		vector<Grid*> row = _gridList[y];
		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			grid->Draw_Base( window );
		}
	}
	for (unsigned int y = 0; y < _gridList.size(); ++y) {
		vector<Grid*> row = _gridList[y];
		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			grid->Draw_Card( window );
		}
	}
	for (unsigned int y = 0; y < _gridList.size(); ++y) {
		vector<Grid*> row = _gridList[y];
		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			grid->Draw_Piece( window );
		}
	}
	for (unsigned int y = 0; y < _gridList.size(); ++y) {
		vector<Grid*> row = _gridList[y];
		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			grid->Draw_Flash( window );
		}
	}
}

//=========================================================
// この位置のマス目をください。
//=========================================================
Grid* Board::GetGrid( int x, int y ) {

	Grid *result = NULL;

	if ( y < _gridList.size() ) {

		vector<Grid*> row = _gridList[y];

		if ( x < row.size() ) {
			result = row[x];
		}
	}

	return result;
}

//=========================================================
// マウスが乗っているグリッドをください
//=========================================================
Grid* Board::GetMouseOverGrid() {

	Grid *result = NULL;

	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];

			if (grid->IsMouseOver() ) {
				result = grid;
				break;
			}

		}
		if ( result != NULL ) {
			break;
		}
	}

	return result;
}

//==========================================================
// マウスが乗っている座標を教えてください
//==========================================================
sf::Vector2i Board::GetMouseOverPosition() {

	sf::Vector2i result = sf::Vector2i(-1, -1);

	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		bool isEnd = false;
		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];

			if (grid->IsMouseOver() ) {
				result.x = x;
				result.y = y;
				isEnd = true;
				break;
			}

		}
		if ( isEnd ) {
			break;
		}
	}

	return result;
}

//==========================================================
// 全駒が欲しいんです。
//==========================================================
vector<Piece*> Board::GetAllPieces() {
	
	vector<Piece*> result;
	
	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			
			if ( grid != NULL ) {
				
				Piece *piece = grid->GetPiece();
				
				if ( piece != NULL ) {
					result.push_back(piece);
				}
			}
		}
	}	
	
	
	return result;
}

//==========================================================
// 全カードが欲しいんです。
//==========================================================
vector<Card*> Board::GetAllCards() {
	
	vector<Card*> result;
	
	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];
			
			if ( grid != NULL ) {
				
				Card *card = grid->GetCard();
				
				if ( card != NULL ) {
					result.push_back(card);
				}
			}
		}
	}	
	
	return result;
}
	

//==========================================================
// ファイルに保存するための文字列出力
//==========================================================
string Board::Output() {

	string result = "";

	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		if ( y > 0 ) { result += "|"; }
		string rowString = "";
		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];

			if ( x > 0 ) { rowString += ","; }
			string gridString = "";

			if ( grid == NULL ) {

				gridString = "";

			} else {

				Piece *piece = grid->GetPiece();
				Card  *card  = grid->GetCard();

				gridString += "PIECE:";
				if ( piece == NULL ) {
					gridString += "VOID";
				}
				else if ( piece->GetColor() == Piece::Color::BLACK ) {
					gridString += "BLACK";
				}
				else if ( piece->GetColor() == Piece::Color::WHITE ) {
					gridString += "WHITE";
				}

				gridString += "/CARD:";
				if (card == NULL) {
					gridString += "VOID";
				} else {

					gridString += card->GetProperty().cardName;
					gridString += "/CARD_DIRECTION:" + Utils_Convert::Int_to_String( card->GetCurrentDirection() );
				}
			}

			rowString += gridString;

		}
		result += rowString;
	}


	return result;
}
//==========================================================
// ファイルに保存するための文字列を出力：サイズ
//==========================================================
string Board::Output_Size() {

	/*
		横幅:縦幅
		例）9:9
	*/

	string result = "";

	result = Utils_Convert::Int_to_String( GetWidth()) + ":" +  Utils_Convert::Int_to_String( GetHeight());

	return result;
}
//==========================================================
// ファイルに保存するための文字列を読み込んで再現
//==========================================================
void Board::Read ( string data ) {
	
	///// 一度掃除する
	for (unsigned int y = 0; y < _gridList.size(); ++y) {

		vector<Grid*> row = _gridList[y];

		for (unsigned int x = 0; x < row.size(); ++x) {

			Grid *grid = row[x];

			delete grid->GetPiece();
			//delete grid->GetCard(); // カードを勝手に解放してはいけないようです。解放させてよー。
			CardFactory::Release( grid->GetCard() );

			grid->SetPiece(NULL);
			grid->SetCard(NULL);
		}

	}
	


	/*
		| で行
		, でグリッド
		/ でグリッド内の各要素
	*/
	vector<string> rowList = Utils_String::Split( data, "|" );
	for (unsigned int y = 0; y < rowList.size(); ++y) {

		string row = rowList[y];
		//printf("%s\n", row.c_str());

		vector<string> gridList = Utils_String::Split( row, "," );
		for ( unsigned int x = 0; x < gridList.size(); ++x ) {

			string gridString = gridList[x];

			vector<string> elems = Utils_String::Split( gridString, "/" );
			for ( unsigned int i = 0; i < elems.size(); ++i ) {

				string elem = elems[i];
				vector<string> elemData = Utils_String::Split( elem, ":" );

				string key   = (elemData.size() > 0) ? elemData[0] : "";
				string value = (elemData.size() > 1) ? elemData[1] : "";

				if ( key == "PIECE" ) {
					if ( value == "BLACK" ) {
						Piece *piece = new Piece( Piece::Color::BLACK );
						_gridList[y][x]->SetPiece( piece );
					}
					else if ( value == "WHITE" ) {
						Piece *piece = new Piece( Piece::Color::WHITE );
						_gridList[y][x]->SetPiece( piece );
					}
				}
				else if ( key == "CARD" ) {

					Card *card = CardFactory::GetCard( value );
					if (card != NULL) {

						_gridList[y][x]->SetCard( card );
					}
				}
				else if ( key == "CARD_DIRECTION" ) {

					int direction = Utils_Convert::String_to_Int( value );
					if (_gridList[y][x]->GetCard() != NULL) {
						_gridList[y][x]->GetCard()->SetCurrentDirection( direction );
					}

				}
			}

		}

	}


}

