//=========================================================
//
// Card
//
//=========================================================

#ifndef _CARD_
#define _CARD_

#include <string>


#include <SFML/Graphics.hpp>

#include "IMouseClingableObject.h"
#include "IDrawable.h"

class Grid;
class CardBackDisplayPanel;

using namespace std;

class Card : public IMouseClingableObject, public IDrawable {

public:
	
	//=========================================================
	// static
	//=========================================================
	static CardBackDisplayPanel *_cardBackDisplayPanel;

	//=========================================================
	// グリッドへの影響
	//=========================================================
	struct GridAffection {

		int x;
		int y;
		int movableDirection;

	};

	//=========================================================
	// プロパティ
	//=========================================================
	struct Property {

		string cardLabel;
		string cardName;
		int    processId;

		string    imageFileName;
		sf::Color transparentColor;

		vector<GridAffection> gridAffectionList;

	};

	Card(Property prop);
	~Card();


	void Update(sf::RenderWindow &window, sf::Vector2f pos);
	void Draw(sf::RenderWindow &window);
	void DrawActual(sf::RenderWindow &window);

	void DrawDemonstration(sf::RenderWindow &window, Grid *grid, int direction);
	void DrawHighlight();

	
	void SetPosition( sf::Vector2f pos);
	sf::Vector2f GetCurrentPosition () { return _position_Current;}
	sf::Vector2f GetGoalPosition    () { return _position_Goal;}

	vector< GridAffection > GetAffectionList();
	vector< GridAffection > GetAffectionList( int direction );

	bool IsMouseOver() { return _isMouseOver; }

	Property GetProperty() { return _prop; }

	int GetCurrentDirection() { return _currentDirection; }
	void SetCurrentDirection( int direstion );
	void RollingRight ();
	void RollingLeft  ();

	float GetMouseDrawOffsetX() { return 0.0f; }
	float GetMouseDrawOffsetY() { return 0.0f; }

	void SetOffset ( int x, int y ) { _position_Offset.x = x; _position_Offset.y = y; }
	void SetOffsetX ( int x ) { _position_Offset.x = x; }
	void SetOffsetY ( int y ) { _position_Offset.y = y; }
	sf::Vector2f GetOffset() { return _position_Offset; }

	bool IsMovingStop() { return ( _position_Current == _position_Goal ); }

	void SetAutonomous ( bool arg ) { _isAutonomous = arg; }
	bool IsAutonomous  ()           { return _isAutonomous; }

	///// 右クリックしたときの回転
	void Rolling_By_Operatiion();

	// 現在のびよーんを省略せよ（このときだけ）
	void WarpToGoal() { _isWarpToGoal = true; _isWarpToGoalAngle = true;}
	
private:

	Property _prop;

	sf::Sprite   _sprite;
	sf::Texture  _texture;
	sf::Image   *_image;

	sf::Sprite  _nameLabel_Sprite;
	sf::Texture _nameLabel_Texture;
	sf::Image   _nameLabel;
	
	sf::RectangleShape _shape_highlight;

	sf::Vector2f _position_Goal;    // あるべき位置
	sf::Vector2f _position_Current; // 実際の位置
	sf::Vector2f _position_Offset;  // オフセット

	int _currentDirection; // 現在の向き：正面を「UP」とする。

	int _rotationAngle_Goal;    // あるべき角度
	int _rotationAngle_Current; // 実際の角度

	bool _isAutonomous;
	
	
	bool _isWarpToGoal;
	bool _isWarpToGoalAngle;
	
	bool _isReservedToDrawHighlight;
	void DrawHighlight(sf::RenderWindow &window);


	//=========================================================
	// 影響を及ぼす位置情報を向きで変換
	//=========================================================
	GridAffection ConvertAffectionGrid_by_Direction( GridAffection src, int direction );

	//=========================================================
	// 向きを向きで回転させる
	//=========================================================
	int RollDirection( int direction_src, int direction_move );


	bool _isMouseOver;

	//=========================================================
	// マウスオーバー判定
	//=========================================================
	void MouseOverCheck();

	//=========================================================
	// 右クリックで回転（マウスで保持されている間）
	// 保持されていない場合には、UPに毎回上書き。
	//=========================================================
	void Rolling();


	//=========================================================
	// 自治処理
	//=========================================================
	void AutonomousUpdate ( sf::RenderWindow &window );
};
#endif










