//=========================================================
//
// カードの裏側に説明文が書いてある、というイメージで。
//
//=========================================================

#include "CardBackDisplayPanel.h"
#include "Card.h"
#include "Application.h"
#include "Define.h"
#include "../GUI_SFML/ImageManager.h"
#include <SFML/Graphics.hpp>


//=========================================================
// コンストラクタ
//=========================================================
CardBackDisplayPanel::CardBackDisplayPanel() {

	_isShow = false;
	_position = sf::Vector2f( 300.0f, 100.0f );

	_panelBase_Shape = sf::RectangleShape (sf::Vector2f( 200.0f, 200.0f));
	_panelBase_Shape.setFillColor(sf::Color(200,200,255,200));
	_panelBase_Shape.setOutlineColor(sf::Color(0,0,0,0));
	_panelBase_Shape.setOutlineThickness(1.0f);

	_panelBase_Shape.setPosition( _position );

}
//=========================================================
// デストラクタ
//=========================================================
CardBackDisplayPanel::~CardBackDisplayPanel() {

}

//=========================================================
// カード（の設定）をセット
//=========================================================
void CardBackDisplayPanel::SetCardProperty(Card::Property cardProp) {

	_gridList.clear();
	_currentShowingCardProp = cardProp;

	///// とりあえず 9×9ぐらいでこしらえる。
	for (int y = 0; y < 9; ++y) {
		for (int x = 0; x < 9; ++x) {

			_gridList[y][x] = GenerateMiniGrid_Blank();

		}
	}

	// 真ん中をカード、ということにする
	_gridList[4][4] = GenerateMiniGrid_Card();


	for ( unsigned int i = 0; i < cardProp.gridAffectionList.size(); ++i) {

		int x = 4 + cardProp.gridAffectionList[i].x;
		int y = 4 + cardProp.gridAffectionList[i].y;

		if ( x >=0 && x < 9 && y >= 0 && y < 9) {
			_gridList[y][x] = GenerateMiniGrid_Througable( cardProp.gridAffectionList[i].movableDirection );
		}
	}


}

//=========================================================
// 表示
//=========================================================
void CardBackDisplayPanel::Show() {

	_isShow = true;
}
//=========================================================
// 隠す
//=========================================================
void CardBackDisplayPanel::Hide() {
	_isShow = false;
}
//=========================================================
// 今表示しているカードのカード名を取得
//=========================================================
string CardBackDisplayPanel::GetCurrentShowingCardName() {

	string result = "";
	if ( _isShow ) {

		result = _currentShowingCardProp.cardName;
	}
	
	return result;
}

//=========================================================
// 更新
//=========================================================
void CardBackDisplayPanel::Update(sf::RenderWindow &window) {

	if (!_isShow) {
		return;
	}

	sf::Vector2f offset( 46.0f, 46.0f );

	int x = 0;
	int y = 0;
	for ( map<int, map<int, MiniGrid>>::iterator itY = _gridList.begin(); itY != _gridList.end() ;++itY ) {

		map<int, MiniGrid> row = itY->second;
		y = itY->first;

		for ( map<int, MiniGrid>::iterator itX = row.begin(); itX != row.end() ;++itX ) {

			MiniGrid miniGrid = itX->second;
			x = itX->first;

			_gridList[y][x].shape.setPosition( _position + offset + sf::Vector2f( x * 12.0f, y * 12.0f )  );

			if (_gridList[y][x].moveDirectionImage != NULL) {

				_gridList[y][x].moveDirectionSprite.setPosition(
					_gridList[y][x].shape.getPosition() + _gridList[y][x].size / 2.0f
				);
			}
		}
	}

	///// マウス操作で表示終了
	if (Application::GetInstance()->_mouseState_Middle == MOUSE::BUTTON::STATE::RELEASE) {

		Application::GetInstance()->_mouseState_Middle = MOUSE::BUTTON::STATE::NONE;
		Hide();
	}


}
//=========================================================
// 描画
//=========================================================
void CardBackDisplayPanel::Draw(sf::RenderWindow &window) {

	Application::GetInstance()->AddLaterDraw( this );
}
void CardBackDisplayPanel::DrawActual(sf::RenderWindow &window) {

	if (!_isShow) {
		return;
	}

	window.draw(_panelBase_Shape);

	for ( map<int, map<int, MiniGrid>>::iterator itY = _gridList.begin(); itY != _gridList.end() ;++itY ) {

		map<int, MiniGrid> row = itY->second;
		for ( map<int, MiniGrid>::iterator itX = row.begin(); itX != row.end() ;++itX ) {

			MiniGrid miniGrid = itX->second;

			window.draw( miniGrid.shape );

			if ( miniGrid.moveDirectionImage != NULL ) {
				window.draw( miniGrid.moveDirectionSprite );
			}
		}
	}
}

//=========================================================
// MiniGrid を生成
//=========================================================
CardBackDisplayPanel::MiniGrid CardBackDisplayPanel::GenerateMiniGrid_Blank() {

	CardBackDisplayPanel::MiniGrid result;

	result.shape = sf::RectangleShape ( result.size );
	result.shape.setFillColor ( sf::Color(255,255,255,255) );
	result.shape.setOutlineColor ( sf::Color(0,0,0,255) );
	result.shape.setOutlineThickness ( 1.0f );

	return result;
}
CardBackDisplayPanel::MiniGrid CardBackDisplayPanel::GenerateMiniGrid_Card() {

	CardBackDisplayPanel::MiniGrid result;


	result.shape = sf::RectangleShape ( result.size );
	result.shape.setFillColor ( sf::Color(255,50,50,255) );
	result.shape.setOutlineColor ( sf::Color(0,0,0,255) );
	result.shape.setOutlineThickness ( 1.0f );

	return result;
}
CardBackDisplayPanel::MiniGrid CardBackDisplayPanel::GenerateMiniGrid_Througable( int moveDirection ) {

	CardBackDisplayPanel::MiniGrid result;

	result.shape = sf::RectangleShape ( result.size );
	result.shape.setFillColor ( sf::Color(255,255,50,255) );
	result.shape.setOutlineColor ( sf::Color(0,0,0,255) );
	result.shape.setOutlineThickness ( 1.0f );


	if ( moveDirection != DIRECTION::NONE ) {

		result.moveDirectionImage = ImageManager::GetImage( "./image/moveDirectionMini.png", sf::Color(0,255,0));
		result.moveDirectionTexture.loadFromImage( *(result.moveDirectionImage) );
		result.moveDirectionSprite.setTexture( result.moveDirectionTexture );

		result.moveDirectionSprite.setOrigin(
			sf::Vector2f(
			result.moveDirectionImage->getSize().x / 2.0f,
			result.moveDirectionImage->getSize().y / 2.0f
			)
		);

		if ( moveDirection == DIRECTION::LEFT  ) { result.moveDirectionSprite.setRotation(90);  }
		if ( moveDirection == DIRECTION::DOWN  ) { result.moveDirectionSprite.setRotation(180); }
		if ( moveDirection == DIRECTION::RIGHT ) { result.moveDirectionSprite.setRotation(270); }
		if ( moveDirection == DIRECTION::UP    ) { result.moveDirectionSprite.setRotation(0);   }

	}


	return result;
}

