	//=========================================================
//
// カード置き場
//
//=========================================================

#include <SFML/Graphics.hpp>
#include <vector>

#include "../ScriptEngine/Application.h"

using namespace std;

#include "IMouseClingableObject.h"
#include "Card.h"
#include "CardBox.h"
#include "CardFactory.h"
#include "Define.h"
#include "../Utils/Utils_Others.h"
#include "../Utils/Utils_String.h"

//=========================================================
// コンストラクタ
//=========================================================
CardBox::CardBox ( bool isDisplayReverse ) {

	_boxSize = sf::Vector2f( 150.0f, 320.0f );

	_box = sf::RectangleShape ( _boxSize );
	_box.setFillColor ( sf::Color(255,200,255,200) );
	_box.setOutlineColor ( sf::Color(0,0,0,0) );
	_box.setOutlineThickness ( 0.0f );

	_isMouseOver   = false;
	_isAutonomous  = true;
	_mouseOverCard = NULL;
	_isDisplayReverse = isDisplayReverse;
}

//=========================================================
// デストラクタ
//=========================================================
CardBox::~CardBox() {

}

//=========================================================
// 更新
//=========================================================
void CardBox::Update(sf::RenderWindow &window, sf::Vector2f pos) {

	_box.setPosition( pos );

	MouseOverCheck();

	sf::Vector2f currentCardPos = ( !_isDisplayReverse) ? 
		pos + sf::Vector2f(30.0f, 30.0f) :
		pos + sf::Vector2f( _boxSize.x - 30.0f, 30.0f);

	_mouseOverCard = NULL;

	//=========================================================
	// 各カードを更新
	//=========================================================
	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		sf::Vector2f currentPosition_RowOffset( 0.0f, 0.0f );
		for ( unsigned int n = 0; n < _cardFolderList[i].cardList.size(); ++n) {


			_cardFolderList[i].cardList[n]->Update( window, currentCardPos + currentPosition_RowOffset);

			currentPosition_RowOffset += ( !_isDisplayReverse ) ?
				sf::Vector2f( 10.0f, 0.0f ) :
				sf::Vector2f( -10.0f, 0.0f );

			// マウスが乗っていたら、覚えておく。
			if ( _cardFolderList[i].cardList[n]->IsMouseOver() ) {
				_mouseOverCard = _cardFolderList[i].cardList[n];
			}
		}
		currentCardPos += sf::Vector2f( 0.0f, 60.0f );
	}


	AutonomousUpdate ( window );

}

//=========================================================
// 自治処理
//=========================================================
void CardBox::AutonomousUpdate ( sf::RenderWindow &window ) {

	if ( !_isAutonomous ) {
		return;
	}


	//=========================================================
	// マウス操作でカードを置いたり取ったり。
	//=========================================================
	// この駒置き場にマウスが乗っていて、
	if (_isMouseOver) {

		// マウスに駒がくっついている
		if (Application::GetInstance()->IsMouseClingObj_Card()) {

			// 左押下
			if (Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

				// マウスから駒を放して、この駒置き場に追加
				IMouseClingableObject *obj = Application::GetInstance()->DrawMouseClingObj();
				Card *card = dynamic_cast<Card*>( obj );

				AddCard(card);
			}

		// マウスに空きがある
		} else if (!Application::GetInstance()->IsMouseClingObjExist()){

			// 駒置き場内の駒にマウスが乗っていて、左押下
			if ( _mouseOverCard != NULL && Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

				// マウスにくっつけて、その駒は駒置き場から放す
				Application::GetInstance()->AddMouseClingObj( _mouseOverCard );
				RemoveCard ( _mouseOverCard );
			}

		}

	}

}


//=========================================================
// 描画
//=========================================================
void CardBox::Draw(sf::RenderWindow &window) {

	window.draw(_box);

	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		for ( unsigned int n = 0; n < _cardFolderList[i].cardList.size(); ++n) {
			_cardFolderList[i].cardList[n]->DrawActual( window );
		}
	}
}

//=========================================================
// Add AnotherCard To This Object
//=========================================================
void CardBox::AddCard(Card *obj) {

	bool isAdd = false;

	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		if ( _cardFolderList[i].cardList.size() > 0 ) {

			if ( _cardFolderList[i].cardList[0]->GetProperty().processId == obj->GetProperty().processId) {

				_cardFolderList[i].cardList.push_back( obj );
				isAdd = true;
				break;
			}
		}
	}

	if (!isAdd) {

		CardFolder newFolder;
		newFolder.cardList.push_back( obj );

		_cardFolderList.push_back( newFolder );
	}

}

//=========================================================
// 全カードのIDリスト
//=========================================================
vector<int> CardBox::GetAllCardIdList() {

	vector<int> result;

	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		for ( unsigned int n = 0; n < _cardFolderList[i].cardList.size(); ++n) {
			
			result.push_back( _cardFolderList[i].cardList[n]->GetProperty().processId );
		}
	}

	return result;
}

//=========================================================
// 一枚引く
//=========================================================
Card* CardBox::TakeCard( int cardId ) {

	Card *result = NULL;

	//for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {
	for ( vector<CardFolder>::iterator it_Folder = _cardFolderList.begin(); it_Folder != _cardFolderList.end(); ++it_Folder ) {

		for ( vector<Card*>::iterator it = (*(it_Folder)).cardList.begin(); it != (*(it_Folder)).cardList.end(); ++it) {
			
			Card *card = *it;
			if ( card->GetProperty().processId == cardId) {

				result = card;
				(*(it_Folder)).cardList.erase( it );

				if ( (*(it_Folder)).cardList.size() == 0 ) {
					_cardFolderList.erase(it_Folder);
				}
				break;
			}
		}

		if (result != NULL ) {
			break;
		}
	}

	return result;
}
//=========================================================
// 一枚引く：オブジェクト指定版
//=========================================================
Card* CardBox::TakeCard( Card *obj ) {

	Card *result = NULL;

	//for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {
	for ( vector<CardFolder>::iterator it_Folder = _cardFolderList.begin(); it_Folder != _cardFolderList.end(); ++it_Folder ) {

		for ( vector<Card*>::iterator it = (*(it_Folder)).cardList.begin(); it != (*(it_Folder)).cardList.end(); ++it) {
			
			Card *card = *it;
			if ( card == obj) {

				result = card;
				(*(it_Folder)).cardList.erase( it );

				if ( (*(it_Folder)).cardList.size() == 0 ) {
					_cardFolderList.erase(it_Folder);
				}
				break;
			}
		}

		if (result != NULL ) {
			break;
		}
	}

	return result;
}

//=========================================================
// マウスオーバー判定
//=========================================================
void CardBox::MouseOverCheck() {

	int mouseX = Application::GetInstance()->_mouseX;
	int mouseY = Application::GetInstance()->_mouseY;

	sf::Vector2f position = _box.getPosition();

	if
	(
		position.x < mouseX && mouseX < (position.x + _boxSize.x) 
		&&
		position.y < mouseY && mouseY < (position.y + _boxSize.y) 
	) 
	{
		//printf("mouse on\n");

		_isMouseOver = true;

	} else {
		_isMouseOver = false;
	}
}

//=========================================================
// カードを消す（オブジェクトを指定）
//=========================================================
void CardBox::RemoveCard( Card *obj ) {

	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		bool isErase = false;

		for ( vector<Card*>::iterator it = _cardFolderList[i].cardList.begin(); it != _cardFolderList[i].cardList.end(); ++it) {

			Card *card = *(it);

			if (card == obj) {
			
				_cardFolderList[i].cardList.erase(it);
				isErase = true;
				break;
			}
		}

		if ( isErase ) {

			if (_cardFolderList[i].cardList.size() == 0) {

				UTILS_OTHERS::deleteFromVector<CardFolder>(_cardFolderList, i);
			}
			break;
		}
	}


}

//==========================================================
// ファイルに保存するための文字列を出力
//==========================================================
string CardBox:: Output() {

	string result = "";

	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {


		for ( vector<Card*>::iterator it = _cardFolderList[i].cardList.begin(); it != _cardFolderList[i].cardList.end(); ++it) {

			Card *card = *(it);

			if ( result != "" ) {
				result += ",";
			}
			result += card->GetProperty().cardName;
		}
	}

	return result;
}


//==========================================================
// ファイルに保存するための文字列から再現
//==========================================================
void CardBox::Read( string data ) {

	///// 一度掃除
	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		bool isErase = false;

		for ( vector<Card*>::iterator it = _cardFolderList[i].cardList.begin(); it != _cardFolderList[i].cardList.end(); ++it) {

			Card *card = *(it);
			CardFactory::Release( card );

		}

	}
	_cardFolderList.clear();

	///// 入れなおす
	vector<string> cardNameList = Utils_String::Split( data, "," );
	for ( unsigned int i = 0; i < cardNameList.size(); ++i ) {


		Card *card = CardFactory::GetCard( cardNameList[i] );
		if ( card != NULL ) {

			AddCard( card );
		}
	}

}

//=========================================================
// 全カードを取得
//=========================================================
vector<Card*> CardBox::GetAllCards() {

	vector<Card*> result;

	for ( unsigned int i = 0; i < _cardFolderList.size(); ++i ) {

		for ( unsigned int n = 0; n < _cardFolderList[i].cardList.size(); ++n) {
			
			result.push_back( _cardFolderList[i].cardList[n] );
		}
	}

	return result;
}










