//=========================================================
//
// Grid
//
//=========================================================

#include "Grid.h"
#include "Application.h"
#include "Piece.h"
#include "Card.h"
#include "Define.h"

#include "../GUI_SFML/ImageManager.h"

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>



//=========================================================
// コンストラクタ
//=========================================================
Grid::Grid() {

	_position = sf::Vector2f( 0.0f, 0.0f );
	_piece = NULL;
	_card  = NULL;

	///// _shape
	_shape = sf::RectangleShape (_prop.size);
	_shape.setFillColor (_prop.color_fill);
	_shape.setOutlineColor (_prop.color_line);
	_shape.setOutlineThickness (1.0f);

	///// _shape_Cursor
	_shape_Cursor = sf::RectangleShape ( _prop.size - sf::Vector2f( 5.0f,  5.0f) );
	_shape_Cursor.setFillColor (sf::Color(0,0,0,0));
	_shape_Cursor.setOutlineColor (sf::Color(255,0,0,255));
	_shape_Cursor.setOutlineThickness (2.0f);
	_shape_Cursor.setOrigin ( -2, -2 );

	///// _shape_Flash
	_shape_Flash = sf::RectangleShape ( _prop.size );
	_shape_Flash.setFillColor (sf::Color(255,255,200,200));
	_shape_Flash.setOutlineColor (sf::Color(0,0,0,0));
	_shape_Flash.setOutlineThickness (0.0f);

	///// _shape_Selectable
	_shape_Selectable = sf::RectangleShape ( _prop.size );
	_shape_Selectable.setFillColor ( sf::Color(100,255,100,100));
	_shape_Selectable.setOutlineColor (sf::Color(0,0,0,0));
	_shape_Selectable.setOutlineThickness (0.0f);

	///// _shape_Selected
	_shape_Selected = sf::RectangleShape ( _prop.size - sf::Vector2f( 5.0f,  5.0f));
	_shape_Selected.setFillColor ( sf::Color(0,0,0,0));
	_shape_Selected.setOutlineColor (sf::Color(255,0,0,255));
	_shape_Selected.setOutlineThickness (2.0f);
	_shape_Selected.setOrigin ( -2, -2 );

	///// _shape_Throuable
	_shape_Throuable = sf::RectangleShape ( _prop.size - sf::Vector2f( 5.0f,  5.0f));
	_shape_Throuable.setFillColor ( sf::Color(255,255,200,200));
	_shape_Throuable.setOutlineColor (sf::Color(0,0,0,0));
	_shape_Throuable.setOutlineThickness (0.0f);
	_shape_Throuable.setOrigin ( -2, -2 );

	///// _shape_AffectedDisplay
	_shape_AffectedDisplay = sf::RectangleShape ( _prop.size - sf::Vector2f( 5.0f,  5.0f));
	_shape_AffectedDisplay.setFillColor ( sf::Color(0,0,0,0));
	_shape_AffectedDisplay.setOutlineColor (sf::Color(255,0,0,255));
	_shape_AffectedDisplay.setOutlineThickness (1.0f);
	_shape_AffectedDisplay.setOrigin ( -2, -2 );
	
	
	///// 最終着手のハイライト
	_shape_LastMove = sf::RectangleShape ( _prop.size - sf::Vector2f( 3.0f,  3.0f) );
	_shape_LastMove.setFillColor ( sf::Color(0,0,0,0));
	_shape_LastMove.setOutlineColor (sf::Color(0,0,255,255));
	_shape_LastMove.setOutlineThickness (1.0f);
	_shape_LastMove.setOrigin ( -1, -1 );

	_isMouseOver         = false;
	_state_isThrouable   = false;
	_state_isMustGoRight = false;
	_state_isMustGoLeft  = false;
	_state_isMustGoUp    = false;
	_state_isMustGoDown  = false;
	_state_isAffectedDisplay = false;
	_state_isSelectable  = false;
	_state_isSelected    = false;


	_moveDrectionImage = ImageManager::GetImage( "./image/gameBoard/moveDirection.png", sf::Color(0,255,0) );
	_moveDrectionTexture.loadFromImage ( *_moveDrectionImage );
	_moveDirectionSprite.setTexture( _moveDrectionTexture );
	_moveDirectionSprite.setOrigin( sf::Vector2f( _moveDrectionImage->getSize().x / 2.0f, _moveDrectionImage->getSize().y / 2.0f ) );


	_isAutonomous = true;
}

//=========================================================
// デストラクタ
//=========================================================
Grid::~Grid() {

	// 駒、カードが乗っていても、解放しない。

}

//=========================================================
// 更新
//=========================================================
void Grid::Update(sf::RenderWindow &window) {

	_shape           .setPosition( _position );
	_shape_Cursor    .setPosition( _position );
	_shape_Flash     .setPosition( _position );
	_shape_Selectable.setPosition( _position );
	_shape_Selected  .setPosition( _position );
	_shape_Throuable .setPosition( _position );
	_shape_LastMove  .setPosition( _position );
	_shape_AffectedDisplay.setPosition( _position );
	_moveDirectionSprite.setPosition( _position + _prop.size / 2.0f );

	if      ( _state_isMustGoRight ) { _moveDirectionSprite.setRotation ( -270 ); }
	else if ( _state_isMustGoDown  ) { _moveDirectionSprite.setRotation (  180 ); }
	else if ( _state_isMustGoLeft  ) { _moveDirectionSprite.setRotation (  -90 ); }
	else if ( _state_isMustGoUp    ) { _moveDirectionSprite.setRotation (    0 ); }

	MouseOverCheck();

	///// 自治更新処理
	AutonomousUpdate ( window );


	if (_piece != NULL) {
		_piece->Update(window, _position + _prop.size / 2.0f);
	}
	if (_card != NULL) {
		_card->Update(window, _position + _prop.size / 2.0f);
	}

}

//=========================================================
// 自治処理
//=========================================================
void Grid::AutonomousUpdate ( sf::RenderWindow &window ) {

	if ( !_isAutonomous ) {
		return;
	}

	/*
		マウスが乗っていて、
		マウスにくっついているモノがPieceならば、
		マウスにくっついているモノがCardならば、
	*/
	if ( _isMouseOver ) {

		// このマス目に、今、駒なし
		if ( _piece == NULL ) {

			// マウスに駒がくっついている
			if (Application::GetInstance()->IsMouseClingObj_Piece()) {

				// 左押下
				if (Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

					// マウスから駒を放して、このマス目に設置
					IMouseClingableObject *piece = Application::GetInstance()->DrawMouseClingObj();
					_piece = dynamic_cast<Piece*>(piece);

					// カードもこのマス目にある場合、操作がかぶらないように。
					Application::GetInstance()->_mouseState_Left = MOUSE::BUTTON::STATE::NONE;
				}

			}

		// このマス目に、今、駒あり
		} else {

			// マウスに何もくっつけてない
			if (!Application::GetInstance()->IsMouseClingObjExist()) {

				// 駒にマウスを乗せている
				if (_piece->IsMouseOver()) {

					// 左リリース
					if (Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

						// このマス目から駒を放して、マウスにくっつける
						Application::GetInstance()->AddMouseClingObj( _piece );
						_piece = NULL;
					}

				}
			}
		}


		// このマス目に、今、カードなし
		if ( _card == NULL ) {

			// マウスにカードがくっついている
			if (Application::GetInstance()->IsMouseClingObj_Card()) {

				// 左押下
				if (Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

					// マウスからカードを放して、このマス目に設置
					IMouseClingableObject *card = Application::GetInstance()->DrawMouseClingObj();
					_card = dynamic_cast<Card*>(card);
				}

			}

		// このマス目に、今、カードあり
		} else {

			// マウスに何もくっつけてない
			if (!Application::GetInstance()->IsMouseClingObjExist()) {

				// カードにマウスを乗せている
				if (_card->IsMouseOver()) {

					// 左リリース
					if (Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE) {

						// このマス目からカードを放して、マウスにくっつける
						Application::GetInstance()->AddMouseClingObj( _card );
						_card = NULL;
					}

				}
			}
		}
	}
}


//=========================================================
// マウスオーバー判定
//=========================================================
void Grid::MouseOverCheck() {

	int mouseX = Application::GetInstance()->_mouseX;
	int mouseY = Application::GetInstance()->_mouseY;

	if
	(
		_position.x < mouseX && mouseX < (_position.x + _prop.size.x) 
		&&
		_position.y < mouseY && mouseY < (_position.y + _prop.size.y) 
	) 
	{
		//printf("mouse on\n");

		_isMouseOver = true;

	} else {
		_isMouseOver = false;
	}
}


//=========================================================
// カードの効果を反映させる
//=========================================================
void Grid::ReflectCardEffect() {

	Card *useCard = _card;
	if (useCard == NULL) {

		if (_isMouseOver && Application::GetInstance()->IsMouseClingObj_Card()) {
			useCard = dynamic_cast<Card*>(Application::GetInstance()->RefferMouseClingObj());
		}
	}

	if (useCard != NULL) {

		/*
			カードの影響範囲のグリッドに、影響を反映させる。

			カードの移動が定着してから？
			それはカード側で、移動終了するまでは影響範囲を返さないようにしましょう。
		*/
		vector<Card::GridAffection> affectionList = useCard->GetAffectionList();

		for (unsigned int i = 0; i < affectionList.size(); ++i) {

			Grid *relativeGrid = GetRelativeGrid( affectionList[i].x, affectionList[i].y );

			//printf("%p (%d, %d)\n", relativeGrid, affectionList[i].x, affectionList[i].y);

			if (relativeGrid != NULL) {
				relativeGrid->SetThrouable(true);
				relativeGrid->SetMustGoDirection( affectionList[i].movableDirection );

				if (_isMouseOver) {
					relativeGrid->SetAffectedDisplay(true);
				}
			}

		}
	}
}

//=========================================================
// カードの効果をクリアする
//=========================================================
void Grid::ClearCardEffect() {

	/*
		自分の面倒だけみる。
	*/
	_state_isThrouable   = false;
	_state_isMustGoRight = false;
	_state_isMustGoLeft  = false;
	_state_isMustGoUp    = false;
	_state_isMustGoDown  = false;
	_state_isAffectedDisplay = false;
}


//=========================================================
// 描画
//=========================================================
void Grid::Draw(sf::RenderWindow &window) {

	///// 下地
	Draw_Base ( window );

	///// アクター
	Draw_Card ( window );
	Draw_Piece ( window );

	///// かぶせたいフラッシュ
	Draw_Flash ( window );
}

//=========================================================
// 描画：指し手の仮表示（カードの影響範囲）
//=========================================================
void Grid::Draw_CardEffectDemonstration (sf::RenderWindow &window, int direction) {
	
	sf::Color color_org  =  _shape_Throuable.getFillColor();
	sf::Color color_temp =  color_org; color_temp.a = 150;
	_shape_Throuable.setFillColor(color_temp);

	window.draw( _shape_Throuable );

	_shape_Throuable.setFillColor(color_org);

	if ( direction != DIRECTION::NONE ) {
		
		color_org  = _moveDirectionSprite.getColor();
		color_temp =  color_org; color_temp.a = 150;
		float rotation_org = _moveDirectionSprite.getRotation();
		float rotation_temp = 0;
		
		if      (direction == DIRECTION::LEFT ) { rotation_temp  = 270;}
		else if (direction == DIRECTION::RIGHT) { rotation_temp  = 90;}
		else if (direction == DIRECTION::DOWN ) { rotation_temp  = 180;}
		else if (direction == DIRECTION::UP   ) { rotation_temp  = 0;}

		_moveDirectionSprite.setColor(color_temp);
		_moveDirectionSprite.setRotation(rotation_temp);
		
		window.draw( _moveDirectionSprite );
		
		_moveDirectionSprite.setColor(color_org);

		
	}
	//if (_state_isAffectedDisplay) {
	//	window.draw( _shape_AffectedDisplay );
	//}
	
}



//=========================================================
// 描画：下地
//=========================================================
void Grid::Draw_Base  (sf::RenderWindow &window) {

	window.draw( _shape );

	if (_state_isThrouable) {
		window.draw( _shape_Throuable );

		if ( _state_isMustGoRight || _state_isMustGoLeft || _state_isMustGoUp || _state_isMustGoDown) {
			window.draw( _moveDirectionSprite );
		}

	}
	if (_state_isAffectedDisplay) {
		window.draw( _shape_AffectedDisplay );
	}

	// マウスオーバーカーソル
	if (_isMouseOver) {
		window.draw( _shape_Cursor );
	}
}
//=========================================================
// 描画：カード
//=========================================================
void Grid::Draw_Card (sf::RenderWindow &window) {

	// カード
	if (_card != NULL) {
		_card->DrawActual(window);
	}
}
//=========================================================
// 描画：駒
//=========================================================
void Grid::Draw_Piece (sf::RenderWindow &window) {

	// 駒
	if (_piece != NULL) {
		_piece->DrawActual(window);
	}
}
//=========================================================
// 描画：フラッシュ
//=========================================================
void Grid::Draw_Flash (sf::RenderWindow &window) {

	// 選択可能
	if (_state_isSelectable ) {
		window.draw( _shape_Selectable );
	}
	// 選択されている
	if (_state_isSelected ) {
		window.draw( _shape_Selected );
	}
}

//=========================================================
// 描画：最終着手のハイライト
//=========================================================
void Grid::Draw_LastMoveHighlight (sf::RenderWindow &window) {
	
	window.draw( _shape_LastMove );

	
}


//=========================================================
// 相対位置のお仲間を取得
//=========================================================
Grid* Grid::GetRelativeGrid( int x, int y ) {

	Grid *result = NULL;

	map<int, map<int, Grid*>>::iterator it_Y = _relativeGridList.find( y );
	if ( it_Y != _relativeGridList.end() ) {

		map<int, Grid*>::iterator it_X = _relativeGridList[y].find( x );
	
		if ( it_X != _relativeGridList[y].end() ) {

			result = it_X->second;
		}

	}

	return result;
}
//=========================================================
// 移動強制方向をセット
//=========================================================
void Grid::SetMustGoDirection( int direction ) {

	// 一度全部クリアするのですね？
	_state_isMustGoLeft  = false;
	_state_isMustGoRight = false;
	_state_isMustGoUp    = false;
	_state_isMustGoDown  = false;

	if      (direction == DIRECTION::LEFT ) { _state_isMustGoLeft  = true;}
	else if (direction == DIRECTION::RIGHT) { _state_isMustGoRight = true;}
	else if (direction == DIRECTION::DOWN ) { _state_isMustGoDown  = true;}
	else if (direction == DIRECTION::UP   ) { _state_isMustGoUp    = true;}

}


