//=========================================================
//
// Grid
//
//=========================================================

#ifndef _GRID_
#define _GRID_

#include <map>

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>

using namespace std;

class Piece;
class Card;

class Grid {

public:
	/*
	//=========================================================
	// 定数：マス目の状態
	//=========================================================
	struct State {
		static const int FLAT        = 0; // 無印
		static const int THROUGHABLE = 1; // 通行可能：全方向
		static const int THROUGHABLE_RIGHT = 2;  // 通行可能：右へ
		static const int THROUGHABLE_DOWN  = 4;  // 通行可能：下へ
		static const int THROUGHABLE_LEFT  = 8;  // 通行可能：左へ
		static const int THROUGHABLE_UP    = 16; // 通行可能：上へ
	};
	*/

	//=========================================================
	// プロパティ
	//=========================================================
	struct Property {
		
		sf::Vector2f size;
		sf::Color color_line;
		sf::Color color_fill;

		int decorationValue;

		Property() {
			size.x = 38;
			size.y = 38;
			//decorationValue = GRID::DECORATION::NONE;
			color_line = sf::Color( 0,0,0,255 );
			color_fill = sf::Color( 240,240,240,255 );
		}

	};


	Grid();
	~Grid();

	void Update(sf::RenderWindow &window);
	void Draw(sf::RenderWindow &window);

	void Draw_Base  (sf::RenderWindow &window);
	void Draw_Card  (sf::RenderWindow &window);
	void Draw_Piece (sf::RenderWindow &window);
	void Draw_Flash (sf::RenderWindow &window);
	void Draw_LastMoveHighlight (sf::RenderWindow &window);

	
	void Draw_CardEffectDemonstration (sf::RenderWindow &window, int direction);

	///// カードの効果を反映させる
	void ReflectCardEffect();

	///// カードの効果をクリアする
	void ClearCardEffect();


	sf::Vector2f GetSize()     { return _prop.size;   }
	sf::Vector2f GetPosition() { return _position;    }
	sf::Vector2f GetCenterPosition() { return _position + (_prop.size / 2.0f);}
	bool         IsMouseOver() { return _isMouseOver; }

	void SetPosition (sf::Vector2f pos) { _position   = pos; }
	void SetPositionX( float arg      ) { _position.x = arg; }
	void SetPositionY( float arg      ) { _position.y = arg; }

	void   SetPiece( Piece *obj) { _piece = obj; }
	Piece* GetPiece()            { return _piece; }

	void  SetCard( Card *obj) { _card = obj; }
	Card* GetCard()           { return _card; }

	map<int,map<int, Grid*>> _relativeGridList;

	bool IsThrouable() { return _state_isThrouable; }
	void SetThrouable( bool arg ) { _state_isThrouable = arg;}
	void SetMustGoDirection( int direction );
	void SetAffectedDisplay ( bool arg ) { _state_isAffectedDisplay = arg; }
	void SetSelectable      ( bool arg ) { _state_isSelectable      = arg; }
	void SetSelected        ( bool arg ) { _state_isSelected        = arg; }

	bool IsMustGoLeft () { return _state_isMustGoLeft;  }
	bool IsMustGoRight() { return _state_isMustGoRight; }
	bool IsMustGoUp   () { return _state_isMustGoUp;    }
	bool IsMustGoDown () { return _state_isMustGoDown;  }

	void SetAutonomous ( bool arg ) { _isAutonomous = arg; }
	bool IsAutonomous  ()           { return _isAutonomous; }

	//=========================================================
	// 相対位置のお仲間を取得
	//=========================================================
	Grid* GetRelativeGrid( int x, int y );

private:

	Property _prop;
	sf::Vector2f _position;

	bool _isMouseOver;
	bool _state_isThrouable;
	bool _state_isMustGoRight;
	bool _state_isMustGoLeft;
	bool _state_isMustGoUp;
	bool _state_isMustGoDown;
	bool _state_isAffectedDisplay;
	bool _state_isSelectable;
	bool _state_isSelected;

	sf::RectangleShape _shape;
	sf::RectangleShape _shape_Cursor;
	sf::RectangleShape _shape_Flash;
	sf::RectangleShape _shape_Selectable;
	sf::RectangleShape _shape_Selected;
	sf::RectangleShape _shape_LastMove;

	sf::RectangleShape _shape_Throuable;
	sf::RectangleShape _shape_AffectedDisplay;

	sf::Image *_moveDrectionImage;
	sf::Texture _moveDrectionTexture;
	sf::Sprite _moveDirectionSprite;

	Piece *_piece;
	Card *_card;

	bool _isAutonomous;

	//=========================================================
	// マウスオーバー判定
	//=========================================================
	void MouseOverCheck();

	//=========================================================
	// 自治処理
	//=========================================================
	void AutonomousUpdate ( sf::RenderWindow &window );
};
#endif
