//=========================================================
//
// 駒
//
//=========================================================

#include "Piece.h"
#include "Application.h"

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>
#include "../GUI_SFML/TextImageFactory.h"
#include "../Utils/Utils_Cal.h"
#include "Define.h"

//=========================================================
// コンストラクタ
//=========================================================
Piece::Piece( int color ) {

	_prop.color = color;
	/*
	_pieceBlack = sf::CircleShape   ( _prop.size );
	_pieceBlack.setFillColor        ( sf::Color(0,0,0,255) );
	_pieceBlack.setOutlineColor     ( sf::Color(0,0,0,255) );
	_pieceBlack.setOutlineThickness ( 1.0f );
	_pieceBlack.setOrigin ( _prop.size, _prop.size );

	_pieceWhite = sf::CircleShape   ( _prop.size );
	_pieceWhite.setFillColor        ( sf::Color(255,255,255,255) );
	_pieceWhite.setOutlineColor     ( sf::Color(0,0,0,255) );
	_pieceWhite.setOutlineThickness ( 1.0f );
	_pieceWhite.setOrigin ( _prop.size, _prop.size );
	sf::Color charColor;
	int rotation = 0;
	
	if      (color == Piece::Color::BLACK) { charColor = sf::Color( 255,255,255,255 );rotation = 0;} 
	else if (color == Piece::Color::WHITE) { charColor = sf::Color( 0,0,0,255 ); rotation = 180;}

	//TextImageFactory::_isAntiAlias = false;
	TextImageFactory::Create(_nameLabel, "意", 12, charColor);
	_nameLabel_Texture.loadFromImage(_nameLabel);
	_nameLabel_Sprite.setTexture(_nameLabel_Texture);
	_nameLabel_Sprite.setOrigin( _nameLabel.getSize().x / 2.0f, _nameLabel.getSize().y / 2.0f );

	//_nameLabel_Sprite.SetRotation(rotation);
	//TextImageFactory::_isAntiAlias = true;
	*/

	_texture_Black.loadFromFile ( "./image/gameBoard/pieceBlack.png" );
	_texture_White.loadFromFile ( "./image/gameBoard/pieceWhite.png" );
	_sprite.setTexture (_texture_Black);
	_sprite.setOrigin ( _sprite.getGlobalBounds().width / 2, _sprite.getGlobalBounds().height / 2 );

	_prop.size = _sprite.getGlobalBounds().width / 2;
	
	
	_shape_lastCollidedHighlight = sf::CircleShape ( _sprite.getGlobalBounds().width/2 - 3 );
	_shape_lastCollidedHighlight.setOrigin (
		_shape_lastCollidedHighlight.getGlobalBounds().width/2,
		_shape_lastCollidedHighlight.getGlobalBounds().height/2
	);
	_shape_lastCollidedHighlight.setFillColor ( sf::Color(0,0,0,0) );
	_shape_lastCollidedHighlight.setOutlineThickness(2.0f);	
	_shape_lastCollidedHighlight.setOutlineColor ( sf::Color(0,0,255,255) );

	_isReservedToDrawLastCollidedHighlight = false;

	
	_position_Offset.x = 0;
	_position_Offset.y = 0;
	
	_isWarpToGoal = true;
	
}

//=========================================================
// デストラクタ
//=========================================================
Piece::~Piece() {

}

//=========================================================
// 更新
//=========================================================
void Piece::Update(sf::RenderWindow &window, sf::Vector2f pos) {

	_position_Goal = pos;
	
	if ( _isWarpToGoal ) {
		
		_position_Current = _position_Goal;	
		_isWarpToGoal = false;
	
	} else {
		
		// 現在地と目標地の距離の、
		float distance = Utils_Cal::GetDistance( _position_Current.x, _position_Current.y, _position_Goal.x, _position_Goal.y );

		// 半分の距離を、
		float moveDistance = ( distance < 3.0f ) ? distance : distance / 3.0f;

		// 移動する
		pair<float, float> moveResult = 
			Utils_Cal::MoveTo ( _position_Current.x, _position_Current.y, _position_Goal.x, _position_Goal.y, moveDistance );

		_position_Current.x = moveResult.first;
		_position_Current.y = moveResult.second;
	}

	// 両色ともに反映
	_sprite.setPosition ( _position_Current + _position_Offset );
	_shape_lastCollidedHighlight.setPosition ( _sprite.getPosition()  );

	/*
	_pieceBlack.setPosition( _position_Current + _position_Offset );
	_pieceWhite.setPosition( _position_Current + _position_Offset );
	_nameLabel_Sprite.setPosition( _position_Current + _position_Offset );
	*/

	// マウスオーバー判定
	MouseOverCheck();

}


//=========================================================
// マウスオーバー判定
//=========================================================
void Piece::MouseOverCheck() {

	int mouseX = Application::GetInstance()->_mouseX;
	int mouseY = Application::GetInstance()->_mouseY;

	float distance = Utils_Cal::GetDistance( _position_Current.x, _position_Current.y, (float)mouseX, (float)mouseY );

	if ( distance <= _prop.size ) 
	{
		//printf("mouse on\n");

		_isMouseOver = true;

	} else {
		_isMouseOver = false;
	}

}

//=========================================================
// 描画
//=========================================================
void Piece::Draw(sf::RenderWindow &window) {

	// マウスにくっついているときには普通に描画する。
	if (Application::GetInstance()->RefferMouseClingObj() == this) {
		DrawActual(window);
		return;
	}

	// 後で前面に描画してもらう → マス目の裏側に入ってしまうのを防ぐ。
	Application::GetInstance()->AddLaterDraw_Piece( this );

}
//=========================================================
// 後で前面に描画
//=========================================================
void Piece::DrawActual(sf::RenderWindow &window) {

	/*
	sf::CircleShape currentShape;
	if      ( _prop.color == Color::BLACK ) { currentShape = _pieceBlack; }
	else if ( _prop.color == Color::WHITE ) { currentShape = _pieceWhite; }

	window.draw( currentShape );
	window.draw( _nameLabel_Sprite );
	*/

	sf::Texture currentTexture;
	if      ( _prop.color == Color::BLACK ) { currentTexture = _texture_Black; }
	else if ( _prop.color == Color::WHITE ) { currentTexture = _texture_White; }
	currentTexture.setSmooth ( false );
	_sprite.setTexture ( currentTexture );

	window.draw( _sprite );
	
	// ぶつけられの強調表示の描画が予約されていたら描画する
	if ( _isReservedToDrawLastCollidedHighlight ) {
			DrawLastCollidedHighlight(window);
	}

}

//=========================================================
// 最後の指し手強調表示、を、予約
//
// アクター画像と重なると、この強調表示だけアクターの手前側に出てしまうので、このような仕組みにする。
//
//=========================================================
void Piece::DrawLastCollidedHighlight() {

	_isReservedToDrawLastCollidedHighlight = true;
	

}
//=========================================================
// 最後の指し手強調表示：ぶつけられた駒として
//=========================================================
void Piece::DrawLastCollidedHighlight(sf::RenderWindow &window) {
	
	window.draw ( _shape_lastCollidedHighlight );
	
	_isReservedToDrawLastCollidedHighlight = false;
}

//=========================================================
// 指し手の仮表示用描画処理
//=========================================================
void Piece::DrawDemonstration(sf::RenderWindow &window, sf::Vector2f pos) {
	
	sf::Vector2f pos_org   = _sprite.getPosition();
	sf::Color    color_org = _sprite.getColor();
	
	sf::Vector2f pos_temp   = pos;
	sf::Color    color_temp = color_org; color_temp.a = 100;
	
	_sprite.setPosition(pos_temp);
	_sprite.setColor(color_temp);
	
	DrawActual ( window );

	_sprite.setPosition(pos_org);
	_sprite.setColor(color_org);
	
}


//=========================================================
// 位置をセット
//=========================================================
void Piece::SetPosition( sf::Vector2f pos) {

	///// びよーん なしで指定の位置になる。
	_position_Goal    = pos;
	_position_Current = pos;

	// 反映
	_sprite.setPosition (_position_Current + _position_Offset);
	/*
	_pieceBlack.setPosition( _position_Current + _position_Offset );
	_pieceWhite.setPosition( _position_Current + _position_Offset );
	_nameLabel_Sprite.setPosition( _position_Current + _position_Offset );
	*/
}

