//=========================================================
//
// 投了ボタン
//
//=========================================================

#include "Define.h"
#include "Application.h"
#include "ResignButton.h"
#include "ClickableImage.h"
#include "../GUI_SFML/TextImageFactory.h"
#include "../GUI_SFML/Button.h"



const string ResignButton::STATE_HIDDEN         = "";
const string ResignButton::STATE_RESIGN_BUTTON  = "RESIGN_BUTTON";
const string ResignButton::STATE_CONFIRM_WINDOW = "CONFIRM_WINDOW";

	
//=========================================================
// コンストラクタ
//=========================================================
ResignButton::ResignButton() {
	
	_state = STATE_HIDDEN;

	///// 投了ボタン
	{
		ClickableImage::Property closeButtonProp;
		
		closeButtonProp.imageFile = "./image/gameBoard/resignButton.png";
		closeButtonProp.size = sf::Vector2f(60.0f, 20.0f);
		closeButtonProp.rect_Normal = sf::IntRect ( 0, 0,  60, 20 );
		closeButtonProp.rect_Over   = sf::IntRect ( 0, 20, 60, 20 );
		closeButtonProp.rect_Push   = sf::IntRect ( 0, 40, 60, 20 );
		
		_resignButton = new ClickableImage(closeButtonProp);
		
		_defaultPosition = sf::Vector2f ( 730.0f, 560.0f );
		_resignButton->SetPosition ( _defaultPosition );
	}	

	// 確認ウィンドウ用
	_rectForConfirmWindow = sf::RectangleShape ( sf::Vector2f ( 400, 200 ) );
	_rectForConfirmWindow.setFillColor ( sf::Color ( 50,50,150,150 ) );
	_rectForConfirmWindow.setOutlineThickness (0.0f);
	_rectForConfirmWindow.setOutlineColor ( sf::Color ( 0,0,0,0 ) );
	_rectForConfirmWindow.setPosition ( sf::Vector2f ( APPLICATION::WINDOW_SIZE::WIDTH/2, APPLICATION::WINDOW_SIZE::HEIGHT/2 ) );
	_rectForConfirmWindow.setOrigin ( _rectForConfirmWindow.getGlobalBounds().width/2, _rectForConfirmWindow.getGlobalBounds().height/2 );
	
	_labelForConfirmWindow = TextImageFactory::Create ( "投了します", 24);
	_labelForConfirmWindow.setFillColor   ( sf::Color::White );
	_labelForConfirmWindow.setOutlineColor( sf::Color::Black );
	_labelForConfirmWindow.setOutlineThickness ( 1.0f );
	_labelForConfirmWindow.setPosition (
			_rectForConfirmWindow.getPosition().x, 
			250 );
	_labelForConfirmWindow.setOrigin ( (int)(_labelForConfirmWindow.getGlobalBounds().width/2), (int)(_labelForConfirmWindow.getGlobalBounds().height/2) );
	
	_buttonYes = new Button ( "する",     sf::Vector2f ( 150, 40 ), 24 );
	_buttonNo  = new Button ( "しない", sf::Vector2f ( 150, 40 ), 24 );
	_buttonYes->SetPosition ( sf::Vector2f ( 235, 320 ) );
	_buttonNo ->SetPosition ( sf::Vector2f ( 405, 320 ) );
	
	_buttonYes->SetColor_Nomal ( sf::Color(255,255,255,240) );
	_buttonYes->SetColor_Push  ( sf::Color(255,  0,  0,200) );
	_buttonYes->SetColor_Over  ( sf::Color(  0,255,  0,200) );
	
	_buttonNo->SetColor_Nomal ( sf::Color(255,255,255,240) );
	_buttonNo->SetColor_Push  ( sf::Color(255,  0,  0,200) );
	_buttonNo->SetColor_Over  ( sf::Color(  0,255,  0,200) );
	
	_isResignDecided = false;
}

//=========================================================
// デストラクタ
//=========================================================
ResignButton::~ResignButton() {
	
	delete _buttonYes;
	delete _buttonNo;
	
}
	

//=========================================================
// 表示位置設定
//=========================================================
void ResignButton::SetPosition ( int x, int y ) {
	
	sf::Vector2f pos = _resignButton->GetPosition();
	
	pos.x = x;
	pos.y = y;
	
	_resignButton->SetPosition (pos);
}
void ResignButton::SetPositionX ( int x ) {
	
	sf::Vector2f pos = _resignButton->GetPosition();
	
	pos.x = x;
	
	_resignButton->SetPosition (pos);
}
void ResignButton::SetPositionY ( int y ) {
	
	sf::Vector2f pos = _resignButton->GetPosition();
	
	pos.y = y;
	
	_resignButton->SetPosition (pos);
}
void ResignButton::SetPositionToDefault () {
	_resignButton->SetPosition ( _defaultPosition );
}	
	
	
//=========================================================
// アクティブにするかどうか
//=========================================================
void ResignButton::SetActive ( bool arg ) {
	
	if ( arg ) {

		///// 非表示のとき
		if ( _state == STATE_HIDDEN ) { _state = STATE_RESIGN_BUTTON;}

		///// 投了ボタンを表示中
		else if ( _state == STATE_RESIGN_BUTTON ) { _state = STATE_RESIGN_BUTTON; }

		///// 確認ウィンドウを表示中
		else if ( _state == STATE_CONFIRM_WINDOW ) { _state = STATE_CONFIRM_WINDOW; }

	} else {

		///// 非表示のとき
		if ( _state == STATE_HIDDEN ) { _state = STATE_HIDDEN;}

		///// 投了ボタンを表示中
		else if ( _state == STATE_RESIGN_BUTTON ) { _state = STATE_HIDDEN; }

		///// 確認ウィンドウを表示中
		else if ( _state == STATE_CONFIRM_WINDOW ) { _state = STATE_HIDDEN; }

	}
	
}

//=========================================================
// 確認ウィンドウが出ているところかどうか
//=========================================================
bool ResignButton::IsComfirmWindowDisplaying() {

	return (_state == STATE_CONFIRM_WINDOW);
}


//=========================================================
// 更新
//=========================================================
void ResignButton::Update ( sf::RenderWindow &window ) {
	
	///// 決定フラグは毎フレームリセット
	_isResignDecided = false;
	
	///// 非表示のとき
	if ( _state == STATE_HIDDEN ) {
		// 何もしない
	} 
	///// 投了ボタンを表示中
	else if ( _state == STATE_RESIGN_BUTTON ) {

		_resignButton->Update(window);
		
		if ( _resignButton->IsClick() ) {
			
			_state = STATE_CONFIRM_WINDOW;
		}
		
	}
	///// 確認ウィンドウを表示中
	else if ( _state == STATE_CONFIRM_WINDOW ) {
		
		_buttonYes->Update(window);
		_buttonNo->Update(window);
		
		if ( _buttonYes->IsRelease() ) {
			_isResignDecided = true;
			_state = STATE_RESIGN_BUTTON;
		}
		if ( _buttonNo->IsRelease() ) {
			_isResignDecided = false;
			_state = STATE_RESIGN_BUTTON;
		}


		// 右クリックで確認状態解除
		if ( Application::GetInstance()->_mouseState_Right == MOUSE::BUTTON::STATE::RELEASE ) {
			_state = STATE_RESIGN_BUTTON;
		}
		
		int mouseX = Application::GetInstance()->_mouseX;
		int mouseY = Application::GetInstance()->_mouseY;
		
		sf::Vector2i mousePos = sf::Vector2i(mouseX, mouseY);
		sf::Vector2f mouseInnerPos =  Application::GetInstance()->_window.mapPixelToCoords( mousePos );
		mouseX = mouseInnerPos.x;
		mouseY = mouseInnerPos.y;

		
		int confirmWindowLeftTopX =_rectForConfirmWindow.getGlobalBounds().left;
		int confirmWindowLeftTopY =_rectForConfirmWindow.getGlobalBounds().top;

		int confirmWindowRightBottomX = confirmWindowLeftTopX + _rectForConfirmWindow.getGlobalBounds().width;
		int confirmWindowRightBottomY = confirmWindowLeftTopY + _rectForConfirmWindow.getGlobalBounds().height;
		
		bool isMouseOn = false;
		if 
		( 
			confirmWindowLeftTopX < mouseX && mouseX < confirmWindowRightBottomX 
			&&
			confirmWindowLeftTopY < mouseY && mouseY < confirmWindowRightBottomY 
		) 
		{
			isMouseOn = true;
		}
		
		//printf (" マウス %s \n", isMouseOn?"ON":"OFF");
		
		// マウスが乗っていないときに左クリックすると確認ウィンドウ解除
		if ( !isMouseOn ) {
			
			if ( Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::RELEASE ) {
				_state = STATE_RESIGN_BUTTON;
			}
		}
	}

}


//=========================================================
// 描画
//=========================================================
void ResignButton::Draw   ( sf::RenderWindow &window ) {

	///// 非表示のとき
	if ( _state == STATE_HIDDEN ) {
		// 何もしない
	} 
	///// 投了ボタンを表示中
	else if ( _state == STATE_RESIGN_BUTTON ) {

		_resignButton->Draw(window);
		
	}
	///// 確認ウィンドウを表示中
	else if ( _state == STATE_CONFIRM_WINDOW ) {
		
		_resignButton->Draw(window); // 投了ボタンも描画はする

		window.draw (_rectForConfirmWindow);
		window.draw (_labelForConfirmWindow);
		_buttonYes->Draw(window);
		_buttonNo->Draw(window);
		
	}
}

	
	
	
	


