//=========================================================
//
// 棋譜ビューワー
//
//=========================================================

#ifndef _SCORE_SHEET_VIEWER_
#define _SCORE_SHEET_VIEWER_

#include <map>
#include <vector>
#include <SFML/Graphics.hpp>
#include "../GUI_SFML/ISelectBoxItem.h"
#include "../GUI_SFML/SelectBox.h"
#include "ClickableImage.h"

using namespace std;
using namespace GUI_SFML;

class Grid;

//=========================================================
// ビューワーの１行
//=========================================================
class ScoreSheetCell : public ISelectBoxItem {
	
public:
	
	ScoreSheetCell( string moveNotation, const sf::Font &font,
					sf::Sprite *backGroundImage,
					sf::Sprite *focusImage,
					sf::Sprite *mouseoverImage,
					int index  // 何手目か、という情報を、文言とは別に受け取る
	);
	~ScoreSheetCell();
	
	void Update (sf::RenderWindow &window);
	void Draw   (sf::RenderWindow &window);

	sf::Vector2f GetPosition();
	void         SetPosition( sf::Vector2f pos);

	sf::Vector2f GetSize();
	bool IsMouseOver();
	
	void Focus(bool f) { _isFocus = f; }
	
	string GetText() { return _moveNotation_text; }
	
private:
	
	sf::Vector2f _position;
	
	string   _moveNotation_text;
	sf::Text _moveNotation_image;
	
	bool _isFocus;
	
	sf::Sprite *_backGroundImage;
	sf::Sprite *_focusImage;
	sf::Sprite *_mouseoverImage;
};

//=========================================================
// ビューワー本体
//=========================================================
class ScoreSheetViewer {

public:

	ScoreSheetViewer();
	~ScoreSheetViewer();

	void Update (sf::RenderWindow &window);
	void Draw   (sf::RenderWindow &window);
	
	void Add   ( string moveNotation, bool isEndDescription = false);
	void Clear ();
	
	void MoveToBottom();
	void MoveToTop();
	
	void SetPosition ( sf::Vector2f pos );
	
	void ShowGuideButtons() { _guideButtonsOn = true; }
	void HideGuideButtons() { _guideButtonsOn = false; }
	
	// 外部にはボタンの情報を直接には公開しない。
	/*
	bool IsClick_toFirst();
	bool IsClick_toPrev();
	bool IsClick_toNext();
	bool IsClick_toLast();
	*/
	
	int GetCurrentIndex() { return _currentIndex; } // 現在位置
	bool IsIndexChanged() { return _isIndexChanged; } // インデックスに変化があった？ あったら現在位置を取得してね
	bool IsNextPushed() { return _isNextPushed; }
	void SetCurrentIndex(int index);
	
	
	string Serialize();
	void DeSerialize( string data );
	
private:
	
	sf::Texture _backGroundTexture;
	sf::Sprite  _backGroundSprite;
	
	sf::Vector2f _position;
	
	vector<ScoreSheetCell*> _scoreSheetCellList;
	
	SelectBox *_selectBox;
	
	sf::Font _font;
	
	// 棋譜の現在位置
	int  _currentIndex;
	bool _isIndexChanged;
	bool _isNextPushed;
	
	// ナビゲーションパネル
	bool _guideButtonsOn;
	ClickableImage *_guideButton_toFirst;
	ClickableImage *_guideButton_toPrev;
	ClickableImage *_guideButton_toNext;
	ClickableImage *_guideButton_toLast;
	
	sf::Texture _guideButtonsBackGroundTexture;
	sf::Sprite  _guideButtonsBackGroundSprite;
	
	
	// 棋譜の一行分の見た目
	sf::Sprite *_cellBackGroundImage;
	sf::Sprite *_cellFocusImage;
	sf::Sprite *_cellMouseOverImage;
	
	sf::Texture _cellBackGroundTexture;
	sf::Texture _cellFocusTexture;
	sf::Texture _cellMouseOverTexture;
	
	int GetClickedIndex(); // クリックしてなければ負の数を返す

	void GuideButtonWatching();
};
#endif

