//=========================================================
//
// Util_Convert (declare)
//
//=========================================================

#include "Utils_Convert.h"
#include "Utils_String.h"

#include <stdio.h>
#include <time.h> 
#include <climits>

//=========================================================
// string -> int
//=========================================================
int Utils_Convert::String_to_Int( string str, int defaultValue ) {

	if (str == "") {
		return defaultValue;
	}

	int result = 0;
	sscanf(str.c_str(), "%d", &result);
	return result;
}
long Utils_Convert::String_to_Long( string str, long defaultValue ) {

	if (str == "") {
		return defaultValue;
	}

	long result = 0;
	sscanf(str.c_str(), "%ld", &result);
	return result;
}

//=========================================================
// int -> string
//=========================================================
string Utils_Convert::Int_to_String( int arg ) {

	string result = "";

	char buf[1024];
	sprintf( buf, "%d", arg );

	result = buf;

	return result;
}
string Utils_Convert::Long_to_String( long arg ) {

	string result = "";

	char buf[1024];
	sprintf( buf, "%ld", arg );

	result = buf;

	return result;
}

//=========================================================
// float -> string
//=========================================================
string Utils_Convert::Float_to_String( float arg ) {
	
	string result = "";

	char buf[1024];
	sprintf( buf, "%f", arg );

	result = buf;

	return result;
}


//=========================================================
// string -> float
//=========================================================
float Utils_Convert::String_to_Float( string str, float defaultValue ) {

	if (str == "") {	
		return defaultValue;
	}

	float result = 0;
	sscanf(str.c_str(), "%f", &result);
	return result;
}


/*
//=========================================================
// 文字列 → sf::Color
//=========================================================
sf::Color Utils_Convert::String_to_sfColor(string str, string splitter) {

	sf::Color result;

	vector<string> list = Utils_String::Split( str, splitter );

	///// number of elements is 4 :: with alpha
	if (list.size() == 4) {

		result.r = Utils_Convert::String_to_Int ( list[0] );
		result.g = Utils_Convert::String_to_Int ( list[1] );
		result.b = Utils_Convert::String_to_Int ( list[2] );
		result.a = Utils_Convert::String_to_Int ( list[3] );

	///// number of elements is 3 :: without alpha
	} else if (list.size() == 3) {

		result.r = Utils_Convert::String_to_Int ( list[0] );
		result.g = Utils_Convert::String_to_Int ( list[1] );
		result.b = Utils_Convert::String_to_Int ( list[2] );

	///// Invalid argument
	} else {
		// anything error processing?
	}

	return result;
}
*/

//==========================================================
// string -> bool
//==========================================================
bool Utils_Convert::String_to_bool( string str ) {

	bool result = false;

	if      ( str == "true") { result = true;}
	else if ( str == "false") { result = false;}
	else if ( str == "on") { result = true;}
	else if ( str == "off") { result = false;}
	else if ( str == "yes") { result = true;}
	else if ( str == "no") { result = false;}
	else if ( str == "ok") { result = true;}
	else if ( str == "ng") { result = false;}
	else if ( str == "1") { result = true;}
	else if ( str == "0") { result = false;}

	else if ( str == "TRUE") { result = true;}
	else if ( str == "FALSE") { result = false;}
	else if ( str == "ON") { result = true;}
	else if ( str == "OFF") { result = false;}
	else if ( str == "YES") { result = true;}
	else if ( str == "NO") { result = false;}
	else if ( str == "OK") { result = true;}
	else if ( str == "NG") { result = false;}


	return result;
}
//==========================================================
// bool -> string
//==========================================================
string Utils_Convert::Bool_to_string( bool arg ) {

	return ( (arg)?"true":"false" );
}




//=========================================================
// map < string, string > -> string
//=========================================================
string Utils_Convert::map_string_string_to_string ( map < string, string > src ) {

	///// 空なら空で。
	if ( src.size() == 0 ) {
		return "";
	}

	/*
		前半、キーの長さと値の長さを交互につなげた数値の列。「,」区切り。
		「/」が出たら、前半終了。
		
		つまり前半は、数字と、「,」と「/」以外は出てこないことが保証されている。

		後半は、キーと値を交互につなげたもの。
		セパレータもなにもなく、ただつなげただけ。
		前半の数字をもとに、切り分けてください。

	*/

	string result = "";

	///// 前半も後半も。
	string firstHalf  = "";
	string secondHalf = "";


	for ( map<string, string>::iterator it = src.begin(); it != src.end(); ++it ) {

		string key   = it->first;
		string value = it->second;

		// 後半部分には、やみくもにつなげていく。
		secondHalf += key + value;

		string keyLength   = Int_to_String ( key.size() );
		string valueLength = Int_to_String ( value.size() );

		// 最初じゃないなら、前回との区切りとして、「,」を１つ入れておくよ。
		if ( firstHalf != "" ) {
			firstHalf += ",";
		}

		// 前半さん、後半さんを導いて。
		firstHalf += keyLength + "," + valueLength;
	}

	// こうなるのですね。。
	result = firstHalf + "/" + secondHalf;

	return result;
}

//=========================================================
// string -> map < string, string >
//=========================================================
 map < string, string > Utils_Convert::string_to_map_string_string ( string src ) {

	map < string, string > result;

	///// 空なら空で。
	if ( src == "" ) {
		return result;
	}

	///// まずは前半と後半を分けましょう。
	string firstHalf  = "";
	string secondHalf = "";
	{
		vector<string> temp = Utils_String::Split ( src, "/" );

		firstHalf  = temp[0];
		//secondHalf = ( temp.size() >= 2 ) ? temp[1] : ""; // 
		secondHalf = ( src.size() > firstHalf.size() + 1 ) ? src.substr ( firstHalf.size() + 1 ) : "";
	}

	///// 前半の情報を、数値のリストとしましょう。
	vector<int> lengthInformationList;
	{
		vector<string> temp = Utils_String::Split ( firstHalf, "," );

		for ( unsigned int i = 0; i < temp.size(); ++i ) {

			lengthInformationList.push_back ( String_to_Int ( temp[i] ) );
		}
	}

	string sub = secondHalf.substr(0, 10);

	///// 後半を解析：
	int currentCursorPos = 0;
	vector < string > secondHalfAnalysed;
	for ( unsigned int i = 0; i < lengthInformationList.size(); ++i ) {

		string sub = secondHalf.substr ( currentCursorPos, lengthInformationList[i] );

		secondHalfAnalysed.push_back ( sub );

		currentCursorPos += lengthInformationList[i];
	}

	// ついに、中身が手に入る！
	for ( unsigned int i = 0; i < secondHalfAnalysed.size(); i += 2 ) {

		string key   = secondHalfAnalysed[i];
		string value = ( (i + 1) < secondHalfAnalysed.size() ) ? secondHalfAnalysed[i + 1] : "";

		result[key] = value;
	}

	 return result;
 }

//=========================================================
// map < string, string > -> string
//=========================================================
string Utils_Convert::pairList_string_string_to_string ( vector < pair < string, string > > src ) {

	///// 空なら空で
	if ( src.size() == 0 ) {
		return "";
	}

	string result = "";

	///// 前半も後半も。
	string firstHalf  = "";
	string secondHalf = "";

	for ( unsigned int i = 0; i < src.size(); ++i) {

		string key   = src[i].first;
		string value = src[i].second;

		// 後半部分には、やみくもにつなげていく。
		secondHalf += key + value;

		string keyLength   = Int_to_String ( key.size() );
		string valueLength = Int_to_String ( value.size() );

		// 最初じゃないなら、前回との区切りとして、「,」を１つ入れておくよ。
		if ( firstHalf != "" ) {
			firstHalf += ",";
		}

		// 前半さん、後半さんを導いて。
		firstHalf += keyLength + "," + valueLength;
	}

	// こうなるのですね。。
	result = firstHalf + "/" + secondHalf;

	return result;

}

//=========================================================
// string -> vector < pair < string, string > >
//=========================================================
vector < pair <string, string > > Utils_Convert::string_to_pairList_string_string ( string src ) {

	vector < pair <string, string > > result;

	///// 空なら空で
	if ( src == "" ) {
		return result;
	}

	///// まずは前半と後半を分けましょう。
	string firstHalf  = "";
	string secondHalf = "";
	{
		vector<string> temp = Utils_String::Split ( src, "/" );

		firstHalf  = temp[0];
		//secondHalf = ( temp.size() >= 2 ) ? temp[1] : ""; // 
		secondHalf = ( src.size() > firstHalf.size() + 1 ) ? src.substr ( firstHalf.size() + 1 ) : "";
	}

	///// 前半の情報を、数値のリストとしましょう。
	vector<int> lengthInformationList;
	{
		vector<string> temp = Utils_String::Split ( firstHalf, "," );

		for ( unsigned int i = 0; i < temp.size(); ++i ) {

			lengthInformationList.push_back ( String_to_Int ( temp[i] ) );
		}
	}

	//string sub = secondHalf.substr(0, 10);

	///// 後半を解析：
	int currentCursorPos = 0;
	vector < string > secondHalfAnalysed;
	for ( unsigned int i = 0; i < lengthInformationList.size(); ++i ) {

		string sub = secondHalf.substr ( currentCursorPos, lengthInformationList[i] );

		secondHalfAnalysed.push_back ( sub );

		currentCursorPos += lengthInformationList[i];
	}

	// ついに、中身が手に入る！
	for ( unsigned int i = 0; i < secondHalfAnalysed.size(); i += 2 ) {

		string key   = secondHalfAnalysed[i];
		string value = ( (i + 1) < secondHalfAnalysed.size() ) ? secondHalfAnalysed[i + 1] : "";

		pair < string, string > p ( key, value );

		result.push_back ( p );
	}

	return result;
}

//=========================================================
// string -> map<string, string>
//=========================================================
map<string, string> Utils_Convert::ReadStringAsMap ( string &src, string delimiter, string equal) {

	map<string, string> result;

	vector<string> elements = Utils_String::Split ( src, delimiter );
	for ( unsigned int i = 0; i < elements.size(); ++i ) {

		string element = elements[i];

		vector<string> analyzed = Utils_String::Split ( element, equal );

		string key   = ( analyzed.size() >= 1) ? analyzed[0]: "";
		string value = ( analyzed.size() >= 2) ? analyzed[1]: "";;

		result.insert( pair<string, string>( key, value ) );
	}

	return result;
}

//=========================================================
// vector<string> -> comma separated string
//=========================================================
string Utils_Convert::vectorString_to_string ( vector<string> &list, string delimiter) {

	string result = "";

	for ( unsigned  int i = 0; i < list.size(); ++i) {

		if ( i > 0 ) {
			result += delimiter;
		}

		result += list[i];
	}

	return result;
}

//=========================================================
// vector<int> -> comma separated string
//=========================================================
string Utils_Convert::vectorInt_to_string ( vector<int> &list, string delimiter) {

	string result = "";


	for ( unsigned  int i = 0; i < list.size(); ++i) {

		if ( i > 0 ) {
			result += delimiter;
		}

		result += Int_to_String ( list[i] );
	}

	return result;
}


//=========================================================
// UNIXタイムスタンプ文字列から日付情報を取得
//=========================================================
map<string, int> Utils_Convert::String_to_TimeInfo ( string unixTimestamp ) {
	
	long t;
	sscanf( unixTimestamp.c_str(), "%ld", &t);
	return  Long_to_TimeInfo ( t );
}

map<string, int> Utils_Convert::Long_to_TimeInfo ( long unixTimestamp ) {
	
	map<string, int> result;
	
	time_t t = (time_t)unixTimestamp;
	
	
	//printf ( "t : %ld\n", t );
	
	int offset_year = 1900;
	int offset_mon  = 1;
	int offset_mday = 0;
	int offset_hour = 0;
	int offset_min  = 0;
	int offset_sec  = 0;
	
	if ( t < 0 ) {
	
		t += LONG_MAX; //2147483647;
		
		if ( t < 0 ) {
			t *= -1;
		}

		offset_year = 2038 - 70;
		offset_mon  = 1;
		offset_mday = 18;
		offset_hour = 14 - 11;
		offset_min  = 14;
		offset_sec  = 7;
	}
	//printf ( "t : %ld\n", t );
	
	tm *localTime = localtime(&t);

	int year  = localTime->tm_year + offset_year;
	int month = localTime->tm_mon + offset_mon;
	int day   = localTime->tm_mday + offset_mday;
	int hour  = localTime->tm_hour + offset_hour;
	int min   = localTime->tm_min + offset_min;
	int sec   = localTime->tm_sec + offset_sec;
	
	result["year" ] = year;
	result["month"] = month;
	result["day"  ] = day;
	result["hour" ] = hour;
	result["min"  ] = min;
	result["sec"  ] = sec;
	
	return result;
}







