//=========================================================
//
// Utils_Others
//
//=========================================================

#ifndef _UTILS_OTHERS_
#define _UTILS_OTHERS_

#include <vector>
#include <string>
#include <map>
#include <algorithm>

#include <time.h>
#include "Utils_Cal.h"

using namespace std;


namespace UTILS_OTHERS {

	//=========================================================
	// indexOf for STL::vector
	//=========================================================
	template<class T> int indexOf(vector<T> list, T value) {

		int result = -1;

		// 含まれてる？
		for (unsigned int i = 0; i < list.size(); ++i) {

			if ( value == list[i]) {
				result = i;
				break;
			}
		}
		return result;
	}

	//=========================================================
	// 任意の位置の要素を削除 for STL::vector
	//=========================================================
	template<class T> void deleteFromVector(vector<T> &list, unsigned int index) {

		// いきなりオーバー？
		if ( index >= list.size() || index < 0) {
			return;
		}

		unsigned int count = 0;
		typename vector<T>::iterator it = list.begin();
		while ( count < index ) {
			++it;
			++count;
		}

		list.erase(it);
	}
	//=========================================================
	// 任意の位置の要素を削除 for STL::vector
	//=========================================================
	template<class T> void deleteElementFromVector(vector<T> &list, T value) {

		// 一致する要素を全て消す。

		bool isExist = true;

		while (isExist) {

			isExist = false;
			for (typename vector<T>::iterator it = list.begin(); it != list.end(); ++it) {

				if ( *it == value ) {

					list.erase(it);
					isExist = true;
					break;
				}
			}
		}
	}
	//=========================================================
	// STL::vector 一致判定
	//=========================================================
	template<class T> bool isEqual(vector<T> &listA, vector<T> &listB) {

		bool result = false;

		// サイズが違うなら、もう、違う。
		if (listA.size() != listB.size()) {
			return false;
		}

		result = true;
		for (unsigned int i = 0; i < listB.size(); ++i) {

			// 含まれてないなら、もう、違う。
			if ( indexOf<T>(listA, listB[i]) == -1 ) {
				result = false;
				break;
			}
		}

		return result;
	}

	//=========================================================
	// random_shuffle for vector
	//=========================================================
	class Random
	{
	public:

		// コンストラクタ
		Random()
		{
			srand( static_cast<unsigned int>( time(NULL) ) );
		}

		// 関数オブジェクト
		unsigned int operator()(unsigned int max)
		{
			double tmp = static_cast<double>( rand() ) / static_cast<double>( RAND_MAX );
			return static_cast<unsigned int>( tmp * max );
		}

	};

	template<class T> void RandomShuffle(vector<T> &list) {
		
		
		Random r;
		std::random_shuffle( list.begin(), list.end(), r);
		
		/*
		//for ( unsigned int n = 0; n < list.size(); ++n ) {
			for ( unsigned int i = 0; i < list.size(); ++i ) {
				
				int lot = Utils_Cal::GetRandom(100);
				if ( lot > 50 ) {
					
					int index = Utils_Cal::GetRandom(list.size());
					
					T tmp = list[i];
					list[i] = list[index];
					list[index] = tmp;
				}
			}
		//}
		*/
		
	}

	//=========================================================
	// vector + vector
	//=========================================================
	template<class T> vector<T> MargeList(const vector<T> &listA, const vector<T> &listB) {

		vector<T> result = listA;

		for (unsigned int i = 0; i < listB.size(); ++i ) {

			result.push_back( listB[i] );
		}

		return result;
	}


	//=========================================================
	// パーミテーション：実装・重複なし
	//=========================================================
	template<class T>
	void GeneratePermutation_noDuplicate( vector<vector<T>> &result, const vector<T> &srcList, vector<int> &element, int num ) {

		for (unsigned int i = 0; i < srcList.size(); ++i) {

			vector<int> element_this = element;

			// まだ含まれてないならば
			if ( UTILS_OTHERS::indexOf<T>(element_this, i) == -1) {

				element_this.push_back( i );

				if (num == 1) {
				
					vector<T> item;
					for (unsigned int n = 0; n < element_this.size(); ++n) {

						item.push_back( srcList[element_this[n]] );
					}
					result.push_back(item);

				} else {
					GeneratePermutation_noDuplicate( result, srcList, element_this, num - 1 );
				}

			// すでにあるならば
			} else {
				continue;
			}

		}
	}
	//=========================================================
	// パーミテーション：実装・重複あり
	//=========================================================
	template<class T>
	void GeneratePermutation( vector<vector<T>> &result, const vector<T> &srcList, vector<T> &element, int num ) {

		for (unsigned int i = 0; i < srcList.size(); ++i) {

			vector<T> element_this = element;
			element_this.push_back( srcList[i] );

			if (num == 1) {
				result.push_back(element_this);
			} else {
				GeneratePermutation( result, srcList, element_this, num - 1 );
			}
		}
	}

	//=========================================================
	// パーミテーション
	//=========================================================
	template<class T>
	vector<vector<T>> GetPermutation( const vector<T> &srcList, int num, bool isNoduplicate = true) {

		vector<vector<T>> result;
		vector<T> element;

		if (isNoduplicate) {
			GeneratePermutation_noDuplicate( result, srcList, element, num );
		} else {
			GeneratePermutation( result, srcList, element, num );
		}
		return result;
	}


	//=========================================================
	// コンビネーション
	//=========================================================
	template<class T>
	vector<vector<T>> GetCombination( const vector<T> &srcList, int num) {

		vector<vector<T>> result;
		vector<int> element;


		vector<vector<int>> result_temp;
		if (num > 0) {
			GenerateCombination( result_temp, srcList, element, num );
		}


		for ( unsigned int i = 0; i < result_temp.size(); ++i ) {

			vector<T> item;
			for (unsigned int n = 0; n < result_temp[i].size(); ++n) {

				item.push_back( srcList[result_temp[i][n]] );
			}

			result.push_back( item );
		}

		return result;
	}

	//=========================================================
	// コンビネーション：実装
	//=========================================================
	template<class T>
	void GenerateCombination( vector<vector<int>> &result, const vector<T> &srcList, vector<int> &element, int num ) {

		// 保険。
		if (num <= 0) {
			return;
		}

		for (unsigned int i = 0; i < srcList.size(); ++i) {

			vector<int> element_this = element;

			// まだ含まれてないならば
			if ( UTILS_OTHERS::indexOf<int>(element_this, i) == -1) {

				element_this.push_back( i );

				if (num == 1) {

					sort(element_this.begin(), element_this.end());
					if ( UTILS_OTHERS::indexOf< vector<int> >(result, element_this) == -1) {
						result.push_back(element_this);
					}


				} else {
					GenerateCombination( result, srcList, element_this, num - 1 );
				}

			// すでにあるならば
			} else {
				continue;
			}

		}
	}

	//=========================================================
	// map から値を取得。キーがなければデフォルト値
	//=========================================================
	std::string GetValueFromMap ( std::map<string, string> &list, std::string key, std::string defaultValue );

};
#endif
