//=========================================================
//
// 文字列関係のサブルーチン集（宣言）
//
//=========================================================

#include <string>
#include <vector>
#include <map>
//#include "Utils.h"
#include "Utils_String.h"
#include <stdio.h>

using namespace std;

//=========================================================
// 前後Trim
//【引数】
// string
//【戻り値】
// 前後の半角スペース、\n、\rを取り除いた文字列
//=========================================================
string Utils_String::Trim(string str){

	string result = "";

	int firstIndex = 0;
	int lastIndex = str.length() - 1;
	
	if ( lastIndex < 0 ) { return "";}

	// 開始位置を見つける
	while(true){

		// 全部スペースor改行
		if(firstIndex == str.length()){
			return "";
		}

		if(str[firstIndex] == ' ' || str[firstIndex] == '\r' || str[firstIndex] == '\n' || str[firstIndex] == '\t'){
			++firstIndex;
		} else {
			break;
		}
	}

	// 終了位置を見つける
	while(true){

		if(str[lastIndex] == ' ' || str[lastIndex] == '\r' || str[lastIndex] == '\n' || str[lastIndex] == '\t'){
			--lastIndex;
		} else {
			break;
		}
	}

	// 開始位置から終了位置までを取り出す
	for(int i = firstIndex; i <= lastIndex; ++i){
		result += str[i];
	}

	return result;
}
//=========================================================
// Split
// https://gist.github.com/shnya/735066#file-c-split-find-version
//=========================================================
vector<string> Utils_String::Split(string str, string delim){

  vector<string> res;

  size_t current = 0, found, delimlen = delim.size();

  while ( (found = str.find(delim, current)) != string::npos )
  {
    res.push_back(string(str, current, found - current));
    current = found + delimlen;
  }
  res.push_back ( string (str, current, str.size() - current) );
  return res;
}

//=========================================================
// 指定バイト数ごとに区切ったリストを取得
//【引数】
// string str  : 元の文字列
// int    byte : このバイト数で区切る
//【戻り値】
// vector<string>
//=========================================================
vector<string> Utils_String::SplitByByte(string value, int byte){

	vector<string> result;

	int currentPosition = 0;

	while( true ){

		// もうはみ出るときは、
		if((currentPosition + byte) >= value.length()){

			// 残りの文字数全部を入れて終了
			result.push_back(
				value.substr(currentPosition)
			);
			break;
		}

		// ちょうど終わるときは
		if(currentPosition == (value.length() - 1)){

			// 終了
			break;
		}

		// 取得して
		result.push_back(
			value.substr(currentPosition, byte)
		);

		// 進める
		currentPosition += byte;
	}
	return result;
}

//=========================================================
// 肯定を表す文字列かどうか。
//【戻り値】
// true で 肯定を表す文字列ですよ。
//=========================================================
bool Utils_String::isPositiveWord(string arg) {

	bool result = (
		arg == "true" || 
		arg == "ok" || 
		arg == "on" || 
		arg == "yes" || 
		arg == "TRUE" || 
		arg == "ON" || 
		arg == "OK" || 
		arg == "YES"
	);

	return result;
}

//=========================================================
// 否定を表す文字列かどうか。
//【戻り値】
// true で 否定を表す文字列ですよ。
//=========================================================
bool Utils_String::isNegativeWord(string arg) {

	bool result = (
		arg == "false" || 
		arg == "cancel" || 
		arg == "off" || 
		arg == "no" || 
		arg == "FALSE" || 
		arg == "OFF" || 
		arg == "CANCEL" || 
		arg == "NO"
	);

	return result;
}
//=========================================================
// vector<string>の中に指定の要素があるか？
//=========================================================
bool Utils_String::isContain(vector<string> list, string element) {

	bool result = false;

	vector<string>::iterator it = list.begin();
	while (it != list.end()) {

		if (element == (*it)) {
			result = true;
			break;
		}

		++it;
	}

	return result;
}
//=========================================================
// 文字列中に指定のパターンが含まれているかどうか
//【戻り値】
// true で 含まれている。
//=========================================================
bool Utils_String::IsContainsPattern(string dest, string pattern){

	return Split(dest, pattern).size() >= 2;

}

//=========================================================
// 指定位置の次の文字の先頭バイトのインデックスを取得
//=========================================================
int Utils_String::CharNext ( std::string str, int index  ) {

	int result = 0;

	int i = index;
	int charStride = 0;
	while ( i >= 0 )
	{
		unsigned char c = str[i];
		if      ( (c >= 0b00000000) && (c <= 0b01111111) ) { charStride = 1; break; }
		else if ( (c >= 0b11000010) && (c <= 0b11011111) ) { charStride = 2; break; }
		else if ( (c >= 0b11100000) && (c <= 0b11101111) ) { charStride = 3; break; }
		else if ( (c >= 0b11110000) && (c <= 0b11110111) ) { charStride = 4; break; }
		else if ( (c >= 0b11111000) && (c <= 0b11111011) ) { charStride = 5; break; }
		else if ( (c >= 0b11111100) && (c <= 0b11111101) ) { charStride = 6; break; }
		else                                               { printf("[Utils_String::CharNext] else\n");}

		// さかのぼる
		i -= 1;

	}

	result = i + charStride;

	//if ( str[index] & 0b1) { stride += 1;}

	return result;
}
//=========================================================
// 指定位置の文字の先頭バイトのインデックスを取得
//=========================================================
int Utils_String::CharCurrent ( std::string str, int index  ) {

	int result = 0;

	int i = index;
	while ( i >= 0 )
	{
		unsigned char c = str[i];
		if      ( (c >= 0b00000000) && (c <= 0b01111111) ) { break; } // charStride = 1
		else if ( (c >= 0b11000010) && (c <= 0b11011111) ) { break; } // charStride = 2
		else if ( (c >= 0b11100000) && (c <= 0b11101111) ) { break; } // charStride = 3
		else if ( (c >= 0b11110000) && (c <= 0b11110111) ) { break; } // charStride = 4
		else if ( (c >= 0b11111000) && (c <= 0b11111011) ) { break; } // charStride = 5
		else if ( (c >= 0b11111100) && (c <= 0b11111101) ) { break; } // charStride = 6
		else                                               { printf("else\n");}

		// さかのぼる
		i -= 1;

	}

	result = i;

	//if ( str[index] & 0b1) { stride += 1;}

	return result;
}

//=========================================================
// 全角/半角 判定（引数は１文字である前提。文字数チェックは入れていない）
// 参考：
//	http://orange-factory.com/dnf/utf-8.html
//	http://tech.nitoyon.com/ja/blog/2014/03/14/utf8-str-count/
//	http://www.unicode.org/Public/UCD/latest/ucd/EastAsianWidth.txt
//=========================================================
bool Utils_String::isHalfWidth ( const std::string c ) {

	//printf ( "[isHalfWidth] %s （１文字ですか？）\n", c.c_str() );

	bool result = false;

	// 1 byte 文字の場合：無条件で半角と判定
	if ( c.size() == 1 ) {
		//printf("1 byte\n");
		result = true;
	}
	// 2 byte 文字の場合（怪しい。2 byte で半角になる文字の存在を確認できず）
	else if ( c.size() == 2 ) {
		//printf("2 byte : ");

		unsigned char b1 = c[0];
		unsigned char b2 = c[1];

		//printf ( "%x %x\n", b1,b2 );

		if ( b1 >= 0xc2 && b1 <= 0xef ) {
			if ( b2 == 0x80 && b2 == 0xbf ) {
				result = true;
			}
		}
	}
	// 3 byte 文字の場合（つまり半角カナ判定）
	else if ( c.size() == 3) {
		//printf("3 byte : ");

		unsigned char b1 = c[0];
		unsigned char b2 = c[1];
		unsigned char b3 = c[2];

		//printf ( "%x %x %x\n", b1,b2,b3 );

		if ( b1 == 0xef ) {
			if ( b2 == 0xbd ) {
				if ( b3 >= 0xa1 && b3 <= 0xbf) {
					result = true;
				}
			}
			else if ( b2 == 0xbe ) {
				if ( b3 >= 0x80 && b3 <= 0x9f) {
					result = true;
				}
			}
		}
	}

	//printf ( "%s\n", ( result ? "半角" : "全角" ) );

	return result;
}

//=========================================================
// 半角か全角かで幅に変換する（数値）
//=========================================================
int Utils_String::GetCharWidth ( const std::string c ) {

	// 半角なら 1、全角なら 2
	return ( isHalfWidth(c) ? 1 : 2);

}

//=========================================================
// 禁則処理をしつつ指定のバイト数で改行したい。
//=========================================================
vector<string> Utils_String::SplitByByte_Hyphenation(string str, int byte) {
	vector<string> result;

	int stride = 0;

	unsigned int index = 0;
	while ( index < str.size() )
	{
		// 現在位置の文字を取得
		int currentCharIndex = CharCurrent ( str, index );
		int nextCharIndex    = CharNext    ( str, index );
		std::string currentChar = str.substr ( currentCharIndex, nextCharIndex - currentCharIndex );

		stride += nextCharIndex - index;

		index = nextCharIndex;


		// 区切って要素に追加する条件：指定の長さに達した
		if ( stride >= byte ) {


			// 次の文字を取得
			int nextNextCharIndex = CharNext    ( str, nextCharIndex );
			std::string nextChar   = str.substr ( nextCharIndex, nextNextCharIndex - nextCharIndex );

			// 禁則文字じゃないなら区切る
			if ( !isHypenationChar(nextChar) ) {
				std::string element = str.substr ( index - stride, stride );
				result.push_back (element);

				stride = 0;
			}

		// 区切って要素に追加する条件：文字列の最後に到達した
		} else if (index >= str.size()) {
			std::string element = str.substr ( index - stride, stride );
			result.push_back (element);

		}
	}


	return result;
}
//=========================================================
// 禁則処理をしつつ指定の幅数で改行したい。
//=========================================================
vector<string> Utils_String::SplitByWidth_Hyphenation(string str, const int width) {

	vector<string> result;


	int stride = 0;

	int separateMark = 0;
	unsigned int index = 0;
	while ( index < str.size() )
	{
		// 現在位置の文字を取得
		int currentCharIndex = CharCurrent ( str, index );
		int nextCharIndex    = CharNext    ( str, index );
		std::string currentChar = str.substr ( currentCharIndex, nextCharIndex - currentCharIndex );

		// 半角 or 全角 で幅を加算する
		stride += GetCharWidth (currentChar);

		index = nextCharIndex;


		// 区切って要素に追加する条件：指定の長さに達した
		if ( stride >= width ) {

			//printf ( "stride(%d) >= width(%d)\n", stride, width );

			// 次の文字を取得
			int nextNextCharIndex = CharNext    ( str, nextCharIndex );
			std::string nextChar   = str.substr ( nextCharIndex, nextNextCharIndex - nextCharIndex );

			// 禁則文字じゃないなら区切る
			if ( !isHypenationChar(nextChar) ) {

				std::string element = str.substr ( separateMark, index - separateMark );
				result.push_back (element);

				separateMark = index;
				stride = 0;
			} else {
				printf("禁則文字だから区切らない : [%s]\n", nextChar.c_str());
			}

		// 区切って要素に追加する条件：文字列の最後に到達した
		} else if (index >= str.size()) {

			std::string element = str.substr ( separateMark, index - separateMark );
			result.push_back (element);

			separateMark = index;
			stride = 0;
		}
	}


	return result;
}
//=========================================================
// 改行しない禁則処理文字であるかどうか
//【引数】
// 一文字であると期待。
//【戻り値】
// 改行しない文字であれば true
//=========================================================
bool Utils_String::isHypenationChar(string str) {

	bool result = (
				   str != "。"
				&& str != "、"
				&& str != ","
				&& str != "."
				&& str != "！"
				&& str != "？"
				&& str != "!"
				&& str != "?"
				&& str != ")"
				&& str != "）"
				&& str != "]"
				&& str != "」"
				&& str != "}"
				&& str != "っ"
				&& str != "ッ"
				&& str != "-"
				&& str != "ー"
				&& str != " "
			);


	return !result;
}

//=========================================================
// 英大文字を全て小文字にする
//=========================================================
string Utils_String::toLowerString(string str) {

	/*
		全角文字には対応しておりませぬ。

		ファイル名を変換するために使うのであれば、
		ファイル名には全角文字を使わない決まりで。
	*/

	string result = "";

	for (unsigned int i = 0; i < str.size(); ++i) {

		char c = str[i];

		if        (c == 'A') { c = 'a';
		} else if (c == 'B') { c = 'b';
		} else if (c == 'C') { c = 'c';
		} else if (c == 'D') { c = 'd';
		} else if (c == 'E') { c = 'e';
		} else if (c == 'F') { c = 'f';
		} else if (c == 'G') { c = 'g';
		} else if (c == 'H') { c = 'h';
		} else if (c == 'I') { c = 'i';
		} else if (c == 'J') { c = 'j';
		} else if (c == 'K') { c = 'k';
		} else if (c == 'L') { c = 'l';
		} else if (c == 'M') { c = 'm';
		} else if (c == 'N') { c = 'n';
		} else if (c == 'O') { c = 'o';
		} else if (c == 'P') { c = 'p';
		} else if (c == 'Q') { c = 'q';
		} else if (c == 'R') { c = 'r';
		} else if (c == 'S') { c = 's';
		} else if (c == 'T') { c = 't';
		} else if (c == 'U') { c = 'u';
		} else if (c == 'V') { c = 'v';
		} else if (c == 'W') { c = 'w';
		} else if (c == 'X') { c = 'x';
		} else if (c == 'Y') { c = 'y';
		} else if (c == 'Z') { c = 'z';
		}

		result += c;

	}

	return result;
}


//=========================================================
// 文字列をマップとしてパース
//=========================================================
map<string, string> Utils_String::ParseStringAsMap ( string str, string delimiter, string equalSign ) {

	map<string, string> result;
	
	vector<string> elemList = Split ( str, delimiter );
	for ( unsigned int i = 0; i < elemList.size(); ++i ) {
		
		vector<string> elem = Split ( elemList[i], equalSign );
		
		if ( elem.size() >= 2 ) {
			string key   = Trim ( elem[0] );
			string value = Trim ( elem[1] );
			
			result[key] = value;
		}
	}
	
	return result;
}

//=========================================================
// ファイル名をキーリストフィルターを通して取得する
//【引数】
// string ファイル名として渡ってきた文字列。
//=========================================================
string Utils_String::GetFileNameThroughFilter(string fileNameOrKey, map<string, string> &fileNameKeyList) {

	///// アーカイブから読んだとき対策でトリム。
	fileNameOrKey = Trim(fileNameOrKey);

	string result = fileNameOrKey;

	///// １文字目と最終文字を確保
	char firstChar = (fileNameOrKey != "" && fileNameOrKey.length() > 0)?fileNameOrKey.c_str()[0]:NULL;
	char lastChar  = (fileNameOrKey != "" && fileNameOrKey.length() > 0)?fileNameOrKey.c_str()[fileNameOrKey.length() - 1]:NULL;

	///// キーかどうか
	if(firstChar == '[' && lastChar == ']'){

		///// キーを取り出す
		string key = "";
		for(int i = 1; i < fileNameOrKey.length() - 1; ++i){
			key += fileNameOrKey.c_str()[i];
		}

		///// キーがあるかどうか
		if(fileNameKeyList.find(key) != fileNameKeyList.end()){
			result = fileNameKeyList.find(key)->second;
		} else {
			result = "";
		}


	} else {

		///// キーじゃないのでそのままファイル名ということ
		result = fileNameOrKey;
	}

	return result;
}


//=========================================================
// 置換
// src の中の from を to に置換して 結果を返す
//=========================================================
string Utils_String::Replace ( string src, string from, string to ) {
	
	string result = src;
	
	std::string::size_type pos( result.find( from ) );
	
	while( pos != std::string::npos )
	{
	
		result.replace( pos, from.length(), to );
		pos = result.find( from, pos + to.length() );
	}	
	
	return result;

}
	

//=========================================================
// 文字列を二項演算としてパース
//=========================================================
Utils_String::BinaryOperationElements Utils_String::ParseAsBinaryOperation ( string str ) {
	
	Utils_String::BinaryOperationElements result;
	
	
	/*
		"optionSmile = notSelectedYet"
		
		のような文字列を想定。
		各要素の間には半角スペースが入っていることを想定。
		３項目以降があったとしても無視
	*/
	
	vector<string> raw = Split ( str, " " );
	
	for ( unsigned int i = 0; i < raw.size(); ++i ) {
		
		if ( raw[i] != "") {
			
			if ( result.left == "" ) {
					
				result.left = raw[i];
			
			} else if ( result.op == "" ) {
				
				result.op = raw[i];
			
			} else if ( result.right == "" ) {
				
				result.right = raw[i];
				break;
			}
		}
	}
	
	return result;
}
	

//=========================================================
// 関係演算：数値
//=========================================================
bool Utils_String::ExecuteRelationalExpression ( int left, string op, int right ) {
	
	bool result = false;
	
	if      ( op == "<"                ) { result = ( left <  right ); }
	else if ( op == "<=" || op == "=<" ) { result = ( left <= right ); }
	else if ( op == ">"                ) { result = ( left >  right ); }
	else if ( op == ">=" || op == "=>" ) { result = ( left >= right ); }
	else if ( op == "!=" || op == "<>" ) { result = ( left != right ); }
	else if ( op == "="  || op == "==" ) { result = ( left == right ); }

	else if ( op == "lt" ) { result = ( left <  right ); }
	else if ( op == "le" ) { result = ( left <= right ); }
	else if ( op == "gt" ) { result = ( left >  right ); }
	else if ( op == "ge" ) { result = ( left >= right ); }
	else if ( op == "ne" ) { result = ( left != right ); }
	else if ( op == "eq" ) { result = ( left == right ); }
	
	
	return result;


}
//=========================================================
// 関係演算：文字列
//=========================================================
bool Utils_String::ExecuteRelationalExpression ( string left, string op, string right ) {
	
	bool result = false;
	
	if      ( op == "!=" || op == "<>" ) { result = ( left !=  right ); }
	else if ( op == "="  || op == "==" ) { result = ( left ==  right ); }
	
	return result;
}
	
	
	
	
	
	
	
