//=========================================================
//
// 文字列関係のサブルーチン集（宣言）
//
//=========================================================

#ifndef _UTILS_STRING_
#define _UTILS_STRING_


#include <string>
#include <vector>
#include <map>


using namespace std;

class Utils_String {
	
public:
	
	//=========================================================
	// 二項演算用文字列データ
	//=========================================================
	struct BinaryOperationElements {
		
		string left;
		string right;
		string op;
		
		BinaryOperationElements () {
			left = "";
			right = "";
			op = "";
		}
	};

	//=========================================================
	// 前後Trim
	//【引数】
	// string
	//【戻り値】
	// 前後の半角スペース、\n、\rを取り除いた文字列
	//=========================================================
	static string Trim(string str);

	//=========================================================
	// Split
	// https://gist.github.com/shnya/735066#file-c-split-find-version
	//=========================================================
	static vector<string> Split(string str, string delim);


	//=========================================================
	// 指定バイト数ごとに区切ったリストを取得
	//【引数】
	// string str  : 元の文字列
	// int    byte : このバイト数で区切る
	//【戻り値】
	// vector<string>
	//=========================================================
	static vector<string> SplitByByte(string value, int byte);

	//=========================================================
	// 肯定を表す文字列かどうか。
	//【戻り値】
	// true で 肯定を表す文字列ですよ。
	//=========================================================
	static bool isPositiveWord(string arg);

	//=========================================================
	// 否定を表す文字列かどうか。
	//【戻り値】
	// true で 否定を表す文字列ですよ。
	//=========================================================
	static bool isNegativeWord(string arg);

	//=========================================================
	// vector<string>の中に指定の要素があるか？
	//=========================================================
	static bool isContain(vector<string> list, string element);

	//=========================================================
	// 文字列中に指定のパターンが含まれているかどうか
	//【戻り値】
	// true で 含まれている。
	//=========================================================
	static bool IsContainsPattern(string dest, string pattern);


	//=========================================================
	// 指定位置の次の文字の先頭バイトのインデックスを取得
	//=========================================================
	static int CharNext ( std::string str, int index  );

	//=========================================================
	// 指定位置の文字の先頭バイトのインデックスを取得
	//=========================================================
	static int CharCurrent ( std::string str, int index  );


	//=========================================================
	// 全角/半角 判定（引数は１文字である前提。文字数チェックは入れていない）
	// 参考：
	//	http://orange-factory.com/dnf/utf-8.html
	//	http://tech.nitoyon.com/ja/blog/2014/03/14/utf8-str-count/
	//	http://www.unicode.org/Public/UCD/latest/ucd/EastAsianWidth.txt
	//=========================================================
	static bool isHalfWidth ( const std::string c );


	//=========================================================
	// 半角か全角かで幅に変換する（数値）
	//=========================================================
	static int GetCharWidth ( const std::string c );


	//=========================================================
	// 禁則処理をしつつ指定のバイト数で改行したい。
	//=========================================================
	static vector<string> SplitByByte_Hyphenation(string str, int byte);

	//=========================================================
	// 禁則処理をしつつ指定の幅数で改行したい。
	//=========================================================
	static vector<string> SplitByWidth_Hyphenation(string str, const int width);
	
	//=========================================================
	// ファイル名をキーリストフィルターを通して取得する
	//【引数】
	// string ファイル名として渡ってきた文字列。
	//=========================================================
	static string GetFileNameThroughFilter(string fileNameOrKey, map<string, string> &fileNameKeyList);

	//=========================================================
	// 改行しない禁則処理文字であるかどうか
	//【引数】
	// 一文字であると期待。
	//【戻り値】
	// 改行しない文字であれば true
	//=========================================================
	static bool isHypenationChar(string str);

	//=========================================================
	// 英大文字を全て小文字にする
	//=========================================================
	static string toLowerString(string str);
	
	
	//=========================================================
	// 置換
	// src の中の from を to に置換して 結果を返す
	//=========================================================
	static string Replace ( string src, string from, string to );
	
	
	//=========================================================
	// 文字列を二項演算としてパース
	//=========================================================
	static BinaryOperationElements ParseAsBinaryOperation ( string str );

	//=========================================================
	// 文字列をマップとしてパース
	//=========================================================
	static map<string, string> ParseStringAsMap ( string str, string delimiter, string equalSign );
	
	
	//=========================================================
	// 関係演算
	//=========================================================
	static bool ExecuteRelationalExpression ( int left, string op, int right );
	static bool ExecuteRelationalExpression ( string left, string op, string right );
	

};
#endif
