//=========================================================
//
// ゲーム盤管理クラス
//
//=========================================================

#include "Application.h"
#include "Define.h"
#include "GameManager.h"
#include "Board.h"
#include "CardBox.h"
#include "PieceBox.h"
#include "Piece.h"
#include "CardFactory.h"
#include "Card.h"
#include "Grid.h"
#include "Define.h"
#include "Game_Utils.h"
#include "GameRecord.h"
#include "AI_Manager.h"
#include "Actor.h"
#include "Delegate.h"
#include "PsychoGage.h"
#include "Arrow.h"
#include "Image.h"
#include "Select.h"
#include "SelectOption.h"
#include "InterfereModePanel.h"
#include "CurrentTurnNamePanel.h"
#include "Effect_PutPiece_Human.h"
#include "Effect_PutPiece_AI.h"
#include "Effect_PutCard_Human.h"
#include "Effect_PutCard_AI.h"
#include "Effect_TakeCard.h"
#include "Effect_MovePiece.h"
#include "Effect_MovePiece_Collide.h"
#include "Effect_RepetitionMoveMessage.h"
#include "Effect_RepetitionMoveAlertMessage.h"
#include "EffectManager.h"
#include "ScriptEngine.h"

#include "../AI/AI.h"
#include "../AI/Move.h"
#include "../AI/Define.h"
#include "../AI/InnerExpression.h"

#include "../GUI_SFML/SoundPlayer.h"
#include "../GUI_SFML/TextImageFactory.h"

#include "../Utils/Utils_Cal.h"
#include "../Utils/Utils_String.h"
#include "../Utils/Utils_Others.h"
#include "../Utils/Utils_Convert.h"
#include "../Utils/Utils_File.h"
#include "../Utils/mt19937ar.h"

//=========================================================
// コンストラクタ
//=========================================================
GameManager::GameManager(string name, ScriptEngine *scriptEngine) {

	///// 活性ではない。
	_isActive = false;

	_isRendered = true;

	_name = name;
	
	_scriptEngine = scriptEngine;


	// 棋譜オブジェクトは常にある、ということにする
	// _score = NULL;
	_score = new GameRecord ();
	_kifuViewerPosition = sf::Vector2f(515.0f, 410.0f );
	_score->SetPosition(_kifuViewerPosition);
	_score->Hide();

	//_AI          = NULL;
	//_actor       = NULL;
	//_AI_fileName = ""; // セーブ/ロード用のメモ

	///// フリー操作は、基本的に OFF
	_isFreeOperation = false;

	_isAstralProjection = false;
	_astralProjection_physicalState = NULL;

	///// 空の盤面。
	_pieceBox_P1 = new PieceBox();
	_pieceBox_P2 = new PieceBox();
	_cardBox_P1  = new CardBox();
	_cardBox_P2  = new CardBox(true);

	_boardCenterX = APPLICATION::WINDOW_SIZE::WIDTH  / 2.0f;
	_boardCenterY = APPLICATION::WINDOW_SIZE::HEIGHT / 2.0f;

	_board       = NULL;

	// SetUpBoard の前にインスタンスを生成している前提
	_currentTurnNamePanel = new CurrentTurnNamePanel();


	SetUpBoard( 9, 9 );

	// アクターの位置のデフォルト（ここでシステム的な固定値を書いてしまえばいいのかも）
	//_actorPosisionX = 100.0f;
	//_actorPosisionY = APPLICATION::WINDOW_SIZE::HEIGHT / 2.0f;


	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	// 各種スクリプトファイル＿未設定
	//_scriptFile_MoveEffectDecision = "";
	//_scriptFile_EmotionProcess     = "";
	//_scriptFile_Reaction           = "";

	// AIをセットアップ
	//SetUpAI();
	// ルールをセットアップ
	SetUpRule();

	///// ルール上の、操作状態
	_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

	///// 手番
	//_currentTurn = AI_Modules::TURN::BLACK;

	_selectingGrid = NULL;
	_selectingGridX = -1;
	_selectingGridY = -1;

	///// AIスレッド関連
	_AI_Manager = new AI_Manager();
	//_AI_ThreadHandle = NULL;
	//InitializeCriticalSection( &_criticalSection_For_AI );

	_AI_ProgressBar = new GUI_SFML::ProgressBar ( sf::Vector2f ( 150.0f, 20.0f ), 100, "%" );


	///// 気力ゲージ
	_psychoGage  = new PsychoGage();
	_psychoGage->SetPosition ( sf::Vector2f ( 750, 200 ) );

	_arrowForMoveDemonstration = new Arrow();
	_crashForMoveDemonstration = new Image();
	_crashForMoveDemonstration->ChangeImage ( "./image/gameBoard/crash.png" );

	_interfereModePanel   = new InterfereModePanel();

	// ゲーム管理側として出す選択肢。ある場合だけインスタンスを入れる。NULLじゃないなら更新/描画する
	_select = NULL;


	///// 着手エフェクトからの終了通知受け取り
	_endMoveFromEffect   = Delegate<GameManager>::createDelegator_move ( this, &GameManager::EndMoveFromEffect );
	_endMoveFromAICutIn  = Delegate<GameManager>::createDelegator_int  ( this, &GameManager::ExecuteAIMove );

	_is_MoveEffectGoing = false;



	_operationMemo_belongsToWhichPieceBox = "";
	_operationMemo_whatColorThePieceWas   = -1;
	_operationMemo_belongsToWhichCardBox  = "";
	_operationMemo_whichCardTheCardWas    = "";
	
	
	_isScriptingMoving = false; // 指し手が完了するまで待機したいときがある。スクリプトでの着手進行とか。
	
	_isTutorial = true;
}

//=========================================================
// 盤のサイズを設定
//=========================================================
void GameManager::SetUpBoard ( int width, int height ) {

	/*
		すでに盤があった場合、
		盤上のカードと駒は、解放されます。
	*/

	///// あったら、お掃除。
	// カードと駒のリストだけ退避しておいて、新Boardに後で乗せればいいのかも？
	if ( _board != NULL ) {

		vector<vector<Grid*>> gridList = _board->GetAllGrid();
		for ( unsigned int y = 0; y < gridList.size(); ++y ) {
			for ( unsigned int x = 0; x < gridList[y].size(); ++x ) {

				// 駒とカードを解放しておく。
				ReleasePiece ( gridList[y][x]->GetPiece() );
				CardFactory::Release (gridList[y][x]->GetCard());
			}
		}

		delete _board;
	}

	///// 新たに生成。
	_board = new Board ( width, height );


	// 表示位置をセットアップ
	SetUpPartsDrawPositions();

	// フリー操作フラグを周知しておく：盤のサイズを変えるたびに実行する必要あり。各グリッドへ。
	// これは、カードにも設定する必要ありますよね。追加するたびに。
	SetFreeOperation ( _isFreeOperation );

}


//=========================================================
// デストラクタ
//=========================================================
GameManager::~GameManager() {

printf ("0 GameManager::~GameManager()\n");

	ReleaseCardFromMouseIfBelongToHere();

	//DeleteCriticalSection( &_criticalSection_For_AI );

	delete _board;
	delete _pieceBox_P1;
	delete _pieceBox_P2;
	delete _cardBox_P1;
	delete _cardBox_P2;
	delete _AI_Manager;
	//delete _AI;  // スレッド(_AI_Manager)を解放してから、AIを解放する。
	clearAllParticipants();
	delete _AI_ProgressBar;
	delete _endMoveFromEffect;
	delete _endMoveFromAICutIn;
	delete _score;
	//delete _actor;
	delete _psychoGage;
	delete _arrowForMoveDemonstration;
	delete _crashForMoveDemonstration;
	delete _select;
	delete _interfereModePanel;
	delete _currentTurnNamePanel;

printf ("100 GameManager::~GameManager()\n");
	ReleaseAllPiece();

printf ("1000 GameManager::~GameManager()\n");
	TerminateAllOurBeLovedEffects();

printf ("10000 GameManager::~GameManager()\n");

	/*
	// うちで担当したカードをマウスに渡しっぱなしになってましたっけ？
	if ( _operationMemo_belongsToWhichCardBox != "" ) {

		printf ( "マウスでカードを持ったままなんじゃないですか？ [%s]\n", _operationMemo_belongsToWhichCardBox.c_str() );

		if ( Application::GetInstance()->IsMouseClingObj_Card() ) {

			printf ( "マウスから外して解放しておきますよ\n" );

			Card *card = Application::GetInstance()->DrawMouseClingObj();
			CardFactory::Release(card);

		} else {
			printf ( "え？ マウスにはカードがくっついてない？ おかしいなぁ。\n" );
		}
	}
	*/


printf ("1000000000 GameManager::~GameManager()\n");
}

//=========================================================
// つまり、ロード時に、操作用のエフェクトを削除する目的
//=========================================================
void GameManager::LetsRemenberThatIMadeThisEffect ( IEffect *effect ) {

	_takeCareOfTheseEffects.push_back ( effect );

}

//=========================================================
// つまり、ロード時に、操作用のエフェクトを削除する目的
//=========================================================
void GameManager::TerminateAllOurBeLovedEffects() {

	for ( IEffect *effect : _takeCareOfTheseEffects ) {

		Application::GetInstance()->_effectManager->Terminate ( effect );
	}
}

//=========================================================
// うちの子をマウスでつかんでるなら、マウスから外して解放する：カード編
//=========================================================
void GameManager::ReleaseCardFromMouseIfBelongToHere() {

	if ( Application::GetInstance()->IsMouseClingObj_Card() ) {

		printf ( "マウスでカードを持っていますね〜\n"  );
		Card *cardOnMouse = (Card*)Application::GetInstance()->RefferMouseClingObj();

		bool isBelongTo = false;
		for ( unsigned int i = 0; i < _cardList_forManagement.size(); ++i) {

			Card *card = _cardList_forManagement[i];

			if ( card == cardOnMouse ) {

				printf ( "うちの子なので解放しておきましょう\n" );
				isBelongTo = true;

				Application::GetInstance()->DrawMouseClingObj();
				CardFactory::Release ( card );

				UTILS_OTHERS::deleteFromVector<Card*>(_cardList_forManagement, i);

				break;
			}
		}

		if ( !isBelongTo ) {
			printf ( "うちの子じゃないので放っておきます\n" );
		}
	}
}

//=========================================================
// うちの子をマウスでつかんでるなら、マウスから外して解放する：駒編
//=========================================================
void GameManager::ReleasePieceFromMouseIfBelongToHere() {

	if ( Application::GetInstance()->IsMouseClingObj_Piece() ) {

		printf ( "マウスで駒を持っていますね〜\n"  );
		Piece *pieceOnMouse = (Piece*)Application::GetInstance()->RefferMouseClingObj();

		bool isBelongTo = false;
		for ( unsigned int i = 0; i < _cardList_forManagement.size(); ++i) {

			Piece *piece = _pieceList_forManagement[i];

			if ( piece == pieceOnMouse ) {

				printf ( "うちの子なので解放しておきましょう\n" );
				isBelongTo = true;

				Application::GetInstance()->DrawMouseClingObj();
				delete piece;

				UTILS_OTHERS::deleteFromVector<Piece*>(_pieceList_forManagement, i);

				break;
			}
		}

		if ( !isBelongTo ) {
			printf ( "うちの子じゃないので放っておきます\n" );
		}
	}
}



//==========================================================
// 駒オブジェクトを解放
//==========================================================
void GameManager::ReleaseAllPiece() {

	IMouseClingableObject *mouseClingObj = Application::GetInstance()->RefferMouseClingObj();

	while (_pieceList_forManagement.size() > 0 ) {

		Piece *piece = *(_pieceList_forManagement.begin());

		// マウスにくっついていたら外しておく
		if ( mouseClingObj != NULL && mouseClingObj == piece) {

			printf("[GameManager::ReleaseAllPiece] 駒がマウスにくっついていたので外しておく\n");
			Application::GetInstance()->DrawMouseClingObj();
		}

		delete *(_pieceList_forManagement.begin());

		_pieceList_forManagement.erase(_pieceList_forManagement.begin());
	}

}
void GameManager::ReleasePiece ( Piece *piece ) {

	bool isExist = false;

	// マウスにくっついていたら外しておく
	IMouseClingableObject *mouseClingObj = Application::GetInstance()->RefferMouseClingObj();
	if ( mouseClingObj != NULL && mouseClingObj == piece ) {

		printf("[GameManager::ReleasePiece] 駒がマウスにくっついていたので外しておく : [%p]\n", piece);
		Application::GetInstance()->DrawMouseClingObj();
	}


	// 管理リストにあれば、そこからも消しておく
	for ( vector<Piece*>::iterator it = _pieceList_forManagement.begin(); it != _pieceList_forManagement.end(); ++it ) {

		if ( *it == piece ) {
			isExist = true;
			delete piece;
			_pieceList_forManagement.erase ( it );
			break;
		}
	}

	if (!isExist) {
		delete piece;
	}
}

//=========================================================
// 盤の中心の表示位置を設定。これに合わせて他のパーツが並ぶ。
//=========================================================
void GameManager::SetBoardCenterPosition ( int x, int y ) {

	_boardCenterX = x;
	_boardCenterY = y;

	SetUpPartsDrawPositions();
}

//=========================================================
// アクターの基準位置を設定
//=========================================================
void GameManager::SetActorPosition ( int x, int y ) {

	//_actorPosisionX = x;
	//_actorPosisionY = y;
}

//=========================================================
// 試合のタイトルを設定
//=========================================================
void GameManager::SetBattleTitle ( string battleTitle ) {
	
	_battleInfo.battleTitle = battleTitle;
	
	if ( _score != NULL ) {
		_score->SetBattleTitle(_battleInfo.battleTitle);
	}
	
}

//=========================================================
// 各パーツの表示位置を定める
//=========================================================
void GameManager::SetUpPartsDrawPositions() {

	_boardX = _boardCenterX - _board->GetBaseWidth() / 2.0f;
	_boardY = _boardCenterY - _board->GetBaseHeight() / 2.0f;// - 90.0f;

	_carBoxP1X = _boardX + (float)_board->GetBaseWidth()     + 10.0f + ( ((float)_pieceBox_P1->GetBoxWidth() - (float)_cardBox_P1->GetBoxWidth()) / 2.0f);
	_carBoxP2X = _boardX - (float)_cardBox_P2->GetBoxWidth() - 10.0f - ( ((float)_pieceBox_P2->GetBoxWidth() - (float)_cardBox_P2->GetBoxWidth()) / 2.0f);


	// 駒台と、カード置き場を合わせた高さ
	int P1_BoxHeight = _pieceBox_P1->GetBoxHeight() + _cardBox_P1->GetBoxHeight() + 3;
	int P2_BoxHeight = _pieceBox_P2->GetBoxHeight() + _cardBox_P2->GetBoxHeight() + 3;

	_cardBoxP1Y = _boardCenterY - (float)P1_BoxHeight / 2.0f;
	_cardBoxP2Y = _boardCenterY - (float)P2_BoxHeight / 2.0f + _pieceBox_P2->GetBoxHeight() + 3.0f;

	_pieceBoxP1X = _boardX + _board->GetBaseWidth()      + 10.0f;
	_pieceBoxP2X = _boardX - _pieceBox_P2->GetBoxWidth() - 10.0f;

	_pieceBoxP1Y = _cardBoxP1Y + _cardBox_P1->GetBoxHeight()  + 3.0f;
	_pieceBoxP2Y = _cardBoxP2Y - _pieceBox_P2->GetBoxHeight() - 3.0f;

	_AI_ProgresBarX_P1 = _pieceBoxP1X;
	_AI_ProgresBarY_P1 = _cardBoxP1Y - 30;

	_AI_ProgresBarX_P2 = _pieceBoxP2X;
	_AI_ProgresBarY_P2 = _pieceBoxP2Y - 30;

	_interfereModePanelX = 650;
	_interfereModePanelY = 550;



	///// 強引に小数点以下を丸める：小数点以下があると、表示したとき、ぼやける。floatを使うのがすでに間違いか？
	Utils_Cal::RoundWildly ( _boardX      );
	Utils_Cal::RoundWildly ( _boardY      );
	Utils_Cal::RoundWildly ( _carBoxP1X   );
	Utils_Cal::RoundWildly ( _carBoxP2X   );
	Utils_Cal::RoundWildly ( _cardBoxP1Y  );
	Utils_Cal::RoundWildly ( _cardBoxP2Y  );
	Utils_Cal::RoundWildly ( _pieceBoxP1X );
	Utils_Cal::RoundWildly ( _pieceBoxP2X );
	Utils_Cal::RoundWildly ( _pieceBoxP1Y );
	Utils_Cal::RoundWildly ( _pieceBoxP2Y );
	Utils_Cal::RoundWildly ( _AI_ProgresBarX_P1 );
	Utils_Cal::RoundWildly ( _AI_ProgresBarY_P1 );
	Utils_Cal::RoundWildly ( _AI_ProgresBarX_P2 );
	Utils_Cal::RoundWildly ( _AI_ProgresBarY_P2 );
	Utils_Cal::RoundWildly ( _interfereModePanelX );
	Utils_Cal::RoundWildly ( _interfereModePanelY );

	_currentTurnNamePanel->SetPosition ( _boardCenterX, _boardY - 20.0f );



}

//=========================================================
// AIをセットアップ
//=========================================================
//void GameManager::SetUpAI() {

//=========================================================
// ルールをセットアップ
//=========================================================
void GameManager::SetUpRule() {

	/*
	///// 一応、既存のがあれば消しておく。
	if ( _AI != NULL ) {
		delete _AI;
		_AI = NULL;
	}

	///// こんにちわ。
	_AI = new AI_Modules::AI();
	*/

	///// ルールをセット
	AI_Modules::Rule rule;
	rule.boardWidth  = _board->GetWidth();
	rule.boardHeight = _board->GetHeight();

	rule.isRepetitionMoveProhibited = true;
	rule.repetitionMoveMaxCount = 2;

	map<int, AI_Modules::Card> cardListForAI;
	map<int, string>           cardNameListForAI;

	map<int, Card::Property>::iterator cardIt = CardFactory::_cardPropertyList.begin();
	while ( cardIt != CardFactory::_cardPropertyList.end() ) {

		int            cardId   = cardIt->first;
		Card::Property cardProp = cardIt->second;

		AI_Modules::Card cardForAI;
		cardForAI.processId = cardId;

		for (unsigned int i = 0; i < cardProp.gridAffectionList.size(); ++i) {

			AI_Modules::GridAffection gridAffectionForAI;
			gridAffectionForAI.x =  cardProp.gridAffectionList[i].x;
			gridAffectionForAI.y =  cardProp.gridAffectionList[i].y;
			gridAffectionForAI.movableDirection =  cardProp.gridAffectionList[i].movableDirection;

			cardForAI.gridAffectionList.push_back( gridAffectionForAI );
		}

		cardListForAI[ cardId ] = cardForAI;
		cardNameListForAI[ cardId ] = cardProp.cardName;

		++cardIt;
	}
	AI_Modules::AI::SetRule( rule, cardListForAI);
	AI_Modules::AI::SetCardNameList( cardNameListForAI);

	/*
	///// PEE の設定ファイルを読み込む スクリプトで設定できるようにしましょう。
	//string PEE_File = "./data/AI/savePEE_from_Plain.txt";
	string PEE_File = "";
	AI_Modules::Parameters_PieceEvaluation PEE = AI_Modules::AI::PEE_Read_File( PEE_File );
	_AI->PEE_Init(PEE);
	*/
}


//=========================================================
// 参加者登録
//=========================================================
void GameManager::addParticipant ( Participant participant ) {

	// 同じIDのがいたらマズイけど、面倒なのでチェックはしない。気をつけて設定してね。
	_participantsList.push_back (participant);

	printf ("参加者登録（ID:%s SIDE:%s NAME:%s）\n", participant.id.c_str(), participant.side.c_str(), participant.name.c_str());
}

//=========================================================
// 参加者を全部クリア
//=========================================================
void GameManager::clearAllParticipants () {

	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

		delete _participantsList[i].ai;
		delete _participantsList[i].actor;
	}

	_participantsList.clear();
}

//=========================================================
// AI設定
//=========================================================
/*
void GameManager::SetAIFile ( string AI_Name, string fileName ) {

	// AIを複数持つことを見越して、名前を受け取っている。

	AI_Modules::Parameters_PieceEvaluation PEE = AI_Modules::AI::PEE_Read_File( fileName );
	_AI->PEE_Init(PEE);

	// セーブ/ロード用のメモとして設定ファイル名を覚えておく。AI_Name の方はさしあたり忘れていい。
	_AI_fileName = fileName;
}
*/
void GameManager::SetAIFile ( string participantID, string aiFileName ) {

	printf ("%s にAIファイル設定：%s\n", participantID.c_str(), aiFileName.c_str());
	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

		if ( _participantsList[i].id == participantID ) {

			if ( aiFileName != "" ) {

				if (_participantsList[i].ai != NULL) { delete _participantsList[i].ai; }
				_participantsList[i].ai = new AI_Modules::AI();

				AI_Modules::Parameters_PieceEvaluation PEE = AI_Modules::AI::PEE_Read_File( aiFileName );
				_participantsList[i].ai->PEE_Init(PEE);

				_participantsList[i].aiFileName = aiFileName;

			} else {
				if (_participantsList[i].ai != NULL) { delete _participantsList[i].ai; }
				_participantsList[i].ai = NULL;

			}
			break;
		}
	}
	printf ("AIファイル設定end\n");
}

//=========================================================
// アクター設定
//=========================================================
/*
void GameManager::SetActorScriptFile ( string AI_Name, string fileName, string playingSide, GUI_SFML::TextMessageWindow *textWindow ) {

	// AIを複数持つことを見越して、名前を受け取っている。
	// （今は、AIもアクターも１つずつしかいない）

	if ( _actor != NULL ) {
		delete _actor;
	}
	_actor = new Actor ( fileName, textWindow );

	_actor->SetPlayingSide ( playingSide );

	///// 非活性。明示的に「活性」といわれない限り、表示しない方針。ScriptEngine側のフェイズにゆだねる。
	_actor->SetActive(false);

}
*/
void GameManager::SetActorScriptFile ( GameManager::Participant participantInfo, string actorFileName, int posX, int posY, int startPosX, int startPosY, GUI_SFML::TextMessageWindow *textWindow ) {

	string participantID = participantInfo.id;

	printf ("%s にActorファイル設定：%s\n", participantID.c_str(), actorFileName.c_str());

	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

		if ( _participantsList[i].id == participantID ) {

			if ( actorFileName != "" ) {

				if ( _participantsList[i].actor != NULL ) {
					delete _participantsList[i].actor;
				}
				_participantsList[i].actor = new Actor ( actorFileName, textWindow );

				_participantsList[i].actor->SetPlayingSide ( _participantsList[i].side );

				sf::Vector2f startPosition (startPosX, startPosY);
				sf::Vector2f goalPosition  (posX, posY);
				_participantsList[i].actor->SetPosition    ( startPosition );
				_participantsList[i].actor->MoveTo         ( goalPosition );
				_participantsList[i].actor->SetPlayingSide ( participantInfo.side );
				_participantsList[i].actor->SetName        ( participantInfo.name );

				_participantsList[i].actorFileName = actorFileName;


				///// 非活性。明示的に「活性」といわれない限り、表示しない方針。ScriptEngine側のフェイズにゆだねる。
				_participantsList[i].actor->SetActive(false);

			} else {
				if ( _participantsList[i].actor != NULL ) {
					delete _participantsList[i].actor;
				}
				_participantsList[i].actor = NULL;
			}

			break;
		}
	}
	printf ("Actorファイル設定end\n");
}


/*
//=========================================================
// アクターさんにアクセスしたい！
//=========================================================
Actor* GameManager::GetActor( string AI_Name ) {

	// AIを複数持つことを見越して、名前を受け取っている。
	// （今は、AIもアクターも１つずつしかいない）

	return _actor;
}
*/
//=========================================================
///// 全アクターさんの active をセット
//=========================================================
void GameManager::SetAllActorsActivity ( bool isActive ) {

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].actor != NULL ) {
			_participantsList[i].actor->SetActive ( isActive );
		}
	}
}

//=========================================================
///// 全アクターさんを Init
//=========================================================
void GameManager::InitAllActors () {

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].actor != NULL ) {
			_participantsList[i].actor->Init();
		}
	}

}

//=========================================================
// 現在のターンの参加者を取得
//=========================================================
GameManager::Participant GameManager::GetCurrentTurnParticipant () {

	GameManager::Participant result;

	bool isParticipantExit = false;

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].id == _operationData.currentTurn ) {
			result = _participantsList[i];
			isParticipantExit = true;
			break;
		}
	}

	if ( !isParticipantExit ) {
		printf ( "参加者がいない （at GameManager::GetCurrentTurnParticipant）\n" );
	}

	return result;
}

//=========================================================
// 現在のターンのアクターさんを取得
//=========================================================
Actor* GameManager::GetCurrentTurnActor () {

	Actor *result = NULL;

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].id == _operationData.currentTurn ) {
			result = _participantsList[i].actor;
			break;
		}
	}

	return result;
}

//=========================================================
// 全アクターさんを取得
//=========================================================
vector<Actor*> GameManager::GetAllActors () {

	vector<Actor*> result;

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		// アクターを持っている分だけ。
		if ( _participantsList[i].actor != NULL ) {
			result.push_back ( _participantsList[i].actor);
		}
	}

	return result;
}

//=========================================================
// 掛け合い受付
//=========================================================
//void GameManager::AcceptKakeai ( string participantId, string remark, string faceName ) {
void GameManager::AcceptKakeai ( Actor::KakeaiInfo kakeaiInfo ) {

	string participantId = kakeaiInfo.participantId;
	string faceName      = kakeaiInfo.faceName;
	string remark        = kakeaiInfo.remark;
	string operation     = kakeaiInfo.operation;

	///// 指名されたアクターを探してリアクションを実行してもらう
	for (unsigned int i = 0; i < _participantsList.size(); ++i ) {

		if ( _participantsList[i].id == participantId ) {

			if ( _participantsList[i].actor != NULL ) {

				_participantsList[i].actor->AcceptKakeai ( remark, faceName, operation );

			}
			break;

		}
	}

	///// こちらで受け持つ分のOperation
	if ( operation == "FORCE-QUIT" ) {

		_battleInfo.result = "force-quit";

		// つまり、もう指さないで欲しいということ
		_progressData_AfterThinking.progressRate = 0.0f;
		_progressData_AfterThinking.answerMove   = AI_Modules::Move();
		_progressData_AfterThinking.isStop       = true;

		_AI_Manager->ThinkStop();

	}

}

//=========================================================
// チーム認識
//=========================================================
void GameManager::TeamForming() {

	// 既存のものをクリアしまして、
	_teamActorP1.clear();
	_teamActorP2.clear();


	// まずは人数に関係なく、各サイドのチームへ振り分けておく（アクターがいる場合）
	for (unsigned int i = 0; i < _participantsList.size(); ++i ) {

		if ( _participantsList[i].side == "P1" && _participantsList[i].actor != NULL) {
			_teamActorP1.push_back(_participantsList[i].actor);
		}
		if ( _participantsList[i].side == "P2" && _participantsList[i].actor != NULL ) {
			_teamActorP2.push_back(_participantsList[i].actor);
		}
	}

	// １人のチームはチーム扱いはしない
	if (_teamActorP1.size() < 2) { // 2名に満たない場合
		_teamActorP1.clear();
	}
	if (_teamActorP2.size() < 2) { // 2名に満たない場合
		_teamActorP2.clear();
	}
}

//=========================================================
// ペアバトルのチームに対して、ターンの開始を告げる。表示位置などをよしなに
// すでに新ターンになっているものとする
//=========================================================
void GameManager::TeamNewTurnStart() {

	Actor *currentTurnActor = GetCurrentTurnActor();

	// もっとも「前に出ている」アクターを求める
	Actor *mostAdvancedActor = currentTurnActor; // デフォルトで自分としておく

	// アクターの手番じゃないなら何もしない
	if ( currentTurnActor == NULL ) {
		return;
	}

	// どっちサイドのチーム？ P1の場合
	if ( currentTurnActor->GetPlayingSide() == "P1" ) {

		if (_teamActorP1.empty() ) { // チームじゃないなら何もしない
			return;
		}
		// もっとも「前に出ている」アクターを求める
		for ( unsigned int i = 0; i < _teamActorP1.size(); ++i ) {

			if ( _teamActorP1[i] == NULL ) {printf("ん？何これ？ GameManager::TeamNewTurnStart()\n");continue;} // 念のため

			// P1チームなので「前」とは「x値が小さい方」
			if ( mostAdvancedActor->GetPosition().x > _teamActorP1[i]->GetPosition().x ) {
				mostAdvancedActor = _teamActorP1[i];
			}
		}
		// Z表示順として、一番手前に移動
		for ( vector<Actor*>::iterator it = _teamActorP1.begin(); it != _teamActorP1.end(); ++it) {

			Actor *a = *it;
			if ( a == currentTurnActor ) { // リストから一度消した上で、最後尾に追加しなおす
				_teamActorP1.erase(it);
				_teamActorP1.push_back(currentTurnActor);
			}
		}

	}

	// どっちサイドのチーム？ P2の場合
	if ( currentTurnActor->GetPlayingSide() == "P2" ) {

		if (_teamActorP2.empty() ) { // チームじゃないなら何もしない
			return;
		}
		// もっとも「前に出ている」アクターを求める
		for ( unsigned int i = 0; i < _teamActorP2.size(); ++i ) {

			if ( _teamActorP2[i] == NULL ) {printf("ん？何これ？ GameManager::TeamNewTurnStart()\n");continue;} // 念のため

			// P2チームなので「前」とは「x値が大きい方」
			if ( mostAdvancedActor->GetPosition().x < _teamActorP2[i]->GetPosition().x ) {
				mostAdvancedActor = _teamActorP2[i];
			}
		}

		// Z表示順として、一番手前に移動
		for ( vector<Actor*>::iterator it = _teamActorP2.begin(); it != _teamActorP2.end(); ++it) {

			Actor *a = *it;
			if ( a == currentTurnActor ) { // リストから一度消した上で、最後尾に追加しなおす
				_teamActorP2.erase(it);
				_teamActorP2.push_back(currentTurnActor);
			}
		}
	}

	// もっとも前に出ていた人と位置入れ替え
	sf::Vector2f advancedPos = mostAdvancedActor->GetPosition();
	sf::Vector2f currentPos  = currentTurnActor->GetPosition();

	currentTurnActor->MoveTo ( advancedPos );
	mostAdvancedActor->MoveTo ( currentPos );

}


/*
void GameManager::SetScriptFile_MoveEffectDecision  ( string AI_Name, string fileName ) {

	// AIを複数持つことを見越して、名前を受け取っている。

	_scriptFile_MoveEffectDecision = fileName;
}
void GameManager::SetScriptFile_EmotionProcess      ( string AI_Name, string fileName ) {

	// AIを複数持つことを見越して、名前を受け取っている。

	_scriptFile_EmotionProcess = fileName;
}
void GameManager::SetScriptFile_Reaction            ( string AI_Name, string fileName ) {

	// AIを複数持つことを見越して、名前を受け取っている。

	_scriptFile_Reaction = fileName;
}
string GameManager::GetScriptFile_MoveEffectDecision  ( string AI_Name ) {

	// AIを複数持つことを見越して、名前を受け取っている。
	return _scriptFile_MoveEffectDecision;
}
string GameManager::GetScriptFile_EmotionProcess      ( string AI_Name ) {

	// AIを複数持つことを見越して、名前を受け取っている。
	return _scriptFile_EmotionProcess;
}
string GameManager::GetScriptFile_Reaction            ( string AI_Name ) {

	// AIを複数持つことを見越して、名前を受け取っている。
	return _scriptFile_Reaction;
}
*/

//=========================================================
///// 現在のターンの参加者の手番フレーム経過数を進める
//=========================================================
void GameManager::CurrentTurnParticipant_ProgressTurnFrameCount() {

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].id == _operationData.currentTurn ) {

			_participantsList[i].turnFrameCount += 1;

			// 一応上限をつける。が、0でループ。「直後」を表す「-1」にはしない
			if ( _participantsList[i].turnFrameCount > 9999 ) {
				_participantsList[i].turnFrameCount = 0;
			}

			break;
		}
	}
}

//=========================================================
///// 現在のターンの参加者の手番フレーム経過数をリセットする
//=========================================================
void GameManager::CurrentTurnParticipant_ResetTurnFrameCount() {

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].id == _operationData.currentTurn ) {

			_participantsList[i].turnFrameCount = -1;

			break;
		}
	}
}
//=========================================================
///// 現在のターンの参加者の手番フレーム経過数を「開始直後」ではないものにする
//=========================================================
void GameManager::CurrentTurnParticipant_TurnFrameCountStart() {

	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {

		if ( _participantsList[i].id == _operationData.currentTurn ) {

			_participantsList[i].turnFrameCount = 0;

			break;
		}
	}
}


//=========================================================
// 駒台に駒を乗せる
//=========================================================
void GameManager::AddPieceToPieceBox ( PieceBox *pieceBox, int pieceNum, int color ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 駒台に駒を乗せる
	for ( int i = 0; i < pieceNum; ++i ) {

		Piece *piece = new Piece( color );
		_pieceList_forManagement.push_back(piece);

		pieceBox->AddPiece ( piece );
	}

}

//=========================================================
// 更新
//=========================================================
void GameManager::Update ( sf::RenderWindow &window ) {

	if ( !_isActive ) {
		return;
	}
	_isRendered = true;

	// まっさきにアクターを表示しまして、
	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {
		if (_participantsList[i].actor != NULL) {
			_participantsList[i].actor->_gameManager = this;
			_participantsList[i].actor->Update ( window );
		}
	}

	// 試合中かつアクターがリアクション中ならブロックしたい要素
	bool isBattlingAndActorReacting = false;
	GUI_SFML::TextMessageWindow *tw = ScriptEngine::textMessageWindowParaTi;
	if ( _battleInfo.isBattlingNow && !tw->IsAllFinished() ) {
		isBattlingAndActorReacting = true;
	}

	if ( !isBattlingAndActorReacting ) {

		_board->Update( window, sf::Vector2f( _boardX, _boardY ) );

		_pieceBox_P1->Update(window, sf::Vector2f( _pieceBoxP1X, _pieceBoxP1Y ));
		_cardBox_P1 ->Update(window, sf::Vector2f( _carBoxP1X, _cardBoxP1Y ));

		_pieceBox_P2->Update(window, sf::Vector2f( _pieceBoxP2X, _pieceBoxP2Y ));
		_cardBox_P2 ->Update(window, sf::Vector2f( _carBoxP2X, _cardBoxP2Y ));

		if ( Is_AI_Participating_Now() ) {

			//_AI_ProgressBar->SetPosition ( _AI_ProgresBarX, _AI_ProgresBarY );

			string side = GetCurrentTurnParticipant().side;
			if ( side == "P1") {
				_AI_ProgressBar->SetPositionX( _AI_ProgresBarX_P1 );
				_AI_ProgressBar->SetPositionY( _AI_ProgresBarY_P1 );
			}
			else if ( side == "P2") {
				_AI_ProgressBar->SetPositionX( _AI_ProgresBarX_P2 );
				_AI_ProgressBar->SetPositionY( _AI_ProgresBarY_P2 );
			}

			_AI_ProgressBar->Update ( window );
		}
	}

	//// アクター。試合中のみ動作させる。もちろん表示も。
	//if ( _battleInfo.isBattlingNow && _actor != NULL ) {

	// アクター。インスタンスがあれば更新する。あとは内部で活性になっているかどうか。
	//if ( _actor != NULL ) {
	//	_actor-> Update ( window );
	//}

	// 試合中でなくても表示されるが、いいのか？
	_psychoGage->Update(window);


	if ( _select != NULL ) {
		_select->Update(window);
	}

	///// ルールに沿った操作を受け付ける
	//Operation_With_Rule (window);

	// 棋譜には描画機能があるのでした。
	if (_score != NULL) {
		_score->Update(window);
	}

	// 口出しモードの試合中ならば、状況パネルを表示する
	if ( IsInterferenceBattle() && _battleInfo.isBattlingNow) {

		if ( _interfereModePanel != NULL ) {

			_interfereModePanel->SetPosition ( _interfereModePanelX, _interfereModePanelY );
			_interfereModePanel->Update(window);
		}
	}

	// 現在の手番パネル：試合中、または結果が出ている（試合後）場合に表示
	if ( _battleInfo.isBattlingNow || _battleInfo.result != "") {

		_currentTurnNamePanel->Update(window);
	}
	
	
	if ( Application::GetInstance()->_releasedKeyCode == sf::Keyboard::C ) {
        
        string board = BoardToString();
        
        printf ( "\n\n%s\n\n", board.c_str() ); 
	}
	
	// 距離比較思考処理の動作テスト
	/*
	if ( Application::GetInstance()->_releasedKeyCode == sf::Keyboard::P ) {
        
		Grid *mouseOverGrid = _board->GetMouseOverGrid();

		if ( mouseOverGrid != NULL ) {

			int x =  _board->GetMouseOverPosition().x;
			int y =  _board->GetMouseOverPosition().y;

			AI_Modules::InnerExpression *ieBlackTurn = GenerateInnerExpressionFromCurrentState ( AI_Modules::TURN::BLACK );
			AI_Modules::InnerExpression *ieWhiteTurn = GenerateInnerExpressionFromCurrentState ( AI_Modules::TURN::WHITE );

			AI_Modules::AI *aiTemp = NULL;
			
			for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {
				
				Participant p = _participantsList[i];
				if ( p.side == "P2" && p.ai != NULL ) {
					aiTemp = p.ai;
				}
				break;
			}
			
			bool isAICreated = false;
			if ( aiTemp == NULL ) {
				aiTemp = new AI_Modules::AI();
				isAICreated = true;
			
				AI_Modules::Parameters_PieceEvaluation settings = AI_Modules::AI::PEE_Read_File ( "./data/AI/PEE_Level1.txt" );

				aiTemp->PEE_Init(settings);
			}

			AI_Modules::AI::PositionValueData resultBlaskTurn = aiTemp->Evaluate_Position ( ieBlackTurn, AI_Modules::Position(x, y) );
			AI_Modules::AI::PositionValueData resultWhiteTurn = aiTemp->Evaluate_Position ( ieWhiteTurn, AI_Modules::Position(x, y) );
			
			int distanceToGoal_black = aiTemp->GetDistanceToGoal ( ieBlackTurn, AI_Modules::TURN::BLACK, x, y, DIRECTION::NONE, 0, 0 );
			int distanceToGoal_white = aiTemp->GetDistanceToGoal ( ieWhiteTurn, AI_Modules::TURN::WHITE, x, y, DIRECTION::NONE, 0, 0 );
			
			printf ( "Grid(%d, %d) Black : Advantage[%d], Risk[%d]\n", x, y, resultBlaskTurn.advantage, resultBlaskTurn.risk ); 
			printf ( "Grid(%d, %d) White : Advantage[%d], Risk[%d]\n", x, y, resultWhiteTurn.advantage, resultWhiteTurn.risk );
			
			printf ( "DistanceToGoal : Black[%d]  White[%d]\n", distanceToGoal_black, distanceToGoal_white );
			
			printf ( "\n" );
			
			delete ieBlackTurn;
			delete ieWhiteTurn;
			
			if ( isAICreated ) {
				delete aiTemp;
			}
        }
	}
	*/

}


//=========================================================
// 現局面の簡易スクリプト化
//=========================================================
string GameManager::BoardToString() {
    
    string result = "";
    
    result += "<Put game=\"\">\n";
    
    ///// カード置き場 P1
    vector<Card*> cardListP1 = _cardBox_P1->GetAllCards();
    for ( unsigned int i = 0; i < cardListP1.size(); ++i ) {
        
        Card *card = cardListP1[i];
        
        string cardName = card->GetProperty().cardName;
        int direction = card->GetCurrentDirection();

        string side = "p1";
        
        string directionStr = "";
        if ( direction == DIRECTION::UP    ) { directionStr = "UP";}
        if ( direction == DIRECTION::DOWN  ) { directionStr = "DOWN";}
        if ( direction == DIRECTION::LEFT  ) { directionStr = "LEFT";}
        if ( direction == DIRECTION::RIGHT ) { directionStr = "RIGHT";}
        
        result += "\t<Card name=\"" + cardName + "\" direction=\"" + directionStr + "\" box=\"" + side + "\" />\n";
    }
    
    result += "\t\n";

    ///// カード置き場 P2
    vector<Card*> cardListP2 = _cardBox_P2->GetAllCards();
    for ( unsigned int i = 0; i < cardListP2.size(); ++i ) {
        
        Card *card = cardListP2[i];
        
        string cardName = card->GetProperty().cardName;
        int direction = card->GetCurrentDirection();

        string side = "p2";
        
        string directionStr = "";
        if ( direction == DIRECTION::UP    ) { directionStr = "UP";}
        if ( direction == DIRECTION::DOWN  ) { directionStr = "DOWN";}
        if ( direction == DIRECTION::LEFT  ) { directionStr = "LEFT";}
        if ( direction == DIRECTION::RIGHT ) { directionStr = "RIGHT";}
        
        result += "\t<Card name=\"" + cardName + "\" direction=\"" + directionStr + "\" box=\"" + side + "\" />\n";
    }
    
    result += "\t\n";
    
    ///// 駒台 P1
    vector<Piece*> pieceListP1 = _pieceBox_P1->GetPieceList();
    for ( unsigned int i = 0; i < pieceListP1.size(); ++i ) {
        
        Piece *piece = pieceListP1[i];
        
        if ( piece->GetColor() == AI_Modules::TURN::BLACK) {
            result += "\t<Piece color=\"BLACK\" box=\"p1\" />\n";
        }

        if ( piece->GetColor() == AI_Modules::TURN::WHITE) {
            result += "\t<Piece color=\"WHITE\" box=\"p1\" />\n";
        }
    }

    result += "\t\n";
   
    ///// 駒台 P2
    vector<Piece*> pieceListP2 = _pieceBox_P2->GetPieceList();
    for ( unsigned int i = 0; i < pieceListP2.size(); ++i ) {
        
        Piece *piece = pieceListP2[i];
        
        if ( piece->GetColor() == AI_Modules::TURN::BLACK) {
            result += "\t<Piece color=\"BLACK\" box=\"p2\" />\n";
        }

        if ( piece->GetColor() == AI_Modules::TURN::WHITE) {
            result += "\t<Piece color=\"WHITE\" box=\"p2\" />\n";
        }
    }

    result += "\t\n";
    
    
    int maxY = _board->GetHeight();
    int maxX = _board->GetWidth();
    
    for ( unsigned int y = 0; y < maxY; ++y ) {
        for ( unsigned int x = 0; x < maxX; ++x ) {
            
            
            Grid *grid = _board->GetGrid(x,y);
            
            Piece *piece = grid->GetPiece();
            Card  *card  = grid->GetCard();
            
            
            if (piece != NULL) {
                
                string pieceColor = "";
                
                if ( piece->GetColor() == AI_Modules::TURN::BLACK) { pieceColor = "BLACK"; }
                if ( piece->GetColor() == AI_Modules::TURN::WHITE) { pieceColor = "WHITE"; }
                
                result += "\t<Piece color=\"" + pieceColor +  "\" x=\"" + to_string(x) + "\" y=\"" + to_string(y) + "\" />\n";
                
            }
            
            
            if ( card != NULL ) {
                
                string cardName = card->GetProperty().cardName;
                int direction = card->GetCurrentDirection();

                string directionStr = "";
                if ( direction == DIRECTION::UP    ) { directionStr = "UP";}
                if ( direction == DIRECTION::DOWN  ) { directionStr = "DOWN";}
                if ( direction == DIRECTION::LEFT  ) { directionStr = "LEFT";}
                if ( direction == DIRECTION::RIGHT ) { directionStr = "RIGHT";}    

                result += "\t<Card name=\"" + cardName + "\" direction=\"" + directionStr + "\" x=\"" + to_string(x) + "\" y=\"" + to_string(y) + "\" />\n";
                
            }
            
        }
    }
    
    
    
    result += "</Put>";
        
    return result;
    
}




//=========================================================
// 今、AIの番？
//=========================================================
bool GameManager::Is_AI_Turn_Now() {

	/*
		試合中じゃない場合も、false
	*/

	bool isAI_Turn = false;
	if ( _battleInfo.isBattlingNow ) {

		// 現在の手番の参加者がAIを持っているなら、AIの番、という判定。
		for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

			if ( _participantsList[i].id == _operationData.currentTurn ) {

				if ( _participantsList[i].ai != NULL ) {
					isAI_Turn = true;
				}

				break;
			}
		}

		/*

		if ( _operationData.currentTurn == "FIRST") {
			if (_operationData.first == "P1") {
				if ( _operationData.operatorP1 == "AI" ) {
					isAI_Turn = true;
				}
			}
			else if (_operationData.first == "P2") {
				if ( _operationData.operatorP2 == "AI" ) {
					isAI_Turn = true;
				}
			}
		}
		else if ( _operationData.currentTurn == "SECOND") {
			if (_operationData.second == "P1") {
				if ( _operationData.operatorP1 == "AI" ) {
					isAI_Turn = true;
				}
			}
			else if (_operationData.second == "P2") {
				if ( _operationData.operatorP2 == "AI" ) {
					isAI_Turn = true;
				}
			}
		}
		*/
	}

	return isAI_Turn;
}

//=========================================================
// 試合中で、どっちかがAI？
//=========================================================
bool GameManager::Is_AI_Participating_Now() {

	bool result = false;

	if ( _battleInfo.isBattlingNow ) {

		//if ( _operationData.operatorP1 == "AI" || _operationData.operatorP2 == "AI" ) {
		//	result = true;
		//}

		// 現在の手番の参加者がAIを持っているなら、AIの番、という判定。
		for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

			if ( _participantsList[i].id == _operationData.currentTurn ) {

				if ( _participantsList[i].ai != NULL ) {
					result = true;
				}

				break;
			}
		}
	}

	return result;
}

//=========================================================
// ズバリ、カードをつかんでいる？：つまり、右クリックがメニュー表示とカード回転でかぶるので、一時的にメニュー表示を抑制する意図
//=========================================================
bool GameManager::IsHoldingCard() {

	bool result = false;

	result = Application::GetInstance()->IsMouseClingObj_Card();

	return result;
}


//=========================================================
// 更新：操作だけしたいとき。結果的にチュートリアル専用か？
//=========================================================
void GameManager::Update_Operation () {

	if ( !_isActive ) {
		return;
	}

	///// ルールに沿った操作を受け付ける
	Operation_With_Rule ();

}

//=========================================================
///// 棋譜を読み込んでください。
//=========================================================
void GameManager::ReadScoreSheetFile ( string fileName ) {

	printf("棋譜を読み込んでください。%s\n", fileName.c_str());

	///// 既存の棋譜はリセットされる。
	if ( _score != NULL ) {

		// TODO kifuViewerPosition = _score->GetPosition(); ってやって位置を引き継ごうかと思ったけど、GetPosition がないぃぃぃ。ないなら別にいい？
		delete _score;
	}

	///// 棋譜ファイルを指定して生成
	_score = new GameRecord ( fileName );

	_score->SetPosition ( _kifuViewerPosition );

	// 気を利かせて表示フラグを立てておく。マニュアル操作に徹するべきか？
	_score->Show();

	// 読み込んだ棋譜の初期局面にするよ
	GameRecord::Record record = _score->GetCurrentRecord();
	RearrangeAllPiecesAndCards ( record.innerExpression );
	
	
	// 試合のタイトルを棋譜クラスから取得して画面表示できるようにラベルを作成
	string battleTitle = _score->GetBattleTitle();

	// こっち側でも試合タイトルを確保しておくよ
	_battleInfo.battleTitle = _score->GetBattleTitle();
	
	// ラベル作成
	SetupBattleTitleForKifuViewing ( battleTitle );
	
}


//=========================================================
// 棋譜再生モード用の試合タイトルラベルをセット（セーブ/ロード時にGamaManager側で再現できるよう別メソッドにしておく）
//=========================================================
void GameManager::SetupBattleTitleForKifuViewing ( string battleTitle ) {
	
	_battleTitleForKifuViewing = TextImageFactory::Create ( battleTitle, 22 );
	_battleTitleForKifuViewing.setOutlineThickness(1);
	_battleTitleForKifuViewing.setOutlineColor (sf::Color::Black);
	_battleTitleForKifuViewing.setFillColor    (sf::Color::White);
	
	sf::Vector2f battleTitleOrigin;
	battleTitleOrigin.x = _battleTitleForKifuViewing.getGlobalBounds().width/2;
	battleTitleOrigin.y = _battleTitleForKifuViewing.getGlobalBounds().height/2;
	Utils_Cal::RoundWildly ( battleTitleOrigin.x );
	Utils_Cal::RoundWildly ( battleTitleOrigin.y );
	_battleTitleForKifuViewing.setOrigin ( battleTitleOrigin );
	
	_battleTitleForKifuViewing.setPosition ( 270, 480 );

	// 念には念を入れておくよ
	_battleInfo.battleTitle = battleTitle;

}

//=========================================================
///// 棋譜の現在位置を指定：検討モード用、ということだと思います。
//=========================================================
void GameManager::SetScoreIndex ( string index ) {

	if ( _score != NULL ) {

		if ( index == "END" ) {

			_score->MoveToBottom();

		} else {

			_score->SetCurrentIndex ( std::stoi(index) );
		}
	}

	// 盤面に反映させればいいのかな？
	GameRecord::Record record = _score->GetCurrentRecord();
	RearrangeAllPiecesAndCards ( record.innerExpression );
}

//=========================================================
// 検討モードの準備
//=========================================================
void GameManager::AstralProjectionStart () {

	_isAstralProjection = true;

	// フリー操作としておく
	SetFreeOperation(true);

	// 現在の盤面の状態：フリー操作でいじっても変化しない。が、棋譜を再生するときには、この状態が基点となる。
	_astralProjection_physicalState = GenerateInnerExpressionFromCurrentState();

	if ( _score == NULL ) {

		_score = new GameRecord();
		_score->SetPosition ( _kifuViewerPosition );

	}

	_score->ShowGuideButtons();

	// これは気を利かせてるというより、検討モードなのだから、棋譜はあって当然なのでありまして、ですよね？ と、上目遣いで訊ねる。猫なで声で訊ねる。
	_score->Show();
}


//=========================================================
// 検討モード後片付け
//=========================================================
void GameManager::AstralProjectionEnd () {

	_isAstralProjection = false;

	// フリー操作を解除。
	// TODO もともとフリー操作だったら？ うーんうーん
	SetFreeOperation(false);

	_score->HideGuideButtons();


	// カードをマウスでつかんだままだったら困るのです。
	if ( Application::GetInstance()->IsMouseClingObj_Card()) {

		// 外しておく
		Card *card = Application::GetInstance()->DrawMouseClingObj();

		// P1のカード置き場にでも置いておく。
		_cardBox_P1->AddCard(card);

		// 解放はしない。CardFactoryの管轄。
	}

	// 駒をマウスでつかんだままだったら困るのです。
	if ( Application::GetInstance()->IsMouseClingObj_Piece()) {

		// 外しておく
		Piece *piece = Application::GetInstance()->DrawMouseClingObj();

		// P1の駒置き場にでも置いておく。
		_pieceBox_P1->AddPiece(piece);

	}

	/*
		マウスで駒やカードをつかんだまま検討が終わった場合：

		スクリプトの側で、検討が一度終わったら盤面を解放するとは限らない。
		なんやかやとスクリプトを実行したあと、また同じ盤面で検討するかもしれない。
		したがって、マウスにつかんでいたオブジェクトは、場所はどこでもいいので、ともかく盤面に戻しておくのです。
		解放しない。行方不明にしない。
	*/

	delete _astralProjection_physicalState;
	_astralProjection_physicalState = NULL;
}

//=========================================================
// 検討モード後片付け＿強制終了
//=========================================================
void GameManager::AstralProjectionEnd_Forcebly() {

	/*
		どういう状況での強制終了かにもよる、はず。
		ぶっちゃけ、検討中にセーブデータをロードした場合用、ということ。
	*/

	// 通常の終了と同じ処理
	AstralProjectionEnd_Forcebly();

}


//=========================================================
// 更新：検討モード
//=========================================================
void GameManager::Update_AstralProjection () {
	
	///// 棋譜ビューワーの操作を担当したりもする。

	if ( _score->IsIndexChanged()) {

		printf ( "棋譜をクリックしたよー！\n" );

		// マウスに駒がくっついている？ そんなのおかしいですよ！
		ReleaseCardFromMouseIfBelongToHere();

		/* // ↓これだと、この盤面に属していないものまで削除してしまう。
		if ( Application::GetInstance()->IsMouseClingObj_Piece()) {

			Piece *piece = (Piece*)Application::GetInstance()->DrawMouseClingObj();
			ReleasePiece(piece);
		}
		*/

		// マウスにカードがくっついている？ そんなのおかしいですよ！
		ReleasePieceFromMouseIfBelongToHere();

		/* // ↓これだと、この盤面に属していないものまで削除してしまう。
		if ( Application::GetInstance()->IsMouseClingObj_Card()) {

			Card *card = (Card*)Application::GetInstance()->DrawMouseClingObj();
			CardFactory::Release(card);
		}
		*/

		GameRecord::Record record = _score->GetCurrentRecord();

		printf ("指し手：%s\n", _score->RecordToString(record).c_str());


		AI_Modules::InnerExpression *ie = record.innerExpression;

		// 「次へ」をクリックしたのだった？
		if ( _score->IsNextPushed() ) {

			printf ( "「次へ」をクリックしたよ！\n" );

			AI_Modules::InnerExpression *ie_astral  = GenerateInnerExpressionFromCurrentState();

			if ( ! IsEqual_EasyVersion ( ie_astral, _astralProjection_physicalState ) ) {

				printf("あ〜、ズレてますね〜。\n");

				// このとき、盤面を修正しているにも関わらず、「着手」がワープになってしまうのはなぜかというと、
				// RearrangeAllPiecesAndCards の中でびよーん省略フラグを立てていて、
				// Updateの前に Move を実行しているからです。ね。多分。
				// つまり、バグではない、ということにする。
				// 説明ができると悪は悪ではなくなってしまうのですね？ 悪とは説明のつかないもののこと、ということですか。
				// しかし、そのゆえにこの世には嘘の言葉がはびこるのではないのか？
				RearrangeAllPiecesAndCards ( _astralProjection_physicalState );
			}

			// エフェクトを使わずに着手
			Move(record.move, false);

		} else {

			RearrangeAllPiecesAndCards ( ie );
		}


		_astralProjection_physicalState = GenerateInnerExpressionFromCurrentState();
	}


}

//=========================================================
// 検討モードでの再生用。盤面が一致しているかどうか。
// つまり、「次へ」ボタンで駒などを動かすとき、フリー操作で盤面がいじられた状態かどうかを判定したい。
// InnerExpression の IsEqual() ほど厳密に同一性を調査する必要はない
//=========================================================
bool GameManager::IsEqual_EasyVersion ( AI_Modules::InnerExpression *ie1, AI_Modules::InnerExpression *ie2 ) {

	/*
		InnerExpression.h から、関係しそうな箇所をコピペ。資料として。

		///// 盤面情報
		int         _gridList_Piece[15][15];         // 駒レイヤー      ：黒/白/なし GRID
		CardOnBoard _gridList_Card[15][15];          // カードレイヤー  ：カードと向き
		int _boardSizeX;
		int _boardSizeY;

		///// 駒台情報：何個あるか、だけ。
		int _handPiece_BLACK;
		int _handPiece_WHITE;

		///// 手札情報：カード番号のリスト
		int _handCards_BLACK[100];
		int _handCards_WHITE[100];
		int _handCards_BLACK_Index;
		int _handCards_WHITE_Index;
	*/

	bool result = true;

	if ( ie1->_handPiece_BLACK != ie2->_handPiece_BLACK) {
		printf("不一致：手駒数・黒 [%d, %d]\n", ie1->_handPiece_BLACK, ie2->_handPiece_BLACK);
		return false;
	}
	if ( ie1->_handPiece_WHITE != ie2->_handPiece_WHITE) {
		printf("不一致：手駒数・白 [%d, %d]\n", ie1->_handPiece_WHITE, ie2->_handPiece_WHITE);
		return false;
	}

	if ( ie1->_handCards_BLACK_Index  != ie2->_handCards_BLACK_Index) {
		printf("不一致：手札数・黒 [%d, %d]\n", ie1->_handCards_BLACK_Index, ie2->_handCards_BLACK_Index);
		return false;
	}
	if ( ie1->_handCards_WHITE_Index  != ie2->_handCards_WHITE_Index) {
		printf("不一致：手札数・白 [%d, %d]\n", ie1->_handCards_WHITE_Index, ie2->_handCards_WHITE_Index);
		return false;
	}

	if ( memcmp(ie1->_gridList_Card         , ie2->_gridList_Card         , sizeof(ie2->_gridList_Card         )) != 0) {
		printf("不一致：盤上のカード\n");
		return false;
	}
	if ( memcmp(ie1->_gridList_Piece        , ie2->_gridList_Piece        , sizeof(ie2->_gridList_Piece        )) != 0) {
		printf("不一致：盤上の駒\n");
		return false;
	}

	///// 手札置き場を比べるには…：ソートされてしまいますけど、構いませんよね？
	qsort ( ie1->_handCards_BLACK, ie1->_handCards_BLACK_Index, sizeof(int), AI_Modules::InnerExpression::compare_int);
	qsort ( ie2->_handCards_BLACK, ie2->_handCards_BLACK_Index, sizeof(int), AI_Modules::InnerExpression::compare_int);
	if ( memcmp(ie1->_handCards_BLACK       , ie2->_handCards_BLACK       , sizeof(ie2->_handCards_BLACK       )) != 0) {
		printf("不一致：カード置き場・黒\n");
		return false;
	}

	qsort ( ie1->_handCards_WHITE, ie1->_handCards_WHITE_Index, sizeof(int), AI_Modules::InnerExpression::compare_int);
	qsort ( ie2->_handCards_WHITE, ie2->_handCards_WHITE_Index, sizeof(int), AI_Modules::InnerExpression::compare_int);
	if ( memcmp(ie1->_handCards_WHITE       , ie2->_handCards_WHITE       , sizeof(ie2->_handCards_WHITE       )) != 0) {
		printf("不一致：カード置き場・白\n");
		return false;
	}


	return result;
}

//=========================================================
// 更新：試合
//=========================================================
void GameManager::Update_Battle () {

	/*

		さて、何をすればいいのかな？

		勝敗を監視するとか？→ 操作後の事後処理内で、試合中なら、勝敗判定を呼ぶ。

		操作を受け付けるのは、勝負がついてないときだけ、か。
	*/


	if ( !_isActive ) {
		return;
	}

	// 試合中で、テキストが表示され切ってない＝アクターがしゃべっている、ということのはず。
	// つまりアクターのリアクション中の待機ということ。
	// GameManagerのUpdate自体も停止させたいところでありまして、それはScriptEngine側で対応している。
	// → ScriptEngine::GameManager_Update()
	GUI_SFML::TextMessageWindow *textWindow = ScriptEngine::textMessageWindowParaTi;
	if ( _battleInfo.isBattlingNow && !textWindow->IsAllFinished() ) {
		return;
	}

	///// まだ勝敗が決まってないなら、ルールに沿った操作を受け付ける
	if ( _battleInfo.result == "" ) {

		Operation_With_Rule ();

		/*
			着手はエフェクトを介して行われるので、
			Operation_With_Rule () の直後のこの位置では、試合結果は、得られない。

			_battleInfo.isBattlingNow = false; は、
			ScriptEngine の側で BattleEnd()を呼んで、その中で行うこととする。

			///// 勝敗が決まったら、試合終了
			if ( _battleInfo.result != "" ) {
				_battleInfo.isBattlingNow = false;
			}
		*/

		///// 思考のプログレスバー
		if ( Is_AI_Turn_Now() ) {
			_AI_ProgressBar->SetValue ( _AI_Manager->GetThinkProgress().progressRate );
		}
	}
}


//=========================================================
// 試合開始の準備
//=========================================================
void GameManager::BattleStart() {

	///// 試合情報を初期化
	_battleInfo.isBattlingNow = true;
	_battleInfo.result        = "";

	///// 棋譜を初期化：開始局面を登録（）
	AI_Modules::InnerExpression *firstState = GenerateInnerExpressionFromCurrentState();
	delete _score;
	_score = new GameRecord (firstState);
	_score->SetPosition ( sf::Vector2f( 515.0f, 410.0f ) );

	///// アクターを どっち側・どの位置 に表示するか？
	/*
		この時点でアクター自体は生成およびセットアップされている前提。
		アクターの設定は「ゲーム盤定義タグ」のAI設定の項目で行っている。
	*/
	/*
	if ( _actor != NULL ) {

		//	反転について(_actor->SetReverse)
		//	P1だから or P2だから、という理由での反転は、しない。
		//
		//	というか結局、反転は、しない。
		//	対応する向きの画像を、画像として、用意して設定する。

		// P1 が AI
		if ( _operationData.operatorP1 == "AI") {

			//_actor->SetReverse ( false ); // 反転しない
			_actor->SetPosition ( sf::Vector2f(
				APPLICATION::WINDOW_SIZE::WIDTH - _actorPosisionX, //APPLICATION::WINDOW_SIZE::WIDTH - 100,
				_actorPosisionY //APPLICATION::WINDOW_SIZE::HEIGHT / 2
			) );
		}
		// P2 が AI
		else if ( _operationData.operatorP2 == "AI") {

			//_actor->SetReverse ( true ); // 反転しない
			_actor->SetPosition ( sf::Vector2f(
				_actorPosisionX, //100,
				_actorPosisionY //APPLICATION::WINDOW_SIZE::HEIGHT / 2
			) );
		}
	}
	*/
}
//=========================================================
// 試合終了処理
//=========================================================
void GameManager::BattleEnd() {

	///// 試合中ではない、のフラグを立てる。
	_battleInfo.isBattlingNow = false;

	/*
		試合結果 _battleInfo.result には、何らかの値が入っているはず。
		これをクリアすると、まるで試合がまだ終わっていない、みたいになってしまう。

		Update_Battle() の中では、下記のようなif判定を行っている

		// まだ試合結果が出ていないならば
		if (_battleInfo.result == "") {
			...
		}
	*/

	//_score->OutputScoreSheet ( "./record/scoreSheetTestOut.txt" );

    time_t t = time(NULL);
    string timeStamp = Utils_Convert::Long_to_String(t);
	
    string battleTitle = "";
	string afterAliasApplied = _battleInfo.battleTitle;

    if ( _battleInfo.battleTitle == "" ) {
		battleTitle = "";
    } else {
		
		TextMessageWindow::ReplaceWithAlias ( afterAliasApplied );
		
		battleTitle = "(" + afterAliasApplied + ")";
    }
    
	// 試合のタイトルをファイル名に入れると、Windows環境では文字化けしてしまう。コノヤロ～。
	//string scoreSheetFileName = "./record/kifu_" + timeStamp + battleTitle + ".txt";
	
	string scoreSheetFileName = "./record/kifu_" + timeStamp + ".mkf";

	_score->OutputScoreSheet ( scoreSheetFileName, t,  afterAliasApplied );
	

#if defined(_WIN32)
	
	// 棋譜ファイルが増えた場合にフォルダにまとめる処理
	ArchiveRecordFiles ();

#elif defined(__linux__)


#endif

	
	// 選択肢を片付ける
	delete _select;
	_select = NULL;

}
//=========================================================
// 試合終了処理＿強制終了
//=========================================================
void GameManager::BattleEnd_Forcebly() {

	///// 試合中ではない、のフラグを立てる。
	_battleInfo.isBattlingNow = false;


	/*
		棋譜に出力しないこと以外は通常の終了処理と同じ
	*/
}


#if defined(_WIN32)

//=========================================================
// 棋譜ファイルが増えた場合にフォルダにまとめる処理
//=========================================================
void GameManager::ArchiveRecordFiles () {
	
	int limitFilesNum = 110;  // ファイル数がこの数に達したら
	int arcFilesNum   = 100;  // 古い方のファイルをこの件数分、フォルダにまとめる
	
	//string scoreSheetFileName = "./record/kifu_" + timeStamp + ".mkf";
	
	//GetFileList ( string dirName, FolderEntity &result, vector<string> path = vector<string>() );

	Utils_File::FolderEntity recordDir;
	
	Utils_File::GetFileList ( "record", recordDir );
	
	vector< pair<time_t, Utils_File::FolderEntity> > recordFileList;
	
	for ( unsigned int i = 0; i < recordDir.childList.size(); ++i) {
		
		Utils_File::FolderEntity c = recordDir.childList[i];
		
		if ( c.isFolder ) {
			continue;
		}
		
		// kifu_1570915920.mkf
		
		// 最低でも10文字以上のはず
		if ( c.name.size() < 10 ) {
			continue;
		}
		
		string extention = c.name.substr ( c.name.size()-4, 4);
		string prefix    = c.name.substr (0, 4);
		
		// プレフィックスと拡張子を確認
		if ( extention != ".mkf" ) { continue; }
		if ( prefix    != "kifu" ) { continue; }
		
		// UNIXタイムスタンプ部分を求める
		string unixtimestamptStr = "";
		
		vector<string> ana1 = Utils_String::Split ( c.name, "_" );
		if ( ana1.size() != 2 ) { continue; }
		
		vector<string> ana2 = Utils_String::Split ( ana1[1], "." );
		if ( ana2.size() != 2 ) { continue; }
		
		unixtimestamptStr = ana2[0];
		
		
		pair< time_t, Utils_File::FolderEntity > recordFileData;
		recordFileData.first  = Utils_Convert::String_to_Long ( unixtimestamptStr );
		recordFileData.second = c;

		recordFileList.push_back ( recordFileData );
		
		
		//printf ( "timestamp : %ld\n", recordFileData.first );
	}
	
	int recordCount = recordFileList.size();
	printf ( "mkf record count : %d\n" , recordCount );
	
	
	///// 多くなってないなら何もせずに終了
	if ( recordCount < limitFilesNum ) { return; }
	
	
	///// 日付順に並び替え：リストの後ろの方が数字が大きい＝新しい、ということにする
	bool isSwitched = true;
	while ( isSwitched ) {
		
		isSwitched = false;
		
		for ( unsigned int i = 0; i < recordFileList.size() - 1; ++i ) {
			
			pair< time_t, Utils_File::FolderEntity > current = recordFileList[i];
			pair< time_t, Utils_File::FolderEntity > next    = recordFileList[i + 1];
			
			if ( current.first > next.first ) {
				
				recordFileList [i]     = next;
				recordFileList [i + 1] = current;
				
				isSwitched = true;
				
				//break;
			}
		}
		
	}
	
	// long max 2147483647
	
	int remainingFilesNum = limitFilesNum - arcFilesNum;
	
	time_t indexA = recordFileList[0].first;
	time_t indexB = recordFileList[ recordFileList.size() - (remainingFilesNum + 1) ].first;
	
	printf ( "%ld to %ld\n", indexA, indexB );
	
	map<string, int> timeInfoA = Utils_Convert::Long_to_TimeInfo (indexA);
	string yearA  = Utils_Convert::Int_to_String ( timeInfoA["year"] );
	string monthA = Utils_Convert::Int_to_String ( timeInfoA["month"] );
	string dayA   = Utils_Convert::Int_to_String ( timeInfoA["day"] );
	monthA = ( monthA.size() == 1 ) ? "0" + monthA : monthA;
	dayA   = ( dayA.size()   == 1 ) ? "0" + dayA : dayA;
	
	map<string, int> timeInfoB = Utils_Convert::Long_to_TimeInfo (indexB);
	string yearB  = Utils_Convert::Int_to_String ( timeInfoB["year"] );
	string monthB = Utils_Convert::Int_to_String ( timeInfoB["month"] );
	string dayB   = Utils_Convert::Int_to_String ( timeInfoB["day"] );
	monthB = ( monthB.size() == 1 ) ? "0" + monthB : monthB;
	dayB   = ( dayB.size()   == 1 ) ? "0" + dayB : dayB;
	
	// フォルダを作成するよ
	string newDirName = "./record/arc_" + yearA + monthA + dayA + "_" + yearB + monthB + dayB;
	
	printf ( "new dir name : %s\n", newDirName.c_str() );
	
	CreateDirectory (
		newDirName.c_str(), // ディレクトリ名
		NULL // セキュリティ識別子
	);
	
	// ファイルを移動させるよ
	int movedCount = 0;
	for ( unsigned int i = 0; i <= recordFileList.size() - (remainingFilesNum + 1); ++i ) {
		
		Utils_File::FolderEntity f = recordFileList[i].second;
		
		string fileNameFrom = "./record/" + f.name;
		string fileNameTo   = newDirName + "/" + f.name;
		
		MoveFile (
			fileNameFrom.c_str(),  // ファイル名
			fileNameTo.c_str()     // 新しいファイル名
		);
		
		++movedCount;
		
		//printf ( "%s -> %s\n", fileNameFrom.c_str(), fileNameTo.c_str() );
	}
	
	printf ( "moved : %d\n", movedCount );
	
}

#elif defined(__linux__)


#endif

//=========================================================
// 勝敗判定：着手のたびに呼び出す。
//=========================================================
void GameManager::Judgement() {

	/*
		引き分け判定をどうにかしましょう。

		つまり千日手ってことですよね。棋譜を見て？ どうのこうの。
	*/
	
	///// 現在の盤面を見る。
	AI_Modules::InnerExpression *ie = GenerateInnerExpressionFromCurrentState();

	// 色で、勝敗を判定
	/*
		これは、P1が黒で、P2が白、という前提。
		P1が白駒を持っているかもしれない。課題。
	*/
	bool isBlackWin = IsWin ( ie, AI_Modules::GRID::BLACK );
	bool isWhiteWin = IsWin ( ie, AI_Modules::GRID::WHITE );


	if ( isBlackWin ) {
		_battleInfo.result = "P1Win";
		_battleInfo.resultDescription = "P1の勝ち";

	}
	else if ( isWhiteWin ) {
		_battleInfo.result = "P2Win";
		_battleInfo.resultDescription = "P2の勝ち";
	}
	
	///// 合法手がない場合は負けという判定
	vector<AI_Modules::Move> legalMoveList;

	// 今、P1/P2どっちの手番なのか？（この勝敗判定処理には手番交代直後にきているはず。そんなこと気にしなきゃならない実装なの？ っぽい）
	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	if ( currentTurnParticipant.side == "P1") {
	
		ie->_currentTurn = AI_Modules::TURN::BLACK;
	
	}
	else if ( currentTurnParticipant.side == "P2") {

		ie->_currentTurn = AI_Modules::TURN::WHITE;

	}
	
	// 合法手リストを求める
	ie->GetLegalNextMoveList(legalMoveList);
	printf ("合法手数[%d]\n", legalMoveList.size());
	
	
	// パス以外の手があるか？
	bool IsMoveExistOtherThanPASS = false;
	for ( unsigned int i = 0; i < legalMoveList.size(); ++i ) {
		
		AI_Modules::Move m = legalMoveList[i];
		if ( m.action != AI_Modules::MOVE::ACTION::PASS ) {
		
			IsMoveExistOtherThanPASS = true;
			break;
		}
	}
	
	if ( !IsMoveExistOtherThanPASS ) {
		
		printf ("（パス以外の）合法手がない\n");
				
		if ( currentTurnParticipant.side == "P1") {
			printf ("P1の手番なので、P2の勝ち\n");
			_battleInfo.result = "P2Win";
			_battleInfo.resultDescription = "P2の勝ち（P1の合法手がなくなった）";
		}
		else if ( currentTurnParticipant.side == "P2") {
			printf ("P2の手番なので、P1の勝ち\n");
			_battleInfo.result = "P1Win";
			_battleInfo.resultDescription = "P1の勝ち（P2の合法手がなくなった）";
		}
	}
	
	// AIが投了しているかもしれない
	for ( unsigned int i = 0; i < _participantsList.size(); ++i) {
		
		Participant p = _participantsList[i];
		
		if ( p.ai != NULL && p.ai->_isResign ) {
			
			printf ( "AI Resigned.\n" );
			
			if ( p.side == "P1") {
				printf ("P1の手番なので、P2の勝ち\n");
				_battleInfo.result = "P2Win";
				_battleInfo.resultDescription = "P2の勝ち（P1投了）";
			}
			else if ( p.side == "P2") {
				printf ("P2の手番なので、P1の勝ち\n");
				_battleInfo.result = "P1Win";
				_battleInfo.resultDescription = "P1の勝ち（P2投了）";
			}
			
			break;
		}
	}
	

	/*
		その他、アクターが投了していたら、そのアクターの所属しているサイドの負け、としましょうか。
		それだとアクターのないプレイヤーはどうしましょう？
	*/


	printf ( "勝敗判定（GameManager::Judgement） : %s\n", _battleInfo.result.c_str() );

}

//=========================================================
// この色、勝ってる？
//=========================================================
bool GameManager::IsWin ( AI_Modules::InnerExpression *ie, int color ) {

	bool result = false;

	int opponentAreaY = 0;
	int myPiece       = color;

	///// どっちの手番として見ますか？
	if ( color == AI_Modules::GRID::BLACK ) {
		opponentAreaY = 0;
	}
	else if ( color == AI_Modules::GRID::WHITE ) {
		opponentAreaY = ie->_boardSizeY - 1;

	// え？
	} else {
		return false;
	}


	/////  一つでも相手陣に自分の駒があれば、勝ち。
	for (int x = 0; x < ie->_boardSizeX; ++x) {

		if ( ie->_gridList_Piece[opponentAreaY][x] == myPiece) {
			result = true;
			break;
		}
	}


	return result;
}


//=========================================================
// 描画
//=========================================================
void GameManager::Draw ( sf::RenderWindow &window ) {

	if ( !_isActive ) {
		return;
	}

	_pieceBox_P1->Draw(window);
	_cardBox_P1->Draw(window);
	_pieceBox_P2->Draw(window);
	_cardBox_P2->Draw(window);

	_board->Draw(window);

	if ( Is_AI_Participating_Now() ) {
		_AI_ProgressBar->Draw ( window );
	}

	// アクター。インスタンスがあれば更新する。あとは内部で活性になっているかどうか。
	//if ( _actor != NULL ) {
	//	_actor->Draw ( window );
	//}
	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

		if ( _participantsList[i].actor != NULL ) {

			// チームに入っている場合はチーム単位での描画
			if ( UTILS_OTHERS::indexOf( _teamActorP1, _participantsList[i].actor ) == -1 && UTILS_OTHERS::indexOf( _teamActorP2, _participantsList[i].actor ) == -1) {

				_participantsList[i].actor->Draw ( window );
			}
		}
	}

	// チーム単位での描画
	for ( unsigned int i = 0; i < _teamActorP1.size(); ++i ) { _teamActorP1[i]->Draw ( window ); }
	for ( unsigned int i = 0; i < _teamActorP2.size(); ++i ) { _teamActorP2[i]->Draw ( window ); }


	// 試合中でなくても表示されるが、いいのか？
	_psychoGage->Draw(window);

	// 現在の手番：試合中、または結果が出ている（試合後）場合に表示
	if ( _battleInfo.isBattlingNow || _battleInfo.result != "") {
		_currentTurnNamePanel->Draw(window);
	}

	// 指し手の仮表示設定があれば表示
	if ( _moveForDemonstration.action != 0 ) {
		DrawMoveDemonstration ( window );
	}
	
	
	// 最後の指し手を強調表示：条件式の意味 [ 試合中のみ表示。ただし、決着した後も表示が消えないで欲しい ]
	if ( _score != NULL && (_battleInfo.isBattlingNow || _battleInfo.result != "" )  ) {
		
		vector<GameRecord::Record> recordList = _score->GetRecordList();
		if ( recordList.size() > 0 ) {
			
			vector < pair<int, int> > positionList;
			
			Card  *takenCard     = NULL;
			Piece *collidedPiece = NULL;
			
			GameRecord::Record lastRecord = recordList [ recordList.size() - 1 ];
			
			if ( lastRecord.move.action == AI_Modules::MOVE::ACTION::PUT_PIECE) {
				
				pair<int, int> position1;
				position1.first  = lastRecord.move.putPieceX;
				position1.second = lastRecord.move.putPieceY;
				positionList.push_back ( position1 );
			}
			else if ( lastRecord.move.action == AI_Modules::MOVE::ACTION::PUT_CARD) {
				
				pair<int, int> position1;
				position1.first  = lastRecord.move.putCardX;
				position1.second = lastRecord.move.putCardY;
				positionList.push_back ( position1 );
			}
			else if ( lastRecord.move.action == AI_Modules::MOVE::ACTION::MOVE_PIECE) {
				
				pair<int, int> position1;
				position1.first  = lastRecord.move.movePieceFromX;
				position1.second = lastRecord.move.movePieceFromY;
				positionList.push_back ( position1 );

				pair<int, int> position2;
				position2.first  = lastRecord.move.movePieceToX;
				position2.second = lastRecord.move.movePieceToY;
				positionList.push_back ( position2 );
				
				
				///// ぶつけられた駒を探す
				if ( lastRecord.move.movePieceAndIsCllide ) {

					// 駒ぶつけエフェクトのメモを参照。
					Piece *lastCollidedPiece = Effect_MovePiece_Collide::LAST_COLLIDED_PIECE;					

					// 最低限、現在も存在しているオブジェクトであることを確認
					// カード取得とちがって、手番からはどちらの駒台にあるかは予測できないので、両方見る
					if ( _pieceBox_P1 != NULL ) {
						
						vector<Piece*> pieceList = _pieceBox_P1->GetPieceList();
						for ( unsigned int i = 0; i < pieceList.size(); ++i ) {
							
							Piece *p = pieceList[i];
							
							if ( lastCollidedPiece ==  p ) {
								collidedPiece = p;
								break;
							}
						}
					}
					if ( _pieceBox_P2 != NULL && collidedPiece == NULL ) {
						
						vector<Piece*> pieceList = _pieceBox_P2->GetPieceList();
						for ( unsigned int i = 0; i < pieceList.size(); ++i ) {
							
							Piece *p = pieceList[i];
							
							if ( lastCollidedPiece ==  p ) {
								collidedPiece = p;
								break;
							}
						}
					}
					
				}
				
			}
			else if ( lastRecord.move.action == AI_Modules::MOVE::ACTION::TAKE_CARD) {
				
				pair<int, int> position1;
				position1.first  = lastRecord.move.takeCardX;
				position1.second = lastRecord.move.takeCardY;
				positionList.push_back ( position1 );
				
				// カード取得エフェクトのメモを参照。
				Card *lastTakenCard = Effect_TakeCard::LAST_TAKEN_CARD;
				
				// 最低限、現在も存在しているオブジェクトであることを確認
				// ついでに、カード置き場から選んでマウスでつかみ上げると強調表示が消えるという（望ましいかもしれない）効果もある。なお、実際には二度指しはありえないので、この効果を目にすることもない
				CardBox *cardBox = NULL;
				if ( lastRecord.move.turn == AI_Modules::TURN::BLACK ) { cardBox = _cardBox_P1;}
				if ( lastRecord.move.turn == AI_Modules::TURN::WHITE ) { cardBox = _cardBox_P2;}
				
				if ( cardBox != NULL ) {
					
					vector<Card*> cardList = cardBox->GetAllCards();
					for ( unsigned int i = 0; i < cardList.size(); ++i ) {
						
						Card *c = cardList[i];
						
						if ( lastTakenCard ==  c ) {
							takenCard = c;
							break;
						}
					}
				}
			}
			
			///// ボード上の該当グリッドを強調表示
			if ( _board != NULL && positionList.size() > 0) {
				
				for ( unsigned int i = 0; i < positionList.size(); ++i ) {
					
					int highLightGridX = positionList[i].first;
					int highLightGridY = positionList[i].second;
				
					Grid *grid = _board->GetGrid ( highLightGridX, highLightGridY );
					
					if ( grid != NULL ) { 
						
						grid->Draw_LastMoveHighlight(window);
						
					}
				}
			}
			
			///// 取得したカードを強調表示（の予約。アクターと重なる場合の対策）
			if ( takenCard != NULL ) {
				takenCard->DrawHighlight();
			}
			
			///// ぶつけられた駒を強調表示（の予約。アクターと重なる場合の対策）
			if ( collidedPiece != NULL ) {
				collidedPiece->DrawLastCollidedHighlight();
				
			}
			
			
		}
		
		
	}

}


//=========================================================
// 口出しモードの進行パネルを表示するかもしれない
// テキスト欄がScriptEngineの管轄で、その上（Z的に）に表示して欲しいから
//=========================================================
void GameManager::Draw_InterfereModePanel ( sf::RenderWindow &window ) {

	// 口出しモードの試合中ならば、状況パネルを表示する
	if ( IsInterferenceBattle() && _battleInfo.isBattlingNow) {

		if ( _interfereModePanel != NULL ) {
			_interfereModePanel->Draw(window);
		}
	}

	// ここで表示する選択肢とは、つまり口出しモード用の選択肢のハズ。
	if ( _select != NULL ) {
		_select->Draw(window);
	}

}


//=========================================================
// 指し手を仮表示する
//=========================================================
void GameManager::DrawMoveDemonstration ( sf::RenderWindow &window ) {

	//printf ( "指し手を仮表示\n" );


	///// 駒設置の場合
	if ( _moveForDemonstration.action == AI_Modules::MOVE::ACTION::PUT_PIECE ) {

		sf::Vector2f arrowFrom;
		sf::Vector2f arrowTo;

		Piece *piece = NULL;

		if (_moveForDemonstration.turn == 1) {

			arrowFrom = _pieceBox_P1->GetPosition();
			arrowFrom.x += _pieceBox_P1->GetBoxWidth() / 2.0f;
			arrowFrom.y += _pieceBox_P1->GetBoxHeight() / 2.0f;

			vector<Piece*> pieceList = _pieceBox_P1->GetPieceList();
			if ( pieceList.size() > 0 ) { piece = pieceList[0]; }
		}
		if (_moveForDemonstration.turn == 2) {

			arrowFrom = _pieceBox_P2->GetPosition();
			arrowFrom.x += _pieceBox_P2->GetBoxWidth() / 2.0f;
			arrowFrom.y += _pieceBox_P2->GetBoxHeight() / 2.0f;

			vector<Piece*> pieceList = _pieceBox_P2->GetPieceList();
			if ( pieceList.size() > 0 ) { piece = pieceList[0]; }
		}

		Grid *grid = _board->GetGrid ( _moveForDemonstration.putPieceX, _moveForDemonstration.putPieceY );
		arrowTo = grid->GetPosition();
		arrowTo.x += grid->GetSize().x / 2.0f;
		arrowTo.y += grid->GetSize().y / 2.0f;

		_arrowForMoveDemonstration->Init ( arrowFrom, arrowTo );
		_arrowForMoveDemonstration->Draw(window);

		if ( piece != NULL ) {
			piece->DrawDemonstration(window, arrowTo);
		}

	///// カード設置の場合
	} else if ( _moveForDemonstration.action == AI_Modules::MOVE::ACTION::PUT_CARD ) {

		sf::Vector2f arrowFrom;
		sf::Vector2f arrowTo;

		Grid *grid = _board->GetGrid ( _moveForDemonstration.putCardX, _moveForDemonstration.putCardY );
		arrowTo = grid->GetPosition();
		arrowTo.x += grid->GetSize().x / 2.0f;
		arrowTo.y += grid->GetSize().y / 2.0f;

		Card *putCard = NULL;
		CardBox *cardBoxFrom = NULL;

		if (_moveForDemonstration.turn == 1) { cardBoxFrom = _cardBox_P1;}
		if (_moveForDemonstration.turn == 2) { cardBoxFrom = _cardBox_P2;}

		if ( cardBoxFrom != NULL ) {

			vector<Card*> cardList = cardBoxFrom->GetAllCards();
			for ( unsigned int i = 0; i < cardList.size(); ++i ) {

				Card *card = cardList[i];
				if (card->GetProperty().processId == _moveForDemonstration.putCardId) {

					putCard = card;
					arrowFrom = card->GetCurrentPosition();
				}
			}
		}

		_arrowForMoveDemonstration->Init ( arrowFrom, arrowTo );
		_arrowForMoveDemonstration->Draw(window);

		int direction = putCard->GetCurrentDirection();

		putCard->DrawDemonstration ( window, grid, _moveForDemonstration.putCardDirection );

		//vector<Card::GridAffection> affectionList = putCard->GetAffectionList();
		vector<Card::GridAffection> affectionList = putCard->GetAffectionList ( _moveForDemonstration.putCardDirection );
		for ( unsigned int i = 0; i < affectionList.size(); ++i ) {

			Card::GridAffection affection = affectionList[i];

			int affectionRotatedX = affection.x; int affectionRotatedY = affection.y; int moveDRotated = affection.movableDirection;
			
			/*
			int affectionRotatedX = 0; int affectionRotatedY = 0; int moveDRotated = 0;
			if (_moveForDemonstration.putCardDirection == DIRECTION::UP) {
				affectionRotatedX = affection.x; affectionRotatedY = affection.y;

				if ( affection.movableDirection == DIRECTION::UP    ) { moveDRotated = DIRECTION::UP;}
				if ( affection.movableDirection == DIRECTION::RIGHT ) { moveDRotated = DIRECTION::RIGHT;}
				if ( affection.movableDirection == DIRECTION::DOWN  ) { moveDRotated = DIRECTION::DOWN;}
				if ( affection.movableDirection == DIRECTION::LEFT  ) { moveDRotated = DIRECTION::LEFT;}
			}
			else if (_moveForDemonstration.putCardDirection == DIRECTION::RIGHT) {
				affectionRotatedX = affection.y * -1; affectionRotatedY = affection.x;

				if ( affection.movableDirection == DIRECTION::UP    ) { moveDRotated = DIRECTION::RIGHT;}
				if ( affection.movableDirection == DIRECTION::RIGHT ) { moveDRotated = DIRECTION::DOWN;}
				if ( affection.movableDirection == DIRECTION::DOWN  ) { moveDRotated = DIRECTION::LEFT;}
				if ( affection.movableDirection == DIRECTION::LEFT  ) { moveDRotated = DIRECTION::UP;}
			}
			else if (_moveForDemonstration.putCardDirection == DIRECTION::DOWN) {
				affectionRotatedX = affection.x * -1; affectionRotatedY = affection.y * -1;

				if ( affection.movableDirection == DIRECTION::UP    ) { moveDRotated = DIRECTION::DOWN;}
				if ( affection.movableDirection == DIRECTION::RIGHT ) { moveDRotated = DIRECTION::LEFT;}
				if ( affection.movableDirection == DIRECTION::DOWN  ) { moveDRotated = DIRECTION::UP;}
				if ( affection.movableDirection == DIRECTION::LEFT  ) { moveDRotated = DIRECTION::RIGHT;}
			}
			else if (_moveForDemonstration.putCardDirection == DIRECTION::LEFT) {
				affectionRotatedX = affection.y; affectionRotatedY = affection.x * -1;

				if ( affection.movableDirection == DIRECTION::UP    ) { moveDRotated = DIRECTION::LEFT;}
				if ( affection.movableDirection == DIRECTION::RIGHT ) { moveDRotated = DIRECTION::UP;}
				if ( affection.movableDirection == DIRECTION::DOWN  ) { moveDRotated = DIRECTION::RIGHT;}
				if ( affection.movableDirection == DIRECTION::LEFT  ) { moveDRotated = DIRECTION::DOWN;}
			}
			*/

			Grid *rGrid = grid->GetRelativeGrid ( affectionRotatedX, affectionRotatedY);
			if ( rGrid != NULL ) {
				rGrid->Draw_CardEffectDemonstration (window, moveDRotated);
			}
		}


	///// 駒移動の場合
	} else if ( _moveForDemonstration.action == AI_Modules::MOVE::ACTION::MOVE_PIECE ) {

		sf::Vector2f arrowFrom;
		sf::Vector2f arrowTo;

		Grid *gridFrom = _board->GetGrid ( _moveForDemonstration.movePieceFromX, _moveForDemonstration.movePieceFromY );
		Grid *gridTo   = _board->GetGrid ( _moveForDemonstration.movePieceToX,   _moveForDemonstration.movePieceToY );

		arrowFrom = gridFrom->GetPosition();
		arrowFrom.x += gridFrom->GetSize().x / 2.0f;
		arrowFrom.y += gridFrom->GetSize().y / 2.0f;

		arrowTo = gridTo->GetPosition();
		arrowTo.x += gridTo->GetSize().x / 2.0f;
		arrowTo.y += gridTo->GetSize().y / 2.0f;

		Piece *piece = gridFrom->GetPiece();

		_arrowForMoveDemonstration->Init ( arrowFrom, arrowTo );
		_arrowForMoveDemonstration->Draw(window);
		piece->DrawDemonstration(window, arrowTo);

		// ぶつける場合
		if (_moveForDemonstration.movePieceAndIsCllide ) {

			Grid *gridCollide = _board->GetGrid (
				_moveForDemonstration.movePieceToX + Utils_Cal::Sign ( _moveForDemonstration.movePieceToX - _moveForDemonstration.movePieceFromX ),
				_moveForDemonstration.movePieceToY + Utils_Cal::Sign ( _moveForDemonstration.movePieceToY - _moveForDemonstration.movePieceFromY )
			);

			if ( gridCollide != NULL ) {
				Piece *collidedPiece = gridCollide->GetPiece();

				if ( collidedPiece != NULL ) {
					sf::Vector2f collideGridPos = collidedPiece->GetCurrentPosition();

					collideGridPos.x = (collideGridPos.x + arrowTo.x) / 2.0f;
					collideGridPos.y = (collideGridPos.y + arrowTo.y) / 2.0f;

					_crashForMoveDemonstration->SetPosition(collideGridPos);
					_crashForMoveDemonstration->Update(window);
					_crashForMoveDemonstration->Draw(window);
				}
			}
		}


	///// カード取得の場合
	} else if ( _moveForDemonstration.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

		sf::Vector2f arrowFrom;
		sf::Vector2f arrowTo;

		Grid *grid = _board->GetGrid ( _moveForDemonstration.takeCardX, _moveForDemonstration.takeCardY );
		arrowFrom = grid->GetPosition();
		arrowFrom.x += grid->GetSize().x / 2.0f;
		arrowFrom.y += grid->GetSize().y / 2.0f;

		Card *takenCard = grid->GetCard();

		CardBox *cardBoxTo = NULL;

		if (_moveForDemonstration.turn == 1) { cardBoxTo = _cardBox_P1;}
		if (_moveForDemonstration.turn == 2) { cardBoxTo = _cardBox_P2;}

		if ( cardBoxTo != NULL ) {

			arrowTo = cardBoxTo->GetPosition();
			arrowTo.x += cardBoxTo->GetBoxWidth() / 2.0f;
			arrowTo.y += cardBoxTo->GetBoxHeight() / 2.0f;

			vector<Card*> cardList = cardBoxTo->GetAllCards();
			for ( unsigned int i = 0; i < cardList.size(); ++i ) {

				Card *card = cardList[i];
				if (card->GetProperty().cardName == takenCard->GetProperty().cardName) {

					arrowTo = card->GetCurrentPosition();
				}

			}
		}

		_arrowForMoveDemonstration->Init ( arrowFrom, arrowTo );
		_arrowForMoveDemonstration->Draw(window);

	}

}


//=========================================================
// 口出しモード進行：戻り値＝待機orNot
//
// 進行フロー：
//     ・AIから候補手が出てきたところ
//     ・「これでいい？」の質問を表示してプレイヤーの回答を待っているところ
//     ・プレイヤーの操作（着手）を待機しているところ
//
//=========================================================
void GameManager::Update_InterferenceMode ( AI_Modules::ProgressData AI_thinkResult ) {

	//bool isWait = false;

	//ClearMoveDemonstration ();
	/*
	printf ("口出しモード更新 : action[%d]\n", _interferenceModeInfo.moveSuggestedByAI.action);
	printf ("  口出しモード更新 : action[%d]\n", _interferenceModeInfo.moveSuggestedByAI.action);
	printf ( "    maxTimes : %d\n", _interferenceModeInfo.maxTimes );
	printf ( "    currentTimes : %d\n", _interferenceModeInfo.currentTimes );
	printf ("  口出しモード更新 : action[%d]\n", _interferenceModeInfo.moveSuggestedByAI.action);
	printf ("口出しモード更新 : action[%d]\n", _interferenceModeInfo.moveSuggestedByAI.action);
	*/


	// 現在の手番が口出され対象じゃないなら何もしない
	if ( !InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode() ) {
		printf ("現在の手番が口出され対象じゃないので何もしない\n");
		return;
	}

	// 進行フェイズ：次の自分の手番を待っているところ
	if (_interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN) {

		// 処理なし
		printf ("進行フェイズ：次の自分の手番を待っているところ\n");

		// 口出し直後ってことはないですよね、すでに
		//_interferenceModeInfo.isRightAfterInterferring = false;

		return;

	}
	// 進行フェイズ：AIからの候補手を待っていたところ
	else if (_interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_AI_MOVE) {

		printf ("口出しモード：AIから候補手が来たところ\n");

		// ここでのみ、候補手を受け取る
		_interferenceModeInfo.moveSuggestedByAI = AI_thinkResult.answerMove;

		// 「カードを取る」の場合、カード自体の情報が入ってないので、ここで入れておく
		if ( _interferenceModeInfo.moveSuggestedByAI.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

			int x = _interferenceModeInfo.moveSuggestedByAI.takeCardX;
			int y = _interferenceModeInfo.moveSuggestedByAI.takeCardY;

			Card *card = _board->GetGrid(x, y)->GetCard();
			if ( card != NULL ) {
				_interferenceModeInfo.moveSuggestedByAI.takeCardCardId = card->GetProperty().processId;
			}
		}

		// 仮表示をセット
		SetMoveDemonstration ( _interferenceModeInfo.moveSuggestedByAI );

		// 選択肢を出して、
		Actor *actor = GetCurrentTurnActor();
		Actor::InterfereMode_AskConfirmationReactionData reactionData =
			actor->InterfereMode_AskConfirmationReaction ( _interferenceModeInfo.moveSuggestedByAI, AI_thinkResult.evaluationSumUp );

		string selectOption_agree = reactionData.selectOptionLabel_agree;
		string selectOption_no    = reactionData.selectOptionLabel_no;

		TextMessageWindow::ReplaceWithAlias(selectOption_agree);
		TextMessageWindow::ReplaceWithAlias(selectOption_no);

		SelectOption *option_agree = new SelectOption( selectOption_agree, "agree" );
		SelectOption *option_no    = new SelectOption( selectOption_no   , "no" );



		//TextMessageWindow::ReplaceWithAlias(text);

		vector<SelectOption*> optionList;
		optionList.push_back(option_agree);
		optionList.push_back(option_no);

		if ( _select != NULL ) {
			delete _select;
		}
		_select = new Select ( _scriptEngine );

		_select->Clear(); // 一応。
		_select->SetUpSelectOptions( optionList );

		// 選択肢の表示位置
		int moveX = 0;
		int moveY = 0;
		vector<string> p = Utils_String::Split(_interferenceModeInfo.selectOptionsPos, ",");
		( p.size() >= 2  ) ? moveX = Utils_Convert::String_to_Int(p[0]) : 0;
		( p.size() >= 2  ) ? moveY = Utils_Convert::String_to_Int(p[1]) : 0;
		_select->Move ( moveX, moveY );

		// 回答待ちに移行
		_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_CONFIRMATION;

		return;

	}
	// 進行フェイズ：プレイヤーに「これでいい？」と確認して回答を待っているところ
	else if (_interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_CONFIRMATION) {

		//printf ("口出しモード：プレイヤーに「これでいい？」と確認して回答を待っているところ\n");

		// 一応、常に最新情報を、ということで。
		SetMoveDemonstration ( _interferenceModeInfo.moveSuggestedByAI );


		// 回答をモニタリングして、
		bool isAgreedWithAIMove = true;

		string answerFromPlayer = _select->GetSelectOption();

		// 回答が来た
		if ( answerFromPlayer != "" ) {

			isAgreedWithAIMove = ( answerFromPlayer == "agree" );

			// AIの候補手でOKなら、指してね。
			if ( isAgreedWithAIMove ) {

				// 次の自分の手番まで待機
				//_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN;
				_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_AI_MOVE;

				// 同意してもらったよ！
				_interferenceModeInfo.isAgreedWithAIMove = true;

			// AIの候補手に同意を得られず。プレイヤーの手番にする
			} else {

				_interferenceModeInfo.isAgreedWithAIMove = false;

				// この時点で回数を消費する
				_interferenceModeInfo.currentTimes += 1;
				_interfereModePanel->SetCurrentTimes ( _interferenceModeInfo.currentTimes );

				// Lua側から「じゃあお願いね」のセリフおよび表情を取得してリアクションを実行
				Actor *actor = GetCurrentTurnActor();
				actor->InterfereMode_AskPlayersMoveReaction (_interferenceModeInfo.moveSuggestedByAI, AI_thinkResult.evaluationSumUp);

				// プレイヤーの着手を待機するフェイズに移行
				_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE;

			}

			// 選択肢を片付ける
			delete _select;
			_select = NULL;

			// 候補手仮表示終了
			ClearMoveDemonstration ();

		// まだ回答来てない
		} else {
			return;
		}


	}
	// 進行フェイズ：プレイヤーの着手を待機しているところ
	else if (_interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE) {

		//printf ("口出しモード：プレイヤーの着手を待機しているところ\n");


		return;

	} else {
		printf ("口出しモード：どのフェイズでもない。\n");
		return;
	}

	return;
}

/*
//=========================================================
// 口出しモードでプレイヤーが着手を決定した
//=========================================================
void GameManager::PlayerMoveDone_InterferenceMode () {

	// プレイヤーの操作は完了
	_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN;

}
*/

//=========================================================
// 口出しモード：口出しモードでのプレイヤー操作中であるかどうか
//=========================================================
bool GameManager::InterferenceMode_IsPlayersDecision () {

	bool result = false;

	if ( _interferenceModeInfo.isActive && _interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE ) {
		result = true;
	}

	return result;
}

//=========================================================
// 現在の手番の参加者は口出され対象ですか？
//=========================================================
bool GameManager::InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode () {

	bool result = false;

	Participant participant = GetCurrentTurnParticipant();

	// 現在の手番の参加者は口出され対象ですか？
	if ( participant.interferenceModeInfo.isActive ) {

		result = true;
	}

	return result;
}

//=========================================================
// 口出しモードとして処理しますか？
//=========================================================
bool GameManager::IsInterferenceMode () {

	bool result = false;


	// 口出しモードでの試合ですか？
	// 現在の手番の参加者は、口出され対象のキャラですか？

	if
	(
		_interferenceModeInfo.isActive
		&&
		InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode()
		//&&
		//Is_AI_Turn_Now()

	)
	{

		// AIの候補手が「それでよし」と同意を得られた場合は、口出しモードとしては処理しない

		// 未同意である場合、口出しモードとして処理せよ
		if ( !_interferenceModeInfo.isAgreedWithAIMove ) {
			result = true;
		}

		// 口出し回数、残ってますか？
		if ( _interferenceModeInfo.currentTimes >= _interferenceModeInfo.maxTimes ) {

			result = false;

		}

		// プレイヤーの着手待ちのときはまさに口出し中ではありませんか
		if ( _interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE ) {

			result = true;
		}


	}

	return result;
}

//=========================================================
// 口出しモードの試合ですか？（試合自体が口出しモードであるか否か）
//=========================================================
bool GameManager::IsInterferenceBattle () {

	// 参加者の中に一人でも口出しモードオンの人がいたら「これは口出しモードの試合」という判定。

	bool result = false;

	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {

		Participant p = _participantsList[i];
		if ( p.interferenceModeInfo.isActive ) {

			result = true;
			break;
		}
	}

	return result;
}

//=========================================================
// 口出しモード：プレイヤーが口出し案を決定した
//=========================================================
void GameManager::InterferenceMode_PlayerMoveDecided ( AI_Modules::Move moveByPlayer ) {

	//_interferenceModeInfo.moveSuggestedByPlayer = moveByPlayer;

	if ( _interferenceModeInfo.isActive ) {

		// 現在の手番の参加者が口出され対象ならば、口出しモード進行関連の処理をする
		if ( InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode() ) {

			// 口出しモードで口出し着手をした直後だぞ、ということ
			/*
				処理なしですか。
			*/
		}
	}
}

//=========================================================
// 口出しモード的にも着手完了通知をもらうよ：手番に関係なく（手番交代前に）
//=========================================================
void GameManager::InterferenceMode_ModeDone() {

	if ( _interferenceModeInfo.isActive ) {

		// 現在の手番の参加者が口出され対象ならば、口出しモード進行関連の処理をする
		if ( InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode() ) {

			//// プレイヤーの着手を待っていた場合に、すること
			if ( _interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE ) {

				// 口出しモードで口出し着手をした直後だぞ、ということ
				//_interferenceModeInfo.isRightAfterInterferring = true;
				/*
					処理なしですか。
				*/
			}

			_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN;

		}
	}
}

//=========================================================
// 口出しモード管理者さんへの手番開始通知：手番に関係なく（手番交代前に）
//=========================================================
void GameManager::InterferenceMode_NewTurnStarts() {

	// 新手番が口出され対象ならば
	if ( InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode() ) {

		InterferenceMode_StartWaitingForAIMove();
	}

}

//=========================================================
// 口出しモード：AIの指し手候補待ち受け開始
//=========================================================
void GameManager::InterferenceMode_StartWaitingForAIMove () {

	if ( _interferenceModeInfo.isActive ) {

		// AIの指し手案を待ち受けるモードにする
		_interferenceModeInfo.phase = _interferenceModeInfo.PHASE_WAIT_FOR_AI_MOVE;

		// 未同意ですよ。
		_interferenceModeInfo.isAgreedWithAIMove = false;

		// 思考前ですよ。
		_progressData_AfterThinking.progressRate = 0.0f;

		// 進行パネルに情報を表示
		if ( _interfereModePanel != NULL ) {

			_interfereModePanel->SetMaxTimes ( _interferenceModeInfo.maxTimes );
			_interfereModePanel->SetCurrentTimes ( _interferenceModeInfo.currentTimes );
		}
	}
}


//=========================================================
// 指定の参加者の口出しモードを設定する
//=========================================================
void GameManager::SetInterferenceModeToParticipant ( string participantId, int maxTimes, string selectOptionsPos ) {

	for (unsigned int i = 0; i < _participantsList.size(); ++i) {

		if (_participantsList[i].id == participantId) {

			// すでに候補手からの選択モードが設定されていたら設定しないこととする
			if ( !_participantsList[i].choiceModeInfo.isActive ) {

				_participantsList[i].interferenceModeInfo.isActive = true;
				_participantsList[i].interferenceModeInfo.maxTimes = maxTimes;
				_participantsList[i].interferenceModeInfo.selectOptionsPos = selectOptionsPos;
				
				
				// 進行パネルに情報を表示してしまう。口出しの相方が後手番のとき、最初の手番待ちの間「回数 0」の表示のままなのが気になるので、ごまかす意図。
				if ( _interfereModePanel != NULL ) {

					_interfereModePanel->SetMaxTimes ( maxTimes );
					_interfereModePanel->SetCurrentTimes ( 0 );
				}
			}

			break;
		}

	}
}

//=========================================================
// 参加者の口出しモード情報をGamaManager側に展開する
//=========================================================
void GameManager::DeployInterferenceModeFromParticipant ( string participantId ) {

	for (unsigned int i = 0; i < _participantsList.size(); ++i) {

		if (_participantsList[i].id == participantId) {

			_interferenceModeInfo = _participantsList[i].interferenceModeInfo;

			break;
		}

	}
}

//=========================================================
// 参加者の口出しモード情報に、GamaManager側から保存する
//=========================================================
void GameManager::WriteInterferenceModeToParticipant ( string participantId ) {

	for (unsigned int i = 0; i < _participantsList.size(); ++i) {

		if (_participantsList[i].id == participantId) {

			_participantsList[i].interferenceModeInfo = _interferenceModeInfo;

			break;
		}

	}

}


//=========================================================
// 指定の参加者の候補手からの選択モードの設定をする
//=========================================================
void GameManager::SetChoiceModeToParticipant ( string participantId, int numberOfChoices, string selectOptionsPos, vector< pair<int, int> > extraChoicesRange ) {

	for (unsigned int i = 0; i < _participantsList.size(); ++i) {

		if (_participantsList[i].id == participantId ) {

			printf ( "%s の候補手からの選択モードを有効にする\n", participantId.c_str() );

			_participantsList[i].choiceModeInfo.isActive          = true;
			_participantsList[i].choiceModeInfo.numberOfChoices   = numberOfChoices;
			_participantsList[i].choiceModeInfo.selectOptionsPos  = selectOptionsPos;
			_participantsList[i].choiceModeInfo.extraChoicesRange = extraChoicesRange;

			// 口出しモードが先に設定されているかどうかは知る由もありませんが、念のため無効にしておく
			_participantsList[i].interferenceModeInfo.isActive = false;

			break;
		}

	}
}

//=========================================================
// 参加者の候補手からの選択モード情報をGamaManager側に展開する
//=========================================================
void GameManager::DeployChoiceModeFromParticipant ( string participantId ) {

	for (unsigned int i = 0; i < _participantsList.size(); ++i) {

		if (_participantsList[i].id == participantId) {

			_choiceModeInfo = _participantsList[i].choiceModeInfo;

			break;
		}

	}
}

//=========================================================
// 参加者の口候補手からの選択モード情報に、GamaManager側から保存する
//=========================================================
void GameManager::WriteChoiceModeToParticipant ( string participantId ) {

	for (unsigned int i = 0; i < _participantsList.size(); ++i) {

		if (_participantsList[i].id == participantId) {

			_participantsList[i].choiceModeInfo = _choiceModeInfo;

			break;
		}

	}
}


//=========================================================
// 候補手からの選択モードとして処理しますか？（ScriptEngine側での進行処理の問題）
//=========================================================
bool GameManager::IsChoiceMode () {

	bool result = false;
;
	if ( _choiceModeInfo.isActive ) {

		// プレイヤーが選択肢を選択した後は、それをAIの指し手ということにして進めてね、ということ。
		if ( _choiceModeInfo.phase != _choiceModeInfo.PHASE_AFTER_PLAYER_DECIDED ) {

			result = true;
		}
	}

	return result;
}

//=========================================================
// 手番交代後の、候補手からの選択モードの進行処理（新手番の人が対象者ならば云々……）
//=========================================================
void GameManager::ChoiceMode_NewTurnStarts() {

	if ( _choiceModeInfo.isActive ) {
		_choiceModeInfo.phase = _choiceModeInfo.PHASE_WAIT_FOR_AI_MOVE;
	}
}

//=========================================================
// 候補手からの選択モード進行
//=========================================================
void GameManager::Update_ChoiceMode ( AI_Modules::ProgressData AI_thinkResult ) {

	// 参考情報として
	Participant currentTurnParticipant = GetCurrentTurnParticipant();

	if ( !_choiceModeInfo.isActive ) {
		return;
	}

	///// 進行フェイズ：自分の手番を待機
	if ( _choiceModeInfo.phase == _choiceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN )
	{
		// 待機
		// 結局ここに来ることはないのではないか？ そうかもしれません。

	}
	///// 進行フェイズ：AIの思考待ち
	else if ( _choiceModeInfo.phase == _choiceModeInfo.PHASE_WAIT_FOR_AI_MOVE )
	{

		// AI、思考完了している？
		if ( AI_thinkResult.progressRate >= 100.0f && !IsMoveEffectGoing() ) {

			printf ( "来てますか？[%s]\n", currentTurnParticipant.id.c_str() );
			printf ( "%f\n", AI_thinkResult.progressRate );
			printf ( "%s\n", MoveToString( AI_thinkResult.answerMove).c_str() );

			// Actorさんにリアクションしてもらう
			GetCurrentTurnActor()->ChoiceMode_AskPlayerDecisionReaction ( AI_thinkResult.answerMove, AI_thinkResult.evaluationSumUp );
			
			_choiceModeInfo.moveListSuggestedByAI.clear();
			
			// まずはAI選定の候補手を１番目の選択肢として入れる。
			_choiceModeInfo.moveListSuggestedByAI.push_back ( AI_thinkResult );
			
			// ２番目以降の選択肢を作成
			if ( currentTurnParticipant.ai != NULL ) { // AIさん、いますよね？ 一応確認。
				
				// AI側で並べ替え済みの候補手を確保
				vector< pair<AI_Modules::Move, int> > evaluatedMoveList_theLatest = currentTurnParticipant.ai->_evaluatedMoveList_theLatest;
				
				// 現時点での合法手も確保。AI側で絞り込んで足りなくなっているときに備えて。
				vector<AI_Modules::Move> nextMoveList;
				{
					AI_Modules::InnerExpression *ie = GenerateInnerExpressionFromCurrentState();
					ie->GetLegalNextMoveList(nextMoveList);
					UTILS_OTHERS::RandomShuffle ( nextMoveList );
					delete ie;
				}
					
				// 指定の選択肢数になるまで作成しますよ
				init_genrand( static_cast<unsigned int>( time(NULL) ) );
				while ( _choiceModeInfo.moveListSuggestedByAI.size() < _choiceModeInfo.numberOfChoices ) {
					
					bool forceQuit = false; // 強制breakフラグ。一応
					
					// 今から何番目の選択肢を作るところですか？
					int index = _choiceModeInfo.moveListSuggestedByAI.size();
					
					// 範囲の指定がありますか？
					if ( (index-1) < _choiceModeInfo.extraChoicesRange.size() ) {
						
						int from = _choiceModeInfo.extraChoicesRange[ index-1 ].first;
						int to   = _choiceModeInfo.extraChoicesRange[ index-1 ].second;
						
						// to に　-1 で「最後まで」の意味。
						if ( to < 0 ) { to = nextMoveList.size() - 1; }
						
						// to の上限越えに対応
						if ( to >= nextMoveList.size() ) { to = nextMoveList.size() - 1; }
						
						// from も上限越えする場合がある：もうゼロにしてしまう。
						if ( from >= nextMoveList.size() ) { from = 0; }
						
						printf ( "\nfrom - to : %d - %d \n", from, to );
						
						// この範囲で候補手のリストを作成
						vector<AI_Modules::Move> candidateList;
						for ( int n = from; n <= to; ++n ) {
							
							// AIでソート済みのリストから取得できるなら取得
							if ( n < evaluatedMoveList_theLatest.size() ) {
								
								//printf ("out of evaluatedMoveList_theLatest \n");
								
								candidateList.push_back ( evaluatedMoveList_theLatest[n].first );
								
							}
							// 可能な候補手リストから取得できるなら取得
							else if ( n < nextMoveList.size() ) {
								
								printf ("out of nextMoveList \n");
								
								candidateList.push_back ( nextMoveList[n] );
							}
							// できない。お手上げですね。
							else {
								forceQuit = true;
								printf ( "evaluatedMoveList_theLatest.size() : %d, nextMoveList.size() : %d\n", evaluatedMoveList_theLatest.size(), nextMoveList.size() );
							}
							
						}
						
						// 作成したリストからランダムに取得
						if ( candidateList.size() > 0 ) { 
							
							int randomIndex = genrand_int31() % candidateList.size();
							
							AI_Modules::ProgressData option;
							option.answerMove = candidateList[randomIndex];
							option.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::EVEN;

							_choiceModeInfo.moveListSuggestedByAI.push_back ( option );
							
							
						} else {
							// おかしいですね。
							printf ( "konnano okasii desuyo!!!\n" );
							
							forceQuit = true;
						}
						
						
					// 範囲の指定がなかった
					} else {
						
						int randomIndex = genrand_int31() % nextMoveList.size();
						
						AI_Modules::ProgressData option;
						option.answerMove = nextMoveList[randomIndex];
						option.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::EVEN;
						
						_choiceModeInfo.moveListSuggestedByAI.push_back ( option );
						
					}
					
					if ( forceQuit ) { break;}
				} // while
				
				
				// 重複しているものがあったら、できるだけ別のものに変更する
				for ( unsigned int i = 0; i < _choiceModeInfo.moveListSuggestedByAI.size(); ++i ) {
					
					AI_Modules::ProgressData candidateA = _choiceModeInfo.moveListSuggestedByAI[i];
					
					bool isDoubled = false;
					
					for ( unsigned int ii = 0; ii < _choiceModeInfo.moveListSuggestedByAI.size(); ++ii ) {
						
						if ( i != ii ) { 
							AI_Modules::ProgressData candidateB = _choiceModeInfo.moveListSuggestedByAI[ii];
							
							if ( IsEqualMove ( candidateA.answerMove, candidateB.answerMove ) ) {
								
								isDoubled = true;
								break;
							}
						}
					}
					
					if ( isDoubled ) {
					
						printf ( "isDoubled (%s) \n", MoveToString ( candidateA.answerMove ).c_str() );
						
						for ( unsigned int n = 0; n < nextMoveList.size(); ++n ) {
							
							AI_Modules::Move moveForAlt = nextMoveList[n];
							
							// 現リストのどれとも重複しないことを確認
							bool isUnique = true;
							for ( unsigned int ii = 0; ii < _choiceModeInfo.moveListSuggestedByAI.size(); ++ii ) {
								
								AI_Modules::ProgressData oneOfCurrentChoices = _choiceModeInfo.moveListSuggestedByAI[ii];
								
								if ( IsEqualMove ( oneOfCurrentChoices.answerMove, moveForAlt ) ) {
									
									isUnique = false;
									break;
								}
							}
							
							if ( isUnique ) {
								
								candidateA.answerMove      = moveForAlt;
								candidateA.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::EVEN;
								
								_choiceModeInfo.moveListSuggestedByAI[i] = candidateA;
								
								printf ( "Replaced (%s) \n", MoveToString ( moveForAlt ).c_str() );
								
								break;
							}
							
						} // 代替候補探索 の for
						
					}
				
				} // 重複避け
				

				
			// 現在のターンの参加者にAIが設定されていない。おかしいですね。
			} else {
				printf ("CurrentTurnParticipant does not have AI.\n");
			}
			
			/*
			// TODO とりあえず、AIの指し手と、後は適当に埋めて選択肢を作っておく
			_choiceModeInfo.moveListSuggestedByAI.clear();
			
			_choiceModeInfo.moveListSuggestedByAI.push_back ( AI_thinkResult );

			AI_Modules::InnerExpression *ie = GenerateInnerExpressionFromCurrentState();

			vector<AI_Modules::Move> nextMoveList;
			ie->GetLegalNextMoveList(nextMoveList);
			UTILS_OTHERS::RandomShuffle ( nextMoveList );

			for ( unsigned int i = 0; i < nextMoveList.size(); ++i ) {

				if ( _choiceModeInfo.moveListSuggestedByAI.size() >= _choiceModeInfo.numberOfChoices ) {
					break;
				}

				if ( !IsEqualMove ( nextMoveList[i], AI_thinkResult.answerMove ) ) {

					AI_Modules::ProgressData option;
					option.answerMove = nextMoveList[i];
					option.evaluationSumUp = AI_Modules::EVALUATION::SUM_UP::EVEN;

					_choiceModeInfo.moveListSuggestedByAI.push_back ( option );
				}
			}
			*/


			///// 画面に表示する選択肢を作成
			if (_select != NULL) { 
				delete _select; 
			}
			_select = new Select ( _scriptEngine );
			_select->SetGameBoardId(_name);

			vector<SelectOption*> optionList;

			SelectOption::InitProperties optionInitProp;
			optionInitProp.fontSize = 24;
			optionInitProp.marginSize = sf::Vector2f ( 5.0f, 10.0f );
			optionInitProp.backRectWidthAdjustValue = 5;

			for ( unsigned int i = 0; i < _choiceModeInfo.moveListSuggestedByAI.size(); ++i ) {

				AI_Modules::ProgressData moveChoice = _choiceModeInfo.moveListSuggestedByAI[i];

				// Lua側に、指し手の文字列表現を問い合わせる

				// 「カードを取る」の場合、カード自体の情報が入ってないので、ここで入れておく
				if ( moveChoice.answerMove.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

					int x = moveChoice.answerMove.takeCardX;
					int y = moveChoice.answerMove.takeCardY;

					Card *card = _board->GetGrid(x, y)->GetCard();
					if ( card != NULL ) {
						moveChoice.answerMove.takeCardCardId = card->GetProperty().processId;
					}
				}

				string choiceString = GetCurrentTurnActor()->GetMoveString(moveChoice.answerMove); // MoveToString ( moveChoice.answerMove );
				string choiceData = std::to_string(i);

				//string choiceData = "index=" + std::to_string(i) + "sumup=" + moveChoice.evaluationSumUp;

				SelectOption *option = new SelectOption ( choiceString, choiceData, optionInitProp );

				optionList.push_back ( option );
			}

			// シャッフルしますよ
			//UTILS_OTHERS::RandomShuffle(optionList);
			init_genrand( static_cast<unsigned int>( time(NULL) ) );
			for ( int iii = 0; iii < 100; ++iii ) {
				
				//int randomIndex = Utils_Cal::GetRandom ( optionList.size() );
				int randomIndex = genrand_int31() % optionList.size();
				
				// 先頭の要素と乱数番目の要素を入れ替える
				SelectOption *option_0 = optionList[0];
				SelectOption *option_r = optionList[randomIndex];
				
				optionList[0] = option_r;
				optionList[randomIndex] = option_0;
				
			}

			
			_select->Clear(); // 一応。
			_select->Move ( 0, 0 ); // 位置もクリア
			
			_select->SetRowGap (1);
			_select->SetUpSelectOptions ( optionList );

			// 選択肢の表示位置
			int moveX = 0;
			int moveY = 0;
			vector<string> p = Utils_String::Split(_choiceModeInfo.selectOptionsPos, ",");
			( p.size() >= 2  ) ? moveX = Utils_Convert::String_to_Int(p[0]) : 0;
			( p.size() >= 2  ) ? moveY = Utils_Convert::String_to_Int(p[1]) : 0;
			_select->Move ( moveX, moveY );
	
			// 回答待ちに移行
			_choiceModeInfo.phase = _choiceModeInfo.PHASE_WAIT_FOR_PLAYER_DECISION;

		// AIの思考完了待ち
		} else {
			// 待機。
		}

	}
	///// 進行フェイズ：プレイヤーの決断待ち
	else if ( _choiceModeInfo.phase == _choiceModeInfo.PHASE_WAIT_FOR_PLAYER_DECISION )
	{

		///// 選択をモニタリング
		string selectedOption = _select->GetSelectOption();

		///// マウスが乗っている選択肢を取得して、盤上に仮表示する
		string mouseoverOption = _select->GetMouseoverOptionJumpLabel();

		if ( mouseoverOption != "" ) {

			ClearMoveDemonstration();

			int index = std::stoi ( mouseoverOption );
			if ( index < _choiceModeInfo.moveListSuggestedByAI.size() ) {

				SetMoveDemonstration ( _choiceModeInfo.moveListSuggestedByAI[index].answerMove );
			}


		} else {

			ClearMoveDemonstration();
		}


		///// 選択されている
		if ( selectedOption != "" ) {

			//map<string, string> selectData = Utils_String::ParseStringAsMap ( selectedOption, ",", "=" );
			//string index = UTILS_OTHERS::GetValueFromMap ( selectData, "index", "" );

			printf ( "%sを選択した\n", selectedOption.c_str() );

			// 何を受け取ったのかを記録して、
			_choiceModeInfo.playersSelectChoiceNumber = selectedOption;

			// 選択肢を消して、
			delete _select;
			_select = NULL;

			// 仮表示も消して
			ClearMoveDemonstration();

			// フェイズを「回答受け取り済み」に移行
			_choiceModeInfo.phase = _choiceModeInfo.PHASE_AFTER_PLAYER_DECIDED;

		}

	}

	///// 進行フェイズ：プレイヤーの決断を受け取った後
	else if ( _choiceModeInfo.phase == _choiceModeInfo.PHASE_AFTER_PLAYER_DECIDED )
	{

		// 処理なし。というか、ここにも来ないのではないか。

	} else {
		printf ("候補手からの選択フェイズ：どのフェイズでもない。\n");
	}

}



//=========================================================
// 候補手からの選択モードであれば、プレイヤーが選んだものに置き換える
//=========================================================
AI_Modules::ProgressData GameManager::ReplaceToPlayersChoiceIfNeeded ( AI_Modules::ProgressData AI_think_result ) {

	AI_Modules::ProgressData result = AI_think_result;

	// モードが有効で、
	if ( _choiceModeInfo.isActive ) {

		// プレイヤーが選択済みならば、
		if ( _choiceModeInfo.phase == _choiceModeInfo.PHASE_AFTER_PLAYER_DECIDED ) {

			if (_choiceModeInfo.playersSelectChoiceNumber != "") {

				int index = std::stoi(_choiceModeInfo.playersSelectChoiceNumber);

				if ( index < _choiceModeInfo.moveListSuggestedByAI.size() ) {

					printf ( "AI思考の指し手[%s]をプレイヤーが選択したもの[%s]に置き換えました\n", MoveToString(result.answerMove).c_str(),  MoveToString( _choiceModeInfo.moveListSuggestedByAI[index].answerMove).c_str() );

					result.answerMove = _choiceModeInfo.moveListSuggestedByAI[index].answerMove;

					// sumup評価値も置き換える
					result.evaluationSumUp = _choiceModeInfo.moveListSuggestedByAI[index].evaluationSumUp;



				} else {
					printf ( "プレイヤーが選択した選択肢番号(%d)が選択肢リストからハミ出てるらしいじゃないですか。\n", index );
				}


			} else {
				printf ("ここに来てプレイヤーが選択した選択肢番号が空文字だなんて！\n");
			}

			// じゃあ、フェイズを進めてしまいますか。これはよしておく。随所で呼び出される前提で、進行とは無関係としておく。
			//_choiceModeInfo.phase = _choiceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN;
		}

	}

	return result;
}









//=========================================================
// 指し手の簡易文字列表現
//=========================================================
string GameManager::MoveToString ( AI_Modules::Move &mv ) {

	string result = "";

	if ( mv.action == AI_Modules::MOVE::ACTION::PUT_PIECE ) {

		result  = "Put Piece (" + std::to_string ( mv.putPieceX );
		result += "," + std::to_string ( mv.putPieceY );
		result += ")";
	}
	else if ( mv.action == AI_Modules::MOVE::ACTION::MOVE_PIECE ) {

		result  = "Move Piece (" + std::to_string ( mv.movePieceFromX );
		result += "," + std::to_string ( mv.movePieceFromY );
		result += ") -> ";
		result += "(" + std::to_string ( mv.movePieceToX );
		result += "," + std::to_string ( mv.movePieceToY );
		result += ")";
		
		if ( mv.movePieceAndIsCllide ) {
            result += "collide";
		} else {
            result += "not collide";
		}

	}
	else if ( mv.action == AI_Modules::MOVE::ACTION::PUT_CARD ) {

		result  = "Put Card";
		result += ":" + CardFactory::GetCardProperty( mv.putCardId ).cardName;
		result += "(" + std::to_string ( mv.putCardX );
		result += "," + std::to_string ( mv.putCardY );
		result += ")";
		result += "Direction :" + std::to_string ( mv.putCardDirection );
	}
	else if ( mv.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

		result  = "Take Card";
		result += "(" + std::to_string ( mv.takeCardX );
		result += "," + std::to_string ( mv.takeCardY );
		result += ") ";

	}


	return result;
}



//=========================================================
// 棋譜を表示する
//=========================================================
void GameManager::Draw_ScoreSheet ( sf::RenderWindow &window ) {

	/*
		なぜ棋譜だけ分けて描画するのかと言うと、
		メッセージウィンドウの下になってしまうからなのでした。
	*/


	// 棋譜には描画機能があるのでした。
	if (_score != NULL) {
	
		_score->Draw(window);
		
		// じゃあ、ここで試合タイトルを表示しますか。（棋譜再生モード用のラベルなので、ここ）
		if ( _score->IsShown() ) {
			window.draw(_battleTitleForKifuViewing);
		}
	}
	

}

//=========================================================
// 指し手をGUI上の盤面で実行する
//=========================================================
void GameManager::Move ( AI_Modules::Move &move, bool isUseEffect ) {

	ExecuteMove(move, isUseEffect);
}

//=========================================================
// 指し手をGUI上の盤面で実行する：タグで受け取る版
//=========================================================
void GameManager::Move ( NotXML::Tag &moveTag ) {

	AI_Modules::Move move = Game_Utils::Move_From_Tag ( moveTag );

	ExecuteMove(move);
}

///// スクリプトで指し手を進める場合に、早送りをしても不整合が生じないようにするための待機用フラグ
void GameManager::StartScriptingMoving()      { _isScriptingMoving = true; }
bool GameManager::IsScriptingMovingFinished() { return !_isScriptingMoving;}

//=========================================================
// 指し手をGUI上の盤面で実行する：内部処理
//=========================================================
void GameManager::ExecuteMove( const AI_Modules::Move &move, bool isUseEffect ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 駒設置の場合
	if ( move.action == AI_Modules::MOVE::ACTION::PUT_PIECE) {

		Piece *piece = NULL;
		if      ( move.turn == AI_Modules::TURN::BLACK ) { piece = _pieceBox_P1->TakePiece();}
        else if ( move.turn == AI_Modules::TURN::WHITE ) { piece = _pieceBox_P2->TakePiece();}
        else { 
            printf("ExecuteMove（駒設置）どちらの手番でもない\n");
        }

		if (piece != NULL) {
            
			Grid *grid = _board->GetGrid( move.putPieceX, move.putPieceY );
			if (grid != NULL) {
                
                if (isUseEffect) {

                    ///// 駒設置エフェクト
                    IEffect *effect = new Effect_PutPiece_AI ( 
                            _endMoveFromEffect,  // 終了時の通知先
                            move,                // この着手ですよ。
                            0,                  // エフェクトレベル
                            piece,               // 設置する駒
                            grid                 // 設置先グリッド
                    );
                    Application::GetInstance()->_effectManager->Add ( effect );						
                
                } else {
                    grid->SetPiece( piece );
                    
                    _isScriptingMoving = false;
                }

			}
			
		} else {

            printf("駒 NULL\n");
	
		}

	}
	///// 駒移動の場合
	else if ( move.action == AI_Modules::MOVE::ACTION::MOVE_PIECE) {
           
		if ( isUseEffect ) {

            //　ぶつける場合
			if (move.movePieceAndIsCllide) {

				int targetX = move.movePieceToX + Utils_Cal::Sign( move.movePieceToX - move.movePieceFromX );
				int targetY = move.movePieceToY + Utils_Cal::Sign( move.movePieceToY - move.movePieceFromY );

				Grid *grid = _board->GetGrid( targetX, targetY );

				///// 駒移動エフェクト
				IEffect *effect = new Effect_MovePiece_Collide (
					_endMoveFromEffect,
					move,
					-1, // エフェクトレベルに負の数：キー操作で切り替えてもらう
					_board->GetGrid ( move.movePieceFromX, move.movePieceFromY ),
					_board->GetGrid ( move.movePieceToX,   move.movePieceToY ),
					_board->GetGrid ( move.movePieceFromX, move.movePieceFromY )->GetPiece(),
					grid->GetPiece(),
					GetPieceBox_BelongToPiece ( grid->GetPiece() ),
					grid
				);
				Application::GetInstance()->_effectManager->Add ( effect );

			//　ぶつけない場合
			} else {

				///// 駒移動エフェクト
				IEffect *effect = new Effect_MovePiece (
					_endMoveFromEffect,
					move,
					-1, // エフェクトレベルに負の数：キー操作で切り替えてもらう
					_board->GetGrid ( move.movePieceFromX, move.movePieceFromY ),
					_board->GetGrid ( move.movePieceToX,   move.movePieceToY ),
					_board->GetGrid ( move.movePieceFromX, move.movePieceFromY )->GetPiece()
				);
				Application::GetInstance()->_effectManager->Add ( effect );
			}



		}
		else
		///// エフェクトを使わない移動
		{
			Grid *gridFrom = _board->GetGrid( move.movePieceFromX, move.movePieceFromY );
			Grid *gridTo   = _board->GetGrid( move.movePieceToX, move.movePieceToY );

			if ( gridFrom != NULL && gridTo != NULL ) {

				Piece *piece = gridFrom->GetPiece();

				gridFrom->SetPiece ( NULL  );
				gridTo  ->SetPiece ( piece );
			}

			// ぶつける場合
			if (move.movePieceAndIsCllide) {

				int targetX = move.movePieceToX + Utils_Cal::Sign( move.movePieceToX - move.movePieceFromX );
				int targetY = move.movePieceToY + Utils_Cal::Sign( move.movePieceToY - move.movePieceFromY );

				Grid *grid = _board->GetGrid( targetX, targetY );
				if (grid != NULL) {

					Piece *targetPiece = grid->GetPiece();
					grid->SetPiece(NULL);

					if (targetPiece != NULL) {

						if (targetPiece->GetColor() == Piece::Color::BLACK) {
							_pieceBox_P1->AddPiece( targetPiece );
						}
						else if (targetPiece->GetColor() == Piece::Color::WHITE) {
							_pieceBox_P2->AddPiece( targetPiece );
						}

					}
				}
			}
			
			_isScriptingMoving = false;
		}


	}
	///// カード設置の場合
	else if ( move.action == AI_Modules::MOVE::ACTION::PUT_CARD) {

		Card *card = NULL;
		if (move.turn == AI_Modules::TURN::BLACK) {
			card = _cardBox_P1->TakeCard( move.putCardId );
		}
		else if (move.turn == AI_Modules::TURN::WHITE) {
			card = _cardBox_P2->TakeCard( move.putCardId );
		}

		if (card != NULL) {

			if      ( move.putCardDirection == AI_Modules::DIRECTION::UP    ) { card->SetCurrentDirection( DIRECTION::UP    ); }
			else if ( move.putCardDirection == AI_Modules::DIRECTION::LEFT  ) { card->SetCurrentDirection( DIRECTION::LEFT  ); }
			else if ( move.putCardDirection == AI_Modules::DIRECTION::RIGHT ) { card->SetCurrentDirection( DIRECTION::RIGHT ); }
			else if ( move.putCardDirection == AI_Modules::DIRECTION::DOWN  ) { card->SetCurrentDirection( DIRECTION::DOWN  ); }

			Grid *grid = _board->GetGrid( move.putCardX, move.putCardY );
			if (grid != NULL) {
				grid->SetCard( card );
			}
		}
		
		_isScriptingMoving = false;
	}
	///// カードを取る場合
	else if ( move.action == AI_Modules::MOVE::ACTION::TAKE_CARD) {

		Grid *grid = _board->GetGrid( move.takeCardX, move.takeCardY );
		Card *card = NULL;
		if ( grid != NULL ) {
			card = grid->GetCard();
			grid->SetCard( NULL );
		}

		if ( card != NULL ) {

			if (move.turn == AI_Modules::TURN::BLACK) {
				_cardBox_P1->AddCard( card );
			}
			else if (move.turn == AI_Modules::TURN::WHITE) {
				_cardBox_P2->AddCard( card );
			}

		}
		
		_isScriptingMoving = false;
	}
	///// パス
	if ( move.action == AI_Modules::MOVE::ACTION::PASS) {
		// 処理なし
		
		_isScriptingMoving = false;
	}
}

//=========================================================
// 現在の局面を内部表現化する
//【引数】
// 今、この手番だとして。
//=========================================================
AI_Modules::InnerExpression* GameManager::GenerateInnerExpressionFromCurrentState() {


	// 現在の手番を「黒（P1）/白（P2）」という形式で取得したい
	int turn = 0;

	Participant currentTunParticipant = GetCurrentTurnParticipant();
	if      ( currentTunParticipant.side == "P1" ) { turn = AI_Modules::TURN::BLACK;}
	else if ( currentTunParticipant.side == "P2" ) { turn = AI_Modules::TURN::WHITE;}
	else { printf ("どちらサイドでもない（currentTunParticipant.side：%s）。\n", currentTunParticipant.side.c_str()); }

	// ↓これは違うか。色としては、P1側が常に黒。P2側が常に白。
	/*
		現在の手番の参加者が手番オーダーの中で奇数番目ならば「黒番」、偶数番目ならば「白番」

		最初は「1」として判定しましょう。
	*/
	/*
	int turnOrderIndex = 1;

	for ( unsigned int i = 0 ; i < _operationData.turnOrder.size(); ++i ) {
		if ( _operationData.turnOrder[i] == _operationData.currentTurn ) {
			break;
		}
		++turnOrderIndex;
	}

	// 奇数か偶数か
	if ( turnOrderIndex%2 == 1 ) {  turn = AI_Modules::TURN::BLACK; }
	else                         {  turn = AI_Modules::TURN::WHITE; }
	*/

	/*
	if ( _operationData.currentTurn == "FIRST" ) {

		if      ( _operationData.first == "P1" ) { turn = AI_Modules::TURN::BLACK;}
		else if ( _operationData.first == "P2" ) { turn = AI_Modules::TURN::WHITE;}
	}
	else if ( _operationData.currentTurn == "SECOND" ) {

		if      ( _operationData.second == "P1" ) { turn = AI_Modules::TURN::BLACK;}
		else if ( _operationData.second == "P2" ) { turn = AI_Modules::TURN::WHITE;}
	}
	*/

	return GenerateInnerExpressionFromCurrentState(turn);
}
AI_Modules::InnerExpression* GameManager::GenerateInnerExpressionFromCurrentState( int turn ) {

	AI_Modules::InnerExpression *innerExpression = new AI_Modules::InnerExpression();
	innerExpression->Init();
	innerExpression->Init_SetCurrentTurn( turn );

	///// 盤のサイズ：放っておくと、AIに設定しているルールで設定される。
	innerExpression->_boardSizeX = _board->GetWidth();
	innerExpression->_boardSizeY = _board->GetHeight();

	///// 手札 黒
	vector<int> handCardList_Black = _cardBox_P1->GetAllCardIdList();
	for (unsigned int i = 0; i < handCardList_Black.size(); ++i) {

		innerExpression->Init_AddHandCard_BLACK( handCardList_Black[i] );
	}
	///// 手札 白
	vector<int> handCardList_White = _cardBox_P2->GetAllCardIdList();
	for (unsigned int i = 0; i < handCardList_White.size(); ++i) {

		innerExpression->Init_AddHandCard_WHITE( handCardList_White[i] );
	}
	///// 駒台 黒
	innerExpression->Init_SetHandPiece_BLACK( _pieceBox_P1->GetNumberOfPieces() );

	///// 駒台 白
	innerExpression->Init_SetHandPiece_WHITE( _pieceBox_P2->GetNumberOfPieces() );

	///// 盤面
	int boardSizeX = _board->GetWidth();
	int boardSizeY = _board->GetHeight();
	for ( int y = 0; y < boardSizeY; ++y) {
		for ( int x = 0; x < boardSizeX; ++x) {

			Grid *grid =  _board->GetGrid( x, y );

			///// 駒のあるなし・色
			Piece *piece = grid->GetPiece();
			if (piece != NULL) {

				int pieceColor = piece->GetColor();

				if (pieceColor == Piece::Color::BLACK ) {
					innerExpression->Init_PutPiece( x, y, AI_Modules::GRID::BLACK );
				}
				else if (pieceColor == Piece::Color::WHITE ) {
					innerExpression->Init_PutPiece( x, y, AI_Modules::GRID::WHITE );
				}
			}

			///// カード
			Card *card = grid->GetCard();
			if (card != NULL) {

				int direction = card->GetCurrentDirection();
				int direction_forAI = AI_Modules::DIRECTION::UP;

				if      (direction == DIRECTION::UP    ) { direction_forAI = AI_Modules::DIRECTION::UP;    }
				else if (direction == DIRECTION::LEFT  ) { direction_forAI = AI_Modules::DIRECTION::LEFT;  }
				else if (direction == DIRECTION::RIGHT ) { direction_forAI = AI_Modules::DIRECTION::RIGHT; }
				else if (direction == DIRECTION::DOWN  ) { direction_forAI = AI_Modules::DIRECTION::DOWN;  }

				innerExpression->Init_PutCard( x, y, card->GetProperty().processId, direction_forAI );
			}

			///// 通行可能情報
			int throuableInfo_forAI = AI_Modules::THROUGHABLE_INFO::NOT_THROUGHABLE;
			if ( grid->IsThrouable() ) {

				throuableInfo_forAI = AI_Modules::THROUGHABLE_INFO::THROUGHABLE;

				if      ( grid->IsMustGoUp   () ) { throuableInfo_forAI = AI_Modules::THROUGHABLE_INFO::MUST_TO_UP;   }
				else if ( grid->IsMustGoLeft () ) { throuableInfo_forAI = AI_Modules::THROUGHABLE_INFO::MUST_TO_LEFT; }
				else if ( grid->IsMustGoRight() ) { throuableInfo_forAI = AI_Modules::THROUGHABLE_INFO::MUST_TO_RIGHT;}
				else if ( grid->IsMustGoDown () ) { throuableInfo_forAI = AI_Modules::THROUGHABLE_INFO::MUST_TO_DOWN; }

				innerExpression->Init_SetThrouableInfo( x, y, throuableInfo_forAI );
			}
		}
	}

	///// これは二度手間？
	innerExpression->Init_ReflectCardAffection();

	return innerExpression;
}



//=========================================================
// 駒を追加：駒台に
//=========================================================
void GameManager::AddPiece ( string color, string player ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;


	///// 色を適当に解釈
	int pieceColor = Piece::Color::BLACK;
	if ( color == "WHITE" ) {
		pieceColor = Piece::Color::WHITE;
	}


	///// P1の駒台へ
	if ( player == "P1" || player == "p1") {

		if ( _pieceBox_P1 != NULL) {

			Piece *piece = new Piece ( pieceColor );
			_pieceList_forManagement.push_back ( piece );

			_pieceBox_P1->AddPiece ( piece );
		}
	}
	///// P2の駒台へ
	else if ( player == "P2" || player == "p2" ) {

		if ( _pieceBox_P2 != NULL) {

			Piece *piece = new Piece ( pieceColor );
			_pieceList_forManagement.push_back ( piece );

			_pieceBox_P2->AddPiece ( piece );
		}
	}

}
//=========================================================
// 駒を追加：盤面に
//=========================================================
void GameManager::AddPiece ( string color, int x, int y ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 色を適当に解釈
	int pieceColor = Piece::Color::BLACK;
	if ( color == "WHITE" ) {
		pieceColor = Piece::Color::WHITE;
	}

	///// 盤面、ある？
	if ( _board != NULL ) {

		///// 指定位置のグリッド、ある？ （盤が思ったよりも小さいとか）
		Grid *grid = _board->GetGrid (x, y);
		if ( grid != NULL ) {

			// 先客がいるなら、消しておく
			if ( grid->GetPiece() != NULL ) {

				ReleasePiece ( grid->GetPiece() );
				grid->SetPiece (NULL);
			}

			// 駒を生成して、管理リストにも登録しまして、
			Piece *piece = new Piece ( pieceColor );
			_pieceList_forManagement.push_back ( piece );

			// グリッドさん、よろしくね。
			grid->SetPiece ( piece );
		}

	}

}

//=========================================================
// カードを追加：カード置き場に
//=========================================================
void GameManager::AddCard  ( string cardName, string direction, string player ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	Card *card = CardFactory::GetCard ( cardName );

	// このカードはうちで生成したものですよの印。→ マウスで持っているものがどこ由来なのかを判定したいときがある。ロード時とか。デストラクタとか。
	_cardList_forManagement.push_back(card);

	// ありますよね？ 信じてる！ いやいや、疑ってますよね。
	if ( card == NULL ) {
		printf ( "カード %s は ありません。\n", cardName.c_str()  );
		return;
	}

	// カード置き場に置くときにだって、向きを指定したいんだからね！
	card->SetCurrentDirection ( Game_Utils::Direction_From_String ( direction ) );

	if ( player == "p1" || player == "P1") {
		_cardBox_P1->AddCard ( card );
	}
	else if ( player == "p2" || player == "P2") {
		_cardBox_P2->AddCard ( card );
	}

	SetFreeOperation ( _isFreeOperation );

}

//=========================================================
// カードを追加：盤面に
//=========================================================
void GameManager::AddCard  ( string cardName, string direction, int x, int y ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	Card *card = CardFactory::GetCard ( cardName );

	// このカードはうちで生成したものですよの印。→ マウスで持っているものがどこ由来なのかを判定したいときがある。ロード時とか。デストラクタとか。
	_cardList_forManagement.push_back(card);

	// ありますよね？ 信じてる！ いやいや、疑ってますよね。
	if ( card == NULL ) {
		printf ( "カード %s は ありません。\n", cardName.c_str()  );
		return;
	}

	// 盤面に置くのだから、向きは もちろん 指定するのですよー。
	card->SetCurrentDirection ( Game_Utils::Direction_From_String ( direction ) );

	// グリッド、ある？
	Grid *grid = _board->GetGrid ( x, y );
	if ( grid != NULL ) {

		// 先客いる？
		Card *pre = grid->GetCard ();
		if ( pre != NULL ) {
			CardFactory::Release ( pre );
			grid->SetCard (NULL);
			printf ( "Warning : 【カードを設置】 (%d, %d)に先客がいらしたので、お引取りいただきました。\n", x, y );
		}

		// 晴れてセット。
		grid->SetCard (card);

	// あっ、グリッドがないよ！
	} else {

		// またねー。
		CardFactory::Release ( card );
		printf ( "Warning : 【カードを設置】 (%d, %d)にはグリッドがないので、設置できませんでした。\n", x, y );
	}

	SetFreeOperation ( _isFreeOperation );
}

//=========================================================
// 駒を削除：駒台から
//=========================================================
void GameManager::RemovePiece ( string player ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// P1の駒台から
	if ( player == "P1" || player == "p1") {

		if ( _pieceBox_P1 != NULL) {

			Piece *piece = _pieceBox_P1->TakePiece();
			ReleasePiece ( piece );
		}
	}
	///// P2の駒台から
	else if ( player == "P2" || player == "p2" ) {

		if ( _pieceBox_P2 != NULL) {

			Piece *piece = _pieceBox_P2->TakePiece();
			ReleasePiece ( piece );
		}
	}


}

//=========================================================
// 駒を削除：盤面から
//=========================================================
void GameManager::RemovePiece ( int x, int y ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	Grid *grid = _board->GetGrid ( x, y );

	if ( grid != NULL ) {

		Piece *piece = grid->GetPiece();
		if ( piece != NULL ) {

			ReleasePiece ( piece );
			grid->SetPiece ( NULL );
		}
	}

	/*
		このとき、delete はしないのですね。
		_pieceList_forManagement にはポインタを残しているので、メモリリークにはならない、ということでしょうか？
		多分そうなのだと思います。
		GameManagerのデストラクタで _pieceList_forManagement を見て解放してますよね？
	*/

}

//=========================================================
// カードを削除：カード置き場から
//=========================================================
void GameManager::RemoveCard  ( string cardName, string player ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	if ( player == "p1" || player == "P1") {

		Card *card = _cardBox_P1->TakeCard ( CardFactory::GetCardID ( cardName ) );

		CardFactory::Release ( card );

	}
	else if ( player == "p2" || player == "P2") {

		Card *card = _cardBox_P2->TakeCard ( CardFactory::GetCardID ( cardName ) );

		CardFactory::Release ( card );
	}

}

//=========================================================
// カードを削除；盤面から
//=========================================================
void GameManager::RemoveCard  ( int x, int y ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	Grid *grid = _board->GetGrid ( x, y );

	if ( grid != NULL ) {

		Card *card = grid->GetCard();
		if ( card != NULL ) {

			CardFactory::Release ( card );
			grid->SetCard ( NULL );
		}
	}

}

//=========================================================
// 駒をまとめて削除：駒台から
//=========================================================
void GameManager::RemoveAllPiece_fromBox   ( string player ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 対象となる駒台
	PieceBox *pieceBox = NULL;

	///// P1の駒台から
	if ( player == "P1" || player == "p1") {

		pieceBox = _pieceBox_P1;
	}
	///// P2の駒台から
	else if ( player == "P2" || player == "p2" ) {

		pieceBox = _pieceBox_P2;
	}

	///// 対象となる駒台があるかな？
	if ( pieceBox != NULL ) {

		while (true) {

			Piece *piece = pieceBox->TakePiece();
			if ( piece == NULL ) {
				break;
			}

			ReleasePiece ( piece );

		}

	}
}

//=========================================================
// 駒をまとめて削除：盤面から
//=========================================================
void GameManager::RemoveAllPiece_fromBoard () {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 全グリッド取得するぜぇぇぇ！ 今がそのときだ！
	vector< vector<Grid*> > allGridList = _board->GetAllGrid();

	for ( unsigned int y = 0; y < allGridList.size(); ++y ) {

		for ( unsigned int x = 0; x < allGridList[y].size(); ++x ) {


			Piece *piece = allGridList[y][x]->GetPiece();

			if ( piece != NULL ) {

				ReleasePiece ( piece );
				allGridList[y][x]->SetPiece ( NULL );
			}
		}
	}

}

//=========================================================
// カードをまとめて削除：カード置き場から
//=========================================================
void GameManager::RemoveAllCard_fromBox    ( string player ) {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 対象となるカード置き場
	CardBox *cardBox = NULL;

	///// P1のカード置き場から
	if ( player == "P1" || player == "p1") {

		cardBox = _cardBox_P1;
	}
	///// P2のカード置き場から
	else if ( player == "P2" || player == "p2" ) {

		cardBox = _cardBox_P2;
	}

	///// 対象となるカード置き場があるかな？
	if ( cardBox != NULL ) {


		vector<int> cardIDList = cardBox->GetAllCardIdList();

		for ( unsigned int i = 0; i < cardIDList.size(); ++i ) {

			while ( true ) {

				Card *card = cardBox->TakeCard ( cardIDList[i] );
				if ( card == NULL ) {
					break;
				}

				CardFactory::Release ( card );
			}

		}
	}

}

//=========================================================
// カードをまとめて削除：盤面から
//=========================================================
void GameManager::RemoveAllCard_fromBoard  () {

	///// 盤面に対して、可能な指し手リストが最新ではなくなる。
	_IsSelectableMoveListNew = false;

	///// 全グリッド取得するぜぇぇぇ！ もはや躊躇は不要！
	vector< vector<Grid*> > allGridList = _board->GetAllGrid();

	for ( unsigned int y = 0; y < allGridList.size(); ++y ) {

		for ( unsigned int x = 0; x < allGridList[y].size(); ++x ) {


			Card *card = allGridList[y][x]->GetCard();

			if ( card != NULL ) {

				CardFactory::Release ( card );
				allGridList[y][x]->SetCard ( NULL );
			}
		}
	}

}

//=========================================================
// 局面を再生
//=========================================================
void GameManager::RearrangeAllPiecesAndCards ( AI_Modules::InnerExpression *innerExpression ) {

	///// まずは全部片付けて
	RemoveAllPiece_fromBox   ( "P1" );
	RemoveAllPiece_fromBox   ( "P2" );
	RemoveAllPiece_fromBoard ();
	ReleaseAllPiece();
	RemoveAllCard_fromBox    ( "P1" );
	RemoveAllCard_fromBox    ( "P2" );
	RemoveAllCard_fromBoard  ();

	///// 駒台
	for (int i = 0; i < innerExpression->_handPiece_BLACK; ++i) {
		AddPiece ( "BLACK", "P1" );
	}
	for (int i = 0; i < innerExpression->_handPiece_WHITE; ++i) {
		AddPiece ( "WHITE", "P2" );
	}

	///// 手札
	for (int i = 0; i < innerExpression->_handCards_BLACK_Index; ++i) {

		int cardId = innerExpression->_handCards_BLACK[i];
		string cardName = CardFactory::GetCardProperty(cardId).cardName;
		AddCard ( cardName, "UP", "P1" );
	}

	for (int i = 0; i < innerExpression->_handCards_WHITE_Index; ++i) {

		int cardId = innerExpression->_handCards_WHITE[i];
		string cardName = CardFactory::GetCardProperty(cardId).cardName;
		AddCard ( cardName, "UP", "P2" );
	}

	///// 盤上
	for ( int y = 0; y < innerExpression->_boardSizeY; ++y ) {
		for ( int x = 0; x < innerExpression->_boardSizeX; ++x ) {

			///// 駒
			int pieceOnGrid = innerExpression->_gridList_Piece[y][x];

			if (pieceOnGrid == 1) {
				AddPiece ( "BLACK", x, y );
			}
			else if (pieceOnGrid == 2) {
				AddPiece ( "WHITE", x, y );
			}

			///// カード
			AI_Modules::InnerExpression::CardOnBoard cardOnBoard = innerExpression->_gridList_Card[y][x];

			if (cardOnBoard.cardId != -1 ) {

				string direction = "";

				if      (cardOnBoard.direction == AI_Modules::DIRECTION::UP    ) { direction = "UP"; }
				else if (cardOnBoard.direction == AI_Modules::DIRECTION::DOWN  ) { direction = "DOWN"; }
				else if (cardOnBoard.direction == AI_Modules::DIRECTION::LEFT  ) { direction = "LEFT"; }
				else if (cardOnBoard.direction == AI_Modules::DIRECTION::RIGHT ) { direction = "RIGHT"; }

				AddCard (
					CardFactory::GetCardProperty(cardOnBoard.cardId).cardName,
					direction,
					x,y
				);

			}
		}
	}


	///// 目的地へのびよーんを省略いたします
	for ( Piece *piece : _pieceList_forManagement) {

		piece->WarpToGoal();

	}

	for ( vector<Grid*> row : _board->GetAllGrid() ) {
		for ( Grid * grid : row ) {

			Card *card = grid->GetCard();

			if ( card != NULL ) {
				card->WarpToGoal();
			}
		}
	}

	for ( Card *card : _cardBox_P1->GetAllCards() ) {
		card->WarpToGoal();
	}
	for ( Card *card : _cardBox_P2->GetAllCards() ) {
		card->WarpToGoal();
	}



	_isRendered = false;
}


//=========================================================
// 自治更新処理モードをセット
//=========================================================
void GameManager::SetFreeOperation ( bool arg ) {

	_isFreeOperation = arg;

	///// カード置き場
	if ( _cardBox_P1 != NULL ) {
		_cardBox_P1->SetAutonomous ( arg );
	}
	if ( _cardBox_P2 != NULL ) {
		_cardBox_P2->SetAutonomous ( arg );
	}

	///// 駒台
	if ( _pieceBox_P1 != NULL ) {
		_pieceBox_P1->SetAutonomous ( arg );
	}
	if ( _pieceBox_P2 != NULL ) {
		_pieceBox_P2->SetAutonomous ( arg );
	}

	///// グリッド
	if ( _board != NULL ) {

		vector<vector<Grid*>> allGridList = _board->GetAllGrid();

		for ( unsigned int y = 0; y < allGridList.size(); ++y ) {
			for ( unsigned int x = 0; x < allGridList[y].size(); ++x ) {
				allGridList[y][x]->SetAutonomous( arg );
			}
		}
	}

	///// カード
	vector<Card*> allCardList = GetAllCards();
	for ( unsigned int i = 0; i < allCardList.size(); ++i ) {

		allCardList[i]->SetAutonomous ( arg );
	}

}
//=========================================================
// 自治更新処理モード？
//=========================================================
bool GameManager::IsFreeOperation  () {
	return _isFreeOperation;
}

//=========================================================
// 全カードを取得：盤面＆両カード置き場
//=========================================================
vector<Card*> GameManager::GetAllCards() {

	vector<Card*> result;

	///// P1のカード置き場から
	if ( _cardBox_P1 != NULL ) {

		vector<Card*> list = _cardBox_P1->GetAllCards();
		result.insert ( result.end(),  list.begin(), list.end() ) ;
	}
	///// P2のカード置き場から
	if ( _cardBox_P2 != NULL ) {

		vector<Card*> list = _cardBox_P2->GetAllCards();
		result.insert ( result.end(),  list.begin(), list.end() ) ;
	}
	///// 盤面から
	vector < vector <Grid*> > allGrids = _board->GetAllGrid();
	for ( unsigned int y = 0; y < allGrids.size(); ++y ) {
		for ( unsigned int x = 0; x < allGrids[y].size(); ++x ) {

			if ( allGrids[y][x]->GetCard() != NULL) {
				result.push_back ( allGrids[y][x]->GetCard() );
			}
		}
	}

	return result;
}

//=========================================================
// ルールにそった操作を受付
//=========================================================
void GameManager::Operation_With_Rule () {

	///// 着手エフェクトが出ている間は、何もしない。
	if (_is_MoveEffectGoing) {
		return;
	}

	///// フリー操作中なら、何もしませんよ。
	if ( _isFreeOperation ) {
		return;
	}

	///// 操作がロックされているときにも、何もしませんよ？
	if ( _operationData.isLock ) {
		return;
	}

	///// 現在の手番の操作者は誰？
	/*
	string currentOperator = "";
	if      ( _operationData.currentTurn == "FIRST"  && _operationData.first  == "P1") { currentOperator = _operationData.operatorP1; }
	else if ( _operationData.currentTurn == "FIRST"  && _operationData.first  == "P2") { currentOperator = _operationData.operatorP2; }
	else if ( _operationData.currentTurn == "SECOND" && _operationData.second == "P1") { currentOperator = _operationData.operatorP1; }
	else if ( _operationData.currentTurn == "SECOND" && _operationData.second == "P2") { currentOperator = _operationData.operatorP2; }
	*/

	Participant currentTurnParticipant = GetCurrentTurnParticipant();


	/*
		ここで、現在の手番の参加者である間の経過フレーム数をカウントしますか。
		手番になった直後のリアクション問い合わせ専用かもしれない。
		いつまでも着手が決まらない場合のリアクション、なんかも求めることができたりするのかもしれない。
	*/
	//if ( currentTurnParticipant.turnFrameCount < 0 ) {
	if ( currentTurnParticipant.turnFrameCount == -1 ) {

		vector<Actor*> actors = GetAllActors();
		for ( unsigned int i = 0; i < actors.size(); ++i ) {

			Actor *a = actors[i];
			a->Notice_TurnStart ( currentTurnParticipant.id );
		}

		// 「直後」ではなくなる、と。
		//currentTurnParticipant.turnFrameCount = 0; // これだと反映されない。
		CurrentTurnParticipant_TurnFrameCountStart();

		// チーム的な手番開始処理も。
		TeamNewTurnStart();
	}

	// 現在の手番の参加者の経過フレーム数を進める（リストの中身なので、更新系は直接触れない）
	CurrentTurnParticipant_ProgressTurnFrameCount();




	///// 人間の操作を受け付ける（現在の手番の参加者にAIオブジェクトが存在するかどうかで、）
	//if ( currentOperator == "HUMAN") {

	bool isHumanTurn = (
		// AI未設定
		currentTurnParticipant.ai == NULL
		||
		// 口出し中
		(_interferenceModeInfo.isActive && _interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE)
	);

	// 千日手の動作テストのため、両者ともプレイヤーの操作としておく
	//isHumanTurn = true;

	if ( isHumanTurn ) {

		///// 現在の局面で可能な指し手が最新じゃないなら作成する
		MakeSelectableList_IfNeccesary();

		///// 気力ゲージ
		if ( Application::GetInstance()->_mouseState_Left == MOUSE::BUTTON::STATE::PRESS ) {
			_psychoGage->Press();
		}

		///// 現在の状態ごとの操作を行う。
		if      ( _operationState == OperationState::TO_DETERMIN_WHAT_TO_DO ) { Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO (); }
		else if ( _operationState == OperationState::TO_PUT_PIECE           ) { Operation_With_Rule_TO_PUT_PIECE           (); }
		else if ( _operationState == OperationState::TO_MOVE_PIECE          ) { Operation_With_Rule_TO_MOVE_PIECE          (); }
		else if ( _operationState == OperationState::TO_PUT_CARD            ) { Operation_With_Rule_TO_PUT_CARD            (); }

	///// AI の操作を受け付ける
	//} else if ( currentOperator == "AI" ) {
	} else {


		///// 進行状況はいかがですか？
		AI_Modules::ProgressData progressData = _AI_Manager->GetThinkProgress();


		///// 思考スレッドが立っていなければ、立てる
		if ( progressData.progressRate < 0.0f) {

			// プログレスバーをクリア
			_AI_ProgressBar->SetValue_Forsibly(0);
			_AI_ProgressBar->SetVanishCountDown(-1);

			// 現在の局面を「出題」
			AI_Modules::InnerExpression *probremState = GenerateInnerExpressionFromCurrentState ();

			// 千日手回避のため、これまでの局面リスト（棋譜）をAIに渡しておく
			if (_score != NULL) {
				currentTurnParticipant.ai->SetRecordList( _score->GetInnerExpressionsList() );
			}

			// 思考開始
			_AI_Manager->ThinkStart ( currentTurnParticipant.ai, probremState );

			// 「問題」は、保存しなくていい。
			delete probremState;



		///// 思考スレッドを立てた後なら、結果または進行状況を見る
		} else {

			///// 思考中？
			if ( progressData.progressRate < 100.0f) {

				// プログレスバーを更新
				float progressRate = progressData.progressRate;

			///// 思考完了してる？
			} else if ( progressData.progressRate == 100.0f) {

				// プログレスバーを 100に。
				_AI_ProgressBar->SetValue_Forsibly(100);
				_AI_ProgressBar->SetVanishCountDown(60);


				// 着手実行を待機
				_progressData_AfterThinking = progressData;

				// 次回にむけてクリア
				//_AI_Manager->Clear();

			///// 思考完了後。
			} else {

				// 処理なし（着手完了を待機している状態）
			}
		}
	}
}

//=========================================================
// 着手エフェクトオブジェクトに渡す終了通知受け取りメソッド
//=========================================================
void GameManager::EndMoveFromEffect ( AI_Modules::Move move ) {

	printf ("着手エフェクトオブジェクトからの終了通知。\n");

	// エフェクト実行中フラグを降ろす。
	_is_MoveEffectGoing = false;

	Operation_AfterExecute_Process ( move );

}

//=========================================================
// AIは思考を完了している？
//=========================================================
bool GameManager::IsAIThinkFinished() {

	bool result = false;

	/*
		AIの手番で、思考が完了しているなら、true
	*/
	if ( Is_AI_Turn_Now() ) {

		if ( _progressData_AfterThinking.progressRate == 100.0f) {
			result = true;
		}
	}


	return result;
}

//=========================================================
// AIさん、着手を実行してね
// カットインあり
//=========================================================
void GameManager::ExecuteAIMoveWithCutIn ( IEffect *cutInEffect ) {

	///// 着手している最中ですよ。
	_is_MoveEffectGoing = true;

	///// カットインエフェクトを開始：カットインの中身は呼び出し元で設定しておいてね。
	Application::GetInstance()->_effectManager->Add (cutInEffect);
}


//=========================================================
// AIさん、着手を実行してね：エフェクトレベルを指定（0〜100値。内部で、用意しているエフェクト数に換算）
// カットインなし
//=========================================================
void GameManager::ExecuteAIMove ( int effectLevel ) {

	// AIが思考を完了していないなら、何もしない
	if ( !IsAIThinkFinished() ) {
		printf ("Warning : AIが思考を完了していないので何もしない。\n");
	}

	///// エフェクトレベルを 0〜3 に換算
	if ( effectLevel > 0) {

		if      (  1 <= effectLevel && effectLevel <= 30 ) { effectLevel = 1;}
		else if ( 31 <= effectLevel && effectLevel <= 60 ) { effectLevel = 2;}
		else if ( 61 <= effectLevel                      ) { effectLevel = 3;}

	} else {
		effectLevel = 0;
	}

	///// 候補手からの選択モードで必要なら、プレイヤーが選択したものに置き換える
	_progressData_AfterThinking = ReplaceToPlayersChoiceIfNeeded(_progressData_AfterThinking);


	///// 試行結果として保持しておいたものを使いますよ。
	AI_Modules::ProgressData progressData = _progressData_AfterThinking;

	///// 指し手に対する総括評価値を退避。
	_evaluationSumUp_FromAI = _progressData_AfterThinking.evaluationSumUp;

	///// 思考結果 を クリアしておく：これで「思考完了しているので着手せよ」が無効になる。
	_progressData_AfterThinking = AI_Modules::ProgressData();

	///// エフェクトつきの着手。駒設置。
	if ( progressData.answerMove.action == AI_Modules::MOVE::ACTION::PUT_PIECE	) {

		IEffect *effect = new Effect_PutPiece_AI (
				_endMoveFromEffect,  progressData.answerMove,

				effectLevel, //0, // エフェクトレベルを何で決めるか？ 評価値を元に「気合」を求める的な何か。

				GetPieceBox_CurrentTurn()->TakePiece(), // もう駒台から駒を１つ取ってしまう。
				_board->GetGrid ( progressData.answerMove.putPieceX, progressData.answerMove.putPieceY )
			);
		Application::GetInstance()->_effectManager->Add ( effect );

		// 着手している最中ですよ。
		_is_MoveEffectGoing = true;


	///// エフェクトつきの着手。カード設置。
	} else if ( progressData.answerMove.action == AI_Modules::MOVE::ACTION::PUT_CARD	) {

		IEffect *effect = new Effect_PutCard_AI (
				_endMoveFromEffect,  progressData.answerMove,

				effectLevel, //3, // エフェクトレベルを何で決めるか？ 評価値を元に「気合」を求める的な何か。

				GetCardBox_CurrentTurn()->TakeCard ( progressData.answerMove.putCardId ),
				_board->GetGrid ( progressData.answerMove.putCardX, progressData.answerMove.putCardY )
			);
		Application::GetInstance()->_effectManager->Add ( effect );

		// 着手している最中ですよ。
		_is_MoveEffectGoing = true;

	///// エフェクトつきの着手。カード取得。
	} else if ( progressData.answerMove.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

		///// カード取得エフェクト
		IEffect *effect = new Effect_TakeCard (
			_endMoveFromEffect,
			progressData.answerMove,
			effectLevel, //2, // エフェクトレベルを何で決めるか？ 評価値を元に「気合」を求める的な何か。
			_board->GetGrid ( progressData.answerMove.takeCardX, progressData.answerMove.takeCardY ),
			_board->GetGrid ( progressData.answerMove.takeCardX, progressData.answerMove.takeCardY )->GetCard(),
			_board->GetGrid ( progressData.answerMove.takeCardX, progressData.answerMove.takeCardY )->GetPiece(),
			GetCardBox_CurrentTurn()
		);
		Application::GetInstance()->_effectManager->Add ( effect );

		// 着手している最中ですよ。
		_is_MoveEffectGoing = true;

	///// エフェクトつきの着手。駒移動・ぶつけない
	} else if ( progressData.answerMove.action == AI_Modules::MOVE::ACTION::MOVE_PIECE && !progressData.answerMove.movePieceAndIsCllide) {

		///// 駒移動エフェクト
		IEffect *effect = new Effect_MovePiece (
			_endMoveFromEffect,
			progressData.answerMove,
			effectLevel, //0, // エフェクトレベルを何で決めるか？ 評価値を元に「気合」を求める的な何か。
			_board->GetGrid ( progressData.answerMove.movePieceFromX, progressData.answerMove.movePieceFromY ),
			_board->GetGrid ( progressData.answerMove.movePieceToX,   progressData.answerMove.movePieceToY ),
			_board->GetGrid ( progressData.answerMove.movePieceFromX, progressData.answerMove.movePieceFromY )->GetPiece()
		);
		Application::GetInstance()->_effectManager->Add ( effect );

		// 着手している最中ですよ。
		_is_MoveEffectGoing = true;

	///// エフェクトつきの着手。駒移動・ぶつける
	} else if ( progressData.answerMove.action == AI_Modules::MOVE::ACTION::MOVE_PIECE && progressData.answerMove.movePieceAndIsCllide) {

		Grid *collidedGrid = GetCollidedGrid (
			progressData.answerMove.movePieceFromX, progressData.answerMove.movePieceFromY,
			progressData.answerMove.movePieceToX,   progressData.answerMove.movePieceToY
		);

		///// 駒移動エフェクト
		IEffect *effect = new Effect_MovePiece_Collide (
			_endMoveFromEffect,
			progressData.answerMove,
			effectLevel, //1, // エフェクトレベルを何で決めるか？ 評価値を元に「気合」を求める的な何か。
			_board->GetGrid ( progressData.answerMove.movePieceFromX, progressData.answerMove.movePieceFromY ),
			_board->GetGrid ( progressData.answerMove.movePieceToX,   progressData.answerMove.movePieceToY ),
			_board->GetGrid ( progressData.answerMove.movePieceFromX, progressData.answerMove.movePieceFromY )->GetPiece(),
			collidedGrid->GetPiece(),
			GetPieceBox_BelongToPiece ( collidedGrid->GetPiece() ),
			collidedGrid
		);
		Application::GetInstance()->_effectManager->Add ( effect );

		// 着手している最中ですよ。
		_is_MoveEffectGoing = true;

	///// エフェクト未対応
	} else {

		/*
			ここにはこないはず
		*/

		printf ("あれ？ エフェクト未対応のAI着手あった？\n");

		// 指し手を実行。
		Move (progressData.answerMove);

		// 事後処理：手番交代とか棋譜とか、そのへん。
		Operation_AfterExecute_Process ( progressData.answerMove );

	}
}

//=========================================================
// 現在の局面で可能な指し手が最新じゃないなら作成する
//=========================================================
void GameManager::MakeSelectableList_IfNeccesary () {

	if ( !_IsSelectableMoveListNew ) {

		AI_Modules::InnerExpression *ie = GenerateInnerExpressionFromCurrentState ();

		_selectableMoveList.clear();
		ie->GetLegalNextMoveList ( _selectableMoveList );

		delete ie;

		_IsSelectableMoveListNew = true;
	}
}

//=========================================================
// ルールにそって、指し手を実行したあとの事後処理
//=========================================================
void GameManager::Operation_AfterExecute_Process ( AI_Modules::Move move ) {
    
    printf ("指し手実行後の事後処理：GameManager::Operation_AfterExecute_Process()\n");
    
    string moveString = MoveToString(move);
    printf ( "%s\n", moveString.c_str() );
    
    // スクリプトでの指し手だったかどうかは知る由もないけれど、ともかくフラグを下ろしておく。
    _isScriptingMoving = false;
    
	/*
		AIの場合も人間の場合も、
		エフェクトレベル０の場合も、
		着手後は必ずここにくる（はず）
	*/

	// 検討モードのときには何もしない（次へ押下で棋譜に書き込まれてしまうのを防ぐ意味）
	if ( _isAstralProjection ) {
		return;
	}
	
	// 試合中じゃない場合も何もしない
	if ( !_battleInfo.isBattlingNow && !_isTutorial ) {
        printf("試合中じゃないので何もしない\n");

        return;
	}

	// 棋譜に記録する
	GameRecord::Record record;
	record.move = move;                                                 // こう指して
	record.innerExpression = GenerateInnerExpressionFromCurrentState(); // こうなりました

	///// 着手が行われた、というイベントを アクターに知らせる
	/*
		自分の手番だった場合に本人にだけ知らせる？
		それとも、誰の手番でも関係なく全員に知らせる？
		とりあえず、全員とする。
	*/
	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {
		if ( _participantsList[i].actor != NULL ) {

			// 局面。_actor->Notice_MoveIsDone() 内では、参照するだけ
			AI_Modules::InnerExpression *ie = record.innerExpression;

			_participantsList[i].actor->Notice_MoveIsDone ( _operationData.currentTurn, move, ie );
		}
	}

	// どっちサイドの番なのかを記録したい、ということ
	//if      ( _operationData.currentTurn == "FIRST"  ) { record.side = _operationData.first;}
	//else if ( _operationData.currentTurn == "SECOND" ) { record.side = _operationData.second;}
	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	record.side = currentTurnParticipant.side;


	// AIの手番なら、思考結果に入っている自己評価値を入れる。
	if ( Is_AI_Turn_Now () ) {

		// _progressData_AfterThinking.evaluationSumUp を使いたいところなのですが、
		// 着手エフェクトの前にクリアしているので、退避しておいたものを使う。_evaluationSumUp_FromAI
		// 着手エフェクトを経由して、エフェクトオブジェクトからここにきている。
		// 着手エフェクトにこの値を渡すのが正統な気がするけど、エフェクトごとに別クラスになってて、
		// 全部のクラスにインターフェイスを追加するのはめんどくさいので、このようにするものなり。

		// 口出しモードで口出しをした場合は「口出しされた」か。
		if ( _interferenceModeInfo.isActive ) {
			if ( _interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE ) {
				_evaluationSumUp_FromAI = "INTERFERED";
			}
		}

		//現在の手番の参加者の side で記録を分ける
		if ( currentTurnParticipant.side == "P1" ) {
			record.evaluationSumUp_From_P1 = _evaluationSumUp_FromAI;
		}
		else if ( currentTurnParticipant.side == "P2" ) {
			record.evaluationSumUp_From_P2 = _evaluationSumUp_FromAI;
		}
	}

	///// 試合中ではない、という可能性がある。その場合、中身に深い意味はないはず、ということもあり、適当に生成してしまう。
	//if ( _score == NULL ) {
	//	_score = new GameRecord ( GenerateInnerExpressionFromCurrentState() );
	//}
	// ↑これをやると、チュートリアルで盤面を操作した途端に棋譜が出現する。してしまう。
	// ↓ゆえに、こうする。試合開始時（BattleStart()) で生成している。

	///// 試合中ではないなら棋譜はつけない。
	//if (_score != NULL) {
	//	_score->Add ( record );
	//}
	// ただし、試合後もBattleEnd で棋譜を削除しているわけではないので、試合後も　_score　!= NULL となる
	// つまり検討モードということ。検討モードについては別判定でここへこないようにしている。

	// ↑と、思ったけれど、チュートリアルにて、着手を待ち受けるイベントで着手を取得できなくなってしまった。
	// つまりチュートリアルは試合ではないから棋譜を持っていない、のだけれど、
	// 着手イベント待ち受けのところで棋譜をアテにして「着手されたかな？」とワクワクしている。
	//
	// 道は２通り。
	// 試合中ではなくとも棋譜をつける、または、最終着手だけは棋譜オブジェクトとは別個に持っておく
	//

	///// そういうわけで、一応 NULL 判定はしておくのですが、NULLになることはないはず。
	// 「試合中かどうか」という意味合いの判定ではない、ということ。
	if (_score != NULL) {
		_score->Add ( record );
	}





	///// 口出しモード的にも着手完了通知をもらうよ（手番交代前に）
	InterferenceMode_ModeDone();

	// 手番交代前：現手番の参加者の口出しモードを保存
	WriteInterferenceModeToParticipant ( GetCurrentTurnParticipant().id );

	// 手番交代前：現手番の参加者の候補手からの選択モードを保存
	WriteChoiceModeToParticipant ( GetCurrentTurnParticipant().id );

	// ここらで手番のフレーム経過数をリセット
	CurrentTurnParticipant_ResetTurnFrameCount();


	// 手番交代
	//_currentTurn = AI_Modules::AI::SwitchTurn ( _currentTurn );
	//_operationData.SwitchTurn();
	_operationData.TurnToNext();

	// 手番交代後：新手番の候補手からの選択モードを展開
	DeployChoiceModeFromParticipant ( GetCurrentTurnParticipant().id );

	// 手番交代後の、候補手からの選択モードの進行処理（新手番の人が対象者ならば云々……）
	ChoiceMode_NewTurnStarts();

	// 手番交代後：新手番の参加者の口出しモードを展開
	DeployInterferenceModeFromParticipant ( GetCurrentTurnParticipant().id );

	// 手番交代後の、口出しモードの進行処理（新手番の人が口出され対象ならば云々……）
	InterferenceMode_NewTurnStarts();

    printf("思考前ですよ\n");
	// 思考前ですよ。
	/*
		口出し着手をした場合に、新手番の参加者が口出され対象じゃないAIの場合、
		ここを0にしておかないと「思考済み」とみなされ、口出しで却下したAIの思考結果が新手番の人に使われることになってしまう。
		いずれにせよ、このタイミングでクリアしておいて間違いではない値のハズ。
	*/
	_progressData_AfterThinking.progressRate = 0.0f;

	// 操作の状態を「まだ何をするか決めていない」にする
	_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

	// 可能な指し手リストは最新じゃなくなっている。
	_IsSelectableMoveListNew = false;

	// グリッドのフラッシュを消しておく
	GridsFlashCancel();

	// ゲージの底値固定を解除
	_psychoGage->SetMin_Clear();
	_psychoGage->ClearValue(); // 値もクリア：普通に減っていくと、相手の手番の間、意欲が萎えているみたいでイヤ。

	// 試合中なら、勝敗判定をする
	if ( _battleInfo.isBattlingNow ) {
		Judgement();
	}

	// 勝敗が出ていないならば、次の番（さきほど交代した新手番）の名前にする
	if (_battleInfo.result == "") {
		RenewCurentTurnNamePanel();
	}

	// 次回にむけてクリア：AIの番じゃないときに実行しても副作用はない、はず。
	// ここ（＝着手エフェクトの後）で実行しないと、着手エフェクト終了待ちの間に進捗が０に戻ってしまう。
	_AI_Manager->Clear();
	
	
	// 実質チュートリアル用の「試合じゃないけど終了処理をする例外ですよフラグ」は毎回下ろすこととする。
	_isTutorial = false;
	
    printf("終了：GameManager::Operation_AfterExecute_Process\n");
}

//=========================================================
// 現在の手番表示パネルに現在の手番の名前を設定
//=========================================================
void GameManager::RenewCurentTurnNamePanel () {

	string currentTurnName = "";

	// プレイヤーかどうかの判定 ID "player" ならプレイヤー、というファンキーな判定。
	if ( GetCurrentTurnParticipant().ai == NULL || GetCurrentTurnParticipant().id == "player" ) {

		currentTurnName = "プレイヤー";

		// 設定ファイルで指定されている名前
		string designatedName = GetCurrentTurnParticipant().name;
		TextMessageWindow::ReplaceWithAlias ( designatedName );

		// 「主人公」として登録されているのとは違う名前だった、それにする。「キャラ２」とか。
		bool isProtagonist = false;
		vector<string> protagonists = TextMessageWindow::GetProtagonistsNames();
		for ( unsigned int i = 0; i < protagonists.size(); ++i ) {

			if ( protagonists[i] == designatedName ) {
				isProtagonist = true;
				break;
			}
		}

		// 「主人公」ではなく、空文字でもないなら、指定の名前とする
		if ( !isProtagonist && designatedName != "" ) {
			currentTurnName = GetCurrentTurnParticipant().name;
		}


	} else {

		currentTurnName =  GetCurrentTurnParticipant().name;
	}

	// ここでキャラ名前エイリアスを通しているのですね
	TextMessageWindow::ReplaceWithAlias(currentTurnName);

	_currentTurnNamePanel->SetName ( currentTurnName );
}

//=========================================================
// 棋譜を表示する？ しない？
//=========================================================
void GameManager::ShowScoreSheetViewer() {
	if ( _score != NULL ) {
		_score->Show();
	}
}
void GameManager::HideScoreSheetViewer() {
	if ( _score != NULL ) {
		_score->Hide();
	}
}


//=========================================================
// グリッドのフラッシュを消す
//=========================================================
void GameManager::GridsFlashCancel() {

	vector< vector<Grid*> > allGridList = _board->GetAllGrid();
	for ( unsigned int y = 0; y < allGridList.size(); ++y ) {
		for ( unsigned int x = 0; x < allGridList[y].size(); ++x ) {

			allGridList[y][x]->SetSelectable (false);
			allGridList[y][x]->SetSelected   (false);
		}
	}
}


//=========================================================
// ルールにそった操作を受付：まだ操作を決めていない
//=========================================================
void GameManager::Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO () {

	///// 着手エフェクトが出ている間は、操作を受け付けない
	if ( _is_MoveEffectGoing ) {
		return;
	}

	///// クリックしてないときは、操作を受け付けない
	if ( Application::GetInstance()->_mouseState_Left != MOUSE::BUTTON::STATE::RELEASE) {
		return;
	}

	///// 駒設置かも。駒移動かも。カード設置かも。（カード取得判定は駒移動判定内で行っている）
	if ( Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_PUT_PIECE  () ) { return; }
	if ( Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_MOVE_PIECE () ) { return; }
	if ( Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_PUT_CARD   () ) { return; }
}

//=========================================================
// ルールにそった操作を受付：まだ操作を決めていない：駒設置、かもしれない。
//=========================================================
bool GameManager::Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_PUT_PIECE  () {

	/*
		駒台の駒を選んでいるかどうかを判定。
	*/

	bool result = false;

	///// 対象の駒台
	PieceBox *pieceBox = NULL;

	///// 手番で切り替え：どっちサイドの駒台を使うか、ということ
	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	if      ( currentTurnParticipant.side == "P1" )  { pieceBox = _pieceBox_P1; }
	else if ( currentTurnParticipant.side == "P2" )  { pieceBox = _pieceBox_P2; }
	else { printf ( "Error：どっちサイドでもないので駒台取得できず（currentTurnParticipant.side = %s）。\n", currentTurnParticipant.side.c_str() ); }

	/*
	if      ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P1") { pieceBox = _pieceBox_P1; }
	else if ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P2") { pieceBox = _pieceBox_P2; }
	else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P1") { pieceBox = _pieceBox_P1; }
	else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P2") { pieceBox = _pieceBox_P2; }
	*/

	///// 対象の駒
	Piece *selectedPiece = NULL;

	///// 駒台の駒をクリック？
	vector<Piece*> pieceList = pieceBox->GetPieceList();
	for ( unsigned int i = 0; i < pieceList.size(); ++i ) {

		if ( pieceList[i]->IsMouseOver() ) {

			selectedPiece = pieceList[i];

			// 駒台から、はずしておく
			pieceBox->TakePiece ( selectedPiece );

			break;
		}
	}

	///// 駒台そのものをクリック？
	if ( selectedPiece == NULL && pieceBox->IsMouseOver() ) {

		// 駒をひとついただく。：ここで駒がないならNULLになる。以降の処理と矛盾しない。
		selectedPiece = pieceBox->TakePiece();
	}

	///// 選択している駒があった？
	if ( selectedPiece != NULL ) {

		result = true;

		// ここで、エフェクトを出すためのモニタリングオブジェクト？ を生成！
		IEffect *moveEffect = new Effect_PutPiece_Human ( _endMoveFromEffect, selectedPiece, pieceBox, _board, GetTurn_in_tarms_of_TURN_FOR_AI(), _psychoGage );
		Application::GetInstance()->_effectManager->Add ( moveEffect );

		// 操作用エフェクトを出したことを覚えておく → ロード（＝この盤面を強制終了）時の後片付け目的
		LetsRemenberThatIMadeThisEffect(moveEffect);

		// ここでセーブされちゃった場合のためのメモ → つまり駒を生成しなおして、駒台に戻すために必要な情報、ということ。
		if ( pieceBox == _pieceBox_P1 ) { _operationMemo_belongsToWhichPieceBox = "P1"; }
		if ( pieceBox == _pieceBox_P2 ) { _operationMemo_belongsToWhichPieceBox = "P2"; }
		_operationMemo_whatColorThePieceWas = selectedPiece->GetColor();

		// マウスにくっつける。
		IMouseClingableObject *previousOne = Application::GetInstance()->AddMouseClingObj ( selectedPiece );

		if ( previousOne != NULL ) {
			printf ( "[Error] マウスに先客がいました。どうしましょうどうしましょう。オロオロ。\n" );
		}

		// 設置可能なグリッドをフラッシュさせる
		FlashSelectableGrids_ToPutPiece ();

		// 操作の状態を「駒の設置先を模索中」に切り替える
		_operationState = OperationState::TO_PUT_PIECE;

		// ゲージの底値を現在値に固定
		_psychoGage->SetMin_CurrentValue();
	}

	return result;
}

//=========================================================
// ルールにそった操作を受付：まだ操作を決めていない：駒移動、かもしれない
//=========================================================
bool GameManager::Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_MOVE_PIECE  () {

	/*
		盤面上の駒を選んでいるかどうかを判定。

		えっと、何をするんでしたっけ？
		すべきことを整理しよう！

		手番の色の、盤上にある駒にマウスオーバーしている。
		いや、グリッドに、の方がいいかも？
		うん、そうしましょう。

		そのとき、マウスにその駒をくっつけて、
		可能な選択肢を取得して（移動、ぶつける、ぶつけない、カード取得）
		選択できる場所を光らせて、
		操作の状態を「移動先模索中」に切り替える。

	*/

	bool result = false;
	Grid *selectingGrid = NULL;

	///// どこかを選んだ？
	int selectedX = -1;
	int selectedY = -1;

	///// 選択可能な指し手リストを信じて、ループする。
	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i) {

		/*
			つまり、自分の駒があって、通行可能またはカードに乗っている位置
			が選択可能、ということになるはず。
		*/

		///// 駒移動に対して
		if ( _selectableMoveList[i].action == AI_Modules::MOVE::ACTION::MOVE_PIECE ) {

			// そこにグリッドがあると信じます。
			if ( _board->GetGrid ( _selectableMoveList[i].movePieceFromX, _selectableMoveList[i].movePieceFromY )->IsMouseOver()) {
				selectedX = _selectableMoveList[i].movePieceFromX;
				selectedY = _selectableMoveList[i].movePieceFromY;

				selectingGrid = _board->GetGrid ( _selectableMoveList[i].movePieceFromX, _selectableMoveList[i].movePieceFromY );

				break;
			}
		}

		///// カード取得
		if ( _selectableMoveList[i].action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

			// そこにグリッドがあると信じます。
			if ( _board->GetGrid ( _selectableMoveList[i].takeCardX, _selectableMoveList[i].takeCardY )->IsMouseOver()) {
				selectedX = _selectableMoveList[i].takeCardX;
				selectedY = _selectableMoveList[i].takeCardY;
				selectingGrid = _board->GetGrid ( _selectableMoveList[i].takeCardX, _selectableMoveList[i].takeCardY );


				break;
			}
		}

	}

	///// どこか選んだ？
	if ( selectingGrid != NULL ) {

		result = true;
		_selectingGrid = selectingGrid;

		// セーブ/ロード用のメモ情報
		_selectingGridX = selectedX;
		_selectingGridY = selectedY;

		///// その位置を光らせる
		FlashSelectableGrids_ToMovePiece ( selectedX, selectedY );

		// 操作の状態を「駒の移動先を模索中」に切り替える
		_operationState = OperationState::TO_MOVE_PIECE;


		///// ここで「システム音」という意味で、選択音を出す
		GUI_SFML::SoundPlayer::Play ( "./sound/se_good_pk01/on01.wav" );


		///// 意志ゲージの底値を現在値に固定
		_psychoGage->SetMin_CurrentValue();
	}


	return result;
}

//=========================================================
// ルールにそった操作を受付：まだ操作を決めていない：カード設置、かもしれない。
//=========================================================
bool GameManager::Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_PUT_CARD  () {

	/*
		盤面上のカードを選んでいるかどうかを判定。
	*/

	bool result = false;

	///// 対象となるカード置き場
	CardBox *cardBox = NULL;

	///// 手番で切り替え：どっちサイドのカード置き場を使うかということ
	Participant currentTurnParticipant = GetCurrentTurnParticipant();

	if      ( currentTurnParticipant.side == "P1" )  { cardBox = _cardBox_P1; }
	else if ( currentTurnParticipant.side == "P2" )  { cardBox = _cardBox_P2; }
	else { printf ( "Error：どっちサイドでもないのでカード置き場取得できず（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }

	/*
	if      ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P1") { cardBox = _cardBox_P1; }
	else if ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P2") { cardBox = _cardBox_P2; }
	else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P1") { cardBox = _cardBox_P1; }
	else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P2") { cardBox = _cardBox_P2; }
	*/
	///// カード置き場のカードを取得：このリストを後ろから見て、先にマウスオーバーしてる子を見れば、Z値の問題はないはず。
	vector<Card*> cardList = cardBox->GetAllCards();

	///// カードを選ぶかもしれないのです。
	Card *selectedCard = NULL;

	for ( int i = cardList.size() - 1; i >= 0; --i ) {

		if ( cardList[i]->IsMouseOver() ) {
			selectedCard = cardList[i];
			break;
		}

	}

	///// カードを選んだ？
	if ( selectedCard != NULL ) {

		result = true;

		// ここで、エフェクトを出すためのモニタリングオブジェクト？ を生成！
		IEffect *moveEffect = new Effect_PutCard_Human ( _endMoveFromEffect, selectedCard, cardBox, _board, GetTurn_in_tarms_of_TURN_FOR_AI(), _psychoGage );
		Application::GetInstance()->_effectManager->Add ( moveEffect );

		// 操作用エフェクトを出したことを覚えておく → ロード（＝この盤面を強制終了）時の後片付け目的
		LetsRemenberThatIMadeThisEffect(moveEffect);

		// ここでセーブ/ロードされちゃったときのために、元に戻すための情報を覚えておかなきゃいけないだなんて！
		if      ( cardBox == _cardBox_P1 ) { _operationMemo_belongsToWhichCardBox = "P1";}
		else if ( cardBox == _cardBox_P2 ) { _operationMemo_belongsToWhichCardBox = "P2";}
		_operationMemo_whichCardTheCardWas = selectedCard->GetProperty().cardName;

		///// マウスにくっつけて、カード置き場からははずしておく。
		Application::GetInstance()->AddMouseClingObj ( selectedCard );
		cardBox->TakeCard ( selectedCard );

		///// 設置先候補を光らせるには、カードID と 向き が必要なのです。
		FlashSelectableGrids_ToPutCard ( selectedCard->GetProperty().processId, selectedCard->GetCurrentDirection() );

		// 操作の状態を「カードの設置先を模索中」に切り替える
		_operationState = OperationState::TO_PUT_CARD;

		///// ゲージの底値を現在値に固定
		_psychoGage->SetMin_CurrentValue();
	}


	return result;
}


//=========================================================
// 指し手が同じかどうか
//=========================================================
bool GameManager::IsEqualMove ( AI_Modules::Move move1, AI_Modules::Move move2 ) {

	bool result = false;

	if ( move1.action == move2.action ) {

		if ( move1.action == AI_Modules::MOVE::ACTION::PUT_PIECE) {

			if
			(
				move1.putPieceX == move2.putPieceX &&
				move1.putPieceY == move2.putPieceY
			)
			{
				result = true;
			}
		}
		else if ( move1.action == AI_Modules::MOVE::ACTION::MOVE_PIECE) {

			if
			(
				move1.movePieceFromX                == move2.movePieceFromX &&
				move1.movePieceFromY                == move2.movePieceFromY &&
				move1.movePieceToX                  == move2.movePieceToX &&
				move1.movePieceToY                  == move2.movePieceToY &&
				move1.movePieceAndIsCllide          == move2.movePieceAndIsCllide &&
				move1.movePieceAndCllidedPieceColor == move2.movePieceAndCllidedPieceColor
			)
			{
				result = true;
			}
		}
		else if ( move1.action == AI_Modules::MOVE::ACTION::PUT_CARD) {

			if
			(
				move1.putCardX         == move2.putCardX &&
				move1.putCardY         == move2.putCardY &&
				move1.putCardDirection == move2.putCardDirection &&
				move1.putCardId        == move2.putCardId
			)
			{
				result = true;
			}
		}
		else if ( move1.action == AI_Modules::MOVE::ACTION::TAKE_CARD) {

			if
			(
				move1.takeCardCardId == move2.takeCardCardId &&
				move1.takeCardX      == move2.takeCardX &&
				move1.takeCardY      == move2.takeCardY
			)
			{
				result = true;
			}
		}

	}

	return result;
}

//=========================================================
// 千日手判定
// プレイヤーが着手しようとしているタイミングでの同一局面カウント
//
// つまり、たとえば駒を設置しようとしているなら、
// 駒はマウスにくっついているのであり、駒台にはない、というようなこと
//
//=========================================================
int GameManager::CountRepetition ( AI_Modules::Move attemptingMove ) {

	int occurenceCount = 0;

	// 現在の盤面
	AI_Modules::InnerExpression *currentState = GenerateInnerExpressionFromCurrentState();

	///// 指した状態の局面を作る
	//
	///// しようとしている動作：「駒設置」
	if ( attemptingMove.action == AI_Modules::MOVE::ACTION::PUT_PIECE ) {

		// 駒をマウスに持った状態なので、戻しておく
		if ( GetCurrentTurnParticipant().side == "P1") { currentState->_handPiece_BLACK += 1;}
		if ( GetCurrentTurnParticipant().side == "P2") { currentState->_handPiece_WHITE += 1;}

		// 指した状態とする
		currentState->GoNext(attemptingMove);


	///// しようとしている動作：「駒移動」
	} else if ( attemptingMove.action == AI_Modules::MOVE::ACTION::MOVE_PIECE ) {

		// 指した状態とする
		currentState->GoNext(attemptingMove);

	///// しようとしている動作：「カード設置」
	} else if ( attemptingMove.action == AI_Modules::MOVE::ACTION::PUT_CARD ) {


		///// マウスでカードを持っている状態のハズ
		if ( !Application::GetInstance()->IsMouseClingObj_Card() ) {
			printf ( "千日手判定 GameManager::CountRepetition マウスでカードを持っている状態ではないらしいですよ？\n" );
			return 0;
		}

		IMouseClingableObject *obj = Application::GetInstance()->RefferMouseClingObj();
		Card *card = dynamic_cast<Card*>(obj);

		// カードをマウスに持った状態なので、戻しておく
		if ( GetCurrentTurnParticipant().side == "P1") {
			currentState->Init_AddHandCard_BLACK( card->GetProperty().processId );
		}
		if ( GetCurrentTurnParticipant().side == "P2") {
			currentState->Init_AddHandCard_WHITE( card->GetProperty().processId );
		}

		// 指した状態とする
		currentState->GoNext(attemptingMove);

	///// しようとしている動作：「カード取得」
	} else if ( attemptingMove.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

		// 指した状態とする
		currentState->GoNext(attemptingMove);

	}

	/*
		同じ局面で、同じ指し手を実行した、という過去があるかどうか、という判定。
		同じ局面が発生するのだとしても、指し手が異なる場合は「同一」とは判定されないこととする。
	*/

	vector<GameRecord::Record> recordList = _score->GetRecordList();
	for ( unsigned int i = 0; i < recordList.size(); ++i ) {

		GameRecord::Record record = recordList[i];
		//record.innerExpression // 記録にある局面
		//record.move;           // その局面を作った指し手

		// 形だけ比較するので、対象外の値をそろえてしまう
		currentState->_currentTurn           = record.innerExpression->_currentTurn;
		currentState->_totalMoveCount        = record.innerExpression->_totalMoveCount;
		currentState->_isThrouableInfoLatest = record.innerExpression->_isThrouableInfoLatest;

		// 局面が同じで、指し手も同じの場合に、同一手とみなす。
		if ( currentState->IsEquals( record.innerExpression ) && IsEqualMove ( attemptingMove, record.move ) ) {
			occurenceCount += 1;
		}

	}

	/*
	vector<AI_Modules::InnerExpression*> recordList = _score->GetInnerExpressionsList();
	for ( unsigned int i = 0; i < recordList.size(); ++i ) {

		// 形だけ比較するので、対象外の値をそろえてしまう
		currentState->_currentTurn           = recordList[i]->_currentTurn;
		currentState->_totalMoveCount        = recordList[i]->_totalMoveCount;
		currentState->_isThrouableInfoLatest = recordList[i]->_isThrouableInfoLatest;

		if ( currentState->IsEquals( recordList[i] ) ) {
			occurenceCount += 1;
			//break;
		}
	}
	*/

	delete currentState;

	// ここで返す値は「これまでの発生回数」
	// なので、今回を含めると、千日手の判定は↓このようになる。
	//int repeatCount = occurenceCount + 1;
	//isRepetitionMove = ( repeatCount > AI_Modules::AI::_rule.repetitionMoveMaxCount);

	return occurenceCount;
}

//=========================================================
// ルールにそった操作を受付：駒設置先模索中
//=========================================================
void GameManager::Operation_With_Rule_TO_PUT_PIECE () {

	///// クリックしてないときは、操作を受け付けない
	if ( Application::GetInstance()->_mouseState_Left != MOUSE::BUTTON::STATE::RELEASE) {
		return;
	}

	///// 選択結果
	AI_Modules::Move selectedMove;
	bool isSelect = false;

	///// 可能な選択先のグリッドを選択しているか？
	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i ) {

		const AI_Modules::Move &move = _selectableMoveList[i];

		///// 駒設置？
		if ( move.action == AI_Modules::MOVE::ACTION::PUT_PIECE ) {

			// 設置先を選択している？
			if ( _board->GetGrid ( move.putPieceX, move.putPieceY )->IsMouseOver() ) {

				isSelect     = true;
				selectedMove = move;
				break;
			}
		}

	}

	///// 指し手を決めた？
	if ( isSelect ) {

		///// ここで千日手の確認
		bool isRepetitionMove = false;

		// 千日手禁止のルールがある場合は。
		if ( AI_Modules::AI::_rule.isRepetitionMoveProhibited ) {

			int occurenceCount = CountRepetition (selectedMove);
			int repeatCount = occurenceCount + 1;
			isRepetitionMove = ( repeatCount > AI_Modules::AI::_rule.repetitionMoveMaxCount);

			// 千日手未満だった？
			if ( !isRepetitionMove && occurenceCount > 0 ) {

				IEffect *effect = new Effect_RepetitionMoveAlertMessage ( repeatCount, AI_Modules::AI::_rule.repetitionMoveMaxCount );
				Application::GetInstance()->_effectManager->Add ( effect );

			}
		}

		// 千日手だった？
		if ( isRepetitionMove ) {

			IEffect *effect = new Effect_RepetitionMoveMessage();
			Application::GetInstance()->_effectManager->Add ( effect );

		// 千日手じゃなかった？
		} else {

			humanMoveDecided (selectedMove);

			///// マウスからはずして、選んだ位置に設置。
			IMouseClingableObject *obj = Application::GetInstance()->DrawMouseClingObj();

			Piece *piece = dynamic_cast<Piece*>(obj);

			_board->GetGrid ( selectedMove.putPieceX, selectedMove.putPieceY )->SetPiece ( piece );

			///// 操作状態を戻しておく：この時点で、もう操作を受け付けない、ということ。エフェクトの最中に操作を重複して受け付けるないように。
			_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;
			_is_MoveEffectGoing = true;

			///// 事後処理
			//Operation_AfterExecute_Process ( selectedMove );

			///// ゲージの底値を現在値に固定
			_psychoGage->SetMin_CurrentValue();
		}

	///// 決めてない
	} else {

		///// 盤面をクリックしていないなら、キャンセルとする。
		if ( _board->GetMouseOverGrid() == NULL ) {

			///// マウスからはずして、駒台に戻す
			IMouseClingableObject *obj = Application::GetInstance()->DrawMouseClingObj();

			Piece *piece = dynamic_cast<Piece*>(obj);

			PieceBox *currentTurnPieceBox = NULL;

			Participant currentTurnParticipant = GetCurrentTurnParticipant();
			if      ( currentTurnParticipant.side == "P1" ) { currentTurnPieceBox = _pieceBox_P1; }
			else if ( currentTurnParticipant.side == "P2" ) { currentTurnPieceBox = _pieceBox_P2; }
			else { printf ( "どっちサイドでもないので駒台取得できず（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }

			/*
			if      ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P1") { currentTurnPieceBox = _pieceBox_P1; }
			else if ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P2") { currentTurnPieceBox = _pieceBox_P2; }
			else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P1") { currentTurnPieceBox = _pieceBox_P1; }
			else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P2") { currentTurnPieceBox = _pieceBox_P2; }
			*/

			if ( currentTurnPieceBox != NULL ) {
				currentTurnPieceBox->AddPiece ( piece );
			}

			///// フラッシュを消しておく
			GridsFlashCancel();

			///// 操作状態を戻しておく
			_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

			///// ゲージの底値固定を解除
			_psychoGage->SetMin_Clear();
		}

	}
}
//=========================================================
// ルールにそった操作を受付：駒移動先模索中
//=========================================================
void GameManager::Operation_With_Rule_TO_MOVE_PIECE () {

	///// クリックしてないときは、操作を受け付けない
	if ( Application::GetInstance()->_mouseState_Left != MOUSE::BUTTON::STATE::RELEASE) {
		return;
	}

	///// 選択結果
	AI_Modules::Move selectedMove;
	bool isSelect = false;

	Grid *grid_Collided = NULL;

	///// 可能な移動先のグリッドを選択しているか？
	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i ) {

		const AI_Modules::Move &move = _selectableMoveList[i];

		///// 駒移動？
		if ( move.action == AI_Modules::MOVE::ACTION::MOVE_PIECE ) {

			// 選択したグリッドからの移動先？
			if ( _selectingGrid == _board->GetGrid ( move.movePieceFromX, move.movePieceFromY )) {

				// ぶつけない
				if ( !move.movePieceAndIsCllide ) {

					// 移動先を選択している？
					if (_board->GetGrid ( move.movePieceToX, move.movePieceToY )->IsMouseOver() ) {

						isSelect     = true;
						selectedMove = move;
						break;
					}

				// ぶつける
				} else {

					// ぶつけ先を選択している？
					pair<int, int> collidedPos = Game_Utils::GetCollidedPiecePosition ( move );

					// ぶつけられる駒が乗っているグリッド
					Grid *grid = _board->GetGrid ( collidedPos.first, collidedPos.second );

					if ( grid != NULL && grid->IsMouseOver() ) {

						isSelect      = true;
						selectedMove  = move;
						grid_Collided = grid;
						break;
					}

				}
			}

		///// カード取得？
		} else if ( move.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

			// 取得元を選択している？
			if ( _selectingGrid == _board->GetGrid ( move.takeCardX, move.takeCardY )) {

				if ( _selectingGrid->IsMouseOver() ) {

					isSelect     = true;
					selectedMove = move;
					break;
				}
			}
		}

	}

	///// 指し手を決めた？
	if ( isSelect ) {

		///// ここで千日手の確認
		bool isRepetitionMove = false;

		// 千日手禁止のルールがある場合は。
		if ( AI_Modules::AI::_rule.isRepetitionMoveProhibited ) {

			int occurenceCount = CountRepetition (selectedMove);
			int repeatCount = occurenceCount + 1;
			isRepetitionMove = ( repeatCount > AI_Modules::AI::_rule.repetitionMoveMaxCount);

			// 千日手未満だった？
			if ( !isRepetitionMove && occurenceCount > 0 ) {

				IEffect *effect = new Effect_RepetitionMoveAlertMessage ( repeatCount, AI_Modules::AI::_rule.repetitionMoveMaxCount );
				Application::GetInstance()->_effectManager->Add ( effect );

			}
		}

		// 千日手だった？
		if ( isRepetitionMove ) {

			IEffect *effect = new Effect_RepetitionMoveMessage();
			Application::GetInstance()->_effectManager->Add ( effect );

		// 千日手じゃなかった？
		} else {

			humanMoveDecided (selectedMove);

			///// 駒移動（ぶつけない）でした
			if ( selectedMove.action == AI_Modules::MOVE::ACTION::MOVE_PIECE && !selectedMove.movePieceAndIsCllide ) {

				///// 駒移動エフェクト
				IEffect *effect = new Effect_MovePiece (
					_endMoveFromEffect,
					selectedMove,
					//-1, // エフェクトレベルに負の数：キー操作で切り替えてもらう
					_psychoGage->GetValueAsEffectLevel(), // エフェクトレベル：意志ゲージで。
					_board->GetGrid ( selectedMove.movePieceFromX, selectedMove.movePieceFromY ),
					_board->GetGrid ( selectedMove.movePieceToX,   selectedMove.movePieceToY ),
					_board->GetGrid ( selectedMove.movePieceFromX, selectedMove.movePieceFromY )->GetPiece()
				);
				Application::GetInstance()->_effectManager->Add ( effect );

				// 操作用エフェクトを出したことを覚えておく → ロード（＝この盤面を強制終了）時の後片付け目的
				LetsRemenberThatIMadeThisEffect(effect);

				///// 操作状態を戻しておく
				_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;
				_is_MoveEffectGoing = true;

				///// フラッシュを消しておく
				GridsFlashCancel();

				///// ゲージ底値を固定
				_psychoGage->SetMin_CurrentValue();

			///// 駒移動（ぶつける）でした
			} else if ( selectedMove.action == AI_Modules::MOVE::ACTION::MOVE_PIECE && selectedMove.movePieceAndIsCllide  ) {

				///// エフェクトを使わない場合のコード
				/*
				// 指し手を実行。
				Move (selectedMove);

				// 事後処理
				Operation_AfterExecute_Process ( selectedMove );
				*/

				///// 駒移動エフェクト
				IEffect *effect = new Effect_MovePiece_Collide (
					_endMoveFromEffect,
					selectedMove,
					//-1, // エフェクトレベルに負の数：キー操作で切り替えてもらう
					_psychoGage->GetValueAsEffectLevel(), // エフェクトレベル：意志ゲージで。
					_board->GetGrid ( selectedMove.movePieceFromX, selectedMove.movePieceFromY ),
					_board->GetGrid ( selectedMove.movePieceToX,   selectedMove.movePieceToY ),
					_board->GetGrid ( selectedMove.movePieceFromX, selectedMove.movePieceFromY )->GetPiece(),
					grid_Collided->GetPiece(),
					GetPieceBox_BelongToPiece ( grid_Collided->GetPiece() ),
					grid_Collided
				);
				Application::GetInstance()->_effectManager->Add ( effect );

				// 操作用エフェクトを出したことを覚えておく → ロード（＝この盤面を強制終了）時の後片付け目的
				LetsRemenberThatIMadeThisEffect(effect);

				///// 操作状態を戻しておく
				_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;
				_is_MoveEffectGoing = true;

				///// フラッシュを消しておく
				GridsFlashCancel();

				///// ゲージ底値を固定
				_psychoGage->SetMin_CurrentValue();


			///// カード取得でした
			} else if ( selectedMove.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

				///// カード取得エフェクト
				IEffect *effect = new Effect_TakeCard (
					_endMoveFromEffect,
					selectedMove,
					//-1, // エフェクトレベルに負の数：キー操作で切り替えてもらう
					_psychoGage->GetValueAsEffectLevel(), // エフェクトレベル：意志ゲージで。
					_board->GetGrid ( selectedMove.takeCardX, selectedMove.takeCardY ),
					_board->GetGrid ( selectedMove.takeCardX, selectedMove.takeCardY )->GetCard(),
					_board->GetGrid ( selectedMove.takeCardX, selectedMove.takeCardY )->GetPiece(),
					GetCardBox_CurrentTurn()
				);
				Application::GetInstance()->_effectManager->Add ( effect );

				// 操作用エフェクトを出したことを覚えておく → ロード（＝この盤面を強制終了）時の後片付け目的
				LetsRemenberThatIMadeThisEffect(effect);

				///// 操作状態を戻しておく
				_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;
				_is_MoveEffectGoing = true;

				///// フラッシュを消しておく
				GridsFlashCancel();

				///// ゲージ底値を固定
				_psychoGage->SetMin_CurrentValue();

			///// それ以外？ そんなぁ〜
			} else {

				printf ("駒移動操作で、おかしなことになっています。selectedMove.action=%d\n", selectedMove.action);
			}

		} // 千日手判定の閉じ

	///// 決めてない
	} else {

		///// 盤以外のところをクリック → キャンセル
		if ( _board->GetMouseOverGrid() == NULL ) {

			///// 操作状態を戻しておく
			_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

			///// フラッシュを消しておく
			GridsFlashCancel();

			///// ゲージ底値固定を解除
			_psychoGage->SetMin_Clear();
		}

		///// 選択可能な他の駒をクリック → それを選びなおす
		Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_MOVE_PIECE();
	}

}


//=========================================================
// ルールにそった操作を受付：カード設置先模索中
//=========================================================
void GameManager::Operation_With_Rule_TO_PUT_CARD () {

	///// 右クリックで、マウスにくっついているはずのカードを回転させる。
	if ( Application::GetInstance()->_mouseState_Right == MOUSE::BUTTON::STATE::HAVE_PRESSED ) {

		IMouseClingableObject *obj = Application::GetInstance()->RefferMouseClingObj();

		Card *card = dynamic_cast<Card*>(obj);

		if ( card != NULL ) {

			card->Rolling_By_Operatiion();

			// 設置可能位置のフラッシュをやりなおす。
			GridsFlashCancel();
			FlashSelectableGrids_ToPutCard (
				card->GetProperty().processId,
				card->GetCurrentDirection()
			);
		}
	}


	///// クリックしてないときは、操作を受け付けない
	if ( Application::GetInstance()->_mouseState_Left != MOUSE::BUTTON::STATE::RELEASE) {
		return;
	}

	///// マウスにカードがくっついているはずです。
	IMouseClingableObject *obj = Application::GetInstance()->RefferMouseClingObj();
	Card *selectingCard = dynamic_cast<Card*>(obj);

	if ( selectingCard == NULL ) {
		printf ("GameManager::Operation_With_Rule_TO_PUT_CARD() カードがマウスにくっついているはずなのに！\n");
		return;
	}

	///// 選択結果
	AI_Modules::Move selectedMove;
	bool isSelect = false;

	///// 可能な選択先のグリッドを選択しているか？
	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i ) {

		const AI_Modules::Move &move = _selectableMoveList[i];

		///// カード設置？
		if ( move.action == AI_Modules::MOVE::ACTION::PUT_CARD ) {

			// 効果範囲がない場合、可能な指し手リストから向きが除外されている（UPだけになっている）
			int currentDirection = selectingCard->GetCurrentDirection();
			if ( selectingCard->GetProperty().gridAffectionList.size() == 0 ) {
				currentDirection = AI_Modules::DIRECTION::UP;
			}

			// カードID と 向きが一致している？
			if ( move.putCardId != selectingCard->GetProperty().processId || move.putCardDirection != currentDirection ) {
				continue;
			}

			// 設置先を選択している？
			if ( _board->GetGrid ( move.putCardX, move.putCardY )->IsMouseOver() ) {

				isSelect     = true;
				selectedMove = move;
				break;
			}
		}

	}

	///// 指し手を決めた？
	if ( isSelect ) {

		///// ここで千日手の確認
		bool isRepetitionMove = false;

		// 千日手禁止のルールがある場合は。
		if ( AI_Modules::AI::_rule.isRepetitionMoveProhibited ) {

			int occurenceCount = CountRepetition (selectedMove);
			int repeatCount = occurenceCount + 1;
			isRepetitionMove = ( repeatCount > AI_Modules::AI::_rule.repetitionMoveMaxCount);

			// 千日手未満だった？
			if ( !isRepetitionMove && occurenceCount > 0 ) {

				IEffect *effect = new Effect_RepetitionMoveAlertMessage ( repeatCount, AI_Modules::AI::_rule.repetitionMoveMaxCount );
				Application::GetInstance()->_effectManager->Add ( effect );

			}
		}

		// 千日手だった？
		if ( isRepetitionMove ) {

			IEffect *effect = new Effect_RepetitionMoveMessage();
			Application::GetInstance()->_effectManager->Add ( effect );

		// 千日手じゃなかった？
		} else {

			humanMoveDecided (selectedMove);

			///// マウスからはずして、選んだ位置に設置。
			IMouseClingableObject *obj = Application::GetInstance()->DrawMouseClingObj();

			Card *card = dynamic_cast<Card*>(obj);

			_board->GetGrid ( selectedMove.putCardX, selectedMove.putCardY )->SetCard ( card );

			///// 操作状態を戻しておく：この時点で、もう操作を受け付けない、ということ。エフェクトの最中に操作を重複して受け付けるないように。
			_operationState     = OperationState::TO_DETERMIN_WHAT_TO_DO;
			_is_MoveEffectGoing = true;

			// マウスから外れた、というわけで、ロード用のメモを忘れておく。
			_operationMemo_belongsToWhichCardBox = "";
			_operationMemo_whichCardTheCardWas   = "";

			///// 事後処理
			//Operation_AfterExecute_Process ( selectedMove );

			///// ゲージの底値を現在値に固定
			_psychoGage->SetMin_CurrentValue();
		}

	///// 決めてない
	} else {

		///// 盤の外側をクリック → キャンセル
		if ( _board->GetMouseOverGrid() == NULL ) {

			// マウスから外して、元あったカード置き場に戻す

			IMouseClingableObject *obj = Application::GetInstance()->DrawMouseClingObj();

			Card *card = dynamic_cast<Card*>(obj);

			CardBox *currentTurnCardBox = NULL;

			Participant currentTurnParticipant = GetCurrentTurnParticipant();
			if      ( currentTurnParticipant.side == "P1" ) { currentTurnCardBox = _cardBox_P1; }
			else if ( currentTurnParticipant.side == "P2" ) { currentTurnCardBox = _cardBox_P2; }
			else { printf ( "どっちサイドでもないのでカード置き場取得できず（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }
			/*
			if      ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P1") { currentTurnCardBox = _cardBox_P1; }
			else if ( _operationData.currentTurn == "FIRST"  &&  _operationData.first  == "P2") { currentTurnCardBox = _cardBox_P2; }
			else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P1") { currentTurnCardBox = _cardBox_P1; }
			else if ( _operationData.currentTurn == "SECOND" &&  _operationData.second == "P2") { currentTurnCardBox = _cardBox_P2; }
			*/

			if ( currentTurnCardBox != NULL ) {
				currentTurnCardBox->AddCard ( card );
			}

			// マウスから外れた、というわけで、ロード用のメモを忘れておく。
			_operationMemo_belongsToWhichCardBox = "";
			_operationMemo_whichCardTheCardWas   = "";

			/*
			if ( _currentTurn == AI_Modules::TURN::BLACK ) {
				_cardBox_P1->AddCard ( card );
			}
			else if ( _currentTurn == AI_Modules::TURN::WHITE ) {
				_cardBox_P2->AddCard ( card );
			}
			*/

			///// フラッシュを消しておく
			GridsFlashCancel();

			///// 操作状態を戻しておく
			_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

			///// ゲージの底値固定を解除
			_psychoGage->SetMin_Clear();
		}
	}

}


//=========================================================
// 人間の指し手が決定したタイミングで指し手情報を受け取って何かするかもしれない
//=========================================================
void GameManager::humanMoveDecided ( AI_Modules::Move decidedMove ) {

	///// 口出しモードでの口出し着手だった場合
	if ( _interferenceModeInfo.isActive && _interferenceModeInfo.phase == _interferenceModeInfo.PHASE_WAIT_FOR_PLAYER_MOVE ) {

		InterferenceMode_PlayerMoveDecided ( decidedMove );
	}


	/*

	// 人間の指し手ということで、総括評価値は「未知」で。
	string sumUpEval = AI_Modules::EVALUATION::SUM_UP::UNKNOWN;

	vector<Actor*> actors = GetAllActors();
	for ( unsigned int i = 0; i < actors.size(); ++i ) {
		Actor *a = actors[i];
		a->Notice_MoveIsReady ( GetOperationData().currentTurn, decidedMove, sumUpEval );
	}
	*/
}


//=========================================================
// 選択可能なグリッドをフラッシュさせる：駒設置
//=========================================================
void GameManager::FlashSelectableGrids_ToPutPiece  () {

	/*
		ここでは、指し手リストを信じて、見た目として、光らせるだけ。
	*/

	///// フラッシュを消しておく
	GridsFlashCancel();

	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i ) {

		const AI_Modules::Move &move = _selectableMoveList[i];

		// 指し手の種類：駒設置だけ、お相手する。
		if ( move.action != AI_Modules::MOVE::ACTION::PUT_PIECE ) {
			continue;
		}

		Grid *grid = _board->GetGrid ( move.putPieceX, move.putPieceY );
		if ( grid != NULL ) {
			grid->SetSelectable ( true );
		}
	}
}

//=========================================================
// 選択可能なグリッドをフラッシュさせる：駒移動先
//=========================================================
void GameManager::FlashSelectableGrids_ToMovePiece  (  int selectedX, int selectedY  ) {

	/*
		ここでは、指し手リストを信じて、見た目として、光らせるだけ。

		操作方法の都合として、
		ここでは、「移動」「ぶつける移動」「カード取得」が混ざっている。

		「ぶつける移動」では、ぶつける相手の駒の位置を光らせる。

		選択した位置も、光らせる。

	*/

	///// フラッシュを消しておく
	GridsFlashCancel();

	///// 選択元（移動元）を光らせる（「選んだ位置」、として）
	Grid *selectedGrid = _board->GetGrid ( selectedX, selectedY );
	if ( selectedGrid != NULL ) {
		selectedGrid->SetSelected ( true );
	}

	///// 選択可能な指し手リストを信じてループする
	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i ) {

		const AI_Modules::Move &move = _selectableMoveList[i];

		// 指し手の種類：移動
		if ( move.action == AI_Modules::MOVE::ACTION::MOVE_PIECE ) {

			// 選んだ位置が一致している？
			if ( move.movePieceFromX != selectedX || move.movePieceFromY != selectedY ) {
				continue;
			}

			// ぶつけない
			if ( !move.movePieceAndIsCllide ) {

				// 移動先のグリッドを光らせる
				Grid *grid = _board->GetGrid ( move.movePieceToX, move.movePieceToY );

				if ( grid != NULL ) {
					grid->SetSelectable ( true );
				}

			// ぶつける
			} else {

				// ぶつける相手の駒の位置を光らせる
				pair<int, int> collidedPos = Game_Utils::GetCollidedPiecePosition ( move );

				Grid *grid = _board->GetGrid ( collidedPos.first, collidedPos.second );

				if ( grid != NULL ) {
					grid->SetSelectable ( true );
				}

			}

		// 指し手の種類：カード取得
		} else if ( move.action == AI_Modules::MOVE::ACTION::TAKE_CARD ) {

			// 選んだ位置が一致している？
			if ( move.takeCardX != selectedX || move.takeCardY != selectedY ) {
				continue;
			}

			// 取得するカードがある位置のグリッドを光らせる
			Grid *grid = _board->GetGrid ( move.takeCardX, move.takeCardY );

			if ( grid != NULL ) {
				grid->SetSelectable ( true );
			}

		}

	}
}
//=========================================================
// 選択可能なグリッドをフラッシュさせる：カード設置先
//=========================================================
void GameManager::FlashSelectableGrids_ToPutCard  ( int cardId, int direction ) {

	/*
		ここでは、指し手リストを信じて、見た目として、光らせるだけ。
	*/

	/*
		えっと、効果範囲がないカードは、可能な指し手リストから、
		回転が除外されているので、
		そういうカードの場合、引数の direction を、UP　に書き換えて以下の処理を行う。
	*/
	if ( CardFactory::GetCardProperty ( cardId ).gridAffectionList.size() == 0) {
		direction = AI_Modules::DIRECTION::UP;
	}

	///// フラッシュを消しておく
	GridsFlashCancel();

	for ( unsigned int i = 0; i < _selectableMoveList.size(); ++i ) {

		const AI_Modules::Move &move = _selectableMoveList[i];

		// 指し手の種類を確認：カード設置じゃないなら、およびでない。
		if ( move.action != AI_Modules::MOVE::ACTION::PUT_CARD ) {
			continue;
		}

		// カードIDと向きが一致？
		if ( move.putCardId != cardId || move.putCardDirection != direction ) {
			continue;
		}

		// 設置先のグリッドを光らせる
		Grid *grid = _board->GetGrid ( move.putCardX, move.putCardY );

		if ( grid != NULL ) {
			grid->SetSelectable ( true );
		}

	}
}

//=========================================================
// 最終着手をください。
//=========================================================
AI_Modules::Move GameManager::GetLastMove() {

	AI_Modules::Move result;

	///// 試合じゃないとき、という可能性もある。いや、ない。常に棋譜オブジェクトを持つことにした。表示しないときは非表示フラグを使う。
	if ( _score != NULL) {

		vector<GameRecord::Record> recordList = _score->GetRecordList();

		if ( recordList.size() > 0 ) {

			result = recordList[ recordList.size() - 1 ].move;
		}

	} else {
		printf ("最終着手をください、ですって？ だけど棋譜オブジェクトがないのですのよー！ ウワオー！\n");
	}
	return result;
}
//=========================================================
// 何手目かをください。
//=========================================================
int GameManager::GetScoreRecordLength() {

	int result = 0;

	///// 試合じゃないとき、という可能性もある。いや、ない。常に棋譜オブジェクトを持つことにした。表示フラグにて表示を切り替える
	if ( _score != NULL) {

		vector<GameRecord::Record> recordList = _score->GetRecordList();

		result = recordList.size();

	} else {
		printf("何手目かをください？ でも棋譜オブジェクトがないんですよー！ ウオー！\n");
	}

	return result;
}

//=========================================================
// 現在の手番は、AI の TURN としてはどっちか？
// AI::TURN の 黒・白に、GameManager のFirst・Second を変換
//=========================================================
int GameManager::GetTurn_in_tarms_of_TURN_FOR_AI () {

	int result = AI_Modules::TURN::BLACK;

	/*
		実際の駒の色が何であろうと、P1サイドはBLACK、P2サイドはWHITE とする。
	*/

	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	if      ( currentTurnParticipant.side == "P1" ) { result = AI_Modules::TURN::BLACK; }
	else if ( currentTurnParticipant.side == "P2" ) { result = AI_Modules::TURN::WHITE; }
	else { printf ( "Error in GameManager::GetTurn_in_tarms_of_TURN_FOR_AI どっちサイドでもない（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }

	/*
	if ( _operationData.currentTurn == "FIRST" ) {
		if      ( _operationData.first == "P1" ) { result = AI_Modules::TURN::BLACK; }
		else if ( _operationData.first == "P2" ) { result = AI_Modules::TURN::WHITE; }
	}
	else if ( _operationData.currentTurn == "SECOND" ) {
		if      ( _operationData.second == "P1" ) { result = AI_Modules::TURN::BLACK; }
		else if ( _operationData.second == "P2" ) { result = AI_Modules::TURN::WHITE; }
	}
	*/

	return result;
}

//=========================================================
// 現在の手番の駒台をください
//=========================================================
PieceBox* GameManager::GetPieceBox_CurrentTurn() {

	PieceBox *result = NULL;

	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	if      ( currentTurnParticipant.side == "P1" ) { result = _pieceBox_P1; }
	else if ( currentTurnParticipant.side == "P2" ) { result = _pieceBox_P2; }
	else { printf ( "Error in GameManager::GetPieceBox_CurrentTurn どっちサイドでもない（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }

	/*
	if ( _operationData.currentTurn == "FIRST" ) {
		if      ( _operationData.first == "P1" ) { result = _pieceBox_P1; }
		else if ( _operationData.first == "P2" ) { result = _pieceBox_P2; }
	}
	else if ( _operationData.currentTurn == "SECOND" ) {
		if      ( _operationData.second == "P1" ) { result = _pieceBox_P1; }
		else if ( _operationData.second == "P2" ) { result = _pieceBox_P2; }
	}
	*/
	return result;
}
//=========================================================
// 現在の手番のカード置き場をください
//=========================================================
CardBox* GameManager::GetCardBox_CurrentTurn() {

	CardBox *result = NULL;

	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	if      ( currentTurnParticipant.side == "P1" ) { result = _cardBox_P1; }
	else if ( currentTurnParticipant.side == "P2" ) { result = _cardBox_P2; }
	else { printf ( "Error in GameManager::GetCardBox_CurrentTurn どっちサイドでもない（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }

	/*
	if ( _operationData.currentTurn == "FIRST" ) {
		if      ( _operationData.first == "P1" ) { result = _cardBox_P1; }
		else if ( _operationData.first == "P2" ) { result = _cardBox_P2; }
	}
	else if ( _operationData.currentTurn == "SECOND" ) {
		if      ( _operationData.second == "P1" ) { result = _cardBox_P1; }
		else if ( _operationData.second == "P2" ) { result = _cardBox_P2; }
	}
	*/

	return result;
}
//=========================================================
// 現在の手番じゃない方の駒台をください
//=========================================================
PieceBox* GameManager::GetPieceBox_NotCurrentTurn() {

	PieceBox *result = NULL;

	Participant currentTurnParticipant = GetCurrentTurnParticipant();
	if      ( currentTurnParticipant.side == "P1" ) { result = _pieceBox_P2; }
	else if ( currentTurnParticipant.side == "P2" ) { result = _pieceBox_P1; }
	else { printf ( "Error in GameManager::GetPieceBox_NotCurrentTurn どっちサイドでもない（currentTurnParticipant.side：%s）。\n", currentTurnParticipant.side.c_str() ); }

	/*
	if ( _operationData.currentTurn == "FIRST" ) {
		if      ( _operationData.first == "P1" ) { result = _pieceBox_P2; }
		else if ( _operationData.first == "P2" ) { result = _pieceBox_P1; }
	}
	else if ( _operationData.currentTurn == "SECOND" ) {
		if      ( _operationData.second == "P1" ) { result = _pieceBox_P2; }
		else if ( _operationData.second == "P2" ) { result = _pieceBox_P1; }
	}
	*/

	return result;
}

//=========================================================
// この駒が属している方の駒台をください
//=========================================================
PieceBox* GameManager::GetPieceBox_BelongToPiece (  Piece *piece ) {

	PieceBox *result = NULL;

	if ( piece == NULL ) {
		return NULL;
	}

	///// 黒だったら P1、白だったら P2
	/*
		P1に白を持たせて後手にしたい、という件につきましては、

		駒の「論理色」と、「見た目色」という概念で対処することとしましょう。

		内部的に「BLACK」「WHITE」というのは、「論理色」

		見た目を変えたいだけのときは、そのように設定する。
		見た目は白だけど、論理色はBLACK、という状態を作り出す。
	*/

	if      ( piece->GetColor() == Piece::Color::BLACK) { result = _pieceBox_P1; }
	else if ( piece->GetColor() == Piece::Color::WHITE) { result = _pieceBox_P2; }


	return result;
}

//=========================================================
// ぶつける先のグリッドが欲しい。
//【引数】from to、では、その先のグリッドは？
//=========================================================
Grid* GameManager::GetCollidedGrid ( int fromX, int fromY, int toX, int toY ) {

	Grid *result = NULL;

	int collidedX = toX + Utils_Cal::Sign ( toX - fromX );
	int collidedY = toY + Utils_Cal::Sign ( toY - fromY );

	result = _board->GetGrid ( collidedX, collidedY );

	return result;
}


//=========================================================
// 全駒を取得
//=========================================================
vector<Piece*> GameManager::GetAllPieces_forManagement() {
        
    return _pieceList_forManagement;
}
//=========================================================
// 全札を取得
//=========================================================
vector<Card*> GameManager::GetAllCards_forManagement() {
        
    return _cardList_forManagement;
}



//=========================================================
// セーブ用のデータを作成
//=========================================================
GameManager::SaveData GameManager::GetSaveData() {

	GameManager::SaveData result;

	result.name = _name;
	result.gameBoardSerialized = SerializeGameBoard();
	result.isActive = _isActive;
	result.isFreeOperation = _isFreeOperation;
	result.boardCenterX = _boardCenterX;
	result.boardCenterY = _boardCenterY;
	//result.actorPosisionX = _actorPosisionX;
	//result.actorPosisionY = _actorPosisionY;
	result.isRendered              = _isRendered;
	result.operationState          = _operationState;
	result.IsSelectableMoveListNew = _IsSelectableMoveListNew;
	result.isAstralProjection      = _isAstralProjection;
	result.is_MoveEffectGoing      = _is_MoveEffectGoing;
	result.kifuViewerPosition      = _kifuViewerPosition;
	result.evaluationSumUp_FromAI  = _evaluationSumUp_FromAI;
	result.operationData = _operationData;
	result.battleInfo    = _battleInfo;

	result.selectableMoveList         = _selectableMoveList;
	result.progressData_AfterThinking = _progressData_AfterThinking;
	result.astralProjection_physicalState = ToString_astralProjection_physicalState();
	result.gameRecordSerialized = (_score == NULL) ? "NULL" : _score->Serialize();

	// 負の数なら選択中のグリッドはないものとする。
	result.selectingGridX = _selectingGridX;
	result.selectingGridY = _selectingGridY;

	result.operationMemo_belongsToWhichPieceBox = _operationMemo_belongsToWhichPieceBox;
	result.operationMemo_whatColorThePieceWas   = _operationMemo_whatColorThePieceWas;
	result.operationMemo_belongsToWhichCardBox  = _operationMemo_belongsToWhichCardBox;
	result.operationMemo_whichCardTheCardWas    = _operationMemo_whichCardTheCardWas;

	// TODO さしあたり、試合中のセーブのことは考えないこととする
	//result.AI_fileName = "";
	//result.actorSerialized = "NULL";

	//result.AI_fileName = _AI_fileName;
	//result.actorSerialized = ( _actor == NULL ) ? "NULL" : _actor->Serialize();


	///// 参加者情報
	string participantsString    = "";
	string participantsDelimitor = "<pDelim>";
	for ( unsigned int i = 0; i < _participantsList.size(); ++i ) {


		Participant p = _participantsList[i];

		// TODO 参加者ごとにシリアライズ/デシリアライズすることとしましょうか。

		string participantSeriarized = SerializeParticipant ( p );

		if ( participantsString != "" ) { participantsString += participantsDelimitor; }
		participantsString += participantSeriarized;

	}
	result.participantsString = participantsString;

	///// 手番の順番情報
	//_operationData.currentTurn;
	//_operationData.currentTurnIndex;
	//_operationData.turnOrder;
	result.operationData = _operationData; // これは SaveData.cpp の側でよしなにしている
	
	
	//sf::String s1 = _battleTitleForKifuViewing.getString();
	// できれば _battleTitleForKifuViewing から string を取得したいところではあるけれど、、、
	//result.battleTitle = "";
	result.battleTitle = _battleInfo.battleTitle;
	/*
	if ( _score != NULL ) {
		result.battleTitle = _score->GetBattleTitle();
		
		
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
	printf ("試合タイトルをセーブデータに入れます：%s\n", result.battleTitle.c_str());
		
	}
	*/

	return result;
}


//=========================================================
// ロード
//=========================================================
void GameManager::Load ( GameManager::SaveData &saveData ) {

printf("ロードするぞ！！！！ GameManager::Load()\n");

	///// ロードする前にしておくべき後片付け処理
	if ( _isAstralProjection ) {
		AstralProjectionEnd_Forcebly();
	}
	if (_battleInfo.isBattlingNow) {
		BattleEnd_Forcebly();
	}

	// 操作用エフェクトを出していたなら消しておく。
	TerminateAllOurBeLovedEffects();

	// マウスにカードがくっついている？ そんなのおかしいですよ！
	ReleaseCardFromMouseIfBelongToHere();

	//if ( Application::GetInstance()->IsMouseClingObj_Card()) {

	//	Card *card = (Card*)Application::GetInstance()->DrawMouseClingObj();
	//	CardFactory::Release(card);

		/*
			そのカードが、この盤面がいじっていいオブジェクトであるという保証はあんまりない。
			たとえば GameManager を複数同時に動かしているときはどうなのか？
			もしかすると、他の盤面から持ち上げたカードかもしれない。

			しかし「ロード」ということは、他の盤面も含めて、一旦全て消されているわけで、
			仮に他の盤面由来のものであるにしても、同じことなのかもしれない。

			つまり、検討モードでカードを持ったままセーブ→ロードするとどうなるのか？ ってことです。
			棋譜を読み込んで該当の局面にする、という処理があるので、
			マウスのカードを無視して局面だけは元の状態になる。
			そして、マウスのカードが一枚余分に増えた勘定になる。

			ちなみに駒の場合はどうかというと、ReleaseAllPiece() の中で、マウスからも外しているのですよね。
			なぜそれができるのかと言えば、駒はGameManager内で new したオブジェクトのリストに入れているから。

			カードでも同じことをするとよい？



		*/
	//}

	///// データを取り込む
	_name                    = saveData.name;
	_isActive                = saveData.isActive;
	_isFreeOperation         = saveData.isFreeOperation;
	_isRendered              = saveData.isRendered;
	_operationState          = saveData.operationState;
	_IsSelectableMoveListNew = saveData.IsSelectableMoveListNew;
	_isAstralProjection      = saveData.isAstralProjection;
	_is_MoveEffectGoing      = saveData.is_MoveEffectGoing;
	_kifuViewerPosition      = saveData.kifuViewerPosition;

	//_operationData.isLock      = saveData.operationData.isLock;

	// TODO さしあたり、試合中のセーブ/ロードのことは考えないこととする
	//_operationData.currentTurn = saveData.operationData.currentTurn;
	//_operationData.first       = saveData.operationData.first;
	//_operationData.second      = saveData.operationData.second;
	//_operationData.operatorP1  = saveData.operationData.operatorP1;
	//_operationData.operatorP2  = saveData.operationData.operatorP2;
	_operationData = saveData.operationData; // これでいいはず（SaveData.cppの側で面倒を見ている）

	_battleInfo.isBattlingNow  = saveData.battleInfo.isBattlingNow;
	_battleInfo.result         = saveData.battleInfo.result;
	_evaluationSumUp_FromAI    = saveData.evaluationSumUp_FromAI;
	_selectableMoveList        = saveData.selectableMoveList;

	_progressData_AfterThinking = saveData.progressData_AfterThinking;

	_operationMemo_belongsToWhichPieceBox = saveData.operationMemo_belongsToWhichPieceBox;
	_operationMemo_whatColorThePieceWas   = saveData.operationMemo_whatColorThePieceWas;
	_operationMemo_belongsToWhichCardBox  = saveData.operationMemo_belongsToWhichCardBox;
	_operationMemo_whichCardTheCardWas    = saveData.operationMemo_whichCardTheCardWas;

	//_actorPosisionX = saveData.actorPosisionX;
	//_actorPosisionY = saveData.actorPosisionY;

	string participantsString = saveData.participantsString;

printf("ロードできたかな？？？\n");


	///// データを取り込んだらば、行うべき準備処理
	DeSerializeGameBoard   ( saveData.gameBoardSerialized);
	SetBoardCenterPosition ( saveData.boardCenterX, saveData.boardCenterY);
	SetFreeOperation       ( saveData.isFreeOperation );

	FromString_astralProjection_physicalState ( saveData.astralProjection_physicalState );

	if ( _score != NULL ) {delete _score;}
	_score = NULL;
	_score = new GameRecord(); // 棋譜オブジェクトは常にあってくれる前提ですよね
	if ( saveData.gameBoardSerialized != "NULL" ) {
		//_score = new GameRecord();
		_score->DeSerialize ( saveData.gameRecordSerialized );
		
	}
	if ( _score != NULL ) {
		_score->SetPosition    (_kifuViewerPosition);
	}


	//SetAIFile ( "", saveData.AI_fileName ); // TODO AI名はダミー。今はAIを一つしか持ってないのでセーブ/ロード対象からは省いている。

	// TODO さしあたり試合中のセーブ/ロードのことは考えないこととする
	//if ( _actor != NULL ) { delete _actor;}
	//_actor = Actor::DeSerialize ( saveData.actorSerialized, ScriptEngine::textMessageWindowParaTi->GetSameSetuppedInstance());

	///// 参加者情報を復元するよ
	clearAllParticipants(); // 念のため、既存のものをクリアしておく

	vector<string> participantsStringList = Utils_String::Split ( participantsString, "<pDelim>" );
	for ( unsigned int i = 0; i < participantsStringList.size(); ++i ) {


		Participant p = DeSerializeParticipant ( participantsStringList[i] );
		addParticipant ( p );

		SetAIFile ( p.id, p.aiFileName );
		SetActorScriptFile ( p, p.actorFileName, p.actorGoalPosX, p.actorGoalPosY, p.actorCurrentPosX, p.actorCurrentPosY, ScriptEngine::textMessageWindowParaTi );
	}



	// 選択中のグリッド
	_selectingGridX = saveData.selectingGridX;
	_selectingGridY = saveData.selectingGridY;

	if ( _selectingGridX < 0 || _selectingGridY < 0 ) {

		// 参照を持つだけなので delete はしない
		_selectingGrid = NULL;

	} else {

		_selectingGrid = _board->GetGrid ( _selectingGridX, _selectingGridY );

	}


	///// ルールに則った操作の途中でしたか？
	if ( _operationState == OperationState::TO_PUT_PIECE ) {

		printf ( "駒を設置しようとしていたのですね\n" );

		_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

		// マウスで操作中の駒、というものがあったはず。
		// しかし、それはセーブ/ロード対象ではないのでした。
		// というわけで、ここで作り直そうではありませんか。
		// ちなみに、作り直してもマウスにくっつけ直したりはしませんよ。
		// 「操作エフェクト」も消滅しているので、結局、「操作をする前」の状態に戻すしかないのです。

		Piece *piece = new Piece ( _operationMemo_whatColorThePieceWas );
		piece->WarpToGoal();

		// 管理リストにも入れる必要があるはず。
		_pieceList_forManagement.push_back ( piece );

		// さて、この駒はどこの駒台にあったのでしょう？
		if ( _operationMemo_belongsToWhichPieceBox == "P1" ) {
			_pieceBox_P1->AddPiece(piece);
		}
		else if ( _operationMemo_belongsToWhichPieceBox == "P2" ) {
			_pieceBox_P2->AddPiece(piece);
		}
		else  {
			printf ( "えっ、どこの駒台にも属してないだなんて！！！！！\n" );
		}


	}
	else if ( _operationState == OperationState::TO_PUT_CARD ) {

		printf ( "カードを設置しようとしていたのですね\n" );

		// じゃあ、カードを生成しなおしましょうか
		Card *card = CardFactory::GetCard(_operationMemo_whichCardTheCardWas);
		card->WarpToGoal();

		// 生成しなおしたら、カード置き場に戻したいですよね？ P1/P2 どっちに戻せばいいのかな？
		if ( _operationMemo_belongsToWhichCardBox == "P1" ) {
			_cardBox_P1->AddCard(card);
		}
		else if ( _operationMemo_belongsToWhichCardBox == "P2" ) {
			_cardBox_P2->AddCard(card);
		} else {
			printf ( "どこのカード置き場にも属してないだなんて！！！！！！\n" );
		}

		_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

		/*
			右クリックでのカード回転とメニュー表示が衝突する関係で、
			カードをマウスで持っているときにはメニューは出ないようにしている。
			つまり、セーブもできない。

			操作方法がこのようである限りにおいて、このつじつま合わせ処理の出番はない、はず。

			ただし、たとえば、メニューを介さずにショートカットキーなどでセーブするような機能を追加したならば、
			カードを持った状態でのセーブ、というものが可能になってしまうでしょう。おそろしい！
		*/

	}
	else if ( _operationState == OperationState::TO_MOVE_PIECE ) {

		printf ( "駒を動かそうとしていたのですね\n" );

		_operationState = OperationState::TO_DETERMIN_WHAT_TO_DO;

		// つじつま合わせのために必要な処理はないはず。マウスにくっつけているわけではないですからね。
		// ちなみに言えば、グリッドの発光状態もセーブ/ロード対象ではない。着手前に戻るのであるから、それでよい。
	}


	///// 着手エフェクトが進行中でしたか？
	if ( _is_MoveEffectGoing ) {

		printf ( "着手エフェクトが進行中だったのですね\n" );

		// エフェクトはセーブ/ロード対象ではない
		_is_MoveEffectGoing = false;
	}
	
	
	if ( saveData.battleTitle != "") {
	
		SetupBattleTitleForKifuViewing ( saveData.battleTitle );
		
		// GameManager側での試合タイトル確保
		_battleInfo.battleTitle = saveData.battleTitle;
		
		if ( _score != NULL ) {
			
			// 棋譜側での試合タイトル確保
			_score->SetBattleTitle ( saveData.battleTitle );

		} else {
			printf ( "_score が NULL でした。\n" );
		}
	}



printf ("ロードできた！！！！！！\n");
}


//=========================================================
// 盤面の状態を文字列化
//=========================================================
string GameManager::SerializeGameBoard() {

	/*
		盤のサイズが違っているときには未対応。_board 内の再現関数をいじる必要あり。
		カードの定義が違うときも未対応。
		せいぜい、「カードの定義書」を記録しておく程度か？
	*/

	string data = "";
	/*
	Board    *_board;
	PieceBox *_pieceBox_P1;
	CardBox  *_cardBox_P1;
	PieceBox *_pieceBox_P2;
	CardBox  *_cardBox_P2;
	*/

	string outPut_Board_Size = _board->Output_Size();
	string outPut_Board      = _board->Output();
	string outPut_CardBoxP1  = _cardBox_P1->Output();
	string outPut_CardBoxP2  = _cardBox_P2->Output();
	string outPut_PieceBoxP1 = _pieceBox_P1->Output();
	string outPut_PieceBoxP2 = _pieceBox_P2->Output();

	data += outPut_Board_Size + "*";
	data += outPut_Board      + "*";
	data += outPut_CardBoxP1  + "*";
	data += outPut_CardBoxP2  + "*";
	data += outPut_PieceBoxP1 + "*";
	data += outPut_PieceBoxP2;

	//printf("%s\n" , data.c_str());


	return data;
}

//=========================================================
// 盤面の状態を文字列から復元
//=========================================================
void   GameManager::DeSerializeGameBoard( string data ) {

	///// 文字列を解釈しまして、
	vector<string> list = Utils_String::Split(data, "*");
	string outPut_Board_Size = list[0];
	string outPut_Board      = list[1];
	string outPut_CardBoxP1  = list[2];
	string outPut_CardBoxP2  = list[3];
	string outPut_PieceBoxP1 = list[4];
	string outPut_PieceBoxP2 = list[5];

	///// 現状のものを全部片付けて
	RemoveAllPiece_fromBox   ( "P1" );
	RemoveAllPiece_fromBox   ( "P2" );
	RemoveAllPiece_fromBoard ();
	RemoveAllCard_fromBox    ( "P1" );
	RemoveAllCard_fromBox    ( "P2" );
	RemoveAllCard_fromBoard  ();

	// これでマウスにくっついているオブジェクトも外れる
	ReleaseAllPiece();

	delete _board;
	_board = Board::GetInstance( outPut_Board_Size );
	_board->Read( outPut_Board );
	_cardBox_P1->Read( outPut_CardBoxP1 );
	_cardBox_P2->Read( outPut_CardBoxP2 );
	_pieceBox_P1->Read( outPut_PieceBoxP1 );
	_pieceBox_P2->Read( outPut_PieceBoxP2 );

	///// 管理用の駒リストを合わせる
	if ( _pieceList_forManagement.size() > 0 ) {
		printf ( "[Warning] 管理用の駒リストに駒が残っていた！[%d個] GameManager::DeSerializeGameBoard( string [%s] )\n", _pieceList_forManagement.size(), data.c_str() );
	}
	_pieceList_forManagement.clear();

	// 盤面から、駒台から、駒のポインタを持ち寄って、管理リストに入れる。
	for (Piece *piece : _board->GetAllPieces() ) {
		_pieceList_forManagement.push_back(piece);
	}
	for (Piece *piece : _pieceBox_P1->GetPieceList() ) {
		_pieceList_forManagement.push_back(piece);
	}
	for (Piece *piece : _pieceBox_P2->GetPieceList() ) {
		_pieceList_forManagement.push_back(piece);
	}

	///// 管理用のカードリストを合わせる
	if ( _cardList_forManagement.size() > 0 ) {
		printf ( "[Warning] 管理用のカードリストにカードが残っていた！[%d個] GameManager::DeSerializeGameBoard( string [%s] )\n", _pieceList_forManagement.size(), data.c_str() );
	}
	_cardList_forManagement.clear();

	// 盤面から、カード置き場から、カードのポインタを持ち寄って、管理リストに入れる。
	for (Card *card : _board->GetAllCards() ) {
		_cardList_forManagement.push_back(card);
	}
	for (Card *card : _cardBox_P1->GetAllCards() ) {
		_cardList_forManagement.push_back(card);
	}
	for (Card *card : _cardBox_P2->GetAllCards() ) {
		_cardList_forManagement.push_back(card);
	}

	///// 目的地へのびよーんを省略いたします
	for ( Piece *piece : _pieceList_forManagement) {

		piece->WarpToGoal();

	}
	for ( vector<Grid*> row : _board->GetAllGrid() ) {
		for ( Grid * grid : row ) {

			Card *card = grid->GetCard();

			if ( card != NULL ) {
				card->WarpToGoal();
			}
		}
	}
	for ( Card *card : _cardBox_P1->GetAllCards() ) {
		card->WarpToGoal();
	}
	for ( Card *card : _cardBox_P2->GetAllCards() ) {
		card->WarpToGoal();
	}

	_isRendered = false;
}

//=========================================================
// 検討モードでの盤面いじり判定用のデータ。セーブロード対応のための処理
//=========================================================
string GameManager::ToString_astralProjection_physicalState () {

	string result = "";

	if (_astralProjection_physicalState == NULL) {
		return "";
	}

	// つまり、GameManager::IsEqual_EasyVersion() で見ている分だけ、ということ。
	/*
		///// 盤面情報
		int         _gridList_Piece[15][15];         // 駒レイヤー      ：黒/白/なし GRID
		CardOnBoard _gridList_Card[15][15];          // カードレイヤー  ：カードと向き
		int _boardSizeX;
		int _boardSizeY;

		///// 駒台情報：何個あるか、だけ。
		int _handPiece_BLACK;
		int _handPiece_WHITE;

		///// 手札情報：カード番号のリスト
		int _handCards_BLACK_Index;
		int _handCards_WHITE_Index;
		int _handCards_BLACK[100];
		int _handCards_WHITE[100];
	*/
	/*
		//=========================================================
		// 盤上のカード
		//=========================================================
		struct CardOnBoard {

			int cardId;       // どのカードが
			int direction;    // どの向きか
		}
	*/

	string interVariableDelim = ",";
	string withinGridDelim = "*";
	string interGridDelim  = "<@>";

	result += std::to_string (_astralProjection_physicalState->_boardSizeX) + interVariableDelim;
	result += std::to_string (_astralProjection_physicalState->_boardSizeY) + interVariableDelim;
	result += std::to_string (_astralProjection_physicalState->_handPiece_BLACK) + interVariableDelim;
	result += std::to_string (_astralProjection_physicalState->_handPiece_WHITE) + interVariableDelim;
	result += std::to_string (_astralProjection_physicalState->_handCards_BLACK_Index) + interVariableDelim;
	result += std::to_string (_astralProjection_physicalState->_handCards_WHITE_Index) + interVariableDelim;

	for ( int y = 0; y < _astralProjection_physicalState->_boardSizeY; ++y ) {
		for ( int x = 0; x < _astralProjection_physicalState->_boardSizeX; ++x ) {

			if (x > 0 || y > 0) {
				result += interGridDelim;
			}

			result += std::to_string ( _astralProjection_physicalState->_gridList_Piece[y][x]           ) + withinGridDelim;
			result += std::to_string ( _astralProjection_physicalState->_gridList_Card [y][x].cardId    ) + withinGridDelim;
			result += std::to_string ( _astralProjection_physicalState->_gridList_Card [y][x].direction );
		}
	}
	result +=interVariableDelim;


	for ( int i = 0; i <  _astralProjection_physicalState->_handCards_BLACK_Index; ++i) {
		result += std::to_string (_astralProjection_physicalState->_handCards_BLACK[i]) + interGridDelim;
	}
	result +=interVariableDelim;

	for ( int i = 0; i <  _astralProjection_physicalState->_handCards_WHITE_Index; ++i) {
		result += std::to_string (_astralProjection_physicalState->_handCards_WHITE[i]) + interGridDelim;
	}

	return result;
}

//=========================================================
// 検討モードでの盤面いじり判定用のデータ。セーブロード対応のための処理
//=========================================================
void GameManager::FromString_astralProjection_physicalState ( string fromSaveData ) {

	//printf ("いくよ！ GameManager::FromString_astralProjection_physicalState ( string fromSaveData[%s] )\n", fromSaveData.c_str());

	if ( _astralProjection_physicalState != NULL ) {
		delete _astralProjection_physicalState;
		_astralProjection_physicalState = NULL;
	}

	if ( fromSaveData == "" ) {

		//printf("なかったよ！\n");
		return;
	}

	_astralProjection_physicalState = new AI_Modules::InnerExpression();
	_astralProjection_physicalState->Init();

	string interVariableDelim = ",";
	string withinGridDelim = "*";
	string interGridDelim  = "<@>";


	vector<string> dataList = Utils_String::Split ( fromSaveData, interVariableDelim );
	if ( dataList.size() < 9) {
		printf ("セーブデータが壊れています：GameManager::FromString_astralProjection_physicalState ( %s )\n", fromSaveData.c_str());

		delete _astralProjection_physicalState;
		_astralProjection_physicalState = NULL;
		return;
	}

	_astralProjection_physicalState->_boardSizeX = std::stoi( dataList[0] );
	_astralProjection_physicalState->_boardSizeY = std::stoi( dataList[1] );
	_astralProjection_physicalState->_handPiece_BLACK = std::stoi( dataList[2] );
	_astralProjection_physicalState->_handPiece_WHITE = std::stoi( dataList[3] );
	_astralProjection_physicalState->_handCards_BLACK_Index = std::stoi( dataList[4] );
	_astralProjection_physicalState->_handCards_WHITE_Index = std::stoi( dataList[5] );

	string gridListData = dataList[6];
	string handCards_BLACK_ListData = dataList[7];
	string handCards_WHITE_ListData = dataList[8];

	vector<string> gridDataList = Utils_String::Split ( gridListData, interGridDelim );

	int gridIndex = 0;
	for ( int y = 0; y < _astralProjection_physicalState->_boardSizeY; ++y ) {
		for ( int x = 0; x < _astralProjection_physicalState->_boardSizeX; ++x ) {

			string gridData = gridDataList[gridIndex];

			vector<string> gridInfo = Utils_String::Split ( gridData, withinGridDelim );

			if ( gridInfo.size() < 3 ) {
				printf ("セーブデータが壊れています（グリッド情報）：GameManager::FromString_astralProjection_physicalState ( %s )\n", fromSaveData.c_str());
				continue;
			}

			_astralProjection_physicalState->_gridList_Piece[y][x]          = std::stoi ( gridInfo[0] );
			_astralProjection_physicalState->_gridList_Card[y][x].cardId    = std::stoi ( gridInfo[1] );
			_astralProjection_physicalState->_gridList_Card[y][x].direction = std::stoi ( gridInfo[2] );

			++gridIndex;
		}
	}

	vector<string> aaaaaa = Utils_String::Split ( handCards_BLACK_ListData, interGridDelim );
	for ( int i = 0; i < _astralProjection_physicalState->_handCards_BLACK_Index; ++i ) {

		_astralProjection_physicalState->_handCards_BLACK[i] = std::stoi ( aaaaaa[i] );
	}

	vector<string> bbbbbb = Utils_String::Split ( handCards_WHITE_ListData, interGridDelim );
	for ( int i = 0; i < _astralProjection_physicalState->_handCards_WHITE_Index; ++i ) {

		_astralProjection_physicalState->_handCards_WHITE[i] = std::stoi ( bbbbbb[i] );
	}

}

//=========================================================
// Participantのシリアライズ
//=========================================================
string GameManager::SerializeParticipant ( Participant participant ) {

	string result = "";

	string equalSign = "{=}";
	string delimitor = "<qp>";

	result +=             "participant.actorFileName"  + equalSign + participant.actorFileName;
	result += delimitor + "participant.aiFileName"     + equalSign + participant.aiFileName;
	result += delimitor + "participant.id"             + equalSign + participant.id;
	result += delimitor + "participant.name"           + equalSign + participant.name;
	result += delimitor + "participant.side"           + equalSign + participant.side;
	result += delimitor + "participant.turnFrameCount" + equalSign + to_string ( participant.turnFrameCount );

	if ( participant.actor != NULL ) {
		result += delimitor + "actorCurrentPosX" + equalSign + Utils_Convert::Int_to_String ( (int)(participant.actor->GetImagePositionCurrent().x) );
		result += delimitor + "actorCurrentPosY" + equalSign + Utils_Convert::Int_to_String ( (int)(participant.actor->GetImagePositionCurrent().y) );
		result += delimitor + "actorGoalPosX"    + equalSign + Utils_Convert::Int_to_String ( (int)(participant.actor->GetImagePositionGoal().x) );
		result += delimitor + "actorGoalPosY"    + equalSign + Utils_Convert::Int_to_String ( (int)(participant.actor->GetImagePositionGoal().y) );
	} else {
		// ないなら設定しない。復元時、キーがないならないで、よしなに。
	}
	
	//vector< pair<int,int> > extraChoicesRange;
	string extraChoicesRangeStr = "";
	for ( unsigned int i = 0; i < participant.choiceModeInfo.extraChoicesRange.size(); ++i) {
		
		if ( i > 0 ) {
			extraChoicesRangeStr += "@qp@";
		}
		pair<int,int> range = participant.choiceModeInfo.extraChoicesRange[i];
		
		extraChoicesRangeStr += Utils_Convert::Int_to_String ( range.first );
		extraChoicesRangeStr += "p@q";
		extraChoicesRangeStr += Utils_Convert::Int_to_String ( range.second );
	}
	
	result += delimitor + "participant.choiceModeInfo.isActive"                  + equalSign + Utils_Convert::Bool_to_string ( participant.choiceModeInfo.isActive );
	result += delimitor + "participant.choiceModeInfo.numberOfChoices"           + equalSign + to_string ( participant.choiceModeInfo.numberOfChoices );
	result += delimitor + "participant.choiceModeInfo.phase"                     + equalSign + participant.choiceModeInfo.phase;
	result += delimitor + "participant.choiceModeInfo.playersSelectChoiceNumber" + equalSign + participant.choiceModeInfo.playersSelectChoiceNumber;
	result += delimitor + "participant.choiceModeInfo.selectOptionsPos"          + equalSign + participant.choiceModeInfo.selectOptionsPos;
	result += delimitor + "participant.choiceModeInfo.extraChoicesRange"         + equalSign + extraChoicesRangeStr;
	//participant.choiceModeInfo.moveListSuggestedByAI; // 試合中にはセーブしない仕様なので、指し手情報は保存対象としない

	result += delimitor + "participant.interferenceModeInfo.isActive"           + equalSign + Utils_Convert::Bool_to_string ( participant.interferenceModeInfo.isActive );
	result += delimitor + "participant.interferenceModeInfo.currentTimes"       + equalSign + to_string ( participant.interferenceModeInfo.currentTimes );
	result += delimitor + "participant.interferenceModeInfo.isAgreedWithAIMove" + equalSign + Utils_Convert::Bool_to_string ( participant.interferenceModeInfo.isAgreedWithAIMove );
	result += delimitor + "participant.interferenceModeInfo.maxTimes"           + equalSign + to_string ( participant.interferenceModeInfo.maxTimes );
	result += delimitor + "participant.interferenceModeInfo.phase"              + equalSign + participant.interferenceModeInfo.phase;
	result += delimitor + "participant.interferenceModeInfo.selectOptionsPos"   + equalSign + participant.interferenceModeInfo.selectOptionsPos;
	//participant.interferenceModeInfo.moveSuggestedByAI; // 試合中にはセーブしない仕様なので、指し手情報は保存対象としない



	return result;
}
//=========================================================
// Participantのデシリアライズ
//=========================================================
GameManager::Participant GameManager::DeSerializeParticipant( string dataString ) {

	Participant participant;

	string equalSign = "{=}";
	string delimitor = "<qp>";

	map<string, string> data = Utils_Convert::ReadStringAsMap ( dataString, delimitor, equalSign );


	participant.actorFileName  =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.actorFileName"  , "" );
	participant.aiFileName     =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.aiFileName"     , "" );
	participant.id             =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.id"             , "" );
	participant.name           =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.name"           , "" );
	participant.side           =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.side"           , "" );
	participant.turnFrameCount = Utils_Convert::String_to_Int  ( UTILS_OTHERS::GetValueFromMap ( data, "participant.turnFrameCount" , "-1" ) );

	if ( participant.actorFileName != "" ) {
		participant.actorCurrentPosX = Utils_Convert::String_to_Int ( UTILS_OTHERS::GetValueFromMap ( data, "actorCurrentPosX", "0") );
		participant.actorCurrentPosY = Utils_Convert::String_to_Int ( UTILS_OTHERS::GetValueFromMap ( data, "actorCurrentPosY", "0") );
		participant.actorGoalPosX    = Utils_Convert::String_to_Int ( UTILS_OTHERS::GetValueFromMap ( data, "actorGoalPosX"   , "0") );
		participant.actorGoalPosY    = Utils_Convert::String_to_Int ( UTILS_OTHERS::GetValueFromMap ( data, "actorGoalPosY"   , "0") );
	}

	//vector< pair<int,int> > extraChoicesRange;
	participant.choiceModeInfo.extraChoicesRange.clear(); //一応
	string extraChoicesRangeStr = UTILS_OTHERS::GetValueFromMap ( data, "participant.choiceModeInfo.extraChoicesRange", "");
	vector<string> extraChoicesRangeStrList = Utils_String::Split ( extraChoicesRangeStr, "@qp@" );
	for ( unsigned int i = 0; i < extraChoicesRangeStrList.size(); ++i) {
		
		string rangeStr = extraChoicesRangeStrList[i];
		vector<string> rangeStrAna = Utils_String::Split ( rangeStr, "p@q" );
		
		pair<int,int> range;
		range.first  = ( rangeStrAna.size() >= 1 ) ? Utils_Convert::String_to_Int ( rangeStrAna[0] ) : 0;
		range.second = ( rangeStrAna.size() >= 2 ) ? Utils_Convert::String_to_Int ( rangeStrAna[1] ) : -1;
		
		participant.choiceModeInfo.extraChoicesRange.push_back ( range );
	}
	
		
	participant.choiceModeInfo.isActive                  = Utils_Convert::String_to_bool ( UTILS_OTHERS::GetValueFromMap ( data, "participant.choiceModeInfo.isActive"                 , "false" ) );
	participant.choiceModeInfo.numberOfChoices           = Utils_Convert::String_to_Int  ( UTILS_OTHERS::GetValueFromMap ( data, "participant.choiceModeInfo.numberOfChoices"          , "0" ) );
	participant.choiceModeInfo.phase                     =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.choiceModeInfo.phase"                    , participant.choiceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN );
	participant.choiceModeInfo.playersSelectChoiceNumber =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.choiceModeInfo.playersSelectChoiceNumber", "0" );
	participant.choiceModeInfo.selectOptionsPos          =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.choiceModeInfo.selectOptionsPos"         , "0,0" );
	//participant.choiceModeInfo.moveListSuggestedByAI; // 試合中にはセーブしない仕様なので、指し手情報は保存対象としない

	participant.interferenceModeInfo.isActive           = Utils_Convert::String_to_bool ( UTILS_OTHERS::GetValueFromMap ( data, "participant.interferenceModeInfo.isActive"          , "false" ) );
	participant.interferenceModeInfo.currentTimes       = Utils_Convert::String_to_Int  ( UTILS_OTHERS::GetValueFromMap ( data, "participant.interferenceModeInfo.currentTimes"      , "0" ) );
	participant.interferenceModeInfo.isAgreedWithAIMove = Utils_Convert::String_to_bool ( UTILS_OTHERS::GetValueFromMap ( data, "participant.interferenceModeInfo.isAgreedWithAIMove", "false" ) );
	participant.interferenceModeInfo.maxTimes           = Utils_Convert::String_to_Int  ( UTILS_OTHERS::GetValueFromMap ( data, "participant.interferenceModeInfo.maxTimes"          , "0" ) );
	participant.interferenceModeInfo.phase              =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.interferenceModeInfo.phase"             , participant.interferenceModeInfo.PHASE_WAIT_FOR_NEXT_MY_TURN );
	participant.interferenceModeInfo.selectOptionsPos   =                                 UTILS_OTHERS::GetValueFromMap ( data, "participant.interferenceModeInfo.selectOptionsPos"  , "0,0" );
	//participant.interferenceModeInfo.moveSuggestedByAI; // 試合中にはセーブしない仕様なので、指し手情報は保存対象としない


	return participant;
}












